/************************************************************************
 * FILE NAME   : GpDspIo.c                                              *
 *									*
 * DESCRIPTION : IO routines for communicating with the DSP and framer  *
 *                                                                      *
 *                                                                      *
 * (C) Copyright 2001 by GlobeSpan Inc				        *
 *                                                                      *
 * $Header: /usr/local/cvsroot/wanpipe_common/adsl/Epci/GpDspIo.c,v 1.1.1.1 2002/11/25 09:21:53 ncorbic Exp $ *
 *                                                                      *
 ************************************************************************/

#include "GpInclude.h"


/************************************************************************
* Desc  : The following routines are required by the GTI transceiver code. 
*	: They provide basic read and write services to the DSP core.
*
*                                                                       
************************************************************************/



/************************************************************************
* Func   : WriteStarB
* Desc   : Write a byte of data to the DSP at the specified
*        : Register.
* Input  : GP_BYTE starlet, GP_BYTE reg, GP_BYTE att_data
* Output : None
* Return : None
************************************************************************/
GP_VOID WriteStarB(struct EXECUTE_CMD * command,
                        GP_BYTE starlet,
                        GP_BYTE reg,
                        GP_BYTE att_data)
{
	PDSL_MODEM pDslModem;

	pDslModem = (PDSL_MODEM)command->CustXcvrInfo.pDslModem;	
	
	switch (starlet){
      	case GTI_STAR_A:
		GpWritePciRegB(	pDslModem, 
				(GP_WORD)(DSP_HOST | reg | command->dsp_B_cs), 
				att_data);
		break;

	case GTI_STAR_B:
        	break;

	case GTI_STAR_AB:
	case GTI_STAR_ALL:
		GpWritePciRegB(	pDslModem, 
				(GP_WORD)(DSP_HOST | reg | command->dsp_B_cs), 
				att_data);
		break;

	case GTI_GEMINI_DMA:
        	GpWritePciRegB(pDslModem, DSP_HOST | reg, att_data);
		break;

	default:
        	break;
   }
}

/************************************************************************
* Func   : ReadStarB
* Desc   : This routine reads a byte from the DSP. The
*        : specific register to use is specified and this is added
*        : to the base address to generate the absolute address to read from.
* Input  : GP_BYTE starlet, GP_BYTE reg
* Output : None
* Return : GP_BYTE
************************************************************************/
GP_BYTE ReadStarB(struct EXECUTE_CMD * command,
                              GP_BYTE starlet,
                              GP_BYTE reg)
{
	GP_BYTE  result=0;

	PDSL_MODEM pDslModem;

	pDslModem = (PDSL_MODEM)command->CustXcvrInfo.pDslModem;	

	switch (starlet){
	case GTI_STAR_A:

		if (pDslModem->DslLinkState == GP_TRAINED && (pDslModem->SNRStatusRead==2)){
			/* If in data mode and SNR status has been read at least
			// once since entering data mode, then check DSP address.
			// If it corresponds to SNR margin (i.e., SQ or bit
			// loading), then read data
			// from local array. Reading all SNR margin data directly 
			// from the DSP in one shot takes too long (10's of msec),
			// resulting in the ping delay problem.
			*/
			  
			switch (pDslModem->LocalDspAddH){
			case 0xd7:
				result = pDslModem->DspRamD7_E[pDslModem->LocalDspAddL];    /* Get DS bit loading extended byte */
				break;
			case 0xf4:
				if (pDslModem->LocalDspAddL >= 0xC0)
					result = pDslModem->DspRamF4_E[pDslModem->LocalDspAddL]; /* Get US bit loading extended byte */
				else 
					GpReadPciRegB(pDslModem, (GP_WORD)(DSP_HOST | reg | command->dsp_B_cs), &result);
				break;
			default:
				/* Not an address corresponding to SNR data so read directly */
				/* from DSP */
					
				GpReadPciRegB(pDslModem, (GP_WORD)(DSP_HOST | reg | command->dsp_B_cs), &result);
			}
			/*GpDbgPrint("Read Star Byte from Local, reg. = %d, data = %d\n", reg, result);*/
		}else{
			/* Not in data mode or SNR has not been read once */
			/* since entering data mode so read directly from DSP */
			GpReadPciRegB(pDslModem, (GP_WORD)(DSP_HOST | reg | command->dsp_B_cs), &result);
		}
		break;

	case GTI_STAR_B:
        	break;

	case GTI_GEMINI_DMA:
        	GpReadPciRegB(pDslModem, DSP_HOST | reg, &result);
		break;

	default:
        	break;
   	}
	return result;
}

/************************************************************************
* Func   : WriteStarBlock
* Desc   : This function writes a complete block of data to the DSP.
*        :
*        : All bytes are written to the same address.
*        :
* Input  : GP_BYTE starlet
*        : GP_BYTE reg
*        : GP_WORD length
*        : GP_BYTE * start
* Output : None.
* Return : None.
************************************************************************/
GP_VOID WriteStarBlock(struct EXECUTE_CMD * command,
                           GP_BYTE starlet,
                           GP_BYTE reg,
                           GP_WORD length,
                           GP_BYTE GTI_CONST_QUAL GTI_ADDRESS_QUAL * start)
{
GP_WORD index;
PDSL_MODEM pDslModem;

	pDslModem = (PDSL_MODEM)command->CustXcvrInfo.pDslModem;	

   switch (starlet)
   {
      case GTI_STAR_A:
         for(index = 0; index < length; index++)
         {
            GpWritePciRegB(pDslModem, (GP_WORD)(DSP_HOST | reg | command->dsp_B_cs), (GP_BYTE)start[index]);
         }
         break;

      case GTI_STAR_B:
         break;

      case GTI_STAR_AB:
      case GTI_STAR_ALL:
         for(index = 0; index < length; index++)
         {
            GpWritePciRegB(pDslModem, (GP_WORD)(DSP_HOST | reg | command->dsp_B_cs), (GP_BYTE)start[index]);
         }
         break;

      case GTI_GEMINI_DMA:
         for(index = 0; index < length; index++)
         {
            GpWritePciRegB(pDslModem, DSP_HOST | reg, (GP_BYTE)start[index]);
         }
         break;

      default:
         break;
   }
}


/***************************************************************/
/* Name:WriteDspAddr                                           */
/* Input:                                                      */
/* GTISysPtr:   Store XCVR system in Multi XCVR                */
/*              hardware platforms                             */
/* id:     Select which DSP     to access                      */
/*         Can be one or all                                   */
/* addrl:  Data to write to hard coded DSP     register        */
/*         Can be from 0x00->0xff                              */
/* addrh:  Data to write to hard coded DSP     register        */
/*         Can be from 0x00->0xff                              */
/***************************************************************/
GP_VOID WriteDspAdd(struct EXECUTE_CMD* command,
           GTI_BYTE starlet,
           GTI_BYTE addrl,
           GTI_BYTE addrh)
{
	PDSL_MODEM pDslModem;

	pDslModem = (PDSL_MODEM)command->CustXcvrInfo.pDslModem;	

   #if 0
		if (pDslModem->DslLinkState == GP_TRAINED)
		  DbgPrint("Write DSP Add.= %d  %d \n", addrh, addrl);
	#endif

   switch (starlet)
   {
     case GTI_DSP_AB:
     case GTI_DSP_ALL:
     case GTI_DSP_A:
		  #ifdef FAST_SNR_STATUS
			/* If in data mode and SNR status has been read at least
			// once since entering data mode, then check DSP address.
			// If it corresponds to bit loading, then don't actually
			// write the address becuase it's really being read from
			// a local array.
			// Reading all SNR margin data directly 
			// from the DSP in one shot takes too long (10's of msec),
			// resulting in the ping delay problem.
			*/
		  if (pDslModem->DslLinkState == GP_TRAINED && (pDslModem->SNRStatusRead==2))
		  {
		    if ((addrh==0xD7)|| ((addrh==0xF4) && (addrl>=0xC0)))
			{
				;/* In this case, do nothing. */
			}
			else
			{
		      GpWritePciRegB(pDslModem, (GP_WORD)(DSP_HOST | GTI_ADDL_DSP | command->dsp_B_cs), addrl);
              GpWritePciRegB(pDslModem, (GP_WORD)(DSP_HOST | GTI_ADDH_DSP | command->dsp_B_cs), addrh);
			}
		  }
		  else
		  {
			  GpWritePciRegB(pDslModem, (GP_WORD)(DSP_HOST | GTI_ADDL_DSP | command->dsp_B_cs), addrl);
              GpWritePciRegB(pDslModem, (GP_WORD)(DSP_HOST | GTI_ADDH_DSP | command->dsp_B_cs), addrh);
		  }
		
		  pDslModem->LocalDspAddL = addrl;	/* Store address for use by ReadStarWord and ReadStarByte*/
		  pDslModem->LocalDspAddH = addrh;

		  #else
		    GpWritePciRegB(pDslModem, (GP_WORD)(DSP_HOST | GTI_ADDL_DSP | command->dsp_B_cs), addrl);
            GpWritePciRegB(pDslModem, (GP_WORD)(DSP_HOST | GTI_ADDH_DSP | command->dsp_B_cs), addrh);
		  #endif /* FAST_SNR_STATUS */

         break;
   }
}

/************************************************************************
* Func   : WriteStarW
* Desc   : This routine writes a word out to the addressed STARLET. Unlike
*        : the 'WriteByte' routine, there is no specified register. Instead
*        : GTI want us to write the low byte to device + GTI_DATAL_STAR
*        : (which is register 6) and the high byte to device +
*        : GTI_DATAH_STAR (which is register 7)
* Input  : GP_BYTE starlet, GP_BYTE datal, GP_BYTE datah
* Output : None.
* Return : None.
************************************************************************/
GP_VOID WriteStarW(struct EXECUTE_CMD * command,
                        GP_BYTE starlet,
                        GP_BYTE datal,
                        GP_BYTE datah)
{
   PDSL_MODEM pDslModem;

   pDslModem = (PDSL_MODEM)command->CustXcvrInfo.pDslModem;	
  

   switch (starlet)
   {
      case GTI_STAR_A:
         GpWritePciRegB(pDslModem, (GP_WORD)(DSP_HOST | GTI_DATAL_DSP | command->dsp_B_cs), datal);
         GpWritePciRegB(pDslModem, (GP_WORD)(DSP_HOST | GTI_DATAH_DSP | command->dsp_B_cs), datah);
         break;

      case GTI_STAR_B:
         break;

      case GTI_STAR_AB:
      case GTI_STAR_ALL:
         GpWritePciRegB(pDslModem, (GP_WORD)(DSP_HOST | GTI_DATAL_DSP | command->dsp_B_cs), datal);
         GpWritePciRegB(pDslModem, (GP_WORD)(DSP_HOST | GTI_DATAH_DSP | command->dsp_B_cs), datah);
         break;

      case GTI_GEMINI_DMA:
         GpWritePciRegB(pDslModem, DSP_HOST | GTI_DATAL_DSP, datal);
         GpWritePciRegB(pDslModem, DSP_HOST | GTI_DATAH_DSP, datah);
         break;

      default:
            /*while(1);*/   /* ERROR - Loop Forever here.   */
            break;
   }
}

/************************************************************************
* Func   : ReadStarW
* Desc   : This is the corresponding Read Word routine. This routine reads
*        : a byte from two addresses (device + GTI_DATAL_STAR and device +
*        : GTI_DATAH_STAR) and then combines them into an integer.
* Input  : GP_BYTE starlet
* Output : None
* Return : GP_WORD
************************************************************************/
GP_WORD ReadStarW(struct EXECUTE_CMD * command, GP_BYTE starlet)
{
	GP_BYTE datal;
	GP_BYTE datah;
	GP_WORD  result;
	PDSL_MODEM pDslModem;

   pDslModem = (PDSL_MODEM)command->CustXcvrInfo.pDslModem;	

   switch (starlet)
   {
      case GTI_STAR_A:
		 #ifdef FAST_SNR_STATUS
		   if ((pDslModem->DslLinkState == GP_TRAINED) && (pDslModem->SNRStatusRead==2))
		   {
			/* If in data mode and SNR status has been read at least
			// once since entering data mode, then check DSP address.
			// If it corresponds to SNR margin, then read data
			// from local array. Reading all SNR margin data directly 
			// from the DSP in one shot takes too long (10's of msec),
			// resulting in the ping delay problem.
			*/
			  
			  switch (pDslModem->LocalDspAddH)
			  {
			    case 0xd6:	
					/* SQ data */
					datah = pDslModem->DspRamD6_H[pDslModem->LocalDspAddL];     
					datal = pDslModem->DspRamD6_L[pDslModem->LocalDspAddL];
					/* DbgPrint("Read Star Word from Local, DataH = %d, DataL = %d \n",datah, datal); */
					pDslModem->LocalDspAddL++;	/* Increment address low byte */
					break;
				case 0xd7:
					/* DS bit loading data */
					datah = pDslModem->DspRamD7_H[pDslModem->LocalDspAddL];     
					datal = pDslModem->DspRamD7_L[pDslModem->LocalDspAddL]; 
					/* DbgPrint("Read Star Word from Local, DataH = %d, DataL = %d \n",datah, datal); */
					pDslModem->LocalDspAddL++;
					break;
				case 0xf4:
					if (pDslModem->LocalDspAddL >= 0xC0){
						/* US bit loading data */
					    datah = pDslModem->DspRamF4_H[pDslModem->LocalDspAddL];  
					    datal = pDslModem->DspRamF4_L[pDslModem->LocalDspAddL];
						/*DbgPrint("Read Star Word from Local, DataH = %d, DataL = %d \n",datah, datal); */
						pDslModem->LocalDspAddL++;
					}
					else
					{
						/* Not an address corresponding to SNR data so read directly*/
						/* from DSP */
					    GpReadPciRegB(pDslModem, (GP_WORD)(DSP_HOST | GTI_DATAL_DSP | command->dsp_B_cs), &datal);
					    GpReadPciRegB(pDslModem, (GP_WORD)(DSP_HOST | GTI_DATAH_DSP | command->dsp_B_cs), &datah);
						/* DbgPrint("Read Star Word from DSP  , DataH = %d, DataL = %d \n",datah, datal); */
					}
					
					break;
				default:
					/* Not an address corresponding to SNR data so read directly */
					/* from DSP */
					GpReadPciRegB(pDslModem, (GP_WORD)(DSP_HOST | GTI_DATAL_DSP | command->dsp_B_cs), &datal);
					GpReadPciRegB(pDslModem, (GP_WORD)(DSP_HOST | GTI_DATAH_DSP | command->dsp_B_cs), &datah);
					/* DbgPrint("Read Star Word from DSP, DataH = %d, DataL = %d \n",datah, datal); */
			  } 
		   }
		   else 
		   {
			/*  Not in data mode or SNR has not been read once */
			/* since entering data mode so read directly from DSP */
			GpReadPciRegB(pDslModem, (GP_WORD)(DSP_HOST | GTI_DATAL_DSP | command->dsp_B_cs), &datal);
			GpReadPciRegB(pDslModem, (GP_WORD)(DSP_HOST | GTI_DATAH_DSP | command->dsp_B_cs), &datah);
			/*DbgPrint("Read Star Word from DSP, DataH = %d, DataL = %d \n",datah, datal);*/
		   }
        
		#else
			GpReadPciRegB(pDslModem, (GP_WORD)(DSP_HOST | GTI_DATAL_DSP | command->dsp_B_cs), &datal);
			GpReadPciRegB(pDslModem, (GP_WORD)(DSP_HOST | GTI_DATAH_DSP | command->dsp_B_cs), &datah);
			#if 0
			 if((pDslModem->LocalDspAddH==0xD6)|(pDslModem->LocalDspAddH==0xD7)||(pDslModem->LocalDspAddH==0xF4))
			 {
				DbgPrint("Read Word from DSP, AddH = %d, AddL = %d\n",pDslModem->LocalDspAddH, pDslModem->LocalDspAddL);
				DbgPrint("Read Word ,                               DataH = %d, DataL = %d \n", datah, datal);
				pDslModem->LocalDspAddL++;
			 }
			#endif
		#endif  /* End FAST_SNR_STATUS */
		 break;

      case GTI_STAR_B:
         break;

      case GTI_GEMINI_DMA:
         GpReadPciRegB(pDslModem, DSP_HOST | GTI_DATAL_DSP, &datal);
         GpReadPciRegB(pDslModem, DSP_HOST | GTI_DATAH_DSP, &datah);
         break;

      default:
         /*while(1);*/   /* ERROR - Loop Forever here.   */
         break;
   }

   result = ((datah << 8) | datal) & 0xffff;
   return result;
}

/************************************************************************
* Func   : WaitXms
* Desc   : This routine is for providing a specified delay in milliseconds.
*        : According to GTI, it is not used and simply return.
* Input  : GP_WORD count
* Output : None
* Return : None
************************************************************************/
GP_VOID WaitXms(GP_WORD count)
{
   count = 0;
}

/************************************************************************
* Func   : ResetStar
* Desc   : This routine resets the addressed STARLET device by toggling
*        : the appropriate 8051 Port pins.
*        : Although the GTI code can selectively request a device to be
*        : reset, we go ahead and reset all devices at the same time.
* Input  : GP_BYTE starlet
* Output : None.
* Return : None.
************************************************************************/
GP_VOID ResetStar(struct EXECUTE_CMD * command,
                        GP_BYTE starlet)
{
	PDSL_MODEM pDslModem;

	pDslModem = (PDSL_MODEM)command->CustXcvrInfo.pDslModem;	

    	GpWritePciRegB(pDslModem, PCORE_ADDR_GLUERESETIMAGE, GP_NOVA_RESET_MASK);
    	GpDelay(50);
}

/************************************************************************
* Func   : ResetDSP - Star
* Desc   : This routine resets the addressed STARLET.
*        : Although the GTI code can selectively request a device to be
*        : reset, we go ahead and reset all devices at the same time.
* Input  : GP_BYTE starlet
* Output : None.
* Return : None.
************************************************************************/
GP_VOID ResetDSP(struct EXECUTE_CMD * command,
                        GP_BYTE starlet)
{
	PDSL_MODEM pDslModem;

	pDslModem = (PDSL_MODEM)command->CustXcvrInfo.pDslModem;	

    GpWritePciRegB(pDslModem, PCORE_ADDR_GLUERESETIMAGE, GP_NOVA_RESET_MASK);
    GpDelay(50);
}
/* Stubs for GTI compatibility */
/************************************************************************
* Func   : WriteRAMHBrook
* Desc   : 
* Input  : 
* Output : None.
* Return : None.
************************************************************************/
GP_VOID WriteRAMHBrook(struct EXECUTE_CMD *x,GP_BYTE y,GP_BYTE z)
{
   return;
}

/************************************************************************
* Func   : WriteRAMLBrook
* Desc   : 
* Input  : 
* Output : None.
* Return : None.
************************************************************************/
GP_VOID WriteRAMLBrook(struct EXECUTE_CMD *x,GP_BYTE y,GP_BYTE z)
{
   return;
}

/************************************************************************
* Func   : WriteADSLFramerW
* Desc   : 
* Input  : 
* Output : None.
* Return : None.
************************************************************************/
GP_VOID WriteADSLFramerW(struct EXECUTE_CMD *x,GP_BYTE y,GP_WORD z)
{
   return;
}

/************************************************************************
* Func   : WriteHDSLFramerB
* Desc   : 
* Input  : 
* Output : None.
* Return : None.
************************************************************************/
GP_VOID WriteHDSLFramerB(struct EXECUTE_CMD *x,GP_BYTE y,GP_BYTE z)
{
   return;
}

