/************************************************************************
 * FILE NAME   : GpEeprom.c                                         	*
 *								    	*
 * DESCRIPTION : Routines used for accessing the EEPROM.		*
 *		 Three GPIO leads are used to to control		*
 *		 the clock, data, and chip select signals.           	*
 * (C) Copyright 2001 by GlobeSpan Inc				        *
 *                                                                      *
 * $Header: /usr/local/cvsroot/wanpipe_common/adsl/Epci/GpEeprom.c,v 1.1.1.1 2002/11/25 09:21:53 ncorbic Exp $ *
 *                                                                      *
 ************************************************************************/

#include "GpInclude.h"
/*#include "shell.h"*/


#pragma hdrstop
#pragma warning "disable: 4244 4514"


/*-----------------------------------------------------------------------------
// Procedure:   GpEepromRead
//
// Description: This routine serially reads one word out of the EEPROM.
//              The EEPROM control registers are accessed via virtual
//              memory address, so this function can ONLY be used after the
//              CSRs physical address has been vitually mapped.
//
// Arguments:
//      pDslModem - pointer to DslModem structure
//      Reg - address of EEPROM word to read.
//
// Returns:
//      Contents of EEPROM word (Reg).
//-----------------------------------------------------------------------------
*/

GP_DWORD
GpEepromRead(
              PDSL_MODEM pDslModem,
              GP_WORD Reg
              )
{
    GP_DWORD  data;


	#if GP_DBG
      GpDbgPrint( "====> ReadEEprom" );
	#endif

    /* Put the EEPROM into a known state*/
    GpEepromEnable(pDslModem);

    /* Write the read opcode and register number in that order*/
    GpEepromShiftOut(pDslModem, ((GP_EEPROM_READ_OPCODE << 8) | (Reg & 0xff)), 16);

    GpDelay(250);

    /* Now read the data (16 bits) in from the selected EEPROM word*/
    data = GpEepromShiftIn(pDslModem);

    /* Reset the EEPROM interfaces to something that will not cause problems*/
    GpEepromDisable(pDslModem);

#if GP_DBG
    DbgPrint( "<==== ReadEEprom [%2x]\n", data );
#endif

    return data;
}

/*-----------------------------------------------------------------------------
// Procedure:   GpEepromWrite
//
// Description: This routine serially write one word to the EEPROM.
//
// Arguments:
//      pDslModem - pointer to DslModem structure
//      Reg - EEPROM word to read
//      Data - EEPROM data to write
//
// Returns:
//      None
//-----------------------------------------------------------------------------
*/
GP_VOID
GpEepromWrite(
              PDSL_MODEM pDslModem,
              GP_WORD  Reg,
              GP_DWORD  data
              )
{
#if GP_DBG
    DbgPrint( "====> WriteEEprom (Reg=%2x data=%2x)\n", Reg, data );
#endif

    /* Put the EEProm into a known state */
    GpEepromEnable( pDslModem);
    
    /* Write the WREN opcode */
    GpEepromShiftOut( pDslModem, GP_EEPROM_WREN_OPCODE, 8 );
    GpDelay(500);

    /* Return EEPROM to NORMAL */
    GpWritePciReg(pDslModem, PCORE_ADDR_GPIOOUTPUT2, GPIO_EEPROM_OUTPUT_NORMAL );
	GpDelay(500);

    /* Sent the write opcode with data */
    GpEepromShiftOut( pDslModem, ((GP_EEPROM_WRITE_OPCODE << 16) | (Reg << 8) | (data & 0xff)), 24 );

    /* Reset the EEPROM interfaces to something that will not cause problems */
    GpEepromDisable( pDslModem);
	GpDelay(10000);		/* Need to delay 10 msec. between consecutive writes */

	#if GP_DBG
    DbgPrint( "<==== WriteEEprom\n" );
	#endif
}

/*-----------------------------------------------------------------------------
// Procedure:   GpEepromShiftOut
//
// Description: This routine shifts data bits out to the EEPROM.
//
// Arguments:
//      pDslModem - pointer to DslModem structure
//      data - data to send to the EEPROM.
//      count - number of data bits to shift out.
//
// Returns: (none)
//-----------------------------------------------------------------------------
*/
GP_VOID
GpEepromShiftOut(
                PDSL_MODEM pDslModem,
                GP_DWORD data,
                GP_WORD count
                )
{
    GP_DWORD  mask;
    GP_DWORD  data_to_write;
    GP_WORD  i;

    mask = 0x01 << (count - 1);
    data_to_write = GPIO_EEPROM_OUTPUT_NORMAL;

    GpDelay(250);

    /* Start with NORMAL - something we know about */
    GpWritePciReg(pDslModem, PCORE_ADDR_GPIOOUTPUT2, data_to_write );

    /* Chip Select is Active low - Lower Chip Select */
    data_to_write &= ~GP_EECS;
    GpWritePciReg(pDslModem, PCORE_ADDR_GPIOOUTPUT2, data_to_write );
    for( i = 0; i < count; i++ )
    {
        /* Lower the clock */
        GpEepromLowerClock( pDslModem, &data_to_write );

        /* Get the next bit */
        data_to_write &= ~GP_EEDO;
        if( data & mask )
        {
            data_to_write |= GP_EEDO;
        }
        mask = mask >> 1;

        /* Write the data bit out */
        GpWritePciReg(pDslModem, PCORE_ADDR_GPIOOUTPUT2, data_to_write );

        /* Raise the clock */
        GpEepromRaiseClock( pDslModem, &data_to_write );
    }
}


/*-----------------------------------------------------------------------------
// Procedure:   GpEepromRaiseClock
//
// Description: This routine raises the EEPOM's clock input (EESK)
//
// Arguments:
//      pDslModem - pointer to DslModem structure
//      x - Ptr to the EEPROM control register's current value
//
// Returns: (none)
//-----------------------------------------------------------------------------
*/
GP_VOID
GpEepromRaiseClock(
              PDSL_MODEM pDslModem,
              GP_DWORD *x
              )

{
    *x = *x | GP_EESK;
    GpWritePciReg(pDslModem, PCORE_ADDR_GPIOOUTPUT2, *x );
    GpDelay(250);
}


/*-----------------------------------------------------------------------------
// Procedure:   GpEepromLowerClock
//
// Description: This routine lower's the EEPROM's clock input (EESK)
//
// Arguments:
//      pDslModem - pointer to DslModem structure
//      x - Ptr to the EEPROM control register's current value
//
// Returns: (none)
//-----------------------------------------------------------------------------
*/
GP_VOID
GpEepromLowerClock(
              PDSL_MODEM pDslModem,
              GP_DWORD *x
              )

{
    *x = *x & ~GP_EESK;
    GpWritePciReg(pDslModem, PCORE_ADDR_GPIOOUTPUT2, *x );
    GpDelay( 250);
}


/*-----------------------------------------------------------------------------
// Procedure:   GpEepromShiftIn
//
// Description: This routine shifts data bits in from the EEPROM.
//
// Arguments:
//      pDslModem - pointer to DslModem structure
//
// Returns:
//      The contents of that particular EEPROM word
//-----------------------------------------------------------------------------
*/
GP_DWORD
GpEepromShiftIn(
               PDSL_MODEM pDslModem
               )

{
    GP_DWORD		data;
    GP_DWORD		data_to_read;
    GP_DWORD		x;
    GP_WORD			i;
	GP_BYTE		dummy;
    
    data = 0;
    x = (GPIO_EEPROM_OUTPUT_NORMAL & ~GP_EECS);

    for( i = 0; i < 8; i++ )
    {
        data = data << 1;
        GpEepromLowerClock( pDslModem, &x);
        GpDelay(250);
        GpEepromRaiseClock( pDslModem, &x);
        GpDelay(250);

        data_to_read = (GP_WORD)GpReadPciReg( pDslModem, PCORE_ADDR_GPIOINPUTSTAT2, &dummy );
        if( data_to_read & GP_EEDI )
        {
            data |= 1;
        }
    }

    return data;
}

/*-----------------------------------------------------------------------------
// Procedure:   GpEepromDisable
//
// Description: This routine returns the EEPROM to an idle state
//
// Arguments:
//				pDslModem - pointer to DslModem structure
//
// Returns: (none)
//-----------------------------------------------------------------------------
*/
GP_VOID
GpEepromDisable(
                 PDSL_MODEM pDslModem
                 )

{
    /* Initialize output data register to something
    // that will not cause any problems
    */
    GpWritePciReg(pDslModem, PCORE_ADDR_GPIOOUTPUT2, GPIO_EEPROM_OUTPUT_NORMAL );
    GpDelay(250);
}

/*-----------------------------------------------------------------------------
// Procedure:   GpEepromEnable
//
// Description: This routine setups the EEPROM for any operations
//
// Arguments:
//				pDslModem - pointer to DslModem structure
//
// Returns: (none)
//-----------------------------------------------------------------------------
*/
GP_VOID
GpEepromEnable(
                 PDSL_MODEM pDslModem
                 )
{
    /* Initialize output data register to something
    // that will not cause any problems
    */
    GpWritePciReg(pDslModem, PCORE_ADDR_GPIOOUTPUT2, GPIO_EEPROM_OUTPUT_NORMAL );

    /* Set output configuration bits
    // 'Always enable' except EEDI (input)
    */
    GpWritePciReg(pDslModem, PCORE_ADDR_GPIOOUTPUTCFG2, GPIO_EEPROM_SELECT_CFG );

    /* Select output bits (all reverted to inputs) */
    GpWritePciReg(pDslModem,  PCORE_ADDR_GPIOOUTPUTSEL2, 0 );

    /* EECS HIGH to LOW transition is required */
    GpWritePciReg(pDslModem, PCORE_ADDR_GPIOOUTPUT2, GPIO_EEPROM_OUTPUT_CS_H_TO_L );
    GpDelay(250);
    GpWritePciReg(pDslModem, PCORE_ADDR_GPIOOUTPUT2, GPIO_EEPROM_OUTPUT_NORMAL );
    GpDelay(250);
}


