/************************************************************************
 *
 * FILE NAME   : GpReceiveDmaHandler
 *
 * DESCRIPTION : This function manages the receive DMA buffer
 *
 *
 * Return:       GP_TRUE - new cells were received and no errors were detected.
 * (C) Copyright 2001 by GlobeSpan Inc.
 *
 * $Header: /usr/local/cvsroot/wanpipe_common/adsl/Epci/GpReceiveDmaHandler.c,v 1.1.1.1 2002/11/25 09:21:53 ncorbic Exp $ *
 *                                                                      *
 ************************************************************************/

#include "GpInclude.h"
#include "Gp.h"

GP_VOID GpReceiveDmaHandler(PDSL_MODEM pDslModem)
{
	
	/* Receive buffer read pointer. Points to the
	// next cell to be read from the buffer.
	*/
   	GP_DWORD	RxCellRdPtr;				
	
	
	/* Receive buffer write pointer. Points to the
	// last cell written to the buffer by the DMA
	// engine. This value is read from the PCI device,
	*/
   	GP_DWORD	RxCellWrPtr;
	
	GP_BYTE        	dummy;
   	GP_WORD		NumRxBytes;  /* Number of bytes received this time */
   	GP_WORD		NumRxCells;  /* Number of cells received this time */

   	/* If we are in the Shutdown or Halt path, and a RX Int Occurs
   	// do not process RX Int
	*/
   	if(pDslModem->InitStatus & GP_INIT_SHUTDOWN )
		return;

   	/* Read the current receive DMA buffer pointer from the 
	// hardware and convert the physical
   	// address to a virtual address.
   	// PCORE_ADDR_RXCURRENTADD uses bits[2:31] out of [0:31], 
	// therefore we mask out bits[0:1]
	*/

   	RxCellWrPtr = (GpReadPciReg(pDslModem, PCORE_ADDR_RXCURRENTADD, &dummy )&0xFFFFFFFC) -
      		       pDslModem->RxBuffStartP + pDslModem->RxBuffStartV;

	
	/* Check for DMA buffer pointer exactly at the end of buffer.
	// If at the end, move to the begining.  This is a bug fix for LSI.
	*/
	if (RxCellWrPtr == pDslModem->RxBuffEndV )
		RxCellWrPtr = pDslModem->RxBuffStartV;

   	/* Get the last location address we processed, i.e., 
	// the previous pointer value.
	*/
   	RxCellRdPtr = pDslModem->RxCellRdPtrSav;

	
	/* Has pointer moved since last time, i.e.,
	// is there any new data? If not, just return.
	//
	// This should not happen unless this handler
	// is being called from a periodic timer
	// function (i.e., polling) rather than the
	// receive  cell interrupt.
	// Note, this check assumes that we can't get an
	// entire buffer of data in one interrupt.
	// The buffer size must be chosen to ensure that
	// this cannot happen.
	*/
	
   	if (RxCellRdPtr == RxCellWrPtr){			
   		return;			
	}
												
   	/* Check for buffer misalignment. If misalignment is found 
	// re-initialize the DMA engine.
   	// Two types of misalignment are theoretically possible:
   	//		1) pointer is out of bounds
   	//		2) the no. of bytes received does not 
	//		   correspond to an integer no. of ATM cells.
   	// However, neither of these conditions should ever occur.
	*/

	/* Check for DMA buffer pointer out of range */
	if ((RxCellWrPtr < pDslModem->RxBuffStartV) || 
	    (RxCellWrPtr > pDslModem->RxBuffEndV)){
         
#if GP_DBG
        	GpDbgPrint ("\n****Rcv DMA ERROR - Buffer Pointer out of bounds, reset DMA*****\n" );
           	GpDumpRxBuffInfo (pDslModem);
#endif

		pDslModem->RxDmaErrCnt++;
         	GpResetDma(pDslModem);
		return;
      	}

	/* Check if the write pointer has wrapped around past the start
	// of the buffer since the last interrupt.
	// 
	// Note: if pointer just barely wrapped,
	// i.e., it's = to start of buffer, the
	// rcv. cells are still contiguous, so
	// treat similar to non-wrap case.
	// 
	// No - pointer has not wrapped past start.
	// Determine the number of received cells
	// for the non-wrap case
	*/
      	if((RxCellWrPtr > RxCellRdPtr) || (RxCellWrPtr == pDslModem->RxBuffStartV)){
      											
		
		if(RxCellWrPtr == pDslModem->RxBuffStartV){
			NumRxBytes = pDslModem->RxBuffEndV - RxCellRdPtr;
		}else{
            		NumRxBytes = RxCellWrPtr - RxCellRdPtr;
		}
		
		/* Does the number of received bytes correspond to an
		// integer number of cells?
		*/
		if (NumRxBytes%GP_ATM_CELL_SIZE){
			
			/* No - this is an error condition - reset DMA */
			GpResetDma(pDslModem);	
			pDslModem->RxDmaErrCnt++;
			
#if GP_DBG
              		GpDbgPrint ("\n**Rcv DMA ERROR - Cell Misalignment (non-wrap case), reset DMA**\n" );
              		GpDumpRcvBuffInfo (pDslModem);
			GpDbgPrint(" Number of received bytes = %d \n",RxCellWrPtr-RxCellRdPtr);
#endif

			return;
		 
		}else{
			
			/* Yes - therefore, pointer is aligned properly
			// Call GpReceiveCell call back function to process
			// the cells.
			*/

			NumRxCells = NumRxBytes/GP_ATM_CELL_SIZE;
		
#if GP_DBG
              		GpDbgPrintf("No Wrap: number of Rx cells = %d \n",NumRxCells);
#endif
			GpReceiveCells(pDslModem, RxCellRdPtr,NumRxCells);

			pDslModem->TotalNumRxCells = pDslModem->TotalNumRxCells + NumRxCells;
			
			/* Update the read pointer and save for next
			// time.
			*/
			pDslModem->RxCellRdPtrSav = RxCellWrPtr;
		}
		
	}else{	
		/* Yes pointer has wrapped past the start of the buffer.
	  	// GpReceiveCells will be called twice, once for cells
	  	// at the end of the buffer, and one for cells at the
		// start of the buffer.
		*/

		/* First process the received cells
		// at the end of the buffer
		// Determine the number of received cells
		// at the end of the buffer
		*/
		NumRxBytes = pDslModem->RxBuffEndV - RxCellRdPtr;

		/* Determine the number of received cells
		// at the start of the buffer
		// Does the number of received bytes correspond to an
		// integer number of cells?	
		*/
		if (NumRxBytes%GP_ATM_CELL_SIZE){			
			
			/* No - error condition - reset DMA */
			GpResetDma(pDslModem);				
            		pDslModem->RxDmaErrCnt++;
			
#if GP_DBG
              		GpDbgPrint ("\n**Rcv DMA ERROR - Cell Misalignment (wrap case), reset DMA**\n" );
              		GpDumpRcvBuffInfo (pDslModem);
#endif
			return;
		
		}else{
			/* Yes - therefore, buffer is aligned properly
			// Call GpReceiveCells call back function to process
			// the cells.	
			*/
			NumRxCells = NumRxBytes/GP_ATM_CELL_SIZE;

#if GP_DBG
              		GpDbgPrintf("Wrap 1: number of Rx cells = %d \n",NumRxCells);
#endif

			GpReceiveCells(pDslModem, RxCellRdPtr,NumRxCells);

			/* Update running count of receive cells */
			pDslModem->TotalNumRxCells = pDslModem->TotalNumRxCells + NumRxCells;
		}

		/* Next process the received cells
		// at the start of the buffer
		// Determine the number of received cells
		// at the start of the buffer
		*/

		NumRxBytes = RxCellWrPtr - pDslModem->RxBuffStartV;

		/* Does the number of received bytes correspond to an  
		// integer number of cells?
		*/

		if (NumRxBytes%GP_ATM_CELL_SIZE){			

			GpResetDma(pDslModem);
            		pDslModem->RxDmaErrCnt++;
			
#if GP_DBG
              		GpDbgPrint ("\n**Rcv DMA ERROR - Cell Misalignment (wrap case), reset DMA**\n" );
              		GpDumpRcvBuffInfo (pDslModem);
#endif
		}else{
			/* Yes - therefore, buffer is aligned properly
			// Call GpReceiveCell call back function to process
			// the cells.
			*/
		
			NumRxCells = NumRxBytes/GP_ATM_CELL_SIZE;
			
#if GP_DBG
             		GpDbgPrintf("Wrap 2: number of Rx cells = %d \n",NumRxCells);
#endif

			GpReceiveCells(pDslModem, pDslModem->RxBuffStartV, NumRxCells);

			/* Update running count of receive cells */
			pDslModem->TotalNumRxCells = pDslModem->TotalNumRxCells + NumRxCells;
	
			/* Update the read pointer and save for next
			// time.
			*/
			pDslModem->RxCellRdPtrSav = RxCellWrPtr;
		}
      	}
}/* End of ReceiveDmaHandler */

/************************************************************************************/

GP_DWORD GpFindIdleCells(GP_CELL *thisCell)
{
	GP_DWORD idleCell = 0;
	GP_DWORD clp = 0;

	idleCell = (thisCell->hdr[0]&0x0f)+(thisCell->hdr[1]&0xff)+(thisCell->hdr[2]&0xff)
		+(thisCell->hdr[3]&0xf0);

	clp = thisCell->hdr[3] & 0x01;

	return ((!idleCell)&clp);
}

/********************************************************************************/
/* This routine resets and re-initializes the DMA engine in Panda		*/
GP_VOID GpResetDma(PDSL_MODEM pDslModem)
{
	GpWritePciReg(pDslModem, PCORE_ADDR_GLUERESETIMAGE,1);
	GpWritePciReg(pDslModem, PCORE_ADDR_RXSTARTADD, pDslModem->RxBuffStartP );
	GpWritePciReg(pDslModem, PCORE_ADDR_RXENDADD, pDslModem->RxBuffEndP );
	GpWritePciReg(pDslModem, PCORE_ADDR_TXCOMMAND, PCORE_INITIAL_TXCOMMAND);
	GpWritePciReg(pDslModem, PCORE_ADDR_RXCFG1, (PCORE_INITIAL_RXCFG1&RX_CELL_CNT_MASK)|
		   (pDslModem->RxCellCount<<3));
	GpWritePciReg(pDslModem, PCORE_ADDR_RXCFG2, PCORE_INITIAL_RXCFG2);

	pDslModem->RxCellRdPtrSav = pDslModem->RxBuffStartV;
	/* Start RX DMA Engine */
	GpWritePciReg(pDslModem,PCORE_ADDR_RXCOMMAND, PCORE_INITIAL_RXCOMMAND | 0x00000001 );
}

/************************************************************************************/

#ifdef GP_DBG
GpDumpRcvBuffInfo (PDSL_MODEM pDslModem, GP_DWORD RxCellWrPtr, GP_DWORD RxCellRdPtr)
{
	GpDbgPrint("virt address: RxBuffStartV=0x%x\n", pDslModem->RxBuffStartV );
     	GpDbgPrint("virt address: RxBuffEndV=0x%x\n", pDslModem->RxBuffEndV );
      	GpDbgPrint("virt address: RxCellWrPtr=0x%x\n", RxCellWrPtr );
      	GpDbgPrint("virt address: RxCellRdPtr=0x%x\n", RxCellRdPtr );
      	GpDbgPrint("phys address: RxBuffStartP=0x%x\n", pDslModem->RxBuffStartP );
	GpDbgPrint("Receive DMA Error Count=%d ",pDslModem->RxDmaErrCnt);
}
#endif
