/*+M*************************************************************************
 * GlobeSpan C Library Routines
 *
 *-M*************************************************************************/
/****************************************************************************
@doc INTERNAL GpUtil GpUtil_c

@module GpDebug.c |

    This module implements some of the basic C library functions.  This makes
    porting to some platforms easier, where the C run-time libraries are not
    available.

@head3 Contents |
@index class,mfunc,func,msg,mdata,struct,enum | GpUtil_c

@end
 ****************************************************************************/
#if 0
#include <sys/types.h>
#endif
#include "drvh/GpUtil.h"

#if defined(DEBUG) || defined(_DEBUG) || !defined(NO_TPI_UTILS)

void * Gp_memcpy(void *dst, const void *src, size_t n);
/*
// Copy non-overlapping data from a source to a destination.
*/
void * Gp_memcpy(void *dst, const void *src, size_t n)
{
    unsigned char *tmp = (unsigned char *) dst;
    const unsigned char *s = (const unsigned char *) src;

    while (n--)
    {
        *(tmp++) = *(s++);
    }
    return dst;
}

/*
// Copy overlapping data from a source to a destination.
*/
void * tpi_memmove(void *dst, const void *src, size_t n)
{
    unsigned char *t;
    const unsigned char *s;

    if (dst <= src)
    {
        t = (unsigned char *) dst;
        s = (const unsigned char *) src;
        while (n--)
        {
            *(t++) = *(s++);
        }
    }
    else
    {
        t = (unsigned char *) dst + n;
        s = (const unsigned char *) src + n;
        while (n--)
        {
            *(--t) = *(--s);
        }
    }
    return dst;
}

/*
// Search for the first occurance of val in the pointed to data.
*/
void * tpi_memchr(const void *ptr, int val, size_t n)
{
    const unsigned char *p = (const unsigned char *) ptr;

    while (n--)
    {
        if (*p == val)
        {
            return (void *) p;
        }
        p++;
    }
    return (void *) 0;
}

/*
// Compare the first n bytes referenced by the specified pointers.
// Return 0 if the bytes match.
*/
int tpi_memcmp(const void *ptr1, const void *ptr2, size_t n)
{
    const unsigned char *su1, *su2;
    int res = 0;

    for (su1 = ptr1, su2 = ptr2; 0 < n; ++su1, ++su2, n--)
    {
        if ((res = (*su1 - *su2)) != 0)
        {
            break;
        }
    }
    return res;
}

/*
// Set all the data values to a give value.
*/
void * tpi_memset(void *dst, int val, size_t n)
{
    unsigned char *p = (unsigned char *) dst;

    while (n--)
    {
        *(p++) = val;
    }
    return dst;
}

/*
// Return the minimum of the specified values.
*/
unsigned long tpi_min(unsigned long x, unsigned long y)
{
    return (x <= y) ? x : y;
}

/*
// Return the maximum of the specified values.
*/
unsigned long tpi_max(unsigned long x, unsigned long y)
{
    return (x >= y) ? x : y;
}

/*
// Byte swap the 16bit value.
*/
unsigned short tpi_swaps(unsigned short x)
{
    return (unsigned short) (((x & 0x0FF) << 8) | ((x >> 8) & 0x0FF));
}

/*
// Byte swap the 32bit value.
*/
unsigned long tpi_swapl(unsigned long x)
{
    return ((tpi_swaps((unsigned short) (x & 0x0FFFF)) << 16) |
             tpi_swaps((unsigned short) (x >> 16)));
}

/*
// Return upper case version of the specified character.
*/
char tpi_toupper(char cp)
{
    return tpi_islower(cp) ? (char) (cp - ('a' - 'A')) : cp;
}

/*
// Return lower case version of the specified character.
*/
char tpi_tolower(char cp)
{
    return tpi_isupper(cp) ? (char) (cp - ('A' - 'a')) : cp;
}

/*
// Check if the digit is correct for the given base. Return non-zero if it
// is correct and return its value, otherwise, return zero.
*/
unsigned char tpi_basedigit(char digit, unsigned char base, unsigned char *pValue)
{
    unsigned char      result = 0;
    unsigned char      val = 0;

    if ((digit >= '0') && (digit <= '9'))
    {
        if (digit < '0'+base)
        {
            result = !0;
            val = digit-'0';
        }
    }
    else if ((tpi_toupper(digit) >= 'A') && (tpi_toupper(digit) <= 'Z'))
    {
        if (tpi_toupper(digit) < 'A'-10+base)
        {
            result = !0;
            val = tpi_toupper(digit)-'A'+10;
        }
    }

    /* Return the value */
    if (pValue != (void *) 0)
    {
        *pValue = val;
    }
    return result;
}

/*
// Return the numeric value of the string of characters, in the base specified.
*/
unsigned long tpi_strtoul(const char *cp, char **endp, unsigned char base)
{
    unsigned long       result = 0;
    int                 sign = 1;
    unsigned char       value;

    /* Skip leading white space */
    while (tpi_isspace(*cp)) cp++;

    /* Handle an optional leading negative */
    if (*cp == '-')
    {
        sign = -1;
        cp++;
    }

    /* Handle base conversion if not specified */
    if (!base)
    {
        base = 10;
        if (*cp == '0')
        {
            base = 8;
            cp++;

            if (((*cp == 'x') || (*cp == 'X')) && tpi_isxdigit(cp[1]))
            {
                cp++;
                base = 16;
            }
        }
        else
        {
            /* Handle Microsoft way of specifying values */
            if ((*cp == 'x') && tpi_basedigit(cp[1], 16, (void *) 0))
            {
                cp++;
                base = 16;
            }
            else if ((*cp == 'o') && tpi_basedigit(cp[1], 8, (void *) 0))
            {
                cp++;
                base = 8;
            }
            else if ((*cp == 'b') && tpi_basedigit(cp[1], 2, (void *) 0))
            {
                cp++;
                base = 2;
            }
        }
    }

    /* Collect all the digits */
    while (tpi_basedigit(*cp, base, &value))
    {
        result = (result * base) + value;
        cp++;
    }

    /* Apply the sign to the result */
    result *= sign;

    /* Mark where we stopped processing */
    if (endp)
    {
        *endp = (char *) cp;
    }
    return result;
}

/*
// Concatenate a string onto the end of another string. It is assumed that
// both strings are NULL terminated. The concatenated string is returned.
*/
char * tpi_strcat(char *dst, const char *src)
{
    char *tmp = dst;

    while (*dst)
    {
        dst++;
    }
    while ((*dst++ = *src++) != '\0')
        /* nothing */;

    return tmp;
}

/*
// Concatenate a string onto the end of another string. It is assumed that
// both strings are NULL terminated. The concatenated string is returned.
// In the event that the destination string would be longer than n bytes,
// the destination string will be null terminated and only contain the first
// n-1 characters of the concatenated string.
*/
char * tpi_strncat(char *dst, const char *src, size_t n)
{
    char *tmp = dst;

    while (*dst)
    {
        --n;
        dst++;
    }
    if (n)  /* zero length output buffer! */
    {
        while (--n && (*dst++ = *src++) != '\0')
            /* nothing */;
        *dst = 0;
    }
    return tmp;
}

/*
// Copy a string into the specified destination. It is assumed that the
// source string is NULL terminated. The destination is returned.
*/
char * tpi_strcpy(char *dst, const char *src)
{
    char *tmp = dst;

    while ((*dst++ = *src++) != '\0')
        /* nothing */;

    return tmp;
}

/*
// Copy a string into the specified destination. It is assumed that the
// source string is NULL terminated. The destination is returned.
// In the event that the source string is longer than n bytes, the
// destination string will be null terminated and only contain the first
// n-1 characters of the source string.
*/
char * tpi_strncpy(char *dst, const char *src, size_t n)
{
    char *tmp = dst;

    if (n)  /* zero length output buffer! */
    {
        while (--n && (*dst++ = *src++) != '\0')
            /* nothing */;
        *dst = 0;
    }
    return tmp;
}

/*
// Convert the string to lower case.  The string is converted in place,
// and a pointer to the beginning of the string is returned.  All
// non-alphabetic characters are left unchanged.
*/
char *tpi_strlwr(char *s)
{
    char *sc;

    for (sc = s; *sc != '\0'; ++sc)
    {
        *sc = tpi_tolower(*sc);
    }
    return (s);
}

/*
// Convert the string to upper case.  The string is converted in place,
// and a pointer to the beginning of the string is returned.  All
// non-alphabetic characters are left unchanged.
*/
char *tpi_strupr(char *s)
{
    char *sc;

    for (sc = s; *sc != '\0'; ++sc)
    {
        *sc = tpi_toupper(*sc);
    }
    return (s);
}

/*
// Return the string length of the specified NULL terminated string.
// The length does not include the string terminator.
*/
size_t tpi_strlen(const char *s)
{
    const char *sc;

    for (sc = s; *sc != '\0'; ++sc)
        /* nothing */;

    return (sc - s);
}

/*
// Return the string length of the specified string.
*/
size_t tpi_strnlen(const char *s, size_t n)
{
    const char *sc;

    for (sc = s; (n-- != 0) && (*sc != '\0'); ++sc)
        /* nothing */;

    return (sc - s);
}

/*
// Compare the characters of the given character strings. If the strings
// match, 0 is returned.
*/
int tpi_strcmp(const char *s1, const char *s2)
{
    char res = 0;

    for(;;)
    {
        if (((res = (*s1 - *(s2++))) != 0) || (*(s1++) == '\0'))
        {
            break;
        }
    }
    return res;
}

/*
// Compare n characters of the given character strings. If the strings
// match, 0 is returned.
*/
int tpi_strncmp(const char *s1, const char *s2, size_t n)
{
    char res = 0;

    while (n)
    {
        if (((res = (*s1 - *(s2++))) != 0) || (*(s1++) == '\0'))
        {
            break;
        }
        n--;
    }
    return res;
}

/*
// Compare the characters of the given character strings, ignoring the
// case of the characters. If the strings match, 0 is returned.
*/
int tpi_strcasecmp(const char *s1, const char *s2)
{
    char res = 0;

    for (;;)
    {
        if (((res = (tpi_toupper(*s1) - tpi_toupper(*(s2++)))) != 0) ||
            (*(s1++) == '\0'))
        {
            break;
        }
    }
    return res;
}

/*
// Compare n characters of the given characters strings, ignoring the
// case of the characters. Returns 0 if strings match.
*/
int tpi_strncasecmp(const char *s1, const char *s2, size_t n)
{
    char res = 0;

    while (n)
    {
        if (((res = (tpi_toupper(*s1) - tpi_toupper(*(s2++)))) != 0) ||
            (*(s1++) == '\0'))
        {
            break;
        }
        n--;
    }
    return res;
}

/*
// Return the index of the first occurance of a character in the set
// string in the source string.
*/
size_t tpi_strspn(const char *s, const char *set)
{
    const char *p;
    const char *a;
    size_t n = 0;

    for (p = s; *p != '\0'; ++p)
    {
        for (a = set; *a != '\0'; ++a)
        {
            if (*p == *a)
            {
                break;
            }
        }
        if (*a == '\0')
        {
            return n;
        }
        ++n;
    }
    return n;
}

/*
// Return a pointer to the first character found in the string that is
// in the set of characters specified.
*/
char * tpi_strpbrk(const char *s, const char *set)
{
    const char *p;
    const char *a;

    for (p = s; *p != '\0'; ++p)
    {
        for (a = set; *a != '\0'; ++a)
        {
            if (*p == *a)
            {
                return (char *) p;
            }
        }
    }
    return (void *) 0;
}

/*
// Return a pointer to the first occurance of a character in the give string.
*/
char * tpi_strchr(const char *s, int c)
{
    for (; *s != c; s++)
    {
        if (*s == '\0')
        {
            return (void *) 0;
        }
    }
    return (char *) s;
}

/*
// Return a pointer to the last occurance of a character in the give string.
*/
char * tpi_strrchr(const char *s, int c)
{
    const char *p = s + tpi_strlen(s);

    do {
        if (*p == c)
        {
            return (char *) p;
        }
    } while (--p >= s);

    return (void *) 0;
}

/*
// Return a pointer to the first occurance of s2 in s1.
*/
char * tpi_strstr(const char *s1, const char *s2)
{
    unsigned int l1, l2;

    l2 = tpi_strlen(s2);
    if (l2 == 0)
    {
        return (char *) s1;
    }

    l1 = tpi_strlen(s1);
    while (l1 >= l2)
    {
        l1--;
        if (!tpi_memcmp(s1, s2, l2))
        {
            return (char *) s1;
        }
        s1++;
    }
    return (void *) 0;
}

#define ZEROPAD 0x01    /* pad with zero */
#define SIGN    0x02    /* unsigned/signed long */
#define PLUS    0x04    /* show plus */
#define SPACE   0x08    /* space if plus */
#define LEFT    0x10    /* left justified */
#define SPECIAL 0x20    /* 0x */
#define LARGE   0x40    /* use 'ABCDEF' instead of 'abcdef' */

/*
// Return the integer representation of the incoming character string.
// The incoming string pointer is modified to point at the character
// after the decimal digits.
*/
static int skip_atoi(const char **s)
{
    int i = 0;

    while (tpi_isdigit(**s))
    {
        i = (i * 10) + *((*s)++) - '0';
    }
    return i;
}

/*
// Return the pointer to the data after consuming a number in the
// base, size, number, precision and type specified. This routine
// is used by the string formating functions like printf().
*/
static char * number(char *str, long num, int base, int size, int precision, int type)
{
    char c, sign, tmp[66];
    const char *digits = "0123456789abcdefghijklmnopqrstuvwxyz";
    int i;

    if (type & LARGE)
    {
        digits = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
    }

    if (type & LEFT)
    {
        type &= ~ZEROPAD;
    }

    if (base < 2 || base > 36)
    {
        return 0;
    }

    c = (type & ZEROPAD) ? '0' : ' ';

    sign = 0;
    if (type & SIGN)
    {
        if (num < 0)
        {
            sign = '-';
            num = -num;
            size--;
        }
        else if (type & PLUS)
        {
            sign = '+';
            size--;
        }
        else if (type & SPACE)
        {
            sign = ' ';
            size--;
        }
    }

    if (type & SPECIAL)
    {
        if (base == 16)
        {
            size -= 2;
        }
        else if (base == 8)
        {
            size--;
        }
    }

    i = 0;
    if (num == 0)
    {
        tmp[i++]='0';
    }
    else
    {
        while (num != 0)
        {
            unsigned long res;

            res = (unsigned long) num % (unsigned long) base;
            num = (unsigned long) num / (unsigned long) base;

            tmp[i++] = digits[res];
        }
    }

    if (i > precision)
    {
        precision = i;
    }
    size -= precision;

    if (!(type & (ZEROPAD | LEFT)))
    {
        while(size-- > 0)
        {
            *str++ = ' ';
        }
    }

    if (sign)
    {
        *str++ = sign;
    }

    if (type & SPECIAL)
    {
        if (base == 8)
        {
            *str++ = '0';
        }
        else if (base == 16)
        {
            *str++ = '0';
            *str++ = digits[33];
        }
    }

    if (!(type & LEFT))
    {
        while (size-- > 0)
        {
            *str++ = c;
        }
    }

    while (i < precision--)
    {
        *str++ = '0';
    }

    while (i-- > 0)
    {
        *str++ = tmp[i];
    }

    while (size-- > 0)
    {
        *str++ = ' ';
    }
    return str;
}

/*
// Format the given data represented by the variable argument list into a
// user supplied buffer and return the number of characters formatted.
*/
int tpi_vsprintf(char *buf, const char *fmt, va_list args)
{
    int len;
    unsigned long num;
    int i, base;
    char *str;
    const char *s;
    int flags;
    int field_width;
    int precision;
    int qualifier;

    for (str=buf ; *fmt ; ++fmt)
    {
        if (*fmt != '%')
        {
            *str++ = *fmt;
            continue;
        }

        /* Process flags */
        flags = 0;
    repeat:
        /* Skip leading '%' */
        ++fmt;
        switch (*fmt)
        {
        case '-': flags |= LEFT; goto repeat;
        case '+': flags |= PLUS; goto repeat;
        case ' ': flags |= SPACE; goto repeat;
        case '#': flags |= SPECIAL; goto repeat;
        case '0': flags |= ZEROPAD; goto repeat;
        }

        /* Get field width */
        field_width = -1;
        if (tpi_isdigit(*fmt))
        {
            field_width = skip_atoi(&fmt);
        }
        else if (*fmt == '*')
        {
            ++fmt;
            /* It's the next argument */
            field_width = va_arg(args, int);
            if (field_width < 0)
            {
                field_width = -field_width;
                flags |= LEFT;
            }
        }

        /* Get the precision */
        precision = -1;
        if (*fmt == '.')
        {
            ++fmt;
            if (tpi_isdigit(*fmt))
            {
                precision = skip_atoi(&fmt);
            }
            else if (*fmt == '*')
            {
                ++fmt;
                /* It's the next argument */
                precision = va_arg(args, int);
            }
            if (precision < 0)
            {
                precision = 0;
            }
        }

        /* Get the conversion qualifier */
        qualifier = -1;
        if (*fmt == 'h')
        {
            qualifier = 'h';
            ++fmt;
        }
        else if (*fmt == 'l' || *fmt == 'L')
        {
            qualifier = 'l';
            ++fmt;
        }

        /* Default base */
        base = 10;

        switch (*fmt)
        {
        case 'c':
            if (!(flags & LEFT))
            {
                while (--field_width > 0)
                {
                    *str++ = ' ';
                }
            }
            *str++ = (unsigned char) va_arg(args, int);
            while (--field_width > 0)
            {
                *str++ = ' ';
            }
            continue;

        case 's':
            s = va_arg(args, char *);
            if (!s)
            {
                s = "<NULL>";
            }

            if (qualifier == 'l'
#if defined(UNICODE)
                || qualifier == 'h'
#endif
                )
            {

                len = 0;
                while (((short *)s)[len])
                {
                    len++;
                }
            }
            else if (flags & SPECIAL)
            {
                /* pascal string (first character is length) */
                len = *s++;
                if (precision >0 && len > precision)
                {
                    len = precision;
                }
            }
            else
            {
                len = tpi_strnlen(s, precision);
            }

            if (!(flags & LEFT))
            {
                while (len < field_width--)
                {
                    *str++ = ' ';
                }
            }
            for (i = 0; i < len; ++i)
            {
                if (qualifier == 'l'
#if defined(UNICODE)
                    || qualifier == 'h'
#endif
                    )
                {
                    *str++ = (char) (*(short *)s);
                    s += sizeof(short);
                }
                else
                {
                    *str++ = *s++;
                }
            }
            while (len < field_width--)
            {
                *str++ = ' ';
            }
            continue;

        case 'p':
            if (field_width == -1)
            {
                field_width = 2 * sizeof(void *);
                flags |= ZEROPAD;
            }
            str = number(str, (unsigned long) va_arg(args, void *), 16,
                field_width, precision, flags);
            continue;


        case 'n':
            if (qualifier == 'l')
            {
                long * ip = va_arg(args, long *);
                *ip = (str - buf);
            }
            else
            {
                int * ip = va_arg(args, int *);
                *ip = (str - buf);
            }
            continue;

        /* Integer number formats - set up the flags and "break" */
        case 'b':
            base = 2;
            break;

        case 'o':
            base = 8;
            break;

        case 'X':
            flags |= LARGE;
        case 'x':
            base = 16;
            break;

        case 'd':
        case 'i':
            flags |= SIGN;
        case 'u':
            break;

        default:
            if (*fmt != '%')
            {
                *str++ = '%';
            }
            if (*fmt)
            {
                *str++ = *fmt;
            }
            else
            {
                --fmt;
            }
            continue;
        }
        if (qualifier == 'l')
        {
            num = va_arg(args, unsigned long);
        }
        else if (qualifier == 'h')
        {
            if (flags & SIGN)
            {
                num = va_arg(args, short);
            }
            else
            {
                num = va_arg(args, unsigned short);
            }
        }
        else if (flags & SIGN)
        {
            num = va_arg(args, int);
        }
        else
        {
            num = va_arg(args, unsigned int);
        }
        str = number(str, num, base, field_width, precision, flags);
    }
    *str = '\0';

    return str-buf;
}

/*
// Format the given data into a user supplied buffer and return the
// number of characters formatted.
*/
int tpi_sprintf(char *buffer, const char *fmtString, ...)
{
    va_list  params;
    int      i;

    va_start(params, fmtString);
    i = tpi_vsprintf(buffer, fmtString, params);
    va_end(params);

    return i;
}

/*
// Modified sprintf to format into Str255 pascal string.
*/
int tpi_sprintf255(
    unsigned char *buf255,  /* 255 bytes max! */
    const char *fmt,
    ...
    )
{
    va_list params;

    va_start(params, fmt);
    buf255[0] = (unsigned char)tpi_vsprintf((char *)&buf255[1], fmt, params);
    va_end(params);

    return ((int) buf255[0]);
}

/* Table representing character attributes used by tpi_is...() functions. */
unsigned char const _tpi_ctype[] = {
    _C,_C,_C,_C,_C,_C,_C,_C,                                /* 0-7 */
    _C,_C|_S,_C|_S,_C|_S,_C|_S,_C|_S,_C,_C,                 /* 8-15 */
    _C,_C,_C,_C,_C,_C,_C,_C,                                /* 16-23 */
    _C,_C,_C,_C,_C,_C,_C,_C,                                /* 24-31 */
    _S|_SP,_P,_P,_P,_P,_P,_P,_P,                            /* 32-39 */
    _P,_P,_P,_P,_P,_P,_P,_P,                                /* 40-47 */
    _D,_D,_D,_D,_D,_D,_D,_D,                                /* 48-55 */
    _D,_D,_P,_P,_P,_P,_P,_P,                                /* 56-63 */
    _P,_U|_X,_U|_X,_U|_X,_U|_X,_U|_X,_U|_X,_U,              /* 64-71 */
    _U,_U,_U,_U,_U,_U,_U,_U,                                /* 72-79 */
    _U,_U,_U,_U,_U,_U,_U,_U,                                /* 80-87 */
    _U,_U,_U,_P,_P,_P,_P,_P,                                /* 88-95 */
    _P,_L|_X,_L|_X,_L|_X,_L|_X,_L|_X,_L|_X,_L,              /* 96-103 */
    _L,_L,_L,_L,_L,_L,_L,_L,                                /* 104-111 */
    _L,_L,_L,_L,_L,_L,_L,_L,                                /* 112-119 */
    _L,_L,_L,_P,_P,_P,_P,_C,                                /* 120-127 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,                        /* 128-143 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,                        /* 144-159 */
    _S|_SP,_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,    /* 160-175 */
    _P,_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,_P,        /* 176-191 */
    _U,_U,_U,_U,_U,_U,_U,_U,_U,_U,_U,_U,_U,_U,_U,_U,        /* 192-207 */
    _U,_U,_U,_U,_U,_U,_U,_P,_U,_U,_U,_U,_U,_U,_U,_L,        /* 208-223 */
    _L,_L,_L,_L,_L,_L,_L,_L,_L,_L,_L,_L,_L,_L,_L,_L,        /* 224-239 */
    _L,_L,_L,_L,_L,_L,_L,_P,_L,_L,_L,_L,_L,_L,_L,_L         /* 240-255 */
};

#endif /* DBG || !defined(NO_tpi_UTILS) */

