/*+M*************************************************************************
 * GlobeSpan ADSL WAN Interface
 *
 * This driver bypasses the flip buffer usage since this driver only accepts
 * synchronous PPP data. The flip buffer only holds 512 bytes at a time,
 * so the driver maintains it's own buffer that can hold a whole ethernet
 * frame.
 *
 * Copyright (c) 2000 GlobeSpan Semiconductor Inc. -- http://www.globespan.net
 * All rights reserved.
 *
 * Developed for GlobeSpan Semiconductor Inc. by TriplePoint, Inc.
 *   http://www.triplepoint.com
 *-M*************************************************************************/
#include "Gp.h"
#include "drvh/gsi_pci.h"
#include "drvh/gsi_ioctl.h"

#if 0
//#define GSI_WAN_MAX_ADAPTERS    8
//#define GSI_TTY_MAJOR           240
//#define GSI_TTYAUX_MAJOR        44
#endif

#define GSI_WAN_EVENT_WRITE_WAKEUP  0

#define GSI_WAN_DEFAULT_FLAGS   0 

#if DBG
const char *DbgIoctlCmd(unsigned int cmd);
#endif 

#define TTY_ADSL_MAX_MTU	2000

int GpWanChannelStartup(channel_t  *pChan);

/*+F*************************************************************************
 * Function:
 *   GpWanAlloc
 *
 * Description:
 *   Allocate the WAN interface.
 *-F*************************************************************************/
int GpWanAlloc(adapter_t      *pAdapter)
{
	gsi_wan_t  *pWan =  &(pAdapter->wan);
	channel_t *pChan;
	int i;

	wpabs_debug_init("Calling %s ...\n", __FUNCTION__);

	wpabs_memset(pWan, 0, sizeof(*pWan));

	for (i=0;i<GSI_WAN_MAX_CHANNELS;i++){

		pChan=&(pWan->channel[i]);
		
		pChan->slock = wpabs_spinlock_alloc();
		if (!pChan->slock){
			goto wan_alloc_failed;
		}

		pChan->tqueue = wpabs_taskq_alloc();
		if (!pChan->tqueue){
			goto wan_alloc_failed;
		}

		pChan->normal_termios = adsl_termios_alloc();
		if (!pChan->normal_termios){
			goto wan_alloc_failed;
		}
		
		pChan->callout_termios = adsl_termios_alloc();
		if (!pChan->callout_termios){
			goto wan_alloc_failed;
		}
	}

	pAdapter->tty_skb = wpabs_skb_alloc(TTY_ADSL_MAX_MTU);
	if (!pAdapter->tty_skb){
		wpabs_debug_event("%s: WanAlloc: No memory!\n",pAdapter->devname);
		goto wan_alloc_failed;
	}

	return 0;

wan_alloc_failed:
	wpabs_debug_event("%s: Failed allocate adapter structures (wan)\n",
				pAdapter->devname);
	GpWanFree(pAdapter);
	return -ENOMEM;
}

/*+F*************************************************************************
 * Function:
 *   GpWanFree
 *
 * Description:
 *   Free the WAN interface.
 *-F*************************************************************************/
void GpWanFree(adapter_t  *pAdapter)
{
	gsi_wan_t  *pWan = &(pAdapter->wan);
	channel_t  *pChan;
	
	wpabs_debug_init("Calling %s ...\n", __FUNCTION__);
    	/*FIXME: Dangerous without a spin lock
         */    	

	pChan=&(pWan->channel[0]);
		
	if (pChan->slock){
		wpabs_spinlock_free(pChan->slock);
		pChan->slock=NULL;
	}
	if (pChan->tqueue){
		wpabs_free(pChan->tqueue);
		pChan->tqueue=NULL;
	}
	if (pChan->normal_termios){
		wpabs_free(pChan->normal_termios);
		pChan->normal_termios=NULL;
	}
	if (pChan->callout_termios){
		wpabs_free(pChan->callout_termios);
		pChan->callout_termios=NULL;
	}
	
	if (pAdapter->tty_skb){
		void *tmp_skb=pAdapter->tty_skb;
		pAdapter->tty_skb=NULL;
		wpabs_skb_free(tmp_skb);
    	}
}


/*+F*************************************************************************
 * Function:
 *   GpWanSoftint
 *
 * Description:
 *   Routine to actually process queued channel events.
 *-F*************************************************************************/
static void GpWanSoftint(void *context)
{
    channel_t          *pChan = (channel_t *)context;

    wpabs_debug_event("In GpWanSoftint\n");
    
    if (!pChan->count){
	return;
    }
    
    if (pChan->tty == NULL){
	return;
    }

    adsl_wan_soft_intr(pChan->tty,GSI_WAN_EVENT_WRITE_WAKEUP,&(pChan->event));
}


/*+F*************************************************************************
 * Function:
 *   GpWanRegister
 *
 * Description:
 *   Register the WAN interface of the adapter as a TTY device with the
 *   operating system.
 *-F*************************************************************************/
int
GpWanRegister(adapter_t *pAdapter)
{
	DBG_FUNCX("GpWanRegister", DBG_WAN_SYS)
	gsi_wan_t  *pWan = &(pAdapter->wan);
	int     status = 0;
	channel_t  *pChan;

	wpabs_debug_init("Calling GpWanRegister ...\n");
	status = adsl_wan_register(pAdapter,
				   pAdapter->devname,
			       	   pWan->tty_minor);
	
			 /*      GSI_WAN_MAX_CHANNELS);*/
	    
    	if (status != 0){
		goto gsi_wan_register_exit;
    	}
  
	pWan->flags |= _TTY_ALLOC;

	wpabs_debug_init("pAdapter->wan.channel[0] in GpWanRegister %4x\n", 
			pWan->channel[0]);

	pChan = &(pWan->channel[0]);
	pChan->pAdapter         = pAdapter;
	pChan->count            = 0;
	pChan->flags            = GSI_WAN_DEFAULT_FLAGS;
	pChan->tty		= NULL;
	wpabs_taskq_init(pChan->tqueue,GpWanSoftint,pChan);

gsi_wan_register_exit:
    	if (status != 0){
        	GpWanUnregister(pAdapter);
    	}

    	DBG_RETURNF(DbgInfo, ("status %d", status));
    	return status;
}

/*+F*************************************************************************
 * Function:
 *   GpWanUnregister
 *
 * Description:
 *   Unregister the TTY interface of the adapter with the operating system.
 *-F*************************************************************************/
void
GpWanUnregister(adapter_t *pAdapter)
{
	gsi_wan_t  *pWan = &(pAdapter->wan);
	channel_t *pChan = &pAdapter->wan.channel[0];
	unsigned long flags;

	wpabs_debug_init("Calling GpWanUnregister ...\n");
	
	if (pWan->flags & _TTY_ALLOC){
		pWan->flags &= ~_TTY_ALLOC;
		adsl_card_lock_irq(pAdapter,&flags);
		adsl_wan_unregister(pWan->tty_minor);
		adsl_card_unlock_irq(pAdapter,&flags);
	}

	DBG_LEAVE(DbgInfo);
}


/*+F*************************************************************************
 * Function:
 *   GpWanOpen
 *
 * Description:
 *   Entry point for TTY device open request from the application.
 *-F*************************************************************************/
int GpWanOpen(void *pAdapter_ptr, unsigned char line, void *tty, void **data)
{
    	DBG_FUNCX("GpWanOpen", DBG_WAN_SYS)
    	adapter_t  *pAdapter = (adapter_t*)pAdapter_ptr;
    	channel_t  *pChan = &(pAdapter->wan.channel[0]);
    	int     status = 0;
	DSL_MODEM* pDslModem = pAdapter->custstruct;
	unsigned long flags;

    	wpabs_debug_cfg("wanpipe: WanOpen\n");

    	if (pChan == NULL){
		return NULL;
	}

	if (!pDslModem || pDslModem->InitStatus & GP_INIT_SHUTDOWN){
		return NULL;
	}
	
	wpabs_debug_event("%s: %s%d Open\n",
			pAdapter->devname,GSI_WAN_TTY_NAME,pAdapter->wan.tty_minor);

	adsl_card_lock_irq(pAdapter,&flags);

        pChan->count++;
        if (pChan->count == 1){
		
            	pChan->tty = tty;
		*data=pChan;

            	status = GpWanChannelStartup(pChan);
            	if (status != 0){
                	wpabs_debug_event("wanpipe: gsi_wan_channel_startup failed!\n");
			
			pChan->tty=NULL;
			*data=NULL;
			
			adsl_card_unlock_irq(pAdapter,&flags);
			pChan->count--;
                	goto gsi_wan_open_exit;
            	}
	}

	adsl_card_unlock_irq(pAdapter,&flags);

gsi_wan_open_exit:
	return status;

}

/*+F*************************************************************************
 * Function:
 *   GpWanClose
 *
 * Description:
 *   Entry point for TTY device close request from the application.
 *-F*************************************************************************/
int GpWanClose(void *pAdapter_ptr, void *pChan_ptr)
{
	channel_t  *pChan = (channel_t  *)pChan_ptr;
	adapter_t  *pAdapter = (adapter_t  *)pAdapter_ptr;
    	unsigned long flags;

	wpabs_debug_cfg("wanpipe: GpWanClose\n");

	adsl_card_lock_irq(pAdapter,&flags);

	pChan->count--;
	if (pChan->count == 0){

		wpabs_debug_event("%s: %s%d Close\n",
			pAdapter->devname,GSI_WAN_TTY_NAME,pAdapter->wan.tty_minor);
		
            	pChan->tty = NULL;
        }
	adsl_card_unlock_irq(pAdapter,&flags);

	return pChan->count;
}

/*+F*************************************************************************
 * Function:
 *   GpWanTx
 *
 * Description:
 *   Send a buffer to the adapter. It is assumed that the data has already
 *   been formatted as a synchronous PPP packet. This can be assured by
 *   setting the sync flag to PPP. Since the adapter does not need the PPP
 *   Address/Control bytes, they are stripped before the packet is sent down
 *   to the adapter.
 *-F*************************************************************************/
int GpWanTx(void 		       *pChan_ptr,
       	           int                     from_user,
        	   const unsigned char    *buffer,
        	   int                     bufferLen)
{
	channel_t  *pChan = (channel_t*)pChan_ptr;
	int         length = 0;
	adapter_t  *pAdapter = pChan->pAdapter;

	wpabs_debug_tx("wanpipe: WanTx Len=%d\n",bufferLen);
	
        if (pAdapter->dspState != DspStateTrained){
		wpabs_debug_event("%s: Error: Tx failed, ADSL link is not connected\n", 
					pAdapter->devname,
					pAdapter->dspState);
		return -EIO;
	}

	if (!pAdapter->tty_skb || wpabs_skb_len(pAdapter->tty_skb)){
		wpabs_debug_event("%s: Error: TTY Tx skb %s\n", 
				 pAdapter->devname,
				 pAdapter->tty_skb ? 
				 "Busy":"DeAllocated");
		return -EINVAL;
	}

        /* Copy the data into the NDIS packet if it is coming from
         * the user address space.  
	 *
	 * Also trip of the Address/Control bytes from the sync PPP data
	 */
	wpabs_skb_init(pAdapter->tty_skb, 16);
	wpabs_skb_trim(pAdapter->tty_skb, 0);

	if (from_user){
		wpabs_skb_copyback_user (pAdapter->tty_skb,0, bufferLen-2 , (unsigned long)(buffer+2));
	}else{
		wpabs_skb_copyback (pAdapter->tty_skb,0, bufferLen-2, (unsigned long)(buffer+2));
	}

	wpabs_debug_print_skb(pAdapter->tty_skb,1);
	
	DBG_TX(DbgInfo, ("###### TxWan %d bytes...\n", dataLen));
        switch (GpatmisRfc1577Send(pAdapter, pAdapter->tty_skb, (unsigned int)pChan)){
          
		case NDIS_STATUS_SUCCESS:
			DBG_TX(DbgInfo, ("###### TxWan complete.\n"));
			pAdapter->tx_collisions=0;
			length = bufferLen;
			break;

            	case NDIS_STATUS_PENDING:
			DBG_TX(DbgInfo, ("###### TxWan pending...\n"));
			length = 0;
			break;

            	default:
			DBG_TX(DbgInfo, ("###### TxWan failed!\n"));
			length = -ENOTCONN;
			break;
      	}

	wpabs_skb_init(pAdapter->tty_skb, 16);
	wpabs_skb_trim(pAdapter->tty_skb, 0);

   	return length;
}

/*+F*************************************************************************
 * Function:
 *   GpWanTxComplete
 *
 * Description:
 *   The previous transmit request has finished. Release the buffer(s) back
 *   to the free list and tell the TTY subsystem that we can accept more
 *   transmit data.
 *-F*************************************************************************/
void
GpWanTxComplete(
    adapter_t      *pAdapter,
    channel_t      *pChan,
    int             status
    )
{
    DBG_FUNCX("GpWanTxComplete", DBG_WAN_SYS)

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pAdapter", ("%d (0x%p)", DBG_ID(pAdapter), pAdapter));
    DBG_PARAMF(DbgInfo, "pChan", ("(0x%p)", pChan));
    DBG_PARAMF(DbgInfo, "status", ("%d", status));

    /* Tell the upper layers that we are ready for another packet.*/
    wpabs_taskq_schedule_event(GSI_WAN_EVENT_WRITE_WAKEUP,&pChan->event,pChan->tqueue);

    DBG_LEAVE(DbgInfo);
}

/*+F*************************************************************************
 * Function:
 *   GpWanMediaChange
 *
 * Description:
 *   Used by the adapter to inform this interface layer that link state
 *   has changed.
 *-F*************************************************************************/
void
GpWanMediaChange(
    adapter_t      *pAdapter
    )
{
    DBG_FUNCX("GpWanMediaChange", DBG_WAN_SYS)
    gsi_wan_t      *pWan = &(pAdapter->wan);

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pAdapter", ("%d (0x%p)", DBG_ID(pAdapter), pAdapter));
    wpabs_debug_init("Calling GpWanMediaChange...\n");

    if (pAdapter->dspState != DspStateTrained){
	GpatmisStateDown(pAdapter);	
    }

    if (pWan != NULL){
        /* Nothing to do */
    }

    DBG_LEAVE(DbgInfo);
}


/*+F*************************************************************************
 * Function:
 *   GpWanRx
 *
 * Description:
 *   Handle newly received data coming from the adapter. As the data comes
 *   from the adapter, it will be missing the Address/Control bytes that are
 *   needed for the complete synchronous PPP data packet. Those bytes need
 *   to be re-added to the packet before it is give up to the application.
 *-F*************************************************************************/
void
GpWanRx(
    adapter_t      *pAdapter,
    channel_t      *pChan,
    void           *pPacket,
    u32             packetSize
    )
{
    unsigned long flags;

    adsl_card_lock_irq(pAdapter,&flags);

    if (!pChan->count){
	    adsl_card_unlock_irq(pAdapter,&flags);
	    return;
    }
    
    if (pChan->tty != NULL)
    {
        u8     *pData;
        int     dataLen;
	u8     *pFlags;

        wpabs_debug_rx("%s: RxWan %d bytes\n", pAdapter->devname,packetSize);

        pData = pChan->rxData;
        dataLen = sizeof(pChan->rxData);
	pFlags = pChan->rxFlags;

        if ((packetSize + 2) < dataLen)
        {
            u8     *pPkt = (u8 *) pPacket;

            /*
	    //	 Add on the Address/Control bytes if required. There is an
            // issue with the GlobeSpan CO pAdapter that it transmits these
            // bytes! So, if they are already present, don't add another
            // set. DWG
	    */
            if ((pPkt[0] == 0xFF) && (pPkt[1] == 0x03))
            {
                dataLen = packetSize;
                wpabs_memcpy(pData, pPacket, packetSize);
            
	    }else{
                pData[0] = 0xFF; pData[1] = 0x03;

                dataLen = packetSize + 2;
                wpabs_memcpy(pData + 2, pPacket, packetSize);
            }

	    adsl_card_unlock_irq(pAdapter,&flags);

	    adsl_tty_receive(pChan->tty, pData, pFlags, dataLen);
	    
            wpabs_taskq_schedule_event(GSI_WAN_EVENT_WRITE_WAKEUP,&pChan->event,pChan->tqueue);
        }
        else
        {
	    adsl_card_unlock_irq(pAdapter,&flags);
            wpabs_debug_event("%s: WanRx: Packet size too large to accept: %d\n",
			    pAdapter->devname,packetSize);

        }
    }
    else
    {
	adsl_card_unlock_irq(pAdapter,&flags);
        wpabs_debug_event("%s: WanRx: No channel TTY defined!\n",pAdapter->devname);
    }
 
}

#if 0
/*+F*************************************************************************
 * Function:
 *   GpWanUpdateFlags
 *
 * Description:
 *   Routine to update the internal channel flags based on the state of
 *   the channel.
 *-F*************************************************************************/
DBG_STATIC void
GpWanUpdateFlags(
    channel_t  *pChan
    )
{
    DBG_FUNCX("GpWanUpdateFlags", DBG_WAN_SYS)

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pChan", ("(0x%p)", pChan));

    if (C_BAUD(pChan->tty))
    {
        pChan->mcr |= UART_MCR_DTR;
    }
    else
    {
        pChan->mcr &= ~UART_MCR_DTR;
    }

    if (C_CRTSCTS(pChan->tty))
    {
        pChan->flags |= ASYNC_CTS_FLOW;
    }
    else
    {
        pChan->flags &= ~ASYNC_CTS_FLOW;
    }

    if (C_CLOCAL(pChan->tty))
    {
        pChan->flags &= ~ASYNC_CHECK_CD;
    }
    else
    {
        pChan->flags |= ASYNC_CHECK_CD;
    }

    DBG_LEAVE(DbgInfo);
}
#endif

/*+F*************************************************************************
 * Function:
 *   GpWanChannelStartup
 *
 * Description:
 *   Startup the given channel. This routine allocates the necessary Tx/Rx
 *   resources and makes the TTY ready for operation.
 *   NOTE: The channel spinlock needs to be held by calling routine.
 *-F*************************************************************************/
int GpWanChannelStartup(channel_t  *pChan)
{
	    DBG_FUNCX("GpWanChannelStartup", DBG_WAN_SYS)
	    int             status = 0;

	    DBG_ENTER(DbgInfo);
	    DBG_PARAMF(DbgInfo, "pChan", ("(0x%p)", pChan));

	    pChan->event = 0;
	    pChan->txAvailable = GSI_WAN_NUM_NDIS_PACKETS;
	    pChan->txHead = 0;
	    pChan->txBytes = 0;

#if 0
	    if (!(pChan->flags & ASYNC_INITIALIZED)){
			pChan->mcr = (UART_MCR_DTR | UART_MCR_RTS | UART_MCR_OUT2);
			if (pChan->tty){
				clear_bit(TTY_IO_ERROR, &(pChan->tty->flags));
			}

			GpWanUpdateFlags(pChan);

			pChan->flags |= ASYNC_INITIALIZED;
			pChan->msr |= (UART_MSR_DSR | UART_MSR_CTS);
	    }

	    DBG_RETURNF(DbgInfo, ("status %d", status));
#endif
	    return status;
}

/*+F*************************************************************************
 * Function:
 *   GpWanWriteRoom
 *
 * Description:
 *   Return the number of additional transmit bytes the device can accept.
 *-F*************************************************************************/
int GpWanWriteRoom(void *pChan_ptr)
{
	channel_t  *pChan = (channel_t  *)pChan_ptr; 
	adapter_t  *pAdapter = pChan->pAdapter;

    	if (pAdapter->dspState != DspStateTrained){
		return 0;
	}

	return GSI_WAN_NDIS_BUFFER_SIZE;
}


#if 0
/*+F*************************************************************************
 * Function:
 *   GpWanChannelShutdown
 *
 * Description:
 *   Mark the TTY as down and not ready to receive characters, and release
 *   the Tx/Rx resources.
 *   NOTE: The channel spinlock needs to be held by calling routine.
 *-F*************************************************************************/
DBG_STATIC void
GpWanChannelShutdown(
    channel_t  *pChan
    )
{
    DBG_FUNCX("GpWanChannelShutdown", DBG_WAN_SYS)

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pChan", ("(0x%p)", pChan));

#if 0
    if (pChan->flags & ASYNC_INITIALIZED)
    {
        pChan->msr &= ~UART_MSR_RI;
        if (!pChan->tty || C_HUPCL(pChan->tty))
        {
            pChan->mcr &= ~(UART_MCR_DTR | UART_MCR_RTS);
        }

        if (pChan->tty){
            set_bit(TTY_IO_ERROR, &(pChan->tty->flags));
        }

        pChan->flags &= ~ASYNC_INITIALIZED;
    }
#endif
    
    pChan->txAvailable = 0;
    pChan->event = 0;

    DBG_LEAVE(DbgInfo);
}
#endif


#if 0
/*+F*************************************************************************
 * Function:
 *   GpWanGetModemInfo
 *
 * Description:
 *   Return the modem status flags to the application. The flags needed
 *   to be mapped from the kernel version of the flags into the application
 *   versions.
 *-F*************************************************************************/
DBG_STATIC int
GpWanGetModemInfo(
    channel_t      *pChan,
    unsigned int   *pArg
    )
{
    DBG_FUNCX("GpWanGetModemInfo", DBG_WAN_SYS)
    int         status = 0;
    unsigned int    flags;

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pChan", ("(0x%p)", pChan));
    DBG_PARAMF(DbgInfo, "pArg", ("(0x%p)", pArg));

    flags =
        ((pChan->mcr & UART_MCR_RTS) ? TIOCM_RTS : 0) |
        ((pChan->mcr & UART_MCR_DTR) ? TIOCM_DTR : 0) |
        ((pChan->msr & UART_MSR_DCD) ? TIOCM_CD  : 0) |
        ((pChan->msr & UART_MSR_RI)  ? TIOCM_RI  : 0) |
        ((pChan->msr & UART_MSR_DSR) ? TIOCM_DSR : 0) |
        ((pChan->msr & UART_MSR_CTS) ? TIOCM_CTS : 0);

#if DBG
    {
    char    flagStr[128];

    *flagStr = '\0';
    if (flags & TIOCM_LE) strcat(flagStr, "LE ");
    if (flags & TIOCM_DTR) strcat(flagStr, "DTR ");
    if (flags & TIOCM_RTS) strcat(flagStr, "RTS ");
    if (flags & TIOCM_ST) strcat(flagStr, "ST ");
    if (flags & TIOCM_SR) strcat(flagStr, "SR ");
    if (flags & TIOCM_CTS) strcat(flagStr, "CTS ");
    if (flags & TIOCM_DSR) strcat(flagStr, "DSR ");
    if (flags & TIOCM_CD) strcat(flagStr, "CD ");
    if (flags & TIOCM_RI) strcat(flagStr, "RI ");
    if (flags & TIOCM_OUT1) strcat(flagStr, "OUT1 ");
    if (flags & TIOCM_OUT2) strcat(flagStr, "OUT2 ");
    if (*flagStr != '\0') flagStr[strlen(flagStr) - 1] = '\0';
    DBG_PRINT(DbgInfo, ("#### Modem Info: %s (0x%08x)\n", flagStr, flags));
    }
#endif /* DBG */

    status = put_user(flags, pArg);

    DBG_RETURNF(DbgInfo, ("status %d", status));
    return status;
}
#endif

#if 0
/*+F*************************************************************************
 * Function:
 *   GpWanSetModemInfo
 *
 * Description:
 *   Interface for the application to modify the modem control information.
 *-F*************************************************************************/
DBG_STATIC int
GpWanSetModemInfo(
    channel_t      *pChan,
    unsigned int    cmd,
    unsigned int   *pArg
    )
{
    DBG_FUNCX("GpWanSetModemInfo", DBG_WAN_SYS)
    int         status = 0;
    unsigned int    flags;

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pChan", ("(0x%p)", pChan));
    DBG_PARAMF(DbgInfo, "cmd", ("%d", cmd));
    DBG_PARAMF(DbgInfo, "pArg", ("(0x%p)", pArg));

    status = get_user(flags, pArg);
    if (status == 0)
    {
        switch (cmd)
        {
        case TIOCMBIS:
            DBG_PRINT(DbgInfo, ("#### TIOCMBIS: "));
            if (flags & TIOCM_RTS)
            {
                pChan->mcr |= UART_MCR_RTS;
                DBG_PRINT(DbgInfo, ("RTS "));
            }

            if (flags & TIOCM_DTR)
            {
                pChan->mcr |= UART_MCR_DTR;
                DBG_PRINT(DbgInfo, ("DTR "));
            }
            break;

        case TIOCMBIC:
            DBG_PRINT(DbgInfo, ("#### TIOCMBIC: "));
            if (flags & TIOCM_RTS)
            {
                pChan->mcr &= ~UART_MCR_RTS;
                DBG_PRINT(DbgInfo, ("~RTS "));
            }

            if (flags & TIOCM_DTR)
            {
                pChan->mcr &= ~UART_MCR_DTR;
                DBG_PRINT(DbgInfo, ("~DTR "));
            }
            break;

        case TIOCMSET:
            DBG_PRINT(DbgInfo, ("#### TIOCMSET: "));
            if (flags & TIOCM_RTS)
            {
                pChan->mcr |= UART_MCR_RTS;
                DBG_PRINT(DbgInfo, ("RTS "));
            }
            else
            {
                pChan->mcr &= ~UART_MCR_RTS;
                DBG_PRINT(DbgInfo, ("~RTS "));
            }

            if (flags & TIOCM_DTR)
            {
                pChan->mcr |= UART_MCR_DTR;
                DBG_PRINT(DbgInfo, ("DTR "));
            }
            else
            {
                pChan->mcr &= ~UART_MCR_DTR;
                DBG_PRINT(DbgInfo, ("~DTR "));
            }
            break;

        default:
            DBG_ERROR(DbgInfo, ("Unexpected modem command %d\n", cmd));
            status = -EINVAL;
            break;
        }
        DBG_PRINT(DbgInfo, ("\n"));
    }

    DBG_RETURNF(DbgInfo, ("status %d", status));
    return status;
}
#endif

#if 0

/*+F*************************************************************************
 * Function:
 *   GpWanGetLsrInfo
 *
 * Description:
 *   Return the line status information to the application. The flags need
 *   to be converted from kernel space values to application values.
 *-F*************************************************************************/
DBG_STATIC int
GpWanGetLsrInfo(
    channel_t      *pChan,
    unsigned int   *pArg
    )
{
    DBG_FUNCX("GpWanGetLsrInfo", DBG_WAN_SYS)
    int         status = 0;
    unsigned int    flags;

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pChan", ("(0x%p)", pChan));
    DBG_PARAMF(DbgInfo, "pArg", ("(0x%p)", pArg));

    flags = ((pChan->lsr & UART_LSR_TEMT) ? TIOCSER_TEMT : 0);

#if DBG
    {
    char    flagStr[128];

    *flagStr = '\0';
    if (flags & TIOCSER_TEMT) strcat(flagStr, "TEMT ");
    if (*flagStr != '\0') flagStr[strlen(flagStr) - 1] = '\0';
    DBG_PRINT(DbgInfo, ("#### LSR Info: %s (0x%08x)\n", flagStr, flags));
    }
#endif /* DBG */

    status = put_user(flags, pArg);

    DBG_RETURNF(DbgInfo, ("status %d", status));
    return status;
}
#endif

#if 0
/*+F*************************************************************************
 * Function:
 *   GpWanIoctl
 *
 * Description:
 *   Exported IOCTL interface for the TTY device.
 *-F*************************************************************************/
DBG_STATIC int
GpWanIoctl(
    struct tty_struct  *tty,
    struct file        *pFile,
    unsigned int        cmd,
    unsigned long       arg
    )
{
    DBG_FUNCX("gsi_wan_ioctl", DBG_WAN_SYS)
    adapter_t  *pAdapter;
    channel_t  *pChan;
    int         status = 0;

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "tty", ("(0x%p)", tty));
    DBG_PARAMF(DbgInfo, "pFile", ("(0x%p)", pFile));
    DBG_PARAMF(DbgInfo, "cmd", ("%s (0x%04x)", DbgIoctlCmd(cmd), cmd));
    DBG_PARAMF(DbgInfo, "arg", ("0x%08x", arg));

    if ((pChan = (channel_t *) tty->driver_data) == NULL)
    {
        DBG_WARNING(DbgInfo, ("No channel for operation!\n"));
        status = -ENODEV;
        goto gsi_wan_ioctl_exit;
    }

    pAdapter = pChan->pAdapter;
    DBG_ASSERT(pAdapter != NULL);

    if (!GpatmisIoctl(pAdapter, cmd, (struct gsi_req *) arg, &status)) 
   {
 	wpabs_debug_cfg("wanpipe: WanIoctl Cmd=0x%x\n",cmd);
        switch (cmd)
        {
        case TCSBRK:    /* SVID version: non-zero arg --> no break */
            status = tty_check_change(tty);
            if (status == 0)
            {
                tty_wait_until_sent(tty, 0);
            }
            break;

        case TCSBRKP:   /* support for POSIX tcsendbreak() */
            status = tty_check_change(tty);
            if (status == 0)
            {
                tty_wait_until_sent(tty, 0);
            }
            break;

        case TIOCGSOFTCAR:
            status = put_user((tty->termios->c_cflag & CLOCAL) ? 1 : 0,
                (unsigned long *) arg);
            break;

        case TIOCSSOFTCAR:
            status = get_user(arg, (unsigned long *) arg);
            if (status == 0)
            {
                tty->termios->c_cflag =
                    ((tty->termios->c_cflag & ~CLOCAL) | (arg ? CLOCAL : 0));
            }
            break;

        case TIOCMGET:
            status = GpWanGetModemInfo(pChan, (unsigned int *) arg);
            break;

        case TIOCMBIS:
        case TIOCMBIC:
        case TIOCMSET:
            status = GpWanSetModemInfo(pChan, cmd, (unsigned int *) arg);
            break;

        case TIOCSERGETLSR:
            status = GpWanGetLsrInfo(pChan, (unsigned int *) arg);
            break;

        default:
            wpabs_debug_event("Unhandled IOCTL command: (0x%x)\n",
                 cmd);
            status = -ENOIOCTLCMD;
            break;
        }
    }

gsi_wan_ioctl_exit:
    DBG_RETURNF(DbgInfo, ("status %d", status));
    return status;
}
#endif



