/*****************************************************************************
* wanpipe_atm.c	WANPIPE(tm) ATM ALL5 SAR
* 		
* Authors: 	Nenad Corbic <ncorbic@sangoma.com>
*
* Copyright:	(c) 2003 Sangoma Technologies Inc.
*
*		This program is free software; you can redistribute it and/or
*		modify it under the terms of the GNU General Public License
*		as published by the Free Software Foundation; either version
*		2 of the License, or (at your option) any later version.
* ============================================================================
* Jan 07, 2003	Nenad Corbic	Based on ADSL ATM SAR
*****************************************************************************/
/* 
 * ATM Header 
 * 
 * --8---7---6---5---4---3---2---1--
 * |             VPI               |
 * |-------------------------------|
 * |     VPI       |    VCI        |
 * |-------------------------------|
 * |             VCI               |
 * |-------------------------------|
 * |      VCI      |    PT     |CLP|
 * |-------------------------------|
 * |             HEC               |
 * |-------------------------------|
 * |                               |
 * |     Information Payload       |
 * |       (48 Octects)            |
 * ---------------------------------
 */
		
/* ATM Protocol Headers (LLC) 
 *
 * --1-----1-----1--------3---------2--------<=65,527-----
 * |DSAP|SSAP|Control|Unique Id|Protocol Id|   PDU       |
 * -------------------------------------------------------
 *
 * Bridged LLC: Ethernet over ATM
 * 	DSAP		=0xaa
 * 	SSAP		=0xaa
 * 	Control		=0x03
 *	Unique ID	=0x00 0x80 0xC2 	#Bridged PDU
 *	Protocol Id	=0x00 0x07		#802.3/Ethernet 
 *						#Witout Preserverd FCS 
 *	Pading		=0x00 0x00
 *			
 * Classical IP over ATM
 *	DSAP  		=0xAA
 *	SSAP		=0xAA
 *	Control		=0x03
 *	Unique Id	=0x00 0x00 0x00
 *
 *	Protocol Id 	=0x80 0x00  : IP
 *	Protocol Id	=0x80 0x06  : ARP
 *
 *      MTU=8192
 *
 */


/*==================================================================
 * INCLUES
 *=================================================================*/

#include "wanpipe_abstr.h"
#include "wanpipe_cfg.h"
#include "wanpipe_atm.h" 


/*==================================================================
 * DEFINES
 *=================================================================*/

#define	MAX_BAD_CELLS 10

/*==================================================================
 * GLOBAL VARIABLES
 *
 * Note: This code has been written to work in SMP multi-treaded 
 *       systems, where multiple drivers might simultaneously 
 *       use this code. 
 *       Thus global variables are very dangerous.  
 *       Only use them if you KNOW what you are doning.
 *=================================================================*/

static unsigned char  g_trailer[]={0x00, 0x00, 0xEE, 0xEE, 0xFF, 0xFF, 0xFF, 0xFF};
static unsigned short byte_crc10_table[256];
static unsigned long  crc_table[256];
static unsigned char  global_init_crc;
unsigned char  HEC_table[256];
static unsigned char  FLIP_TABLE[256];


static unsigned char 
atm_oam_req[]={0x18, 0x01, 0x00, 0x03, 0x66, 0xA9, 0xFF, 0xFF, 0xFF, 0xFF, 
	       0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
	       0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};


static unsigned char
atm_cc_req[]={0x14};

//static unsigned char
//atm_cc_act_deact_req[]={0x81};

/*==================================================================
 * FUNCTION PROTOTYPES 
 *=================================================================*/

static void 		gen_crc_table(void);
static void 		gen_byte_crc10_table(void);
static unsigned long 	update_crc(unsigned long, char*, int);
unsigned char		calculate_hec_crc(unsigned char *util_ptr);
static void 		generate_hec_crc_table(void);

static int 		parse_prot_header(wp_sar_dev_t*, void*);

static int 		wphash_sar_dev(wp_atm_reg_t *, wp_sar_dev_t *);
static void 		wpunhash_sar_dev(wp_sar_dev_t *p);
static wp_sar_dev_t 	*wp_find_sar_dev_by_vpivci(wp_atm_reg_t *atm_device, 
							unsigned int key);

static __inline 
unsigned char 		is_idle_cell(unsigned int *data);

static int 		capture_trace_packet (wp_sar_dev_t *, void *, char);
static int 		capture_atm_trace_packet(wp_atm_reg_t*,void*,int,char);

static void* 		handle_atm_admin_packet(wp_sar_dev_t *, wp_atm_cell_t *);
static void* 		handle_oam_cell_loopback(wp_sar_dev_t *, wp_atm_cell_t *);
static void 		sar_tx_oam_loopback_req(wp_atm_reg_t *atm_device, wp_sar_dev_t *pAdapter);
static void* 		handle_oam_alarm(wp_sar_dev_t *, wp_atm_cell_t *);


static int 		handle_atm_arp(wp_sar_dev_t *pAdapter, void* rx_skb);
static void* 		create_inarp_request(wp_sar_dev_t *);
static void 		sar_tx_atm_arp(wp_atm_reg_t *atm_device, wp_sar_dev_t *pAdapter);

static void  		atm_flip_bitstrm(unsigned char *data, int len);
static void 		gen_flip_tables(void);

static void 		sar_tx_continuity_req(wp_atm_reg_t *atm_device, wp_sar_dev_t *pAdapter);
static void* 		sar_tx_oam_packet(wp_atm_reg_t *atm_device, wp_sar_dev_t *pAdapter, 
		 		          unsigned char *data, int len, unsigned char skb_return);

static int 		validate_mgmnt_cell(wp_atm_cell_1_t *tx_data_ptr);
static void 		sar_tx_rdi_req(wp_atm_reg_t *atm_device, wp_sar_dev_t *pAdapter);
static void 		set_atm_sar_state(wp_sar_dev_t* pAdapter, unsigned char state);


/*==================================================================
 * wp_sar_register_device
 *
 *  	o Create ATM device
 *  	o Initialize crc tables
 *  	o Initialize ATM device variables
 */ 

int wp_sar_register_device(void *card, 
		           unsigned char *devname, 
			   void **atm_device_ptr,
			   void *rx_data_list,
			   void *tx_prot_list,
			   void *trace_info_ptr)
{
	int i;
	wp_atm_reg_t *atm_device;
	
	atm_device=wpabs_malloc(sizeof(wp_atm_reg_t));
	if (!atm_device){
		WPATM_MEM_ASSERT();
		return -1;
	}

	wpabs_memset(atm_device,0,sizeof(wp_atm_reg_t));

	atm_device->devname=devname;

	for(i = 0; i < WPATMHASH_SZ; i++){
		atm_device->atm_hash[i] = NULL;
	}

	if (!wpabs_test_bit(0,&global_init_crc)){
		wpabs_set_bit(0,&global_init_crc);
		gen_crc_table();
		gen_byte_crc10_table();
		generate_hec_crc_table();
		gen_flip_tables();
	}
	
	if (wpabs_test_bit(0,&atm_device->init)){
		wpabs_debug_event("Error ATM Device busy: init=1!\n");
		return -1;
	}

	atm_device->card = card;
	atm_device->rx_data_list=rx_data_list;
	atm_device->tx_prot_list=tx_prot_list;
	atm_device->trace_info=trace_info_ptr;
		
	*atm_device_ptr = (void*)atm_device;
	wpabs_set_bit(0,&atm_device->init);

	return 0;
}

/*==================================================================
 * wp_sar_unregister_device
 *
 *  	o Unconfigure atm device
 *  	o Free it
 *
 */ 

void wp_sar_unregister_device(void *atm_device_ptr)
{
#define atm_device   ((wp_atm_reg_t *)atm_device_ptr)

	if (!atm_device_ptr){
		return;
	}	

	atm_device->card = NULL;
	atm_device->rx_data_list = NULL;
	atm_device->tx_prot_list = NULL;
	atm_device->trace_info = NULL;

	wpabs_clear_bit(0,&atm_device->init);

	wpabs_free(atm_device);
#undef atm_device
}


/*==================================================================
 * wp_sar_register_pvc
 *
 *  	o Initialize the ATM Ecapsulation Protocol
 *  	o Allocate a tx cells buffer that will be used to
 *  	  cut up tx packets into ATM cells
 *
 */ 

int wp_sar_register_pvc(void *atm_device_ptr, 
			void **pAdapter_ptr, 
			void **tx_cells_skb,
			void *chan,
			void *chan_name,
			void *dev,
			wan_atm_conf_if_t *atm_cfg,
			unsigned int mtu)
{
#define atm_device   ((wp_atm_reg_t *)atm_device_ptr)
	wp_sar_dev_t *pAdapter;
	int x;

	if (!atm_device_ptr){
		return -1;
	}	
	
	pAdapter=wpabs_malloc(sizeof(wp_sar_dev_t));
	if (!pAdapter){
		WPATM_MEM_ASSERT();
		return -1;
	}

	wpabs_memset(pAdapter,0,sizeof(wp_sar_dev_t));

	wpabs_memcpy(&pAdapter->cfg,atm_cfg,sizeof(wan_atm_conf_if_t));
	
	pAdapter->devname=chan_name;	
	
	pAdapter->max_sdu_size=mtu*ATM_OVERHEAD;
	if (pAdapter->max_sdu_size%ATM_CELL_SIZE){
		pAdapter->max_sdu_size-=pAdapter->max_sdu_size%ATM_CELL_SIZE;
		pAdapter->max_sdu_size+=ATM_CELL_SIZE;	
	}

	wpabs_debug_cfg("%s:%s: Allocating Rx cell buf Len=%d Max SDU=%d\n",
			atm_device->devname,pAdapter->devname,
			pAdapter->max_sdu_size+ATM_CELL_SIZE,			
			pAdapter->max_sdu_size);
	
	*tx_cells_skb = wpabs_skb_alloc(pAdapter->max_sdu_size+ATM_CELL_SIZE);
	if (!*tx_cells_skb){
		wpabs_free(pAdapter);
		return -1;
	}
		
	pAdapter->EncapMode=pAdapter->cfg.encap_mode;
	pAdapter->vpi=pAdapter->cfg.vpi;
	pAdapter->vci=pAdapter->cfg.vci;

	wpabs_debug_event("%s:%s: ATM Config Info\n",
			atm_device->devname,pAdapter->devname);

	wpabs_debug_event("%s:%s:     OAM Loopback Requests   = %s\n",
			atm_device->devname,pAdapter->devname,
			pAdapter->cfg.atm_oam_loopback?"Enabled":"Disabled");
	
	if (pAdapter->cfg.atm_oam_loopback){
		wpabs_debug_event("%s:%s:     OAM Loopback Interval   = %d sec\n",
			atm_device->devname,pAdapter->devname,
			pAdapter->cfg.atm_oam_loopback_intr);	
	}
	
	wpabs_debug_event("%s:%s:     OAM Continuity Check    = %s\n",
			atm_device->devname,pAdapter->devname,
			pAdapter->cfg.atm_oam_continuity?"Enabled":"Disabled");

	if (pAdapter->cfg.atm_oam_continuity){
		wpabs_debug_event("%s:%s:     OAM Continuity Interval = %d sec\n",
			atm_device->devname,pAdapter->devname,
			pAdapter->cfg.atm_oam_continuity_intr);		
	}

	if (pAdapter->EncapMode == RFC_MODE_ROUTED_IP_LLC){
		wpabs_debug_event("%s:%s:     ATM IP ARP              = %s\n",
				atm_device->devname,pAdapter->devname,
				pAdapter->cfg.atm_arp?"Enabled":"Disabled");

		if (pAdapter->cfg.atm_arp){
			wpabs_debug_event("%s:%s:     ATM IP ARP Interval     = %d sec\n",
				atm_device->devname,pAdapter->devname,
				pAdapter->cfg.atm_arp_intr);		
		}
	}
	
			
	pAdapter->atmdevice=atm_device;
	
	switch (pAdapter->EncapMode){

		case RFC_MODE_BRIDGED_ETH_LLC:
			wpabs_debug_event("%s:%s: ATM configured for Bridged LLC ETH/ATM\n",
					atm_device->devname,pAdapter->devname);
			pAdapter->prot_header[0]=0xAA;
			pAdapter->prot_header[1]=0xAA;
			pAdapter->prot_header[2]=0x03;
			pAdapter->prot_header[3]=0x00;
			pAdapter->prot_header[4]=0x80;
			pAdapter->prot_header[5]=0xC2;
			pAdapter->prot_header[6]=0x00;
			pAdapter->prot_header[7]=0x07;
			pAdapter->prot_header[8]=0x00;
			pAdapter->prot_header[9]=0x00;
			pAdapter->prot_header_size=10;	
			break;

		case RFC_MODE_BRIDGED_ETH_VC:
			wpabs_debug_event("%s:%s: ATM configured for Bridged VC ETH/ATM\n",
					atm_device->devname,pAdapter->devname);
			pAdapter->prot_header[0]=0x00;
			pAdapter->prot_header[1]=0x00;
			pAdapter->prot_header_size=2;
			break;

		case RFC_MODE_ROUTED_IP_LLC:
			wpabs_debug_event("%s:%s: ATM configured for Classical IP (LLC) over ATM\n",
					atm_device->devname,pAdapter->devname);
			pAdapter->prot_header[0]=0xAA;
			pAdapter->prot_header[1]=0xAA;
			pAdapter->prot_header[2]=0x03;
			pAdapter->prot_header[3]=0x00;
			pAdapter->prot_header[4]=0x00;
			pAdapter->prot_header[5]=0x00;
			pAdapter->prot_header[6]=0x08;
			pAdapter->prot_header[7]=0x00;
			pAdapter->prot_header_size=8;

			pAdapter->tx_arp_skb=wpabs_skb_alloc(MAX_ARP_SIZE);
			if (!pAdapter->tx_arp_skb){
				wpabs_skb_free(*tx_cells_skb);
				*tx_cells_skb=NULL;
				wpabs_free(pAdapter);	
				return -1;
			}
			break;


		case RFC_MODE_ROUTED_IP_VC:
			wpabs_debug_event("%s:%s: ATM configured for IP (VC) over ATM\n",
					atm_device->devname,pAdapter->devname);
			pAdapter->prot_header_size=0;
			break;
			
		case RFC_MODE_PPP_LLC:
			wpabs_debug_event("%s:%s: ATM configured for PPP (LLC) over ATM\n",
					atm_device->devname,pAdapter->devname);
			pAdapter->prot_header[0]=0xFE;
			pAdapter->prot_header[1]=0xFE;
			pAdapter->prot_header[2]=0x03;
			pAdapter->prot_header[3]=0xCF;
			pAdapter->prot_header_size=4;
			break;

		case RFC_MODE_PPP_VC:
			wpabs_debug_event("%s:%s: ATM configured for PPP (VC) over ATM\n",
					atm_device->devname,pAdapter->devname);
			pAdapter->prot_header_size=0;
			break;


		default:
			wpabs_debug_event("%s:%s: ERROR: Invalid ATM Lan Protocol option %d!\n",
					atm_device->devname,pAdapter->devname,pAdapter->EncapMode);
			wpabs_free(pAdapter);
			wpabs_skb_free(*tx_cells_skb);
			*tx_cells_skb=NULL;
			return -1;
	}


	pAdapter->hash_next=NULL;
	pAdapter->hash_pprev=NULL;
	
	pAdapter->atm_header[0] = (pAdapter->vpi &0xF0)>>4;	
	pAdapter->atm_header[1] = (pAdapter->vpi &0x0F)<<4;	
	pAdapter->atm_header[1] |= (pAdapter->vci&0xF000)>>12;
	pAdapter->atm_header[2] = (pAdapter->vci &0x0FF0)>>4;
	pAdapter->atm_header[3] = (pAdapter->vci &0x000F)<<4;

	
	/* Calculate the hec number now because it will be the
	 * same all cells, since vpi/vci numbers will not change */
	pAdapter->atm_header[4] = calculate_hec_crc(pAdapter->atm_header);
	pAdapter->hec=pAdapter->atm_header[4];

	/* Calculate the hec of atm header when EOF bit
	 * (end of frame bit) is set */
	wpabs_set_bit(1,&pAdapter->atm_header[3]);
	pAdapter->hec_eof=calculate_hec_crc(pAdapter->atm_header);

	/* Reset the EOF frame bit */
	wpabs_clear_bit(1,&pAdapter->atm_header[3]);

	wpabs_memcpy(&pAdapter->trailer[0],&g_trailer[0],sizeof(pAdapter->trailer));
	
	wpabs_debug_cfg("%s:%s: ATM header = %x %x %x %x\n",
			atm_device->devname,
			pAdapter->devname,
			pAdapter->atm_header[0],pAdapter->atm_header[1],
			pAdapter->atm_header[2],pAdapter->atm_header[3]);


	wpabs_memset(pAdapter->pad_len_buf,PAD_CHAR,sizeof(pAdapter->pad_len_buf));

	for (x = 0; x<MAX_ATM_ADDR; x++){
		if (x<=8){
			pAdapter->atm_addr[x]=0x31+x;
		}else if (x<=12){
			pAdapter->atm_addr[x]=0x30+x-9;
		}else if (x<=18){
			pAdapter->atm_addr[x]=0x0; 
		}else{
			pAdapter->atm_addr[x]=0x0f;
		}
	}

	if (wphash_sar_dev(atm_device,pAdapter) != 0){
		wpabs_free(pAdapter);
		wpabs_skb_free(*tx_cells_skb);
		*tx_cells_skb=NULL;
		return -1;
	}

	pAdapter->priv=chan;
	pAdapter->dev=dev;

	set_atm_sar_state(pAdapter,ATM_CONNECTED);

	*pAdapter_ptr=(void*)pAdapter;

	return 0; 
#undef atm_device
}

/*==================================================================
 * wp_sar_unregister_pvc
 *
 *  	o Unconfigure pvc device
 *  	o Free it
 */ 
void wp_sar_unregister_pvc(void *atm_device_ptr, void *pAdapter_ptr, void *tx_skb)
{
#define atm_device   ((wp_atm_reg_t *)atm_device_ptr)
#define pAdapter     ((wp_sar_dev_t *)pAdapter_ptr)

	/* FIXME: This section should be locked */

	if (!atm_device_ptr || !pAdapter_ptr){
		return;
	}	
	
	if (pAdapter->rx_pkt_skb){
		wpabs_skb_free(pAdapter->rx_pkt_skb);
		pAdapter->rx_pkt_skb = NULL;
	}

	if (tx_skb){
		wpabs_skb_free(tx_skb);
	}

	if (pAdapter->tx_arp_skb){
		wpabs_skb_free(pAdapter->tx_arp_skb);
	}

	wpunhash_sar_dev(pAdapter);

	wpabs_free(pAdapter);
	return;

#undef pAdapter
#undef atm_device
}

/*==================================================================
 * wanpipe_sar_rx
 *
 *  	o ATM Device global rx routine
 *  	o The incoming rx_cell_block contains aritary number
 *  	  of ATM cells.  The rx functions parses the ATM
 *  	  cells, and handles ATM managemnt, oam and data packets.
 */ 
int wanpipe_sar_rx (void *atm_device_ptr, void *rx_cell_block)
{
#define atm_device   ((wp_atm_reg_t *)atm_device_ptr)

	volatile wp_atm_cell_t*	atm_cell;
	void*			rx_skb;
	unsigned short 		info_len;
	unsigned char*		data_ptr;
	int 			num_of_cells = 0;
	wp_sar_dev_t		*pAdapter=NULL;
	int 			i;
	int			bad_cells=0;
	unsigned char		hec;
	int			rc;

	if (!atm_device_ptr || !rx_cell_block){
		return -2;
	}

	if (!atm_device->card){
		wpabs_debug_event("%s:%d: Critical Error: atm_device not initialized\n",
				__FUNCTION__,__LINE__);
		return -2;
	}

	atm_flip_bitstrm(wpabs_skb_data(rx_cell_block),
			 wpabs_skb_len(rx_cell_block));

	atm_cell = (wp_atm_cell_t *)wpabs_skb_data(rx_cell_block);
	num_of_cells=wpabs_skb_len(rx_cell_block)/ATM_CELL_SIZE;
	
	for (i=0; i<num_of_cells; i++){
	
		rc = wpabs_tracing_enabled(atm_device->trace_info);
		//wpabs_debug_event("rc: 0x%X\n", rc);

		if (is_idle_cell((unsigned int*)atm_cell)){

			if(rc == 3){
				/* Capture idle cell for ATM trace */
				if(wpabs_get_last_trace_direction(atm_device->trace_info) == TRC_OUTGOING_FRM ){

					capture_atm_trace_packet(atm_device,	
								(void*)atm_cell,
								 sizeof(wp_atm_cell_t),
								 TRC_INCOMING_FRM);

					wpabs_set_last_trace_direction(atm_device->trace_info, TRC_INCOMING_FRM);
				}
			}
			
			/* Discard idle cells */
			goto atm_cell_failure;	
		}

		if (rc == 2 || rc == 3){
			
			/* Capture cell for ATM Cell trace */
			capture_atm_trace_packet(atm_device,	
						 (void*)atm_cell,
						 sizeof(wp_atm_cell_t),
						 TRC_INCOMING_FRM);
		}

#if 0
		DO Not try to alter the ATM header since
		we are re-calculating the hec to determine
		if the packet has correct ATM header. If the
		ATM header is changed the HEC value will not
		match!
		
		/* Clear the data congestion bit , we don't care */
		if (wpabs_test_bit(PT_CONGESTION,(void*)&atm_cell->hdr[3])){
			wpabs_clear_bit(PT_CONGESTION,(void*)&atm_cell->hdr[3]);
		}
			
		/* Clear the cell loss priority , we dont care */
		if (wpabs_test_bit(CLP_BIT,(void*)&atm_cell->hdr[3])){
			wpabs_clear_bit(CLP_BIT,(void*)&atm_cell->hdr[3]);
		}
#endif

		pAdapter = wp_find_sar_dev_by_vpivci(atm_device,*(unsigned int *)atm_cell);
		if (pAdapter == NULL){

			hec=calculate_hec_crc((unsigned char *)atm_cell);
			
			if (hec != ((unsigned char*)atm_cell)[4]){

				if (++bad_cells > MAX_BAD_CELLS){
					return -1;
				}

				if (wpabs_net_ratelimit()){
					wpabs_debug_event("%s: Rx ATM SAR: Cell with invalid HEC=0x%x: 0x%x 0x%x 0x%x 0x%x 0x%x\n",
				        	    atm_device->devname, 
						    hec,
						    ((unsigned char *)atm_cell)[0],
						    ((unsigned char *)atm_cell)[1],
						    ((unsigned char *)atm_cell)[2],
						    ((unsigned char *)atm_cell)[3],
						    ((unsigned char *)atm_cell)[4]);
				}
				goto atm_cell_failure;
			}

			if (wpabs_net_ratelimit()){
				wpabs_debug_event("%s: Rx ATM SAR: Cell with unconfig VPI=%i VCI=%i : 0x%08X\n",
					            atm_device->devname, WPVPI(atm_cell), WPVCI(atm_cell),
						    *(unsigned int*)atm_cell);
			}

			goto atm_cell_failure;
		}

		wpabs_debug_rx("%s: Found Device %s for VPI=%d VCI=%d\n",
				atm_device->devname,pAdapter->devname,WPVPI(atm_cell),WPVCI(atm_cell));
		
		if (!pAdapter->rx_pkt_skb){
			pAdapter->rx_pkt_skb = wpabs_skb_alloc(pAdapter->max_sdu_size+ATM_CELL_SIZE);
			if (!pAdapter->rx_pkt_skb){
				wpabs_debug_event("%s: Rx ATM SAR: Critical error No Memory!\n");
				goto atm_cell_failure;
			}
		}

		rx_skb=pAdapter->rx_pkt_skb;

			
		/* Test if the incoming cell is an ATM management packet.
		 * If it is run it through the ATM management handler */
		if (wpabs_test_bit(PT_ADMIN,(void*)&atm_cell->hdr[3])){
			
			rx_skb=handle_atm_admin_packet(pAdapter,(wp_atm_cell_t*)atm_cell);
			if (!rx_skb){
				goto atm_cell_failure;
			}	

			if (wpabs_skb_queue_len(atm_device->tx_prot_list) > MAX_TX_PROT_LIST){
				wpabs_debug_event("%s: Error tx prot queue full, tx pkt dropped!\n",
						atm_device->devname);
				wpabs_skb_free(rx_skb);
				goto atm_cell_failure;	
			}

			if (wpabs_tracing_enabled(pAdapter->atmdevice->trace_info) == 2 ||
			    wpabs_tracing_enabled(pAdapter->atmdevice->trace_info) == 3){
				/* Capture cell for ATM Cell trace */
				capture_atm_trace_packet(pAdapter->atmdevice,	
					 wpabs_skb_data(rx_skb),
					 wpabs_skb_len(rx_skb),
					 TRC_OUTGOING_FRM);
			}

			atm_flip_bitstrm(wpabs_skb_data(rx_skb),
					 wpabs_skb_len(rx_skb));

			wpabs_skb_set_dev(rx_skb,pAdapter->priv);
			wpabs_skb_queue_tail(atm_device->tx_prot_list,rx_skb);

			goto atm_cell_failure;
		}
		
		/* PT bit not set: Start of the frame.
		 * PT bit set: Last data fragment */
		if (!wpabs_test_bit(PT_SDU1,(void*)&atm_cell->hdr[3])){

			wpabs_debug_rx("%s: RX: New frame\n", 
						pAdapter->devname);

			/* FIXME: No auto vpi/vci config */
			if (wpabs_skb_len(rx_skb) && pAdapter->prot_header_size){
				if (wpabs_memcmp((void*)atm_cell->info,
						 pAdapter->prot_header,
						 pAdapter->prot_header_size)==0){
				wpabs_debug_event("%s: ATM Error: SDU0 PDU(Len=%i) rx cell with PDU header!\n",
						pAdapter->devname,
						wpabs_skb_len(rx_skb));
				}
			}
				
			if (wpabs_skb_len(rx_skb) + 48 > pAdapter->max_sdu_size){
				wpabs_debug_event("%s: ATM Error: SDU0 Frame too big %i > %i !\n",
						pAdapter->devname,
						wpabs_skb_len(rx_skb) + 48,
						pAdapter->max_sdu_size);
				wpabs_skb_init(rx_skb, 16);
				wpabs_skb_trim(rx_skb, 0);
				++pAdapter->atm_stats.rx_atm_pdu_size;
				goto atm_cell_failure;
			}

			data_ptr=(unsigned char*)atm_cell->info;
			wpabs_skb_copyback(rx_skb, 
					 wpabs_skb_len(rx_skb), 
					 48, 
					 (unsigned long)data_ptr);
			++pAdapter->atm_stats.rx_valid;
		}else{
			wpabs_debug_rx("%s: RX: Last data fragment\n", 
					pAdapter->devname);

			info_len = *((unsigned short*)&atm_cell->info[42]);
			info_len = wpabs_htons(info_len);
			
			if (wpabs_skb_len(rx_skb) + 48 > pAdapter->max_sdu_size){

				wpabs_debug_event("%s: ATM Error: SDU1 Frame too big %i > %i !\n",
						pAdapter->devname,
						wpabs_skb_len(rx_skb) + 48,
						pAdapter->max_sdu_size);

				wpabs_skb_init(rx_skb, 16);
				wpabs_skb_trim(rx_skb, 0);
				++pAdapter->atm_stats.rx_atm_pdu_size;
				goto atm_cell_failure;
			}

			data_ptr=(unsigned char*)&atm_cell->info[0];
			wpabs_skb_copyback(rx_skb, 
					 wpabs_skb_len(rx_skb), 
					 48, 
					 (unsigned long)data_ptr);
			
			if (wpabs_skb_len(rx_skb) > info_len){
				wpabs_skb_trim(rx_skb, info_len);
			}

			if (wpabs_tracing_enabled(atm_device->trace_info) >= 0){
				capture_trace_packet(
						pAdapter,	
						rx_skb,
						TRC_INCOMING_FRM);
			}

			if (parse_prot_header(pAdapter, rx_skb)){
				wpabs_skb_init(rx_skb, 16);
				wpabs_skb_trim(rx_skb, 0);
				++pAdapter->atm_stats.rx_invalid_atm_hdr;
				goto atm_cell_failure;
			}
		
			++pAdapter->atm_stats.rx_valid;
			wpabs_skb_pull(rx_skb, pAdapter->prot_header_size);

			if (wpabs_skb_queue_len(atm_device->rx_data_list) > MAX_RX_DATA_LIST){
				wpabs_debug_event("%s: Error rx queue full, rx pkt dropped!\n",
						atm_device->devname);
				wpabs_skb_init(rx_skb, 16);
				wpabs_skb_trim(rx_skb, 0);
				goto atm_cell_failure;	
			}
		
			wpabs_skb_set_dev(rx_skb,pAdapter->priv);
			wpabs_skb_queue_tail(atm_device->rx_data_list,rx_skb);

			pAdapter->rx_pkt_skb=NULL;
			rx_skb=NULL;
		}
atm_cell_failure:
		atm_cell++;

	}/* for */

	return 0;
#undef atm_device
}

/*==================================================================
 * wanpipe_sar_tx
 *
 * 	Main Tx SAR for each configured PVC
 * 	o The tx_skb contains upper layer packet
 * 	o tx_cells_sbk is the user buffer where we
 * 	  should store ATM cells.
 *
 */ 
int wanpipe_sar_tx (void *pAdapter_ptr, void* tx_skb, void *tx_cells_skb)
{
#define pAdapter     ((wp_sar_dev_t *)pAdapter_ptr)

	unsigned char *src_data_ptr;
	unsigned int  src_data_size, src_payload_size;
	unsigned char *target_data_ptr;
	unsigned int  offset=0,pad=0;
	unsigned long crc = INITIAL_FCS_VALUE;
	unsigned long fin_crc;
	int i;

	if (wpabs_test_and_set_bit(SAR_TX,&pAdapter->init)){
		wpabs_debug_event("%s: Warning: Critical in sar tx\n",pAdapter->devname);
		return -1;
	}
	
	/* If the incoming packet has length of zero 
	 * stop the tx procedure */
	if (wpabs_skb_len(tx_skb) == 0){
		wpabs_clear_bit(SAR_TX,&pAdapter->init);
		return -1;
	}
	
	src_data_ptr    = wpabs_skb_data(tx_skb);
        src_data_size   = wpabs_skb_len(tx_skb);

	if (src_data_ptr == NULL || src_data_size == 0){
		wpabs_debug_event("%s: src_data_ptr is NULL or src_data_size is 0!\n", 
						pAdapter->devname);
		wpabs_clear_bit(SAR_TX,&pAdapter->init);
		return -1;
	}

	/* Initialize the Tx Cell buffer */
	wpabs_skb_init(tx_cells_skb, 16);
	wpabs_skb_trim(tx_cells_skb, 0);

	if (wpabs_skb_tailroom(tx_cells_skb) < pAdapter->max_sdu_size){
		wpabs_debug_event("%s: Error Tx cell buffer is too small %i < SAR MTU=%i\n", 
					pAdapter->devname,
					wpabs_skb_tailroom(tx_cells_skb),
					pAdapter->max_sdu_size);
		wpabs_clear_bit(SAR_TX,&pAdapter->init);
		return -1;
	}

	/* If the incoming packet is an ATMARP, we must change the
	 * default IPoATM PDU header to ATMARP header, just for
	 * this pakcet.  This value will be reset after the SAR
	 * is finished with this packet */
	if (pAdapter->EncapMode==RFC_MODE_ROUTED_IP_LLC &&
	    *(unsigned short*)&src_data_ptr[0]==ATMARP_HEADER){
		pAdapter->prot_header[7]=0x06;
	}

	target_data_ptr=wpabs_skb_put(tx_cells_skb,pAdapter->max_sdu_size);

	/* Initialize SAR variables used to cut up the tx packet
	 * into ATM Cells */
	offset=0;
	src_payload_size = src_data_size + pAdapter->prot_header_size;
	*((unsigned short*)&pAdapter->trailer[2])=wpabs_htons(src_payload_size);
	*((unsigned short*)&pAdapter->pad_len_buf[47])=0x00;
	*((unsigned short*)&pAdapter->pad_len_buf[49])=wpabs_htons(src_payload_size);
	

	/* Begining of the SAR. 
	 * Start cutting up the tx frame into ATM 
	 * Cells.  
	 *
	 * Note: For S518 Boards
	 * 	 
	 * 	 Do not calculate the HEC here. The HEC is
	 *       done in hardware. Thus we only have to 
	 *       deal with 52 bytes frames 
	 *
	 *       For S514 Boards
	 *
	 *       HEC is calculated once on setup. Our ATM header
	 *       does not change after that.
	 */
	for (i=0;;){

		/* Each ATM cell starts with a 4 byte (i.e. HEC not included)
		 * ATM header which contains VPI/VCI/PT/CLP values */
		wpabs_memcpy((target_data_ptr+offset),
		             pAdapter->atm_header,
			     sizeof(pAdapter->atm_header));

		offset+=sizeof(pAdapter->atm_header);
	
		/* Check that a tx packet can fit into a single
		 * ATM cell which can accommodate 48 bytes of data */
		if ((i+48) <= src_payload_size){
		
			/* The packet cannot fit into a single cell */
			
			/* Check if this is a first cell of an ATM PDU */
			if (i==0){
				/* If this is a first cell of the ATM PDU
				 * we must append an ATM PDU header. ATM
				 * PDU header defines the above protocol
				 * that is being transported by the ATM
				 * network */
				wpabs_memcpy((target_data_ptr+offset),
					pAdapter->prot_header,
					pAdapter->prot_header_size);
		
				offset+=pAdapter->prot_header_size;
			
				wpabs_memcpy((target_data_ptr+offset),
				        src_data_ptr,
				        (48-pAdapter->prot_header_size)); 
				src_data_ptr += (48-pAdapter->prot_header_size);
				offset += (48-pAdapter->prot_header_size);

			}else{
				/* Not a first cell, thus, just fill the 48 bytes
				 * of data and move to another cell */
				wpabs_memcpy((target_data_ptr+offset),src_data_ptr,48);
				src_data_ptr += 48;
				offset += 48;
			}
			++pAdapter->atm_stats.tx_valid;
		}else{
			
			/* Tx packet can fit into a single cell or 
			 * we have reached the last cell needed to
			 * tx the ATM PDU */
			
			++pAdapter->atm_stats.tx_valid;
			
			if (i==0){
				/* This is the first cell, which means
				 * that the ATM PDU can fit into a
				 * single cell i.e. less than 48 bytes.
				 *
				 * The Payload is smaller than 48 bytes,
				 * there are two scenarios here:
				 * 
				 * 1) Payload + Trailer = PDU fit into a
				 *    single cell. i.e <= 48
				 *    
				 * 2) Payload + Trailer > PDU do not fit
				 *    into a single cell i.e. > 48 thus 
				 *    we must create and pad the second cell
				 */
				
				/* Creating a first cell */
				wpabs_memcpy((target_data_ptr+offset),
					pAdapter->prot_header,
				        pAdapter->prot_header_size);
				offset+=pAdapter->prot_header_size;
			
				wpabs_memcpy((target_data_ptr+offset),
				       src_data_ptr,
				       src_data_size); 
				src_data_ptr += src_data_size;
				offset += src_data_size;
			
			}else{
				/* Copy the last part of the tx frame
				 * into the ATM CELL */
				
				if ((src_payload_size - i) > 0){
					wpabs_memcpy((target_data_ptr+offset),
					       src_data_ptr,
					       (src_payload_size-i)); 
					src_data_ptr+=(src_payload_size-i);
					offset	   +=(src_payload_size-i);
				}
			}

			/* Since this is the last ATM CELL we must append a
			 * ATM PDU trailer to the CELL.  Check if the trailer
			 * will fit into the CELL.  If not we must create
			 * another cell and pad it */
			
			if ((src_payload_size+sizeof(pAdapter->trailer)) > (i+48)){
				
				/* We cannot fit the whole PDU frame
				 * into this cell. We need add another cell. */
				
				/* Pad to the end of the first cell because
				 * the trailer must reside at the end of the
				 * ATM CELL. Thus we'll put it at the end of
				 * the next cell */
				pad=((i+48)-src_payload_size);
				if (pad > 0){
					wpabs_memset((target_data_ptr+offset),PAD_CHAR,pad);
					offset+=pad;
				}

				/* Start the atm header on second cell */
				wpabs_memcpy((target_data_ptr+offset),
					pAdapter->atm_header,
					sizeof(pAdapter->atm_header));

				offset+=sizeof(pAdapter->atm_header);
			
				/* Set the END of FRAME bit in the PT section of
				 * the ATM header, this will indicate to the
				 * remote side that this CELL is the last */
				wpabs_set_bit(1,&((target_data_ptr+offset-2)[0]));

				(target_data_ptr+offset-2)[1]=pAdapter->hec_eof;

				/* Pad the second cell until trailer */
				wpabs_memset((target_data_ptr+offset),PAD_CHAR,40);
				offset+=40;
				pad+=40;
				
				/* Insert trailer as last 8 bytes. The
				 * trailer contains dummy data right now
				 * because we still have to calculate the CRC
				 * and the overall packet length */
				wpabs_memcpy((target_data_ptr+offset),
						pAdapter->trailer,
						sizeof(pAdapter->trailer));
				
				offset+=sizeof(pAdapter->trailer);
			}else{
				/* The whole PDU fits into this frame */
				
				/* PAD all the unused bytes of the CELL that reside
				 * between ATM PDU data and the trailer */
				pad=((i+48)-(src_payload_size+sizeof(pAdapter->trailer)));
				if (pad > 0){
					wpabs_memset((target_data_ptr+offset),PAD_CHAR,pad);
					offset+=pad;
				}
								
				/* Insert trailer as last 8 bytes. The
				 * trailer contains dummy data right now
				 * because we still have to calculate the CRC
				 * and the overall packet length */
				wpabs_memcpy((target_data_ptr+offset),
						pAdapter->trailer,
						sizeof(pAdapter->trailer));
				
				wpabs_memcpy((target_data_ptr+offset+4),&crc,sizeof(long));
				offset+=sizeof(pAdapter->trailer);

				/* Set the END of FRAME bit in the PT section of
				 * the ATM header of this CELL, 
				 * this will indicate to the remote side 
				 * that this CELL is the last */
				wpabs_set_bit(1,&((target_data_ptr+offset-50)[0]));

				(target_data_ptr+offset-50)[1]=pAdapter->hec_eof;
			}
			break;
		}
		i = i+48;
	}/* for */
	
	/* End of SAR:
	 * The tx packet / ATM PDU packet, has been cut up and all 
	 * ATM CELLS are located in target_data_ptr (Tx dma buffer). 
	 * We must now calculate the CRC and LENGTH of the 
	 * ATM PDU and store those values into the trailer of 
	 * the last ATM CELL */
	
	/* Set the size of the tx DMA buffer that contains all
	 * the ATM CELLS. The offset will be equal to 52*number
	 * of cells created */
	wpabs_skb_trim(tx_cells_skb,offset);

	/* Calculate the CRC of the ATM-PDU. 
	 * Starting with ATM-PDU protocol header, tx packet 
	 * and any pad bytes that were inserted to fill in 
	 * the ATM CELL up to the trailer */
	crc=INITIAL_FCS_VALUE;
	crc=update_crc(crc, (char*)pAdapter->prot_header, pAdapter->prot_header_size);
	crc=update_crc(crc, (char*)wpabs_skb_data(tx_skb), wpabs_skb_len(tx_skb));
	crc=update_crc(crc,
		       (char*)(pAdapter->pad_len_buf+(sizeof(pAdapter->pad_len_buf)-(pad+4))),
		       (pad+4));
	crc=~crc;
	FLIP_CRC(crc,fin_crc);

	/* Add the calculated CRC value into the trailer of the
	 * last ATM CELL */
	wpabs_memcpy((target_data_ptr+offset-4),&fin_crc,sizeof(long));

	/* If tracing is enabled store the packet for tracing */
	if (wpabs_tracing_enabled(pAdapter->atmdevice->trace_info) >= 0){
		capture_trace_packet(pAdapter,	
				     tx_skb,
				     TRC_OUTGOING_FRM);
	
	}
	
	if (pAdapter->EncapMode==RFC_MODE_ROUTED_IP_LLC &&
	    *((unsigned short *)wpabs_skb_data(tx_skb)) == ATMARP_HEADER){
		pAdapter->prot_header[7]=0x00;
	}

	if (wpabs_tracing_enabled(pAdapter->atmdevice->trace_info) == 2 ||
	    wpabs_tracing_enabled(pAdapter->atmdevice->trace_info) == 3){
		/* Capture cell for ATM Cell trace */
		capture_atm_trace_packet(pAdapter->atmdevice,	
					 wpabs_skb_data(tx_cells_skb),
					 wpabs_skb_len(tx_cells_skb),
					 TRC_OUTGOING_FRM);
	}

	atm_flip_bitstrm(wpabs_skb_data(tx_cells_skb),
			 wpabs_skb_len(tx_cells_skb));

	wpabs_clear_bit(SAR_TX,&pAdapter->init);
	return 0;

#undef pAdapter
}

/*==================================================================
 * wanpipe_sar_poll
 *
 * 	ATM Polling routine	
 * 	o Tx ATM ARP Request when configured for Classical IP
 * 	o Send OAM loopback request for ATM level debugging.
 */ 

int wanpipe_sar_poll (void *pAdapter_ptr, int timeout)
{	
#define pAdapter     ((wp_sar_dev_t *)pAdapter_ptr)
	int tx_available=0;

	pAdapter->timeout+=timeout;

	if  (pAdapter->state == ATM_AIS){
		if ((wpabs_get_systemticks()-pAdapter->ais_timeout) > 3*wpabs_get_hz()){
			wpabs_debug_event("%s:%s: ATM AIS timeout condition: no AIS in 3 sec\n",
					pAdapter->atmdevice->devname,
					pAdapter->devname);
			set_atm_sar_state(pAdapter,ATM_CONNECTED);
		}else{
			sar_tx_rdi_req(pAdapter->atmdevice,pAdapter);
		}
	}
	
	if (pAdapter->cfg.atm_oam_continuity){	
		if (!(pAdapter->timeout%pAdapter->cfg.atm_oam_continuity_intr)){
			sar_tx_continuity_req(pAdapter->atmdevice,pAdapter);
			tx_available=1;
		}
	}

	if (pAdapter->EncapMode == RFC_MODE_ROUTED_IP_LLC){
		if (pAdapter->cfg.atm_arp){
			if (!(pAdapter->timeout%pAdapter->cfg.atm_arp_intr)){
				sar_tx_atm_arp(pAdapter->atmdevice,pAdapter);	
				tx_available=1;
			}
		}
	}

	if (pAdapter->cfg.atm_oam_loopback){
		if (!(pAdapter->timeout%pAdapter->cfg.atm_oam_loopback_intr)){
			sar_tx_oam_loopback_req(pAdapter->atmdevice,pAdapter);
			tx_available=1;
		}
	}

	return tx_available;
#undef pAdapter
}

/*==================================================================
 * wanpipe_get_atm_state
 *
 */ 

int wanpipe_get_atm_state (void *pAdapter_ptr)
{
#define pAdapter     ((wp_sar_dev_t *)pAdapter_ptr)
	if (!pAdapter){
		return ATM_DISCONNECTED;
	}

	return pAdapter->state;	
#undef pAdapter
}

/*==================================================================
 * wanpipe_set_atm_state
 *
 */ 
void wanpipe_set_atm_state (void *pAdapter_ptr, int state)
{
#define pAdapter     ((wp_sar_dev_t *)pAdapter_ptr)
	if (!pAdapter){
		return;
	}
	set_atm_sar_state(pAdapter_ptr,state);
#undef pAdapter
}



/***SECTOIN**********************************************************
 *
 *  ATM Private Functions
 *  
 ********************************************************************/


/*==================================================================
 * parse_prot_header
 *
 */ 
static int parse_prot_header(wp_sar_dev_t *pAdapter, void* rx_skb)
{
	unsigned char* data = (unsigned char*)wpabs_skb_data(rx_skb);
	int len = wpabs_skb_len(rx_skb);
	int err = 0;
	
	switch (pAdapter->EncapMode){
			
		case RFC_MODE_ROUTED_IP_LLC:
			if (data[6] == 0x08){
				if (data[7] == 0x06){
					wpabs_debug_rx("%s:%s: RX ARP packet, sending reply...\n",
								pAdapter->atmdevice->devname,
								pAdapter->devname);
					
					handle_atm_arp(pAdapter, rx_skb);	
					
					/* We don't want to send this rx packet 
					 * up the stack, since we just replied to the
					 * arp. We should kill this packet thus, we return -1 */
					return -1;
				}
			}else{
				char temp[200];
				err = wpabs_detect_prot_header(data, len, temp);

				if (err==0){
					wpabs_debug_event("%s:%s: Rx Pkt with invalid protocol header!\n",
							pAdapter->atmdevice->devname,
							pAdapter->devname);
					wpabs_debug_event("%s:%s: Not LLC IP/ATM: Detected as: %s!\n",
							pAdapter->atmdevice->devname,
							pAdapter->devname, 
							temp);
				}
				return -1;
			}
			break;

		case RFC_MODE_ROUTED_IP_VC:
			if (data[0] != 0x45){
				char temp[200];
				err=wpabs_detect_prot_header(data, len, temp);
				if (err==0){
					wpabs_debug_event("%s:%s: Rx Pkt with invalid protocol header!\n",
							pAdapter->atmdevice->devname,	
							pAdapter->devname);
					wpabs_debug_event("%s:%s: Not VC IP/ATM: Detected as: %s!\n",
							pAdapter->atmdevice->devname,
							pAdapter->devname, 
							temp);
				}
				return -1;
			}
			break;
			
		case RFC_MODE_BRIDGED_ETH_LLC:
			if (data[6] != 0x00 || data[7] != 0x07){
				char temp[200];
				err=wpabs_detect_prot_header(data, len, temp);
				if (err==0){
					wpabs_debug_event("%s:%s: Rx Pkt with invalid protocol header!\n",
							pAdapter->atmdevice->devname,
							pAdapter->devname);
					wpabs_debug_event("%s:%s: Not Bridged LLC/ATM: Detected as: %s\n",
							pAdapter->atmdevice->devname,
							pAdapter->devname,
							temp);
				}
				return -1;
			}
			break;

		case RFC_MODE_BRIDGED_ETH_VC:
			{
				int err = wpabs_detect_prot_header(data, len, NULL);
				if (err == 0){
					char temp[200];
					err = wpabs_detect_prot_header(data, len, temp);
					if (err==0){
						wpabs_debug_event("%s:%s: Rx Pkt with invalid protocol header!\n",
							pAdapter->atmdevice->devname,
							pAdapter->devname);
						wpabs_debug_event("%s:%s: Not Bridged VC/ATM: Detected as: %s\n",
							pAdapter->atmdevice->devname,
							pAdapter->devname,
							temp);
					}
				}
			}
			break;

		case RFC_MODE_PPP_LLC:

			if (data[0] != 0xFE || data[1] != 0xFE){
				char temp[200];
				err=wpabs_detect_prot_header(data, len, temp);
				if (err==0){
					wpabs_debug_event("%s:%s: Rx Pkt with invalid protocol header!\n",
						pAdapter->atmdevice->devname,
						pAdapter->devname);
					wpabs_debug_event("%s:%s: Not LLC PPP/ATM: Detected as: %s!\n",
						pAdapter->atmdevice->devname,
						pAdapter->devname, temp);
				}
				return -1;
			}
			break;

		case RFC_MODE_PPP_VC:
			if (data[0] == 0xC0 && data[1] == 0x21){
				return 0;
			}else if (data[0] == 0x00 && data[1] == 0x21){
				return 0;
			}else{
				err = wpabs_detect_prot_header(data, len, NULL);
				if (err == 0){
					char temp[200];
					err = wpabs_detect_prot_header(data, len, temp);
					if (err==0){
						wpabs_debug_event("%s:%s: Rx Pkt with invalid protocol header!\n",
							pAdapter->atmdevice->devname,
							pAdapter->devname);
						wpabs_debug_event("%s:%s: Not VC PPP/ATM: Detected as: %s\n",
							pAdapter->atmdevice->devname,
							pAdapter->devname,
							temp);
					}
				}
			}
			break;

		default:
			/* This will never happen, just for completeness sake */
			wpabs_debug_event("%s:%s: Rx Configuration error, unkown mode 0x%x\n",
					pAdapter->atmdevice->devname,
					pAdapter->devname,pAdapter->EncapMode);
			return -1;
	}
	
	return 0;
}


/*==================================================================
 * handle_atm_arp
 *
 */ 

static int handle_atm_arp(wp_sar_dev_t *pAdapter, void* rx_skb)
{		
	void*		tx_skb = NULL;
	char*		rx_data = (char*)wpabs_skb_data(rx_skb);
	atmarp_t*	atmarp = NULL;	
	atmarp_t	atmarp_reply;
	unsigned long 	local_ip;
	void *		tx_cell_skb;
	int 		err;

	atmarp = (atmarp_t*)&rx_data[8];	
	if (atmarp->arphrd != ATMARP_HEADER){
		wpabs_debug_event("%s:%s: Rx invalid arp pkt header 0x%X!\n",
				pAdapter->atmdevice->devname,
				pAdapter->devname,atmarp->arphrd);
		wpabs_debug_event("0x%08X  0x%08X 0x%08X 0x%08X\n",
				*((unsigned long*)rx_data),
				*((unsigned long*)rx_data+1),
				*((unsigned long*)rx_data+2),
				*((unsigned long*)rx_data+3));
		return -1;
	}

	if (atmarp->arppro != ATMARP_PROT_IP){
		wpabs_debug_event("%s:%s: Rx invalid arp pkt: protocol 0x%X not supported!\n",
				pAdapter->atmdevice->devname,
				pAdapter->devname,atmarp->arppro);
		return -1;
	}

	switch (atmarp->arpop){

	case InARP_REQUEST:

		if (atmarp->arpshtl > sizeof(pAdapter->atm_addr)){
			wpabs_debug_event("%s:%s: Rx invalid arp pkt: Source ATM address too big %d\n",
						pAdapter->atmdevice->devname,
						pAdapter->devname,atmarp->arpshtl);
			return -1;
		}
		if (atmarp->arpsstl > sizeof(pAdapter->atm_addr)){
			wpabs_debug_event("%s:%s: Rx invalid arp pkt: Source Sub ATM address too big %d\n",
						pAdapter->atmdevice->devname,
						pAdapter->devname,atmarp->arpsstl);
			return -1;
		}

		if (atmarp->arpspln > 4){
			wpabs_debug_event("%s:%s: Rx invalid arp pkt: Source IP ATM address too long %d\n",
						pAdapter->atmdevice->devname,
						pAdapter->devname,atmarp->arpspln);
			return -1;
		}

		if (wpabs_test_and_set_bit(ATM_ARP,&pAdapter->init)){
			wpabs_debug_event("%s:%s: Critical in ATM ARP\n",
					pAdapter->atmdevice->devname,
					pAdapter->devname);
			return -1;	
		}

		tx_skb = pAdapter->tx_arp_skb;
		if (!tx_skb){
			wpabs_clear_bit(ATM_ARP,&pAdapter->init);
			wpabs_debug_event("%s:%s: Error: no tx arp skb\n",
					pAdapter->atmdevice->devname,
					pAdapter->devname);
			return -1;
		}
	
		wpabs_skb_init(tx_skb,16);
		wpabs_skb_trim(tx_skb,0);
		
		atmarp_reply.arphrd 	= ATMARP_HEADER;
		atmarp_reply.arppro 	= ATMARP_PROT_IP;
		atmarp_reply.arpshtl	= sizeof(pAdapter->atm_addr);
		atmarp_reply.arpsstl	= 0;
		atmarp_reply.arpop  	= InARP_REPLY;
		atmarp_reply.arpspln	= 4;
		atmarp_reply.arpthtl	= atmarp->arpshtl;
		atmarp_reply.arptstl	= atmarp->arpsstl;
		atmarp_reply.arptpln	= atmarp->arpspln;
		wpabs_skb_copyback(tx_skb, 0, sizeof(atmarp_t), (unsigned long)&atmarp_reply);

		local_ip = wpabs_get_ip_addr(pAdapter->dev, WAN_LOCAL_IP);
		wpabs_debug_tx("%s:%s: Source ATM Len=%d, Sub Len=%d, IP Len=%d, %lx\n",
				pAdapter->atmdevice->devname,
				pAdapter->devname,atmarp->arpshtl, 
				atmarp->arpsstl, atmarp->arpspln, local_ip);

		wpabs_skb_copyback(tx_skb, 
				 wpabs_skb_len(tx_skb), 
				 sizeof(pAdapter->atm_addr),
				 (unsigned long)pAdapter->atm_addr);

		wpabs_skb_copyback(tx_skb, 
				 wpabs_skb_len(tx_skb), 
				 sizeof(local_ip),
				 (unsigned long)&local_ip);

		if (atmarp->arpshtl){
			wpabs_skb_copyback(tx_skb, 
				 wpabs_skb_len(tx_skb), 
				 atmarp->arpshtl,
				 (unsigned long)(&rx_data[8]+sizeof(atmarp_t)));
		}

		if (atmarp->arpsstl){
			wpabs_skb_copyback(tx_skb, 
				 wpabs_skb_len(tx_skb), 
				 atmarp->arpsstl,
				 (unsigned long)(&rx_data[8]+sizeof(atmarp_t)+atmarp->arpshtl));
		}

		if (atmarp->arpspln){
			wpabs_skb_copyback(tx_skb, 
				 wpabs_skb_len(tx_skb), 
				 atmarp->arpshtl,
				 (unsigned long)(&rx_data[8]+sizeof(atmarp_t)+atmarp->arpshtl+atmarp->arpsstl));
		}

		/* Pad the outgoing ARP to 78 bytes */
		if (wpabs_skb_len(tx_skb) < 78){
			char* tx_data = (char*)wpabs_skb_put(tx_skb,78 - wpabs_skb_len(tx_skb));
			wpabs_memset(tx_data, 0, 78 - wpabs_skb_len(tx_skb));
		}	

		tx_cell_skb=wpabs_skb_alloc(pAdapter->max_sdu_size+ATM_CELL_SIZE);	
		if (!tx_cell_skb){
			wpabs_clear_bit(ATM_ARP,&pAdapter->init);
			WPATM_MEM_ASSERT();
			return -1;
		}

		err=wanpipe_sar_tx(pAdapter, tx_skb, tx_cell_skb);
		if (err){
			wpabs_skb_free(tx_cell_skb);
			wpabs_clear_bit(ATM_ARP,&pAdapter->init);
			wpabs_debug_event("%s:%s: Error: sar tx failed on tx arp reply \n",
					pAdapter->atmdevice->devname,
					pAdapter->devname);
			return -1;
		}

		wpabs_clear_bit(ATM_ARP,&pAdapter->init);

		wpabs_skb_set_dev(tx_cell_skb,pAdapter->priv);
		wpabs_skb_queue_tail(pAdapter->atmdevice->tx_prot_list,tx_cell_skb);

		break;

	case InARP_REPLY:

		if (atmarp->arpspln != 0){
			unsigned long rmt_addr,remote_ip;
				
			rmt_addr= *(unsigned long*)(&rx_data[8]+sizeof(atmarp_t)+
				                 atmarp->arpshtl+atmarp->arpsstl);

			remote_ip = wpabs_get_ip_addr(pAdapter->dev, WAN_POINTOPOINT_IP);		

			if (rmt_addr != remote_ip){
				unsigned char ip1[25],ip2[25];
				wpabs_decode_ipaddr(rmt_addr,ip1);
				wpabs_decode_ipaddr(remote_ip,ip2);
				wpabs_debug_event("%s:%s: Error Rx InARP Reply !\n",
					pAdapter->atmdevice->devname,
					pAdapter->devname);
				wpabs_debug_event("%s:%s:    Local P-to-P %s != Remote Src %s\n",
					pAdapter->atmdevice->devname,
					pAdapter->devname,ip2,ip1);
				/* FIXME: Change the remote ip address automatically */
			}else{
				wpabs_debug_rx("%s:%s: Rx InARP Reply OK\n",
					pAdapter->atmdevice->devname,
					pAdapter->devname);
			}
		}else{
			wpabs_debug_event("%s:%s: Error Rx InARP Reply: Remote Addr = 0.0.0.0\n",
					pAdapter->atmdevice->devname,
					pAdapter->devname);
		}
		
		break;
	case ARP_NAK:
		wpabs_debug_event("%s:%s: Rx ARP NAK\n",
					pAdapter->atmdevice->devname,
					pAdapter->devname);
		break;

	case ARP_REQUEST:
		wpabs_debug_rx("%s:%s: Rx ARP Request \n",
					pAdapter->atmdevice->devname,
					pAdapter->devname);
		break;
		
	case ARP_REPLY:
		wpabs_debug_rx("%s:%s: Rx ARP Reply \n",
					pAdapter->atmdevice->devname,
					pAdapter->devname);
		break;

	default:
		wpabs_debug_event("%s:%s: Rx Invalid arp pkt: Op Code 0x%x\n",
					pAdapter->atmdevice->devname,
					pAdapter->devname,atmarp->arpop);
		return -1;
	}

	return 0;
}

/*==================================================================
 * create_inarp_request
 *
 */ 

static void* create_inarp_request(wp_sar_dev_t *pAdapter)
{
	void*		tx_skb = NULL;
	atmarp_t	atmarp;	
	unsigned long 	local_ip;
	unsigned int 	len;
	unsigned char*  tx_data;
	
	
	tx_skb = wpabs_skb_alloc(MAX_ARP_SIZE);
	if (!tx_skb){
		return NULL;
	}
	
	atmarp.arphrd 	= ATMARP_HEADER;
	atmarp.arppro 	= ATMARP_PROT_IP;
	atmarp.arpshtl	= sizeof(pAdapter->atm_addr);
	atmarp.arpsstl	= 0;
	atmarp.arpop  	= InARP_REQUEST;
	atmarp.arpspln	= 4;
	atmarp.arpthtl	= 0;
	atmarp.arptstl	= 0;
	atmarp.arptpln	= 0;
	
	wpabs_skb_copyback(tx_skb, 0, sizeof(atmarp_t), (unsigned long)&atmarp);

	local_ip = wpabs_get_ip_addr(pAdapter->dev, WAN_LOCAL_IP);

	wpabs_skb_copyback(tx_skb, 
			 wpabs_skb_len(tx_skb), 
			 sizeof(pAdapter->atm_addr),
			 (unsigned long)pAdapter->atm_addr);

	wpabs_skb_copyback(tx_skb, 
			 wpabs_skb_len(tx_skb), 
			 sizeof(local_ip),
			 (unsigned long)&local_ip);

	if ((len=wpabs_skb_len(tx_skb)) < 78){
		len=78-len;
		tx_data = wpabs_skb_put(tx_skb,len);
		wpabs_memset(tx_data, 0, len);
	}	

	return tx_skb;
}

/*===============================================================
 * sar_tx_atm_arp
 * 
 * Create an ATMARP Request and queue it for tx. 
 *
 * For Classical IP over ATM ONLY !!!
 */
static void sar_tx_atm_arp(wp_atm_reg_t *atm_device, wp_sar_dev_t *pAdapter)
{
	void *tx_cell_skb,*tx_skb;
	int err;
		
	tx_skb=create_inarp_request(pAdapter);
	if (!tx_skb){
		return;
	}

	tx_cell_skb=wpabs_skb_alloc(pAdapter->max_sdu_size+ATM_CELL_SIZE);	
	if (!tx_cell_skb){
		wpabs_skb_free(tx_skb);
		return;
	}

	err=wanpipe_sar_tx(pAdapter, tx_skb, tx_cell_skb);
	if (err){
		wpabs_skb_free(tx_cell_skb);
		wpabs_skb_free(tx_skb);
		return;
	}

	wpabs_debug_event("%s:%s: Sending InARP Request \n",
			pAdapter->atmdevice->devname,
			pAdapter->devname);

	wpabs_skb_set_dev(tx_cell_skb,pAdapter->priv);
	wpabs_skb_queue_tail(atm_device->tx_prot_list,tx_cell_skb);

	wpabs_skb_free(tx_skb);
	return; 
}



#define OAM_F5_END_TO_END_CELL 	0x05
#define OAM_F5_SEGMENT 		0x04

/* OAM Types */
#define OAM_FAULT_MGMT		0x01
#define OAM_PERF_MGMT		0x02
#define OAM_ACT_DEACT		0x80

/* OAM Function Types */
#define OAM_FAULT_FUNC_ALARM_INDIC	0x00
#define OAM_FAULT_FUNC_FERF		0x01
#define OAM_FAULT_FUNC_CELL_LOOPBACK	0x08
#define OAM_FAULT_FUNC_CONTINUITY	0x04

#define OAM_PERF_FUNC_FW_MONITOR	0x00
#define OAM_PERF_FUNC_BW_REPORT		0x01
#define OAM_PERF_FUNC_MON_REPORT	0x02

#define OAM_AD_FUNC_PERF_MONITOR	0x00
#define OAM_AD_FUNC_CONTINUITY		0x01

#define OAM_CELL_LOOPBACK 		0x18
#define OAM_LOOPBACK_INDICATION 	0x01
#define ATM_PAYLOAD_SZ 			48

/*==================================================================
 * handle_atm_admin_packet
 *
 */ 

void *handle_atm_admin_packet(wp_sar_dev_t *pAdapter, wp_atm_cell_t *atm_cell_ptr)
{
	unsigned char oam_type, oam_func;
	unsigned char pt = (atm_cell_ptr->hdr[3] >> 1) & 0x07;
	
	unsigned short vpi = WPVPI(atm_cell_ptr);
 	unsigned short vci = WPVCI(atm_cell_ptr);
	
	if (validate_mgmnt_cell((wp_atm_cell_1_t*)atm_cell_ptr) != 0){
		wpabs_debug_event("%s:%s: Rx invalid ATM Mgmt Cell: Bad CRC\n",
				pAdapter->atmdevice->devname,
				pAdapter->devname);
		return NULL;
	}
	
	if (pAdapter->vpi != vpi){
		wpabs_debug_event("%s:%s: Rx invalid ATM Mgmt Cell: Vpi Mismatch Cfg Vpi=%d In Vpi=%d\n",
				pAdapter->atmdevice->devname,
				pAdapter->devname,pAdapter->vpi,vpi);
		return NULL;
	}

	if ((pt == OAM_F5_END_TO_END_CELL || pt == OAM_F5_SEGMENT) && pAdapter->vci == vci){
		goto oam_f5_packet;
		
	}else if (pAdapter->vci == 3 || pAdapter->vci == 4){
		goto oam_f4_packet;

	}else{
		wpabs_debug_event("%s:%s: Rx invalid ATM Mgmt Cell: Vpi=%d Vci=%d Pt=0x%x\n",
				pAdapter->atmdevice->devname,
				pAdapter->devname,vpi,vci,pt);
		return NULL;
	}

oam_f4_packet:
oam_f5_packet:

	oam_type = (atm_cell_ptr->info[0] >> 4) & 0x0F;
	oam_func = atm_cell_ptr->info[0] & 0x0F;

	switch (oam_type){

	case OAM_FAULT_MGMT:

		switch (oam_func){
			
			case OAM_FAULT_FUNC_CELL_LOOPBACK:
				set_atm_sar_state(pAdapter,ATM_CONNECTED);
				return handle_oam_cell_loopback(pAdapter,atm_cell_ptr);
				
			case OAM_FAULT_FUNC_ALARM_INDIC:
				return handle_oam_alarm(pAdapter,atm_cell_ptr);
			
			case OAM_FAULT_FUNC_CONTINUITY:
				wpabs_debug_rx("%s:%s: Rx Continuity Check\n",
						pAdapter->atmdevice->devname,
						pAdapter->devname);
				set_atm_sar_state(pAdapter,ATM_CONNECTED);
				break;

			case OAM_FAULT_FUNC_FERF:	
				wpabs_debug_event("%s:%s: Rx RDI Indication\n",
						pAdapter->atmdevice->devname,
						pAdapter->devname);
				break;
				
			default:
				wpabs_debug_event("%s:%s: Rx Invalid ATM OAM Fault Mgmt Cell: Type=0x%x Func=0x%x\n",
						pAdapter->atmdevice->devname,
						pAdapter->devname,oam_type,oam_func);
				break;
		}
		break;	
	
	case OAM_PERF_MGMT:
		wpabs_debug_event("%s:%s: Rx unsupported ATM OAM Perf Mgmt Cell: Type=0x%x Func=0x%x\n",
						pAdapter->atmdevice->devname,
						pAdapter->devname,oam_type,oam_func);
		break;

	case OAM_ACT_DEACT:
		wpabs_debug_event("%s:%s: Rx unsupported ATM OAM Act/Deact Cell: Type=0x%x Func=0x%x\n",
						pAdapter->atmdevice->devname,
						pAdapter->devname,oam_type,oam_func);
		break;
		
	default:
		wpabs_debug_event("%s:%s: Rx unsupported ATM OAM Cell: Type=0x%x Func=0x%x\n",
						pAdapter->atmdevice->devname,
						pAdapter->devname,oam_type,oam_func);
		
		break;
	}

	return NULL;
}

/*==================================================================
 * handle_oam_cell_loopback
 *
 */ 

static void *handle_oam_cell_loopback(wp_sar_dev_t *pAdapter, wp_atm_cell_t *atm_cell_ptr)
{

	if (atm_cell_ptr->info[1] != OAM_LOOPBACK_INDICATION){
		wpabs_debug_rx("%s:%s: Rx OAM Loopback Reply\n",
				pAdapter->atmdevice->devname,
				pAdapter->devname);
		return NULL;
	}

	wpabs_debug_rx("%s:%s: Rx ATM OAM Loopback Request\n",
				pAdapter->atmdevice->devname,
				pAdapter->devname);

	
	/* Clear the OAM F5 Loopback request, thus
	 * indicating a reply */
	atm_cell_ptr->info[1]=0x00;

	wpabs_debug_tx("%s:%s: Tx ATM OAM Loopback Reply\n",
					pAdapter->atmdevice->devname,
					pAdapter->devname);

	return sar_tx_oam_packet (pAdapter->atmdevice,pAdapter,atm_cell_ptr->info,ATM_PAYLOAD_SZ,1);
}


static void *handle_oam_alarm (wp_sar_dev_t *pAdapter, wp_atm_cell_t *atm_cell_ptr)
{
	pAdapter->ais_timeout=wpabs_get_systemticks();
	wpabs_memcpy(&pAdapter->ais_data[1],&atm_cell_ptr->info[1],ATM_AIS_DATA_SZ-1);
	pAdapter->ais_data[0]=0x11;

	if (pAdapter->state != ATM_AIS){
		wpabs_debug_event("%s:%s: Rx ATM Alarm AIS Indication\n",
					pAdapter->atmdevice->devname,
					pAdapter->devname);

		set_atm_sar_state(pAdapter,ATM_AIS);
	}


	return NULL;
}

/*==================================================================
 * sar_tx_oam_loopback_req
 *
 */ 

static void sar_tx_oam_loopback_req(wp_atm_reg_t *atm_device, wp_sar_dev_t *pAdapter)
{
	wpabs_debug_tx("%s:%s: Tx OAM Loopback Req\n",
			pAdapter->atmdevice->devname,
			pAdapter->devname);

	sar_tx_oam_packet (atm_device,pAdapter,atm_oam_req,sizeof(atm_oam_req),0);
	return; 
}

static void sar_tx_continuity_req(wp_atm_reg_t *atm_device, wp_sar_dev_t *pAdapter)
{
	wpabs_debug_tx("%s:%s: Tx Continuity Check\n",
			pAdapter->atmdevice->devname,
			pAdapter->devname);

	sar_tx_oam_packet (atm_device,pAdapter,atm_cc_req,sizeof(atm_cc_req),0);
	return; 
}

static void sar_tx_rdi_req(wp_atm_reg_t *atm_device, wp_sar_dev_t *pAdapter)
{
	
#if 0
	wpabs_debug_event("%s:%s: Tx ATM RDI Indication\n",
					pAdapter->atmdevice->devname,
					pAdapter->devname);
#endif

	sar_tx_oam_packet (atm_device,pAdapter,pAdapter->ais_data,ATM_AIS_DATA_SZ,0);
	return; 
}

static void *sar_tx_oam_packet(wp_atm_reg_t *atm_device, wp_sar_dev_t *pAdapter, 
		               unsigned char *data, int len, unsigned char skb_return)
{
	void *tx_cell_skb;
	wp_atm_cell_1_t *tx_data_ptr;
	volatile unsigned short crc10_accum=0;
	int i;
	
	tx_cell_skb=wpabs_skb_alloc(ATM_CELL_SIZE);
	if (!tx_cell_skb){
		return NULL;
	}

	tx_data_ptr=(wp_atm_cell_1_t*) wpabs_skb_put(tx_cell_skb,sizeof(wp_atm_cell_t));
	wpabs_memset(tx_data_ptr,PAD_CHAR,sizeof(wp_atm_cell_t));

	wpabs_memcpy(&tx_data_ptr->hdr,pAdapter->atm_header,ATM_HEADER_SZ);
	wpabs_memcpy(&tx_data_ptr->data,data,len);

	tx_data_ptr->hdr.pt = 0x5;
	tx_data_ptr->hdr.hec = calculate_hec_crc((unsigned char*)tx_data_ptr);
	
	/* Clear the old CRC values */
	set_atm_crc10(tx_data_ptr->data,0);

	for ( i = 0;  i < ATM_PAYLOAD_SZ; i++ ){
        	crc10_accum = ((crc10_accum << 8) & 0x3ff)
                	       ^ byte_crc10_table[( crc10_accum >> 2) & 0xff]
                               ^ tx_data_ptr->data.u.info[i];
    	}

	set_atm_crc10(tx_data_ptr->data,crc10_accum);

	if (skb_return){
		return tx_cell_skb;
	}
	
	if (wpabs_tracing_enabled(pAdapter->atmdevice->trace_info) == 2 ||
	    wpabs_tracing_enabled(pAdapter->atmdevice->trace_info) == 3){
		/* Capture cell for ATM Cell trace */
		capture_atm_trace_packet(pAdapter->atmdevice,	
					 wpabs_skb_data(tx_cell_skb),
					 wpabs_skb_len(tx_cell_skb),
					 TRC_OUTGOING_FRM);
	}

	atm_flip_bitstrm(wpabs_skb_data(tx_cell_skb),
			 wpabs_skb_len(tx_cell_skb));

	wpabs_skb_set_dev(tx_cell_skb,pAdapter->priv);
	wpabs_skb_queue_tail(atm_device->tx_prot_list,tx_cell_skb);
	return NULL; 
}

/*==================================================================
 * capture_atm_trace_packet
 *
 */ 

static int 
capture_atm_trace_packet(
		wp_atm_reg_t*	atm_device,
		void 		*data,
		int		len,
		char		direction)
{
	int atm_packets=0;
	int i;
	wan_trace_pkt_t trc_el;		
	volatile wp_atm_cell_t *atm_cell_ptr=(wp_atm_cell_t*)data;
	void *new_skb=NULL;
	
	if (wpabs_tracing_enabled(atm_device->trace_info) != 3){
		
		if (wpabs_tracing_enabled(atm_device->trace_info) != 2){
			return -1;
		}
	}

	atm_packets = len/sizeof(wp_atm_cell_t);

	if (atm_packets <= 0){
		return -1;
	}
	
	for (i=0;i<atm_packets;i++){

		new_skb = wpabs_skb_alloc(sizeof(wp_atm_cell_t)+sizeof(wan_trace_pkt_t)); 
		if (new_skb == NULL) 
			return -1;
		
		trc_el.status		= direction;
		trc_el.data_avail	= 1;
		trc_el.time_stamp	= (unsigned short)(wpabs_get_systemticks() % 0xFFFF);
		trc_el.real_length	= sizeof(wp_atm_cell_t);
	
		wpabs_skb_copyback(new_skb, 
				  wpabs_skb_len(new_skb), 
				  sizeof(wan_trace_pkt_t), 
				  (unsigned long)&trc_el);

		wpabs_skb_copyback(new_skb, 
				  wpabs_skb_len(new_skb), 
				  sizeof(wp_atm_cell_t),
				  (unsigned long)atm_cell_ptr);

		wpabs_trace_enqueue(atm_device->trace_info, new_skb);

		atm_cell_ptr++;
	}

	return 0;
}

/*==================================================================
 * capture_trace_packet
 *
 */ 

static int 
capture_trace_packet(
		wp_sar_dev_t*	pAdapter,
		void*		skb,
		char		direction)
{
	wp_atm_reg_t*	atm_device=pAdapter->atmdevice;
	int		len = 0;
	void*		new_skb = NULL;
	wan_trace_pkt_t trc_el;	
	int		flag = 0;

	ATM_ASSERT2((!atm_device),0);
	ATM_ASSERT2((!atm_device->trace_info),0);
	
	if ((flag = wpabs_tracing_enabled(atm_device->trace_info)) >= 0){

		if (flag > 1){
			return 0;
		}
			
		/* Allocate a header mbuf */
		len = 	wpabs_skb_len(skb) +  
			sizeof(wan_trace_pkt_t);

		if (!flag && direction == TRC_INCOMING_FRM){
			if ((wpabs_skb_len(skb) - pAdapter->prot_header_size) <= 0){
				return -1;
			}
			len -= pAdapter->prot_header_size;
		}

		new_skb = wpabs_skb_alloc(len+pAdapter->prot_header_size); 
		if (new_skb == NULL) 
			return -1;

		trc_el.status		= direction;
		trc_el.data_avail	= 1;
		trc_el.time_stamp	= (unsigned short)(wpabs_get_systemticks() % 0xFFFF);
		trc_el.real_length	= wpabs_skb_len(skb);
		
		if (direction == TRC_OUTGOING_FRM){
			if (!flag){

				wpabs_skb_copyback(new_skb, 
					  wpabs_skb_len(new_skb), 
				   	  sizeof(wan_trace_pkt_t), 
					  (unsigned long)&trc_el);
	
				wpabs_skb_copyback(new_skb, 
					  wpabs_skb_len(new_skb), 
					  wpabs_skb_len(skb),
					  (unsigned long)wpabs_skb_data(skb));
			}else{
				trc_el.real_length += pAdapter->prot_header_size;

				wpabs_skb_copyback(new_skb, 
				  	wpabs_skb_len(new_skb), 
			   	  	sizeof(wan_trace_pkt_t), 
				  	(unsigned long)&trc_el);

				
				if (pAdapter->prot_header_size){
					wpabs_skb_copyback(new_skb, 
						wpabs_skb_len(new_skb), 
					  	pAdapter->prot_header_size,
					  	(unsigned long)pAdapter->prot_header);
				}
				
				wpabs_skb_copyback(new_skb, 
					  	  wpabs_skb_len(new_skb), 
					  	  wpabs_skb_len(skb),
					  	  (unsigned long)wpabs_skb_data(skb));
			}
		}else{
			if (!flag){
				trc_el.real_length -= pAdapter->prot_header_size;

				wpabs_skb_copyback(new_skb, 
					  wpabs_skb_len(new_skb), 
				   	  sizeof(wan_trace_pkt_t), 
					  (unsigned long)&trc_el);

				wpabs_skb_copyback(new_skb, 
					  wpabs_skb_len(new_skb), 
					  wpabs_skb_len(skb)-pAdapter->prot_header_size,
					  (unsigned long)wpabs_skb_data(skb)+pAdapter->prot_header_size);
			}else{
			
				wpabs_skb_copyback(new_skb, 
					  wpabs_skb_len(new_skb), 
				   	  sizeof(wan_trace_pkt_t), 
					  (unsigned long)&trc_el);

				
				wpabs_skb_copyback(new_skb, 
					  wpabs_skb_len(new_skb), 
					  wpabs_skb_len(skb),
					  (unsigned long)wpabs_skb_data(skb));
			}
		}
		
		wpabs_trace_enqueue(atm_device->trace_info, new_skb);
	}

	return 0;
}




/*========================================================
 * is_idle_cell
 * 
 * 	Check if the incoming ATM Cell is 
 * 	IDLE or UNASSIGNED.
 */

static __inline unsigned char is_idle_cell(unsigned int *data)
{
	unsigned long atm_header = *data;

	/* Idle Cell or Physical Layer OAM Cell */
	if (atm_header == 0x01000000 || atm_header == 0x03000000){
		return 1;
	}

	atm_header &= 0xf1ffff0f;

	/* Reserved for Physical Layer or Unassigned Cell */
	if (atm_header == 0x00000000 || atm_header == 0x01000000){
		return 1;
	}

	/* Data Cell */
	return 0;
}



/*===============================================================
 * wphash_sar_dev
 * wpunhash_sar_dev
 * wp_find_sar_dev_by_vpivci
 *
 *  	Hashing functions used to map VPI/VCI to PVC devices
 */

static int wphash_sar_dev(wp_atm_reg_t *atm_device, wp_sar_dev_t *p)
{
	unsigned int hashid=*((unsigned int*)p->atm_header)&~(ATM_HEADER_MASK);
	wp_sar_dev_t **htable = (wp_sar_dev_t **)&atm_device->atm_hash[vpivci_hashfn(hashid)];

	if (wpabs_test_bit(0,&p->init)){
		wpabs_debug_event("%s: Critical Error: Init Sar in wphash_sar_dev!\n",
				atm_device->devname);
		return -1;
	}	

	if (wp_find_sar_dev_by_vpivci(atm_device, *(unsigned int*)p->atm_header)){
		wpabs_debug_event("%s:%s: Error: device VPI=%i VCI=%i already in use!\n",
				atm_device->devname,p->devname,p->vpi,p->vci);
		return -1;
	}
	
	if((p->hash_next = *htable) != NULL)
		(*htable)->hash_pprev = &p->hash_next;
	
	*htable = p;
	p->hash_pprev = htable;

	wpabs_set_bit(0,&p->init);
	return 0;
}

static void wpunhash_sar_dev(wp_sar_dev_t *p)
{
	if(p->hash_next)
		p->hash_next->hash_pprev = p->hash_pprev;
	
	*p->hash_pprev = p->hash_next;

	wpabs_clear_bit(0,&p->init);
}

static wp_sar_dev_t *wp_find_sar_dev_by_vpivci(wp_atm_reg_t *atm_device, 
		                               unsigned int key)
{
	wp_sar_dev_t *p, **htable;
	
	key&=~(ATM_HEADER_MASK);

	htable = (wp_sar_dev_t **)&atm_device->atm_hash[vpivci_hashfn(key)];
	
	for(p = *htable; p ;p = p->hash_next){
		if (((*(unsigned int*)p->atm_header)&~(ATM_HEADER_MASK)) == key){
			break;
		}
	}

	return p;
}




/*==SECTION=========================================================
 *
 * CRC Routines
 * 
 *=================================================================*/


/* crc32h.c -- package to compute 32-bit CRC one byte at a time using   */
/*             the high-bit first (Big-Endian) bit ordering convention  */
/*                                                                      */
/* Synopsis:                                                            */
/*  gen_crc_table() -- generates a 256-word table containing all CRC    */
/*                     remainders for every possible 8-bit byte.  It    */
/*                     must be executed (once) before any CRC updates.  */
/*                                                                      */
/*  unsigned update_crc(crc_accum, data_blk_ptr, data_blk_size)         */
/*           unsigned crc_accum; char *data_blk_ptr; int data_blk_size; */
/*           Returns the updated value of the CRC accumulator after     */
/*           processing each byte in the addressed block of data.       */
/*                                                                      */
/*  It is assumed that an unsigned long is at least 32 bits wide and    */
/*  that the predefined type char occupies one 8-bit byte of storage.   */
/*                                                                      */
/*  The generator polynomial used for this version of the package is    */
/*  x^32+x^26+x^23+x^22+x^16+x^12+x^11+x^10+x^8+x^7+x^5+x^4+x^2+x^1+x^0 */
/*  as specified in the Autodin/Ethernet/ADCCP protocol standards.      */
/*  Other degree 32 polynomials may be substituted by re-defining the   */
/*  symbol POLYNOMIAL below.  Lower degree polynomials must first be    */
/*  multiplied by an appropriate power of x.  The representation used   */
/*  is that the coefficient of x^0 is stored in the LSB of the 32-bit   */
/*  word and the coefficient of x^31 is stored in the most significant  */
/*  bit.  The CRC is to be appended to the data most significant byte   */
/*  first.  For those protocols in which bytes are transmitted MSB      */
/*  first and in the same order as they are encountered in the block    */
/*  this convention results in the CRC remainder being transmitted with */
/*  the coefficient of x^31 first and with that of x^0 last (just as    */
/*  would be done by a hardware shift register mechanization).          */
/*                                                                      */
/*  The table lookup technique was adapted from the algorithm described */
/*  by Avram Perez, Byte-wise CRC Calculations, IEEE Micro 3, 40 (1983).*/

#define POLYNOMIAL 0x04c11db7L


static void gen_crc_table(void)
{ 
   register int i, j;  register unsigned long crc_accum;
   for ( i = 0;  i < 256;  i++ )
       { crc_accum = ( (unsigned long) i << 24 );
         for ( j = 0;  j < 8;  j++ )
              { if ( crc_accum & 0x80000000L )
                   crc_accum =
                     ( crc_accum << 1 ) ^ POLYNOMIAL;
                else
                   crc_accum =
                     ( crc_accum << 1 ); }
         crc_table[i] = crc_accum; }
   return; 
}

static unsigned long update_crc(unsigned long crc_accum, char *data_blk_ptr,
                                                    int data_blk_size)
{ 	
	register int i, j;
   	for ( j = 0;  j < data_blk_size;  j++ )
       	{ i = ( (int) ( crc_accum >> 24) ^ *data_blk_ptr++ ) & 0xff;
         crc_accum = ( crc_accum << 8 ) ^ crc_table[i]; }
   	return crc_accum; 
}

#define CRC10_POLYNOMIAL 0x633
static void gen_byte_crc10_table(void)
/* generate the table of CRC-10 remainders for all possible bytes */
{
    register int i, j;
    register unsigned short crc10_accum=0;

    for ( i = 0;  i < 256;  i++ )
    {
        crc10_accum = ((unsigned short) i << 2);
        for ( j = 0;  j < 8;  j++ )
        {
            if ((crc10_accum <<= 1) & 0x400) crc10_accum ^= CRC10_POLYNOMIAL;
        }
        byte_crc10_table[i] = crc10_accum;
    }
    return;
}




/*===============================================================
 * generate_HEC_chk_table
 *
 * GENERATE A TABLE FOR CHECKING THE HEC VALUE IN A CELL 
 *
 */
#define HEC_GENERATOR   	   0x107
#define I361_HEC_COSET_FUNCTION    0x55
 
static void generate_hec_crc_table(void)
{
	int HEC_syndrome;
	int i,j;

	for(i = 0; i < 256; i++) {
		HEC_syndrome = i;
		for(j = 0; j < 8; j++) {
			if(HEC_syndrome & 0x80) {
				HEC_syndrome = (HEC_syndrome << 1) ^ HEC_GENERATOR;
			}else{
				HEC_syndrome = (HEC_syndrome << 1);
			}
		}
		HEC_table[i] = (unsigned char)HEC_syndrome;
    }
}



/**** COMPUTE THE HEC VALUE OF THE CELL CURRENTLY IN THE SCRATCHPAD BUFFER ****/

unsigned char calculate_hec_crc(unsigned char *util_ptr)
{

	/* zero the HEC accumulator */
	unsigned char HEC_accum = 0;
	unsigned int i; 
	unsigned char util_char;
	
	/* set a pointer to the cell scratchpad buffer */
	/* process the first four bytes in the ATM header */
	for(i = 0; i < (ATM_HEADER_SZ - 1); i ++) {
		/* update the HEC value for each byte in the cell header */
		HEC_accum = HEC_table[HEC_accum ^ *util_ptr++];
	}

	/* in general, the cell header check sum is XORed with the COSET 
	 * function defined in CCITT I.361 (0x55) 
	 * */
	util_char = I361_HEC_COSET_FUNCTION;		
	
	HEC_accum ^= util_char;		

	return HEC_accum;
}

static void gen_flip_tables(void)
{
	int bitcnt=0,i;
	unsigned char byte=0,fbyte=0;

	for (byte=0;;byte++){
		bitcnt=0;
		fbyte=0;

		for (i=7;i>=0;i--){
			if (wpabs_test_bit(i,&byte)){
				fbyte |= (1<<bitcnt);	
			}
			bitcnt++;
		}

		FLIP_TABLE[byte]=fbyte;

		if (byte==0xFF){
			break;
		}
	}
}

static void atm_flip_bitstrm(unsigned char *data, int len)
{
	int l;
	for (l=0;l<len;l++){
		data[l]=FLIP_TABLE[data[l]];
	}
}

static int validate_mgmnt_cell(wp_atm_cell_1_t *tx_data_ptr)
{
	unsigned short crc10_accum=0;
	int i;
	
	if (calculate_hec_crc((unsigned char *)tx_data_ptr) != tx_data_ptr->hdr.hec){
		return 1;
	}
	
	for ( i = 0;  i < ATM_PAYLOAD_SZ; i++ ){
        	crc10_accum = ((crc10_accum << 8) & 0x3ff)
                	       ^ byte_crc10_table[( crc10_accum >> 2) & 0xff]
                               ^ tx_data_ptr->data.u.info[i];
    	}

	return crc10_accum;
}

static void set_atm_sar_state(wp_sar_dev_t* pAdapter, unsigned char state)
{
	if (pAdapter->state != state){

		switch (state){

		case ATM_CONNECTED:
			wpabs_debug_event("%s:%s: ATM SAR State: Connected\n",
					pAdapter->atmdevice->devname,
					pAdapter->devname);
			break;

		case ATM_DISCONNECTED:
			wpabs_debug_event("%s:%s: ATM SAR State: Disconnected\n",
					pAdapter->atmdevice->devname,
					pAdapter->devname);
			break;
			
		case ATM_AIS:
			wpabs_debug_event("%s:%s: ATM SAR State: Alarm (AIS)\n",
					pAdapter->atmdevice->devname,
					pAdapter->devname);
			break;

		default:
			wpabs_debug_event("%s:%s: ATM SAR State: Unknown (0x%X)\n",
					pAdapter->atmdevice->devname,
					pAdapter->devname,
					state);
			break;	
		};
		
		pAdapter->state=state;
	}
}
