/*****************************************************************************
* wanpipe_atm.c	WANPIPE(tm) ATM ALL5 SAR
* 		
* Authors: 	Nenad Corbic <ncorbic@sangoma.com>
*
* Copyright:	(c) 2003 Sangoma Technologies Inc.
*
*		This program is free software; you can redistribute it and/or
*		modify it under the terms of the GNU General Public License
*		as published by the Free Software Foundation; either version
*		2 of the License, or (at your option) any later version.
* ============================================================================
* Jan 07, 2002	Nenad Corbic	Based on ADSL ATM SAR
*****************************************************************************/
/* 
 * ATM Header 
 * 
 * --8---7---6---5---4---3---2---1--
 * |             VPI               |
 * |-------------------------------|
 * |     VPI       |    VCI        |
 * |-------------------------------|
 * |             VCI               |
 * |-------------------------------|
 * |      VCI      |    PT     |CLP|
 * |-------------------------------|
 * |             HEC               |
 * |-------------------------------|
 * |                               |
 * |     Information Payload       |
 * |       (48 Octects)            |
 * ---------------------------------
 */
		
/* Protocol Headers 
 *
 * --1-----1-----1--------3---------2--------<=65,527-----
 * |DSAP|SSAP|Control|Unique Id|Protocol Id|   PDU       |
 * -------------------------------------------------------
 *
 * Bridged LLC: Ethernet over ATM
 * 	DSAP		=0xaa
 * 	SSAP		=0xaa
 * 	Control		=0x03
 *	Unique ID	=0x00 0x80 0xC2 	#Bridged PDU
 *	Protocol Id	=0x00 0x07		#802.3/Ethernet 
 *						#Witout Preserverd FCS 
 *	Pading		=0x00 0x00
 *			
 * Classical IP over ATM
 *	DSAP  		=0xAA
 *	SSAP		=0xAA
 *	Control		=0x03
 *	Unique Id	=0x00 0x00 0x00
 *
 *	Protocol Id 	=0x80 0x00  : IP
 *	Protocol Id	=0x80 0x06  : ARP
 *
 *      MTU=8192
 */

/*==================================================================
 * INCLUES
 *=================================================================*/

#include "wanpipe_abstr.h"
#include "wanpipe_cfg.h"
#include "wanpipe_atm.h" 


/*==================================================================
 * DEFINES
 *=================================================================*/

#define INITIAL_FCS_VALUE 	0xFFFFFFFF
#define EIGHT_BITS 	  	8

#define FLIP_CRC(a,b)  { b=0; \
			 b |= (a&0x000000FF)<<24 ; \
			 b |= (a&0x0000FF00)<<8; \
			 b |= (a&0x00FF0000)>>8; \
			 b |= (a&0xFF000000)>>24; }

#define CLP_BIT		0
#define PT_SDU1		1
#define PT_CONGESTION 	2			 
#define PT_ADMIN	3

#define PAD_CHAR	0x6A
#define ATM_OVERHEAD    1.1

/* ethernet broadcast address */
#define MAC_BCAST_ADDR "\xff\xff\xff\xff\xff\xff"

#define ETH_P_PPPOE_DISC 0x8863 /* discovery stage */
#define ETH_P_PPPOE_SESS 0x8864 /* session stage */

/* PPPoE codes */
#define CODE_SESS 0x00 /* PPPoE session */
#define CODE_PADI 0x09 /* PPPoE Active Discovery Initiation */
#define CODE_PADO 0x07 /* PPPoE Active Discovery Offer */
#define CODE_PADR 0x19 /* PPPoE Active Discovery Request */
#define CODE_PADS 0x65 /* PPPoE Active Discovery Session-confirmation */
#define CODE_PADT 0xa7 /* PPPoE Active Discovery Terminate */

struct ethhdr 
{
	unsigned char	h_dest[6];	/* destination eth addr	*/
	unsigned char	h_source[6];	/* source ether addr	*/
	unsigned short	h_proto;		/* packet type ID field	*/
} __attribute__((packed));

/* PPPoE packet; includes Ethernet headers and such */
struct wan_pppoe_packet {
    struct ethhdr	ethhdr; /* ethernet header */
    unsigned int	ver:4; /* pppoe version */
    unsigned int	type:4; /* pppoe type */
    unsigned int	code:8; /* pppoe code CODE_* */
    unsigned int	session:16; /* session id */
    unsigned short	length; /* payload length */
    /* payload follows */
};
/* PPPoE tag; the payload is a sequence of these */
struct wan_pppoe_tag {
    unsigned short type; /* tag type TAG_* */
    unsigned short length; /* tag length */
    /* payload follows */
};

/* PPPoE tag types */
#define TAG_END_OF_LIST        0x0000
#define TAG_SERVICE_NAME       0x0101
#define TAG_AC_NAME            0x0102
#define TAG_HOST_UNIQ          0x0103
#define TAG_AC_COOKIE          0x0104
#define TAG_VENDOR_SPECIFIC    0x0105
#define TAG_RELAY_SESSION_ID   0x0110
#define TAG_SERVICE_NAME_ERROR 0x0201
#define TAG_AC_SYSTEM_ERROR    0x0202
#define TAG_GENERIC_ERROR      0x0203

#define WPATM_MEM_ASSERT()  wpabs_debug_event("%s:%d: No Memory!\n",__FUNCTION__,__LINE__)


/*==================================================================
 * GLOBAL VARIABLES
 *
 * Note: This code has been written to work in SMP multi-treaded 
 *       systems, where multiple drivers might simultaneously 
 *       use this code. 
 *       Thus global variables are very dangerous.  
 *       Only use them if you KNOW what you are doning.
 *=================================================================*/

static unsigned char  g_trailer[]={0x00, 0x00, 0xEE, 0xEE, 0xFF, 0xFF, 0xFF, 0xFF};
static unsigned short byte_crc10_table[256];
static unsigned long  crc_table[256];
static unsigned char  global_init_crc;
static unsigned char  HEC_table[256];


/*==================================================================
 * FUNCTION PROTOTYPES 
 *=================================================================*/

static void 		gen_crc_table(void);
static void 		gen_byte_crc10_table(void);
static unsigned long 	update_crc(unsigned long, char*, int);
static unsigned char	calculate_hec_crc(unsigned char *util_ptr);
static void 		generate_hec_crc_table(void);

static int 		parse_prot_header(wp_sar_dev_t*, void*);

static __inline void 	wphash_sar_dev(wp_atm_reg_t *, wp_sar_dev_t *);
static __inline void 	wpunhash_sar_dev(wp_sar_dev_t *p);
static __inline wp_sar_dev_t *wp_find_sar_dev_by_vpivci(wp_atm_reg_t *atm_device, 
							unsigned int key);


#if 0
static int 		capture_trace_packet (adapter_t *, void *, char);
static int 		capture_atm_trace_packet(adapter_t*,void*,int,char);

static void* 		create_inarp_request(adapter_t *);
static void 		handle_atm_admin_packet(adapter_t *, void *);
static void 		handle_oam_cell_loopback(adapter_t *, void *);
static int 		handle_atm_arp(adapter_t *pAdapter, void* rx_skb);

static int 		capture_trace_packet (adapter_t *, void *, char);
static int 		capture_atm_trace_packet(adapter_t*,void*,int,char);
#endif

static __inline unsigned char is_idle_cell(unsigned int *data);

#if defined(__FreeBSD__) || defined(__OpenBSD__)
static void atm_tasklet (void* data, int pending);
#else
static void atm_tasklet (unsigned long data);
#endif


int wp_atm_register_device(void *card, wp_atm_reg_t **atm_device_ptr)
{
	
	wp_atm_reg_t *atm_device=wpabs_malloc(sizeof(wp_atm_reg_t));
	if (!atm_device){
		WPATM_MEM_ASSERT();
		return -1;
	}

	atm_device->atm_task=wpabs_tasklet_alloc();
	if (!atm_device->atm_task){
		WPATM_MEM_ASSERT();
		wpabs_free(atm_device);
		return -1;
	}
	
	if (!wpabs_test_bit(0,&global_init_crc)){
		wpabs_set_bit(0,&global_init_crc);
		gen_crc_table();
		gen_byte_crc10_table();
		generate_hec_crc_table();
	}
	
	if (wpabs_test_bit(0,&atm_device->init)){
		wpabs_debug_event("Error ATM Device busy: init=1!\n");
		return -1;
	}

	wpabs_tasklet_init(atm_device->atm_task,
			   0,
			   (void*)&atm_tasklet,
			   (void*)atm_device); 

	atm_device->card = card;
		
	*atm_device_ptr = atm_device;
	wpabs_set_bit(0,&atm_device->init);

	return 0;
}

void wp_atm_unregister_device(wp_atm_reg_t *atm_device)
{
	wpabs_clear_bit(0,&atm_device->init);

	if (atm_device->atm_task){
		wpabs_tasklet_kill(atm_device->atm_task);
		wpabs_free(atm_device->atm_task);	
	}

	wpabs_free(atm_device);
}


/*==================================================================
 * wanpipe_atm_register
 *
 *  	o Initialize the ATM Ecapsulation Protocol
 *  	o Allocate a tx cells buffer that will be used to
 *  	  cut up tx packets into ATM cells
 *
 */ 

int wp_atm_register_pvc(void *card,
		        wp_atm_reg_t *atm_device, 
			wp_sar_dev_t **pAdapter_ptr, 
			unsigned char mode, unsigned short vpi, 
			unsigned short vci,unsigned int mtu)
{
	wp_sar_dev_t *pAdapter;
	
	pAdapter=wpabs_malloc(sizeof(wp_sar_dev_t));
	if (!pAdapter){
		WPATM_MEM_ASSERT();
		return -1;
	}
		
	
	pAdapter->EncapMode=mode;
	pAdapter->vpi=vpi;
	pAdapter->vci=vci;
	
	switch (pAdapter->EncapMode){

		case RFC_MODE_BRIDGED_ETH_LLC:
			wpabs_debug_event("%s: ATM configured for Bridged LLC ETH/ATM\n",
					pAdapter->devname);
			pAdapter->prot_header[0]=0xAA;
			pAdapter->prot_header[1]=0xAA;
			pAdapter->prot_header[2]=0x03;
			pAdapter->prot_header[3]=0x00;
			pAdapter->prot_header[4]=0x80;
			pAdapter->prot_header[5]=0xC2;
			pAdapter->prot_header[6]=0x00;
			pAdapter->prot_header[7]=0x07;
			pAdapter->prot_header[8]=0x00;
			pAdapter->prot_header[9]=0x00;
			pAdapter->prot_header_size=10;	
			break;

		case RFC_MODE_BRIDGED_ETH_VC:
			wpabs_debug_event("%s: ATM configured for Bridged VC ETH/ATM\n",
					pAdapter->devname);
			pAdapter->prot_header[0]=0x00;
			pAdapter->prot_header[1]=0x00;
			pAdapter->prot_header_size=2;
			break;

		case RFC_MODE_ROUTED_IP_LLC:
			wpabs_debug_event("%s: ATM configured for Classical IP (LLC) over ATM\n",
					pAdapter->devname);
			pAdapter->prot_header[0]=0xAA;
			pAdapter->prot_header[1]=0xAA;
			pAdapter->prot_header[2]=0x03;
			pAdapter->prot_header[3]=0x00;
			pAdapter->prot_header[4]=0x00;
			pAdapter->prot_header[5]=0x00;
			pAdapter->prot_header[6]=0x08;
			pAdapter->prot_header[7]=0x00;
			pAdapter->prot_header_size=8;
			break;


		case RFC_MODE_ROUTED_IP_VC:
			wpabs_debug_event("%s: ATM configured for IP (VC) over ATM\n",
					pAdapter->devname);
			pAdapter->prot_header_size=0;
			break;
			
		case RFC_MODE_PPP_LLC:
			wpabs_debug_event("%s: ATM configured for PPP (LLC) over ATM\n",
					pAdapter->devname);
			pAdapter->prot_header[0]=0xFE;
			pAdapter->prot_header[1]=0xFE;
			pAdapter->prot_header[2]=0x03;
			pAdapter->prot_header[3]=0xCF;
			pAdapter->prot_header_size=4;
			break;

		case RFC_MODE_PPP_VC:
			wpabs_debug_event("%s: ATM configured for PPP (VC) over ATM\n",
					pAdapter->devname);
			pAdapter->prot_header_size=0;
			break;


		default:
			wpabs_debug_event("%s: ERROR: Invalid ATM Lan Protocol option %d!\n",
					pAdapter->devname,pAdapter->EncapMode);
			wpabs_free(pAdapter);
			return -1;
	}
	
	pAdapter->atm_header[0] = (pAdapter->vpi &0xF0)>>4;	
	pAdapter->atm_header[1] = (pAdapter->vpi &0x0F)<<4;	
	pAdapter->atm_header[1] |= (pAdapter->vci&0xF000)>>12;
	pAdapter->atm_header[2] = (pAdapter->vci &0x0FF0)>>4;
	pAdapter->atm_header[3] = (pAdapter->vci &0x000F)<<4;

	/* Calculate the hec number now because it will be the
	 * same all cells, since vpi/vci numbers will not change */
	pAdapter->atm_header[4] = calculate_hec_crc(pAdapter->atm_header);
	pAdapter->hec=calculate_hec_crc(pAdapter->atm_header);

	/* Calculate the hec of atm header when EOF bit
	 * (end of frame bit) is set */
	wpabs_set_bit(1,&pAdapter->atm_header[3]);
	pAdapter->hec_eof=calculate_hec_crc(pAdapter->atm_header);

	/* Reset the EOF frame bit */
	wpabs_clear_bit(1,&pAdapter->atm_header[3]);

	wpabs_memcpy(&pAdapter->trailer[0],&g_trailer[0],sizeof(pAdapter->trailer));
	
	wpabs_debug_cfg("%s: ATM header = %x %x %x %x\n",
			pAdapter->devname,
			pAdapter->atm_header[0],pAdapter->atm_header[1],
			pAdapter->atm_header[2],pAdapter->atm_header[3]);

	pAdapter->max_sdu_size=mtu*ATM_OVERHEAD;

	pAdapter->tx_cells_skb = 
		wpabs_skb_alloc(pAdapter->max_sdu_size+10);
	
	if (!pAdapter->tx_cells_skb){
		wpabs_debug_event("%s: Failed allocate memory for Tx cells packet (%d bytes)!\n",
				pAdapter->devname, pAdapter->max_sdu_size+10);
		wpabs_free(pAdapter);
		return -1;
	}

	wpabs_memset(pAdapter->pad_len_buf,PAD_CHAR,sizeof(pAdapter->pad_len_buf));


	/* FIXME: Trace not supported 
	atm_trace_info_init(pAdapter->trace_info);
	*/

	pAdapter->priv=card;

	wphash_sar_dev(atm_device,pAdapter);

	return 0; 
}




void wp_atm_unregister_pvc(wp_atm_reg_t *atm_device, wp_sar_dev_t *pAdapter)
{

	/* FIXME: This section should be locked */
	
	if (pAdapter->rx_pkt_skb){
		wpabs_skb_free(pAdapter->rx_pkt_skb);
		pAdapter->rx_pkt_skb = NULL;
	}

	if (pAdapter->tx_cells_skb){
		wpabs_skb_free(pAdapter->tx_cells_skb);
		pAdapter->rx_pkt_skb = NULL;
	}

	wpunhash_sar_dev(pAdapter);

	wpabs_free(pAdapter);
	return;
}

	
void wanpipe_atm_rx (wp_atm_reg_t *atm_device)
{
	if (!wpabs_test_bit(0,&atm_device->atm_task_running)){
		
		wpabs_set_bit(0,&atm_device->atm_task_running);
		
		wpabs_debug_rx("(Debug RX): Schedule Rx task.\n");
		
		wpabs_tasklet_schedule(atm_device->atm_task);
	}
	return;
}

#if defined(__FreeBSD__) || defined(__OpenBSD__)
static void atm_tasklet (void* data, int pending)
#else
static void atm_tasklet (unsigned long data)
#endif
{	
	wp_atm_reg_t *		atm_device = (wp_atm_reg_t *)data;
	volatile wp_atm_cell_t*	atm_cell;
	void*			rx_skb;
	void*			rx_cell_block;
	unsigned short 		info_len;
	unsigned char*		data_ptr;
	int 			num_of_cells = 0;
	wp_sar_dev_t		*pAdapter;
	int 			i;


	/* SMP sanity check, this funciton is not
	 * re-entrant */
	if (wpabs_test_bit(1,&atm_device->atm_task_running)){
		wpabs_debug_event("CRITICAL IN TASKLET!\n");
		wpabs_tasklet_end(atm_device->atm_task);
		return;
	}

	wpabs_set_bit(1,&atm_device->atm_task_running);
	wpabs_clear_bit(0,&atm_device->atm_task_running);
	
	if (!atm_device->atm_get_rx_block || !atm_device->card){
		wpabs_debug_event("%s:%d: Critical Error: atm_device not initialized\n",
				__FUNCTION__,__LINE__);
		goto atm_task_exit;
	}
	
	while ((rx_cell_block=atm_device->atm_get_rx_block(atm_device->card)) != NULL){
		
		atm_cell = (wp_atm_cell_t *)wpabs_skb_data(rx_cell_block);
		num_of_cells=wpabs_skb_len(rx_cell_block)/ATM_CELL_SIZE;


		for (i=0; i<num_of_cells; i++){

			if (is_idle_cell((unsigned int*)atm_cell)){
				/* Discard idle cells */
				goto atm_cell_failure;
			}
			
			pAdapter = wp_find_sar_dev_by_vpivci(atm_device,*(unsigned int *)atm_cell);
			if (!pAdapter){
				wpabs_debug_event("%s: Rx ATM SAR: Cell with invalid VPI=%i VCI=%i\n",
					pAdapter->devname, WPVPI(atm_cell), WPVCI(atm_cell));
				goto atm_cell_failure;
			}

			if (!pAdapter->rx_pkt_skb){
				pAdapter->rx_pkt_skb = wpabs_skb_alloc(pAdapter->max_sdu_size+10);
				if (pAdapter->rx_pkt_skb){
					wpabs_debug_event("%s: Rx ATM SAR: Critica error No Memory!\n");
					goto atm_cell_failure;
				}
			}

			rx_skb=pAdapter->rx_pkt_skb;

			wpabs_debug_rx("RX: ATM HEADER %x %x %x %x  CFG: ATM HEADER %x %x %x %x\n",
					atm_cell->hdr[0],
					atm_cell->hdr[1],
					atm_cell->hdr[2],
					atm_cell->hdr[3],
					pAdapter->atm_header[0],
					pAdapter->atm_header[1],
					pAdapter->atm_header[2],
					pAdapter->atm_header[3]);

			/* FIXME: Add tracing support */
#if 0
			if (atm_tracing_enabled(pAdapter->trace_info) == 2){
				/* Capture cell for ATM Cell trace */
				capture_atm_trace_packet(pAdapter,	
					  	         (void*)atm_cell,
							 sizeof(GP_CELL),
					        	 TRC_INCOMING_FRM);
			}
#endif
			
			/* Test if the incoming cell is an ATM management packet.
			 * If it is run it through the ATM management handler */
			if (wpabs_test_bit(PT_ADMIN,(void*)&atm_cell->hdr[3])){
				
				/* FIXME: Add PT ADMIN support
				 * handle_atm_admin_packet(pAdapter,(void*)atm_cell);*/
				
				wpabs_debug_event("%s: Dropping Admin Packet PTI=0x%X\n",
						pAdapter->devname,(atm_cell->hdr[3]&0x0E)>>1);
				goto atm_cell_failure;
			}
			
			/* PT bit not set: Start of the frame.
			 * PT bit set: Last data fragment */
			if (!wpabs_test_bit(PT_SDU1,(void*)&atm_cell->hdr[3])){

				wpabs_debug_rx("%s: RX: New frame\n", 
							pAdapter->devname);

				/* FIXME: No auto vpi/vci config */
				
				if (wpabs_skb_len(rx_skb) + 48 > pAdapter->max_sdu_size){
					wpabs_debug_event("%s: ATM Error: SDU0 Frame too big %i > %i !\n",
							pAdapter->devname,
							wpabs_skb_len(rx_skb) + 48,
							pAdapter->max_sdu_size);
					wpabs_skb_init(rx_skb, 16);
					wpabs_skb_trim(rx_skb, 0);
					++pAdapter->atm_stats.rx_atm_pdu_size;
					goto atm_cell_failure;
				}
	
				data_ptr=(unsigned char*)atm_cell->info;
				wpabs_skb_copyback(rx_skb, 
						 wpabs_skb_len(rx_skb), 
						 48, 
						 (unsigned long)data_ptr);
				++pAdapter->atm_stats.rx_valid;
			}else{
				wpabs_debug_rx("%s: RX: Last data fragment\n", 
						pAdapter->devname);

				info_len = *((unsigned short*)&atm_cell->info[42]);
				info_len = wpabs_htons(info_len);
				
				if (wpabs_skb_len(rx_skb) + 48 > pAdapter->max_sdu_size){

					wpabs_debug_event("%s: ATM Error: SDU1 Frame too big %i > %i !\n",
							pAdapter->devname,
							wpabs_skb_len(rx_skb) + 48,
							pAdapter->max_sdu_size);

					wpabs_skb_init(rx_skb, 16);
					wpabs_skb_trim(rx_skb, 0);
					++pAdapter->atm_stats.rx_atm_pdu_size;
					goto atm_cell_failure;
				}

				data_ptr=(unsigned char*)&atm_cell->info[0];
				wpabs_skb_copyback(rx_skb, 
						 wpabs_skb_len(rx_skb), 
						 48, 
						 (unsigned long)data_ptr);
				
				if (wpabs_skb_len(rx_skb) > info_len){
					wpabs_skb_trim(rx_skb, info_len);
				}

#if 0
				if (atm_tracing_enabled(pAdapter->trace_info) >= 0){
					capture_trace_packet(
							pAdapter,	
							rx_skb,
							TRC_INCOMING_FRM);
				}
#endif

				if (parse_prot_header(pAdapter, rx_skb)){
					wpabs_skb_init(rx_skb, 16);
					wpabs_skb_trim(rx_skb, 0);
					++pAdapter->atm_stats.rx_invalid_atm_hdr;
					goto atm_cell_failure;
				}

				++pAdapter->atm_stats.rx_valid;
				wpabs_skb_pull(rx_skb, pAdapter->prot_header_size);

				if (pAdapter->rx_data_ready){
					pAdapter->rx_data_ready(atm_device->card,rx_skb);
					pAdapter->rx_pkt_skb=NULL;
				}else{
					wpabs_debug_rx("%s: %s:%d: Critical error atm dev not initialized\n",
							pAdapter->devname,
							__FUNCTION__,
							__LINE__);
					wpabs_skb_init(rx_skb, 16);
					wpabs_skb_trim(rx_skb, 0);
					goto atm_cell_failure;
				}
			}
atm_cell_failure:
			atm_cell++;

		}/* for */
	}

atm_task_exit:
	wpabs_clear_bit(1,&atm_device->atm_task_running);
	wpabs_tasklet_end(atm_device->atm_task);
	return;
}


int wanpipe_atm_tx (wp_sar_dev_t *pAdapter, void* tx_skb, void **tx_cell_skb)
{
	unsigned char *src_data_ptr;
	unsigned int  src_data_size, src_payload_size;
	unsigned char *target_data_ptr;
	unsigned int  offset=0,pad=0;
	unsigned long crc = INITIAL_FCS_VALUE;
	unsigned long fin_crc;
	int i;

	/* If the incoming packet has length of zero 
	 * stop the tx procedure */
	if (wpabs_skb_len(tx_skb) == 0){
		return -1;
	}
	
	wpabs_debug_tx("%s: Sending packet (%d bytes)...\n", 
				pAdapter->devname, wpabs_skb_len(tx_skb));

	src_data_ptr    = wpabs_skb_data(tx_skb);
        src_data_size   = wpabs_skb_len(tx_skb);

	if (src_data_ptr == NULL || src_data_size == 0){
		wpabs_debug_event("%s: src_data_ptr is NULL or src_data_size is 0!\n", 
						pAdapter->devname);
		return -1;
	}

	/* If the incoming packet is an ATMARP, we must change the
	 * default IPoATM PDU header to ATMARP header, just for
	 * this pakcet.  This value will be reset after the SAR
	 * is finished with this packet */
/* FIXME
	if (pAdapter->EncapMode==RFC_MODE_ROUTED_IP_LLC &&
	    *(unsigned short*)&src_data_ptr[0]==ATMARP_HEADER){
		pAdapter->prot_header[7]=0x06;
	}
*/
	if (!pAdapter->tx_cells_skb){
		wpabs_debug_event("%s: TX ATM SAR: atm sar device not initialized!\n",
				pAdapter->devname);
		return -1;
	}

	/* Initialize the Tx Cell buffer */
	wpabs_skb_init(pAdapter->tx_cells_skb, 16);
	wpabs_skb_trim(pAdapter->tx_cells_skb, 0);

	target_data_ptr=wpabs_skb_put(pAdapter->tx_cells_skb,pAdapter->max_sdu_size);

	/* Initialize SAR variables used to cut up the tx packet
	 * into ATM Cells */
	offset=0;
	src_payload_size = src_data_size + pAdapter->prot_header_size;
	*((unsigned short*)&pAdapter->trailer[2])=wpabs_htons(src_payload_size);
	*((unsigned short*)&pAdapter->pad_len_buf[47])=0x00;
	*((unsigned short*)&pAdapter->pad_len_buf[49])=wpabs_htons(src_payload_size);
	

	/* Begining of the SAR. 
	 * Start cutting up the tx frame into ATM 
	 * Cells.  
	 *
	 * Note: For S518 Boards
	 * 	 
	 * 	 Do not calculate the HEC here. The HEC is
	 *       done in hardware. Thus we only have to 
	 *       deal with 52 bytes frames 
	 *
	 *       For S514 Boards
	 *
	 *       HEC is calculated once on setup. Our ATM header
	 *       does not change after that.
	 */
	for (i=0;;){

		/* Each ATM cell starts with a 4 byte (i.e. HEC not included)
		 * ATM header which contains VPI/VCI/PT/CLP values */
		wpabs_memcpy((target_data_ptr+offset),
		             pAdapter->atm_header,
			     sizeof(pAdapter->atm_header));

		offset+=sizeof(pAdapter->atm_header);
	
		/* Check that a tx packet can fit into a single
		 * ATM cell which can accommodate 48 bytes of data */
		if ((i+48) <= src_payload_size){
		
			/* The packet cannot fit into a single cell */
			
			/* Check if this is a first cell of an ATM PDU */
			if (i==0){
				/* If this is a first cell of the ATM PDU
				 * we must append an ATM PDU header. ATM
				 * PDU header defines the above protocol
				 * that is being transported by the ATM
				 * network */
				wpabs_memcpy((target_data_ptr+offset),
				        pAdapter->prot_header,
					pAdapter->prot_header_size);
		
				offset+=pAdapter->prot_header_size;
			
				wpabs_memcpy((target_data_ptr+offset),
				        src_data_ptr,
				        (48-pAdapter->prot_header_size)); 
				src_data_ptr += (48-pAdapter->prot_header_size);
				offset += (48-pAdapter->prot_header_size);

			}else{
				/* Not a first cell, thus, just fill the 48 bytes
				 * of data and move to another cell */
				wpabs_memcpy((target_data_ptr+offset),src_data_ptr,48);
				src_data_ptr += 48;
				offset += 48;
			}
			++pAdapter->atm_stats.tx_valid;
		}else{
			
			/* Tx packet can fit into a single cell or 
			 * we have reached the last cell needed to
			 * tx the ATM PDU */
			
			++pAdapter->atm_stats.tx_valid;
			
			if (i==0){
				/* This is the first cell, which means
				 * that the ATM PDU can fit into a
				 * single cell i.e. less than 48 bytes.
				 *
				 * The Payload is smaller than 48 bytes,
				 * there are two scenarios here:
				 * 
				 * 1) Payload + Trailer = PDU fit into a
				 *    single cell. i.e <= 48
				 *    
				 * 2) Payload + Trailer > PDU do not fit
				 *    into a single cell i.e. > 48 thus 
				 *    we must create and pad the second cell
				 */
				
				/* Creating a first cell */
				wpabs_memcpy((target_data_ptr+offset),
				        pAdapter->prot_header,
				        pAdapter->prot_header_size);
				offset+=pAdapter->prot_header_size;
			
				wpabs_memcpy((target_data_ptr+offset),
				       src_data_ptr,
				       src_data_size); 
				src_data_ptr += src_data_size;
				offset += src_data_size;
			
			}else{
				/* Copy the last part of the tx frame
				 * into the ATM CELL */
				
				if ((src_payload_size - i) > 0){
					wpabs_memcpy((target_data_ptr+offset),
					       src_data_ptr,
					       (src_payload_size-i)); 
					src_data_ptr+=(src_payload_size-i);
					offset	   +=(src_payload_size-i);
				}
			}

			/* Since this is the last ATM CELL we must append a
			 * ATM PDU trailer to the CELL.  Check if the trailer
			 * will fit into the CELL.  If not we must create
			 * another cell and pad it */
			
			if ((src_payload_size+sizeof(pAdapter->trailer)) > (i+48)){
				
				/* We cannot fit the whole PDU frame
				 * into this cell. We need add another cell. */
				
				/* Pad to the end of the first cell because
				 * the trailer must reside at the end of the
				 * ATM CELL. Thus we'll put it at the end of
				 * the next cell */
				pad=((i+48)-src_payload_size);
				if (pad > 0){
					wpabs_memset((target_data_ptr+offset),PAD_CHAR,pad);
					offset+=pad;
				}

				/* Start the atm header on second cell */
				wpabs_memcpy((target_data_ptr+offset),
					pAdapter->atm_header,
					sizeof(pAdapter->atm_header));

				offset+=sizeof(pAdapter->atm_header);
			
				/* Set the END of FRAME bit in the PT section of
				 * the ATM header, this will indicate to the
				 * remote side that this CELL is the last */
				wpabs_set_bit(1,&((target_data_ptr+offset-1)[0]));

				(target_data_ptr+offset-1)[1]=pAdapter->hec_eof;

				/* Pad the second cell until trailer */
				wpabs_memset((target_data_ptr+offset),PAD_CHAR,40);
				offset+=40;
				pad+=40;
				
				/* Insert trailer as last 8 bytes. The
				 * trailer contains dummy data right now
				 * because we still have to calculate the CRC
				 * and the overall packet length */
				wpabs_memcpy((target_data_ptr+offset),
						pAdapter->trailer,
						sizeof(pAdapter->trailer));
				
				offset+=sizeof(pAdapter->trailer);
			}else{
				/* The whole PDU fits into this frame */
				
				/* PAD all the unused bytes of the CELL that reside
				 * between ATM PDU data and the trailer */
				pad=((i+48)-(src_payload_size+sizeof(pAdapter->trailer)));
				if (pad > 0){
					wpabs_memset((target_data_ptr+offset),PAD_CHAR,pad);
					offset+=pad;
				}
								
				/* Insert trailer as last 8 bytes. The
				 * trailer contains dummy data right now
				 * because we still have to calculate the CRC
				 * and the overall packet length */
				wpabs_memcpy((target_data_ptr+offset),
						pAdapter->trailer,
						sizeof(pAdapter->trailer));
				
				wpabs_memcpy((target_data_ptr+offset+4),&crc,sizeof(long));
				offset+=sizeof(pAdapter->trailer);

				/* Set the END of FRAME bit in the PT section of
				 * the ATM header of this CELL, 
				 * this will indicate to the remote side 
				 * that this CELL is the last */
				wpabs_set_bit(1,&((target_data_ptr+offset-49)[0]));

				(target_data_ptr+offset-49)[1]=pAdapter->hec_eof;
			}
			break;
		}
		i = i+48;
	}/* for */
	
	/* End of SAR:
	 * The tx packet / ATM PDU packet, has been cut up and all 
	 * ATM CELLS are located in target_data_ptr (Tx dma buffer). 
	 * We must now calculate the CRC and LENGTH of the 
	 * ATM PDU and store those values into the trailer of 
	 * the last ATM CELL */
	
	/* Set the size of the tx DMA buffer that contains all
	 * the ATM CELLS. The offset will be equal to 52*number
	 * of cells created */
	wpabs_skb_trim(pAdapter->tx_cells_skb,offset);

	/* Calculate the CRC of the ATM-PDU. 
	 * Starting with ATM-PDU protocol header, tx packet 
	 * and any pad bytes that were inserted to fill in 
	 * the ATM CELL up to the trailer */
	crc=INITIAL_FCS_VALUE;
	crc=update_crc(crc, pAdapter->prot_header, pAdapter->prot_header_size);
	crc=update_crc(crc, wpabs_skb_data(tx_skb), wpabs_skb_len(tx_skb));
	crc=update_crc(crc,
		       (pAdapter->pad_len_buf+(sizeof(pAdapter->pad_len_buf)-(pad+4))),
		       (pad+4));
	crc=~crc;
	FLIP_CRC(crc,fin_crc);

	/* Add the calculated CRC value into the trailer of the
	 * last ATM CELL */
	wpabs_memcpy((target_data_ptr+offset-4),&fin_crc,sizeof(long));

	/* FIXME: tracing not enabled */
#if 0
	/* If tracing is enabled store the packet for tracing */
	if (atm_tracing_enabled(pAdapter->trace_info) >= 0){
		capture_trace_packet(pAdapter,	/*pLan->dev, */
				     tx_skb,
				     TRC_OUTGOING_FRM);
	}
#endif
	
	/* If the tx packet was an ATMARP, remember to reset
	 * the ATM-PDU header back to the IPoATM */
/* FIXME
	if (pAdapter->EncapMode==RFC_MODE_ROUTED_IP_LLC &&
	    *((unsigned short *)wpabs_skb_data(tx_skb)) == ATMARP_HEADER){
		pAdapter->prot_header[7]=0x00;
	}
*/

#if 0
	wpabs_debug_tx("ATM CELL TOTAL SiZE = %ld PAD=%d SBS=%d SPS=%d\n",
			phy_descr->Buffer->Size,
			/*ALEXphy_descr->Buffer->DmaData.length,*/
			pad, 
			src_data_size, 
			src_payload_size);
	wpabs_debug_tx("TX CELL 1: ");
	for (i=0; i < phy_descr->Buffer->Size; i++){
		if (i && !(i%52)){
			wpabs_debug_tx("\n");
			wpabs_debug_tx("\n");
			wpabs_debug_tx("TX CELL %03d:\n",(i/52)+1);
		}else{
			if (i && !(i%10)){
				wpabs_debug_tx("\n");
				wpabs_debug_tx("             ");
			}
		}
		wpabs_debug_tx("%02X ",*(target_data_ptr+i));
	}
	wpabs_debug_tx("\n");

#if 0
	wpabs_debug_tx("HOPEFULLY  CRC = %04lX ~CRC=%04lX\n",crc,~crc);
	

	crc=INITIAL_FCS_VALUE;
	crc = update_crc(crc,tmp_frame1,sizeof(tmp_frame1));
	wpabs_debug_event("ANOTHER  CRC = %04lX ~CRC=%04lX\n",crc,~crc);

	crc=INITIAL_FCS_VALUE;
	crc = update_crc(crc,pkt_data,(sizeof(pkt_data)-4));
	wpabs_debug_tx("ANOTHER KNOWN  CRC = %04lX ~CRC= %04lX\n",crc,~crc);
#endif
#endif


	/* FIXME: tracing not enabled */
#if 0
	/* If ATM CELL trace is enabled, trace all the ATM CELLS
	 * created in the TX DMA buffer */
	if (atm_tracing_enabled(pAdapter->trace_info) == 2){
		capture_atm_trace_packet(pAdapter,	
					 (void*)target_data_ptr,
					 phy_descr->Buffer->Size,
			        	 TRC_OUTGOING_FRM);
	}
#endif

	*tx_cell_skb = (void*)pAdapter->tx_cells_skb;
	return 0;
}

#if 0

#define MAX_ARP_SIZE  500
static int handle_atm_arp(adapter_t *pAdapter, void* rx_skb)
{		
	void*		tx_skb = NULL;
	char*		rx_data = wpabs_skb_data(rx_skb);
	atmarp_t*	atmarp = NULL;	
	atmarp_t	atmarp_reply;
	unsigned long 	local_ip;
	gsi_lan_t*	pLan = &(pAdapter->lan);


	if (pAdapter->config.interface == WAN_INTERFACE){
		return 0;
	}
	
	atmarp = (atmarp_t*)&rx_data[8];	
	if (atmarp->arphrd != ATMARP_HEADER){
		wpabs_debug_event("%s: Rx invalid arp pkt header 0x%X!\n",
				pAdapter->devname,atmarp->arphrd);
		return -EINVAL;
	}

	if (atmarp->arppro != ATMARP_PROT_IP){
		wpabs_debug_event("%s: Rx invalid arp pkt: protocol 0x%X not supported!\n",
				pAdapter->devname,atmarp->arppro);
		return -EINVAL;
	}

	switch (atmarp->arpop){

	case InARP_REQUEST:

		wpabs_debug_tx("%s: Rx InARP Request \n",
					pAdapter->devname);

		if (atmarp->arpshtl > sizeof(pAdapter->atm_addr)){
			wpabs_debug_event("%s: Rx invalid arp pkt: Source ATM address too big %d\n",
						pAdapter->devname,atmarp->arpshtl);
			return -EINVAL;
		}
		if (atmarp->arpsstl > sizeof(pAdapter->atm_addr)){
			wpabs_debug_event("%s: Rx invalid arp pkt: Source Sub ATM address too big %d\n",
						pAdapter->devname,atmarp->arpsstl);
			return -EINVAL;
		}

		if (atmarp->arpspln > 4){
			wpabs_debug_event("%s: Rx invalid arp pkt: Source IP ATM address too long %d\n",
						pAdapter->devname,atmarp->arpspln);
			return -EINVAL;
		}
			
		tx_skb = wpabs_skb_alloc(MAX_ARP_SIZE);
		if (!tx_skb){
			return -EINVAL;
		}
	
		atmarp_reply.arphrd 	= ATMARP_HEADER;
		atmarp_reply.arppro 	= ATMARP_PROT_IP;
		atmarp_reply.arpshtl	= sizeof(pAdapter->atm_addr);
		atmarp_reply.arpsstl	= 0;
		atmarp_reply.arpop  	= InARP_REPLY;
		atmarp_reply.arpspln	= 4;
		atmarp_reply.arpthtl	= atmarp->arpshtl;
		atmarp_reply.arptstl	= atmarp->arpsstl;
		atmarp_reply.arptpln	= atmarp->arpspln;
		wpabs_skb_copyback(tx_skb, 0, sizeof(atmarp_t), (unsigned long)&atmarp_reply);

		local_ip = wpabs_get_ip_addr(pLan->dev, WAN_LOCAL_IP);
		wpabs_debug_tx("Source ATM Len=%d, Sub Len=%d, IP Len=%d, %lx\n",
				atmarp->arpshtl, atmarp->arpsstl, atmarp->arpspln, local_ip);

		wpabs_skb_copyback(tx_skb, 
				 wpabs_skb_len(tx_skb), 
				 sizeof(pAdapter->atm_addr),
				 (unsigned long)pAdapter->atm_addr);

		wpabs_skb_copyback(tx_skb, 
				 wpabs_skb_len(tx_skb), 
				 sizeof(local_ip),
				 (unsigned long)&local_ip);

		if (atmarp->arpshtl){
			wpabs_skb_copyback(tx_skb, 
				 wpabs_skb_len(tx_skb), 
				 atmarp->arpshtl,
				 (unsigned long)(&rx_data[8]+sizeof(atmarp_t)));
		}

		if (atmarp->arpsstl){
			wpabs_skb_copyback(tx_skb, 
				 wpabs_skb_len(tx_skb), 
				 atmarp->arpsstl,
				 (unsigned long)(&rx_data[8]+sizeof(atmarp_t)+atmarp->arpshtl));
		}

		if (atmarp->arpspln){
			wpabs_skb_copyback(tx_skb, 
				 wpabs_skb_len(tx_skb), 
				 atmarp->arpshtl,
				 (unsigned long)(&rx_data[8]+sizeof(atmarp_t)+atmarp->arpshtl+atmarp->arpsstl));
		}

		/* Pad the outgoing ARP to 78 bytes */
		if (wpabs_skb_len(tx_skb) < 78){
			char* tx_data = wpabs_skb_put(tx_skb,78 - wpabs_skb_len(tx_skb));
			wpabs_memset(tx_data, 0, 78 - wpabs_skb_len(tx_skb));
		}	
		
		adsl_send(pAdapter, tx_skb, 0);

		wpabs_skb_free(tx_skb);	
		break;

	case InARP_REPLY:
	
		if (atmarp->arpspln != 0){
			unsigned long rmt_addr,remote_ip;
				
			rmt_addr= *(unsigned long*)(&rx_data[8]+sizeof(atmarp_t)+
				                 atmarp->arpshtl+atmarp->arpsstl);

			remote_ip = wpabs_get_ip_addr(pLan->dev, WAN_POINTOPOINT_IP);		

			if (rmt_addr != remote_ip){
				unsigned char ip1[25],ip2[25];
				wpabs_decode_ipaddr(rmt_addr,ip1);
				wpabs_decode_ipaddr(remote_ip,ip2);
				wpabs_debug_event("%s: Error Rx InARP Reply: Local P-to-P %s != Remote Src %s\n",
					pAdapter->devname,ip2,ip1);

				/* FIXME: Change the remote ip address automatically */
			}else{
				wpabs_debug_tx("%s: Rx InARP Reply OK\n",
					pAdapter->devname);
			}
		}else{
			wpabs_debug_event("%s: Error Rx InARP Reply: Remote Addr = 0.0.0.0\n",
					pAdapter->devname);
		}
		
		break;
	case ARP_NAK:
		wpabs_debug_event("%s: Rx ARP NAK\n",
					pAdapter->devname);
		break;

	case ARP_REQUEST:
		wpabs_debug_event("%s: Rx ARP Request \n",
					pAdapter->devname);
		break;
		
	case ARP_REPLY:
		wpabs_debug_event("%s: Rx ARP Reply \n",
					pAdapter->devname);
		break;

	default:
		wpabs_debug_event("%s: Rx Invalid arp pkt: Op Code 0x%x\n",
					pAdapter->devname,atmarp->arpop);
		return -EINVAL;
	}

	return 0;
}


static void* create_inarp_request(adapter_t *pAdapter)
{
	void*		tx_skb = NULL;
	atmarp_t	atmarp;	
	unsigned long 	local_ip;
	gsi_lan_t*	pLan = &(pAdapter->lan);
	unsigned int 	len;
	unsigned char*  tx_data;
	

	tx_skb = wpabs_skb_alloc(MAX_ARP_SIZE);
	if (!tx_skb){
		return NULL;
	}
	
	atmarp.arphrd 	= ATMARP_HEADER;
	atmarp.arppro 	= ATMARP_PROT_IP;
	atmarp.arpshtl	= sizeof(pAdapter->atm_addr);
	atmarp.arpsstl	= 0;
	atmarp.arpop  	= InARP_REQUEST;
	atmarp.arpspln	= 4;
	atmarp.arpthtl	= 0;
	atmarp.arptstl	= 0;
	atmarp.arptpln	= 0;
	
	wpabs_skb_copyback(tx_skb, 0, sizeof(atmarp_t), (unsigned long)&atmarp);

	local_ip = wpabs_get_ip_addr(pLan->dev, WAN_LOCAL_IP);

	wpabs_skb_copyback(tx_skb, 
			 wpabs_skb_len(tx_skb), 
			 sizeof(pAdapter->atm_addr),
			 (unsigned long)pAdapter->atm_addr);

	wpabs_skb_copyback(tx_skb, 
			 wpabs_skb_len(tx_skb), 
			 sizeof(local_ip),
			 (unsigned long)&local_ip);

	if ((len=wpabs_skb_len(tx_skb)) < 78){
		len=78-len;
		tx_data = wpabs_skb_put(tx_skb,len);
		wpabs_memset(tx_data, 0, len);
	}	


	return tx_skb;
}

/*===============================================================
 * adsl_arp_check
 * 
 * Called by GpMonitorTransceiver every 
 * GSI_CHECK_TRAIN_TIMEOUT seconds.  
 *
 * For Classical IP over ATM ONLY !!!
 * After 10 seconds we should send another
 * InARP request.
 * 
 * For all other protocols encapsulations
 * return and do nothing.
 *
 */
void adsl_arp_check (adapter_t *pAdapter)
{
	if (pAdapter->config.interface == WAN_INTERFACE){
		return;
	}

	if (wpabs_test_bit(0, &pAdapter->atm_autocfg_running)){
		return;
	}

	if (wpabs_get_state(pAdapter->card) != WAN_CONNECTED){
		pAdapter->arp_timeout=10;
		return;
	}
	
	if (pAdapter->EncapMode != RFC_MODE_ROUTED_IP_LLC){
		return;
	}
	
	pAdapter->arp_timeout += GSI_CHECK_TRAIN_TIMEOUT;

	if (pAdapter->arp_timeout >= 10){
		void *tx_skb=NULL;
		
		wpabs_debug_tx("%s: Sending out InARP Request \n",
				pAdapter->devname);	

		tx_skb=create_inarp_request(pAdapter);
		if (tx_skb){
			adsl_send(pAdapter, tx_skb, 0);
			wpabs_skb_free(tx_skb);
		}
		
		pAdapter->arp_timeout=0;
	}
	return; 
}
#endif


static int parse_prot_header(wp_sar_dev_t *pAdapter, void* rx_skb)
{
	unsigned char* data = (unsigned char*)wpabs_skb_data(rx_skb);
	int len = wpabs_skb_len(rx_skb);
	
	switch (pAdapter->EncapMode){
			
		case RFC_MODE_ROUTED_IP_LLC:
			if (data[6] == 0x08){
				if (data[7] == 0x06){
					wpabs_debug_rx("%s: RX ARP packet, sending reply...\n",
								pAdapter->devname);
					
					/* FIXME atm arp not supported
					 * handle_atm_arp(pAdapter, rx_skb);	
					 */
					
					/* We don't want to send this rx packet 
					 * up the stack, since we just replied to the
					 * arp. We should kill this packet thus, we return -1 */
					return -1;
				}
			}else{
				char temp[200];
				wpabs_debug_event("%s: Rx Pkt with invalid protocol header!\n",
							pAdapter->devname);
				wpabs_detect_prot_header(data, len, temp);
				wpabs_debug_event("%s: Not LLC IP/ATM: Detected as: %s!\n",
							pAdapter->devname, 
							temp);
				return -1;
			}
			break;

		case RFC_MODE_ROUTED_IP_VC:
			if (data[0] != 0x45){
				char temp[200];
				wpabs_debug_event("%s: Rx Pkt with invalid protocol header!\n",
							pAdapter->devname);
				wpabs_detect_prot_header(data, len, temp);
				wpabs_debug_event("%s: Not VC IP/ATM: Detected as: %s!\n",
							pAdapter->devname, 
							temp);
				return -1;
			}
			break;
			
		case RFC_MODE_BRIDGED_ETH_LLC:
			if (data[6] != 0x00 || data[7] != 0x07){
				char temp[200];
				wpabs_debug_event("%s: Rx Pkt with invalid protocol header!\n",
							pAdapter->devname);
				wpabs_detect_prot_header(data, len, temp);
				wpabs_debug_event("%s: Not Bridged LLC/ATM: Detected as: %s\n",
							pAdapter->devname,
							temp);
				return -1;
			}
			break;

		case RFC_MODE_BRIDGED_ETH_VC:
			{
				int err = wpabs_detect_prot_header(data, len, NULL);
				if (err == 0){
					char temp[200];
					err = wpabs_detect_prot_header(data, len, temp);
					wpabs_debug_event("%s: Rx Pkt with invalid protocol header!\n",
							pAdapter->devname);
					wpabs_debug_event("%s: Not Bridged VC/ATM: Detected as: %s\n",
							pAdapter->devname,
							temp);
				}
			}
			break;

		case RFC_MODE_PPP_LLC:

			if (data[0] != 0xFE || data[1] != 0xFE){
				char temp[200];
				wpabs_debug_event("%s: Rx Pkt with invalid protocol header!\n",
						pAdapter->devname);
				wpabs_detect_prot_header(data, len, temp);
				wpabs_debug_event("%s: Not LLC PPP/ATM: Detected as: %s!\n",
						pAdapter->devname, temp);
				return -1;
			}
			break;

		case RFC_MODE_PPP_VC:
			if (data[0] == 0xC0 && data[1] == 0x21){
				return 0;
			}else if (data[0] == 0x00 && data[1] == 0x21){
				return 0;
			}else{
				int err = wpabs_detect_prot_header(data, len, NULL);
				if (err == 0){
					char temp[200];
					err = wpabs_detect_prot_header(data, len, temp);
					wpabs_debug_event("%s: Rx Pkt with invalid protocol header!\n",
							pAdapter->devname);
					wpabs_debug_event("%s: Not VC PPP/ATM: Detected as: %s\n",
							pAdapter->devname,
							temp);
				}
			}
			break;

		default:
			/* This will never happen, just for completeness sake */
			wpabs_debug_event("%s:Rx Configuration error, unkown mode 0x%x\n",
					pAdapter->devname,pAdapter->EncapMode);
			return -1;
	}
	
	return 0;
}


#if 0
static int 
capture_atm_trace_packet(
		adapter_t*	pAdapter,
		void 		*data,
		int		len,
		char		direction)
{
	int atm_packets=0;
	int i;
	wan_trace_pkt_t trc_el;		
	volatile GP_CELL *atm_cell_ptr=(GP_CELL*)data;
	void *new_skb=NULL;
	
	if (adsl_tracing_enabled(pAdapter->trace_info) != 2){
		return 0;
	}

	atm_packets = len/sizeof(GP_CELL);

	if (atm_packets <= 0){
		return -EINVAL;
	}
	
	for (i=0;i<atm_packets;i++){

		new_skb = wpabs_skb_alloc(sizeof(GP_CELL)+sizeof(wan_trace_pkt_t));
		if (new_skb == NULL) 
			return -ENOMEM;
		
		trc_el.status		= direction;
		trc_el.data_avail	= 1;
		trc_el.time_stamp	= (unsigned short)(wpabs_get_systemticks() % 0xFFFF);
		trc_el.real_length	= sizeof(GP_CELL);
	
		wpabs_skb_copyback(new_skb, 
				  wpabs_skb_len(new_skb), 
				  sizeof(wan_trace_pkt_t), 
				  (unsigned long)&trc_el);

		wpabs_skb_copyback(new_skb, 
				  wpabs_skb_len(new_skb), 
				  sizeof(GP_CELL),
				  (unsigned long)atm_cell_ptr);

		adsl_trace_enqueue(pAdapter->trace_info, new_skb);

		atm_cell_ptr++;
	}

	return 0;
}

static int 
capture_trace_packet(
		adapter_t*	pAdapter,
		void*		skb,
		char		direction)
{
	int		len = 0;
	void*		new_skb = NULL;
	wan_trace_pkt_t trc_el;	
	int		flag = 0;
	
	if ((flag = adsl_tracing_enabled(pAdapter->trace_info)) >= 0){

		if (flag > 1){
			return 0;
		}
			
		/* Allocate a header mbuf */
		len = 	wpabs_skb_len(skb) +  
			sizeof(wan_trace_pkt_t);

		if (!flag && direction == TRC_INCOMING_FRM){
			if ((wpabs_skb_len(skb) - pAdapter->prot_header_size) <= 0){
				return -EINVAL;
			}
			len -= pAdapter->prot_header_size;
		}

		new_skb = wpabs_skb_alloc(len+pAdapter->prot_header_size);
		if (new_skb == NULL) return -ENOMEM;

		trc_el.status		= direction;
		trc_el.data_avail	= 1;
		trc_el.time_stamp	= (unsigned short)(wpabs_get_systemticks() % 0xFFFF);
		trc_el.real_length	= wpabs_skb_len(skb);
		
		if (direction == TRC_OUTGOING_FRM){
			if (!flag){

				wpabs_skb_copyback(new_skb, 
					  wpabs_skb_len(new_skb), 
				   	  sizeof(wan_trace_pkt_t), 
					  (unsigned long)&trc_el);
	
				wpabs_skb_copyback(new_skb, 
					  wpabs_skb_len(new_skb), 
					  wpabs_skb_len(skb),
					  (unsigned long)wpabs_skb_data(skb));
			}else{
				trc_el.real_length += pAdapter->prot_header_size;

				wpabs_skb_copyback(new_skb, 
				  	wpabs_skb_len(new_skb), 
			   	  	sizeof(wan_trace_pkt_t), 
				  	(unsigned long)&trc_el);

				
				if (pAdapter->prot_header_size){
					wpabs_skb_copyback(new_skb, 
						wpabs_skb_len(new_skb), 
					  	pAdapter->prot_header_size,
					  	(unsigned long)pAdapter->prot_header);
				}
				
				wpabs_skb_copyback(new_skb, 
					  	  wpabs_skb_len(new_skb), 
					  	  wpabs_skb_len(skb),
					  	  (unsigned long)wpabs_skb_data(skb));
			}
		}else{
			if (!flag){
				trc_el.real_length -= pAdapter->prot_header_size;

				wpabs_skb_copyback(new_skb, 
					  wpabs_skb_len(new_skb), 
				   	  sizeof(wan_trace_pkt_t), 
					  (unsigned long)&trc_el);

				wpabs_skb_copyback(new_skb, 
					  wpabs_skb_len(new_skb), 
					  wpabs_skb_len(skb)-pAdapter->prot_header_size,
					  (unsigned long)wpabs_skb_data(skb)+pAdapter->prot_header_size);
			}else{
			
				wpabs_skb_copyback(new_skb, 
					  wpabs_skb_len(new_skb), 
				   	  sizeof(wan_trace_pkt_t), 
					  (unsigned long)&trc_el);

				
				wpabs_skb_copyback(new_skb, 
					  wpabs_skb_len(new_skb), 
					  wpabs_skb_len(skb),
					  (unsigned long)wpabs_skb_data(skb));
			}
		}
		
		adsl_trace_enqueue(pAdapter->trace_info, new_skb);
	}

	return 0;
}
#endif

#define OAM_F5_END_TO_END_CELL 	0x05
#define OAM_F5_SEGMENT 		0x04

/* OAM Types */
#define OAM_FAULT_MGMT		0x01
#define OAM_PERF_MGMT		0x02
#define OAM_ACT_DEACT		0x80

/* OAM Function Types */
#define OAM_FAULT_FUNC_ALARM_INDIC	0x00
#define OAM_FAULT_FUNC_FERF		0x01
#define OAM_FAULT_FUNC_CELL_LOOPBACK	0x08
#define OAM_FAULT_FUNC_CONTINUITY	0x04

#define OAM_PERF_FUNC_FW_MONITOR	0x00
#define OAM_PERF_FUNC_BW_REPORT		0x01
#define OAM_PERF_FUNC_MON_REPORT	0x02

#define OAM_AD_FUNC_PERF_MONITOR	0x00
#define OAM_AD_FUNC_CONTINUITY		0x01

#define OAM_CELL_LOOPBACK 		0x18
#define OAM_LOOPBACK_INDICATION 	0x01
#define ATM_PAYLOAD_SZ 			48

#if 0
void handle_atm_admin_packet(adapter_t *pAdapter, void *data)
{
	volatile GP_CELL *atm_cell_ptr=(GP_CELL*)data;
	unsigned char oam_type, oam_func;
	unsigned char pt = (atm_cell_ptr->hdr[3] >> 1) & 0x07;
	
	unsigned short vpi = WPVPI(atm_cell_ptr);
 	unsigned short vci = WPVCI(atm_cell_ptr);

	if (pAdapter->vpi != vpi){
		wpabs_debug_event("%s: Rx invalid ATM Mgmt Cell: Vpi Mismatch Cfg Vpi=%d In Vpi=%d\n",
				pAdapter->devname,pAdapter->vpi,vpi);
		return;
	}

	if ((pt == OAM_F5_END_TO_END_CELL || pt == OAM_F5_SEGMENT) && pAdapter->vci == vci){
		goto oam_f5_packet;
		
	}else if (pAdapter->vci == 3 || pAdapter->vci == 4){
		goto oam_f4_packet;

	}else{
		wpabs_debug_event("%s: Rx invalid ATM Mgmt Cell: Vpi=%d Vci=%d Pt=0x%x\n",
				pAdapter->devname,vpi,vci,pt);
		return;
	}

oam_f4_packet:
oam_f5_packet:

	oam_type = (atm_cell_ptr->info[0] >> 4) & 0x0F;
	oam_func = atm_cell_ptr->info[0] & 0x0F;

	switch (oam_type){

	case OAM_FAULT_MGMT:

		switch (oam_func){
			
			case OAM_FAULT_FUNC_CELL_LOOPBACK:
				handle_oam_cell_loopback(pAdapter,(void*)atm_cell_ptr);
				break;
				
			case OAM_FAULT_FUNC_ALARM_INDIC:
			case OAM_FAULT_FUNC_FERF:	
			case OAM_FAULT_FUNC_CONTINUITY:
			default:
				wpabs_debug_event("%s: Rx Invalid ATM OAM Fault Mgmt Cell: Type=0x%x Func=0x%x\n",
						pAdapter->devname,oam_type,oam_func);
				break;
		}
		break;	
	
	case OAM_PERF_MGMT:
		wpabs_debug_event("%s: Rx unsupported ATM OAM Perf Mgmt Cell: Type=0x%x Func=0x%x\n",
						pAdapter->devname,oam_type,oam_func);
		break;

	case OAM_ACT_DEACT:
		wpabs_debug_event("%s: Rx unsupported ATM OAM Act/Deact Cell: Type=0x%x Func=0x%x\n",
						pAdapter->devname,oam_type,oam_func);
		break;
		
	default:
		wpabs_debug_event("%s: Rx unsupported ATM OAM Cell: Type=0x%x Func=0x%x\n",
						pAdapter->devname,oam_type,oam_func);
		
		break;
	}
	
}


void handle_oam_cell_loopback(adapter_t *pAdapter, void *data)
{
	int i;
	volatile GP_CELL *atm_cell_ptr=(GP_CELL*)data;
	PCI_PHY *pPci = &(pAdapter->pciPhy);
	PHY_DESCR *phy_descr;
	unsigned char *tx_data_ptr;
	volatile unsigned short crc10_accum=0;


	if (atm_cell_ptr->info[1] != OAM_LOOPBACK_INDICATION){
		wpabs_debug_event("%s: Rx Invalid ATM OAM Loopback Cell: Not Loobpack 0x%x\n",
				pAdapter->devname,atm_cell_ptr->info[1]);
		return;
	}
	
	/* Clear the OAM F5 Loopback request, thus
	 * indicating a reply */
	atm_cell_ptr->info[1]=0x00;

	if (IS_LIST_EMPTY(&(pPci->TxFreeList))){
		/* If tx list is empty, silently discart the
		 * received packet, and wait for another one */
		return;
	}
	
	phy_descr= (PHY_DESCR*)PHY_REMOVE_LIST(&(pPci->TxFreeList));
	if (!phy_descr){
		/* If tx list is empty, silently discart the
		 * received packet, and wait for another one */
		return;
	}

	phy_descr->Buffer->Size=sizeof(GP_CELL);
	tx_data_ptr=(unsigned char *)phy_descr->Buffer->Data;
	wpabs_memcpy(tx_data_ptr,(unsigned char*)atm_cell_ptr,sizeof(GP_CELL));

	/* Clear the old CRC values */
	tx_data_ptr[4+ATM_PAYLOAD_SZ-2]=0;
	tx_data_ptr[4+ATM_PAYLOAD_SZ-1]=0;

	for ( i = 0;  i < ATM_PAYLOAD_SZ; i++ ){
        	crc10_accum = ((crc10_accum << 8) & 0x3ff)
                	       ^ byte_crc10_table[( crc10_accum >> 2) & 0xff]
                               ^ *((unsigned char *)(tx_data_ptr+4+i));
    	}

	tx_data_ptr[4+ATM_PAYLOAD_SZ-2] = (crc10_accum >> 8);
	tx_data_ptr[4+ATM_PAYLOAD_SZ-1] = (crc10_accum & 0xff);

	if (adsl_tracing_enabled(pAdapter->trace_info) == 2){
		capture_atm_trace_packet(pAdapter,	
					 (void*)tx_data_ptr,
					 phy_descr->Buffer->Size,
			        	 TRC_OUTGOING_FRM);
	}

	PHY_ADD_LIST(&(pPci->TxPendingList),phy_descr);	

	gsi_dma_tx_start(pAdapter);	
}

#endif

/*========================================================
 * is_idle_cell
 * 
 * 	Check if the incoming ATM Cell is 
 * 	IDLE or UNASSIGNED.
 */

static __inline unsigned char is_idle_cell(unsigned int *data)
{
	unsigned long atm_header = *data;

	/* Idle Cell or Physical Layer OAM Cell */
	if (atm_header == 0x01000000 || atm_header == 0x03000000){
		return 1;
	}

	atm_header &= 0xf1ffff0f;

	/* Reserved for Physical Layer or Unassigned Cell */
	if (atm_header == 0x00000000 || atm_header == 0x01000000){
		return 1;
	}

	/* Data Cell */
	return 0;
}



/*===============================================================
 * wphash_sar_dev
 * wpunhash_sar_dev
 * wp_find_sar_dev_by_vpivci
 *
 *  	Hashing functions used to map VPI/VCI to PVC devices
 */

static __inline void wphash_sar_dev(wp_atm_reg_t *atm_device, wp_sar_dev_t *p)
{
	unsigned int hashid=*((unsigned int*)p->atm_header)&~(ATM_HEADER_MASK);
	wp_sar_dev_t **htable = (wp_sar_dev_t **)&atm_device->atm_hash[vpivci_hashfn(hashid)];

	if (wpabs_test_bit(0,&p->init)){
		wpabs_debug_event("Critical Error: Init Sar in wphash_sar_dev!\n");
		return;
	}	

	if((p->hash_next = *htable) != NULL)
		(*htable)->hash_pprev = &p->hash_next;
	
	*htable = p;
	p->hash_pprev = htable;

	wpabs_set_bit(0,&p->init);
}

static __inline void wpunhash_sar_dev(wp_sar_dev_t *p)
{
	if(p->hash_next)
		p->hash_next->hash_pprev = p->hash_pprev;
	*p->hash_pprev = p->hash_next;

	wpabs_clear_bit(0,&p->init);
}

static __inline wp_sar_dev_t *wp_find_sar_dev_by_vpivci(wp_atm_reg_t *atm_device, 
		                               unsigned int key)
{
	wp_sar_dev_t *p, **htable;
	
	key&=~(ATM_HEADER_MASK);
	
	htable = (wp_sar_dev_t **)&atm_device->atm_hash[vpivci_hashfn(key)];
	
	for(p = *htable; p ;p = p->hash_next){
		
		if (((*(unsigned int*)p->atm_header)&~(ATM_HEADER_MASK)) == key)
			break;
	}

	return p;
}




/*==SECTION=========================================================
 *
 * CRC Routines
 * 
 *=================================================================*/


/* crc32h.c -- package to compute 32-bit CRC one byte at a time using   */
/*             the high-bit first (Big-Endian) bit ordering convention  */
/*                                                                      */
/* Synopsis:                                                            */
/*  gen_crc_table() -- generates a 256-word table containing all CRC    */
/*                     remainders for every possible 8-bit byte.  It    */
/*                     must be executed (once) before any CRC updates.  */
/*                                                                      */
/*  unsigned update_crc(crc_accum, data_blk_ptr, data_blk_size)         */
/*           unsigned crc_accum; char *data_blk_ptr; int data_blk_size; */
/*           Returns the updated value of the CRC accumulator after     */
/*           processing each byte in the addressed block of data.       */
/*                                                                      */
/*  It is assumed that an unsigned long is at least 32 bits wide and    */
/*  that the predefined type char occupies one 8-bit byte of storage.   */
/*                                                                      */
/*  The generator polynomial used for this version of the package is    */
/*  x^32+x^26+x^23+x^22+x^16+x^12+x^11+x^10+x^8+x^7+x^5+x^4+x^2+x^1+x^0 */
/*  as specified in the Autodin/Ethernet/ADCCP protocol standards.      */
/*  Other degree 32 polynomials may be substituted by re-defining the   */
/*  symbol POLYNOMIAL below.  Lower degree polynomials must first be    */
/*  multiplied by an appropriate power of x.  The representation used   */
/*  is that the coefficient of x^0 is stored in the LSB of the 32-bit   */
/*  word and the coefficient of x^31 is stored in the most significant  */
/*  bit.  The CRC is to be appended to the data most significant byte   */
/*  first.  For those protocols in which bytes are transmitted MSB      */
/*  first and in the same order as they are encountered in the block    */
/*  this convention results in the CRC remainder being transmitted with */
/*  the coefficient of x^31 first and with that of x^0 last (just as    */
/*  would be done by a hardware shift register mechanization).          */
/*                                                                      */
/*  The table lookup technique was adapted from the algorithm described */
/*  by Avram Perez, Byte-wise CRC Calculations, IEEE Micro 3, 40 (1983).*/

#define POLYNOMIAL 0x04c11db7L


static void gen_crc_table(void)
{ 
   register int i, j;  register unsigned long crc_accum;
   for ( i = 0;  i < 256;  i++ )
       { crc_accum = ( (unsigned long) i << 24 );
         for ( j = 0;  j < 8;  j++ )
              { if ( crc_accum & 0x80000000L )
                   crc_accum =
                     ( crc_accum << 1 ) ^ POLYNOMIAL;
                else
                   crc_accum =
                     ( crc_accum << 1 ); }
         crc_table[i] = crc_accum; }
   return; 
}

static unsigned long update_crc(unsigned long crc_accum, char *data_blk_ptr,
                                                    int data_blk_size)
{ 	
	register int i, j;
   	for ( j = 0;  j < data_blk_size;  j++ )
       	{ i = ( (int) ( crc_accum >> 24) ^ *data_blk_ptr++ ) & 0xff;
         crc_accum = ( crc_accum << 8 ) ^ crc_table[i]; }
   	return crc_accum; 
}

#define CRC10_POLYNOMIAL 0x633
static void gen_byte_crc10_table(void)
/* generate the table of CRC-10 remainders for all possible bytes */
{
    register int i, j;
    register unsigned short crc10_accum;

    for ( i = 0;  i < 256;  i++ )
    {
        crc10_accum = ((unsigned short) i << 2);
        for ( j = 0;  j < 8;  j++ )
        {
            if ((crc10_accum <<= 1) & 0x400) crc10_accum ^= CRC10_POLYNOMIAL;
        }
        byte_crc10_table[i] = crc10_accum;
    }
    return;
}




/*===============================================================
 * generate_HEC_chk_table
 *
 * GENERATE A TABLE FOR CHECKING THE HEC VALUE IN A CELL 
 *
 */
#define HEC_GENERATOR   	   0x107
#define I361_HEC_COSET_FUNCTION    0x55
 
void generate_hec_crc_table(void)
{
	int HEC_syndrome;
	int i,j;

	for(i = 0; i < 256; i++) {
		HEC_syndrome = i;
		for(j = 0; j < 8; j++) {
			if(HEC_syndrome & 0x80) {
				HEC_syndrome = (HEC_syndrome << 1) ^ HEC_GENERATOR;
			}else{
				HEC_syndrome = (HEC_syndrome << 1);
			}
		}
		HEC_table[i] = (unsigned char)HEC_syndrome;
    }
}



/**** COMPUTE THE HEC VALUE OF THE CELL CURRENTLY IN THE SCRATCHPAD BUFFER ****/

unsigned char calculate_hec_crc(unsigned char *util_ptr)
{

	/* zero the HEC accumulator */
	unsigned char HEC_accum = 0;
	unsigned int i; 
	unsigned char util_char;
	
	/* set a pointer to the cell scratchpad buffer */
	/* process the first four bytes in the ATM header */
	for(i = 0; i < (sizeof(ATM_HEADER_SZ) - 1); i ++) {
		/* update the HEC value for each byte in the cell header */
		HEC_accum = HEC_table[HEC_accum ^ *util_ptr++];
	}

	/* in general, the cell header check sum is XORed with the COSET 
	 * function defined in CCITT I.361 (0x55) 
	 * */
	util_char = I361_HEC_COSET_FUNCTION;		
	
	HEC_accum ^= util_char;		

	return HEC_accum;
}
