

#if defined(__LINUX__)
# include <linux/wanpipe_includes.h>
# include <linux/wanpipe.h>
# include <linux/wanpipe_cfg.h>
#else
# error "This Operating System is not supported!"
#endif

#include "wan_aften.h"
#include "wan_aften_api.h"

//#define DEBUG

static char	*wan_drvname = "wan_aften";
static int ncards; 

static sdla_t* card_list;	/* adapter data space */
extern wan_iface_t wan_iface;

static int wan_aftup_init(void);
static int wan_aftup_exit(void);
static int wan_aftup_setup(sdla_t *card, netdevice_t *dev);
static int wan_aftup_shutdown(sdla_t *card);
static int wan_aftup_hwprobe(void);
static int wan_aftup_open(netdevice_t *dev);
static int wan_aftup_close(netdevice_t *dev);
static int wan_aftup_ioctl (netdevice_t *dev, struct ifreq *ifr, int cmd);

WAN_MODULE_DEFINE(
	wan_aftup, 
	"Alex Feldman <al.feldman@sangoma.com>",
	"Wanpipe AFT Update", 
	"GPL",
	wan_aftup_init, wan_aftup_exit);


static int wan_aftup_init(void)
{
	struct wan_aftup_priv	*priv = NULL;
	struct wan_dev_le	*devle;
	sdla_t			*card;
	static int		if_index = 0;
	int 			err = 0, i;
	
	ncards = sdla_hw_probe();
	if (!ncards){
		DEBUG_EVENT("No Sangoma cards found, unloading modules!\n");
		return -ENODEV;
	}

	for (i=0;i<ncards;i++){
		wan_device_t*	wandev;
		netdevice_t*	dev;

		card=wan_malloc(sizeof(sdla_t));
		if (!card){
			DEBUG_EVENT("%s: Failed allocate new card!\n", 
					wan_drvname);
			goto wanpipe_init_done;
		}
		memset(card, 0, sizeof(sdla_t));
		/* Allocate HDLC device */
		if (!wan_iface.alloc || (dev = wan_iface.alloc(1)) == NULL){
			goto wanpipe_init_done;
		}
		devle = wan_malloc(sizeof(struct wan_dev_le));
		if (devle == NULL){
			DEBUG_EVENT("%s: Failed allocate memory!\n",
					wan_drvname);
			goto wanpipe_init_done;
		}
		if ((priv = wan_malloc(sizeof(struct wan_aftup_priv))) == NULL){
			DEBUG_EVENT("%s: Failed to allocate priv structure!\n",
				wan_drvname);
			goto wanpipe_init_done;
		}
		priv->common.is_netdev		= 1;
		priv->common.iface.open		= &wan_aftup_open;
		priv->common.iface.close	= &wan_aftup_close;
		priv->common.iface.ioctl	= &wan_aftup_ioctl;
		priv->common.card		= card;
		wan_netif_set_priv(dev, priv);
		sprintf(card->devname, "hdlc%d", if_index++);
		/* Register in HDLC device */
		if (!wan_iface.attach || wan_iface.attach(dev, card->devname, 1)){
			wan_free(devle);
			if (wan_iface.free) wan_iface.free(dev);
			goto wanpipe_init_done;
		}
		wandev 			= &card->wandev;
		wandev->magic   	= ROUTER_MAGIC;
		wandev->name    	= card->devname;
		wandev->private 	= card;
		devle->dev		= dev;
		/* Set device pointer */
		WAN_LIST_INIT(&wandev->dev_head);
		WAN_LIST_INSERT_HEAD(&wandev->dev_head, devle, dev_link);
		card->list	= card_list;
		card_list	= card;
		if (wan_aftup_setup(card, dev)){
			DEBUG_EVENT("%s: Failed setup new device!\n", 
					card->devname);
			WAN_LIST_REMOVE(devle, dev_link);
			wan_free(devle);
			if (wan_iface.detach) wan_iface.detach(dev, 1);
			if (wan_iface.free) wan_iface.free(dev);
			card_list	= card->list;
			wan_free(card);
			goto wanpipe_init_done;
		}
	}
	wan_aftup_hwprobe();

wanpipe_init_done:
	if (err) wan_aftup_exit();
	return err;
}

static int wan_aftup_exit(void)
{
	struct wan_dev_le	*devle;
	sdla_t *card;
	int 	err = 0;

	for (card=card_list;card_list;){
		devle = WAN_LIST_FIRST(&card->wandev.dev_head);
		if (devle && devle->dev){
			struct wan_aftup_priv	*priv = wan_netif_priv(devle->dev);
			DEBUG_EVENT("%s: Unregistering interface...\n", 
					wan_netif_name(devle->dev));
			if (wan_iface.detach) wan_iface.detach(devle->dev, 1);
			wan_free(priv);
			if (wan_iface.free) wan_iface.free(devle->dev);
			WAN_LIST_REMOVE(devle, dev_link);
			wan_free(devle);
		}
		DEBUG_EVENT("%s: Shutdown device\n", card->devname);
		wan_aftup_shutdown(card);
		card_list = card->list;
		wan_free(card);
		card = card_list;
	}

	card_list=NULL;
	DEBUG_EVENT("\n");
	DEBUG_EVENT("%s: WANPIPE Generic Modules Unloaded.\n",
						wan_drvname);

#if defined(WAN_DEBUG_MEM)
	DEBUG_EVENT("%s: Total Mem %d\n",
			wan_drvname, atomic_read(&wan_debug_mem));
#endif	
	return err;
}

static int wan_aftup_setup(sdla_t *card, netdevice_t *dev)
{
	struct wan_aftup_priv	*priv = wan_netif_priv(dev);
	int			err;

	card->hw = sdla_register(&card->hw_iface, NULL, card->devname);
	if (card->hw == NULL){
		DEBUG_EVENT("%s: Failed to register hw device\n",
				card->devname);
		goto wan_aftup_setup_done;
	}

	err = card->hw_iface.setup(card->hw, NULL);
	if (err){
		DEBUG_EVENT("%s: Hardware setup Failed %d\n",
				card->devname,err);
		sdla_unregister(&card->hw, card->devname);
		goto wan_aftup_setup_done;
	}

	card->hw_iface.getcfg(card->hw, SDLA_CPU, &priv->irq);
	DEBUG_EVENT("ALEX> CPU %d\n", priv->irq);

	card->hw_iface.pci_read_config_dword(
			card->hw, 0x4, &priv->base_class);
	card->hw_iface.pci_read_config_dword(
			card->hw, PCI_IO_BASE_DWORD, &priv->base_addr0);
	card->hw_iface.pci_read_config_dword(
			card->hw, PCI_MEM_BASE0_DWORD, &priv->base_addr1);
	card->hw_iface.getcfg(card->hw, SDLA_IRQ, &priv->irq);	
	//DEBUG_EVENT("%s: BaseClass %X BaseAddr 0x%X IRQ %d\n", 
	//		wan_netif_name(dev),
	//		priv->base_class,
	//		priv->base_addr,
	//		priv->irq);
	return 0;

wan_aftup_setup_done:
	return -EINVAL;

}

static int wan_aftup_shutdown(sdla_t *card)
{
	
	if (card->hw_iface.down){
		card->hw_iface.down(card->hw);
	}
	sdla_unregister(&card->hw, card->devname);

	return 0;
}

static int wan_aftup_hwprobe(void)
{
	sdla_t		*card;
	unsigned char	*hwprobe;
	int		err;

	DEBUG_EVENT("\n\nList of available Sangoma devices:\n");
	for (card=card_list; card; card = card->list){
		struct wan_dev_le	*devle;
		devle = WAN_LIST_FIRST(&card->wandev.dev_head);
		if (devle && devle->dev){
			err = card->hw_iface.get_hwprobe(
					card->hw, 0, (void**)&hwprobe); 
			if (!err){
				hwprobe[strlen(hwprobe)] = '\0';
				DEBUG_EVENT("%s: %s\n",
						wan_netif_name(devle->dev),
						hwprobe);
			}
		}
	}
	return 0;
}

static int wan_aftup_open(netdevice_t *dev)
{
	return 0;
}

static int wan_aftup_close(netdevice_t *dev)
{
	return 0;
}

static int wan_aftup_ioctl (netdevice_t *dev, struct ifreq *ifr, int cmd)
{
	sdla_t			*card;
	struct wan_aftup_priv	*priv= wan_netif_priv(dev);
	struct wan_aftup_api	api_cmd;
	unsigned char		*str;
	int			err=-EINVAL;
		
	if (!priv || !priv->common.card){
		DEBUG_EVENT("%s: Invalid structures!\n", wan_netif_name(dev));
		return -ENODEV;
	}

	if (!ifr->ifr_data){
		DEBUG_EVENT("%s: No API data!\n", wan_netif_name(dev));
		return -EINVAL;
	}

	card = priv->common.card;
	if (WAN_COPY_FROM_USER(&api_cmd,ifr->ifr_data,sizeof(struct wan_aftup_api))){
		return -EFAULT;
	}

	/* Hardcode bar access FELD */
	switch (cmd){

	case SIOC_WAN_READ_REG:

//		if (api_cmd.bar == 0){
//			vector=chan->bar0 + api_cmd.offset;
//		}else{
//			vector=chan->bar1 + api_cmd.offset;
//		}
	
		if (api_cmd.len == 1){
			card->hw_iface.bus_read_1(
					card->hw,
				       	api_cmd.offset,
				       	(unsigned char*)&api_cmd.data[0]);
		}else if (api_cmd.len == 2){
			card->hw_iface.bus_read_2(
					card->hw,
				       	api_cmd.offset,
				       	(unsigned short*)&api_cmd.data[0]);
		}else if (api_cmd.len == 4){
			card->hw_iface.bus_read_4(
					card->hw,
				       	api_cmd.offset,
				       	(unsigned int*)&api_cmd.data[0]);
		}else{
			card->hw_iface.peek(
					card->hw,
					api_cmd.offset,
					(unsigned char*)&api_cmd.data[0],
					api_cmd.len);
			//memcpy_fromio((unsigned char*)&api_cmd.data[0],
			//		(unsigned char*)vector, api_cmd.len);
		}
		DEBUG_TEST("%s: Reading Bar%i Offset=0x%X Len=%i Data=0x%X\n",
					wan_netif_name(dev),
					api_cmd.bar,
					api_cmd.offset,
					api_cmd.len,
					api_cmd.data[0]);
		
		err=0;
		break;
	
	case SIOC_WAN_WRITE_REG:

//		if (api_cmd.bar == 0){
//			vector=chan->bar0 + api_cmd.offset;
//		}else{
//			vector=chan->bar1 + api_cmd.offset;
//		}

		DEBUG_TEST("%s: Write  Offset=0x%04X Data=0x%08X\n",
				wan_netif_name(dev),api_cmd.offset,
				*(unsigned int*)&api_cmd.data[0]);
		if (api_cmd.len == 1){
			card->hw_iface.bus_write_1(
					card->hw,
					api_cmd.offset,
					*(unsigned char*)&api_cmd.data[0]);
		}else if (api_cmd.len == 2){
			card->hw_iface.bus_write_2(
					card->hw,
					api_cmd.offset,
					*(unsigned short*)&api_cmd.data[0]);
		}else if (api_cmd.len == 4){
			card->hw_iface.bus_write_4(
					card->hw,
					api_cmd.offset,
					*(unsigned int*)&api_cmd.data[0]);
		}else{
			card->hw_iface.poke(
					card->hw,
					api_cmd.offset,
					(unsigned char*)&api_cmd.data[0],
					api_cmd.len);
			//memcpy_toio((unsigned char*)vector,
			//	(unsigned char*)&api_cmd.data[0],api_cmd.len);
		}
		err=0;
		break;

	case SIOC_WAN_SET_PCI_BIOS:

		DEBUG_TEST("%s: Set PCI BaseClass %X BaseAddr %X IRQ %d\n",
				wan_netif_name(dev),
			       	priv->base_class,
			       	priv->base_addr,
				priv->irq);

		udelay(200);
		card->hw_iface.pci_write_config_dword(
				card->hw, 0x04, priv->base_class);
		card->hw_iface.pci_write_config_dword(
				card->hw, PCI_IO_BASE_DWORD, priv->base_addr0);
		card->hw_iface.pci_write_config_dword(
				card->hw, PCI_MEM_BASE0_DWORD, priv->base_addr1);
		card->hw_iface.pci_write_config_byte(
				card->hw, PCI_INT_LINE_BYTE, priv->irq);
		err = 0;
		break;
		
	case SIOC_WAN_HWPROBE:
		DEBUG_TEST("%s: Read list of Sangoma devices!\n",
					wan_netif_name(dev));
		memset(&api_cmd.data[0], 0, MAX_DATA_SIZE);
		if (card->hw_iface.get_hwprobe){
			err = card->hw_iface.get_hwprobe(
					card->hw, 0, (void**)&str); 
			if (err){
				break;
			}
			str[strlen(str)] = '\0';
			memcpy(api_cmd.data, str, strlen(str));
			api_cmd.len = ncards;	/* set to number of cards */
		}
		break;

	case SIOC_WAN_COREREV:
		if (card->hw_iface.getcfg){
			err = card->hw_iface.getcfg(
					card->hw,
				       	SDLA_COREREV,
					&api_cmd.data[0]);
			api_cmd.len = 1;
		}
		DEBUG_TEST("%s: Get core revision (rev %X)!\n", 
				wan_netif_name(dev), api_cmd.data[0]);
		break;
	default:
		DEBUG_EVENT("%s: Unknown IOCTL command!\n",
					wan_netif_name(dev));
		break;
	}

	if (WAN_COPY_TO_USER(ifr->ifr_data,&api_cmd,sizeof(struct wan_aftup_api))){
		return -EFAULT;
	}
	
	return err;
}

