/***********************************************************
 * wanpipe_fr.c	  Wanpipe Frame Relay Protocol Module
 *
 *
 *
 *
 */

#include "wanpipe_fr.h"

/**********************************************************8
 * FUNCTION PROTOTYPES
 *
 */

static int fr_lmi_recv(fr_prot_t* fr_prot, void *rx_pkt);
static void fr_lmi_send(fr_prot_t *fr_prot, int fullrep,int signalling);
static int fr_data_indication(fr_channel_t *chan, void *skb, int type);


/**********************************************************8
 * PUBLIC FUNCTIONS
 *
 */

void *wp_register_fr_prot(void *link_ptr, char *devname, wan_fr_conf_t *cfg, wp_fr_reg_t *prot_reg)
{
	fr_prot_t *fr_prot;

	fr_prot=wpabs_malloc(sizeof(fr_prot_t));
	if (!fr_prot){
		return NULL;
	}

	wpabs_debug_event("%s: DEBUG: Cfg sig =%d\n",__FUNCTION__,cfg->signalling);

	wpabs_debug_event("%s: Initializing Frame Relay Protocol\n",devname);

	wpabs_memset(fr_prot,0,sizeof(fr_prot_t));

	if (cfg){
		wpabs_memcpy(&fr_prot->cfg,cfg,sizeof(wan_fr_conf_t));
	}
#if 0
	fr_prot->cfg.signalling=WANOPT_FR_ANSI;
	fr_prot->cfg.station=WANOPT_CPE;
#endif

	if (!fr_prot->cfg.t391){
		fr_prot->cfg.t391=10;
	}
	if (!fr_prot->cfg.t392){
		fr_prot->cfg.t392=16;
	}
	if (!fr_prot->cfg.n391){
		fr_prot->cfg.n391=6;
	}
	if (!fr_prot->cfg.n392){
		fr_prot->cfg.n392=3;
	}
	if (!fr_prot->cfg.n393){
		fr_prot->cfg.n393=4;
	}
	
	wpabs_strncpy(fr_prot->name,devname,sizeof(fr_prot->name)-1);
	
	wpabs_memcpy(&fr_prot->reg,prot_reg,sizeof(wp_fr_reg_t));
	
	fr_prot->state = LINK_STATE_CHANGED;
	fr_prot->txseq = fr_prot->rxseq = 0;
	fr_prot->last_errors = 0xFFFFFFFF;
	fr_prot->n391cnt = 0;


	wpabs_debug_event("%s: Configuration: \n",
			 fr_prot->name);
	wpabs_debug_event("%s:     Station   : %s\n",
			 fr_prot->name,
			 fr_prot->cfg.station==WANOPT_CPE?"CPE":"Node");
	wpabs_debug_event("%s:     Signalling: %s\n",
			 fr_prot->name,
			 fr_prot->cfg.signalling==WANOPT_FR_AUTO_SIG?"Auto":
			 fr_prot->cfg.signalling==WANOPT_FR_ANSI?"ANSI":"LMI");
	wpabs_debug_event("%s:     T391      : %d\n",
			fr_prot->name,fr_prot->cfg.t391);
	wpabs_debug_event("%s:     T392      : %d\n",
			fr_prot->name,fr_prot->cfg.t392);
	wpabs_debug_event("%s:     N391      : %d\n",
			fr_prot->name,fr_prot->cfg.n392);
	wpabs_debug_event("%s:     N392      : %d\n",
			fr_prot->name,fr_prot->cfg.n392);
	wpabs_debug_event("%s:     N393      : %d\n",
			fr_prot->name,fr_prot->cfg.n393);

	wpabs_debug_event("\n");

	if (fr_prot->cfg.signalling==WANOPT_FR_AUTO_SIG){
		if (fr_prot->cfg.station==WANOPT_NODE){
			wpabs_debug_event("%s: Error: FR Auto Signalling not supported by NODE!\n",
					fr_prot->name);
			wpabs_free(fr_prot);
			return NULL;
		}
		fr_prot->lmi_dlci=LMI_ANSI_DLCI;
	}else if (fr_prot->cfg.signalling==WANOPT_FR_ANSI){
		fr_prot->lmi_dlci=LMI_ANSI_DLCI;
	}else{
		fr_prot->lmi_dlci=LMI_LMI_DLCI;
	}
	
	if (fr_prot->cfg.station == WANOPT_CPE){
		wpabs_debug_event("%s: Waiting for FR DLCI report from switch ...\n\n",	
				fr_prot->name);
	}else{
		wpabs_debug_event("%s: FR Node waiting for CPE requests ...\n\n",	
				fr_prot->name);
	}

	fr_prot->state |= LINK_STATE_RELIABLE;
	fr_prot->dev=link_ptr;

	wp_dev_hold(fr_prot);

	return fr_prot;

}

int wp_unregister_fr_prot(void *prot_ptr)
{	
	fr_prot_t *fr_prot=(fr_prot_t*)prot_ptr;

	if (fr_prot->chan_cnt){
		wpabs_debug_event("%s: Error fr channels exist %d\n",
			 fr_prot->name,fr_prot->chan_cnt);
		return -1;
	}

	wpabs_set_bit(0, &fr_prot->down);

	wp_dev_put(fr_prot);

	return 0;
}


void *wp_register_fr_chan(void *if_ptr, void *prot_ptr, int dlci, void *devname)
{
	fr_prot_t *fr_prot=(fr_prot_t*)prot_ptr;
	fr_channel_t *chan;
	
	if (dlci < LOWEST_VALID_DLCI || dlci > HIGHEST_VALID_DLCI) {

		if (dlci == 0){
			if (fr_prot->dlci_to_dev_map[dlci]){
				wpabs_debug_event("%s: Auto DLCI already configured!\n",devname);
				return NULL;
			}
		}else{
			wpabs_debug_event("%s: Invalid DLCI specified\n",devname);
			return NULL;
		}
	}
	
	if (fr_prot->dlci_to_dev_map[dlci]){
		wpabs_debug_event("%s: Error DLCI %d already in use\n",devname,dlci);
		return NULL;
	}

	if (dlci==0){

		wpabs_debug_event("%s: Initializing DLCI for Autodetect\n",
				fr_prot->name);

		if (fr_prot->cfg.station == WANOPT_NODE){
			wpabs_debug_event("%s: Error: DLCI autodetect not allowed on Node!\n",
					fr_prot->name);
			return NULL;
		}

	}else{
		wpabs_debug_event("%s: Initializing DLCI %d\n",
				fr_prot->name,dlci);
	}
			
	chan=wpabs_malloc(sizeof(fr_channel_t));
	if (!chan){
		return NULL;
	}

	wpabs_memset(chan,0,sizeof(fr_channel_t));

	wpabs_strncpy(chan->name,devname,sizeof(chan->name)-1);

	chan->dlci=dlci;
	chan->dlci_state=0;
	chan->newstate=0;

	
	chan->dev=if_ptr;

	wp_dev_hold(fr_prot);
	chan->prot=fr_prot;

	wp_dev_hold(chan);
	fr_prot->dlci_to_dev_map[dlci]=chan;

	wp_dev_hold(chan);
	wplist_insert_dev(chan,fr_prot->chan_list);

	fr_prot->chan_cnt++;
	
	wp_dev_hold(chan);	
	return chan;
}

void wp_unregister_fr_chan(void *dev_ptr, void *prot_ptr, void *chan_ptr)
{
	fr_prot_t *fr_prot=(fr_prot_t*)prot_ptr;
	fr_channel_t *chan=(fr_channel_t *)chan_ptr;
	fr_channel_t *cur,*priv;

	if (!fr_prot || !chan){
		return;
	}
	
	wpabs_set_bit(0, &fr_prot->down);
	
	fr_prot->dlci_to_dev_map[chan->dlci]=NULL;
	wp_dev_put(chan);

	wplist_remove_dev(chan,cur,priv,fr_prot->chan_list);
	wp_dev_put(chan);

	fr_prot->chan_cnt--;
	
	chan->prot=NULL;
	wp_dev_put(fr_prot);
	
	chan->dev=NULL;
	
	wp_dev_put(chan);
	return;
}

int wp_fr_open_chan(void *prot_ptr, void *chan_ptr)
{
	fr_prot_t *fr_prot=(fr_prot_t*)prot_ptr;
	fr_channel_t *chan=(fr_channel_t *)chan_ptr;

	if (fr_prot->cfg.station == WANOPT_NODE){
		chan->dlci_state = 0;
		fr_prot->state |= LINK_STATE_CHANGED;
	}else{
		if (chan->dlci_state & PVC_STATE_ACTIVE){
			fr_prot->reg.chan_set_state(chan->dev,WAN_CONNECTED,NULL,0);	
		}else{
			fr_prot->reg.chan_set_state(chan->dev,WAN_CONNECTING,NULL,0);
			chan->dlci_state=0;
		}
	}

	wpabs_clear_bit(0,&chan->down);
	return 0;
}


int wp_fr_close_chan(void *prot_ptr, void *chan_ptr)
{
	fr_prot_t *fr_prot=(fr_prot_t*)prot_ptr;
	fr_channel_t *chan=(fr_channel_t *)chan_ptr;

	wpabs_set_bit(0,&chan->down);
	
	if (fr_prot->cfg.station == WANOPT_NODE){
		chan->dlci_state &= ~(PVC_STATE_ACTIVE|PVC_STATE_NEW);
		fr_prot->state |= LINK_STATE_CHANGED;
	}

	return 0;
}

int wp_fr_ioctl_chan(void *prot_ptr, void *chan_ptr, int cmd)
{
	return 0;
}

int wp_fr_rx (void * prot_ptr, void *rx_pkt, unsigned long time)
{
	fr_prot_t *fr_prot 	= (fr_prot_t *)prot_ptr;
	unsigned short dlci;
	fr_channel_t *chan 	= NULL;
	fr_hdr *fh 		= (fr_hdr*)wpabs_skb_data(rx_pkt);
	unsigned char *data	= wpabs_skb_data(rx_pkt);
	int len			= wpabs_skb_len(rx_pkt);

	if (len<4 || fh->ea1){ 
		fr_prot->link_stats.rx_bad_format++;
		goto fr_rx_error;
	}
	dlci = q922_to_dlci(data);

	if (dlci == LMI_ANSI_DLCI || dlci == LMI_LMI_DLCI) {

		if (fh->ui != FR_UI){
			goto fr_rx_error;
		}
		
		if (fh->prot == LMI_ANSI_PROTO || fh->prot == LMI_PROTO) {
			
			if (fr_lmi_recv(fr_prot, rx_pkt)){
				goto fr_rx_error;
			} else {
				/* No request pending */
				fr_prot->state &= ~LINK_STATE_REQUEST;
				fr_prot->last_rx_poll = time;
				goto fr_rx_error;
			}
		}

		wpabs_debug_event("%s: Received non-LMI frame with LMI DLCI\n",fr_prot->name);
		fr_prot->link_stats.rx_bad_format++;
		goto fr_rx_error;
	}

	chan = find_channel(fr_prot, dlci);
	if (!chan){
		wpabs_debug_event( "%s: Rx data on unconfigured DLCI %d\n",
	       		fr_prot->name, dlci);
		wpabs_debug_event( "%s: DBG: Dlci=%d Len=%d Data: %02X %02X %02X %02X %02X\n",
	       		fr_prot->name, dlci, len, data[0],data[1],data[2],data[3]);
		fr_prot->link_stats.rx_bad_dlci++;
		goto fr_rx_error;
	}

	if ((chan->dlci_state & PVC_STATE_FECN) != (fh->fecn ? PVC_STATE_FECN : 0)) {
		chan->rx_FECN++;
		/* FIXME use dev name */
		wpabs_debug_test( "%s: FECN O%s\n", fr_prot->name,
		       fh->fecn ? "N" : "FF");
		chan->dlci_state ^= PVC_STATE_FECN;
	}

	if ((chan->dlci_state & PVC_STATE_BECN) != (fh->becn ? PVC_STATE_BECN : 0)) {
		chan->rx_BECN++;
		/* FIXME use dev name */
		wpabs_debug_test( "%s: BECN O%s\n", fr_prot->name,
		       fh->becn ? "N" : "FF");
		chan->dlci_state ^= PVC_STATE_BECN;
	}


	/* FIXME: Export this state
	** if (chan->dlci_state & PVC_STATE_BECN){
	**	chan->stats.rx_compressed++;
	** }	*/

	/* This is a data packet, Pass Packet up the STACK */
	if (fh->ui != FR_UI){

		if (fh->ui==CISCO_UI && fh->prot == CISCO_IP){
			chan->cisco_hdr=1;
		}else{
			fr_prot->link_stats.rx_bad_format++;
			goto fr_rx_error;
		}
	}else{
		chan->cisco_hdr=0;
	}

	if (fh->prot == NLPID_IP || fh->prot == CISCO_IP){
		wpabs_skb_pull(rx_pkt, 4);
		fr_data_indication(chan,rx_pkt,FR_MP_IP);

	}else if (fh->prot == NLPID_IPV6){
		wpabs_skb_pull(rx_pkt, 4);
		fr_data_indication(chan,rx_pkt,FR_MP_IPV6);
			
	}else if (data[3] == FR_PAD && data[4] == NLPID_SNAP &&
		  data[5] == FR_PAD &&
	          data[6] == FR_PAD && data[7] == FR_PAD &&
	          ((data[8]<<8) | data[9]) == ETH_P_ARP) {
		
		wpabs_skb_pull(rx_pkt, 10);
		fr_data_indication(chan,rx_pkt,FR_MP_ARP);
		
	}else{
		wpabs_skb_pull(rx_pkt, 2);
		fr_data_indication(chan,rx_pkt,FR_MP_NONE);
	}


	return 0;

fr_rx_error:

	return -1;
}


int wp_fr_timer(void *prot_ptr, unsigned long time)
{
	fr_prot_t *fr_prot = (fr_prot_t *)prot_ptr;
	int i, cnt = 0, reliable;
	fr_channel_t *chan;
	unsigned int list;
	int timeout=fr_prot->cfg.t392;

	if (fr_prot->cfg.station == WANOPT_NODE){
		reliable = ((time - fr_prot->last_rx_poll) < (fr_prot->cfg.t392*wpabs_get_hz()));
		if (!reliable){
			fr_prot->link_stats.T392_timeouts++;
		}
	}else{
		fr_prot->last_errors <<= 1; /* Shift the list */
		
		if (fr_prot->state & LINK_STATE_REQUEST) {
			wpabs_debug_event("%s: Waiting for FR DLCI report from switch ...\n",
			       fr_prot->name);
			fr_prot->last_errors |= 1;
			fr_prot->link_stats.T391_timeouts++;
		}

		for (i = 0, list = fr_prot->last_errors; i < fr_prot->cfg.n393;
		     i++, list >>= 1){
			cnt += (list & 1);	/* errors count */
		}
		reliable = (cnt < fr_prot->cfg.n392);
	}

	if ((fr_prot->state & LINK_STATE_RELIABLE) != (reliable ? LINK_STATE_RELIABLE : 0)){ 

		for (chan=fr_prot->chan_list;chan;chan = chan->next){
			chan->dlci_state &= ~(PVC_STATE_NEW | PVC_STATE_ACTIVE);

			fr_prot->reg.chan_set_state(chan->dev,WAN_DISCONNECTED,NULL,0);
		}
		
		fr_prot->state ^= LINK_STATE_RELIABLE;
	
		if (reliable){
			fr_prot->reg.prot_set_state(fr_prot->dev,WAN_CONNECTED,NULL,0);
		}else{
			wpabs_debug_event( "%s: Link unreliable !\n",
					fr_prot->name);
		
			fr_prot->reg.prot_set_state(fr_prot->dev,WAN_DISCONNECTED,NULL,0);
		}
		
		wpabs_debug_event( "%s: Link %sreliable\n", fr_prot->name,
		     reliable ? "" : "un");

		if (reliable) {
			fr_prot->n391cnt = 0; /* Request full status */
			fr_prot->state |= LINK_STATE_CHANGED;
		}
	}

	if (fr_prot->cfg.station == WANOPT_NODE){
		timeout = fr_prot->cfg.t392;
	}else{
		if (fr_prot->n391cnt)
			fr_prot->n391cnt--;

		if (fr_prot->cfg.signalling == WANOPT_FR_AUTO_SIG){
			if (fr_prot->n391cnt == 0){
				fr_lmi_send(fr_prot, 1, WANOPT_FR_ANSI);
				fr_prot->txseq--;
				fr_lmi_send(fr_prot, 1, WANOPT_FR_LMI);
			}else{
				fr_lmi_send(fr_prot, 0, WANOPT_FR_ANSI);
				fr_prot->txseq--;
				fr_lmi_send(fr_prot, 0, WANOPT_FR_LMI);
			}
		}else{	
			fr_lmi_send(fr_prot, (fr_prot->n391cnt == 0),fr_prot->cfg.signalling);
		}
		
		fr_prot->state |= LINK_STATE_REQUEST;
		timeout = fr_prot->cfg.t391;
	}

	return timeout;
}


int wp_fr_tx_hdr (void * chan_ptr, void *skb, int type)
{
	fr_channel_t *chan = (fr_channel_t *)chan_ptr;
	unsigned char *data;
	int len=0;

	switch(type){

	case FR_MP_NONE:

		if (wpabs_skb_headroom(skb) < 2){
			return -1;
		}
		
		data=wpabs_skb_push(skb,2);
		dlci_to_q922(data, chan->dlci);
		return 2;

	case FR_MP_IP:

		if (wpabs_skb_headroom(skb) < 4){
			return -1;
		}
		
		data=wpabs_skb_push(skb,4);

		if (chan->cisco_hdr){
			data[3]=CISCO_IP;
		}else{
			data[3]=NLPID_IP;
		}
		len=4;
		break;
		
	case FR_MP_IPV6:

		if (wpabs_skb_headroom(skb) < 4){
			return -1;
		}

		data=wpabs_skb_push(skb,4);
		data[3]=NLPID_IPV6;
		len=4;
		break;

	case FR_MP_ARP:

		if (wpabs_skb_headroom(skb) < 10){
			return -1;
		}

		data=wpabs_skb_push(skb,10);
		data[3] = FR_PAD;
		data[4] = NLPID_SNAP;
		data[5] = FR_PAD;
		data[6] = FR_PAD;
		data[7] = FR_PAD;
		data[8] = type>>8;
		data[9] = (unsigned char)type;
		len=10;
		break;

	default:
		wpabs_debug_event("%s: Error: Invalid fr protocol type 0x%X\n",
				__FUNCTION__,type);
		return -1;

	}

	dlci_to_q922(data, chan->dlci);

	if (chan->cisco_hdr){
		data[2] = CISCO_UI;
	}else{
		data[2] = FR_UI;
	}

	return len;
}





/**********************************************************8
 * PRIVATE FUNCTIONS
 *
 */

/*======================================================
 * fr_data_transmit
 *
 * Transmitt the fr packet to the lower layer,
 * vi its master (tx_dev) device.
 * 
 */
static int fr_data_transmit(fr_prot_t* fr_prot, void *skb)
{
	if (fr_prot->dev){
		int err = fr_prot->reg.tx_down(fr_prot->dev,skb);
		if (err != 0){
			wpabs_skb_free(skb);
		}
	}else{
		wpabs_skb_free(skb);
	}
	
	return 0;
}

/*======================================================
 * xdlc_data_indication
 *
 * Transmitt the xdlc packet to the lower layer,
 * vi its master (tx_dev) device.
 * 
 */
static int fr_data_indication(fr_channel_t *chan, void *skb, int type)
{
	fr_prot_t *fr_prot = chan->prot;

	wpabs_debug_test("%s: %s() Len=%d\n",
			chan->name,__FUNCTION__,wpabs_skb_len(skb));

	if (fr_prot && chan->dev){
		int err=fr_prot->reg.rx_up(chan->dev,skb,type);
		if (err != 0){
			wpabs_skb_free(skb);
		}
	}

	return 0;
}





static int fr_lmi_recv(fr_prot_t* fr_prot, void *rx_pkt)
{
	int stat_len;
	int reptype = -1, error;
	unsigned char rxseq, txseq;
	fr_channel_t *chan;
	int i;
	unsigned char *data = wpabs_skb_data(rx_pkt);
	int len = wpabs_skb_len(rx_pkt);


/* 	wpabs_debug_event( "\n"); */
/*	wpabs_debug_event( "FR LMI REC: Len=%d Min=%d\n",len, LMI_ANSI_LENGTH); */

	if (fr_prot->cfg.signalling == WANOPT_FR_AUTO_SIG){
		int dlci;

		if (len < LMI_ANSI_LENGTH ||
		    len < LMI_LENGTH){
			wpabs_debug_event("%s: Short LMI frame\n", fr_prot->name);
				fr_prot->link_stats.rx_dropped++;
				return 1;
		}

		dlci = q922_to_dlci(data);
		
		if (dlci == LMI_ANSI_DLCI){
			fr_prot->lmi_dlci=LMI_ANSI_DLCI;
			fr_prot->cfg.signalling=WANOPT_FR_ANSI;
			wpabs_debug_event("%s: FR Signalling Autodetect: ANSI\n",
					fr_prot->name);
		}else if (dlci == LMI_LMI_DLCI){
			fr_prot->lmi_dlci=LMI_LMI_DLCI;
			fr_prot->cfg.signalling=WANOPT_FR_LMI;
			wpabs_debug_event("%s: FR Signalling Autodetect: LMI\n",
					fr_prot->name);
		}else{
			wpabs_debug_event("%s: Fr Signalling autodetect error: dlci=%d\n",
					fr_prot->name,dlci);
			return 1;
		}
		
	}else{
		if (len < ((fr_prot->cfg.signalling == WANOPT_FR_ANSI) ? LMI_ANSI_LENGTH : LMI_LENGTH)){
			wpabs_debug_event("%s: Short LMI frame\n", fr_prot->name);
			fr_prot->link_stats.rx_dropped++;
			return 1;
		}
	}

	if (data[5] != ((fr_prot->cfg.station != WANOPT_NODE) ? 
			LMI_STATUS : LMI_STATUS_ENQUIRY)) {
		
		wpabs_debug_event("%s: LMI msgtype=%x, Not LMI status %s\n",
		       fr_prot->name,data[2],
		       (fr_prot->cfg.station != WANOPT_NODE) ? "enquiry" : "reply");
		fr_prot->link_stats.rx_bad_format++;
		return 1;
	}
	
	i = (fr_prot->cfg.signalling == WANOPT_FR_ANSI) ? 7 : 6;

	if (data[i] !=
	    (MODE_FR_CCITT ? LMI_CCITT_REPTYPE : LMI_REPTYPE)) {
		wpabs_debug_event("%s: Not a report type=%x\n",
		       fr_prot->name, data[i]);
		fr_prot->link_stats.rx_bad_format++;
		return 1;
	}
	i++;
	i++;				/* Skip length field */

	reptype = data[i++];
	/*wpabs_debug_event( "(Debug) RepType = 0x%x, Inc=%d\n",reptype,i); */
	
	if (data[i] != ((MODE_FR_CCITT) ? LMI_CCITT_ALIVE : LMI_ALIVE)) {
		wpabs_debug_event("%s: Unsupported status element=%x\n",
		       fr_prot->name, data[i]);
		fr_prot->link_stats.rx_bad_format++;
		return 1;
	}
	i++;

	i++;			/* Skip length field */

	fr_prot->rxseq = data[i++]; /* TX sequence from peer */
	rxseq = data[i++];	/* Should confirm our sequence */

	txseq = fr_prot->txseq;

	/* wpabs_debug_event( "TxSeqOrig=%d  TxSeqFrame=%d RxSeqFrame=%d Inc=%d\n",
	**		txseq, fr_prot->rxseq, rxseq,i); */
	
	if (fr_prot->cfg.station == WANOPT_NODE) {
		switch (reptype){
			
		case LMI_INTEGRITY:
			fr_prot->link_stats.node_rx_LIV++;
			/* card->wandev.stats.rx_crc_errors++; */
			break;
		case LMI_FULLREP:
			fr_prot->link_stats.node_rx_FSE++;
			/* card->wandev.stats.rx_frame_errors++; */
			break;

		default:
			wpabs_debug_event("%s: Unsupported report type=%x\n",
			       fr_prot->name, reptype);
			fr_prot->link_stats.rx_bad_format++;
			return 1;
		}
	}else{
		switch (reptype){
			
		case LMI_INTEGRITY:
			fr_prot->link_stats.cpe_rx_LIV++;
			break;
		case LMI_FULLREP:
			fr_prot->link_stats.cpe_rx_FSR++;
			break;
		}

	}

	error = 0;
	if (!(fr_prot->state & LINK_STATE_RELIABLE)){
		/* wpabs_debug_event( "FR Link State ! Reliable\n"); */
		error = 1;
	}

	if (rxseq == 0 || rxseq != txseq) {
		/* wpabs_debug_event( "Rxseq =0 or != TxSeq ask for full status again\n"); */
		fr_prot->n391cnt = 0; /* Ask for full report next time */
		error = 1;
	}

	if (fr_prot->cfg.station == WANOPT_NODE) {
	
		if ((fr_prot->state & LINK_STATE_FULLREP_SENT) && !error) {
			
			/* Stop sending full report - the last one has been confirmed by DTE */
			fr_prot->state &= ~LINK_STATE_FULLREP_SENT;
			
			for (chan = fr_prot->chan_list; chan; chan = chan->next){
			
				if (chan->dlci_state & PVC_STATE_NEW) {
					chan->dlci_state &= ~PVC_STATE_NEW;
					chan->dlci_state |= PVC_STATE_ACTIVE;
					fr_log_dlci_active(chan);

					fr_prot->reg.chan_set_state(chan->dev,WAN_CONNECTED,NULL,0);

					/* Tell DTE that new PVC is now active */
					fr_prot->state |= LINK_STATE_CHANGED;
				}
			}
		}

		if (fr_prot->state & LINK_STATE_CHANGED) {
			reptype = LMI_FULLREP;
			fr_prot->state |= LINK_STATE_FULLREP_SENT;
			fr_prot->state &= ~LINK_STATE_CHANGED;
		}

		fr_lmi_send(fr_prot, (reptype == LMI_FULLREP ? 1 : 0),fr_prot->cfg.signalling);
		return 0;
	}

	/* DTE */
	if (reptype != LMI_FULLREP || error){
		/* wpabs_debug_event( "RepType=%d != LMI_FULLREP(0) Error=%d\n",
		** 		reptype,error); */
		return 0;
	}

	stat_len = (fr_prot->cfg.signalling == WANOPT_FR_ANSI)?3:6;

	/* wpabs_debug_event( "JUST BEFORE DEV LIST Cnt=%d Cnt=%d\n",i,(i + 2 + stat_len)); */
	
	for (chan = fr_prot->chan_list;chan;chan = chan->next){
		chan->newstate=0;
	}

	/*wpabs_debug_event( "Starting to go through dlcis\n");*/
	init_global_dlci_state(fr_prot);

	/* wpabs_debug_event("%s: DLCI Report len=%d, i=%d, stat_len=%d\n",
	**		fr_prot->name,len,i,stat_len); */

	while (len >= (i + 2 + stat_len)) {

		unsigned short dlci;
		unsigned char state = 0;

		if (data[i] != ((MODE_FR_CCITT) ? LMI_CCITT_PVCSTAT : LMI_PVCSTAT)) {
			wpabs_debug_event("%s: Invalid PVCSTAT ID: %x\n",
			       fr_prot->name, data[i]);
			fr_prot->link_stats.rx_bad_format++;
			return 1;
		}
		i++;

		if (data[i] != stat_len) {
			wpabs_debug_event("%s: Invalid PVCSTAT length: %x\n",
			       fr_prot->name, data[i]);
			fr_prot->link_stats.rx_bad_format++;
			return 1;
		}
		i++;

		/*wpabs_debug_event( "DLCI TO STATUS starting at I=%d 0x%x 0x%x\n",
		**		i,data[i],data[i+1]); */

		dlci = status_to_dlci(fr_prot->cfg.signalling,(data+i), &state);

		/* wpabs_debug_event( "DLCI IS %d State 0x%x\n",dlci,state); */
		
		chan = find_channel(fr_prot,dlci);
		if (chan){
			chan->newstate = state;
		
		}else{
			if ((chan=fr_prot->dlci_to_dev_map[0])){
				wpabs_debug_event( "%s: DLCI %d Auto detected and configured\n",
		     			  fr_prot->name, dlci);
				fr_prot->dlci_to_dev_map[dlci]=chan;
				fr_prot->dlci_to_dev_map[0]=NULL;
				chan->dlci=dlci;
				chan->newstate = state;
				
			}else if (state == PVC_STATE_NEW){

				/* FIXME: Callback function should be added for
				**       new dlci */
				wpabs_debug_event("%s: New PVC available, DLCI=%u\n",
				       fr_prot->name, dlci);
			}
		}
		
		set_global_dlci_state(fr_prot,dlci,state);
		
		i += stat_len;
	}
	
	/* wpabs_debug_event( "End of dlci report\n"); */

	for (chan=fr_prot->chan_list;chan;chan=chan->next){
	
		if (chan->newstate == PVC_STATE_NEW)
			chan->newstate = PVC_STATE_ACTIVE;

		chan->newstate |= (chan->dlci_state & ~(PVC_STATE_NEW|PVC_STATE_ACTIVE));
		if (chan->dlci_state != chan->newstate) {
			chan->dlci_state = chan->newstate;
			fr_log_dlci_active(chan);

			if (chan->dlci_state & PVC_STATE_ACTIVE){
				fr_prot->reg.chan_set_state(chan->dev,WAN_CONNECTED,NULL,0);
			}

			if (!(chan->dlci_state & PVC_STATE_ACTIVE)){
				fr_prot->reg.chan_set_state(chan->dev,WAN_DISCONNECTED,NULL,0);
			}
		}else{
			if (!(chan->dlci_state & PVC_STATE_ACTIVE)){
				fr_prot->reg.chan_set_state(chan->dev,WAN_DISCONNECTED,NULL,0);
			}
		}
		
	}

	/* Next full report after N391 polls */
	fr_prot->n391cnt = fr_prot->cfg.n391;
	return 0;
}


static void fr_lmi_send(fr_prot_t *fr_prot,int fullrep, int signalling)
{
	fr_channel_t *chan;
	int len = (signalling == WANOPT_FR_ANSI) ? LMI_ANSI_LENGTH : LMI_LENGTH;
	int stat_len = (signalling == WANOPT_FR_ANSI)?3:6;
	unsigned char *data;
	int i = 0,ch;
	int dlci_cnt=0;
	void *skb;
	unsigned short lmi_dlci = (signalling == WANOPT_FR_ANSI) ? LMI_ANSI_DLCI:LMI_LMI_DLCI;

	if (fr_prot->cfg.station == WANOPT_NODE && fullrep) {
		dlci_cnt=fr_prot->chan_cnt;
		len += dlci_cnt * (2 + stat_len);

		if (len > fr_prot->reg.mtu){
			wpabs_debug_event("\n");
			wpabs_debug_event("%s: Error: Full status report size %d exceeds MTU=%d\n", 
					fr_prot->name,len,fr_prot->reg.mtu);
			wpabs_debug_event("%s: Error: Too many DLCI's configured or increase MTU\n",
					fr_prot->name);
			return;
		}
	}


	skb=wpabs_skb_alloc(len);
	if (!skb){
		wpabs_debug_event("%s: Memory squeeze on %s()\n",
			           fr_prot->name,__FUNCTION__);
		return;
	}

	data=wpabs_skb_data(skb);
	wpabs_memset(data, 0, len);

	wpabs_skb_reserve(skb, 4);
	wpabs_skb_push(skb, 4);
	data[3] = (signalling == WANOPT_FR_ANSI) ? LMI_ANSI_PROTO:LMI_PROTO;
	dlci_to_q922(data, lmi_dlci);
	data[2] = FR_UI;
	
	data = wpabs_skb_tail(skb);
	data[i++] = LMI_CALLREF;
	data[i++] = (fr_prot->cfg.station == WANOPT_NODE) ? LMI_STATUS : LMI_STATUS_ENQUIRY;
	
	if (signalling == WANOPT_FR_ANSI){
		data[i++] = LMI_ANSI_LOCKSHIFT;
	}
	
	data[i++] = MODE_FR_CCITT ? LMI_CCITT_REPTYPE : LMI_REPTYPE;
	
	data[i++] = LMI_REPT_LEN;
	data[i++] = fullrep ? LMI_FULLREP : LMI_INTEGRITY;

	data[i++] = MODE_FR_CCITT ? LMI_CCITT_ALIVE : LMI_ALIVE;
	data[i++] = LMI_INTEG_LEN;
	data[i++] = fr_prot->txseq = fr_lmi_nextseq(fr_prot->txseq);
	data[i++] = fr_prot->rxseq;

	
	if (fr_prot->cfg.station == WANOPT_NODE && fullrep) {

		for (ch=0;ch<MAX_FR_CHANNELS;ch++){
	
			if ((chan=fr_prot->dlci_to_dev_map[ch]) == NULL)
				continue;
		
			if (--dlci_cnt < 0)
				break;
		
			if (!chan)
				continue;
	
			/* wpabs_debug_event( "Interigating %s DLCI=%d\n",chan->if_name,chan->dlci); */
		
			data[i++] = MODE_FR_CCITT ? LMI_CCITT_PVCSTAT:LMI_PVCSTAT;
			data[i++] = stat_len;

			if ((fr_prot->state & LINK_STATE_RELIABLE) &&
			    !wpabs_test_bit(0,&chan->down) &&
			    !(chan->dlci_state & (PVC_STATE_ACTIVE|PVC_STATE_NEW))) {
				
				chan->dlci_state |= PVC_STATE_NEW;
				fr_log_dlci_active(chan);

				if (!(chan->dlci_state & PVC_STATE_ACTIVE)){
					fr_prot->reg.chan_set_state(chan->dev,WAN_DISCONNECTED,NULL,0);
				}
			}

			dlci_to_status(fr_prot, chan->dlci,
				       data+i, chan->dlci_state);
			i += stat_len;
		}
	}

	if (fr_prot->cfg.station == WANOPT_NODE){
		if (fullrep){
			fr_prot->link_stats.node_tx_FSR++;
		}else{
			fr_prot->link_stats.node_tx_LIV++;
		}
	}else{
		if (fullrep){
			fr_prot->link_stats.cpe_tx_FSE++;
		}else{
			fr_prot->link_stats.cpe_tx_LIV++;
		}
	}
	
	wpabs_skb_put(skb, i);
	
	fr_data_transmit(fr_prot,skb);
}

