/*****************************************************************************
* lapb_iface.c	WANPIPE LapB Protocol Module. Main code.
*
* Author:	Nenad Corbic
*
* Copyright:	(c)2003 Sangoma Technologies Inc.
*
*		This program is free software; you can redistribute it and/or
*		modify it under the terms of the GNU General Public License
*		as published by the Free Software Foundation; either version
*		2 of the License, or (at your option) any later version.
* ============================================================================
*	This module:
*		This module is free software; you can redistribute it and/or
*		modify it under the terms of the GNU General Public License
*		as published by the Free Software Foundation; either version
*		2 of the License, or (at your option) any later version.
*/
#include "wanpipe_lapb.h"

static int lapb_setparms(lapb_cb *lapb, wan_lapb_if_conf_t *parms);


void *wp_register_lapb_prot(void *link_ptr, char *devname, wan_lapb_if_conf_t *cfg, wp_lapb_reg_t *prot_reg)
{
	lapb_cb *lapb;

	if ((lapb = lapb_create_cb()) == NULL){
		wpabs_debug_event( "LAPB: Failed to create cb %s\n",
				devname);
		return NULL;
	}

	wpabs_strncpy(lapb->name,devname,sizeof(lapb->name)-1);

	if (lapb_setparms(lapb,cfg)){
		lapb_free_cb(lapb);
		return NULL;
	}

	wpabs_memcpy(&lapb->reg,prot_reg,sizeof(wp_lapb_reg_t));

	lapb->dev = link_ptr;;
	lapb->state = LAPB_STATE_DISCONNECTED;

	return lapb;
}


int wp_unregister_lapb_prot(void *prot_ptr)
{	
	lapb_cb *lapb =(lapb_cb*)prot_ptr;

	wpabs_set_bit(LAPB_UNREG_CRIT, &lapb->critical);

	lapb->dev=NULL;
	lapb_free_cb(lapb);

	return 0;
}

int wp_lapb_open(void *lapb_ptr)
{
	lapb_cb *lapb =(lapb_cb*)lapb_ptr;

	lapb_link_up(lapb);

	return 0;
}

int wp_lapb_close(void *lapb_ptr)
{
	lapb_cb *lapb =(lapb_cb*)lapb_ptr;

	lapb_link_down(lapb);

	lapb_send_control(lapb, LAPB_DM, 0, LAPB_COMMAND);

	return 0;
}

int wp_lapb_rx(void *lapb_ptr, void *skb)
{
	FUNC_BEGIN();
	lapb_data_input(lapb_ptr, skb);
	lapb_bh(lapb_ptr);
	FUNC_END();
	return 0;
}

int wp_lapb_bh(void *lapb_ptr)
{
	FUNC_BEGIN();
	return lapb_bh(lapb_ptr);
}


int wp_lapb_tx(void *lapb_ptr, void *skb)
{
	lapb_cb *lapb = (lapb_cb*)lapb_ptr;

	if (lapb->state != LAPB_STATE_CONNECTED){ 
			
		wpabs_debug_event("Lapb: Invalid LABP State!\n");
		wpabs_skb_free(skb);
		return 0;
	}

	if (wpabs_skb_tailroom(skb) < 4 || wpabs_skb_headroom(skb) < 6){
		wpabs_debug_event("Lapb: No skb room for header H %d T %d!!!\n",
				wpabs_skb_headroom(skb), wpabs_skb_tailroom(skb));
		wpabs_skb_free(skb);
		return 0;
	}
	
	if (wpabs_skb_queue_len(lapb->write_queue) >= lapb->window){
		wpabs_debug_test("%s: Interface busy queue > Max\n",
				lapb->name);
		return 1;
	}

#ifdef __DEBUG_DROP_X25_PACKET_
	if (wpabs_skb_len(skb) >= 3){
		unsigned int data = wpabs_skb_data(skb);
		unsigned short lcn=((data[0]&0x0F)<<8) | data[1];
		
		if (data[2] == 0x13 && lcn == 10){
			wpabs_debug_event("%s: Dropping Clear Call :)\n",lapb->name);
			wpabs_skb_free(skb);
			return 0;
		}
	}	
#endif

	wpabs_skb_unlink(skb);
	
	wpabs_skb_queue_tail(lapb->write_queue, skb);

	return 0;
}

void wp_lapb_timer(void *lapb_ptr)
{
	lapb_cb *lapb = (lapb_cb*)lapb_ptr;
	unsigned long period = wpabs_get_systemticks();
	int kick_bh=1;
	
	FUNC_BEGIN();

	if (wpabs_test_bit(LAPB_UNREG_CRIT, &lapb->critical)){
		return;
	}

	wpabs_debug_test("%s: TIMER: T1=%d T2=%d T4=%d State=%d\n",
			lapb->name,(lapb->t1timer >0),
			(lapb->t2timer >0),(lapb->t4timer >0),
			lapb->state);
	
	if (lapb->t1timer){
		
		if ((period-lapb->t1timer) > lapb->t1){
			lapb->t1timer=0;
			lapb_t1timer_expiry(lapb);		
			kick_bh=1;
		}
	}

	if (lapb->t2timer){
		if ((period-lapb->t2timer) > lapb->t2){
			lapb->t2timer=0;
			lapb_t2timer_expiry(lapb);		
			kick_bh=1;
		}
	}

	if (lapb->t4timer){
		if ((period-lapb->t4timer) > lapb->t4){
			lapb->t4timer=0;
			lapb_t4timer_expiry(lapb);		
			kick_bh=1;
		}
	}

	if (kick_bh){
		lapb_bh(lapb_ptr);
	}

	FUNC_END();
}



/*======================================================
 * lapb_data_transmit
 *
 * Transmitt the lapb packet to the lower layer,
 * vi its master (tx_dev) device.
 * 
 */
int lapb_data_transmit(lapb_cb *lapb, void *skb)
{
	wpabs_debug_test("%s: %s() Len=%d\n",
			lapb->name,__FUNCTION__,wpabs_skb_len(skb));

	if (lapb->dev){
		return lapb->reg.tx_down(lapb->dev,skb);
	}
	
	return 1;
}

/*======================================================
 * lapb_data_indication
 *
 * Transmitt the lapb packet to the lower layer,
 * vi its master (tx_dev) device.
 * 
 */
int lapb_data_indication(lapb_cb *lapb, void *skb)
{
	wpabs_debug_test("%s: %s() Len=%d\n",
			lapb->name,__FUNCTION__,wpabs_skb_len(skb));

	if (lapb->dev){
		return lapb->reg.rx_up(lapb->dev,skb);
	}

	return 1;
}



/**********************************************************
 * Private Functions
 **********************************************************/


/*======================================================
 * lapb_connect_indication
 *
 * Indicate to the upper layer (x25) that the
 * lapb is layer is connected and that it should
 * start its connection process.
 */
void lapb_connect_indication(lapb_cb *lapb, int reason)
{
	wpabs_debug_test("%s: %s() Reason=%d\n",
			lapb->name,__FUNCTION__,reason);

	if (lapb->dev && lapb->reg.chan_state_change){
		lapb->reg.chan_state_change(lapb->dev,WAN_CONNECTED,NULL, 0);
	}
}


/*======================================================
 * lapb_disconnect_indication
 *
 * Indicate to the upper layer (x25) that the
 * lapb is layer is disconnected and that it should
 * drop the connection.
 */
void lapb_disconnect_indication(lapb_cb *lapb, int reason)
{
	wpabs_debug_test("%s: %s() Reason=%d\n",
			lapb->name,__FUNCTION__,reason);

	if (lapb->dev && lapb->reg.chan_state_change){
		lapb->reg.chan_state_change(lapb->dev,WAN_DISCONNECTED,NULL,0);
	}
}


static int lapb_setparms(lapb_cb *lapb, wan_lapb_if_conf_t *parms)
{

	if (!parms){
		return 0;
	}

	wpabs_memcpy(&lapb->cfg,parms,sizeof(wan_lapb_if_conf_t));
	
	if (parms->t1 < 1){
		wpabs_debug_event("%s: Error: Lapb T1 timer is invalid \n",
				lapb->name);
		return -1;
	}
	
	/* If timer is zero the RR response will
	 * be instantaneous */
	if (parms->t2 < 0){
		wpabs_debug_event("%s: Error: Lapb T2 timer is invalid \n",
				lapb->name);
		return -1;
	}
	
	if (parms->n2 < 1){
		wpabs_debug_event( "%s: Error: Lapb N2 counter is invalid \n",
				lapb->name);
		return -1;
	}

	lapb->t4_orig = parms->t4;
	lapb->t3_orig = parms->t3;

	if (parms->t4 && parms->t4 <= parms->t1){
		wpabs_debug_event( "%s: Error: Lapb T4 timer must be greater than T1!\n",
				lapb->name);
		return -1;
	}
	
	if (parms->t4 < 1 && parms->t3 > 0){
		/* Disabled T4 timer but enabled T3 timer 
		 * default T4 to T3 and set T3 to 1, thus
	 	* if line is inactive T3 time, disconnect */
		parms->t4 = parms->t3;
		parms->t3 = 1;
		
	}else if (parms->t4 == 0 && parms->t3 == 0){
		/* Disabled both T4 and T3 Do nothing */
		lapb->t4_orig=0;
	
	}else if (parms->t3 > parms->t4){
		/* Both T4 and T3 are enabled,
		 * make T3 multiple of T4 */
		parms->t3 = (unsigned int)(parms->t3/parms->t4);

	}else if (parms->t3 == 0 && parms->t4 > 0){
		/* T3 timer is disabled, and t4
		 * timer is enabled */
		lapb->t3_orig=0;
		
	}else{
		wpabs_debug_event( "%s: Error: Lapb T3 and T4 values invalid! T3 must be greater than T4\n",
				lapb->name);
		return -1;
	}

		
	if (lapb->state == LAPB_STATE_DISCONNECTED) {
		if (parms->mode & LAPB_EXTENDED) {
			if (parms->window < 1 || parms->window > 127)
				return -1;
		} else {
			if (parms->window < 1 || parms->window > 7)
				return -1;
		}

		lapb->mode    = parms->mode;
		lapb->window  = parms->window;
	}

	lapb->t1    = parms->t1 * wpabs_get_hz();
	lapb->t2    = parms->t2 * wpabs_get_hz();
	lapb->n2    = parms->n2;
	lapb->t4    = parms->t4 * wpabs_get_hz();

	/* T3 is not a timer really, we are simulating it
	 * it is really a counter, but to the user its
	 * a timer. Thus do not multiply it with HZ */
	lapb->t3    = parms->t3;
	lapb->mtu   = parms->mtu;


	wpabs_debug_event("%s: LAPB Configuration \n",lapb->name);
	wpabs_debug_event("%s:     Station      =%s\n",
			lapb->name,(lapb->mode&LAPB_DCE)?"DCE":"DTE");
	wpabs_debug_event("%s:     Mode         =%s\n",
			lapb->name,lapb->mode&LAPB_EXTENDED?"Extended":"Normal");
	wpabs_debug_event("%s:     Window       =%d\n",
			lapb->name,lapb->window);
	wpabs_debug_event("%s:     Max Pkt Size =%d\n",
			lapb->name,lapb->mtu);
	wpabs_debug_event("%s:     T1           =%d\n",
			lapb->name,parms->t1);
	wpabs_debug_event("%s:     T2           =%d\n",
			lapb->name,parms->t2);
	wpabs_debug_event("%s:     T3           =%d\n",
			lapb->name,parms->t3);
	wpabs_debug_event("%s:     T4           =%d\n",
			lapb->name,parms->t4);
	wpabs_debug_event("%s:     N2           =%d\n",
			lapb->name,lapb->n2);

	wpabs_debug_test("%s: Lapb T1=%d T2=%d T3=%d T4=%d\n",
			lapb->name,lapb->t1,lapb->t2,lapb->t3,lapb->t4);

	return 0;
}

