/*
 *	LAPB release 002
 *
 *	This code REQUIRES 2.1.15 or higher/ NET3.038
 *
 *	This module:
 *		This module is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 */

#include "wanpipe_lapb.h"


/*
 *	Create an empty LAPB control block.
 */
lapb_cb *lapb_create_cb(void)
{
	lapb_cb *lapb;
	
	lapb=wpabs_malloc(sizeof(*lapb));
	if (lapb == NULL)
		return NULL;

	wpabs_memset(lapb, 0x00, sizeof(lapb_cb));

	wp_dev_hold(lapb);

	lapb->write_queue = wpabs_skb_alloc_queue();
	if (!lapb->write_queue){
		goto lapb_create_cb_failed;
	}
	wpabs_skb_queue_init(lapb->write_queue);

	lapb->ack_queue   = wpabs_skb_alloc_queue();
	if (!lapb->ack_queue){
		goto lapb_create_cb_failed;
	}
	wpabs_skb_queue_init(lapb->ack_queue);

	lapb->async_queue = wpabs_skb_alloc_queue();
	if (!lapb->async_queue){
		goto lapb_create_cb_failed;
	}
	wpabs_skb_queue_init(lapb->async_queue);
	
	lapb->rx_queue   = wpabs_skb_alloc_queue();
	if (!lapb->rx_queue){
		goto lapb_create_cb_failed;
	}
	wpabs_skb_queue_init(lapb->rx_queue);
	
	lapb->t1      = LAPB_DEFAULT_T1;
	lapb->t2      = LAPB_DEFAULT_T2;
	lapb->n2      = LAPB_DEFAULT_N2;
	lapb->mode    = LAPB_DEFAULT_MODE;
	lapb->window  = LAPB_DEFAULT_WINDOW;
	lapb->state   = LAPB_STATE_DISCONNECTED;

	return lapb;
	
lapb_create_cb_failed:

	lapb_free_cb(lapb);

	return NULL;
}

int lapb_free_cb(lapb_cb *lapb)
{
	if (!lapb){
		return -1;
	}
	
	if (lapb->write_queue){
		wpabs_skb_free_queue(lapb->write_queue);
		lapb->write_queue=NULL;
	}

	if (lapb->ack_queue){
		wpabs_skb_free_queue(lapb->ack_queue);
		lapb->ack_queue=NULL;
	}

	if (lapb->async_queue){
		wpabs_skb_free_queue(lapb->async_queue);
		lapb->async_queue=NULL;
	}
	
	if (lapb->rx_queue){
		wpabs_skb_free_queue(lapb->rx_queue);
		lapb->rx_queue=NULL;
	}

	lapb_stop_t4timer(lapb);
	lapb_stop_t2timer(lapb);
	lapb_stop_t1timer(lapb);

	wp_dev_put(lapb);

	return 0;
}


void lapb_state_change(lapb_cb *lapb, int state)
{
	FUNC_BEGIN();

	switch (state){
	
	case LAPB_STATE_CONNECTING:
		wpabs_debug_event( "%s: Lapb Link connecting...\n",lapb->name);
		lapb_stop_t4timer(lapb);
		break;

	case LAPB_STATE_CONNECTED:
		wpabs_debug_event( "%s: Lapb Link connected\n",lapb->name);
		lapb_start_t4timer(lapb, 1);
		break;

	case LAPB_STATE_DISCONNECTED:
		wpabs_debug_event( "%s: Lapb Link disconnected!\n",lapb->name);
		lapb_stop_t4timer(lapb);
		break;
		
	case LAPB_STATE_FRAME_REJECT:
		wpabs_debug_event( "%s: Lapb Link frame reject!\n",lapb->name);
		break;

	default:
		wpabs_debug_event( "%s: Lapb Invalid link state %d!",
				lapb->name,state);
		lapb_stop_t4timer(lapb);
		break;

	}

	lapb->state = state;
}


void lapb_requeue_frames(lapb_cb *lapb)
{
        void *skb, *skb_prev = NULL;

	/*
	 * Requeue all the un-ack-ed frames on the output queue to be picked
	 * up by lapb_bh called from the timer. This arrangement handles the
	 * possibility of an empty output queue.
	 */

	wpabs_set_bit(LAPB_REQUEUING,&lapb->critical);

#if 0
	if (wpabs_skb_queue_len(lapb->ack_queue)){
		wpabs_debug_event( "%s: Requeing %d non acked frames for Tx, new Vs=%d !\n",
			lapb->name,skb_queue_len(lapb->ack_queue),lapb->va);
	}
#endif
	
	//FIXME: LOCK NO LOCKS HERE, PLEASE CONFIRM!!!!
	
	while ((skb = wpabs_skb_dequeue(lapb->ack_queue)) != NULL) {
		if (skb_prev == NULL){
			wpabs_skb_queue_head(lapb->write_queue, skb);
		}else{
			wpabs_skb_append(skb_prev, skb);
		}
		skb_prev = skb;
	}

#if 0
	wpabs_debug_event( "%s: Requeue setting vs=%d back to va=%d\n",
			lapb->name,lapb->vs,lapb->va);
#endif

	lapb->vs=lapb->va;
	
	wpabs_clear_bit(LAPB_REQUEUING,&lapb->critical);
}


/*
 *	This routine is the centralised routine for parsing the control
 *	information for the different frame formats.
 */
void lapb_decode(lapb_cb *lapb, void* skb, struct lapb_frame *frame)
{
	unsigned char *data = wpabs_skb_data(skb);
	frame->type = LAPB_ILLEGAL;

	if (lapb->mode & LAPB_MLP) {
		if (lapb->mode & LAPB_DCE) {
			if (data[0] == LAPB_ADDR_D){
				frame->cr = LAPB_COMMAND;
			}else if (data[0] == LAPB_ADDR_C){
				frame->cr = LAPB_RESPONSE;
			}else{
				frame->type = LAPB_CORRUPT;
				return;
			}
		} else {
			if (data[0] == LAPB_ADDR_C){
				frame->cr = LAPB_COMMAND;
			}else if (data[0] == LAPB_ADDR_D){
				frame->cr = LAPB_RESPONSE;
			}else{
				frame->type = LAPB_CORRUPT;
				return;
			}
		}
	} else {
		if (lapb->mode & LAPB_DCE) {
			if (data[0] == LAPB_ADDR_B){
				frame->cr = LAPB_COMMAND;
			}else if (data[0] == LAPB_ADDR_A){
				frame->cr = LAPB_RESPONSE;
			}else{
				frame->type = LAPB_CORRUPT;
				return;
			}
		} else {
			if (data[0] == LAPB_ADDR_A){
				frame->cr = LAPB_COMMAND;
			}else if (data[0] == LAPB_ADDR_B){
				frame->cr = LAPB_RESPONSE;
			}else{
				frame->type = LAPB_CORRUPT;
				return;
			}
		}
	}
		
	wpabs_skb_pull(skb, 1);
	data = wpabs_skb_data(skb);

	if (lapb->mode & LAPB_EXTENDED) {
		if ((data[0] & LAPB_S) == 0) {
			frame->type       = LAPB_I;			/* I frame - carries NR/NS/PF */
			frame->ns         = (data[0] >> 1) & 0x7F;
			frame->nr         = (data[1] >> 1) & 0x7F;
			frame->pf         = data[1] & LAPB_EPF;
			frame->control[0] = data[0];
			frame->control[1] = data[1];
			wpabs_skb_pull(skb, 2);
		} else if ((data[0] & LAPB_U) == 1) { 	/* S frame - take out PF/NR */
			frame->type       = data[0] & 0x0F;
			frame->nr         = (data[1] >> 1) & 0x7F;
			frame->pf         = data[1] & LAPB_EPF;
			frame->control[0] = data[0];
			frame->control[1] = data[1];
			wpabs_skb_pull(skb, 2);
		} else if ((data[0] & LAPB_U) == 3) { 	/* U frame - take out PF */
			frame->type       = data[0] & ~LAPB_SPF;
			frame->pf         = data[0] & LAPB_SPF;
			frame->control[0] = data[0];
			frame->control[1] = 0x00;
			wpabs_skb_pull(skb, 1);
		}
	} else {
		if ((data[0] & LAPB_S) == 0) {
			frame->type = LAPB_I;			/* I frame - carries NR/NS/PF */
			frame->ns   = (data[0] >> 1) & 0x07;
			frame->nr   = (data[0] >> 5) & 0x07;
			frame->pf   = data[0] & LAPB_SPF;
		} else if ((data[0] & LAPB_U) == 1) { 	/* S frame - take out PF/NR */
			frame->type = data[0] & 0x0F;
			frame->nr   = (data[0] >> 5) & 0x07;
			frame->pf   = data[0] & LAPB_SPF;
		} else if ((data[0] & LAPB_U) == 3) { 	/* U frame - take out PF */
			frame->type = data[0] & ~LAPB_SPF;
			frame->pf   = data[0] & LAPB_SPF;
		}

		frame->control[0] = data[0];

		wpabs_skb_pull(skb, 1);
	}
}

/* 
 *	This routine is called when the HDLC layer internally  generates a
 *	command or  response  for  the remote machine ( eg. RR, UA etc. ). 
 *	Only supervisory or unnumbered frames are processed, FRMRs are handled
 *	by lapb_transmit_frmr below.
 */
void lapb_send_control(lapb_cb *lapb, int frametype, int poll_bit, int type)
{
	void *skb;
	unsigned char  *dptr;

	FUNC_BEGIN();

	if (lapb->condition & LAPB_RNR_SENT_CONDITION &&
	    frametype == LAPB_RR){
		wpabs_debug_event("%s: MAJORE ERROR !! RR send on RNR Cond!\n",
				lapb->name);
		return;
	}

	
	if (lapb->state ==  LAPB_STATE_CONNECTED   && 
	    frametype == LAPB_SABM){
		wpabs_debug_event("%s: MAJORE ERROR !! SABM send on Connected Cond!\n",
				lapb->name);
		return;
	}



	skb=wpabs_skb_alloc(LAPB_HEADER_LEN + 3);
	if (skb == NULL)
		return;
	
	wpabs_skb_reserve(skb, LAPB_HEADER_LEN + 1);

	if (lapb->mode & LAPB_EXTENDED) {
		if ((frametype & LAPB_U) == LAPB_U) {
			dptr   = wpabs_skb_put(skb, 1);
			*dptr  = frametype;
			*dptr |= (poll_bit) ? LAPB_SPF : 0;
		} else {
			dptr     = wpabs_skb_put(skb, 2);
			dptr[0]  = frametype;
			dptr[1]  = (lapb->vr << 1);
			dptr[1] |= (poll_bit) ? LAPB_EPF : 0;
		}
	} else {
		dptr   = wpabs_skb_put(skb, 1);
		*dptr  = frametype;
		*dptr |= (poll_bit) ? LAPB_SPF : 0;
		if ((frametype & LAPB_U) == LAPB_S)		/* S frames carry NR */
			*dptr |= (lapb->vr << 5);
	}

	lapb_update_stats (lapb, frametype, wpabs_skb_len(skb), 1);

	lapb_transmit_buffer(lapb, skb, type,1);

	FUNC_END();
}

/* 
 *	This routine generates FRMRs based on information previously stored in
 *	the LAPB control block.
 */
void lapb_transmit_frmr(lapb_cb *lapb)
{
	struct sk_buff *skb;
	unsigned char  *dptr;

	skb=wpabs_skb_alloc(LAPB_HEADER_LEN + 7);
	if (skb == NULL)
		return;

	wpabs_skb_reserve(skb, LAPB_HEADER_LEN + 1);

	if (lapb->mode & LAPB_EXTENDED) {
		dptr    = wpabs_skb_put(skb, 6);
		*dptr++ = LAPB_FRMR;
		*dptr++ = lapb->frmr_data.control[0];
		*dptr++ = lapb->frmr_data.control[1];
		*dptr++ = (lapb->vs << 1) & 0xFE;
		*dptr   = (lapb->vr << 1) & 0xFE;
		if (lapb->frmr_data.cr == LAPB_RESPONSE)
			*dptr |= 0x01;
		dptr++;
		*dptr++ = lapb->frmr_type;

	}else{
		dptr    = wpabs_skb_put(skb, 4);
		*dptr++ = LAPB_FRMR;
		*dptr++ = lapb->frmr_data.control[0];
		*dptr   = (lapb->vs << 1) & 0x0E;
		*dptr  |= (lapb->vr << 5) & 0xE0;
		if (lapb->frmr_data.cr == LAPB_RESPONSE)
			*dptr |= 0x10;
		dptr++;
		*dptr++ = lapb->frmr_type;

	}
	
	lapb_transmit_buffer(lapb, skb, LAPB_RESPONSE,1);
}

void lapb_update_stats (lapb_cb *lapb, int type, int len, char direction)
{
	FUNC_BEGIN();

	switch (type){

		case LAPB_I:
			if (!direction)
				++lapb->lapbstats.ifrm_rx;
			else
				++lapb->lapbstats.ifrm_tx;	
			break;
		
		case LAPB_RR:
			if (!direction)
				++lapb->lapbstats.rr_rx;
			else
				++lapb->lapbstats.rr_tx;	
			break;
		
		case LAPB_RNR:
			if (!direction)
				++lapb->lapbstats.rnr_rx;
			else
				++lapb->lapbstats.rnr_tx;	
			break;

		case LAPB_REJ:
			if (!direction)
				++lapb->lapbstats.rej_rx;
			else
				++lapb->lapbstats.rej_tx;	
			break;
			
		case LAPB_SABME:
		case LAPB_SABM:
			if (!direction)
				++lapb->lapbstats.sabm_rx;
			else
				++lapb->lapbstats.sabm_tx;	
			break;

		case LAPB_DISC:
			if (!direction)
				++lapb->lapbstats.disc_rx;
			else
				++lapb->lapbstats.disc_tx;	
			break;
		
		case LAPB_DM:
			if (!direction)
				++lapb->lapbstats.dm_rx;
			else
				++lapb->lapbstats.dm_tx;	
			break;
		
		case LAPB_UA:
			if (!direction)
				++lapb->lapbstats.ua_rx;
			else
				++lapb->lapbstats.ua_tx;	
			break;

		case LAPB_FRMR:
			if (!direction)
				++lapb->lapbstats.frmr_rx;
			else
				++lapb->lapbstats.frmr_tx;	
			break;
	
		default:
			wpabs_debug_event( "%s: %s illegal packet type 0x%02X !\n",
					lapb->name,direction ? "Tx" : "Rx" ,type);
			if (!direction)
				++lapb->lapbstats.ill_rx;
			else
				++lapb->lapbstats.ill_tx;	
			break;
	}

#if 0
	if (!direction){
		++lapb->ifstats.rx_packets;
		lapb->ifstats.rx_bytes += len;
	}else{
		++lapb->ifstats.tx_packets;
		lapb->ifstats.tx_bytes += len;
	}
#endif
}

void lapb_reconnect(lapb_cb *lapb)
{
	lapb_stop_t1timer(lapb);
	lapb_start_t1timer(lapb);
}


void lapb_link_up (lapb_cb *lapb)
{
	if (!lapb)
		return;

	if (wpabs_test_bit(LAPB_DOWN,&lapb->critical)){
		return;
	}

	lapb->condition = 0x00;
	lapb->n2count   = 0;
	lapb_clear_queues(lapb);

	lapb_stop_t1timer(lapb);
	lapb_start_t1timer(lapb);
	lapb_stop_t2timer(lapb);

	/* Do not actually try to establish connection 
	 * here.  Since frame relay needs time to start up
	 * let the T1 timer expire and retry the 
	 * startup sequence */
	lapb_state_change(lapb,LAPB_STATE_CONNECTING);
	return;
}

void lapb_link_down (lapb_cb *lapb)
{
	if (!lapb)
		return;

	lapb_state_change(lapb,LAPB_STATE_DISCONNECTED);
	lapb_stop_t1timer(lapb);
	lapb_stop_t2timer(lapb);
	lapb_stop_t4timer(lapb);

	lapb_disconnect_indication(lapb, 0);
	return;
}

