/*****************************************************************************
 * dsp_iface.c	WANPIPE DSP Protocol Module. Main code.
 *
 * Author:	Alex Feldman
 *              Nenad Corbic
 *
 * Copyright:	(c)2003 Sangoma Technologies Inc.
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 * ============================================================================
 *	This module:
 *		This module is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 */

#include <linux/wanpipe_includes.h>
#include <linux/wanpipe_cfg.h>
#include "wanpipe_dsp.h"

/*
 * ******************************************************************
 * 			DEFINES/MACROS
 * ******************************************************************
 */
#define MAX_TAIL_ROOM 16

/*
 * ******************************************************************
 * 			GLOBAL VARIABLES
 * ******************************************************************
 */
dsp_link_t *dsp_link_list;
spinlock_t dsp_link_lock;

#ifdef DSP_DEBUG
unsigned short	dsp_debug = DSP_DEBUG_ALL;
#endif


/*
 * ******************************************************************
 * 			FUNCTION PROTOTYPES
 * ******************************************************************
 */
static int dsp_if_send (struct sk_buff*, struct net_device*);

static int dsp_if_init(struct net_device *dev);
int dsp_open_dev(struct net_device *dev);
int dsp_stop_dev(struct net_device *dev);
void dsp_svc_down(void* dsp_id);
void dsp_svc_ready(void* dsp_id);

/*
 * ******************************************************************
 * 			FUNCTION DEFINITIONS
 * ******************************************************************
 */
static dsp_link_t* dev_to_dsp(struct net_device *dev)
{
	if (dev == NULL)
		return NULL;
	return (dsp_link_t*)dev->priv;
}

/*
 *	Free an allocated lapb control block. This is done to centralise
 *	the MOD count code.
 */
static void dsp_free_dev(struct net_device *dev)
{
	dev->priv = NULL;
	DSP_MEM_DEC(sizeof(struct net_device));
	KFREE(dev);
}

static void dsp_free_link(dsp_link_t *dsp_link){
	
	KFREE(dsp_link);
	DSP_MEM_DEC(sizeof(dsp_link_t));
#ifndef LINUX_2_6
	MOD_DEC_USE_COUNT;
#endif
}

static void dsp_remove_link(dsp_link_t *dsp_link)
{
	dsp_link_t *cur_link;
	dsp_link_t *prev_link;
	unsigned long flags;
	
	spin_lock_irqsave(&dsp_link_lock,flags);
	
	if ((cur_link = dsp_link_list) == dsp_link) {
		dsp_link_list = cur_link->next;
		spin_unlock_irqrestore(&dsp_link_lock,flags);
		return;
	}

	while (cur_link != NULL && cur_link->next != NULL) {
		prev_link=cur_link;
		cur_link = cur_link->next;
		if (cur_link == dsp_link) {
			prev_link->next = dsp_link->next;
			spin_unlock_irqrestore(&dsp_link_lock,flags);
			return;
		}
	}
	spin_unlock_irqrestore(&dsp_link_lock,flags);
}


/*
 *	Socket removal during an interrupt is now safe.
 */
static void dsp_remove_dev(dsp_link_t *dsp_link,struct net_device *dev)
{
	int err;

	if (!dev)
		return;
	
	dsp_link->dev = NULL;

	if (atomic_read(&dev->refcnt) != 1){
		printk(KERN_INFO "%s: !!!!! Warning: DSP Dev refcnt %i !!!!!\n",
				dev->name,atomic_read(&dev->refcnt));
		atomic_set(&dev->refcnt,1);
	}

	rtnl_lock();
	if ((err=unregister_netdevice(dev)) != 0){
		printk(KERN_INFO "DSP: Failed to unregister net device %s err=%i\n",
				dev->name,err);
	}
	rtnl_unlock();
	return;
}


static void dsp_insert_link(dsp_link_t *dsp_link)
{
	unsigned long flags;
	
	spin_lock_irqsave(&dsp_link_lock,flags);
	
	dsp_link->next = dsp_link_list;
	dsp_link_list = dsp_link;

	spin_unlock_irqrestore(&dsp_link_lock,flags);
}


/*
 *	Add a socket to the bound sockets list.
 */
static void dsp_insert_dev(dsp_link_t *dsp_link, struct net_device *dev)
{

	unsigned long flags;
	
	spin_lock_irqsave(&dsp_link_lock,flags);
	dsp_link->dev = dev;
	spin_unlock_irqrestore(&dsp_link_lock,flags);
}

static dsp_link_t *dsp_link_in_list(char *dev_name)
{
	dsp_link_t *cur_link;
	unsigned long flags;
	
	spin_lock_irqsave(&dsp_link_lock,flags);
	
	for ((cur_link = dsp_link_list); cur_link != NULL;  cur_link = cur_link->next){
		if (strcmp(cur_link->name,dev_name) == 0){
			spin_unlock_irqrestore(&dsp_link_lock,flags);
			return cur_link;
		}
	}
	spin_unlock_irqrestore(&dsp_link_lock,flags);
	return NULL;
}



static int dsp_link_ok(dsp_link_t *dsp_link)
{
	dsp_link_t *cur_link;
	unsigned long flags;
	
	spin_lock_irqsave(&dsp_link_lock,flags);
	
	for ((cur_link = dsp_link_list); cur_link != NULL;  cur_link = cur_link->next){
		if (cur_link == dsp_link){
			spin_unlock_irqrestore(&dsp_link_lock,flags);
			return DSP_OK;
		}
	}
	spin_unlock_irqrestore(&dsp_link_lock,flags);
	return -EINVAL;
}

int dsp_bind_api(struct net_device *master_dev, void *sk_id)
{
	struct net_device* dev = NULL;
	dsp_link_t *dsp_link = NULL;
	unsigned long flags;

	DEBUG1((KERN_INFO "DSP Finding free dev for call setup %s!\n", 
			master_dev->name));

	dev = x25_get_svc_dev(master_dev);
	DEBUG1((KERN_INFO "Debug: DSP Master device %s\n", 
		dev ? dev->name : "NULL")); 

	spin_lock_irqsave(&dsp_link_lock,flags);
	for ((dsp_link = dsp_link_list); dsp_link != NULL;  dsp_link = dsp_link->next){
		DEBUG1((KERN_INFO "Debug: DSP Link Master device %s\n", 
			dsp_link->master ? dsp_link->master->name : "NULL")); 
		if (dsp_link->master == dev){
			break;
		}
	}
	
	if (dsp_link == NULL) {
		spin_unlock_irqrestore(&dsp_link_lock,flags);
		printk(KERN_INFO "DSP Bind: No dsp link has a master device: %s !\n",
				master_dev->name);
		return -ENODEV;
	}

	if (dsp_link->dev == NULL) {
		spin_unlock_irqrestore(&dsp_link_lock,flags);
		printk(KERN_INFO "DSP Bind: No device on link %s\n",
				dsp_link->name);
		return -ENODEV;
	}	

	dev = dsp_link->dev;

	if ((dev->flags & IFF_UP) &&
	     dsp_link->usedby == API && !dsp_link->sk_id && 
	     dsp_link->state == DSP_READY){
		sock_hold(sk_id);
		dsp_link->sk_id = sk_id;
	}else{
		spin_unlock_irqrestore(&dsp_link_lock,flags);
		printk(KERN_INFO "DSP Bind: No Device found in link %s!\n",
				master_dev->name);
		return -ENODEV;
	}

	spin_unlock_irqrestore(&dsp_link_lock,flags);
	return dev->ifindex;
}

static dsp_link_t *dsp_create_link(struct net_device *device, char* dev_name)
{
	dsp_link_t *dsp = NULL;
		
	DEBUG1((KERN_INFO "%s: Creating new DSP link...\n", 
				dev_name));
	KMALLOC(dsp,sizeof(dsp_link_t),GFP_KERNEL);
	IS_NULL(dsp, NULL);
	DSP_MEM_INC(sizeof(dsp_link_t));

	memset(dsp,0,sizeof(dsp_link_t));
	
	dsp->master	= device;
	dsp->usedby	= API;
	dsp->critical	= 0x00;
	
	dsp->rx_sn = 0x00;
	dsp->tx_sn = 0x00;
	memset(dsp->array_sn, 0, DSP_SN_MODULES);
	dsp->last_ack_sn_l = DSP_SN_MODULES-1;
	dsp->last_ack_sn_r = DSP_SN_MODULES-1;
	
	dsp->is_rx_more_data = DSP_FALSE;
	dsp->is_tx_more_data = DSP_FALSE;
	memset(&dsp->dsp_api_event, 0, sizeof(dspapi_t));
	//memset(&dsp->rx_hdr, 0, sizeof(dspapi_hdr_t));
	dsp->rx_skb = NULL;
	dsp->tx_pkt_len = 0;
	
	init_timer(&dsp->T1_timer);
	init_timer(&dsp->T2_timer);
	init_timer(&dsp->T3_timer);

	skb_queue_head_init(&dsp->async_queue);
	skb_queue_head_init(&dsp->write_queue);

	spin_lock_init(&dsp->map_lock);
	spin_lock_init(&dsp->list_lock);

	tasklet_init(&dsp->dsp_tx_task,dsp_link_send,(u32)dsp);
	memset(&dsp->cfg, 0, sizeof(struct dsp_parms));	

	atomic_set(&dsp->mem_in_used, 0);
 
	dsp->state = DSP_DISCONNECTED;

#ifndef LINUX_2_6
	MOD_INC_USE_COUNT;
#endif

	return dsp;
}

static struct net_device *dsp_create_dev(char *dev_name)
{
	struct net_device *dev = NULL;

	DEBUG1((KERN_INFO "%s: Creating new DSP device...\n", 
				dev_name));
	KMALLOC(dev,sizeof(struct net_device), GFP_KERNEL);
	IS_NULL(dev, NULL);
	DSP_MEM_INC(sizeof(struct net_device));

	memset(dev, 0, sizeof(struct net_device));
	strcpy(dev->name, dev_name);
	
	return dev;
}

int dsp_ioctl (struct net_device *dev, struct ifreq *ifr, int cmd)
{
	dsp_link_t*	dsp_link = NULL;
	int		err = 0;
	unsigned long flags;
	
	DSP_ASSERT(dev == NULL);
	DSP_ASSERT(dev->priv == NULL);
	
	if (!(dev->flags & IFF_UP)){
		return -ENODEV;
	}

	dsp_link = dev->priv;
	
	switch (cmd){
		/* Common API calls */
		case SIOC_DSP_SET_CALL_DATA:
			return dsp_api_cmd_setup(dsp_link, ifr);

		case SIOC_DSP_GET_CALL_DATA:
			return dsp_api_get_cmd_result(dsp_link, ifr);

		case SIOC_DSP_PLACE_CALL:
			return dsp_api_place_call(dsp_link, ifr);

		case SIOC_DSP_CLEAR_CALL:
			return dsp_api_clear_call(dsp_link, ifr);

		case SIOC_DSP_ACCEPT_CALL:
			return dsp_api_accept_call(dsp_link,ifr);

		case SIOC_DSP_RESET_CALL:
			return dsp_api_reset_call(dsp_link,ifr);

		/* DSP private API calls */
		case SIOC_DSP_INVITATION_TO_CLEAR:
			return dsp_api_inv_to_clear(dsp_link, ifr);

		case SIOC_DSP_CIRCUIT_ENABLED:
			return dsp_api_circuit_enabled(dsp_link, ifr);

		case SIOC_DSP_CIRCUIT_RESET:
			return dsp_api_circuit_reset(dsp_link, ifr, DSP_FALSE);

		case SIOC_DSP_CIRCUIT_DISCONNECT:
			return dsp_api_circuit_disconnect(dsp_link, ifr);

		case SIOC_DSP_STATUS:
			return dsp_api_status(dsp_link, ifr);

		case SIOC_DSP_SET_PID:
			if (ifr && ifr->ifr_data && 
			    dsp_link->master && dsp_link->master->do_ioctl){
				return dsp_link->master->do_ioctl(
						dsp_link->master,
						ifr,
						SIOC_X25_SET_LCN_PID);
			}
			break;

		case SIOC_DSP_SET_LCN_LABEL:
			if (ifr && ifr->ifr_data && 
			    dsp_link->master && dsp_link->master->do_ioctl){
				return dsp_link->master->do_ioctl(
						dsp_link->master,
						ifr,
						SIOC_X25_SET_LCN_LABEL);
			}
			break;

		case SIOC_DSP_ACK:
			return dsp_api_eeack(dsp_link, ifr);

		case SIOC_DSP_CIRCUIT_REQUEST:
			return dsp_api_circuit_request(dsp_link, ifr);

		case SIOC_DSP_CMD_RES_UNDELIVERED:
			return dsp_api_cmd_res_undelivered(dsp_link, ifr);

		case SIOC_DSP_CMD_RES_ABORTED:
			return dsp_api_cmd_res_aborted(dsp_link, ifr);

		/* Socket private calls */
		case SIOC_ANNEXG_BIND_SK:
			DSP_ASSERT(ifr == NULL);
			spin_lock_irqsave(&dsp_link->list_lock,flags);
			if (ifr->ifr_data){
				sock_hold((struct sock*)ifr->ifr_data);
				dsp_link->sk_id = ifr->ifr_data;
			}
			spin_unlock_irqrestore(&dsp_link->list_lock,flags);
			break;

		case SIOC_ANNEXG_UNBIND_SK:
			spin_lock_irqsave(&dsp_link->list_lock,flags);	
			if (dsp_link->sk_id){
				__sock_put(dsp_link->sk_id);
				dsp_link->sk_id=0;
			}
			spin_unlock_irqrestore(&dsp_link->list_lock,flags);	
			break; 
			
		case SIOC_ANNEXG_GET_SK:
			DSP_ASSERT(ifr == NULL);

			if (dsp_link->sk_id){
				sock_hold(dsp_link->sk_id);
				ifr->ifr_data = (void*)dsp_link->sk_id;
			}else{
				ifr->ifr_data = NULL;
			}
			
			break;

		case SIOC_WANPIPE_SET_NONBLOCK:
			printk(KERN_INFO "SET_NONBLOCK (not supported)\n");
			return -EOPNOTSUPP;

		case SIOC_WANPIPE_CHECK_TX:
			if (skb_queue_len(&dsp_link->write_queue)){
				return 1;
			}
			if (skb_queue_len(&dsp_link->async_queue)){
				return 1;
			}
			break;

		case SIOC_WANPIPE_SOCK_STATE:
			printk(KERN_INFO "SOCK_STATE (not supported)\n");
			return -EOPNOTSUPP;

		case SIOC_ANNEXG_KICK:
			DEBUG6((KERN_INFO "%s: SIOC_ANNEXG_KICK\n",
					dev->name));
			if (dsp_link->master && 
			    dsp_link->master->do_ioctl)
				err = dsp_link->master->do_ioctl(
						dsp_link->master,
						NULL,
						SIOC_ANNEXG_KICK);
			else
				return -ENODEV;
			return err;

		default:
			printk(KERN_INFO "%s: Unsupported IOCTL command\n",
					dev->name);
			return -EOPNOTSUPP;
	}
	return 0;	
}

int dsp_set_config (struct net_device *dev,struct ifmap *map)
{
	printk(KERN_INFO "GOT Set Config COMMAND\n");
	return 0;	
}


/* DSP REGSTRATION FUNCTIONS */

int dsp_register(struct net_device *x25_dev, char *dev_name,struct net_device **new_dev)
{
	dsp_link_t *dsp_link = NULL;
	struct net_device *dev = NULL;
	int err;

	DSP_ASSERT2(x25_dev == NULL, -ENODEV);
	DSP_ASSERT2(dev_name == NULL,-ENODEV);

	DEBUG1((KERN_INFO "%s: Registering DSP Device...\n", dev_name));
	
	/* Create the new DSP link for this X25 
	 * connection.
         */
	if (dsp_link_in_list(dev_name) != NULL){
		printk(KERN_INFO "%s: Failed to register dsp link: Already exists!\n",
				dev_name);
		return -EEXIST;
	}
	
	DEBUG1((KERN_INFO "%s: Creating DSP Link...\n", dev_name));
	dsp_link = dsp_create_link(x25_dev, dev_name);
	if (!dsp_link){
		printk(KERN_INFO "DSP: Failed to create DSP link... %s\n",
				dev_name);
		return -ENOMEM;
	}

	strcpy(dsp_link->name, dev_name);
	

	if ((dev = dsp_create_dev(dev_name)) == NULL){
		printk(KERN_INFO "DSP: Failed to create dev %s\n",
				dev_name);
		dsp_free_link(dsp_link);
		return -ENOMEM;
	}

	dev->init = &dsp_if_init;
	dev->priv = dsp_link;

	rtnl_lock();
	if ((err=register_netdevice(dev)) != 0){
		//DEBUG OPTION
		rtnl_unlock();
		dsp_free_dev(dev);
		dsp_free_link(dsp_link);
		printk(KERN_INFO "DSP: Failed to register net dev %s : err=%i\n",
				dev_name,err);
		return -EFAULT;
	}
	rtnl_unlock();

	dsp_link->dev = dev;
	
#ifdef CONFIG_PROC_FS
	if (dsp_proc_link_init (dsp_link) != DSP_OK){
		dsp_remove_dev(dsp_link,dev);
		dsp_free_dev(dev);
		dsp_free_link(dsp_link);
		printk(KERN_INFO "%s: Failed to create DSP link proc entry..\n",
				dev_name);
		return -EFAULT;
	}
#endif

	dsp_insert_dev(dsp_link, dev);
	dsp_insert_link(dsp_link);

	*new_dev=dev;
	return 0;
}

int dsp_unregister(struct net_device **x25_dev, void* dsp_id)
{
	struct net_device* dsp_dev = (struct net_device*)dsp_id;
	dsp_link_t *dsp_link = NULL;
	unsigned long flags;

	DSP_ASSERT(dsp_dev == NULL);
	DSP_ASSERT(dsp_dev->priv == NULL);

	dsp_link = (dsp_link_t *)dsp_dev->priv;
	if (dsp_dev->flags & IFF_UP){
		printk(KERN_INFO "%s: Failed to unregister: device UP!\n",
				dsp_dev->name);
		return -EBUSY;
	}
	
	if (dsp_link_ok(dsp_link) != DSP_OK){
		printk(KERN_INFO "%s: Failed to unregister: no dsp device!\n",
				dsp_dev->name);
		return -ENODEV;
	}

	set_bit(CRIT_LINK_DOWN,&dsp_link->critical);
	tasklet_kill(&dsp_link->dsp_tx_task);
	schedule();

	printk(KERN_INFO "%s: Unregistering DSP device\n", dsp_dev->name);

	if (x25_dev != NULL){
		*x25_dev = dsp_link->master;
	}

	dsp_clear_link_queues(dsp_link, DSP_DISCONNECTED);

#ifdef CONFIG_PROC_FS
	dsp_proc_link_cleanup(dsp_link);
#endif
	
	dsp_remove_dev(dsp_link, dsp_dev);
	dsp_free_dev(dsp_dev);

	/* Remove the dsp device pointer from the
	 * lower (x25) layer, so it will stop sending
	 * us packets */
	
	spin_lock_irqsave(&dsp_link->list_lock,flags);	
	x25_dsp_id_unregister(dsp_link->master);
	dsp_link->master=NULL;
	spin_unlock_irqrestore(&dsp_link->list_lock,flags);
	
	dsp_remove_link(dsp_link);
	dsp_free_link(dsp_link);
	return 0;
}


int dsp_getparms(struct net_device *dev, struct dsp_parms *prm)
{
#if 0
	dsp_link_t* dsp_link = NULL;
	
	if ((dsp_link = (dsp_link_t*)dev->priv) == NULL){
		return DSP_BADTOKEN;
	}
	
	prm->t1      = dsp_link->cfg.T1 / HZ;
	prm->t2      = dsp_link->cfg.T2 / HZ;
	prm->state   = dsp->state;

	if (!timer_pending(&dsp->t1timer))
		prm->t1timer = 0;
	else
		prm->t1timer = (dsp->t1timer.expires - jiffies) / HZ;

	if (!timer_pending(&dsp->t2timer))
		prm->t2timer = 0;
	else
		prm->t2timer = (dsp->t2timer.expires - jiffies) / HZ;
#endif
	return DSP_OK;
}

int dsp_setparms(void* dsp_id, void *prm)
{
	struct net_device* dsp_dev = (struct net_device*)dsp_id;
	dsp_link_t *dsp_link = NULL;
	struct dsp_parms* dsp_cfg = (struct dsp_parms*)prm;
	
	DSP_ASSERT(dsp_dev == NULL);
	DSP_ASSERT(dsp_dev->priv == NULL);
	DSP_ASSERT(dsp_cfg == NULL);

	dsp_link = (dsp_link_t *) dsp_dev->priv;

	/* Copy all parameters from 'prm' to dsp_link */
	dsp_link->cfg.pad_type 		= dsp_cfg->pad_type;
	dsp_link->cfg.T1 		= dsp_cfg->T1*HZ;
	dsp_link->cfg.T2 		= dsp_cfg->T2*HZ;
	dsp_link->cfg.T3 		= dsp_cfg->T3*HZ;
	dsp_link->cfg.auto_ce 		= dsp_cfg->auto_ce;
	dsp_link->cfg.auto_call_req 	= dsp_cfg->auto_call_req;
	dsp_link->cfg.auto_ack 		= dsp_cfg->auto_ack;

	dsp_dev->mtu = (dsp_cfg->dsp_mtu) ? dsp_cfg->dsp_mtu : DSP_MTU_DEFAULT;
	
	if (dsp_link->cfg.pad_type == DSP_HPAD)
		printk(KERN_INFO "%s: Configuring DSP to HOST PAD\n",
				dsp_link->name);
	else
		printk(KERN_INFO "%s: Configuring DSP to TERM PAD\n",
				dsp_link->name);

	printk(KERN_INFO "\n");
	return DSP_OK;
}

/*==============================================================
 * x25_ifstats
 *
 * Description:
 * 	This fucntion interfaces the /proc file system
 * 	to the svc device.  The svc keeps protocol and
 * 	packet statistcs.  This function passes these
 * 	statistics to the /proc file system.
 *	
 * Usedby:
 * 	Kernel /proc file system
 * 	x25_proc.c function also uses it to obtain svc
 * 	statistics.
 */

static struct net_device_stats* dsp_ifstats (struct net_device *dev)
{
	dsp_link_t *dsp_link;
	
	WAN_ASSERT2((!dev),NULL);
	
	dsp_link = (dsp_link_t *)dev->priv;
	if (!dsp_link)
		return NULL;

	return &dsp_link->ifstats;
}



static int dsp_if_init(struct net_device *dev)
{
	dev->open		= &dsp_open_dev;
	dev->stop		= &dsp_stop_dev;
	dev->hard_header	= NULL;
	dev->rebuild_header	= NULL;
	dev->hard_start_xmit	= &dsp_if_send;
	dev->get_stats		= &dsp_ifstats;
	dev->tx_timeout		= NULL;
	dev->watchdog_timeo	= 0;
	dev->hard_header_len	= 25;
	dev->do_ioctl		= &dsp_ioctl;
	dev->set_config		= &dsp_set_config;
	
	/* Initialize media-specific parameters */
	dev->flags		|= IFF_POINTOPOINT;
	dev->flags		|= IFF_NOARP;

	dev->tx_queue_len = 10;
	// FIXME: ARPHRD type
	dev->type = ARPHRD_VOID;

	/* Initialize socket buffers */
	dev_init_buffers(dev);
	return 0;
}

int dsp_open_dev(struct net_device *dev)
{
	dsp_link_t *dsp_link = NULL;
	
	DSP_ASSERT(dev == NULL);
	DSP_ASSERT(dev->priv == NULL);
	dsp_link = (dsp_link_t*)dev->priv;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,18)
	if (netif_running(dev))
		return -EBUSY;
#endif

	dsp_link->usedby = API;

	if (test_bit(CRIT_LINK_DOWN,&dsp_link->critical)){
		clear_bit(CRIT_LINK_DOWN,&dsp_link->critical);
	}

	if (x25_dsp_is_svc_ready(dsp_link->master) == 0){
		dsp_svc_ready(dev);
		netif_start_queue(dev);
	}
	return 0;
}

int dsp_stop_dev(struct net_device *dev)
{
	dsp_link_t*dsp_link = dev_to_dsp(dev);
	DSP_ASSERT(dev == NULL);
	netif_stop_queue(dev);

	if (dsp_link && dsp_link->sk_id){
		unsigned long flags;
		dsp_api_clear_call(dsp_link, NULL);
		
		spin_lock_irqsave(&dsp_link->list_lock,flags);
		if (dsp_link->sk_id){
			__sock_put(dsp_link->sk_id);
			dsp_link->sk_id=NULL;
		}
		spin_unlock_irqrestore(&dsp_link->list_lock,flags);
		
		set_bit(CRIT_LINK_DOWN,&dsp_link->critical);
	}

	return 0;
}


static int dsp_if_send (struct sk_buff* skb, struct net_device* dev)
{
	struct sk_buff *skb_old = skb;
	struct sk_buff *skb2 = NULL;
	dsp_link_t* 	dsp_link = dev_to_dsp(dev);
	int 		err = DSP_OK, ret = 0;

	DSP_ASSERT(dsp_link == NULL);
	DSP_ASSERT(skb == NULL);

	DEBUG3((KERN_INFO "%s: RX packet (%d) from application...\n", 
			dev->name, skb->len));
	netif_stop_queue(dev);

	DSP_SKB_INC(skb->truesize,dsp_link);
	
	if (dsp_link->state != DSP_CONNECTED ||
	    test_bit(CRIT_LINK_DOWN,&dsp_link->critical)){
		printk(KERN_INFO "%s: Device is not in CONNECTED state! %i\n",
			dsp_link->name, dsp_link->state);
		KFREE_SKB(skb,dsp_link);
		return 0;
	}
	
	/* discard the frame if we are configured for */
	/* 'receive only' mode or if there is no data */
	if (skb->len <= sizeof(dspapi_hdr_t)) {
		KFREE_SKB(skb,dsp_link);
		netif_start_queue(dev);
		printk(KERN_INFO "%s: Invalid packet size (%i)\n",
			dev->name, skb->len);
		return 0;
	}

	if (skb_queue_len(&dsp_link->write_queue) != 0){
		DEBUG3((KERN_INFO "%s: Device is busy\n",
				dsp_link->name));
		dsp_link->dsp_api_event.hdr.dsp_result = EBUSY;
		err = -EBUSY;
		goto dsp_ifsend_exit;
	}

	
	//FIXME THIS iS SOOOOO UGLY !!!!
	if (skb_headroom(skb) < 16 || skb_tailroom(skb) < 16){
	
		skb_unlink(skb);

		skb2 = skb_copy_expand(skb, 16, 16, GFP_ATOMIC);

		if (skb2 == NULL){
			printk(KERN_INFO "%s: Failed in pkt reallocation\n",
					dev->name);
			netif_start_queue(dev);
			KFREE_SKB(skb,dsp_link);
			return 0;
		}
		
		skb = skb2;

		skb_unlink(skb);
		
		DEBUG3((KERN_INFO "After Reallocating packet H %i T %i!!!\n",
				skb_headroom(skb),skb_tailroom(skb)));
	}
	
	if ((err = dsp_split_tx_data(dsp_link, skb) != DSP_OK)){
		netif_start_queue(dev);
		dsp_link->dsp_api_event.hdr.dsp_result = err;

		/* This failure should never happen.  Indicate to the
		 * user that the failure occured.  If we created the
		 * new packet skb2, delete it, and return socket
		 * the old skb packet. */
 		if (skb2){
			skb=skb_old;
 			KFREE_SKB(skb2,dsp_link);
 		}
		DSP_SKB_DEC(skb_old->truesize,dsp_link);
 	}else{
		/* The packet was queued up ok. We will return 0 to the
		 * socket. The socket will asume that the incoming
		 * skb buffer has been removed. 
		 *
		 * Thus, if we reallocated
		 * the original skb to skb2. Then free the original
		 * skb since skb2 has been queued up. */
		
		dsp_link->dsp_api_event.hdr.dsp_sn = dsp_link->tx_sn;
 		if (skb2){
 			KFREE_SKB(skb_old,dsp_link);
 		}
 	}

dsp_ifsend_exit:
	ret = dsp_trigger_link_bh(dsp_link);
	DEBUG7((KERN_INFO "%s: Triggering BH TX (if_send)...%s\n",
			dsp_link->name, (!ret)?"Done":"Failed"));

	return err;
}


//Called by the X.25 protocol to give us the packet.
//
int dsp_receive_pkt(void *dsp_id, struct sk_buff *skb)
{
	struct net_device* 	dsp_dev = (struct net_device*)dsp_id;
	dsp_link_t*		dsp_link = NULL;
	x25api_hdr_t*		x25api_hdr = NULL;
	unsigned char		is_q_bit = DSP_FALSE;

	if (dsp_dev == NULL || skb == NULL){
		printk(KERN_INFO "DSP Error: received data without %s!\n",skb?"link id":"skb pkt"); 	
		if (skb != NULL){
			skb_unlink(skb);
			dev_kfree_skb_any(skb);
		}
		return DSP_OK;
	}

	if (!(dsp_dev->flags&IFF_UP)|| dsp_dev->priv == NULL){
		printk(KERN_INFO "%s: DSP received data dev down!\n",dsp_dev->name); 	
		if (skb != NULL){
			skb_unlink(skb);
			dev_kfree_skb_any(skb);
		}
		return DSP_OK;
	}
	
	dsp_link = (dsp_link_t*)dsp_dev->priv;

	DSP_SKB_INC(skb->truesize, dsp_link);
	skb_unlink(skb);
	
	if (test_bit(CRIT_LINK_DOWN,&dsp_link->critical)){
		KFREE_SKB(skb,dsp_link);
		return DSP_OK;
	}
	
	if (skb->len <= sizeof(x25api_hdr_t)){
		printk(KERN_INFO "%s: Invalid packet length (%i)!\n",
			dsp_dev->name, skb->len);
		KFREE_SKB(skb,dsp_link);
		return DSP_OK;
	}

	x25api_hdr = (x25api_hdr_t*)skb->data;
	is_q_bit = IS_L_Q_BIT(x25api_hdr->qdm);

	/* Stop T2 timer (if was started before) */
	dsp_stop_t2_timer(dsp_link);


	dsp_link->ifstats.rx_bytes+=skb->len;

	DEBUG2((KERN_INFO "%s: Rx packet (%d) from X25!\n", 
			dsp_dev->name, skb->len - sizeof(x25api_hdr_t))); 	
			
	if ((skb = dsp_concat_rx_data(dsp_link, skb)) == NULL){
		/* There are more packet for the same M_DATA */
		/* Start T2 timer again only for Data Message */
		if (is_q_bit == DSP_FALSE)
			dsp_start_t2_timer(dsp_link);
		return DSP_OK;
	}
	DEBUG2((KERN_INFO "%s: Rx M_DATA (%d)!\n", 
			dsp_dev->name, skb->len - sizeof(x25api_hdr_t))); 	
	dsp_data_input(dsp_link, skb);

	return DSP_OK;
}


int dsp_data_tx_up(dsp_link_t *dsp_link, struct sk_buff *skb)
{
	int	skb_len = 0;
	int 	err = DSP_OK;

	// For now just pass it up the socket.
	// If DSP protocol has not bound into us yet.
	//
	DSP_ASSERT(dsp_link == NULL);
	DSP_ASSERT(dsp_link->dev == NULL);
	DSP_ASSERT(dsp_link->sk_id == NULL);
	DEBUG2((KERN_INFO "%s: TX packet (len=%d) up to socket\n",
			dsp_link->name, skb->len-sizeof(dspapi_hdr_t)));
	skb_unlink(skb);

	skb->pkt_type = WAN_PACKET_DATA;
	skb->dev = dsp_link->dev;
	skb->protocol = htons(DSP_PROT);
	skb->mac.raw  = skb->data;
	skb_len = skb->truesize;
	err = wanpipe_api_sock_rx(skb, dsp_link->dev, dsp_link->sk_id);
	if (err != DSP_OK){
		/* Failed to send packet to the socket.. dropping */
		DSP_INC_RX_STAT(rx_dropped);
		err=0;
		printk(KERN_INFO "%s: Major Error in Socket Above: Tx UP (Dropping)!!!\n",
					dsp_link->name);
		if (err == -EINVAL){
			unsigned long flags;
			spin_lock_irqsave(&dsp_link->list_lock,flags);
			if (dsp_link->sk_id){
				__sock_put(dsp_link->sk_id);
			}
			dsp_link->sk_id = NULL;
			spin_unlock_irqrestore(&dsp_link->list_lock,flags);
		}
		KFREE_SKB(skb,dsp_link);
		/* Set reason code */
		dsp_link->reason_code = DSP_RC_UNDEFINED;
		/* Call INVITATION_TO_CLEAR now */
		dsp_api_inv_to_clear(dsp_link, NULL);
	}else{
		DSP_SKB_DEC(skb_len, dsp_link);
	}
	return DSP_OK;
}

int dsp_data_tx_down(dsp_link_t *dsp_link, struct sk_buff *skb)
{
	int err=0;
	int truesize=skb->truesize;
	int len=skb->len;
	
	DSP_ASSERT(dsp_link == NULL);
	DSP_ASSERT(dsp_link->dev == NULL);
	skb_unlink(skb);
	skb->dev = dsp_link->master;

	
	DEBUG3((KERN_INFO "%s: Tx packet (len=%d) down to X25\n", 
			dsp_link->name, skb->len-sizeof(x25api_hdr_t)));

	if (dsp_link->master && dsp_link->master->hard_start_xmit){
		if (netif_queue_stopped(dsp_link->master)){
			return -EBUSY;
		}
		err = dsp_link->master->hard_start_xmit(skb, dsp_link->master);
		if (!err){
			DSP_SKB_DEC(truesize, dsp_link);
			dsp_link->ifstats.tx_bytes +=len;
		}
		return err;
	} else {
		printk(KERN_INFO "%s: No Master Device !\n", dsp_link->name);
		return -ENODEV;
	}
	return 0;
}

int dsp_call_request(void* dsp_id, void* sk_id, struct sk_buff* skb) 
{
	struct net_device* 	dsp_dev = (struct net_device*)dsp_id;
	struct ifreq	  	ifr_x25;	
	dsp_link_t*		dsp_link = NULL;
	x25api_t		x25_api;
	int 			err = 0;

	DSP_ASSERT(dsp_dev == NULL);
	DSP_ASSERT(dsp_dev->priv == NULL);
	DSP_ASSERT(sk_id == NULL);

	dsp_link = (dsp_link_t *)dsp_dev->priv;
	printk(KERN_INFO "%s: RX Incoming Call\n", dsp_dev->name);
	dsp_link->dsp_api_event.hdr.dsp_pad_type = dsp_link->cfg.pad_type;

	x25_get_call_info(dsp_link->master, &dsp_link->lcn, NULL,NULL);
	
	// 11.13.2001 Read incoming call string
	if (dsp_link->master && dsp_link->master->do_ioctl){
		int err = 0;
		memset(&x25_api, 0x00, sizeof(x25api_t));
		memset(&dsp_link->dsp_api_event, 0x00, sizeof(dspapi_t));
		x25_api.hdr.length = strlen(x25_api.data);
		ifr_x25.ifr_data = (char*)&x25_api;
		err = dsp_link->master->do_ioctl(dsp_link->master, 
						 &ifr_x25, 
				   		 SIOC_X25_GET_CALL_DATA);
		if (!err){
			dsp_link->dsp_api_event.hdr.dsp_length = x25_api.hdr.length;
			memcpy(dsp_link->dsp_api_event.data, 
			       x25_api.data, 
			       x25_api.hdr.length);
		}
	}
	DSP_INC_RX_STAT(rx_call_req);
	
	err = dsp_snd_api_call_request(dsp_link, sk_id, skb);
	if (err == 0){
		dsp_set_state(dsp_link, DSP_CONNECTING);
	}
	return err;
}

// FIXME: this function never called!!!!!!!!!!!!!!!!!!!!!!
int dsp_call_accept(void* dsp_id, struct sk_buff* skb) 
{
	struct net_device* dsp_dev = (struct net_device*)dsp_id;
	dsp_link_t *dsp_link = NULL;

	DSP_ASSERT(dsp_dev == NULL);
	DSP_ASSERT(dsp_dev->priv == NULL);
	dsp_link = (dsp_link_t *)dsp_dev->priv;
	printk(KERN_INFO "%s: Rx Call Accepted\n", dsp_dev->name);
			
	DSP_INC_RX_STAT(rx_call_acc);
	KFREE_SKB(skb,dsp_link);
	return 0;
}

void dsp_svc_up(void* dsp_id) 
{
	struct net_device* 	dsp_dev = (struct net_device*)dsp_id;
	dsp_link_t*		dsp_link = NULL;

	DSP_ASSERT1(dsp_dev == NULL);
	DSP_ASSERT1(dsp_dev->priv == NULL);
	dsp_link = (dsp_link_t *)dsp_dev->priv;
	DEBUG1((KERN_INFO "%s: Connection is up\n", dsp_dev->name));
	/* If auto_call_req is ON, stop T1 timer */
	if ((dsp_link->state == DSP_CONNECTING || 
	     dsp_link->state == DSP_CONNECTING_CALL_REQ_AGAIN) && 
	    dsp_link->cfg.auto_call_req)
		dsp_stop_t1_timer(dsp_link);
	dsp_set_state(dsp_link, DSP_CONNECTING_ACC_CALL);
	if (dsp_link->sk_id){
		int err = protocol_connected(dsp_dev, dsp_link->sk_id);
		if (err == -EINVAL){
			unsigned long flags;
			printk(KERN_INFO "%s: Major Error in Socket Above: CONN!!!\n",
					dsp_link->name);
			spin_lock_irqsave(&dsp_link->list_lock,flags);
			if (dsp_link->sk_id){
				__sock_put(dsp_link->sk_id);
			}
			dsp_link->sk_id = NULL;
			spin_unlock_irqrestore(&dsp_link->list_lock,flags);
			/* Set reason code */
			dsp_link->reason_code = DSP_RC_UNDEFINED;
			/* Call INVITATION_TO_CLEAR now */
			dsp_api_inv_to_clear(dsp_link, NULL);
			return;
		}
	}
	if (!x25_get_call_info(dsp_link->master, 
			      &dsp_link->lcn, 
			      &dsp_link->tx_pkt_len, 
			      &dsp_link->rx_pkt_len)){
		/* Reinitialize DSP mtu for real value (7 June 2002) */
		printk(KERN_INFO "%s: Bound to LCN %i X.25 MTU %i X.25 MRU %i\n", 
					dsp_link->name, 
					dsp_link->lcn,
					dsp_link->tx_pkt_len,
					dsp_link->rx_pkt_len);
	}else{
		printk(KERN_INFO "%s: No master device\n", 
				dsp_dev->name);
		return;
	}
	DSP_INC_RX_STAT(rx_call_acc);
}

void dsp_svc_down(void* dsp_id) 
{
	struct net_device* 	dsp_dev = (struct net_device*)dsp_id;
	dsp_link_t*		dsp_link = NULL;
	dspapi_hdr_t		hdr;

	DSP_ASSERT1(dsp_dev == NULL);
	DSP_ASSERT1(dsp_dev->priv == NULL);

	dsp_link = (dsp_link_t *)dsp_dev->priv;
	/* Stop all timers */
	dsp_stop_t1_timer(dsp_link);
	dsp_stop_t2_timer(dsp_link);
	dsp_stop_t3_timer(dsp_link);
	/* Initialize sequence numbers */
	dsp_link->rx_sn = 0;
	dsp_link->tx_sn = 0;
	memset(dsp_link->array_sn,0,DSP_SN_MODULES);
	dsp_link->last_ack_sn_l = DSP_SN_MODULES-1;
	dsp_link->last_ack_sn_r = DSP_SN_MODULES-1;
	/* Initialize RX */
	dsp_free_rx_skb(dsp_link);
	/* Initialize RX */
	dsp_link->is_tx_more_data = DSP_FALSE;
	dsp_link->lcn = 0;

	DEBUG1((KERN_INFO "%s: Connection is down\n", dsp_dev->name));
	dsp_set_state(dsp_link, DSP_DISCONNECTED);

	if (dsp_link->sk_id){
		unsigned long flags;
	
		hdr.dsp_pkt_type = 0x00;
		hdr.dsp_result   = DSP_LINK_DOWN;
		dsp_snd_api_oob_msg(dsp_link, &hdr, NULL, 0);

		
		protocol_disconnected(dsp_link->sk_id);
		spin_lock_irqsave(&dsp_link->list_lock,flags);
		if (dsp_link->sk_id){
			__sock_put(dsp_link->sk_id);
		}
		dsp_link->sk_id = NULL;
		spin_unlock_irqrestore(&dsp_link->list_lock,flags);
	}
}


void dsp_svc_ready(void* dsp_id)
{
	struct net_device* 	dsp_dev = (struct net_device*)dsp_id;
	struct sk_buff* 	skb = NULL;
	dsp_link_t*		dsp_link = NULL;

	DSP_ASSERT1(dsp_dev == NULL);
	DSP_ASSERT1(dsp_dev->priv == NULL);

	dsp_link = (dsp_link_t *)dsp_dev->priv;
	DEBUG1((KERN_INFO "%s: Connection is ready\n", dsp_dev->name));
	if (dsp_link->master){
		dsp_link->tx_pkt_len = dsp_link->master->mtu -
			       	       sizeof(x25api_hdr_t); 
	}else{
		printk(KERN_INFO "%s: No master device\n", dsp_dev->name);
		dsp_link->tx_pkt_len = 0x00;
	}
	// 11.26.2001
	dsp_link->lcn = 0;
	/* Clear all packet in write/async queue */

	while ((skb = skb_dequeue(&dsp_link->write_queue)) != NULL){
		KFREE_SKB(skb,dsp_link);
	}
	while ((skb = skb_dequeue(&dsp_link->async_queue)) != NULL){
		KFREE_SKB(skb,dsp_link);
	}
	/* Clear BH_AWAITING_KICK flag (we don't have packets in queue */
	clear_bit(BH_AWAITING_KICK, &dsp_link->tq_working);

	if (dsp_link->state == DSP_CONNECTING_CALL_REQ_AGAIN)
		dsp_api_place_call(dsp_link, NULL);
	else
		dsp_set_state(dsp_link, DSP_READY);
}

void dsp_svc_reset(void* dsp_id)
{
	struct net_device* 	dsp_dev = (struct net_device*)dsp_id;
	dsp_link_t*		dsp_link = NULL;
	dspapi_hdr_t		dsp_hdr;	

	DSP_ASSERT1(dsp_dev == NULL);
	DSP_ASSERT1(dsp_dev->priv == NULL);
	dsp_link = (dsp_link_t *)dsp_dev->priv;
	printk(KERN_INFO "%s: Connection is reset\n", dsp_dev->name);

	/* Sending OOB message to application */
	dsp_hdr.dsp_pkt_type 	= 0x00;
	dsp_hdr.dsp_result   	= DSP_LINK_RESET;
	dsp_hdr.dsp_reason_code = DSP_RC_RESET;
	dsp_snd_api_oob_msg(dsp_link, &dsp_hdr, NULL, 0);

	/* Sending M_CR message to remote side */
	dsp_link->reason_code = DSP_RC_RESET;
	dsp_api_circuit_reset(dsp_link, NULL, DSP_FALSE);
}

void dsp_x25_bh_kick(void *dsp_id)
{
	struct net_device*	dsp_dev = (struct net_device *)dsp_id;
	dsp_link_t*		dsp_link = NULL;
	int			err = 0;
	
	DSP_ASSERT1(dsp_dev == NULL);
	DSP_ASSERT1(dsp_dev->priv == NULL);
	dsp_link = (dsp_link_t*)dsp_dev->priv;
	DEBUG3((KERN_INFO "%s: X25 kicking DSP TX\n", dsp_link->name));

	if (!test_and_clear_bit(BH_AWAITING_KICK, &dsp_link->tq_working))
		return;
	
	if((err = dsp_trigger_link_bh(dsp_link)) != 0){
		set_bit(BH_AWAITING_KICK, &dsp_link->tq_working);
	}
	
	DEBUG7((KERN_INFO "%s: Triggering BH TX (x25 kick)...%s\n",
			dsp_link->name, (!err)?"Done":"Failed"));
}


int dsp_oob_msg(void* dsp_id, struct sk_buff* skb) 
{
	struct net_device* 	dsp_dev = (struct net_device*)dsp_id;
	dsp_link_t*		dsp_link = NULL;
	x25api_hdr_t*		x25api_hdr = NULL;

	DSP_ASSERT(dsp_dev == NULL);
	DSP_ASSERT(dsp_dev->priv == NULL);
	DSP_ASSERT(skb == NULL);
	dsp_link = (dsp_link_t *)dsp_dev->priv;
	x25api_hdr = (x25api_hdr_t*)skb->data;
			
	switch(x25api_hdr->pktType){
	case RESTART_REQUEST_PKT:
	case RESTART_CONF_PKT:
	case CLEAR_REQUEST_PKT:
	case CLEAR_CONF_PKT:
		dsp_svc_down(dsp_id);
		break;
	case RESET_REQUEST_PKT:	
	case RESET_CONF_PKT:
	case INTERRUPT_PKT:
	case INTERRUPT_CONF_PKT:
		dsp_link->reason_code = DSP_RC_RESET;
		dsp_api_circuit_reset(dsp_link, NULL, DSP_FALSE);
		break;
	}

	KFREE_SKB(skb,dsp_link);
	return 0;
}

void dsp_get_active_inactive(void *dsp_id, wp_stack_stats_t *wp_stats)
{
	struct net_device* 	dsp_dev = (struct net_device*)dsp_id;
	dsp_link_t*		dsp_link = NULL;

	DSP_ASSERT1(dsp_dev == NULL);
	DSP_ASSERT1(dsp_dev->priv == NULL);
	dsp_link = (dsp_link_t *)dsp_dev->priv;

	if (dsp_link->state == DSP_CONNECTED){
		wp_stats->dsp_active++;
	}else{
		wp_stats->dsp_inactive++;
	}

}

int dsp_check_dev(void* dsp_id) 
{
	struct net_device* 	dsp_dev = (struct net_device*)dsp_id;
	dsp_link_t*		dsp_link = NULL;

	DSP_ASSERT(dsp_dev == NULL);
	DSP_ASSERT(dsp_dev->priv == NULL);
	dsp_link = (dsp_link_t *)dsp_dev->priv;

	if (dsp_link->usedby == API && 
	    !dsp_link->sk_id && 
	    dsp_link->state == DSP_READY){
		return 0;
	}
	return -EINVAL;
}

int dsp_rx_buf_check(void* dsp_id, int len)
{
	struct net_device*	dsp_dev = (struct net_device *)dsp_id;
	dsp_link_t*		dsp_link = NULL;
	int			err = 0;
	
	DSP_ASSERT(dsp_dev == NULL);
	DSP_ASSERT(dsp_dev->priv == NULL);

	dsp_link = (dsp_link_t*)dsp_dev->priv;
	DEBUG3((KERN_INFO "%s: X25 Rx buf checking\n", dsp_link->name));

	if (!dsp_link->sk_id)
		return -ENODEV;
	
	err = wanpipe_api_buf_check(dsp_link->sk_id, len);
	if (err == -EINVAL){
		unsigned long flags;
		printk(KERN_INFO "%s:Major Error in Socket Above: CHK BUF!!!\n",
					dsp_link->name);
		
		spin_lock_irqsave(&dsp_link->list_lock,flags);
	
		if (dsp_link->sk_id){
			__sock_put(dsp_link->sk_id);
		}
		dsp_link->sk_id = NULL;

		spin_unlock_irqrestore(&dsp_link->list_lock,flags);

	}
	return err;
}

static int __init wanpipe_dsp_init(void)
{
	int err = 0;
	struct x25_dsp_register_struct dsp_reg;
	struct wanpipe_dsp_register_struct dsp_api_reg;

	dsp_link_list=NULL;

	printk(KERN_INFO "WANPIPE DSP for Linux. Version %s (c) 2003 Sangoma Technologies Inc.\n",WANPIPE_VERSION);
	spin_lock_init(&dsp_link_lock);

	atomic_set(&dsp_mem_in_used, 0);
#ifdef CONFIG_PROC_FS
	if (dsp_proc_init() != DSP_OK)
		return -ENOMEM;
#endif

	dsp_reg.dsp_register 			= dsp_register;
	dsp_reg.dsp_unregister 			= dsp_unregister;
	dsp_reg.dsp_setparms 			= dsp_setparms;

	dsp_reg.dsp_call_request		= dsp_call_request;
	dsp_reg.dsp_call_accept			= dsp_call_accept;
	dsp_reg.dsp_svc_up			= dsp_svc_up;
	dsp_reg.dsp_svc_down			= dsp_svc_down;
	dsp_reg.dsp_svc_ready			= dsp_svc_ready;
	dsp_reg.dsp_svc_reset			= dsp_svc_reset;
	
	dsp_reg.dsp_receive_pkt 		= dsp_receive_pkt;
	dsp_reg.dsp_x25_bh_kick 		= dsp_x25_bh_kick;

	dsp_reg.dsp_rx_buf_check 		= dsp_rx_buf_check;
	dsp_reg.dsp_oob_msg 			= dsp_oob_msg;
	dsp_reg.dsp_check_dev 			= dsp_check_dev;
	dsp_reg.dsp_get_active_inactive		= dsp_get_active_inactive;

#if 0	
	/* FIXME: Initialize call back functions */
	
	dsp_reg.dsp_x25_connect_indication 	= dsp_x25_connect_indication;
	dsp_reg.dsp_x25_connect_confirm 	= dsp_x25_connect_confirm; 

	dsp_reg.dsp_x25_disconnect_confirm 	= dsp_x25_disconnect_confirm;
	dsp_reg.dsp_x25_disconnect_indication 	= dsp_x25_disconnect_indication;
#endif	
	err = register_x25_dsp_protocol(&dsp_reg);
	if (err)
		return err;
	
	// FIXME: Initialize dsp_api_reg structure.
	dsp_api_reg.dsp_bind_api = dsp_bind_api;
	return register_wanpipe_dsp_protocol(&dsp_api_reg);
}

MODULE_AUTHOR("Alex Feldman <al.feldman@sangoma.com>");
MODULE_DESCRIPTION("The Sangoma Display System Protocol");
MODULE_LICENSE("GPL");

void __exit wanpipe_dsp_exit (void)
{
	printk(KERN_INFO "WANPIPE DSP: Unloading module\n");
	if (atomic_read(&dsp_mem_in_used) != 0)
		printk(KERN_INFO "DSP: Warning: %d bytes are still in used by DSP module\n", 
			atomic_read(&dsp_mem_in_used));
#if CONFIG_PROC_FS
	 dsp_proc_cleanup();
#endif
	unregister_x25_dsp_protocol();
	unregister_wanpipe_dsp_protocol();
}



module_init(wanpipe_dsp_init);
module_exit(wanpipe_dsp_exit);

