
/*
 * ******************************************************************
 * 			INCLUDE FILES				    *
 * ******************************************************************
 */
#include <linux/wanpipe_includes.h>
#include <linux/wanpipe_cfg.h>
#include "wanpipe_dsp.h"

#ifdef DSP_DEBUG
extern unsigned short dsp_debug;
#endif
/*
 * ******************************************************************
 * 			FUNCTION DEFINITIONS			    *
 * ******************************************************************
 */
/*
 * ******************************************************************
 * 			FUNCTION PROTOTYPES			    *
 * ******************************************************************
 */
static int dsp_decode_pkt_type(dsp_link_t*, struct sk_buff*);
static int dsp_state_connected(dsp_link_t*, struct sk_buff*);
static int dsp_state_connecting_acc_call(dsp_link_t*, struct sk_buff*);
static int dsp_state_connected_circuit_reset(dsp_link_t*, struct sk_buff*);
static int dsp_state_discon_clear_req(dsp_link_t*, struct sk_buff*);


/*
 * ******************************************************************
 * 			FUNCTION DEFINITIONS			    *
 * ******************************************************************
 */
static int dsp_decode_pkt_type(dsp_link_t* dsp_link, struct sk_buff* skb)
{
	x25api_hdr_t	x25api_hdr;
	dspapi_hdr_t*	dspapi_hdr = NULL;
	char* 		packet = NULL;

	DSP_ASSERT(dsp_link == NULL);
	DSP_ASSERT(skb == NULL);

	if (skb->len <= sizeof(x25api_hdr_t)){
		printk(KERN_INFO "%s: RX packet is too small (%d)!\n",
				dsp_link->name, skb->len);
		return -EINVAL;
	}
	memcpy(&x25api_hdr, skb->data, sizeof(x25api_hdr_t));

	/* Remove X25 API header */
	skb_pull(skb, sizeof(x25api_hdr_t));

	/* */
	if (skb_headroom(skb) < sizeof(dspapi_hdr_t)){
		DSP_SKB_HEAD(skb);
		return -EINVAL;
	}
	/* If Q bit set the received message is DSP control message */
	if (IS_L_Q_BIT(x25api_hdr.qdm)) {
		/* Add DSP API header */
		dspapi_hdr = (dspapi_hdr_t*)skb_push(skb, sizeof(dspapi_hdr_t));
		IS_NULL(dspapi_hdr, -ENOMEM);
		memset(dspapi_hdr, 0, sizeof(dspapi_hdr_t));
		dspapi_hdr->dsp_length	= x25api_hdr.length;
	
		packet = &skb->data[sizeof(dspapi_hdr_t)];
		dspapi_hdr->dsp_pkt_type = packet[1];
		switch(dspapi_hdr->dsp_pkt_type){
			case DSP_INV_TO_CLEAR_MSG:
				dspapi_hdr->dsp_reason_code = packet[2];
				DEBUG2((KERN_INFO "%s: Rx M_ITC (reason_code=%i).\n",
					dsp_link->name, dspapi_hdr->dsp_reason_code));
				DSP_INC_RX_STAT(rx_inv_to_clear);
				break;

			case DSP_CMD_RES_UNDELIVERED_MSG:
				dspapi_hdr->dsp_sn = packet[2];
				dspapi_hdr->dsp_error_code = packet[3];
				if (dspapi_hdr->dsp_length == DSP_RES_UNDELIVERED_MSG_LEN){
					dspapi_hdr->dsp_status = packet[4];
					dspapi_hdr->dsp_sense = packet[5];
					dspapi_hdr->dsp_resp = DSP_TRUE;
				}
				DEBUG2((KERN_INFO "%s: Rx CMD/RES Undelivered (sn=%i,err=%i).\n",
					dsp_link->name, dspapi_hdr->dsp_sn, 
					dspapi_hdr->dsp_error_code));
				DSP_INC_RX_STAT(rx_cmd_res_undelivered);
				break;

			case DSP_CMD_RES_ABORTED_MSG:
				dspapi_hdr->dsp_sn = packet[2];
				dspapi_hdr->dsp_error_code = packet[3];
				if (dspapi_hdr->dsp_length == DSP_RES_ABORTED_MSG_LEN){
					dspapi_hdr->dsp_status = packet[4];
					dspapi_hdr->dsp_sense = packet[5];
					dspapi_hdr->dsp_resp = DSP_TRUE;
				}
				DEBUG2((KERN_INFO "%s: Rx CMD/RES Aborted (sn=%i,err=%i).\n",
					dsp_link->name, dspapi_hdr->dsp_sn, 
					dspapi_hdr->dsp_error_code));
				DSP_INC_RX_STAT(rx_cmd_res_aborted);
				break;

			case DSP_STATUS_MSG:
				dspapi_hdr->dsp_status = packet[2];
				dspapi_hdr->dsp_sense = packet[3];
				DEBUG2((KERN_INFO "%s: Rx Status message (status=%i, sense=%i).\n",
					dsp_link->name, 
					dspapi_hdr->dsp_status, 
					dspapi_hdr->dsp_sense));
				DSP_INC_RX_STAT(rx_status);
				break;

			case DSP_ACK_MSG:
				dspapi_hdr->dsp_sn = packet[2];
				DEBUG2((KERN_INFO "%s: Rx M_EEACK (sn=%i).\n",
					dsp_link->name, dspapi_hdr->dsp_sn));
				DSP_INC_RX_STAT(rx_ack);
				break;

			case DSP_CIRCUIT_ENABLED_MSG:
				dspapi_hdr->dsp_src_cu_addr  = packet[2];
				dspapi_hdr->dsp_src_dev_addr = packet[3];
				dspapi_hdr->dsp_ctrl_byte    = packet[4]; 
				dspapi_hdr->dsp_dev_info     = packet[5];
				dspapi_hdr->dsp_crm_type     = packet[6];
				dspapi_hdr->dsp_dst_cu_addr  = packet[8];
				dspapi_hdr->dsp_dst_dev_addr = packet[9];
				DEBUG2((KERN_INFO "%s: Rx M_CE.\n",
					dsp_link->name));
				DSP_INC_RX_STAT(rx_ce);
				break;

			case DSP_CIRCUIT_RESET_MSG:
				if (dsp_link->state == DSP_CONNECTED){
					/*
					 * packet[2] - POS Device sequence number 
					 * packet[3] - HOST sequence number 
					 */
					if (dsp_link->cfg.pad_type == DSP_HPAD){
						dsp_link->tx_sn = packet[3];
						dsp_link->rx_sn = packet[2];
					}else{
						dsp_link->tx_sn = packet[2];
						dsp_link->rx_sn = packet[3];
					}
				}
				DEBUG2((KERN_INFO "%s: Rx M_CR (tx_sn=%i,rx_sn=%i).\n",
					dsp_link->name, 
					dsp_link->tx_sn, 
					dsp_link->rx_sn));
				DSP_INC_RX_STAT(rx_cr);
				break;

			case DSP_CIRCUIT_REQUEST_MSG:
				dspapi_hdr->dsp_src_cu_addr  = packet[2];
				dspapi_hdr->dsp_src_dev_addr = packet[3];
				dspapi_hdr->dsp_ctrl_byte    = packet[4]; 
				dspapi_hdr->dsp_dev_info     = packet[5];
				dspapi_hdr->dsp_crm_type     = packet[6];
				dspapi_hdr->dsp_dst_cu_addr  = packet[7];
				dspapi_hdr->dsp_dst_dev_addr = packet[8];
				DEBUG2((KERN_INFO "%s: Rx Circuit Request.\n",
					dsp_link->name));
				DSP_INC_RX_STAT(rx_creq);
				break;

			case DSP_CIRCUIT_DISCONNECT_MSG:
				dspapi_hdr->dsp_reason_code = packet[2];
				DEBUG2((KERN_INFO "%s: Rx M_CD (reason_code=%i).\n",
					dsp_link->name, dspapi_hdr->dsp_reason_code));
				DSP_INC_RX_STAT(rx_cd);
				break;
			
			default:
				printk(KERN_INFO "DSP: Unknown message type (%X)\n",
						dspapi_hdr->dsp_pkt_type);
				return -EINVAL;
		}
		/* We not need packet, all information is in DSP API header */
		skb_trim(skb, sizeof(dspapi_hdr_t));
		dspapi_hdr->dsp_length = 0;

	}else{
		unsigned char	rx_sn = 0x00;
		unsigned char	is_ack = DSP_FALSE,
#if 0
		ALEX 10.4.2001
		ETX_ETB_ESC
				et_byte = DSP_FALSE,
#endif
				is_transp = DSP_FALSE;
		unsigned short 	length = 0;

		is_ack 		= IS_R_ACK(skb->data[1]);
		is_transp 	= IS_R_XPR(skb->data[1]);
		rx_sn 		= skb->data[2];
#if 0
		ALEX 10.4.2001 
		ETX_ETB_ESC
		et_byte 	= GET_ET_BYTE(skb->data[skb->len-1]); 
#endif

		/* Remove DSP header and tailer */
		skb_pull(skb, DSP_HDR_LEN);
#if 0
		ALEX 10.4.2001 
		ETX_ETB_ESC
		/*****************************************/
		/* This two bytes are part of data frame */
		/* Do not remove them from frame!!!! 	 */
		/*****************************************/
		skb_pull(skb, DSP_RX_HDR_LEN(dsp_link));
		skb_trim(skb, skb->len-1);
#endif
		
		DEBUG2((KERN_INFO "%s: Rx M_DATA(len=%d,sn=%d,ack=%d)\n",
			dsp_link->name, skb->len, rx_sn, 
			is_ack));

		/* M_DATA length */
		length  = skb->len;
		/* Add DSP API header */
		dspapi_hdr = (dspapi_hdr_t*)skb_push(skb, sizeof(dspapi_hdr_t));
		IS_NULL(dspapi_hdr, -ENOMEM);
		dspapi_hdr->dsp_pkt_type = (is_ack == DSP_TRUE) ? 
						DSP_DATA_ACK_MSG : DSP_DATA_MSG;
		dspapi_hdr->dsp_length 		= length;
		dspapi_hdr->dsp_sn 		= rx_sn;
		dspapi_hdr->dsp_data_transp	= is_transp;
#if 0
		dspapi_hdr->dsp_et_byte		= et_byte;
#endif
		packet = &skb->data[sizeof(dspapi_hdr_t)];

		/* 
		 * If M_DATA requires ACK, set it now.
		 */
		dsp_link->array_sn[dsp_link->rx_sn] |=
				(is_ack == DSP_TRUE) ? 
					DSP_SN_WAIT_ACK_L : DSP_SN_NOT_USED;
		/* Check sequence number */
		if (dsp_link->rx_sn != rx_sn){
			/* Received packet out of sequence */
			printk(KERN_INFO "%s: RX data out of sequence (sn=%i,wait_sn=%d)\n",
				dsp_link->name, rx_sn, dsp_link->rx_sn);
			/* FIXME: Dopping packet.... */
			/* Now, send CIRCUIT RESET */
			dsp_link->reason_code = DSP_RC_DATA_ERROR;
			dsp_set_state(dsp_link, DSP_CONNECTED_CR);
			dsp_api_circuit_reset(dsp_link, NULL, DSP_FALSE);
			return -EINVAL;
		}
		/* Increment sequnce number */
		dsp_link->rx_sn = dsp_get_sn(dsp_link->rx_sn+1);

		if (dspapi_hdr->dsp_pkt_type == DSP_DATA_ACK_MSG){
			if (dsp_link->cfg.auto_ack){
				/* Automatically send acknoledment 
				 * for current packet 
				 */
				dsp_api_eeack(dsp_link, NULL);
			}else{
				/* Start T3 timer */
				dsp_start_t3_timer(dsp_link);
			}
		}
		if (dspapi_hdr->dsp_pkt_type == DSP_DATA_ACK_MSG){
			DSP_INC_RX_STAT(rx_ack_data_msg);
		}else{
			DSP_INC_RX_STAT(rx_data_msg);
		}
	}
	
	return DSP_OK;
}

static int dsp_state_connected(dsp_link_t* dsp_link, 
			       struct sk_buff* skb)
{
	dspapi_hdr_t* 	dspapi_hdr = NULL;
	unsigned char 	next_sn = 0x00;
	
	DSP_ASSERT(dsp_link == NULL);
	DSP_ASSERT(skb == NULL);

	dspapi_hdr = (dspapi_hdr_t*)skb->data;
	switch(dspapi_hdr->dsp_pkt_type){
		case DSP_INV_TO_CLEAR_MSG:
			// Send Clear Request 
			dsp_api_clear_call(dsp_link, NULL);
			dsp_set_state(dsp_link, DSP_DISCONNECTING_CL_REQ);
			break;
			
		case DSP_CIRCUIT_RESET_MSG:
			dsp_link->reason_code = DSP_RC_RESET;
			dsp_api_circuit_reset(dsp_link, NULL, DSP_TRUE);
			//return DSP_TRUE;
			break;

		case DSP_ACK_MSG:
			if (dsp_chk_ack_sn(dsp_link, dspapi_hdr->dsp_sn, &next_sn, DSP_FALSE) == DSP_FALSE){
				/* RX M_EEACK from remote PAD out of 
				 * sequence.
				 * Set 'dsp_result' to DSP_RC_DATA_ERROR
				 * and send up to application.
				 */
				printk(KERN_INFO "%s: RX M_EEACK out of sequence (rx_sn=%i,sn=%i)\n",
					dsp_link->name, dspapi_hdr->dsp_sn, next_sn);
				dspapi_hdr->dsp_result = DSP_RC_DATA_ERROR;
			}
			break;

		case DSP_CMD_RES_UNDELIVERED_MSG:
			/* FIXME: Do I need to something? */
			break;
			
		case DSP_CMD_RES_ABORTED_MSG:
			/* FIXME: Do I need to something? */
			break;

		case DSP_DATA_MSG:
		case DSP_DATA_ACK_MSG:
			dsp_data_tx_up(dsp_link, skb);
			return DSP_OK;

		case DSP_STATUS_MSG:
			break;

		default:
			printk(KERN_INFO "%s: Dropping packet (packet type - %02X, state - %s%s)\n",
					dsp_link->name,
					dspapi_hdr->dsp_pkt_type,
					dsp_get_mj_state(dsp_link->state),
					dsp_get_mn_state(dsp_link->state));
			return -EINVAL;
	}
	dsp_snd_api_oob_msg(dsp_link, dspapi_hdr, NULL, 0);
	if (skb){
		KFREE_SKB(skb, dsp_link);
	}
	return DSP_OK;
}

static int dsp_state_connected_circuit_reset(dsp_link_t* dsp_link, 
				             struct sk_buff* skb)
{
	dspapi_hdr_t* 	dspapi_hdr = NULL;

	DSP_ASSERT(dsp_link == NULL);
	DSP_ASSERT(skb == NULL);

	dspapi_hdr = (dspapi_hdr_t*)skb->data;
	switch(dspapi_hdr->dsp_pkt_type){
		case DSP_INV_TO_CLEAR_MSG:
			/* Stop T2 timer */
			dsp_stop_t2_timer(dsp_link);
			dsp_api_clear_call(dsp_link, NULL);
			dsp_set_state(dsp_link, DSP_DISCONNECTING_CL_REQ);
			break;
			
		case DSP_CIRCUIT_RESET_MSG:
			dsp_set_state(dsp_link, DSP_CONNECTED);
			/* Stop T2 timer */
			dsp_stop_t2_timer(dsp_link);
			break;

		default:
			printk(KERN_INFO "%s: Dropping packet (packet type - %02X, state - %s%s)\n",
					dsp_link->name,
					dspapi_hdr->dsp_pkt_type,
					dsp_get_mj_state(dsp_link->state),
					dsp_get_mn_state(dsp_link->state));
			return -EINVAL;
	}
	dsp_snd_api_oob_msg(dsp_link, dspapi_hdr, NULL, 0);
	if (skb){
		KFREE_SKB(skb, dsp_link);
	}
	return DSP_OK;
}


static int dsp_state_connecting_acc_call(dsp_link_t* dsp_link, 
			 	  	 struct sk_buff* skb)
{
	dspapi_hdr_t* 	dspapi_hdr = NULL;

	DSP_ASSERT(dsp_link == NULL);
	DSP_ASSERT(skb == NULL);

	dspapi_hdr = (dspapi_hdr_t*)skb->data;
	switch(dspapi_hdr->dsp_pkt_type){
		case DSP_CIRCUIT_ENABLED_MSG:
			// FIXME: Read data from M_CE packet
			dsp_set_state(dsp_link, DSP_CONNECTED);
			break;
			
		default:
			printk(KERN_INFO "%s: Dropping packet (packet type - %02X, state - %s%s)\n",
					dsp_link->name,
					dspapi_hdr->dsp_pkt_type,
					dsp_get_mj_state(dsp_link->state),
					dsp_get_mn_state(dsp_link->state));
			return -EINVAL;
	}
	dsp_snd_api_oob_msg(dsp_link, dspapi_hdr, NULL, 0);
	if (skb){
		KFREE_SKB(skb, dsp_link);
	}
	/* Do not free sk buffer */
	return DSP_OK;
}

static int dsp_state_discon_clear_req(dsp_link_t* dsp_link, 
				      struct sk_buff* skb)
{
	dspapi_hdr_t* 	dspapi_hdr = NULL;

	DSP_ASSERT(dsp_link == NULL);
	DSP_ASSERT(skb == NULL);

	dspapi_hdr = (dspapi_hdr_t*)skb->data;
	switch(dspapi_hdr->dsp_pkt_type){
		case DSP_CMD_RES_UNDELIVERED_MSG:
			/* FIXME: Do I need to something? */
			break;
			
		case DSP_CMD_RES_ABORTED_MSG:
			/* FIXME: Do I need to something? */
			break;

		default:
			printk(KERN_INFO "%s: Dropping packet (packet type - %02X, state - %s%s)\n",
					dsp_link->name,
					dspapi_hdr->dsp_pkt_type,
					dsp_get_mj_state(dsp_link->state),
					dsp_get_mn_state(dsp_link->state));
			return -EINVAL;
	}
	dsp_snd_api_oob_msg(dsp_link, dspapi_hdr, NULL, 0);
	if (skb){
		KFREE_SKB(skb, dsp_link);
	}
	return DSP_OK;
}

/*
 *	Process an incoming DSP frame
 */
void dsp_data_input(dsp_link_t *dsp_link, struct sk_buff *skb)
{
	int 	err = DSP_TRUE;
	
	DSP_ASSERT1(dsp_link == NULL);
	DSP_ASSERT1(skb == NULL);

	if ((err = dsp_decode_pkt_type(dsp_link, skb)) != DSP_OK){
		/* Free sk buffer */
		err = -EINVAL;
		goto end_data_input;
	}
	
	DEBUG2((KERN_INFO "%s: RX State %i!\n",
				dsp_link->name, dsp_link->state));
	switch(dsp_link->state){
			
		case DSP_CONNECTING_ACC_CALL:
			err = dsp_state_connecting_acc_call(dsp_link, skb);
			break;
			
		case DSP_CONNECTED:
			err = dsp_state_connected(dsp_link, skb);
			break;
			
		case DSP_CONNECTED_CR:
			err = dsp_state_connected_circuit_reset(dsp_link, skb);
			break;
			
		case DSP_DISCONNECTING_CL_REQ:
			err = dsp_state_discon_clear_req(dsp_link, skb);
			break;
			
		case DSP_DISCONNECTED:
		case DSP_CONNECTING:
		case DSP_DISCONNECTING:
			printk(KERN_INFO "%s: Dropping packet (incorrect state)\n",
					dsp_link->name);
			break;
			
		default:
			printk(KERN_INFO "%s: Dropping packet (unknown state)\n",
					dsp_link->name);
			break;
	}
end_data_input:

	if (err){
		KFREE_SKB(skb,dsp_link);
		DSP_INC_TX_STAT(tx_dropped);
	}

	return;
}

