
#include <linux/wanpipe_includes.h>
#include <linux/wanpipe_cfg.h>
#include "wanpipe_dsp.h"

#ifdef DSP_DEBUG
extern unsigned short dsp_debug;
#endif

/* 
 *  This procedure is passed a buffer descriptor for an iframe. It builds
 *  the rest of the control part of the frame and then writes it out.
 */
int dsp_send_data_frame(dsp_link_t *dsp_link, struct sk_buff *skb)
{
	x25api_hdr_t*		x25api_hdr = NULL;
	unsigned char*		data = NULL;
	unsigned char		is_ack = DSP_FALSE, is_m_bit = DSP_FALSE;
	int 			err=0;

	DSP_ASSERT(skb == NULL);
	DSP_ASSERT(dsp_link == NULL);
	
	x25api_hdr = (x25api_hdr_t*)skb->data;
	data = &skb->data[sizeof(x25api_hdr_t)];
	if (dsp_link->is_tx_more_data == DSP_FALSE){
		/* This is the first packet of M_DATA message */
		is_ack = IS_R_ACK(data[1]);
		data[2] = dsp_link->tx_sn;
	}
	is_m_bit = IS_L_M_BIT(x25api_hdr->qdm);

	err = dsp_data_tx_down(dsp_link, skb);
	if (err == DSP_OK){
		/*
		 * The ACK bit is set to 1 only for the first packet 
		 */
		if (dsp_link->is_tx_more_data == DSP_FALSE){
			if (is_ack == DSP_TRUE)
				dsp_link->array_sn[dsp_link->tx_sn] |= 
							DSP_SN_WAIT_ACK_R;	
		}
		/* 
		 * If m-bit set to 1, set 'is_tx_more_data' to 
		 * DSP_TRUE,otherwise DSP_FALSE. 
		 * Increment sequence number only if M_DATA was 
		 * successful sent (m-bit is 0).
		 */
		if (is_m_bit == DSP_TRUE){
			dsp_link->is_tx_more_data = DSP_TRUE;
		}else{
			dsp_link->is_tx_more_data = DSP_FALSE;
			/* Increment sequence number */
			dsp_link->tx_sn = dsp_get_sn(dsp_link->tx_sn+1);
		}
	}
	return err;
}

void dsp_link_send (unsigned long dsp_link_data)
{
#define dsp_link ((dsp_link_t *)dsp_link_data)
	struct sk_buff*		skb = NULL;
	int 			skb_len = 0;
	int			err = 0;

	DSP_ASSERT1(dsp_link == NULL);
	DSP_ASSERT1(dsp_link->dev == NULL);

	spin_lock(&dsp_link->list_lock);
	
	if (test_and_set_bit(BH_SMP,&dsp_link->critical)){
		DEBUG3((KERN_INFO "%s: CRITICAL in DSP BH Send!!\n",
				dsp_link->name));
		
		goto dsp_link_send_exit;
	}

	clear_bit(MORE_TX, &dsp_link->tq_working);
	clear_bit(DSP_KICKED, &dsp_link->tq_working);

	if (test_bit(CRIT_LINK_DOWN,&dsp_link->critical)){
		clear_bit(BH_SMP,&dsp_link->critical);
		clear_bit(BH_RUNNING, &dsp_link->tq_working);
		goto dsp_link_send_exit;
	}

	dsp_link->tx_kick_cnt++;

	while((skb = skb_dequeue(&dsp_link->async_queue)) != NULL){
		DEBUG3((KERN_INFO "%s: Sending Async data....!\n",
				dsp_link->name));
		skb_len = skb->truesize;
		if ((err = dsp_data_tx_down(dsp_link, skb)) != DSP_OK){
			if (err == -EBUSY){
				DEBUG3((KERN_INFO "%s: Master device is busy.\n",
						dsp_link->name));
				skb_queue_head(&dsp_link->async_queue, skb);
				set_bit(BH_AWAITING_KICK, &dsp_link->tq_working);
				goto link_bh_kick_exit;
			}else{
				KFREE_SKB(skb, dsp_link);
				DSP_INC_TX_STAT(tx_dropped);
			}
		}
	}
		
	while((skb = skb_dequeue(&dsp_link->write_queue)) != NULL){
		DEBUG3((KERN_INFO "%s: Sending packet....!\n",
				dsp_link->name));
		skb_len = skb->truesize;
		if ((err = dsp_send_data_frame(dsp_link, skb)) != DSP_OK) {
			if (err == -EBUSY){
				DEBUG3((KERN_INFO "%s: Master device %s is busy.\n",
						dsp_link->name,dsp_link->master->name));
				skb_queue_head(&dsp_link->write_queue, skb);
				set_bit(BH_AWAITING_KICK, &dsp_link->tq_working);
				goto link_bh_kick_exit;
			}else{
				KFREE_SKB(skb,dsp_link);
				DSP_INC_TX_STAT(tx_dropped);
			}
		}
	}

link_bh_kick_exit:
	if (skb_queue_len(&dsp_link->async_queue) ||
	    skb_queue_len(&dsp_link->write_queue)){
		set_bit(MORE_TX, &dsp_link->tq_working);	
	} 
	
	clear_bit(BH_RUNNING, &dsp_link->tq_working);

	if ((skb_queue_len(&dsp_link->write_queue) == 0)){
		/*
	 	 * If socket waiting for sending packet, kick 
		 * the socket
		 */
		DEBUG3((KERN_INFO "%s: DSP Kick the socket \n", 
				dsp_link->name));

		netif_start_queue(dsp_link->dev);
		if (dsp_link->sk_id){
			wanpipe_api_poll_wake(dsp_link->sk_id);
		}
	}

	/* If the lower layer is busy, then forget about 
	 * currently pending packets and let the lower
	 * layer wake us up when its capable of tx */
	if (test_bit(BH_AWAITING_KICK, &dsp_link->tq_working)){
	
		DEBUG3((KERN_INFO "%s: DSP BH: Failed to TX lower layer busy!\n", 
				dsp_link->name));
		clear_bit(MORE_TX,&dsp_link->tq_working);
	
	}else if (test_bit(MORE_TX, &dsp_link->tq_working)){

		DEBUG3((KERN_INFO "%s: MORE_TX flag is set, trigger link bh\n", 
				dsp_link->name));
		dsp_trigger_link_bh(dsp_link);
	}

dsp_link_send_exit:
	
	clear_bit(BH_SMP,&dsp_link->critical);
	spin_unlock(&dsp_link->list_lock);
	return;
#undef dsp_link 
}


int dsp_send_control(dsp_link_t* dsp_link, unsigned char dsp_type, 
		     char* data_ptr, int data_len)
{
	struct sk_buff*	skb = NULL;
	x25api_hdr_t*	x25api_hdr = NULL;
	unsigned char*	dptr = NULL;
	
	DSP_ASSERT(dsp_link == NULL);
	DSP_ASSERT(dsp_link->dev == NULL);
	
	if (data_ptr == NULL){
		printk(KERN_INFO "%s: Invalid control message (%02X) data\n",
				dsp_link->name, dsp_type);
		goto dsp_send_control_error;
	}
	
	if (skb_queue_len(&dsp_link->async_queue) >= DSP_ASYNC_QUEUE_SIZE){
		DEBUG3((KERN_INFO "%s: Async queue is full...\n",
					dsp_link->name));
		return -EBUSY;	
	}

	/* Allocate skb_buffer with x25 API header */
	ALLOC_SKB(skb,data_len+sizeof(x25api_hdr_t),dsp_link);
	if (skb == NULL){
		printk(KERN_INFO "DSP: Critical Error: No Memory!\n");
		return -ENOMEM;
	}
	
	switch (dsp_type){
		case DSP_CIRCUIT_ENABLED_MSG:
		case DSP_CIRCUIT_RESET_MSG:
		case DSP_CIRCUIT_REQUEST_MSG:
		case DSP_CIRCUIT_DISCONNECT_MSG:
		case DSP_INV_TO_CLEAR_MSG:
		case DSP_ACK_MSG:
		case DSP_STATUS_MSG:
		case DSP_CMD_RES_UNDELIVERED_MSG:
		case DSP_CMD_RES_ABORTED_MSG:
			goto dsp_send_control_send_data;
			break;

		default:
			printk(KERN_INFO "%s: Unknown Control Message (%02X).\n",
					dsp_link->name, dsp_type);
			goto dsp_send_control_error;
	}

dsp_send_control_send_data:
	
	x25api_hdr = (x25api_hdr_t*)skb_put(skb, sizeof(x25api_hdr_t));
	IS_NULL(x25api_hdr, -EINVAL);
	x25api_hdr->qdm = SET_L_Q_BIT(DSP_TRUE);
	x25api_hdr->length = skb->len;

	if (data_len){
		if (skb_tailroom(skb) < data_len){
			DSP_SKB_TAIL(skb);
			goto dsp_send_control_error;
		}
		dptr = skb_put(skb, data_len);
		if (!dptr)
			goto dsp_send_control_error;
		memcpy(dptr, data_ptr, data_len);
	}

	/* Add new control message to the ASYNC queue */
	skb_queue_tail (&dsp_link->async_queue, skb);

	/* Trigger BH */
	dsp_trigger_link_bh(dsp_link);
	return DSP_OK;
	
dsp_send_control_error:
	DSP_INC_TX_STAT(tx_dropped);
	if (skb != NULL){
		KFREE_SKB(skb,dsp_link);
	}
	return -EINVAL;

}

