#include <linux/wanpipe_includes.h>
#include <linux/proc_fs.h>
#include <linux/wanpipe_kernel.h>
#include <linux/wanpipe_cfg.h>
#include <linux/wanproc.h>
#include "wanpipe_dsp.h"

#define PROC_BUFSZ 4000   

#ifdef CONFIG_PROC_FS


#ifdef LINUX_2_6

static int config_get_info(struct seq_file *m, void *v);
static int status_get_info(struct seq_file *m, void *v);

static int dsp_link_config_get_info(struct seq_file *m, void *v);
static int dsp_link_status_get_info(struct seq_file *m, void *v);
static int dsp_link_state_get_info(struct seq_file *m, void *v);

#else
/* Proc filesystem interface */
static int dsp_proc_perms(struct inode *, int);
static ssize_t dsp_proc_read(struct file* file, char* buf, size_t count, loff_t *ppos);

/* Methods for preparing data for reading proc entries */

static int config_get_info(char* buf, char** start, off_t offs, int len);
static int status_get_info(char* buf, char** start, off_t offs, int len);

static int dsp_link_config_get_info(char*, char**, off_t, int);
static int dsp_link_status_get_info(char*, char**, off_t, int);
static int dsp_link_state_get_info(char*, char**, off_t, int);
#endif

atomic_t dsp_mem_in_used;

/* Miscellaneous */

/*
 *	Structures for interfacing with the /proc filesystem.
 *	Router creates its own directory /proc/net/router with the folowing
 *	entries:
 *	config		device configuration
 *	status		global device statistics
 *	<device>	entry for each WAN device
 */

/*
 *	Generic /proc/net/router/<file> file and inode operations 
 */

#ifdef LINUX_2_6

static int config_get_info_open(struct inode *inode, struct file *file)
{
	return single_open(file, config_get_info, PDE(inode)->data);
}

static struct file_operations config_get_info_fops = {
	.owner	 = THIS_MODULE,
	.open	 = config_get_info_open,
	.read	 = seq_read,
	.llseek	 = seq_lseek,
	.release = seq_release,
};

static int status_get_info_open(struct inode *inode, struct file *file)
{
	return single_open(file, status_get_info, PDE(inode)->data);
}

static struct file_operations status_get_info_fops = {
	.owner	 = THIS_MODULE,
	.open	 = status_get_info_open,
	.read	 = seq_read,
	.llseek	 = seq_lseek,
	.release = seq_release,
};


static int dsp_link_config_get_info_open(struct inode *inode, struct file *file)
{
	return single_open(file, dsp_link_config_get_info, PDE(inode)->data);
}

static struct file_operations dsp_link_config_get_info_fops = {
	.owner	 = THIS_MODULE,
	.open	 = dsp_link_config_get_info_open,
	.read	 = seq_read,
	.llseek	 = seq_lseek,
	.release = seq_release,
};

static int dsp_link_status_get_info_open(struct inode *inode, struct file *file)
{
	return single_open(file, dsp_link_status_get_info, PDE(inode)->data);
}

static struct file_operations dsp_link_status_get_info_fops = {
	.owner	 = THIS_MODULE,
	.open	 = dsp_link_status_get_info_open,
	.read	 = seq_read,
	.llseek	 = seq_lseek,
	.release = seq_release,
};

static int dsp_link_state_get_info_open(struct inode *inode, struct file *file)
{
	return single_open(file, dsp_link_state_get_info, PDE(inode)->data);
}

static struct file_operations dsp_link_state_get_info_fops = {
	.owner	 = THIS_MODULE,
	.open	 = dsp_link_state_get_info_open,
	.read	 = seq_read,
	.llseek	 = seq_lseek,
	.release = seq_release,
};



#else

static struct file_operations dsp_fops =
{
	read:		dsp_proc_read,
};

static struct inode_operations dsp_inode =
{
	permission:	dsp_proc_perms,
};

/*
 *	/proc/net/wanrouter/dsp 
 */

#endif

static struct proc_dir_entry *proc_dsp;
extern struct proc_dir_entry *proc_router;

/* Strings */



/*
 *	Interface functions
 */

/*
 *	Initialize router proc interface.
 */
#define DSP_DIR_NAME "dsp"

int dsp_proc_init (void)
{
	struct proc_dir_entry *p;

	if (!proc_router)
		return -ENODEV;
	
	proc_dsp = proc_mkdir(DSP_DIR_NAME, proc_router);
	if (!proc_dsp)
		goto fail;

	p = create_proc_entry("config",0,proc_dsp);
	if (!p)
		goto fail_config;
	
#ifdef LINUX_2_6
	p->proc_fops = &config_get_info_fops;
#else
	p->proc_fops = &dsp_fops;
	p->proc_iops = &dsp_inode;
	p->get_info = config_get_info;
#endif

	p = create_proc_entry("status",0,proc_dsp);
	if (!p)
		goto fail_stat;
	
#ifdef LINUX_2_6
	p->proc_fops = &status_get_info_fops;
#else
	p->proc_fops = &dsp_fops;
	p->proc_iops = &dsp_inode;
	p->get_info = status_get_info;
#endif
	return 0;
fail_stat:
	remove_proc_entry("config", proc_dsp);
fail_config:
	remove_proc_entry(DSP_DIR_NAME, proc_router);
fail:
	return -ENOMEM;
}

/*
 *	Clean up router proc interface.
 */

void dsp_proc_cleanup (void)
{
	remove_proc_entry("config", proc_dsp);
	remove_proc_entry("status", proc_dsp);
	remove_proc_entry(DSP_DIR_NAME,proc_router);
}



int dsp_proc_link_init (dsp_link_t *dsp_link)
{
	struct proc_dir_entry *p;

	if (!proc_router)
		return -ENODEV;
	
	dsp_link->proc_dir = proc_mkdir(dsp_link->name, proc_dsp);
	if (!dsp_link->proc_dir)
		goto fail;

	p = create_proc_entry("config",0,dsp_link->proc_dir);
	if (!p)
		goto fail_config;
	
#ifdef LINUX_2_6
	p->proc_fops = &dsp_link_config_get_info_fops;
#else
	p->proc_fops = &dsp_fops;
	p->proc_iops = &dsp_inode;
	p->get_info = dsp_link_config_get_info;
#endif
	p->data	= dsp_link;

	p = create_proc_entry("status",0,dsp_link->proc_dir);
	if (!p)
		goto fail_status;
	
#ifdef LINUX_2_6
	p->proc_fops = &dsp_link_status_get_info_fops;
#else
	p->proc_fops = &dsp_fops;
	p->proc_iops = &dsp_inode;
	p->get_info = dsp_link_status_get_info;
#endif
	p->data	= dsp_link;

	p = create_proc_entry("state",0,dsp_link->proc_dir);
	if (!p)
		goto fail_state;
	
#ifdef LINUX_2_6
	p->proc_fops = &dsp_link_state_get_info_fops;
#else
	p->proc_fops = &dsp_fops;
	p->proc_iops = &dsp_inode;
	p->get_info = dsp_link_state_get_info;
#endif
	p->data	= dsp_link;
	return 0;
fail_state:
	remove_proc_entry("status", dsp_link->proc_dir);
fail_status:
	remove_proc_entry("config", dsp_link->proc_dir);
fail_config:
	remove_proc_entry(dsp_link->name, proc_dsp);
fail:
	return -ENOMEM;

}


int dsp_proc_link_cleanup(dsp_link_t *dsp_link)
{
	remove_proc_entry("config", dsp_link->proc_dir);
	remove_proc_entry("status", dsp_link->proc_dir);
	remove_proc_entry("state", dsp_link->proc_dir);
	remove_proc_entry(dsp_link->name,proc_dsp);
	return 0;
}


/****** Proc filesystem entry points ****************************************/

/*
 *	Verify access rights.
 */

#ifndef LINUX_2_6


static int dsp_proc_perms (struct inode* inode, int op)
{
	return 0;
}

/*
 *	Read router proc directory entry.
 *	This is universal routine for reading all entries in /proc/net/wanrouter
 *	directory.  Each directory entry contains a pointer to the 'method' for
 *	preparing data for that entry.
 *	o verify arguments
 *	o allocate kernel buffer
 *	o call get_info() to prepare data
 *	o copy data to user space
 *	o release kernel buffer
 *
 *	Return:	number of bytes copied to user space (0, if no data)
 *		<0	error
 */

static ssize_t dsp_proc_read(struct file* file, char* buf, size_t count,
				loff_t *ppos)
{
	struct inode *inode = file->f_dentry->d_inode;
	struct proc_dir_entry* dent;
	char* page;
	int len;

	if (count <= 0)
		return 0;
		
	dent = inode->u.generic_ip;
	if ((dent == NULL) || (dent->get_info == NULL))
		return 0;
		
	page = kmalloc(count, GFP_KERNEL);
	DSP_MEM_INC(count);
	if (page == NULL)
		return -ENOBUFS;
		
	len = dent->get_info(page, dent->data, file->f_pos, count);
	if (len) {
		if(copy_to_user(buf, page, len)){
			DSP_MEM_DEC(count);
			return -EFAULT;
		}
		file->f_pos += len;
	}
	kfree(page);
	DSP_MEM_DEC(count);
	return len;
}

#endif

/*
 *	Prepare data for reading 'Config' entry.
 *	Return length of data.
 */
extern  dsp_link_t *dsp_link_list;


#define PROC_CONF_TEMP "| %-9s| %-13s| %-5s| %-16s|\n"
static char conf_hdr[] =
        "----------------------------------------------------\n"
	"| DSP Link |  master dev  | TERM |     State       |\n"
	"----------------------------------------------------\n";

static char end_hdr[] = 
	"----------------------------------------------------\n";

#ifdef LINUX_2_6
static int config_get_info(struct seq_file *m, void *v)
#else
static int config_get_info(char* buf, char** start, off_t offs, int len)
#endif
{
	dsp_link_t* 	dsp_link=NULL;
	PROC_ADD_DECL(m);
	
	PROC_ADD_INIT(m,buf,offs,len);
	PROC_ADD_LINE(m, conf_hdr);

	for (dsp_link = dsp_link_list; dsp_link; 
 	     dsp_link = dsp_link->next){

		PROC_ADD_LINE(m,
			 PROC_CONF_TEMP,
			 dsp_link->name,
			 dsp_link->master->name,
			 (dsp_link->cfg.pad_type==DSP_HPAD)?"HOST":"TERM",
			 dsp_get_mj_state(dsp_link->state));
	}

	PROC_ADD_LINE(m, end_hdr);

	PROC_ADD_LINE(m,"\nTotal Memory Used=%d\n",
		  	atomic_read(&dsp_mem_in_used));

	PROC_ADD_RET(m);
}

/*
 *	Prepare data for reading 'Status' entry.
 *	Return length of data.
 */
#define PROC_STATUS_TEMP "| %-13d|\n"
static char status_hdr[] =
        "----------------\n"
	"| Mem. in used |\n"
	"----------------\n";

static char status_end_hdr[] = 
	"----------------\n";

#ifdef LINUX_2_6
static int status_get_info(struct seq_file *m, void *v)
#else
static int status_get_info(char* buf, char** start, off_t offs, int len)
#endif
{
	PROC_ADD_DECL(m);
	
	PROC_ADD_INIT(m,buf,offs,len);
	PROC_ADD_LINE(m,status_hdr);

	PROC_ADD_LINE(m, 
		 PROC_STATUS_TEMP,
		 atomic_read(&dsp_mem_in_used));

	PROC_ADD_LINE(m, status_end_hdr);
	PROC_ADD_RET(m);
}


#define PROC_LINK_CONF_DS_FORMAT "| %-17s| %-10d| %-17s| %-10s|\n"

static char conf_link_hdr1[] =
        "---------------------------------------------------------------\n"
	"| DSP Link Configuration                                      |\n"
	"---------------------------------------------------------------\n";

static char conf_link_line[] =
	"---------------------------------------------------------------\n";

#ifdef LINUX_2_6
static int dsp_link_config_get_info(struct seq_file *m, void *v)
#else
static int dsp_link_config_get_info(char* buf, char** start, off_t offs, int len)
#endif
{
	dsp_link_t *dsp_link = PROC_GET_DATA();
	PROC_ADD_DECL(m);
		
	if (!dsp_link)
		return 0;

	PROC_ADD_INIT(m,buf,offs,len);
	PROC_ADD_LINE(m,conf_link_hdr1);

	PROC_ADD_LINE(m,
		 PROC_LINK_CONF_DS_FORMAT,
		 "T1", dsp_link->cfg.T1/HZ,
		 "AUTO_CE", CFG_DECODE(dsp_link->cfg.auto_ce==DSP_TRUE));
	
	PROC_ADD_LINE(m,
		 PROC_LINK_CONF_DS_FORMAT,
		 "T2", dsp_link->cfg.T2/HZ,
		 "AUTO_CALL_REQ", CFG_DECODE(dsp_link->cfg.auto_call_req==DSP_TRUE));

	PROC_ADD_LINE(m,
		 PROC_LINK_CONF_DS_FORMAT,
		 "T3", dsp_link->cfg.T3/HZ,
		 "AUTO_ACK", CFG_DECODE(dsp_link->cfg.auto_ack==DSP_TRUE));

	PROC_ADD_LINE(m, conf_link_line);
	PROC_ADD_RET(m);
#if 0
	strcpy(buf,conf_link_hdr1);
	cnt += sizeof(conf_link_hdr1)-1;


	cnt += sprintf(&buf[cnt],PROC_LINK_CONF_DS_FORMAT,
			"T1", dsp_link->cfg.T1/HZ,
			"AUTO_CE", (dsp_link->cfg.auto_ce==DSP_TRUE)?"YES":"NO");
	
	cnt += sprintf(&buf[cnt],PROC_LINK_CONF_DS_FORMAT,
			"T2", dsp_link->cfg.T2/HZ,
			"AUTO_CALL_REQ", (dsp_link->cfg.auto_call_req==DSP_TRUE)?"YES":"NO");

	cnt += sprintf(&buf[cnt],PROC_LINK_CONF_DS_FORMAT,
			"T3", dsp_link->cfg.T3/HZ,
			"AUTO_ACK", (dsp_link->cfg.auto_ack==DSP_TRUE)?"YES":"NO");

	strcpy(&buf[cnt],conf_link_line);
	cnt += sizeof(conf_link_line)-1;

	return cnt;
#endif
}

#define PROC_LINK_STATS_FORMAT "| %-24s|  %-17d|   %-17d|\n"
unsigned char link_line[]= "--------------------------------------------------------------------\n";
unsigned char link_hdr[] = "| Packet Type             |     Received      |    Transmitted     |\n";
	

#ifdef LINUX_2_6
static int dsp_link_status_get_info(struct seq_file *m, void *v)
#else
static int dsp_link_status_get_info(char* buf, char** start, off_t offs, int len)
#endif
{
	dsp_link_t *dsp_link = PROC_GET_DATA();
	PROC_ADD_DECL(m);

	if (!dsp_link)
		return 0;
	PROC_ADD_INIT(m,buf,offs,len);
	PROC_ADD_LINE(m,link_line);
	PROC_ADD_LINE(m,link_hdr);
	PROC_ADD_LINE(m,link_line);

	PROC_ADD_LINE(m, 
		 PROC_LINK_STATS_FORMAT,
		 "Data Msg (with ACK)", 
		 dsp_link->dspstats.rx_ack_data_msg,
		 dsp_link->dspstats.tx_ack_data_msg);
	PROC_ADD_LINE(m,
		 PROC_LINK_STATS_FORMAT,
		 "Data Msg (without ACK)", 
		 dsp_link->dspstats.rx_data_msg,
		 dsp_link->dspstats.tx_data_msg);
#if 0
	PROC_ADD_LINE(m,
		(buf, &cnt, len, offs, &stop_cnt, &size, 
		 PROC_LINK_STATS_FORMAT,
		 "Data Pkt ", 
		 dsp_link->dspstats.rx_data_pkt,
		 dsp_link->dspstats.tx_data_pkt);
#endif
	PROC_ADD_LINE(m,
		 PROC_LINK_STATS_FORMAT,
		 "Call Request Msg",
		 dsp_link->dspstats.rx_call_req,
		 dsp_link->dspstats.tx_call_req);
	PROC_ADD_LINE(m,
		 PROC_LINK_STATS_FORMAT,
		 "Call Accept Msg", 
		 dsp_link->dspstats.rx_call_acc,
		 dsp_link->dspstats.tx_call_acc);
	PROC_ADD_LINE(m,
		 PROC_LINK_STATS_FORMAT,
		 "Circuit Enabled Msg",
		 dsp_link->dspstats.rx_ce,
		 dsp_link->dspstats.tx_ce);
	PROC_ADD_LINE(m,
		 PROC_LINK_STATS_FORMAT,
		 "Circuit Reset Msg", 
		 dsp_link->dspstats.rx_cr,
		 dsp_link->dspstats.tx_cr);
	PROC_ADD_LINE(m,
		 PROC_LINK_STATS_FORMAT,
		 "Circuit Disconnect Msg", 
		 dsp_link->dspstats.rx_cd,
		 dsp_link->dspstats.tx_cd);
	PROC_ADD_LINE(m,
		 PROC_LINK_STATS_FORMAT,
		 "Circuit Request Msg", 
		 dsp_link->dspstats.rx_creq,
		 dsp_link->dspstats.tx_creq);
	PROC_ADD_LINE(m,
		 PROC_LINK_STATS_FORMAT,
		 "End-to-End Ack Msg", 
		 dsp_link->dspstats.rx_ack,
		 dsp_link->dspstats.tx_ack);
	PROC_ADD_LINE(m,
		 PROC_LINK_STATS_FORMAT,
		 "Status Msg", 
		 dsp_link->dspstats.rx_status,
		 dsp_link->dspstats.tx_status);
	PROC_ADD_LINE(m,
		 PROC_LINK_STATS_FORMAT,
		 "Cmd/Res Undelivered Msg", 
		 dsp_link->dspstats.rx_cmd_res_undelivered,
		 dsp_link->dspstats.tx_cmd_res_undelivered);
	PROC_ADD_LINE(m,
		 PROC_LINK_STATS_FORMAT,
		 "Cmd/Res Aborted Msg", 
		 dsp_link->dspstats.rx_cmd_res_aborted,
		 dsp_link->dspstats.tx_cmd_res_aborted);
	PROC_ADD_LINE(m,
		 PROC_LINK_STATS_FORMAT,
		 "Invitation to Clear Msg",
		 dsp_link->dspstats.rx_inv_to_clear,
		 dsp_link->dspstats.tx_inv_to_clear);
	PROC_ADD_LINE(m,
		 PROC_LINK_STATS_FORMAT,
		 "Dropped Msg",
		 dsp_link->dspstats.rx_dropped,
		 dsp_link->dspstats.tx_dropped);

	PROC_ADD_LINE(m,link_line);
	PROC_ADD_LINE(m,
		 PROC_LINK_STATS_FORMAT,
		 "Total Pkts", 
		 dsp_link->dspstats.rx_total,
		 dsp_link->dspstats.tx_total);

	PROC_ADD_LINE(m,link_line);

	PROC_ADD_RET(m);

#if 0
	
	strcpy(buf,link_line);
	cnt += sizeof(link_line)-1;

	strcpy(&buf[cnt],link_hdr);
	cnt += sizeof(link_hdr)-1;
	
	strcpy(&buf[cnt],link_line);
	cnt += sizeof(link_line)-1;

	cnt += sprintf(&buf[cnt],PROC_LINK_STATS_FORMAT,
			"Data Msg (with ACK)", 
			dsp_link->dspstats.rx_ack_data_msg,
			dsp_link->dspstats.tx_ack_data_msg);
	cnt += sprintf(&buf[cnt],PROC_LINK_STATS_FORMAT,
			"Data Msg (without ACK)", 
			dsp_link->dspstats.rx_data_msg,
			dsp_link->dspstats.tx_data_msg);
#if 0
	cnt += sprintf(&buf[cnt],PROC_LINK_STATS_FORMAT,
			"Data Pkt ", 
			dsp_link->dspstats.rx_data_pkt,
			dsp_link->dspstats.tx_data_pkt);
#endif
	cnt += sprintf(&buf[cnt],PROC_LINK_STATS_FORMAT,
			"Call Request Msg", 
			dsp_link->dspstats.rx_call_req,
			dsp_link->dspstats.tx_call_req);
	cnt += sprintf(&buf[cnt],PROC_LINK_STATS_FORMAT,
			"Call Accept Msg", 
			dsp_link->dspstats.rx_call_acc,
			dsp_link->dspstats.tx_call_acc);
	cnt += sprintf(&buf[cnt],PROC_LINK_STATS_FORMAT,
			"Circuit Enabled Msg", 
			dsp_link->dspstats.rx_ce,
			dsp_link->dspstats.tx_ce);
	cnt += sprintf(&buf[cnt],PROC_LINK_STATS_FORMAT,
			"Circuit Reset Msg", 
			dsp_link->dspstats.rx_cr,
			dsp_link->dspstats.tx_cr);
	cnt += sprintf(&buf[cnt],PROC_LINK_STATS_FORMAT,
			"Circuit Disconnect Msg", 
			dsp_link->dspstats.rx_cd,
			dsp_link->dspstats.tx_cd);
	cnt += sprintf(&buf[cnt],PROC_LINK_STATS_FORMAT,
			"Circuit Request Msg", 
			dsp_link->dspstats.rx_creq,
			dsp_link->dspstats.tx_creq);
	cnt += sprintf(&buf[cnt],PROC_LINK_STATS_FORMAT,
			"End-to-End Ack Msg", 
			dsp_link->dspstats.rx_ack,
			dsp_link->dspstats.tx_ack);
	cnt += sprintf(&buf[cnt],PROC_LINK_STATS_FORMAT,
			"Status Msg", 
			dsp_link->dspstats.rx_status,
			dsp_link->dspstats.tx_status);
	cnt += sprintf(&buf[cnt],PROC_LINK_STATS_FORMAT,
			"Cmd/Res Undelivered Msg", 
			dsp_link->dspstats.rx_cmd_res_undelivered,
			dsp_link->dspstats.tx_cmd_res_undelivered);
	cnt += sprintf(&buf[cnt],PROC_LINK_STATS_FORMAT,
			"Cmd/Res Aborted Msg", 
			dsp_link->dspstats.rx_cmd_res_aborted,
			dsp_link->dspstats.tx_cmd_res_aborted);
	cnt += sprintf(&buf[cnt],PROC_LINK_STATS_FORMAT,
			"Invitation to Clear Msg",
			dsp_link->dspstats.rx_inv_to_clear,
			dsp_link->dspstats.tx_inv_to_clear);
	cnt += sprintf(&buf[cnt],PROC_LINK_STATS_FORMAT,
			"Dropped Msg",
			dsp_link->dspstats.rx_dropped,
			dsp_link->dspstats.tx_dropped);

	strcpy(&buf[cnt],link_line);
	cnt += sizeof(link_line)-1;
	
	cnt += sprintf(&buf[cnt],PROC_LINK_STATS_FORMAT,
			"Total Pkts", 
			dsp_link->dspstats.rx_total,
			dsp_link->dspstats.tx_total);

	strcpy(&buf[cnt],link_line);
	cnt += sizeof(link_line)-1;

	return cnt;
#endif
}



/*
 *	Prepare data for reading <device> entry.
 *	Return length of data.
 *
 *	On entry, the 'start' argument will contain a pointer to WAN device
 *	data space.
 */


#define PROC_DEV_SS_FORMAT "| %-15s : %-13s | %-15s : %-13s |\n"
#define PROC_DEV_SD_FORMAT "| %-15s : %-13s | %-15s : %-13d |\n"
#define PROC_DEV_DS_FORMAT "| %-15s : %-13d | %-15s : %-13s |\n"
#define PROC_DEV_DD_FORMAT "| %-15s : %-13d | %-15s : %-13d |\n"
#define PROC_DEV_D_FORMAT  "| %-15s : %-13d |                 :               |\n"
unsigned char dev_line[]= "---------------------------------------------------------------------\n";

#ifdef LINUX_2_6
static int dsp_link_state_get_info(struct seq_file *m, void *v)
#else
static int dsp_link_state_get_info(char* buf, char** start, off_t offs, int len)
#endif
{
	dsp_link_t *dsp_link = PROC_GET_DATA();
	PROC_ADD_DECL(m);


	if (!dsp_link)
		return 0;

	PROC_ADD_INIT(m,buf,offs,len);
	PROC_ADD_LINE(m,dev_line);
	PROC_ADD_LINE(m, 
		 "|      DSP Protocol Driver Stats                                    |\n");
	PROC_ADD_LINE(m, dev_line);

	PROC_ADD_LINE(m, PROC_DEV_SS_FORMAT, 
		 "Link State", 
		 dsp_get_mj_state(dsp_link->state),
		 "Dev State", netif_queue_stopped(dsp_link->dev) ? "Stp" : "Run");
	
	PROC_ADD_LINE(m, PROC_DEV_DD_FORMAT, 
		 "Bound Lcn", dsp_link->lcn,
		 "DSP kicked", test_bit(DSP_KICKED, &dsp_link->tq_working));

	PROC_ADD_LINE(m, PROC_DEV_DD_FORMAT, 
		 "Write Queue", skb_queue_len(&dsp_link->write_queue),
		 "TX BH kick Cnt", dsp_link->tx_kick_cnt);

	PROC_ADD_LINE(m, PROC_DEV_DD_FORMAT, 
		 "Async Queue", (unsigned int)skb_queue_len(&dsp_link->async_queue),
		 "DSP Critical", (unsigned int)dsp_link->critical);

	PROC_ADD_LINE(m, PROC_DEV_DD_FORMAT, 
		 "BH Running", test_bit(BH_RUNNING, &dsp_link->tq_working),
		 "BH Wt Kick", test_bit(BH_AWAITING_KICK, &dsp_link->tq_working));

	PROC_ADD_LINE(m, PROC_DEV_DD_FORMAT, 
		 "Mem in used", atomic_read(&dsp_link->mem_in_used),
		 "TQ Working", (unsigned int)dsp_link->tq_working);

	PROC_ADD_LINE(m, PROC_DEV_SS_FORMAT, 
		 "Sock bound", dsp_link->sk_id?"YES":"NO",
		 "N/A", "");

	PROC_ADD_LINE(m, dev_line);

	PROC_ADD_RET(m);

#if 0
	strcpy(&buf[cnt],dev_line);
	cnt += sizeof(dev_line)-1;

	cnt += sprintf(&buf[cnt],"|      DSP Protocol Driver Stats                                    |\n");

	strcpy(&buf[cnt],dev_line);
	cnt += sizeof(dev_line)-1;

	cnt += sprintf(&buf[cnt],PROC_DEV_SS_FORMAT, 
			"Link State", 
			dsp_get_mj_state(dsp_link->state),
			"Dev State", netif_queue_stopped(dsp_link->dev) ? "Stp" : "Run");
	
	cnt += sprintf(&buf[cnt],PROC_DEV_DD_FORMAT, 
			"Bound Lcn", dsp_link->lcn,
			"DSP kicked", test_bit(DSP_KICKED, &dsp_link->tq_working);

	cnt += sprintf(&buf[cnt],PROC_DEV_DD_FORMAT, 
			"Write Queue", skb_queue_len(&dsp_link->write_queue),
			"TX BH kick Cnt", dsp_link->tx_kick_cnt);

	cnt += sprintf(&buf[cnt],PROC_DEV_DD_FORMAT, 
			"Async Queue", skb_queue_len(&dsp_link->async_queue),
			"DSP Critical", dsp_link->critical);

	cnt += sprintf(&buf[cnt],PROC_DEV_DD_FORMAT, 
			"BH Running", test_bit(BH_RUNNING, &dsp_link->tq_working),
			"BH Wt Kick", test_bit(BH_AWAITING_KICK, &dsp_link->tq_working);

	cnt += sprintf(&buf[cnt],PROC_DEV_D_FORMAT, 
			"Mem in used", atomic_read(&dsp_link->mem_in_used);

	strcpy(&buf[cnt],dev_line);
	cnt += sizeof(dev_line)-1;
	
	return cnt;
#endif
}

#endif


