#include <linux/wanpipe_includes.h>
#include <linux/wanpipe_cfg.h>
#include "wanpipe_dsp.h"

#ifdef DSP_DEBUG
extern unsigned short dsp_debug;
#endif

/*
 *	This routine purges all the queues of frames.
 */
void dsp_clear_link_queues(dsp_link_t *dsp_link, int dsp_state)
{
	struct sk_buff *skb = NULL;

	DSP_ASSERT1(dsp_link == NULL);
	dsp_set_state(dsp_link, dsp_state);
	
	spin_lock_bh(&dsp_link->list_lock);
	while ((skb = skb_dequeue(&dsp_link->async_queue)) != NULL){
		KFREE_SKB(skb,dsp_link);
	}

	while ((skb = skb_dequeue(&dsp_link->write_queue)) != NULL){
		KFREE_SKB(skb,dsp_link);
	}

	//rx queue is not used!
	//while ((skb = skb_dequeue(&dsp->rx_queue)) != NULL)
	//	KFREE_SKB(skb,dsp_link);
	
	spin_unlock_bh(&dsp_link->list_lock);

	dsp_stop_t1_timer(dsp_link);
	dsp_stop_t2_timer(dsp_link);
}

int dsp_trigger_link_bh(dsp_link_t *dsp_link)
{
	DSP_ASSERT(dsp_link == NULL);
	DSP_ASSERT(dsp_link->dev == NULL);

	if (test_bit(CRIT_LINK_DOWN,&dsp_link->critical)){
		printk(KERN_INFO "%s: Failed to trigger BH, link down!\n",
				dsp_link->dev ? dsp_link->dev->name:"DSP");
		return 1;
	}

	if (test_bit(BH_AWAITING_KICK, &dsp_link->tq_working)){
		if (dsp_link->master && !netif_queue_stopped(dsp_link->master)){
			printk(KERN_INFO "%s: DSP trigger waiting for kick with master open %s!\n",
					dsp_link->name,dsp_link->master->name);
			clear_bit(BH_AWAITING_KICK, &dsp_link->tq_working);
		}
	}
		
	if (!test_bit(BH_AWAITING_KICK, &dsp_link->tq_working) &&
	    !test_bit(BH_RUNNING, &dsp_link->tq_working)){
	    	set_bit(BH_RUNNING, &dsp_link->tq_working);

		tasklet_schedule(&dsp_link->dsp_tx_task);
		return 0;
	}
	DEBUG7((KERN_INFO "%s: Failed to trigger BH (%i,%i)\n",
		dsp_link->name, 
		test_bit(BH_AWAITING_KICK, &dsp_link->tq_working),
		test_bit(BH_RUNNING, &dsp_link->tq_working))); 
	return 1;
}

void dsp_set_state(dsp_link_t* dsp_link, int state)
{
	DSP_ASSERT1(dsp_link == NULL);
	DSP_ASSERT1(dsp_link->dev == NULL);

	if (dsp_link->state == state)
		return;

	dsp_link->state = state;

	printk(KERN_INFO "%s: Link %s%s\n",
			dsp_link->name,
			dsp_get_mj_state(dsp_link->state),
			dsp_get_mn_state(dsp_link->state));

	if (state == DSP_CONNECTED && 
	    (dsp_link->dev->flags&IFF_UP) &&
	    netif_queue_stopped(dsp_link->dev)){
		netif_start_queue(dsp_link->dev);
	}
}

char* dsp_get_mj_state(int state)
{
	switch(state & DSP_MJ_MASK){
		case DSP_MJ_DISCONNECTED:
			return ("disconnected");

		case DSP_MJ_READY:
			return ("ready");

		case DSP_MJ_CONNECTING:
			return ("connecting ...");
			
		case DSP_MJ_CONNECTED:
			return ("connected");
			
		case DSP_MJ_DISCONNECTING:
			return ("disconnecting ...");
	}
	return ("unknown!");
}

char* dsp_get_mn_state(int state)
{
	switch(state & DSP_MN_MASK){
		case DSP_MN_ACC_CALL:
			return (" (Acc Call)");
		case DSP_MN_CALL_REQ_AGAIN:
			return (" (Call Req. Again)");
		case DSP_MN_CIRUIT_RESET:
			return (" (M_CR)");
		case DSP_MN_ITC:
			return (" (M_ITC)");
		case DSP_MN_CLEAR_REQ:
			return (" (Clear Req)");
	}
	return (" ");
}

unsigned char dsp_get_sn(unsigned char sn)
{
	unsigned char new_sn = 0x00;
	new_sn = sn % DSP_SN_MODULES;
	return new_sn; 
}

int dsp_chk_ack_sn(dsp_link_t* dsp_link, unsigned char sn, 
		   unsigned char* next_sn, int from_local)
{
	unsigned char 	tmp_sn = 0x00;
	unsigned char 	need_ack = (from_local == DSP_TRUE) ? 
				DSP_SN_WAIT_ACK_L : DSP_SN_WAIT_ACK_R;
	unsigned char 	last_ack = 0;
	int		i = 0;

	DSP_ASSERT(dsp_link == NULL);
	if (from_local == DSP_TRUE){
		last_ack = dsp_link->last_ack_sn_l;
		DEBUG20((KERN_INFO "%s: Last ack from local device %d\n",
				dsp_link->name, last_ack));
	}else{
		last_ack = dsp_link->last_ack_sn_r;
		DEBUG20((KERN_INFO "%s: Last ack from remote device %d\n",
				dsp_link->name, last_ack));
	}
	tmp_sn = dsp_get_sn(last_ack+1);
	while(tmp_sn != last_ack){
		DEBUG20((KERN_DEBUG "%s: Checking sn=%d (array_sn=%d)\n", 
				dsp_link->name, tmp_sn, 
				dsp_link->array_sn[tmp_sn]));
		if (dsp_link->array_sn[tmp_sn] & need_ack)
			break;
		tmp_sn = dsp_get_sn(tmp_sn+1);
		if (++i == DSP_SN_MODULES)
			break;
	}
	if (tmp_sn == last_ack || sn != tmp_sn || i == DSP_SN_MODULES) {
		DSP_ASSERT(next_sn == NULL);
		*next_sn = tmp_sn;
		return DSP_FALSE;
	}

	/* Received ACK for next waiting SN */
	if (from_local == DSP_TRUE)
		dsp_link->last_ack_sn_l = sn;
	else
		dsp_link->last_ack_sn_r = sn;
	dsp_link->array_sn[sn] &= ~need_ack;
	return DSP_TRUE;
}

int dsp_wait4ack(dsp_link_t* dsp_link)
{
	unsigned char 	tmp_sn = 0x00;
	unsigned char 	last_ack = dsp_link->last_ack_sn_l;
	int		i = 0;

	DSP_ASSERT(dsp_link == NULL);
	tmp_sn = dsp_get_sn(last_ack+1);
	while(tmp_sn != last_ack){
		DEBUG20((KERN_DEBUG "%s: Checking sn=%d (array_sn=%d)\n", 
				dsp_link->name, tmp_sn, 
				dsp_link->array_sn[tmp_sn]));
		if (dsp_link->array_sn[tmp_sn] & DSP_SN_WAIT_ACK_L)
			break;
		if (dsp_link->rx_sn == tmp_sn)
			break;
		tmp_sn = dsp_get_sn(tmp_sn+1);
		if (++i == DSP_SN_MODULES)
			break;
	}
	/* There are no Data Messages waiting for EEACK */
	if (tmp_sn == last_ack || dsp_link->rx_sn == tmp_sn || i == DSP_SN_MODULES) {
		return DSP_FALSE;
	}

	/* There is at least one Data Message waiting for EEACK */
	return DSP_TRUE;
}

static void dsp_free_skb_msg(dsp_link_t* dsp_link)
{
	struct sk_buff*	skb = NULL;

	while((skb = skb_dequeue_tail(&dsp_link->write_queue)) != NULL){
		KFREE_SKB(skb,dsp_link);
	}
}

void dsp_free_rx_skb(dsp_link_t* dsp_link)
{
	if (test_and_set_bit(RX_SKB_FREE,(void*)&dsp_link->critical))
		return;
	if (dsp_link->rx_skb != NULL){
		KFREE_SKB(dsp_link->rx_skb,dsp_link);
		dsp_link->rx_skb= NULL;
	}
	clear_bit(RX_SKB_FREE,(void*)&dsp_link->critical);
}

int dsp_split_tx_data(dsp_link_t* dsp_link, struct sk_buff* skb)
{
	struct sk_buff*	new_skb = NULL;
	dspapi_hdr_t	dspapi_hdr;
	x25api_hdr_t*	x25api_hdr = NULL;
	char*		packet = NULL;
	int 		is_ack = 0, data_len = 0, pkt_len = 0;
	int 		hdr_len = 0, max_tx_len = 0;
	
	DSP_ASSERT(dsp_link == NULL);
	DSP_ASSERT(dsp_link->dev == NULL);
	DSP_ASSERT(skb == NULL);

	/* Get the maximum packet size */
	max_tx_len = dsp_link->dev->mtu;
	/* Copy/Remove DSP API header */
	memcpy(&dspapi_hdr, skb->data, sizeof(dspapi_hdr_t));
	skb_pull(skb, sizeof(dspapi_hdr_t));

	is_ack = (dspapi_hdr.dsp_pkt_type == DSP_DATA_ACK_MSG) ? 
					DSP_TRUE : DSP_FALSE;

	/* Save Tx packet info in 'dsp_api_event.hdr' */
	dsp_link->dsp_api_event.hdr.dsp_pkt_type = dspapi_hdr.dsp_pkt_type;
	dsp_link->dsp_api_event.hdr.dsp_length = dspapi_hdr.dsp_length;

	/* Add DSP header and tail for M_DATA */
	hdr_len = DSP_HDR_LEN;
#if 0
	ALEX 10.4.2001 
	ETX_ETB_ESC
	/**************************************/
	/* The ESC byte is part of data frame */
	/**************************************/
	hdr_len = DSP_TX_HDR_LEN(dsp_link);
#endif
	
	if ((skb_headroom(skb) < hdr_len) || (skb_tailroom(skb) < 1)){
		DSP_SKB_HEAD(skb);
		DSP_SKB_TAIL(skb);
		return -ENOMEM;
	}
	/* Add one byte for sequence number */
	packet = skb_push(skb, hdr_len);
	IS_NULL(packet, -ENOMEM);
	/* Update DSP header */
	packet[0x00] = 0x00;
	/* Set ACK bit if end-to-end acknoledgment is required for this M_DATA
	 * Set XPR bit if the user data contains transparent data.
	 */
	packet[0x01] = SET_R_ACK(is_ack) | SET_R_XPR(dspapi_hdr.dsp_data_transp);
	/* Sequence number will update before sending this packet */
	packet[0x02] = dsp_link->tx_sn; //0x00;
#if 0
	ALEX 10.4.2001 
	ETX_ETB_ESC
	/*************************************************/
	/* This two bytes are already in data frame	 */
	/*************************************************/
	/* For Host PAD, add ESC character at the header */
	if (dsp_link->cfg.pad_type == DSP_HPAD){
		packet[0x03] = DSP_ESC;
	}
	/* Add DSP tail (ETX/ETB)*/	
	packet = skb_put(skb, 1);
	IS_NULL(packet, -ENOMEM);
	packet[0] = SET_ET_BYTE(dspapi_hdr.dsp_et_byte);
#endif
	
	/* Update DSP statistics */
	if (is_ack == DSP_TRUE){
		DSP_INC_TX_STAT(tx_ack_data_msg);
	}else{
		DSP_INC_TX_STAT(tx_data_msg);
	}

	data_len = skb->len;
	do {
		if (data_len > dsp_link->tx_pkt_len){
			pkt_len = dsp_link->tx_pkt_len;
		}else{
			pkt_len = data_len;
		}
		
		data_len -= pkt_len;
		if (data_len) {
			ALLOC_SKB(new_skb,pkt_len+sizeof(x25api_hdr_t),dsp_link);
			if (new_skb == NULL){
				printk(KERN_INFO "%s: Critical Error: No Memory!\n",
						dsp_link->name);
				dsp_free_skb_msg(dsp_link);
				return -ENOMEM;
			}
			new_skb->dev = skb->dev;
			new_skb->protocol = skb->protocol;
			if (skb->sk)
				skb_set_owner_w(new_skb, skb->sk);

			packet = skb_put(new_skb, pkt_len + sizeof(x25api_hdr_t));
			IS_NULL(packet, -ENOMEM);			
			packet = &new_skb->data[sizeof(x25api_hdr_t)];
			memcpy(packet, skb->data, pkt_len);
		}else{
			if (skb_headroom(skb) < hdr_len){
				DSP_SKB_HEAD(skb);
				dsp_free_skb_msg(dsp_link);
				return -ENOMEM;
			}
			/* Reuse the original skb buffer and don not call `kfree` for it */
			packet = skb_push(skb, sizeof(x25api_hdr_t));
			IS_NULL(packet, -ENOMEM);
			new_skb = skb;
		}
		/* Set X25 API header */
		x25api_hdr = (x25api_hdr_t*)new_skb->data;
		x25api_hdr->length = new_skb->len;
		x25api_hdr->qdm = 0x00;
		if (data_len){
			/* Set X25 M-bit */
			x25api_hdr->qdm |= SET_L_M_BIT(DSP_TRUE);

			/* Remove copied data from original skb */
			skb_pull(skb, pkt_len);
		}else{
			/* Clear X25 M-bit */
			x25api_hdr->qdm |= SET_L_M_BIT(DSP_FALSE);
		}

		skb_queue_tail(&dsp_link->write_queue, new_skb);

	} while(data_len != 0);

	return DSP_OK;
}

struct sk_buff* dsp_concat_rx_data(dsp_link_t* dsp_link, struct sk_buff* skb)
{
	x25api_hdr_t*	x25api_hdr = NULL;
	int 		is_m_bit = DSP_FALSE;

	DSP_ASSERT2(dsp_link == NULL, NULL);
	DSP_ASSERT2(skb == NULL, NULL);
	DSP_ASSERT2(skb->data == NULL, NULL);

	if (test_bit(RX_SKB_FREE,(void*)&dsp_link->critical)){
		KFREE_SKB(skb,dsp_link);
		return NULL;
	}
	x25api_hdr = (x25api_hdr_t*)skb->data;
	is_m_bit = IS_L_M_BIT(x25api_hdr->qdm);
	if (dsp_link->rx_skb == NULL){
		if (is_m_bit == DSP_TRUE){
			/* Wait for other packets of the same M_DATA */
			dsp_link->rx_skb = skb;
			return NULL;
		}
		/* Received single packet */
		return skb;
	}else{
		struct sk_buff*	new_skb = NULL;
		struct sk_buff*	rx_skb = dsp_link->rx_skb;
		char*		packet = NULL;
		int 		len = 0;

		/* Remove X25 API header */
		skb_pull(skb, sizeof(x25api_hdr_t));
		len = rx_skb->len + skb->len;

		ALLOC_SKB(new_skb,len,dsp_link);
		if (new_skb == NULL){
			printk(KERN_INFO "%s: Critical Error: No Memory!\n",
					dsp_link->name);
			KFREE_SKB(skb,dsp_link);
			dsp_free_rx_skb(dsp_link);
			return NULL;
		}
		packet = skb_put(new_skb, len);
		IS_NULL(packet, NULL);
		memcpy(packet, rx_skb->data, rx_skb->len);
		memcpy(&packet[rx_skb->len], skb->data, skb->len);
		KFREE_SKB(rx_skb,dsp_link);
		KFREE_SKB(skb,dsp_link);

		if (is_m_bit == DSP_TRUE){
			dsp_link->rx_skb = new_skb;
			return NULL;
		}else{
			dsp_link->rx_skb = NULL;
			return new_skb;
		}
	}
	dsp_link->rx_skb = NULL;
	return NULL;
}
