#include <linux/wanpipe_includes.h>
#include <linux/wanpipe_cfg.h>
#include "wanpipe_dsp.h"

#ifdef DSP_DEBUG
extern unsigned short dsp_debug;
#endif

/* DSP's own task queue, used for all API's.
 * All protocol specific tasks will be instered
 * into "dsp_tq_custom" task_queue. 

 * On each rx_interrupt, the whole task queue
 * (dsp_tq_custom) will be queued into 
 * IMMEDIATE_BH via dsp_mark_bh() call. 
 
 * The IMMEDIATE_BH will execute run_dsp_tq() 
 * function, which will execute all pending,
 * tasks in dsp_tq_custom queue */


#if 0
static int dsp_bh_critical=0;


static void run_dsp_tq (unsigned long data)
{
	task_queue *tq_queue = (task_queue *)data;
	if (test_and_set_bit(2,(void*)&dsp_bh_critical))
		printk(KERN_INFO "CRITICAL IN RUNNING TASK QUEUE\n");
	run_task_queue (tq_queue);
	clear_bit(2,(void*)&dsp_bh_critical);
}

DECLARE_TASK_QUEUE(dsp_tq_custom);
static struct tq_struct dsp_tq_task = 
{
	routine: (void (*)(void *)) run_dsp_tq,
	data: &dsp_tq_custom
};


void dsp_queue_tq (struct tq_struct *bh_pointer)
{
	if (test_and_set_bit(1,(void*)&dsp_bh_critical))
		printk(KERN_INFO "CRITICAL IN QUEUING TASK\n");

	queue_task(bh_pointer,&dsp_tq_custom);
	clear_bit(1,(void*)&dsp_bh_critical);
}

void dsp_mark_bh (void)
{
	if (!test_and_set_bit(0,(void*)&dsp_bh_critical)){
		queue_task(&dsp_tq_task,&tq_immediate);
		mark_bh(IMMEDIATE_BH);
		clear_bit(0,(void*)&dsp_bh_critical);
	}else{
		printk(KERN_INFO "CRITICAL: Set mark BH \n");
	}
} 
#endif

/*============================================================
 * RESTART TIMER
 * ==========================================================*/

static void dsp_t1_timer_expiry(unsigned long dsp_id)
{
	dsp_link_t *dsp_link = (dsp_link_t*)dsp_id;
	int err = 0;

	DSP_ASSERT1(dsp_link == NULL);

	if (test_bit(CRIT_LINK_DOWN,&dsp_link->critical)){
		return;
	}
	
	if (!test_and_clear_bit(T1_RUNNING,(void*)&dsp_link->critical))
		return;
	printk(KERN_INFO "%s: DSP Service Timeout Period is expired.\n",
			dsp_link->name);

	/* Clear call */
	// 11.26.2001
	err = dsp_api_clear_call(dsp_link, NULL);

	/* Send PLACE CALL again when the link will be ready */
	// 11.25.2001
	if (!err && dsp_link->state == DSP_DISCONNECTING_CL_REQ)
		dsp_set_state(dsp_link, DSP_CONNECTING_CALL_REQ_AGAIN);
}

void dsp_start_t1_timer(dsp_link_t *dsp_link)
{
	DSP_ASSERT1(dsp_link == NULL);

	if (test_bit(CRIT_LINK_DOWN,&dsp_link->critical)){
		return;
	}
	
	/* Prevent to start T2 timer more than once */
	if (test_bit(T1_RUNNING,(void*)&dsp_link->critical))
		return;
	if (dsp_link->cfg.T1 == 0)
		return;
	DEBUG4((KERN_INFO "%s: Starting DSP Service Timeout period...\n",
			dsp_link->name));
	dsp_link->T1_timer.data     = (unsigned long)dsp_link;
	dsp_link->T1_timer.function = &dsp_t1_timer_expiry;
	dsp_link->T1_timer.expires  = jiffies + dsp_link->cfg.T1;
	
	del_timer(&dsp_link->T1_timer);
	add_timer(&dsp_link->T1_timer);
	set_bit(T1_RUNNING,(void*)&dsp_link->critical);
}

void dsp_stop_t1_timer(dsp_link_t *dsp_link)
{
	DSP_ASSERT1(dsp_link == NULL);
	if (!test_and_clear_bit(T1_RUNNING,(void*)&dsp_link->critical))
		return;
	del_timer(&dsp_link->T1_timer);
	DEBUG4((KERN_INFO "%s: Stopping DSP Service Timeout period...\n",
			dsp_link->name));
}

/*============================================================
 * RESTART TIMER
 * ==========================================================*/

static void dsp_t2_timer_expiry(unsigned long dsp_id)
{
	dsp_link_t *dsp_link = (dsp_link_t*)dsp_id;

	DSP_ASSERT1(dsp_link == NULL);

	if (test_bit(CRIT_LINK_DOWN,&dsp_link->critical)){
		return;
	}
	
	/* Test if T2 is still running */
	if (!test_and_clear_bit(T2_RUNNING,(void*)&dsp_link->critical))
		return;
	printk(KERN_INFO "%s: DSP PAD Protocol Timeout period expired.\n",
			dsp_link->name);

	/* 1. The subsequent packet of the same Data Message
	 * didn't arrive with the PAD protocol timeout period,
	 * the entire message is dropped.
	 * 2. A response to a M_CR is not received within the PAD
	 * protocol timeout.
	 * The PAD initiates the clearning of the user circuit.
	 */
	dsp_free_rx_skb(dsp_link);
	
	//dsp_set_state(dsp_link, DSP_DISCONNECTING);
	/* Set reason code */
	dsp_link->reason_code = DSP_RC_TIMEOUT;
	/* If this timer started, call INVITATION_TO_CLEAR now */
	dsp_api_inv_to_clear(dsp_link, NULL);
}

void dsp_start_t2_timer(dsp_link_t *dsp_link)
{
	DSP_ASSERT1(dsp_link == NULL);

	if (test_bit(CRIT_LINK_DOWN,&dsp_link->critical)){
		return;
	}
	
	/* Prevent to start T2 timer more than once */
	if (test_bit(T2_RUNNING,(void*)&dsp_link->critical))
		return;
	if (dsp_link->cfg.T2 == 0)
		return;
	DEBUG4((KERN_INFO "%s: Starting DSP PAD Protocol Timeout period...\n",
			dsp_link->name));
	dsp_link->T2_timer.data     = (unsigned long)dsp_link;
	dsp_link->T2_timer.function = &dsp_t2_timer_expiry;
	dsp_link->T2_timer.expires  = jiffies + dsp_link->cfg.T2;

	del_timer(&dsp_link->T2_timer);
	add_timer(&dsp_link->T2_timer);
	set_bit(T2_RUNNING,(void*)&dsp_link->critical);
}

void dsp_stop_t2_timer(dsp_link_t *dsp_link)
{
	DSP_ASSERT1(dsp_link == NULL);
	
	if (!test_and_clear_bit(T2_RUNNING,(void*)&dsp_link->critical))
		return;
	del_timer(&dsp_link->T2_timer);
	DEBUG4((KERN_INFO "%s: Stopping DSP PAD Protocol Timeout preiod...\n",
			dsp_link->name));
}


/* ============================================================
 * RESTART TIMER
 * ===========================================================
 */

static void dsp_t3_timer_expiry(unsigned long dsp_id)
{
	dsp_link_t*	dsp_link = (dsp_link_t*)dsp_id;
	dspapi_hdr_t	hdr;
	unsigned char	next_sn = 0x00;

	DSP_ASSERT1(dsp_link == NULL);

	if (test_bit(CRIT_LINK_DOWN,&dsp_link->critical)){
		return;
	}
	
	/* Clear T3_RUNNING bit */
	if (!test_and_clear_bit(T3_RUNNING,(void*)&dsp_link->critical))
		return;

	/* Send OOB message t application: Waiting for ack.... */
	memset(&hdr, 0, sizeof(dspapi_hdr_t));
	hdr.dsp_pkt_type = DSP_ACK_MSG;
	dsp_chk_ack_sn(dsp_link, dsp_link->last_ack_sn_l, &next_sn, DSP_TRUE);
	hdr.dsp_sn = next_sn;
	dsp_snd_api_oob_msg(dsp_link, &hdr, NULL, 0);
	dsp_start_t3_timer(dsp_link);
	printk(KERN_INFO "%s: Waiting for M_EEACK(%i)...Start T3 timer again\n",
			dsp_link->name, next_sn);
}

void dsp_start_t3_timer(dsp_link_t *dsp_link)
{
	DSP_ASSERT1(dsp_link == NULL);
	if (dsp_link->cfg.T3 == 0)
		return;
	
	if (test_bit(CRIT_LINK_DOWN,&dsp_link->critical)){
		return;
	}
	
	if (test_bit(T3_RUNNING,(void*)&dsp_link->critical))
		return;
	
	DEBUG4((KERN_INFO "%s: Starting T3 Timeout period...\n",
			dsp_link->name));
	dsp_link->T3_timer.data     = (unsigned long)dsp_link;
	dsp_link->T3_timer.function = &dsp_t3_timer_expiry;
	dsp_link->T3_timer.expires  = jiffies + dsp_link->cfg.T3;
	
	del_timer(&dsp_link->T3_timer);
	add_timer(&dsp_link->T3_timer);
	set_bit(T3_RUNNING,(void*)&dsp_link->critical);
}

void dsp_stop_t3_timer(dsp_link_t *dsp_link)
{
	DSP_ASSERT1(dsp_link == NULL);
	if (!test_and_clear_bit(T3_RUNNING,(void*)&dsp_link->critical))
		return;
	del_timer(&dsp_link->T3_timer);
	DEBUG4((KERN_INFO "%s: Stopping T3 Timeout period...\n",
			dsp_link->name));
}

