/*
 *	LAPB release 002
 *
 *	This code REQUIRES 2.1.15 or higher/ NET3.038
 *
 *	This module:
 *		This module is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 */

#include <linux/wanpipe_includes.h>
#include <linux/wanpipe_cfg.h>
#include "wanpipe_lapb.h"


int lapb_trigger_kick(lapb_cb *lapb, char kick)
{
	if (test_bit(LAPB_UNREG_CRIT, &lapb->critical)){
		return 1;
	}

	if (test_bit(LAPB_AWAITING_KICK,&lapb->critical)){
		if (lapb->tx_dev && !netif_queue_stopped(lapb->tx_dev)){
			//printk(KERN_INFO "%s: Lapb waiting for open master device %s!\n",
			//		lapb->dev->name,lapb->tx_dev->name);
			clear_bit(LAPB_AWAITING_KICK,&lapb->critical);
		}
	}
	
	if (!test_bit(LAPB_AWAITING_KICK,&lapb->critical) &&
	    !test_and_set_bit(LAPB_BH_RUNNING,&lapb->critical)){
		if (kick){
			tasklet_schedule(&lapb->lapb_task);
		}else{
			lapb_kick((unsigned long)lapb);
		}
		return 0;
	}	

//	printk(KERN_INFO "%s: (Debug): Lap failed to trigger: %i\n",
//			lapb->dev->name,test_bit(LAPB_AWAITING_KICK,&lapb->critical));
	return 1;
}

void lapb_state_change(lapb_cb *lapb, int state)
{

	switch (state){
	
	case LAPB_STATE_CONNECTING:
		printk(KERN_INFO "%s: LapB Link connecting...\n",lapb->dev->name);
		lapb_stop_t4timer(lapb);
		break;

	case LAPB_STATE_CONNECTED:
		printk(KERN_INFO "%s: LapB Link connected\n",lapb->dev->name);
		lapb_start_t4timer(lapb, 1);
		break;

	case LAPB_STATE_DISCONNECTED:
		printk(KERN_INFO "%s: LapB Link disconnected!\n",lapb->dev->name);
		lapb_stop_t4timer(lapb);
		break;
		
	case LAPB_STATE_FRAME_REJECT:
		printk(KERN_INFO "%s: LapB Link frame reject!\n",lapb->dev->name);
		break;

	default:
		printk(KERN_INFO "%s: LapB Invalid link state %i!",
				lapb->dev->name,state);
		lapb_stop_t4timer(lapb);
		break;

	}

	lapb->state = state;
}

/*
 *	This routine purges all the queues of frames.
 */
inline void lapb_clear_queues(lapb_cb *lapb)
{
	struct sk_buff *skb;

	while ((skb = skb_dequeue(&lapb->write_queue)) != NULL)
		KFREE_SKB(skb);

	while ((skb = skb_dequeue(&lapb->ack_queue)) != NULL)
		KFREE_SKB(skb);

	while ((skb = skb_dequeue(&lapb->async_queue)) != NULL)
		KFREE_SKB(skb);

	while ((skb = skb_dequeue(&lapb->rx_queue)) != NULL)
		KFREE_SKB(skb);
}

/*
 * This routine purges the input queue of those frames that have been
 * acknowledged. This replaces the boxes labelled "V(a) <- N(r)" on the
 * SDL diagram.
 */
inline void lapb_frames_acked(lapb_cb *lapb, unsigned short nr)
{
	struct sk_buff *skb;
	volatile int modulus;

	modulus = (lapb->mode & LAPB_EXTENDED) ? LAPB_EMODULUS : LAPB_SMODULUS;

	/*
	 * Remove all the ack-ed frames from the ack queue.
	 */

	if (lapb->va != nr) {
		while (skb_peek(&lapb->ack_queue) != NULL && lapb->va != nr) {
		        skb = skb_dequeue(&lapb->ack_queue);
			lapb->va = (lapb->va + 1) % modulus;
			KFREE_SKB(skb);
		}
	}
}

void lapb_requeue_frames(lapb_cb *lapb)
{
        struct sk_buff *skb, *skb_prev = NULL;

	/*
	 * Requeue all the un-ack-ed frames on the output queue to be picked
	 * up by lapb_kick called from the timer. This arrangement handles the
	 * possibility of an empty output queue.
	 */

	set_bit(LAPB_REQUEUING,&lapb->critical);

	if (skb_queue_len(&lapb->ack_queue)){
		//printk(KERN_INFO "%s: Requeing %i non acked frames for Tx, new Vs=%i !\n",
		//	lapb->dev->name,skb_queue_len(&lapb->ack_queue),lapb->va);
	}

	spin_lock(&lapb->bh_lock);
	
	while ((skb = skb_dequeue(&lapb->ack_queue)) != NULL) {
		if (skb_prev == NULL){
			skb_queue_head(&lapb->write_queue, skb);
		}else{
			skb_append(skb_prev, skb);
		}
		skb_prev = skb;
	}
	//printk(KERN_INFO "%s: Requeue setting vs=%i back to va=%i\n",
	//		lapb->dev->name,lapb->vs,lapb->va);
	lapb->vs=lapb->va;
	
	spin_unlock(&lapb->bh_lock);

	clear_bit(LAPB_REQUEUING,&lapb->critical);
	
	lapb_trigger_kick(lapb,1);
}

/*
 *	Validate that the value of nr is between va and vs. Return true or
 *	false for testing.
 */
inline int lapb_validate_nr(lapb_cb *lapb, volatile unsigned short nr)
{
	volatile unsigned short vc = lapb->va;
	volatile int modulus;
	
	modulus = (lapb->mode & LAPB_EXTENDED) ? LAPB_EMODULUS : LAPB_SMODULUS;

	while (vc != lapb->vs) {
		if (nr == vc) {
			lapb->old_nr=nr;
			return 1;
		}
		vc = (vc + 1) % modulus;
	}
	
	if (nr == lapb->vs) {
		lapb->old_nr=nr;
		return 1;
	}
	printk(KERN_INFO "%s: Error : Rx Frame=%i Invalid NR=%i, PrevNR=%i PrevType=%i Vs=%i, Va=%i\n",
			lapb->dev->name, lapb->nr_type, nr, lapb->old_nr,
			lapb->old_nr_type,lapb->vs,lapb->va);
	
	return 0;
}

/*
 *	This routine is the centralised routine for parsing the control
 *	information for the different frame formats.
 */
void lapb_decode(lapb_cb *lapb, struct sk_buff *skb, struct lapb_frame *frame)
{
	frame->type = LAPB_ILLEGAL;

#if LAPB_DEBUG > 2
	printk(KERN_DEBUG "lapb: (%p) S%d RX %02X %02X %02X\n", lapb->dev, lapb->state, skb->data[0], skb->data[1], skb->data[2]);
#endif

	if (lapb->mode & LAPB_MLP) {
		if (lapb->mode & LAPB_DCE) {
			if (skb->data[0] == LAPB_ADDR_D){
				frame->cr = LAPB_COMMAND;
			}else if (skb->data[0] == LAPB_ADDR_C){
				frame->cr = LAPB_RESPONSE;
			}else{
				frame->type = LAPB_CORRUPT;
				return;
			}
		} else {
			if (skb->data[0] == LAPB_ADDR_C){
				frame->cr = LAPB_COMMAND;
			}else if (skb->data[0] == LAPB_ADDR_D){
				frame->cr = LAPB_RESPONSE;
			}else{
				frame->type = LAPB_CORRUPT;
				return;
			}
		}
	} else {
		if (lapb->mode & LAPB_DCE) {
			if (skb->data[0] == LAPB_ADDR_B){
				frame->cr = LAPB_COMMAND;
			}else if (skb->data[0] == LAPB_ADDR_A){
				frame->cr = LAPB_RESPONSE;
			}else{
				frame->type = LAPB_CORRUPT;
				return;
			}
		} else {
			if (skb->data[0] == LAPB_ADDR_A){
				frame->cr = LAPB_COMMAND;
			}else if (skb->data[0] == LAPB_ADDR_B){
				frame->cr = LAPB_RESPONSE;
			}else{
				frame->type = LAPB_CORRUPT;
				return;
			}
		}
	}
		
	skb_pull(skb, 1);

	if (lapb->mode & LAPB_EXTENDED) {
		if ((skb->data[0] & LAPB_S) == 0) {
			frame->type       = LAPB_I;			/* I frame - carries NR/NS/PF */
			frame->ns         = (skb->data[0] >> 1) & 0x7F;
			frame->nr         = (skb->data[1] >> 1) & 0x7F;
			frame->pf         = skb->data[1] & LAPB_EPF;
			frame->control[0] = skb->data[0];
			frame->control[1] = skb->data[1];
			skb_pull(skb, 2);
		} else if ((skb->data[0] & LAPB_U) == 1) { 	/* S frame - take out PF/NR */
			frame->type       = skb->data[0] & 0x0F;
			frame->nr         = (skb->data[1] >> 1) & 0x7F;
			frame->pf         = skb->data[1] & LAPB_EPF;
			frame->control[0] = skb->data[0];
			frame->control[1] = skb->data[1];
			skb_pull(skb, 2);
		} else if ((skb->data[0] & LAPB_U) == 3) { 	/* U frame - take out PF */
			frame->type       = skb->data[0] & ~LAPB_SPF;
			frame->pf         = skb->data[0] & LAPB_SPF;
			frame->control[0] = skb->data[0];
			frame->control[1] = 0x00;
			skb_pull(skb, 1);
		}
	} else {
		if ((skb->data[0] & LAPB_S) == 0) {
			frame->type = LAPB_I;			/* I frame - carries NR/NS/PF */
			frame->ns   = (skb->data[0] >> 1) & 0x07;
			frame->nr   = (skb->data[0] >> 5) & 0x07;
			frame->pf   = skb->data[0] & LAPB_SPF;
		} else if ((skb->data[0] & LAPB_U) == 1) { 	/* S frame - take out PF/NR */
			frame->type = skb->data[0] & 0x0F;
			frame->nr   = (skb->data[0] >> 5) & 0x07;
			frame->pf   = skb->data[0] & LAPB_SPF;
		} else if ((skb->data[0] & LAPB_U) == 3) { 	/* U frame - take out PF */
			frame->type = skb->data[0] & ~LAPB_SPF;
			frame->pf   = skb->data[0] & LAPB_SPF;
		}

		frame->control[0] = skb->data[0];

		skb_pull(skb, 1);
	}
}

/* 
 *	This routine is called when the HDLC layer internally  generates a
 *	command or  response  for  the remote machine ( eg. RR, UA etc. ). 
 *	Only supervisory or unnumbered frames are processed, FRMRs are handled
 *	by lapb_transmit_frmr below.
 */
void lapb_send_control(lapb_cb *lapb, int frametype, int poll_bit, int type)
{
	struct sk_buff *skb;
	unsigned char  *dptr;

	ALLOC_SKB(skb,LAPB_HEADER_LEN + 3);
	if (skb == NULL)
		return;
	
	skb_reserve(skb, LAPB_HEADER_LEN + 1);

	if (lapb->mode & LAPB_EXTENDED) {
		if ((frametype & LAPB_U) == LAPB_U) {
			dptr   = skb_put(skb, 1);
			*dptr  = frametype;
			*dptr |= (poll_bit) ? LAPB_SPF : 0;
		} else {
			dptr     = skb_put(skb, 2);
			dptr[0]  = frametype;
			dptr[1]  = (lapb->vr << 1);
			dptr[1] |= (poll_bit) ? LAPB_EPF : 0;
		}
	} else {
		dptr   = skb_put(skb, 1);
		*dptr  = frametype;
		*dptr |= (poll_bit) ? LAPB_SPF : 0;
		if ((frametype & LAPB_U) == LAPB_S)		/* S frames carry NR */
			*dptr |= (lapb->vr << 5);
	}

	lapb_update_stats (lapb, frametype, skb->len, 1);

	lapb_transmit_buffer(lapb, skb, type,1);
}

/* 
 *	This routine generates FRMRs based on information previously stored in
 *	the LAPB control block.
 */
void lapb_transmit_frmr(lapb_cb *lapb)
{
	struct sk_buff *skb;
	unsigned char  *dptr;

	ALLOC_SKB(skb,LAPB_HEADER_LEN + 7);
	if (skb == NULL)
		return;

	skb_reserve(skb, LAPB_HEADER_LEN + 1);

	if (lapb->mode & LAPB_EXTENDED) {
		dptr    = skb_put(skb, 6);
		*dptr++ = LAPB_FRMR;
		*dptr++ = lapb->frmr_data.control[0];
		*dptr++ = lapb->frmr_data.control[1];
		*dptr++ = (lapb->vs << 1) & 0xFE;
		*dptr   = (lapb->vr << 1) & 0xFE;
		if (lapb->frmr_data.cr == LAPB_RESPONSE)
			*dptr |= 0x01;
		dptr++;
		*dptr++ = lapb->frmr_type;

#if LAPB_DEBUG > 1
	printk(KERN_DEBUG "lapb: (%p) S%d TX FRMR %02X %02X %02X %02X %02X\n", lapb->dev, lapb->state, skb->data[1], skb->data[2], skb->data[3], skb->data[4], skb->data[5]);
#endif
	}else{
		dptr    = skb_put(skb, 4);
		*dptr++ = LAPB_FRMR;
		*dptr++ = lapb->frmr_data.control[0];
		*dptr   = (lapb->vs << 1) & 0x0E;
		*dptr  |= (lapb->vr << 5) & 0xE0;
		if (lapb->frmr_data.cr == LAPB_RESPONSE)
			*dptr |= 0x10;
		dptr++;
		*dptr++ = lapb->frmr_type;

#if LAPB_DEBUG > 1
	printk(KERN_DEBUG "lapb: (%p) S%d TX FRMR %02X %02X %02X\n", lapb->dev, lapb->state, skb->data[1], skb->data[2], skb->data[3]);
#endif
	}
	
	lapb_transmit_buffer(lapb, skb, LAPB_RESPONSE,1);
}

void lapb_update_stats (lapb_cb *lapb, int type, int len, char direction)
{
	switch (type){

		case LAPB_I:
			if (!direction)
				++lapb->lapbstats.ifrm_rx;
			else
				++lapb->lapbstats.ifrm_tx;	
			break;
		
		case LAPB_RR:
			if (!direction)
				++lapb->lapbstats.rr_rx;
			else
				++lapb->lapbstats.rr_tx;	
			break;
		
		case LAPB_RNR:
			if (!direction)
				++lapb->lapbstats.rnr_rx;
			else
				++lapb->lapbstats.rnr_tx;	
			break;

		case LAPB_REJ:
			if (!direction)
				++lapb->lapbstats.rej_rx;
			else
				++lapb->lapbstats.rej_tx;	
			break;
			
		case LAPB_SABME:
		case LAPB_SABM:
			if (!direction)
				++lapb->lapbstats.sabm_rx;
			else
				++lapb->lapbstats.sabm_tx;	
			break;

		case LAPB_DISC:
			if (!direction)
				++lapb->lapbstats.disc_rx;
			else
				++lapb->lapbstats.disc_tx;	
			break;
		
		case LAPB_DM:
			if (!direction)
				++lapb->lapbstats.dm_rx;
			else
				++lapb->lapbstats.dm_tx;	
			break;
		
		case LAPB_UA:
			if (!direction)
				++lapb->lapbstats.ua_rx;
			else
				++lapb->lapbstats.ua_tx;	
			break;

		case LAPB_FRMR:
			if (!direction)
				++lapb->lapbstats.frmr_rx;
			else
				++lapb->lapbstats.frmr_tx;	
			break;
	
		default:
			printk(KERN_INFO "%s: %s illegal packet type 0x%02X !\n",
					lapb->dev->name,direction ? "Tx" : "Rx" ,type);
			if (!direction)
				++lapb->lapbstats.ill_rx;
			else
				++lapb->lapbstats.ill_tx;	
			break;
	}

	if (!direction){
		++lapb->ifstats.rx_packets;
		lapb->ifstats.rx_bytes += len;
	}else{
		++lapb->ifstats.tx_packets;
		lapb->ifstats.tx_bytes += len;
	}

}

void lapb_reconnect(lapb_cb *lapb)
{
	lapb_stop_t1timer(lapb);
	lapb_start_t1timer(lapb);
}
