/*
 *	LAPB release 002
 *
 *	This code REQUIRES 2.1.15 or higher/ NET3.038
 *
 *	This module:
 *		This module is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 *	History
 */

#include <linux/wanpipe_includes.h>
#include <linux/wanpipe_cfg.h>
#include "wanpipe_lapb.h"

static void lapb_t1timer_expiry(unsigned long);
static void lapb_t2timer_expiry(unsigned long);
static void lapb_t4timer_expiry(unsigned long);

#undef DEL_TIMER_DEBUG
#ifdef DEL_TIMER_DEBUG
#define DEL_TIMER(x,func) {if (!del_timer(x)){ \
	                 	printk(KERN_INFO "%s: %s: %i Timer del failed\n", \
			     		__FILE__, func, __LINE__);}  \
			  }
#else
#define DEL_TIMER(x,func) del_timer(x)
#endif


/* X25's own task queue, used for all API's.
 * All protocol specific tasks will be instered
 * into "x25_tq_custom" task_queue. 

 * On each rx_interrupt, the whole task queue
 * (x25_tq_custom) will be queued into 
 * IMMEDIATE_BH via x25_mark_bh() call. 
 
 * The IMMEDIATE_BH will execute run_x25_tq() 
 * function, which will execute all pending,
 * tasks in x25_tq_custom queue */


#if 0
static void run_lapb_tq (unsigned long data)
{
	task_queue *tq_queue = (task_queue *)data;
	run_task_queue (tq_queue);
}

DECLARE_TASK_QUEUE(lapb_tq_custom);
static struct tq_struct lapb_tq_task = 
{
	routine: (void (*)(void *)) run_lapb_tq,
	data: &lapb_tq_custom
};


__inline__ int lapb_queue_tq (struct tq_struct *bh_pointer)
{
	if (!queue_task(bh_pointer,&lapb_tq_custom)){
		return 1;
	}
	return 0;
}

__inline__ int lapb_tq_mark_bh (void)
{
	if (!queue_task(&lapb_tq_task,&tq_immediate)){
		return 1;
	}
	mark_bh(IMMEDIATE_BH);
	return 0;
} 
#endif


void lapb_start_t1timer(lapb_cb *lapb)
{
	if (test_and_set_bit(LAPB_TIMER_CRIT, &lapb->critical)){
		//printk(KERN_INFO "%s: lapb_start_t1timer: %i, Timer Critical\n",__FILE__,__LINE__);
	}
	
	DEL_TIMER(&lapb->t1timer,"lapb_start_t1timer");

	lapb->t1timer.data     = (unsigned long)lapb;
	lapb->t1timer.function = &lapb_t1timer_expiry;
	lapb->t1timer.expires  = jiffies + lapb->t1;

	add_timer(&lapb->t1timer);
	
	clear_bit(LAPB_TIMER_CRIT, &lapb->critical);
}

void lapb_start_t2timer(lapb_cb *lapb)
{
	if (test_and_set_bit(LAPB_TIMER_CRIT, &lapb->critical)){
		//printk(KERN_INFO "%s: lapb_start_t2timer: %i, Timer Critical\n",__FILE__,__LINE__);
	}

	DEL_TIMER(&lapb->t2timer,"lapb_start_t2timer");

	lapb->t2timer.data     = (unsigned long)lapb;
	lapb->t2timer.function = &lapb_t2timer_expiry;
	lapb->t2timer.expires  = jiffies + lapb->t2;

	add_timer(&lapb->t2timer);
	clear_bit(LAPB_TIMER_CRIT, &lapb->critical);
}

void lapb_start_t4timer(lapb_cb *lapb, unsigned char reset_t4_cnt)
{
	/* The T4 timer is not enabled by the user */
	if (test_and_set_bit(LAPB_TIMER_CRIT, &lapb->critical)){
		//printk(KERN_INFO "%s: lapb_start_t2timer: %i, Timer Critical\n",__FILE__,__LINE__);
	}

	DEBUG_TEST("%s: Staring t4 timer\n",lapb->dev->name);
	
	DEL_TIMER(&lapb->t4timer,"lapb_start_t4timer");

	if (!lapb->t4){
		clear_bit(LAPB_TIMER_CRIT, &lapb->critical);
		return;
	}
	
	lapb->t4timer.data     = (unsigned long)lapb;
	lapb->t4timer.function = &lapb_t4timer_expiry;
	lapb->t4timer.expires  = jiffies + lapb->t4;

	if (reset_t4_cnt){
		lapb->n4count=0;
	}

	add_timer(&lapb->t4timer);
	clear_bit(LAPB_TIMER_CRIT, &lapb->critical);
}

void lapb_stop_t4timer(lapb_cb *lapb)
{
	DEL_TIMER(&lapb->t4timer,"lapb_stop_t4timer");
}

void lapb_stop_t1timer(lapb_cb *lapb)
{
	DEL_TIMER(&lapb->t1timer,"lapb_stop_t1timer");
}

void lapb_stop_t2timer(lapb_cb *lapb)
{
	DEL_TIMER(&lapb->t2timer,"lapb_stop_t2timer");
}

int lapb_t1timer_running(lapb_cb *lapb)
{
	return timer_pending(&lapb->t1timer);
}

static void lapb_t2timer_expiry(unsigned long param)
{
	lapb_cb *lapb = (lapb_cb *)param;

	if (test_bit(LAPB_UNREG_CRIT,&lapb->critical)){
		printk(KERN_INFO "Critical in T2 timer\n");
		return; 
	}
	
	if (lapb->condition & LAPB_ACK_PENDING_CONDITION) {
		lapb->condition &= ~LAPB_ACK_PENDING_CONDITION;
		lapb_timeout_response(lapb);
	}
}

/*=========================================================
 * lapb_t4timer_expiry
 *
 * 	This timer simulates both T4 and T3 Timers.
 * 	
 * Case1: T3 and T4 defined by user:
 * 
 * 	The T3 value would be calcualted as multiple of T4,
 * 	(ie. T3 =  T3/T4) thus if T4 restarts T3 timers the 
 * 	line will be disconneted.
 *
 * Case2: T3 not set, but T4 set by user:
 *
 * 	Not allowed:
 *
 * Case3: T3 is set by T4 is not:
 *
 *     In this case T4 will default to T3 value give by
 *     the user, and the T3 count will be 1.  
 *     The poll frames will not be sent out but the
 *     if the line is inactive for T3*T4 time, the
 *     line will be disconnected.
 *
 * Case4: T3 is not set and T4 is not set:
 *
 * 	This timer will not be used.
 */

static void lapb_t4timer_expiry(unsigned long param)
{
	lapb_cb *lapb = (lapb_cb *)param;

	if (lapb->t3 && ++lapb->n4count >= lapb->t3){
		printk(KERN_INFO "%s: T3 Timer Expiry: Line inactive, disconnecting ...\n",
					lapb->dev->name);
		lapb->n4count=0;
		lapb_clear_queues(lapb);
		lapb_state_change(lapb,LAPB_STATE_DISCONNECTED);
		lapb_disconnect_indication(lapb, LAPB_TIMEDOUT);

		lapb_reconnect(lapb);
		return;
	}


	if (!lapb->t4_orig){
		lapb_start_t4timer(lapb,0);
		return;
	}
	
	switch (lapb->state) {
		
		case LAPB_STATE_CONNECTED:
			if (++lapb->n2count > lapb->n2){

				printk(KERN_INFO "%s: LapB N2 (%d) polls exceeded with no response\n",
					lapb->dev->name,lapb->n2);

				lapb->n2count=0;
				lapb_clear_queues(lapb);
				lapb_state_change(lapb,LAPB_STATE_DISCONNECTED);
				lapb_disconnect_indication(lapb, LAPB_TIMEDOUT);

				lapb_reconnect(lapb);
				return;
			}
			
			if (skb_queue_len(&lapb->rx_queue) > LAPB_RX_Q_H_LIMIT){
				lapb_send_control(lapb, LAPB_RNR, LAPB_POLLON, LAPB_COMMAND);
			}else{
				lapb_send_control(lapb, LAPB_RR, LAPB_POLLON, LAPB_COMMAND);
			}

			break;

		case LAPB_STATE_FRAME_REJECT:
			lapb_transmit_frmr(lapb);
			break;

		default:
			printk(KERN_INFO "%s: T4 Timer Expiry: Invalid State 0x%02X\n",
					lapb->dev->name, lapb->state);

			lapb_clear_queues(lapb);
			lapb_state_change(lapb,LAPB_STATE_DISCONNECTED);
			lapb_disconnect_indication(lapb, LAPB_TIMEDOUT);

			lapb_reconnect(lapb);
			return;
	}

	lapb_start_t4timer(lapb,0);
}


static void lapb_t1timer_expiry(unsigned long param)
{
	lapb_cb *lapb = (lapb_cb *)param;

	//printk(KERN_INFO "%s: T1 Timer Expired \n",lapb->dev->name);
	
	if (test_bit(LAPB_UNREG_CRIT,&lapb->critical)){
		printk(KERN_INFO "Critical in T1 timer\n");
		return; 
	}

	switch (lapb->state) {

		/*
		 *	If we are a DCE, keep going DM .. DM .. DM
		 */
		case LAPB_STATE_DISCONNECTED:
		
			lapb_clear_queues(lapb);
			lapb_state_change(lapb,LAPB_STATE_CONNECTING);
				
		/*
		 *	Awaiting connection state, send SABM(E), up to N2 times.
		 */
		case LAPB_STATE_CONNECTING: 			

			if (lapb->n2count == lapb->n2) {

				lapb_send_control(lapb, LAPB_DM, LAPB_POLLOFF, LAPB_RESPONSE);
#if LAPB_DEBUG > 0
				printk(KERN_DEBUG "lapb: (%p) S1 -> TX DM(0)\n", lapb->dev);
#endif
				lapb->n2count=0;
			} else {
				lapb->n2count++;

				if (lapb->mode & LAPB_EXTENDED) {
#if LAPB_DEBUG > 1
					printk(KERN_DEBUG "lapb: (%p) S1 TX SABME(1)\n", lapb->dev);
#endif
					lapb_send_control(lapb, LAPB_SABME, LAPB_POLLON, LAPB_COMMAND);
				} else {
#if LAPB_DEBUG > 1
					printk(KERN_DEBUG "lapb: (%p) S1 TX SABM(1)\n", lapb->dev);
#endif
					lapb_send_control(lapb, LAPB_SABM, LAPB_POLLON, LAPB_COMMAND);
				}
			}
			break;

		/*
		 *	Awaiting disconnection state, send DISC, up to N2 times.
		 */
		case LAPB_STATE_DISCONNECTING:
			if (lapb->n2count == lapb->n2) {

				lapb->n2count=0;
				lapb_clear_queues(lapb);
				lapb_state_change(lapb,LAPB_STATE_DISCONNECTED);
				lapb_disconnect_confirmation(lapb, LAPB_TIMEDOUT);

#if LAPB_DEBUG > 0
				printk(KERN_DEBUG "lapb: (%p) S2 -> S0\n", lapb->dev);
#endif

				/* NC: Never turn of T1 timer always
				 * restart the link */
				//return;
			} else {
				lapb->n2count++;
#if LAPB_DEBUG > 1
				printk(KERN_DEBUG "lapb: (%p) S2 TX DISC(1)\n", lapb->dev);
#endif
				lapb_send_control(lapb, LAPB_DISC, LAPB_POLLON, LAPB_COMMAND);
			}
			break;

		/*
		 *	Data transfer state, restransmit I frames, up to N2 times.
		 */
		case LAPB_STATE_CONNECTED:
			if (lapb->n2count == lapb->n2) {

				lapb->n2count=0;
				lapb_clear_queues(lapb);
				lapb_state_change(lapb,LAPB_STATE_DISCONNECTED);
				lapb_stop_t2timer(lapb);
				lapb_disconnect_indication(lapb, LAPB_TIMEDOUT);

#if LAPB_DEBUG > 0
				printk(KERN_DEBUG "lapb: (%p) S3 -> S0\n", lapb->dev);
#endif
				/* NC: Never turn of T1 timer always
				 * restart the link */
				//return;
			}else{
				DEBUG_TEST("LAPB: T1 Expiry, Requeueing Frames!\n");
				lapb->n2count++;
				lapb_requeue_frames(lapb);
			}
			break;

		/*
		 *	Frame reject state, restransmit FRMR frames, up to N2 times.
		 */
		case LAPB_STATE_FRAME_REJECT:
			if (lapb->n2count == lapb->n2) {

				lapb->n2count=0;
				lapb_clear_queues(lapb);
				lapb_state_change(lapb,LAPB_STATE_DISCONNECTED);
				lapb_disconnect_indication(lapb, LAPB_TIMEDOUT);
#if LAPB_DEBUG > 0
				printk(KERN_DEBUG "lapb: (%p) S4 -> S0\n", lapb->dev);
#endif

				/* NC: Never turn of T1 timer always
				 * restart the link */
				//return;
			} else {
				lapb->n2count++;
				lapb_transmit_frmr(lapb);
			}
			break;
	}

	lapb_start_t1timer(lapb);
}

