/*
 *	X25 release 002
 *
 *	This code REQUIRES 2.1.15 or higher/ NET3.038
 *
 *	This module:
 *		This module is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 *	History
 */

#include <linux/wanpipe_includes.h>
#include <linux/wanpipe_cfg.h>
#include "wanpipe_x25.h"

extern struct x25_dsp_register_struct x25_dsp_protocol;
 
extern x25_lcn_t* dev_to_x25(struct net_device *dev);
// FIXME DSP_20 int lapb_enqueue_data(struct net_device *lapb_dev, struct sk_buff *skb);


int x25_dsp_register(struct net_device *x25_dev,
		     char *dev_name,
		     struct net_device **new_dev)
{
	x25_lcn_t *x25_lcn = NULL;
	unsigned long flags;
	int err;
	
	if (!IS_FUNC_CALL(x25_dsp_protocol, dsp_register)){
		printk(KERN_INFO "%s: X25 Protocol has is not supported!\n",
			x25_dev->name);
		return -EPROTONOSUPPORT;
	}
	
	if ((x25_lcn = dev_to_x25(x25_dev)) == NULL){
		printk(KERN_INFO "Error: X25 DSP Reg: No X25 struct on %s\n",
			x25_dev->name);
		return -ENODEV;
	}

	err=x25_dsp_protocol.dsp_register(x25_dev,dev_name,new_dev);
	if (*new_dev){
		spin_lock_irqsave(&x25_lcn->x25_link->bh_lock,flags);
		x25_lcn->dsp_id = *new_dev;
		spin_unlock_irqrestore(&x25_lcn->x25_link->bh_lock,flags);
	}

	return err;
}

int x25_dsp_unregister(void* dsp_id)
{
	/* We don't know what lapb device is the master
	 * for the x25_dev being supplied.  
	 *
	 * The x25_unregister function will 
	 * provide the lapb_dev.  Furthermore, if its return
	 * code is 0 then the x25_link has been unregistered.
	 * Therefore break the link to the x25.
	 */

	x25_lcn_t* x25_lcn = NULL;
	struct net_device *x25_dev=NULL;
	int dsp_link_status=0;

	/* Protocol not supported */
	if (!IS_FUNC_CALL(x25_dsp_protocol, dsp_unregister)){
		printk(KERN_INFO "X25: DSP Protocol has is not supported!\n");
		return  -EPROTONOSUPPORT;
	}

	if ((dsp_link_status = x25_dsp_protocol.dsp_unregister(&x25_dev, dsp_id)) < 0)
		return dsp_link_status;

	if (!x25_dev)
		return -EINVAL;
		
	if ((x25_lcn = dev_to_x25(x25_dev)) == NULL){
		printk(KERN_INFO "Error: X25 DSP Un Reg: No X25 struct on %s\n",x25_dev->name);
		return -EINVAL;
	}

	if (dsp_link_status == 0){
		if (x25_lcn->dsp_id){
			unsigned long flags;
			printk(KERN_INFO "%s: Warning: (x25_dsp_unregister) DSP unregister failed to remove dsp_id\n",
					x25_lcn->x25_link->name);
			spin_lock_irqsave(&x25_lcn->x25_link->bh_lock,flags);
			x25_lcn->dsp_id = NULL;
			spin_unlock_irqrestore(&x25_lcn->x25_link->bh_lock,flags);
		}
	}
	return 0;
}


int x25_dsp_setparms (struct net_device *x25_dev, 
			void* dsp_id, 
			void *dsp_parms)
{
	WAN_ASSERT2((!x25_dev),-ENODEV);
	WAN_ASSERT2((!dsp_id),-ENODEV);
	WAN_ASSERT2((!dsp_parms),-ENODEV);
			
	if (!IS_FUNC_CALL(x25_dsp_protocol, dsp_setparms)) {
		printk(KERN_INFO "%s: DSP Protocol has is not supported!\n",
				x25_dev->name);
		return -EPROTONOSUPPORT;
	}
	
	return x25_dsp_protocol.dsp_setparms(dsp_id, 
					     dsp_parms);
}


int send_dsp_call_request (x25_lcn_t *x25_lcn, struct x25_frame *frame,
				struct sk_buff *skb)
{
	unsigned char *buf;
	int len = skb->len;
	x25api_hdr_t *x25_api_hdr;
		
	if (x25_lcn->usedby != API || !x25_lcn->dsp_id ||!x25_lcn->x25_link->dsp_sk_id)
		return -ENODEV;
		
	if (!IS_FUNC_CALL(x25_dsp_protocol, dsp_call_request)) {
		printk(KERN_INFO "%s: DSP Protocol is not supported!\n",
				x25_lcn->dev->name);
		return -EPROTONOSUPPORT;
	}
	
	
	if (skb_headroom(skb) < sizeof(x25api_hdr_t)){
		printk(KERN_INFO "DSP Call request no skb room !\n");
		return -EINVAL;
	}
	
	buf = skb_push(skb,sizeof(x25api_hdr_t));
	if (!buf)
		return -ENOMEM;

	x25_api_hdr = (x25api_hdr_t *)buf;

	x25_api_hdr->qdm = (frame->gfi >> 1);
	if (frame->mbit){
		set_bit(0,(unsigned long*)&x25_api_hdr->qdm);
	}else{
		clear_bit(0,(unsigned long*)&x25_api_hdr->qdm);	
	}
	x25_api_hdr->cause = frame->cause;
	x25_api_hdr->diagn = frame->diag;
	x25_api_hdr->pktType = frame->type;
	x25_api_hdr->length  = len;
	x25_api_hdr->result = 0;
	x25_api_hdr->lcn = frame->lcn;

	skb->pkt_type = WAN_PACKET_ERR;
	skb->protocol=htons(DSP_PROT);
		
	return x25_dsp_protocol.dsp_call_request(x25_lcn->dsp_id,x25_lcn->x25_link->dsp_sk_id,skb);
}


int send_dsp_call_accept (x25_lcn_t *x25_lcn, struct x25_frame *frame,
				struct sk_buff *skb)
{
	unsigned char *buf;
	int len = skb->len;
	x25api_hdr_t *x25_api_hdr;
		
	if (x25_lcn->usedby != API || !x25_lcn->dsp_id)
		return -ENODEV;
		
	if (!IS_FUNC_CALL(x25_dsp_protocol, dsp_call_accept)) {
		printk(KERN_INFO "%s: DSP Protocol has is not supported!\n",
				x25_lcn->dev->name);
		return -EPROTONOSUPPORT;
	}
	
	
	if (skb_headroom(skb) < sizeof(x25api_hdr_t)){
		printk(KERN_INFO "DSP Call request no skb room !\n");
		return -EINVAL;
	}
	
	buf = skb_push(skb,sizeof(x25api_hdr_t));
	if (!buf)
		return -ENOMEM;

	x25_api_hdr = (x25api_hdr_t *)buf;

	x25_api_hdr->qdm = (frame->gfi >> 1);
	if (frame->mbit){
		set_bit(0,(unsigned long*)&x25_api_hdr->qdm);
	}else{
		clear_bit(0,(unsigned long*)&x25_api_hdr->qdm);	
	}
	x25_api_hdr->cause = frame->cause;
	x25_api_hdr->diagn = frame->diag;
	x25_api_hdr->pktType = frame->type;
	x25_api_hdr->length  = len;
	x25_api_hdr->result = 0;
	x25_api_hdr->lcn = frame->lcn;

	skb->pkt_type = WAN_PACKET_ERR;
	skb->protocol=htons(ETH_P_X25);
		
	return x25_dsp_protocol.dsp_call_accept(x25_lcn->dsp_id,skb);
}

void x25_update_dsp_state(x25_lcn_t *x25_lcn)
{

	/* If the LCN state changes from Connected to Disconnected, and
	 * we are in the API mode, then notify the socket that the 
	 * connection has been lost */

	/* If the state change due to reset, do not bring
	 * the DSP down, the dsp reset has already been reported */
	if (x25_lcn->state == X25_ST_RESET_REQ_REC_D3 || x25_lcn->state == X25_ST_RESET_REQ_SNT_D2){
		return;
	}
	
	if (x25_lcn->usedby == API && x25_lcn->dsp_id){

		if (x25_lcn->state != X25_ST_DATA_TRANSFER_P4 && test_bit(1,&x25_lcn->api_state)){
			clear_bit(1,&x25_lcn->api_state);

			if (!IS_FUNC_CALL(x25_dsp_protocol, dsp_svc_down)){
				return;
			}
			x25_dsp_protocol.dsp_svc_down(x25_lcn->dsp_id);

		}

			
		if (x25_lcn->state == X25_ST_DATA_TRANSFER_P4 && !test_bit(1,&x25_lcn->api_state)){
			set_bit(1,&x25_lcn->api_state);
	
			if (!IS_FUNC_CALL(x25_dsp_protocol, dsp_svc_up)){
				return;
			}
			x25_dsp_protocol.dsp_svc_up(x25_lcn->dsp_id);
			return;
		}
		
		if (x25_lcn->state == X25_ST_PKT_LVL_RDY_P1){
			if (!IS_FUNC_CALL(x25_dsp_protocol, dsp_svc_ready)){
				return;
			}
			x25_dsp_protocol.dsp_svc_ready(x25_lcn->dsp_id);
		}
	}
	return;
}

