#include <linux/wanpipe_includes.h>
#include <linux/wanpipe_cfg.h>
#include "wanpipe_x25.h"


/********************************************************************************/
/*************************** FACILITY CHECK *************************************/
/********************************************************************************/


/* CHECK THAT A FACILITY IS PERMITTED IN THE ASYNCHRONOUS PACKET RECEIVED ***/

int facil_permitted_in_Rx_pkt(x25_lcn_t *x25_lcn, unsigned char asy_record_type,unsigned char facility_code)
{

	x25_link_t *x25_link = x25_lcn->x25_link;
	
	/* This logic is taken from ISO 8208, 
	 * Tables 17 (pg 77) and Table 19 (pg 82) */

	if(facility_code != FACILITY_MARKER) {		
		
		/* a facility marker is permitted in all 
		 * packets in which facilities are permitted */

		switch(asy_record_type) {	
			
			/* switch according to the type of packet received */

			case CALL_REQUEST_PKT:							
				
				/* Call Request/Incoming Call */

				switch(facility_code) {

					case NUI_SELECTION_FACIL:
					case CHRG_INFO_REQ_SERVICE_FACIL:
					case RPOA_SELECT_BASIC_FORMAT_FACIL:
					case RPOA_SELECT_EXT_FORMAT_FACIL:
						if(x25_link->cfg.dte)
							/* if we are a DTE (the packet is an Incoming Call), 
							 * then these facilities are not permitted 
							 * in this packet */
							 return -EINVAL;		
						break;

					case CALL_REDIR_DEFLEC_NOTIF_FACIL:
						if (x25_link->cfg.dte)		
							/* if we are a DTE (the packet is an 
							 * Incoming Call), then this facility 
							 * is permitted */
							break;		

					case CHRG_INFO_IND_MON_UNIT_FACIL:
					case CHRG_INFO_IND_SEG_COUNT_FACIL:
					case CHRG_INFO_IND_CALL_DUR_FACIL:
					case CALL_DEFLECT_SELECTION_FACIL:
					case CALLED_LINE_ADDR_MOD_NOT_FACIL:
						/* these facilities are not permitted in a Call 
						 * Request or an Incoming Call packet */
						return -EINVAL;	
					default:
						/* all other facilities are permitted */
						break;
				}
				break;


			case CALL_ACCEPT_PKT:	
				
				/* Call Accepted/Call Connected */

				switch(facility_code) {

					case CHRG_INFO_REQ_SERVICE_FACIL:
					case NUI_SELECTION_FACIL:

						/* if we are a DTE (the packet is a 
						 * Call Connected), then these facilities 
						 * are not permitted in this packet*/

						if(x25_link->cfg.dte)	
							return -EINVAL;			
						
					case PKT_SIZE_NEGOTIATION_FACIL:
					case WIN_SIZE_NEGOTIATION_FACIL:
					case THROUGHPUT_CLASS_NEG_FACIL:
					case CALLED_LINE_ADDR_MOD_NOT_FACIL:
					case CCITT_CALLED_ADDR_EXT_FACIL:
					case CCITT_ETE_TX_DELAY_NEG_FACIL:
					case CCITT_PRIORITY_FACIL:
					case CCITT_PROTECTION_FACIL:
					case CCITT_EXPIDITED_DATA_NEG_FACIL:

						/* these facilities are permitted */
						break;		

					case TRANSIT_DELAY_SELECT_IND_FACIL:

						/* if we are a DTE (the packet is a Call Connected), 
						 * then this facility is permitted */
						if(x25_link->cfg.dte)	
							break;			

					default:
						/* all other facilities are not permitted */
						return -EINVAL;
				}
				break;


			case CLEAR_REQUEST_PKT:		
				
				/* Clear Request/Indication */

				switch(facility_code) {

					case CHRG_INFO_IND_MON_UNIT_FACIL:
					case CHRG_INFO_IND_SEG_COUNT_FACIL:
					case CHRG_INFO_IND_CALL_DUR_FACIL:
						if(!x25_link->cfg.dte)	
							/* if we are a DCE (the packet is a 
							 * Clear Request), then these facilities are not permitted */
							return -EINVAL;		
						break;

					case CCITT_CALLING_ADDR_EXT_FACIL:
					case CCITT_MIN_THRU_CLS_NEG_FACIL:
					case CCITT_ETE_TX_DELAY_NEG_FACIL:
					case CCITT_PRIORITY_FACIL:
					case CCITT_PROTECTION_FACIL:
					case CCITT_EXPIDITED_DATA_NEG_FACIL:
					case CALL_DEFLECT_SELECTION_FACIL:
						
						if(x25_link->cfg.dte)
							/* if we are a DTE (the packet is a 
							 * Clear Indication), then these 
							 * facilities are not permitted */
							return -EINVAL;		
						break;

					case CALLED_LINE_ADDR_MOD_NOT_FACIL:
					case CCITT_CALLED_ADDR_EXT_FACIL:
						/* these facilities are permitted */	
						break;	

					default:	
						/* all other facilities are not permitted */
						return -EINVAL;
				}
				break;


			case CLEAR_CONF_PKT:		
				
				/* Clear Confirmation */

				switch(facility_code) {

					case CHRG_INFO_IND_MON_UNIT_FACIL:
					case CHRG_INFO_IND_SEG_COUNT_FACIL:
					case CHRG_INFO_IND_CALL_DUR_FACIL:
						
						/* these facilities are permitted */
						break;						

					default:
						/* all other facilities are not permitted */
						return -EINVAL;
				}
				break;


			default:
				break;

		}
	}

	/* the facility is permitted in this packet */
	return X25_OK;				
}



/* CHECK THE VALIDITY OF THE WINDOW SIZE PARAMETER IN A FLOW CONTROL PARAMETER NEGOTIATION FACILITY FIELD */

int chk_win_size_in_facils(x25_lcn_t *x25_lcn,
		           unsigned int facils_parm_limit_type, 
			   unsigned int ut_count,
			   unsigned char util_char)
{
	unsigned int min_facils_parm_value,max_facils_parm_value;
	unsigned char facils_char;
	
	/* extract our current internal window size value for this LCN structure */
	if (!ut_count){
		facils_char = x25_lcn->rx_window_size; 
	}else{
		facils_char = x25_lcn->tx_window_size;
	}
	/* set the default minimum window size (1) */
	min_facils_parm_value = MIN_WIN_SIZE_IN_NEG_FACILS;

	/* set the default maximum window size (7) */
	max_facils_parm_value = MAX_WIN_SIZE_IN_NEG_FACILS;

	/* The logic for checking the validity of the window 
	 * size negotiation facilities in a Call Accepted/Connected 
	 * packet is taken from ISO 8208, Tables 11 and 12 (pg 60) and is as
	 * follows: 
	 *
	 * window size in Call Request permitted negototiation windowpacket size
	 * W(Call Req) = 1	 W(Call Acc) = 1 or W(Call Acc) = 2
	 * W(Call Req) >= 2 	 W(Call Req) >= W(Call Acc) >= 2
	 */
	
	if(facils_parm_limit_type == CCITT_PKT_WIN_LIMIT) {
		
		/* the current window size is 1 */
		if (facils_char == MIN_WIN_SIZE_IN_NEG_FACILS){
		
			/* the maximum permitted window size is 2 
			 * (minimum = 1, as set above) */
			max_facils_parm_value = WIN_SIZE_OF_2;
		}else{		
			/* the current window size is greater than or equal to 2 */
			/* the minimum permitted window size is 2 */
			min_facils_parm_value = WIN_SIZE_OF_2;
			/* the maximum permitted window size is our current window size */
			max_facils_parm_value = facils_char;
		}
	}

	/* check to see that the negotiated window size is within the defined limits */
	if ((util_char < min_facils_parm_value) || (util_char > max_facils_parm_value)){
		/* we have an error in the negotiated window size */
		return -EINVAL;
	}
	
	/* update the window size for this LCN */
	if (!ut_count){
#ifdef LAPB_DEBUG
		printk(KERN_INFO "%s: Setting RX WINDOW SIZE TO %i\n",x25_lcn->dev->name,util_char);
#endif
		x25_lcn->rx_window_size = util_char;
	}else{
#ifdef LAPB_DEBUG
		printk(KERN_INFO "%s: Setting TX WINDOW SIZE TO %i\n",x25_lcn->dev->name,util_char);
#endif
		x25_lcn->tx_window_size = util_char;
	}
	/* the negotiated packet size is acceptable */
	return X25_OK;				
}



/* CHECK THE VALIDITY OF THE PACKET SIZE PARAMETER IN A FLOW CONTROL PARAMETER NEGOTIATION FACILITY FIELD */

int chk_pkt_size_in_facils(x25_lcn_t *x25_lcn,
			   unsigned char Rx_pkt_type,
		           unsigned int facils_parm_limit_type, 
			   unsigned int ut_count,
			   unsigned char util_char)
{
	unsigned int min_facils_parm_value,max_facils_parm_value;
	unsigned char facils_char;
	
	/* extract our current internal packet size 
	 * value for this LCN structure */
	if (!ut_count){
		facils_char = pkt_size_no(x25_lcn->rx_pkt_size);
	}else{
		facils_char = pkt_size_no(x25_lcn->tx_pkt_size);
	}
	
	/* The logic for checking the validity of the packet 
	 * size negotiation facilities in a Call Accepted/Connected
	 * packet is taken from ISO 8208, Tables 11 and 12 
	 * (pg 60) and is as	follows:
	 *
	 * packet size in Call Request	 permitted negototiation packet size
	 * P(Call Req) < 128		 P(Call Req) <= P(Call Acc) <= 128
	 * P(Call Req) >= 128		 P(Call Req) >= P(Call Acc) >= 128
	 */
	
	if(facils_parm_limit_type == CCITT_PKT_WIN_LIMIT) {
	
		/* the current packet size is less than 128 bytes */
		if(facils_char < PKT_SIZE_ISO_FORMAT_128_BYTES) {
		
			/* the minimum permitted packet size is our current packet size */
			min_facils_parm_value = facils_char;
			/* the maximum permitted packet size is 128 bytes */
			max_facils_parm_value = PKT_SIZE_ISO_FORMAT_128_BYTES;
		
		}else {	
			/* the current packet size is greater than or equal to 128 bytes */
			/* the minimum permitted packet size is 128 bytes */
			
			min_facils_parm_value = PKT_SIZE_ISO_FORMAT_128_BYTES;   
			
			/* CCITT 128 bytes */
			/* the maximum permitted packet size is our current packet size */
			max_facils_parm_value = facils_char;
		}
	
	}else {
		/* we are checking the packet sizes in a Call Request packet 
		 * the minimum permitted packet size is 16 bytes */
		min_facils_parm_value = PKT_SIZE_ISO_FORMAT_16_BYTES;
		
		/* the maximum permitted packet size is dependant on our configured 
		 * 'maximum_packet_size' parameter */
		max_facils_parm_value = pkt_size_no(x25_lcn->x25_link->cfg.maximum_packet_size);
		
		/* if we are processing an incoming call request packet and the 
		 * requested packet size is greater than our maximum packet 
		 * size, then we negotiate the packet size downwards to our 
		 * defined maximum size */
		if((Rx_pkt_type == CALL_REQUEST_PKT) && (util_char > max_facils_parm_value))
			util_char = max_facils_parm_value;
	}

	/* check to see that the negotiated packet size is within the defined limits */
	if((util_char < min_facils_parm_value) || (util_char > max_facils_parm_value))
		return -EINVAL;	
	
	/* we have an error in the negotiated packet size */
	/* update the packet size for this LCN */

	if (decode_pkt_size(&util_char) == X25_OK){
		if (!ut_count){
#ifdef LAPB_DEBUG
			printk(KERN_INFO "%s: Setting RX PACKET SIZE TO %i\n",x25_lcn->dev->name,util_char);
#endif
			x25_lcn->rx_pkt_size = util_char;
		}else{
#ifdef LAPB_DEBUG
			printk(KERN_INFO "%s: Setting TX PACKET SIZE TO %i\n",x25_lcn->dev->name,util_char);
#endif
			x25_lcn->tx_pkt_size = util_char;
		}
	}else{
		printk(KERN_INFO "%s: Invalid Packet Size in negotiation!\n",x25_lcn->dev->name);
	}

	
	/* the negotiated packet size is acceptable */
	return X25_OK;				
}


/* CHECK THAT THE PACKET OR WINDOW SIZES INCLUDED IN THE FACILITIES FIELD ARE VALID */

int win_pkt_size_in_Rx_pkt_OK(x25_lcn_t *x25_lcn, 
		              unsigned char *util_ptr, 
			      unsigned char Rx_pkt_type)
{
	unsigned char util_char;
	unsigned int facils_parm_limit_type;
	int ut_count=ZERO;
	
	if (Rx_pkt_type == CALL_REQUEST_PKT) {	
		
		/* for a Call Request packet, we swap the called 
		 * and calling packet/window size parameters 
		 * before checking them */
		
		util_char = *(util_ptr + ONE_BYTE);			
		*(util_ptr + ONE_BYTE) = *(util_ptr + TWO_BYTES);
		*(util_ptr + TWO_BYTES) = util_char;
		facils_parm_limit_type = DEFAULT_PARM_LIMIT;	
		
		/* the window/packet sizes must be limited by our configuration 
		 * file (and basic ISO requirements)*/
	}else {	
		/* for a Call Accept, check the packet and 
		 * window sizes against CCITT limits defined 
		 * in Tables 11 and 12 of ISO 8208 */
		facils_parm_limit_type = CCITT_PKT_WIN_LIMIT;
	}
	
	do { 
		/* read the next byte in the facility parameter field */
		util_char = *(util_ptr + ONE_BYTE + ut_count);

		/* if the facility is for packet size negotiation, 
		 * then we check the validity of the included 
		 * packet size */
		
		if(*util_ptr == PKT_SIZE_NEGOTIATION_FACIL) {
			if(chk_pkt_size_in_facils(x25_lcn,Rx_pkt_type,
						  facils_parm_limit_type,ut_count,util_char) != X25_OK)
				/* the packet size was invalid */
				return -EINVAL;
		}

		/* if the facility is for window size negotiation, then we check 
		 * the validity of the included window size */
		
		else if (chk_win_size_in_facils(x25_lcn,facils_parm_limit_type,
					 	 ut_count,util_char) != X25_OK){		
			return -EINVAL;		
		}

		/* process the complete Flow Control Parameter 
		 * Negotiation facility field */
	} while ((++ ut_count) < LGTH_FLOW_CTRL_FACIL_PARM_FLD);

	/* the Flow Control Parameter Negotiation facility field is OK */
	return X25_OK;
}

/****** CHECK FOR DUPLICATE FACILITIES IN AN INCOMING ASYNCHRONOUS PACKET *****/

int duplicate_facil_req_async_pkt(x25_lcn_t *x25_lcn, 
		                  unsigned char facility_code)
{
	int l;
	x25_link_t* x25_link = x25_lcn->x25_link;
	
	/* check all received facility codes for a match */
	for(l = ZERO; l < x25_link->no_diff_facilty_codes_received; l++) {
		if(x25_link->list_received_facilty_codes[l] == facility_code)
			/* a facility code match has been found, 
			 * so we have a duplicate */
			return -EINVAL;				
	}

	if (x25_link->no_diff_facilty_codes_received >= MAX_RX_FACIL_CODE){
		printk(KERN_INFO "%s: Error MAX array count reached for 'no_diff_facilty_codes_received'\n",
				x25_lcn->dev->name);
		return -EINVAL;
	}
	
	/* store the current facility code and update the 
	 * number of different facilities received 
	 * in this asynch packet */
	x25_link->list_received_facilty_codes[x25_link->no_diff_facilty_codes_received ++] = facility_code;

	/* no facility duplication was found */
	return X25_OK;							
}




/* CHECK FOR ERRORS WHEN DECODING THE FACILITIES 
 * FIELD OF AN INCOMING PACKET */

int error_in_decoding_facils(x25_lcn_t *x25_lcn, 
			     unsigned char type,        
			     unsigned char *X25_Rx_pkt_Data_ptr,
			     int facils_lgth,
			     int off_facils)
{
	x25_link_t *x25_link = x25_lcn->x25_link;
	int i=0,j=0;
	unsigned char *util_ptr;
	unsigned char facils_char, facility_code;
	
	/* if we are configured to not check and process incoming 
	 * facilities, then we are done no error was found in the received facilities */
	if(x25_link->cfg.X25_protocol_options & NO_CHK_PROC_INCOMING_FACILS)
		return X25_OK;
	
	/* reset the number of different facility codes received in 
	 * this packet and reset the flag indicating that non */
	/* X25 facilities have been received */
	x25_link->no_diff_facilty_codes_received = x25_link->non_X25_facilities_received = ZERO;

	/* for this LCN, in case of packet/window negotiation facilities 
	 * recover the actual length of the facilities field - note that
	 * we must compensate for having multiplied the length by 2 in */

	i = facils_lgth / TWO_BYTES;				
	
	/* our string conversion routine set a pointer to the first 
	 * byte of the facilities field in the actual received packet - 
	 * note that we must compensate for the facilities offset having 
	 * been multiplied by 2, and we add one byte for the facilities 
	 * length field */
	
	util_ptr = X25_Rx_pkt_Data_ptr + (off_facils / TWO_BYTES  + ONE_BYTE);

	while (i) {
		
		/* process the complete facility field 
		 * we have found a facilty marker field */

		if(*util_ptr == FACILITY_MARKER) {

			/* read and process the facility parameter field */	
			switch(*(util_ptr + ONE_BYTE)) {
	
				/* the X.25 - 1980 specifications precluded the use of CCITT */
				case CCITT_SPECIFIED_DTE_FACILS:			
	
					/* specified facilities (IS) 8208, 3.1.2 i), pg 4 */
					if(x25_link->cfg.CCITT_compatibility == CCITT_1980_COMPATIBILITY)
						return (FACIL_REGSTR_CODE_NOT_ALLOWED);

				/* these two non-X.25 facility categories are valid */
				case NON_X25_FACILS_LOCAL_NETWORK:		
				case NON_X25_FACILS_REMOTE_NETWORK:
					
					/* indicate that non-X25 facilities have been received 
					 * reset the number of different facility codes used - 
					 * this must be done as, for example, we may receive 
					 * the CCITT facility 0101 (reverse charging) and 
					 * then the non-CCITT facility 0101 (tariff class. In 
					 * such a case, we must not clear the call with diagnostic 
					 * code 'duplicate facility requested'. */

					x25_link->non_X25_facilities_received = 1;
					x25_link->no_diff_facilty_codes_received = ZERO;
					break;

				default:
					/* any other facility marker is invalid, 
					 * so return the error */
					return FACIL_REGSTR_CODE_NOT_ALLOWED;
			}

			/* adjust the facility pointer to examine the facilities field */
			util_ptr += LGTH_FACILITY_MARKER_FIELD;	
			
			/* adjust the number of characters processed */
			i -= LGTH_FACILITY_MARKER_FIELD;		
		}

		/* read the facility code */
		facility_code = *util_ptr;

		/* read the first byte of the facility parameter field */
		facils_char = *(util_ptr + ONE_BYTE);			

		/* check that this facility is indeed permitted in this type of
		 *  packet - if not, then return an error code */
#ifdef LAPB_DECODE
		printk(KERN_INFO "%s: Facility Code %x, Facility Char %x\n",
				x25_lcn->dev->name,facility_code,facils_char);
#endif
		if(facil_permitted_in_Rx_pkt(x25_lcn,type,facility_code) != X25_OK) 				
			return(PKT_TYPE_NOT_COMPAT_WITH_FACIL);	

		/* use the two high bits of the facility code to determine 
		 * the format of the facility element */
		switch (facility_code & MASK_2_HIGH_BITS_OF_FACILS_CODE) {

			/* the facility element is 2 bytes in length */
			case 0x00:			
				
				/* handle the different facility codes */
				switch(facility_code) {	

					case REV_CHRG_FAST_SELECT_FACIL:

						/* bits 1 to 5 are reserved in this field 
						 * (ISO 8208, 15.2.2.6, pg 79), so if they are set, then
						 * we have an error */
						
						if (facils_char & BTS_1_5_SET_REV_CHR_FST_SEL)
							return(FACILITY_PARM_NOT_ALLOWED);

						/* if Reverse Charging is included, then we must 
						 * check that we are configured to handle this 
						 * facility */
						if(facils_char & REV_CHARGING_REQUESTED) {
							if(!(x25_link->cfg.genl_facilities_supported_1 & 
										REV_CHARGING_SUPP))
								return(FACIL_REGSTR_CODE_NOT_ALLOWED);
							
							/* change the value of the 'facility_code' so that 
							 * if we have '-f01010108', then 
							 * 'duplicate_facil_req_async_pkt()' will not 
							 * return an error */
							facility_code = INTERNAL_FACILS_REP_REV_CHRG;
						}
				
						/* if Fast Select is included, then we must check that 
						 * we are configured to handle this facility - if not, 
						 * then we issue a Clear Request with a diagnostic 
						 * code #164 (DTE) or #65 (DCE) (C3-D1, Note 5) */
						if((facils_char & FAST_SEL_REQUESTED) && 
						   !(x25_link->cfg.genl_facilities_supported_1 & FAST_SELECT_SUPP)){
							return ((x25_link->cfg.dte) ? 
								FAST_SELECT_NOT_SUBSCRIBED : 
								FACIL_REGSTR_CODE_NOT_ALLOWED);
						}
						break;

					case THROUGHPUT_CLASS_NEG_FACIL: 
						
						/* if Throughput Class Negotiation is included, 
						 * then we must check that we are configured to 
						 * handle this facility */
						if (!(x25_link->cfg.genl_facilities_supported_1 & 
									THROUGHPUT_CLASS_NEG_SUPP))
							return(FACIL_REGSTR_CODE_NOT_ALLOWED);

						/* only check the validity of this facility if 
						 * it appears before a non-X25 facilities marker, so 
						 * as to handle the Sprint network */
						if (!x25_link->non_X25_facilities_received) {
							
							/* the valid Throughput Class parameters are as follows:
							 * the minimum throughput class is 75 bps
							 * the maximum throughput class for X.25 - 1984 is 48000 bps
							 * the maximum throughput class for X.25 - 1988 is 64000 bps
							 */
							
							if ((facils_char & HANDLE_LOW_BITS_OF_THRU_CLASS) < 
									THROUGHPUT_CLS_75_LOW_BITS) 
								return FACILITY_PARM_NOT_ALLOWED;
		
	     
							if ((facils_char & HANDLE_LOW_BITS_OF_THRU_CLASS) > 
									((x25_link->cfg.CCITT_compatibility == CCITT_1988_COMPATIBILITY)?
									THROUGHPUT_CLS_64K_LOW_BITS : 
									THROUGHPUT_CLS_48K_LOW_BITS))
								return FACILITY_PARM_NOT_ALLOWED;

							if ((facils_char & HANDLE_HIGH_BITS_OF_THRU_CLASS) < 
									THROUGHPUT_CLS_75_HIGH_BITS) 
								return FACILITY_PARM_NOT_ALLOWED;

							if ((facils_char & HANDLE_HIGH_BITS_OF_THRU_CLASS) > 
									((x25_link->cfg.CCITT_compatibility == CCITT_1988_COMPATIBILITY)? 
									THROUGHPUT_CLS_64K_HIGH_BITS : 
									THROUGHPUT_CLS_48K_HIGH_BITS));
								return FACILITY_PARM_NOT_ALLOWED;
						}
   							   
   						break;

					case CUG_SELECT_BASIC_FACIL:
					case CUG_OUT_SELECT_BASIC_FACIL:
						
						/* if Closed User Group Selection is included, 
						 * then we must check that we are configured to 
						 * handle this facility */
						if (!(x25_link->cfg.genl_facilities_supported_1 & 
						    (CUG_SELECT_BASIC_FACIL | CUG_OUT_SELECT_BASIC_FACIL)))
							return(FACIL_REGSTR_CODE_NOT_ALLOWED);
						break;

					default:
						/* we have an error, unless we must handle all facilities */
						if(!(x25_link->cfg.X25_protocol_options & HANDLE_ALL_INCOMING_FACILS)){
							//????????????????
							printk(KERN_INFO "%s: Error: Facils Not Allowed by User!\n",
									x25_lcn->dev->name);
							return(FACIL_REGSTR_CODE_NOT_ALLOWED);
						}
				}

				/* record the length of this facility element */		
				j = 2;		
				break;


			case 0x40:
				/* the facility element is 3 bytes in length */

				switch(facility_code) {	
					
					/* handle the different facility codes */

					case PKT_SIZE_NEGOTIATION_FACIL:
					case WIN_SIZE_NEGOTIATION_FACIL: 
						
						/* if Flow Control Parameter Negotiation is included, 
						 * then we must check that we are configured to 
						 * handle this facility */
						
						if (!(x25_link->cfg.genl_facilities_supported_1 & 
						      FLOW_CTRL_PARM_NEG_SUPP)){
							printk(KERN_INFO "%s: Packet and Window facils not supported by user!\n",x25_lcn->dev->name);
							return(FACIL_REGSTR_CODE_NOT_ALLOWED);
						
						}
						/* check that the packet or window sizes 
						 * included in the facilities field are valid */
						
						if (win_pkt_size_in_Rx_pkt_OK(x25_lcn,util_ptr,type) != X25_OK)
							return(FACILITY_PARM_NOT_ALLOWED);


						/* Indicate to the replying code that the 
						 * packet window facilites were negotiated */
						if (type == CALL_REQUEST_PKT)
							x25_lcn->config |= INCOMING_PKT_WIN_FACILS;
						break;

					case TRANSIT_DELAY_SELECT_IND_FACIL:
						
						/* if Transit Delay Selection and Indication 
						 * is included, then we must check that we are 
						 * configured to handle this facility */
						if (!(x25_link->cfg.genl_facilities_supported_1 & 
						     TRANSIT_DELAY_SELECT_IND_SUPP))
							return(FACIL_REGSTR_CODE_NOT_ALLOWED);
						
						break;

					case CUG_SELECT_EXT_FACIL:
					case CUG_OUT_SELECT_EXT_FACIL:
					case BI_CUG_SELECT_FACIL:
						
						/* if Closed User Group Selection is included, 
						 * then we must check that we are configured to 
						 * handle this facility */
						if (!(x25_link->cfg.genl_facilities_supported_1 & 
						       (CUG_SELECT_EXT_SUPP |
							CUG_OUT_ACC_SEL_EXT_SUPP | 
							BI_CUG_SEL_SUPP)))
							return(FACIL_REGSTR_CODE_NOT_ALLOWED);
						
						break;

					default:
						/* we have an error, unless we must handle all facilities */
						if (!(x25_link->cfg.X25_protocol_options & HANDLE_ALL_INCOMING_FACILS))
							return (FACIL_REGSTR_CODE_NOT_ALLOWED);

				}
				
				/* record the length of this facility element */
				j = 3;			
				break;


			case 0x80:
				
				/* the facility element is 4 bytes in length 
				 * handle the different facility codes */

				switch(facility_code) {			
					
					default:	
						/* we have an error, unless we must handle all facilities */
						if(!(x25_link->cfg.X25_protocol_options & HANDLE_ALL_INCOMING_FACILS))
							return (FACIL_REGSTR_CODE_NOT_ALLOWED);
				}
				
				j = 4;	
				/* record the length of this facility element */
				break;


			case 0xC0:
				/* the facility element is of variable length handle 
				 * the different facility codes */

				switch(facility_code) {				
					
					case NUI_SELECTION_FACIL:
						/* if NUI Selection is included, then we 
						 * must check that we are configured to handle this
						 * facility */
						if(!(x25_link->cfg.genl_facilities_supported_1 & NUI_SELECTION_SUPP))
							return (FACIL_REGSTR_CODE_NOT_ALLOWED);
						break;

					case CCITT_CALLING_ADDR_EXT_FACIL:
					case CCITT_CALLED_ADDR_EXT_FACIL:
						/* if CCITT Calling/Called Address Extension is 
						 * included, then we must check that we are 
						 * configured to handle this facility */
						if(!(x25_link->cfg.CCITT_facilities_supported & 
							(CCITT_CALLING_ADDR_EXT_SUPP |
							CCITT_CALLED_ADDR_EXT_SUPP)))
							return (FACIL_REGSTR_CODE_NOT_ALLOWED);
						break;

					default:			
						/* we have an error, unless we must handle all facilities */
						if(!(x25_link->cfg.X25_protocol_options & HANDLE_ALL_INCOMING_FACILS))
							return (FACIL_REGSTR_CODE_NOT_ALLOWED);

				}
				
				/* calculate and record the length of this facility element */
				j = (*(util_ptr + ONE_BYTE) + TWO_BYTES);
				break;
		}

		/* point to the next facility element to be processed */
		util_ptr += j;	

		if (i < j){	
			/* check that we have not gone past the end of the 
			 * facility field in the case of invalid included facilities */
			return (FACIL_REGSTR_CODE_NOT_ALLOWED);		
		}
		
		if (duplicate_facil_req_async_pkt(x25_lcn,facility_code) != X25_OK){
			/* check for duplicate facilities in the received packet */
			return (DUPLICATE_FACILITY_REQUESTED);
		}
		
		/* adjust the total number of characters processed in the */
		i -= j;
	}
		
	/* facility field no error was found in the received facilies */
	return X25_OK;
}


/** INSERT FLOW CONTROL FACILITIES IN AN OUTGOING CALL REQUEST/ACCEPT PACKET **/

void insert_flow_control_facils(x25_lcn_t *x25_lcn, unsigned char type, struct sk_buff *skb)
{
	unsigned char *dest_ptr;

	/* insert the '6' into the argument formatting string to follow
	 * the already inserted '0' - we now indicate a facilities length 
	 * of '0', '6' for Flow Control Negotiation */
	dest_ptr = skb_put(skb,1);
	*dest_ptr = 6;
	
	dest_ptr = skb_put(skb,1);
	*dest_ptr = 0x43;

	dest_ptr = skb_put(skb,2);
	dest_ptr[0] = (type == CALL_REQUEST_PKT) ? x25_lcn->rx_window_size : x25_lcn->tx_window_size;
	dest_ptr[1] = (type == CALL_REQUEST_PKT) ? x25_lcn->tx_window_size : x25_lcn->rx_window_size;
	
	dest_ptr = skb_put(skb,1);
	*dest_ptr = 0x42;
	
	dest_ptr = skb_put(skb,2);
	dest_ptr[0] = (type == CALL_REQUEST_PKT) ? 
			pkt_size_no(x25_lcn->rx_pkt_size) : pkt_size_no(x25_lcn->tx_pkt_size);
	dest_ptr[1] = (type == CALL_REQUEST_PKT) ? 
		        pkt_size_no(x25_lcn->tx_pkt_size) : pkt_size_no(x25_lcn->rx_pkt_size);

}




