#include <linux/wanpipe_includes.h>
#include <linux/wanpipe_cfg.h>
#include "wanpipe_x25.h"


extern struct x25_dsp_register_struct x25_dsp_protocol;

/* 
 *  This procedure is passed a buffer descriptor for an iframe. It builds
 *  the rest of the control part of the frame and then writes it out.
 */
static int x25_send_data_frame(x25_lcn_t *x25, struct sk_buff *skb)
{
	unsigned char *frame;
	unsigned char qd_bits=0, m_bit=0;
	int err=0;

	if (skb == NULL)
		return -EINVAL;
	
	if (x25->usedby == API){
		/* Remove the API Header */
		x25api_hdr_t *api_hdr = (x25api_hdr_t *)skb->data;

		if (skb->len <= sizeof(x25api_hdr_t)){
			printk(KERN_INFO "Send Data: Invalid Api Pkt length %i\n",skb->len);
			return -EINVAL;
		}
		
		memcpy(&x25->x25_api_tx_hdr,skb->data,sizeof(x25api_hdr_t));
		
		/* Set the qdm bits from the packet header 
                 * User has the option to set the qdm bits
                 */
		qd_bits = (api_hdr->qdm >> 1)&0x03;
		m_bit = api_hdr->qdm & 0x01;

		skb_pull(skb,sizeof(x25api_hdr_t));
	}
	
	if (skb_tailroom(skb) < 16 || skb_headroom(skb) < 4){
		printk(KERN_INFO "BAD ROOM IN SEND FRAME T %i H %i\n",
				skb_tailroom(skb), skb_headroom(skb));
		return -EINVAL;
	}

	
	if (x25->x25_link->cfg.mode & GFI_MOD_8_BIT_SET){

		frame = skb_push(skb, 3);
		if (!frame){
			return -ENOMEM;
		}
		/* GFI Identifier*/
		frame[0] =  (qd_bits << 6) | (GFI_MOD_8_BIT_SET <<4);
		
		/* LCN number */
		frame[0] |= (unsigned char)((atomic_read(&x25->lcn) >> 8) & 0x0f); 
		frame[1] = (unsigned char)(atomic_read(&x25->lcn) & 0xff);
		
		/*Pr M Ps data */
		if (x25->x25_sw_lcn == NULL){
			frame[2] = (x25->Pr_count << 5) | (m_bit << 4) | (x25->Ps_count << 1);
		}else{
			/* X25_SW X.25 switching
			 * This LCN used for switching.
			 * Use switching Pr count (received from remote switching device)
			 * instead of real Pr count.
			 */
			frame[2] = (x25->sw_Pr_count << 5) | (m_bit << 4) | (x25->Ps_count << 1);
//			printk(KERN_INFO "%s: Debug X25 Event: Tx Data :Lcn=%d:Pr=0x%02X:Ps=0x%02X\n",
//					x25->dev->name,
//					atomic_read(&x25->lcn), 
//					x25->sw_Pr_count,
//					x25->Ps_count);
		}
	} else {
		frame = skb_push(skb, 4);
		if (!frame){
			return -ENOMEM;
		}

		/* GFI Identifier */
		frame[0] = (qd_bits << 6) | (GFI_FOR_MOD_128 << 4);

		/* LCN number */
		frame[0] |= (unsigned char)((atomic_read(&x25->lcn) >> 8) & 0x0f); 
		frame[1] = (unsigned char)(atomic_read(&x25->lcn) & 0xff);
		
		frame[2] = (x25->Ps_count << 1);
		if (x25->x25_sw_lcn == NULL){
			frame[3] = (x25->Pr_count << 1) | (m_bit);
		}else{
			/* X25_SW X.25 switching
			 * This LCN used for switching.
			 * Use switching Pr count (received from remote switching device)
			 * instead of real Pr count.
			 */
			frame[3] = (x25->sw_Pr_count << 1) | (m_bit);
		}
	}

	if ((err = x25_data_tx_down(x25->x25_link, skb)) != X25_OK){
		err = 1;
		skb_pull(skb,((x25->x25_link->cfg.mode & GFI_MOD_8_BIT_SET) ? 
					X25_HEADER_LEN : X25_HEADER_LEN+1));

		if (x25->usedby == API){
			frame = skb_push(skb,sizeof(x25api_hdr_t));
			if (!frame)
				return -EINVAL;

			memcpy(frame,&x25->x25_api_tx_hdr,sizeof(x25api_hdr_t));			
		}
	}

	return err;
}

int x25_lcn_send(x25_lcn_t *x25_lcn)
{
	struct sk_buff *skb;
	volatile unsigned short modulus, start, end, err=0;
	int len;
	unsigned char data;
	unsigned char pkt_type;
	unsigned char mbit;
	
	WAN_ASSERT2((x25_lcn==NULL),0);
	
	if (test_bit(LCN_DOWN,&x25_lcn->critical) || 
	    test_bit(LCN_UNREGISTER,&x25_lcn->critical)){
		return 0;
	}
	
#if BH_DEBUG > 1
	printk(KERN_INFO "Sending Device %s\n",x25_lcn->dev->name);
#endif
	
	if (x25_lcn->state != X25_ST_DATA_TRANSFER_P4){
		if (!IS_PVC(x25_lcn)){
			return 0;
		}else{
			/* It is legal for PVC to be in non
			 * transfer state */
		}
	}

	if ((skb=skb_dequeue(&x25_lcn->async_queue)) != NULL){

		/* Update the PR count of the outgoing RR packet
		 * since the Pr count might have been changed from
		 * the time the packet was created. */

		if (x25_lcn->x25_link->cfg.mode & GFI_MOD_8_BIT_SET){
			pkt_type = skb->data[2]&0x1F;
		}else{
			pkt_type = skb->data[3];
		}
		
		switch (pkt_type){
			
			case RR_PKT:
			case RNR_PKT:
			case REJ_PKT:
				
			if (x25_lcn->x25_sw_lcn == NULL){
				if (x25_lcn->x25_link->cfg.mode & GFI_MOD_8_BIT_SET){	
					skb->data[2] = (skb->data[2]&0x1F) | (x25_lcn->Pr_count<<5);
				}else{
					mbit=skb->data[3]&0x01;
					skb->data[3] = (x25_lcn->Pr_count<<1)|mbit;
				}
			}else{
				/* X25_SW X.25 switching
				 * This LCN used for switching.
				 * Use switching Pr count (received from remote switching device)
				 * instead of real Pr count.
				 */
				if (x25_lcn->x25_link->cfg.mode & GFI_MOD_8_BIT_SET){
					skb->data[2] = (skb->data[2]&0x1F) | (x25_lcn->sw_Pr_count<<5);
				}else{
					mbit=skb->data[3]&0x01;
					skb->data[3] = (x25_lcn->sw_Pr_count<<1)|mbit;
				}
//				printk(KERN_INFO "%s: Debug X25 Event: Tx RR :Lcn=%d:Pr=0x%02X\n",
//					x25_lcn->dev->name,
//					atomic_read(&x25_lcn->lcn), x25_lcn->sw_Pr_count);
			}
			break;
		}

		data=skb->data[2];
		if (x25_data_tx_down(x25_lcn->x25_link, skb) != X25_OK){
			//printk(KERN_INFO "Failed to send LCN RR Asyc data!\n");
			skb_queue_head(&x25_lcn->async_queue,skb);
			return -EBUSY;
		}
	}
	
	modulus = (x25_lcn->x25_link->cfg.mode & GFI_MOD_8_BIT_SET) ? 
				MOD8_MODULUS : MOD128_MODULUS;

	/* If ack queue is empty:
	 * 	The low edge window = Ps_ack
	 * 	all the packets have already been acknowledeged.
	 * Otherwise:
	 *      The low edge window = Ps_count
	 *      we already sent some packets that have not been
	 *      acked.  Thus, the lower edge must start at the last 
	 *      packet we sent.
	 */
	start = x25_lcn->Ps_count;
	end   = (x25_lcn->Ps_ack + x25_lcn->tx_window_size) % modulus;

	if (test_bit(TX_WINDOW_CLOSED,&x25_lcn->critical)){
		printk(KERN_INFO "%s: TX WINDOW CLOSED !!!!\n",
				x25_lcn->dev->name);
		return 0;
	}
	
	if (start == end){
#if BH_DEBUG > 2
		printk(KERN_INFO "%s: TX WINDOW CLOSED Start=%i End=%i\n",
				x25_lcn->dev->name, start, end);
#endif
		return 0;
	}

	set_bit(LCN_SENDING,&x25_lcn->critical);
	
	if (x25_lcn->state == X25_ST_DATA_TRANSFER_P4 && 
	    start != end &&
	    (skb=skb_dequeue(&x25_lcn->write_queue)) != NULL){

		/*
		 * Transmit the frame copy.
		 */
		len=skb->len;
		if ((err=x25_send_data_frame(x25_lcn, skb)) != X25_OK){
				
			if (err > 0){
				//printk(KERN_INFO "%s: Failed to send data!\n",
				//		x25_lcn->dev->name);
				skb_queue_head(&x25_lcn->write_queue, skb);
				clear_bit(LCN_SENDING,&x25_lcn->critical);
				return -EBUSY;
			}else{
				printk(KERN_INFO "CRITICAL SYSTEM ERROR in BH !!!\n");
				KFREE_SKB(skb);
				clear_bit(LCN_SENDING,&x25_lcn->critical);
				return -EINVAL;
			}
		}else{
			x25_lcn->Ps_count = (x25_lcn->Ps_count + 1) % modulus;
			x25_update_stats(x25_lcn->x25_link,DATA_PKT,1);
			x25_lcn->ifstats.tx_packets++;
			x25_lcn->ifstats.tx_bytes += len;
		}
	}

	clear_bit(LCN_SENDING,&x25_lcn->critical);

	
	if (test_bit(IF_SEND_AWAITING_KICK,&x25_lcn->critical)){
		if (skb_queue_len(&x25_lcn->write_queue) < x25_lcn->tx_window_size){
#if BH_DEBUG > 2
			printk(KERN_INFO "SVC BH TX: Kicking dev %s Queue %i\n",
					x25_lcn->dev->name,skb_queue_len(&x25_lcn->write_queue));
#endif
			clear_bit (IF_SEND_AWAITING_KICK,&x25_lcn->critical);
			netif_wake_queue(x25_lcn->dev);
			
			if (test_bit(LCN_SK_ID,&x25_lcn->used) && x25_lcn->sk_id){
				wanpipe_api_poll_wake(x25_lcn->sk_id);
				
			}else if (x25_lcn->dsp_id){
				if (IS_FUNC_CALL(x25_dsp_protocol,dsp_x25_bh_kick)){
					x25_dsp_protocol.dsp_x25_bh_kick(x25_lcn->dsp_id);
				}	
			}
		}
	}

	if (skb_queue_len(&x25_lcn->write_queue) || 
	    skb_queue_len(&x25_lcn->async_queue)){
		//printk(KERN_INFO "More Data on %s %i %i\n",
		//		x25_lcn->dev->name,
		//		skb_queue_len(&x25_lcn->write_queue),
		//		skb_queue_len(&x25_lcn->async_queue));
		set_bit(MORE_LCN_TX,&x25_lcn->x25_link->tq_working);
	}
	
	return 0;
}

#define MAX_SOFTIRQ_TIMEOUT 3

void x25_link_send (unsigned long x25_link_data)
{
#define x25_link ((x25_link_t*)x25_link_data)
	struct sk_buff *skb=NULL;
	x25_lcn_t *x25_lcn;
	struct net_device *x25_dev;
	volatile unsigned long start_time=jiffies;
		
		
	if (test_and_set_bit(SMP_CONFLICT,&x25_link->tq_working)){
		printk(KERN_INFO " !!!!!!!!!!! ASYNC TX BH CRITICAL : SMP !!!!!!!!!!!!! \n");
		return;
	}

	if (test_bit(LINK_DOWN,&x25_link->tq_working)){
		return;
	}
	
	if (x25_link->state == X25_ST_HDLC_NOT_IN_ABM){
		printk(KERN_INFO "X25 BH TX: Link not in ABM Mode!\n");
		clear_bit(MORE_LCN_TX,&x25_link->tq_working);
		clear_bit(BH_AWAITING_KICK,&x25_link->tq_working);
		clear_bit(SMP_CONFLICT,&x25_link->tq_working);
		clear_bit(BH_RUNNING,&x25_link->tq_working);
		return;
	}

	++x25_link->kick_cnt;

	/* Handle X25 Rx Queue First */
	while ((skb=skb_dequeue(&x25_link->rx_queue))){
		if (x25_data_input(x25_link, skb) != 0){
			skb_queue_head(&x25_link->rx_queue,skb);
			break;
		}

		if ((jiffies-start_time) > MAX_SOFTIRQ_TIMEOUT){
			printk(KERN_INFO "%s: RxBH Time squeeze\n",x25_link->name);
			goto link_bh_kick_exit;
		}
	}

	if (test_bit(BH_AWAITING_KICK,&x25_link->tq_working)){
		goto link_bh_kick_exit;
	}

	clear_bit(MORE_LCN_TX,&x25_link->tq_working);

x25_async_available:
	
	/* Handle X25 Async Tx Queue */
	while ((skb = skb_dequeue(&x25_link->async_queue)) != NULL){
		if (x25_data_tx_down(x25_link, skb) != X25_OK){
			skb_queue_head(&x25_link->async_queue,skb);
			set_bit(BH_AWAITING_KICK,&x25_link->tq_working);
			goto link_bh_kick_exit;
		}

		if ((jiffies-start_time) > MAX_SOFTIRQ_TIMEOUT){
			printk(KERN_INFO "%s: TxBH Async Time squeeze\n",x25_link->name);
			goto link_bh_kick_exit;
		}
	}
	
	if (x25_link->tx_dev == NULL){
		x25_link->tx_dev = get_first_tx_dev(x25_link);
		if (!x25_link->tx_dev){
			goto link_bh_kick_exit;
		}
	}else{
		x25_lcn = (x25_lcn_t*)x25_link->tx_dev->priv;		
		if (!atomic_read(&x25_lcn->lcn)){
			X25_PUT(x25_link->tx_dev);
			x25_link->tx_dev = get_first_tx_dev(x25_link);
			if (!x25_link->tx_dev){
				goto link_bh_kick_exit;
			}
		}
	}

	x25_dev = x25_link->tx_dev;
	for (;;){
	
		if ((jiffies-start_time) > MAX_SOFTIRQ_TIMEOUT){
			printk(KERN_INFO "%s: TxBH Data Time squeeze\n",x25_link->name);
			goto link_bh_kick_exit;
		}
		
		if (skb_peek(&x25_link->async_queue) != NULL){
			x25_link->tx_dev = x25_dev;
			goto x25_async_available;
		}
		
		x25_lcn = (x25_lcn_t*)x25_dev->priv;
		
		switch (x25_lcn_send(x25_lcn)){

		case X25_OK:
			/* This LCN transmitted its packet OK, move 
			 * to the next LCN and try transmitting for it*/

			X25_PUT(x25_dev);
			x25_dev = get_next_tx_dev(x25_link,atomic_read(&x25_lcn->lcn));
			if (!x25_dev){
				X25_PUT(x25_link->tx_dev);
				x25_link->tx_dev = x25_dev;
				break;
			}
			
			if (x25_dev == x25_link->tx_dev){
				
				/* We tried all devices, get out, and wait
				 * for a kick again. Start the tx from the next
			 	 * configured device the next time we are in
			 	 * TX */
				X25_PUT(x25_dev);
				x25_dev = get_next_tx_dev(x25_link,atomic_read(&x25_lcn->lcn));

				x25_link->tx_dev = x25_dev;
				break;
			}
			continue;
		
		case -EBUSY:
			/* The HDLC link is busy, thus, we must wait for it
			 * to kick us once it frees up.
			 *
			 * Since this LCN didn't send, next time we are kick,
			 * we will start transmitting from it*/
			
			x25_link->tx_dev = x25_dev;
			set_bit(BH_AWAITING_KICK,&x25_link->tq_working);
			break;
			
			
		default:
			/* Critical error while sending. Something is really
			 * wrong, like no memory !!! */
			printk(KERN_INFO "CRITICAL ERROR FROM LCN SEND\n");
			x25_link->tx_dev = x25_dev;
			break;
		}

		break;
	}
		
link_bh_kick_exit:	

	/* NC Note:
	 *
	 * Don't use the rx_queue to indicate MORE_LCN_TX 
	 * condition, because, we could be stuck and awaiting
	 * kick from the bottom or top layer.  
	 *
	 * The result of the stuck rx_queueu would result in 
	 * constantly kicking the BH until that condition 
	 * occurs.  i.e. 100% CPU utilitzation */

	if (skb_peek(&x25_link->async_queue)/* || skb_peek(&x25_link->rx_queue)*/){
		set_bit(MORE_LCN_TX,&x25_link->tq_working);	
	} 

	if (test_bit(BH_AWAITING_KICK,&x25_link->tq_working)){
		clear_bit(MORE_LCN_TX,&x25_link->tq_working);
	}
	
	clear_bit(SMP_CONFLICT,&x25_link->tq_working);
	clear_bit(BH_RUNNING,&x25_link->tq_working);

	/* Re-trigger Tx if there is more data to tx */
	if (test_bit(MORE_LCN_TX,&x25_link->tq_working)){
		x25_trigger_link_bh(x25_link,0);
	}

	return;
#undef x25_link
}


/* 
 *	This routine is called when the HDLC layer internally  generates a
 *	command or  response  for  the remote machine ( eg. RR, UA etc. ). 
 *	Only supervisory or unnumbered frames are processed, FRMRs are handled
 *	by x25_transmit_frmr below.
 */
void x25_send_control (x25_link_t *x25_link, unsigned char qd_bit, 
		       unsigned short lcn, unsigned char type, 
		       unsigned char cause, unsigned char diag, 
		       unsigned char *data_ptr, int data_len)
{
	struct sk_buff *skb;
	unsigned char  *dptr;
	x25_lcn_t *x25_lcn=NULL;
	
	/* If ether the pointer is missing then
	 * do not allocat the data ptr size */
	if (!data_ptr)
		data_len=0;
	
	ALLOC_SKB(skb,(X25_HEADER_LEN + 15 + data_len));
	if (skb == NULL){
		printk(KERN_INFO "X25: Critical Error: No Memory!\n");
		return;
	}

	dptr = skb_put(skb, X25_HEADER_LEN);
	if (!dptr){
		printk(KERN_INFO "X25: Critical Error: No Memory!\n");
		goto x25_send_control_error;
	}
	
	dptr[0] = ((qd_bit << 6) | ((x25_link->cfg.mode&0x3) << 4) | (unsigned char)((lcn&0x0f00) >> 8) );
	dptr[1] = (unsigned char)(lcn&0x00ff);
	dptr[2] = type;

	switch (type){

		case RNR_PKT:
		case RR_PKT:
			x25_lcn = get_lcn_device(x25_link,lcn);
			if (!x25_lcn){
				KFREE_SKB(skb);
				return;
			}
			
			if (x25_link->cfg.mode & GFI_MOD_8_BIT_SET){
				dptr[2] = ((cause << 5) | (type & 0x1F));
			}else{
				dptr=skb_put(skb,1);
				dptr[0] = (cause<<1)&0xFE;
			}

			skb_queue_tail(&x25_lcn->async_queue,skb);
			goto x25_send_control_trigger;	

			
		case RESTART_REQUEST_PKT:
			x25_start_t10_t20_timer(x25_link);
			x25_log_proc_event(x25_link,NULL,"Tx Restart Request",
			   		   lcn,cause,diag,NULL);
			break;

		case CLEAR_REQUEST_PKT:
			x25_lcn = get_lcn_device(x25_link,lcn);
			if (x25_lcn){
				CALL_LOG(x25_link, "%s: %s: X25 Event: Tx Clear Request: Lcn=%i, Cause 0x%02X, Diag 0x%02X\n",
					x25_link->name,x25_lcn->dev->name,lcn,cause,diag);

				x25_log_proc_event(x25_link,x25_lcn->dev->name,"Tx Clear Request",
			   		   lcn,cause,diag,NULL);
				x25_start_t13_t23_timer(x25_lcn);
			}else{
				x25_log_proc_event(x25_link,NULL,"Tx Clear Request",
			   		   lcn,cause,diag,NULL);
			}
			break;

		case RESET_REQUEST_PKT:
			x25_lcn = get_lcn_device(x25_link,lcn);
			if (x25_lcn){
				printk(KERN_INFO "%s: %s: X25 Event: Tx Reset Request: Lcn=%i, Cause 0x%02X, Diag 0x%02X\n",
					x25_link->name,x25_lcn->dev->name,lcn,cause,diag);
				x25_log_proc_event(x25_link,x25_lcn->dev->name,"Tx Reset Request",
			   		   lcn,cause,diag,NULL);

				x25_start_t12_t22_timer(x25_lcn);
			}else{
				x25_log_proc_event(x25_link,NULL,"Tx Reset Request",
			   		   lcn,cause,diag,NULL);
			}
			break;

		case CALL_REQUEST_PKT:
			x25_lcn = get_lcn_device(x25_link,lcn);
			if (x25_lcn){
				x25_start_t11_t21_timer(x25_lcn);
				x25_log_proc_event(x25_link,x25_lcn->dev->name,"Tx Call Request",
			   		   lcn,cause,diag,x25_lcn->call_string);
			}else{
				x25_log_proc_event(x25_link,NULL,"Tx Call Request",
			   		   lcn,cause,diag,NULL);
			}
			goto x25_send_control_send_data;
			
		case DIAGNOSTIC_PKT:

			dptr=skb_put(skb,1);
			dptr[0] = diag;

			goto x25_send_control_send;	

		case INTERRUPT_PKT:
			goto x25_send_control_send_data;
			
		case CALL_ACCEPT_PKT:
			x25_lcn = get_lcn_device(x25_link,lcn);
			if (x25_lcn){
				x25_log_proc_event(x25_link,x25_lcn->dev->name,"Tx Call Accept",
			   		   lcn,cause,diag,x25_lcn->call_string);
			}else{	
				x25_log_proc_event(x25_link,NULL,"Tx Call Accept",
			   		   lcn,cause,diag,NULL);
			}	
			goto x25_send_control_send_data;
	
		case CLEAR_CONF_PKT:
			x25_lcn = get_lcn_device(x25_link,lcn);
			if (x25_lcn){
				x25_log_proc_event(x25_link,x25_lcn->dev->name,"Tx Clear Conf",
			   		   lcn,cause,diag,NULL);
			}else{
				x25_log_proc_event(x25_link,NULL,"Tx Clear Conf",
			   		   lcn,cause,diag,NULL);
			}

			/* Jun 18 2003 NC
			 * The clear confirmation does not include
			 * cause or diag. */
			goto x25_send_control_send;
			
		/* All other packets contain do not contain a 
		 * cause or diag as well as no data. 
		 * RESTART and RESET CONFIRM
		 */
		default:
			goto x25_send_control_send;	
	}
	
	dptr = skb_put(skb,2);
	if (!dptr)
		goto x25_send_control_error;

	dptr[0] = cause;
	dptr[1] = diag;
	
x25_send_control_send_data:
	
	if (data_ptr && data_len){
		dptr = skb_put(skb,data_len);
		if (!dptr)
			goto x25_send_control_error;

		memcpy(dptr,data_ptr,data_len);
	}
	
x25_send_control_send:
	skb_queue_tail (&x25_link->async_queue,skb);

x25_send_control_trigger:

	x25_update_stats(x25_link,type,1);
	x25_trigger_link_bh(x25_link,0);
	return;
	
x25_send_control_error:
	printk(KERN_INFO "X25: x25_send_control FAILED !\n");
	KFREE_SKB(skb);
	return;
}



int x25_place_call (x25_lcn_t *x25_lcn)
{
	struct sk_buff *call_skb=NULL;
	int err;
	
	if (atomic_read(&x25_lcn->lcn) != 0 && atomic_read(&x25_lcn->lcn) != 0xFFFF){
		printk(KERN_INFO "%s: Error: Tx Call Request failed, dev already bound to Lcn=%i\n",
				x25_lcn->dev->name,atomic_read(&x25_lcn->lcn));
		return -EBUSY;
	}

	if (x25_lcn->state != X25_ST_PKT_LVL_RDY_P1){
		return -EINVAL;
	}

	ALLOC_SKB(call_skb,X25_HEADER_LEN + MAX_CALL_REQ_ASYNC_PKT);
	if (call_skb == NULL){
		printk(KERN_INFO "X25: Critical Error: No Memory!\n");
		return -EINVAL;
	}

	err = x25_parse_call_string(x25_lcn, CALL_REQUEST_PKT, call_skb);

	if (err != X25_OK){
		printk(KERN_INFO "%s: Place Call Failed rc=%x\n",
				x25_lcn->dev->name,err);
		KFREE_SKB(call_skb);
		return err;
	}

	if (select_outgoing_call_LCN(x25_lcn) != X25_OK){
		printk(KERN_INFO "%s: Error: Tx Call Request failed no free lcn's!\n",
				x25_lcn->dev->name);
		KFREE_SKB(call_skb);
		return -EBUSY;
	}

	lcn_state_change(x25_lcn,X25_ST_CALL_REQ_SNT_P2);

	CALL_LOG(x25_lcn->x25_link, "%s: X25 Event: Tx Call Request  :Lcn=%i: %s: %s\n\n",
			x25_lcn->x25_link->name,atomic_read(&x25_lcn->lcn),
			x25_lcn->dev->name,
			x25_lcn->call_string);
	
	if (x25_lcn->usedby == API){
		x25_send_control(x25_lcn->x25_link, 
				 (x25_lcn->x25_api_cmd_hdr.qdm>>1), 
				 atomic_read(&x25_lcn->lcn),
				 CALL_REQUEST_PKT, 
			 	 x25_lcn->x25_api_cmd_hdr.cause, 
				 x25_lcn->x25_api_cmd_hdr.diagn,
			 	 call_skb->data,
			 	 call_skb->len);

		memcpy(&x25_lcn->x25_api_event,
		       &x25_lcn->x25_api_cmd_hdr,
		       sizeof(x25api_hdr_t));
		
		x25_lcn->x25_api_event.hdr.pktType = CALL_REQUEST_PKT;
		x25_lcn->x25_api_event.hdr.lcn = atomic_read(&x25_lcn->lcn);
		x25_lcn->x25_api_event.hdr.mtu = x25_lcn->tx_pkt_size;
		x25_lcn->x25_api_event.hdr.mru = x25_lcn->rx_pkt_size;
			
		if (call_skb && (x25_lcn->x25_api_event.hdr.length=call_skb->len)){
			memcpy(x25_lcn->x25_api_event.data,
			       call_skb->data,
			       call_skb->len);
		}

	}else{
		x25_send_control(x25_lcn->x25_link, 
				 0, atomic_read(&x25_lcn->lcn),
				 CALL_REQUEST_PKT, 
			 	 0, 0,
			 	 call_skb->data,
			 	 call_skb->len);
	}
	
	/* These are temporary buffers that 
	 * must be discarded */
	KFREE_SKB(call_skb);

	return 0;
}

int x25_accept_call (x25_lcn_t *x25_lcn)
{
	struct sk_buff *call_skb=NULL;
	int err=0;

	if (atomic_read(&x25_lcn->lcn) == 0 || atomic_read(&x25_lcn->lcn)==0xFFFF){
		return 0;
	}

	if (x25_lcn->state != X25_ST_CALL_REQ_REC_P3){
		return -EINVAL;
	}

	ALLOC_SKB(call_skb,X25_HEADER_LEN + MAX_CALL_REQ_ASYNC_PKT);
	if (call_skb == NULL){
		printk(KERN_INFO "%s: X25: Critical Error: No Memory!\n",x25_lcn->dev->name);
		return -EINVAL;
	}

	err = x25_parse_call_string(x25_lcn, CALL_ACCEPT_PKT, call_skb);

	if (err != X25_OK){
		printk(KERN_INFO "%s: Accept Call Failed, format error, rc=0x%02X\n",
			x25_lcn->dev->name,err);
		KFREE_SKB(call_skb);
		return err;
	}
	
	CALL_LOG(x25_lcn->x25_link, "%s: X25 Event: Tx Accept Call   :Lcn=%i\n",
			x25_lcn->dev->name,atomic_read(&x25_lcn->lcn));
	
	lcn_state_change(x25_lcn,X25_ST_DATA_TRANSFER_P4);
	x25_send_control(x25_lcn->x25_link, 
		 	(x25_lcn->x25_api_cmd_hdr.qdm >> 1), 
		 	atomic_read(&x25_lcn->lcn),
		 	CALL_ACCEPT_PKT, 
		 	x25_lcn->x25_api_cmd_hdr.cause, 
		 	x25_lcn->x25_api_cmd_hdr.diagn,
			call_skb->data,
		 	call_skb->len);

	memcpy(&x25_lcn->x25_api_event,
	       &x25_lcn->x25_api_cmd_hdr,
	       sizeof(x25api_hdr_t));

	x25_lcn->x25_api_event.hdr.pktType = CALL_ACCEPT_PKT;
	x25_lcn->x25_api_event.hdr.lcn = atomic_read(&x25_lcn->lcn);
	x25_lcn->x25_api_event.hdr.mtu = x25_lcn->tx_pkt_size;
	x25_lcn->x25_api_event.hdr.mru = x25_lcn->rx_pkt_size;
	
	if (call_skb && (x25_lcn->x25_api_event.hdr.length=call_skb->len)){
		memcpy(x25_lcn->x25_api_event.data,
		       call_skb->data,
		       call_skb->len);
	}

	
	KFREE_SKB(call_skb);
	
	return 0;
}




int x25_clear_call (x25_lcn_t *x25_lcn,unsigned char force)
{
	struct sk_buff *call_skb=NULL;
	int err=0;
	
	if (atomic_read(&x25_lcn->lcn) == 0 || atomic_read(&x25_lcn->lcn)==0xFFFF){
		return 0;
	}

	if (x25_lcn->state == X25_ST_CLEAR_REQ_SNT_P6){
		printk(KERN_INFO "%s: X25 Api Tx Clear Request: Clear call already in progress!\n",
				x25_lcn->dev->name);
		return 0;
	}

	if (!force && x25_lcn->x25_api_cmd_hdr.length){
		ALLOC_SKB(call_skb,X25_HEADER_LEN + MAX_CALL_REQ_ASYNC_PKT);
		if (call_skb == NULL){
			printk(KERN_INFO "%s: X25: Critical Error: No Memory!\n",
					x25_lcn->dev->name);
			return -EINVAL;
		}

		err = x25_parse_call_string(x25_lcn, CLEAR_REQUEST_PKT, call_skb);

		if (err != X25_OK){
			printk(KERN_INFO "%s: X25 Tx Clear Call Failed: packet format error : rc=%x\n",
				x25_lcn->dev->name,err);
			KFREE_SKB(call_skb);
			return err;
		}
	}

	lcn_state_change(x25_lcn,X25_ST_CLEAR_REQ_SNT_P6);
	x25_send_control(x25_lcn->x25_link, 
			 (x25_lcn->x25_api_cmd_hdr.qdm >> 1), 
			 atomic_read(&x25_lcn->lcn),
			 CLEAR_REQUEST_PKT, 
			 force ? 0:x25_lcn->x25_api_cmd_hdr.cause, 
			 force ? 0:x25_lcn->x25_api_cmd_hdr.diagn,
			 call_skb ? call_skb->data : NULL,
			 call_skb ? call_skb->len : 0);

	if (!force){
		memcpy(&x25_lcn->x25_api_event,
		       &x25_lcn->x25_api_cmd_hdr,
		       sizeof(x25api_hdr_t));

		x25_lcn->x25_api_event.hdr.pktType = CLEAR_REQUEST_PKT;
		x25_lcn->x25_api_event.hdr.lcn = atomic_read(&x25_lcn->lcn);
		x25_lcn->x25_api_event.hdr.mtu = x25_lcn->tx_pkt_size;
		x25_lcn->x25_api_event.hdr.mru = x25_lcn->rx_pkt_size;
			

		if (call_skb && (x25_lcn->x25_api_event.hdr.length=call_skb->len)){
			memcpy(x25_lcn->x25_api_event.data,
			       call_skb->data,
			       call_skb->len);
		}

		
		if (call_skb){
			KFREE_SKB(call_skb);
		}
	}
	return 0;
}


int x25_reset_call (x25_lcn_t *x25_lcn)
{

	if (atomic_read(&x25_lcn->lcn) == 0 || atomic_read(&x25_lcn->lcn)==0xFFFF){
		return 0;
	}

	if (x25_lcn->state == X25_ST_PKT_LVL_RDY_P1){
		/* The link is already in ready state thus,
		 * we can HARD initialize it */
		x25_initialize_lcn(x25_lcn,X25_ST_PKT_LVL_RDY_P1,HARD_INIT);
		return 0;

	}else if (x25_lcn->state == X25_ST_RESET_REQ_SNT_D2){
		return 0;
	}

	printk(KERN_INFO "%s: X25 Event: Api Tx Reset Request :Lcn=%i:Cause=0x%02X:Diag=0x%02X\n",
			x25_lcn->dev->name,atomic_read(&x25_lcn->lcn),
			x25_lcn->x25_api_cmd_hdr.cause,
			x25_lcn->x25_api_cmd_hdr.diagn);

	lcn_state_change(x25_lcn,X25_ST_RESET_REQ_SNT_D2);
	x25_send_control(x25_lcn->x25_link, 
			 (x25_lcn->x25_api_cmd_hdr.qdm >> 1), 
			 atomic_read(&x25_lcn->lcn),
			 RESET_REQUEST_PKT, 
			 x25_lcn->x25_api_cmd_hdr.cause, 
			 x25_lcn->x25_api_cmd_hdr.diagn,
			 NULL,
			 0);

	memcpy(&x25_lcn->x25_api_event,
	       &x25_lcn->x25_api_cmd_hdr,
	       sizeof(x25api_hdr_t));

	x25_lcn->x25_api_event.hdr.pktType = RESET_REQUEST_PKT;
	x25_lcn->x25_api_event.hdr.lcn = atomic_read(&x25_lcn->lcn);
	x25_lcn->x25_api_event.hdr.mtu = x25_lcn->tx_pkt_size;
	x25_lcn->x25_api_event.hdr.mru = x25_lcn->rx_pkt_size;
	x25_lcn->x25_api_event.hdr.length = 0;
	

	DEBUG_TEST("%s: %s State=%i\n",__FUNCTION__,x25_lcn->dev->name,x25_lcn->state);

	return 0;
}


