#include <linux/wanpipe_includes.h>
#include <linux/wanpipe_cfg.h>
#include "wanpipe_x25.h"

unsigned short pkt_size_array[] = {0, 0, 0, 0, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096};

unsigned char __inline__ gfi_decode (unsigned int config)
{
	if (config & D_BIT_PRAGMATICS_FOR_SVC){
		return DBIT_SET;
	}
	return QOFF_DOFF;
}

/* CHECK THAT A RECEIVED PACKET IS WITHIN THE PERMITTED MINIMUM AND MAXIMUM LENGTHS */

int pkt_lgth_limits(unsigned int X25_pkt_lgth, UCHAR min_lgth, USHORT max_lgth)
{

	/* check for a packet too short or too long */
	
	if(X25_pkt_lgth < min_lgth){							
		return PKT_TOO_SHORT;
	
	}else if(X25_pkt_lgth > max_lgth){						
		return PKT_TOO_LONG;
	}

	return 0;
}



/******************* CHECK THE LENGTH OF AN INCOMING PACKET *******************/

int packet_length_error(x25_link_t *x25, struct x25_frame *frame)
{
	int err=0;

	/* process each packet type */
	switch(frame->type) {	

		case RR_PKT:	/* an RR and RNR */
		case RNR_PKT:
			{
			int max_len = (x25->cfg.mode & GFI_MOD_8_BIT_SET)?
						MAX_LGTH_SUP_PKT:MAX_LGTH_SUP_PKT+1;

			err = pkt_lgth_limits(frame->len, 
					      (UCHAR)MIN_LGTH_SUP_PKT, 
					       max_len);
			}
			break;

		case CALL_REQUEST_PKT:	/* a Call Request/Incoming Call packet */
			err = pkt_lgth_limits(frame->len, 
					      (UCHAR)MIN_LGTH_CALL_REQ_PKT, 
					      MAX_LGTH_CALL_REQ_PKT);
			break;

		case CALL_ACCEPT_PKT:	/* a Call Accepted/Connected packet */
			
			/* the minimum length of this packet depends on whether or not we 
			 * are configured to include the address and */
			/* facility length fields in the packet */
			
			err = pkt_lgth_limits(frame->len,
					      (x25->cfg.X25_protocol_options & 
					       			CALLAC_DOES_NOT_INCL_ADDR_LGTH) ?
								(UCHAR)MIN_LGTH_CALLAC_NO_ADDR_FACILS : 
								(UCHAR)MIN_LGTH_CALLAC_ADDR_FACILS,
						MAX_LGTH_CALL_ACCEPT_PKT);
			break;

		case CLEAR_REQUEST_PKT:	/* a Clear Request/Indication packet */
			
			/* the minimum length of this packet depends on whether or not we are 
			 * configured to include a diagnostic
			 * field in the packet */
			 
			err = pkt_lgth_limits(frame->len,
					      (x25->cfg.X25_protocol_options & 
					       		NO_DIAG_FIELD_IN_ASYNC_PKTS) ? 
							(UCHAR)MIN_LGTH_CLEAR_REQ_PKT_NO_DIAG : 
							(UCHAR)MIN_LGTH_CLRRQ_PKT_WITH_DIAG, 
						MAX_LGTH_CLEAR_REQ_PKT);
			break;

		case CLEAR_CONF_PKT: /* a Clear Confirmation packet */
			err = pkt_lgth_limits(frame->len,
					      (UCHAR)MIN_LGTH_CLEAR_CONF_PKT, 
					       MAX_LGTH_CLEAR_CONF_PKT);
			break;

		case RESET_REQUEST_PKT: 	/* a Reset Request/Indication packet */
		case RESTART_REQUEST_PKT:	/* a Restart Request/Indication packet */
			
			/* the minimum length of these packets depends on whether or not we 
			 * are configured to include a diagnostic 
			 * field in these packets */
			err = pkt_lgth_limits(frame->len,
					      (x25->cfg.X25_protocol_options & 
					       		NO_DIAG_FIELD_IN_ASYNC_PKTS) ? 
							(UCHAR)MIN_LGTH_RESET_REQ_PKT_NO_DIAG :
							(UCHAR)MIN_LGTH_RESET_REQ_PKT_DIAG, 
						MAX_LGTH_RESET_REQ_PKT);
			break;

		case RESET_CONF_PKT:		/* a Reset Confirmation packet */
		case RESTART_CONF_PKT:		/* a Restart Confirmation packet */
		case INTERRUPT_CONF_PKT:	/* an Interrupt Confirmation packet */
			err = pkt_lgth_limits(frame->len,
					      (UCHAR)MIN_LGTH_RESET_CONF_PKT, 
					      MAX_LGTH_RESET_CONF_PKT);
			break;

		case INTERRUPT_PKT:		/* an Interrupt packet */
			/* the maximum length of this packet depends on the CCITT 
			 * compatibility configuration, as 1980 limits the
			 * data length to 1 byte, and 1984/8 permits 32 bytes */
			err = pkt_lgth_limits(frame->len,
					      (UCHAR)MIN_LGTH_INTERRUPT_PKT, 
					      (x25->cfg.CCITT_compatibility == 
					       CCITT_1980_COMPATIBILITY) ?
					      		MAX_LGTH_INT_PKT_CCITT_1980 : 
							MAX_LGTH_INT_PKT_CCITT_1984_88);
			break;

		case DIAGNOSTIC_PKT:	/* a Diagnostic packet */
			err = pkt_lgth_limits(frame->len,
					      (UCHAR)MIN_LGTH_DIAGNOSTIC_PKT, 
					      MAX_LGTH_DIAGNOSTIC_PKT);
			break;

		case REGISTRATION_REQUEST_PKT:	/* a Registration Request packet */
			err = pkt_lgth_limits(frame->len,
					      (UCHAR)MIN_LGTH_REGISTR_REQ_PKT, 
					      MAX_LGTH_REGISTR_REQ_PKT);
			break;

		case REGISTRATION_CONF_PKT: /* a Registration Confirmation packet */
			err = pkt_lgth_limits(frame->len,
					      (UCHAR)MIN_LGTH_REGISTR_CONF_PKT, 
					      MAX_LGTH_REGISTR_CONF_PKT);
			break;

		default:
			break;
	}

	
	return err;
}

/**** CHECK RESTART, REGISTRATION AND DIAGNOSTIC PACKETS FOR A FORMAT ERROR ***/

int Restrt_Registr_Diag_pkt_error(x25_link_t *x25, struct sk_buff *skb, struct x25_frame *frame)
{
	/* handle the different packet types */
	switch(frame->type) {	
	
		/* we discard a Diagnostic packet received on a non-zero LCN */
		case DIAGNOSTIC_PKT:
			if(frame->lcn){		/* (C2-K1, C2-K2, C2-K3) */
				printk(KERN_INFO "ResRegDiag_pkt: Diagnostic with non 0 lcn!\n");
				return -EINVAL;	/* discard the packet */
			}
			break;

		/* a Registration Request may only be received at a station */
		case REGISTRATION_REQUEST_PKT:
		
		/* configured as a DCE, and a Registration Confirmation may only 
		 * be received at a station configured as a DTE - note that is */
		case REGISTRATION_CONF_PKT:
			if(!frame->lcn) {	
			
				/* LCN is non-zero, we will handle this packet below */
				if ((x25->cfg.dte && (frame->type == REGISTRATION_REQUEST_PKT)) ||
				    (!x25->cfg.dte && (frame->type == REGISTRATION_CONF_PKT))) {
					
					/* handle the unsupported packet type, using the 
					 * diagnostic code #33 in any asynchronous packet issued */
					printk(KERN_INFO "ResRegDiag_pkt: Registration conf invalied\n");
					return -EINVAL;		/* discard the packet */
				}
			}
			break;
			
		default:
			/* Restart Request/Confirmation packets */
			break;
	}

	if(frame->lcn) {	/* if the packet has a non-zero LCN then handle this error, using */
		/* the diagnostic code #41 in any asynchronous packet issued */
		printk(KERN_INFO "ResRegDiag_pkt: Packet with non 0 lcn\n");
		return -EINVAL;		/* discard the packet */
	}
	
	return (packet_length_error(x25,frame));
}


/*****************************************************************************/

unsigned char hexdig(unsigned char n)
{
	n &= 0x0F;														
	/* only operate on the low four bits 
	 * return a single hexadecmal ASCII digit */
	return((n >= 10) ? (n - 10 + 'A') : (n + '0'));
}



void semi_oct_to_oct(UCHAR *src, UCHAR *dest, USHORT lgth)
{

	do {
		/* read the high four bits of the packed byte, 
		 * convert the hexadecimal value to a digit and store 
		 * the byte */
		
		*dest = hexdig((unsigned char)((*src & 0xF0) >> FOUR_BITS));
		dest++;
		
		/* read the low four bits of the packed byte, 
		 * convert the hexadecimal value to a digit and 
		 * store the byte */
		
		*dest = hexdig((UCHAR)(*src & 0x0F));
		dest++;
		src++;

	} while (lgth --);	
	
	/* process all characters in the packed string */
}


/* INSERT AN ASCII PREFIX INTO OUR RECEIVED ASYNCHRONOUS PACKET FORMATTING BUFFER */

void insert_ASCII_prefix(unsigned char **dest_ptr, unsigned char ASCII_prefix_char)
{

	unsigned char *d_ptr = *dest_ptr;
	
	if (ASCII_prefix_char != 'd'){
	  	/* if the prefix is a 's', 'f' or 'u', then insert a space */
		/* character before the '-' for formatting purposes */
		*d_ptr = SPACE_CHAR;
		d_ptr++;
	}
	
	/* insert the leading '-' character */
	*d_ptr = '-';
	d_ptr++;

	/* insert the passed 'd', 's', 'f' or 'u' 
	 * character into the formatting string */
	*d_ptr = ASCII_prefix_char;
	d_ptr++;

	*dest_ptr = d_ptr;
}


int check_insert_address_fields(unsigned char ASCII_prefix, 
				unsigned char *src_ptr, 
				unsigned char *dest_ptr,
				unsigned char address_lgth)
{
	int i;
	unsigned char util_char;
	
	/* insert the '-d' or '-s' address marker 
	 * in the packet arguments */	
	insert_ASCII_prefix(&dest_ptr, ASCII_prefix); 					

	/* read and check each character in the address string */
	for(i = ZERO; i < (USHORT)address_lgth; i ++) {

		/* read the next character */
		util_char = *(src_ptr + i);	
		
		/* read the next character check that this byte in a number */
		if((util_char < '0') || (util_char > '9')) {
			return -EINVAL;
		}else{
			/* store the address character in the packet formatting buffer 
			 * and increment the pointer to this buffer */
			*dest_ptr = util_char;		
			dest_ptr ++;
		}
	}

	/* the address field was OK */
	return 0;
}


int pkt_size_no (unsigned short pkt_size)
{
	int j;

	for(j = 4; j < SIZEOF_PKT_SIZE_ARRAY; j ++) {	
		
		/* search all elements of the packet size array - note that we 
		 * start searching at element #4, due to the ISO packet size 
		 * definitions get the current packet size in the array and check for a match */
		if(pkt_size_array[j] == pkt_size)
			/* if a match is found, then we are done */
			break;
	}
	
	/* return the array element offset */
	return(j);	
}

int decode_pkt_size (unsigned char *index)
{
	if (*index >= SIZEOF_PKT_SIZE_ARRAY)
		return -EINVAL;

	*index = pkt_size_array[*index];
	
	return X25_OK;
}



int asy_pkt_Data_lgth_excessive(unsigned int x25lcn_cfg, int X25_Data_lgth)
{

	/* the user data length is excessive if:
	 * fast select is not used and the data length 
	 * is greater than 16 bytes.
	 * fast select is used and the data length is 
	 * greater than 128 bytes. 
	 */

	if ((X25_Data_lgth > MAX_ASY_PKT_FST_SEL_DLGTH_X_2) ||
	   ((X25_Data_lgth > MAX_ASY_PKT_DLGTH_X_2) && !(x25lcn_cfg & FAST_SEL_REQUESTED))){
	
		return -EINVAL; /* the user data length is excessive */
	}
	
	return 0; /* the user data length is OK */
}


// ALEX: Add x25_link argument to x25_decode_asy_pkt function 
int x25_decode_asy_pkt (x25_link_t* x25_link,
			struct sk_buff *newskb, 
			struct sk_buff *asy_skb, 
			struct x25_frame *frame,
			int *off_facils_ptr,
			int *facils_lgth_ptr,
			int *user_data_lgth_ptr)
{
	int X25_Data_lgth = asy_skb->len;
	int off_facils = 0, facils_lgth = 0, calling_lgth = 0, called_lgth = 0;
	unsigned char *X25_Rx_pkt_Data_ptr = asy_skb->data;
	unsigned char *dest_ptr=NULL;
	
	unsigned char decode_buf[MAX_DECODE_BUF_SZ];
	unsigned char *decode_ptr = decode_buf;

	*off_facils_ptr=0;
	*facils_lgth_ptr=0;
	*user_data_lgth_ptr=0;

	// ALEX SHORT_ACCEPT_CALL:
	//		Check the Accept Call option. The Accept Call packet 
	//		cannot includes Address/Facilities/User Data.
	if (X25_Data_lgth == 0){
		return X25_OK;
	}
	
	if (frame->type == CALL_ACCEPT_PKT && 
		x25_link->cfg.X25_protocol_options & CALLAC_DOES_NOT_INCL_ADDR_LGTH){
		/* this conversion results in the doubling of the length of 
		 * the packet arguments  */
		X25_Data_lgth *= TWO_BYTES;		
		goto decode_user_data;
	}
	// ALEX ^^^^

	/* we use the 'decode_buf' to decode the 
	 * incoming arguments by converting them  from the received 
	 * semi-octet form, to an octec form */

	semi_oct_to_oct(X25_Rx_pkt_Data_ptr, 
			decode_ptr,
			X25_Data_lgth);

	/* this conversion results in the doubling of the length of 
	 * the packet arguments */
	X25_Data_lgth *= TWO_BYTES;		

	/* read the length of the calling address */
	calling_lgth = (*X25_Rx_pkt_Data_ptr & MASK_CALLING_ADDR_LGTH_BITS) >> FOUR_BITS;
	
	/* read the length of the called address */
	called_lgth = *X25_Rx_pkt_Data_ptr & MASK_CALLED_ADDR_LGTH_BITS;

	/* calculate the offset of the facility length byte within 
	 * the converted data string - note that we add two bytes 
	 * to account for the two address length fields */
	off_facils = TWO_BYTES + called_lgth + calling_lgth;

	/* if the total length of the calling and called address fields 
	 * is odd, then zeros are inserted to gain an integral number of 
	 * octets, so adjust the offset of the facility length field */
	if (off_facils % 2)
		off_facils ++;	

	/* read the length of the facility field, and process an 
	 * included facility field */
	facils_lgth = *(X25_Rx_pkt_Data_ptr + (off_facils / TWO_BYTES)) * TWO_BYTES;

	if (facils_lgth){
		/* We have an error condition under the following circumstances:
		 * a) The facility field is one byte in length 
		 *    (two bytes here as we have muliplied this length by two already).
		 * b) The length of the facility field is greater 
		 *    than the permitted maximum.
		 * c) The length of the actual facilities field is less 
		 *    than the defined facilities length, 
		 *    eg. the facilities length is defined as 3 bytes, 
		 *        but the facilities field is 4206 instead of 420606.
		 */
		if ((facils_lgth == TWO_BYTES) || 
		    (facils_lgth > MAX_ASY_PKT_FACILS_LGTH) ||
		    (X25_Data_lgth < (off_facils + facils_lgth + TWO_BYTES))) {
	
			/* for a packet on a non-zero LCN, issue a Clear 
			 * Request with a diagnostic code #64 (C3-D1, 
			 * note 5), and for a Registration Request/Confirmation, 
			 * issue a Restart Request or a Diagnostic 
			 * packet with a diagnostic code #64 */
			return CALL_SETUP_CLEAR_OR_REGSTR_PROB;
		}

		*off_facils_ptr = off_facils;
		*facils_lgth_ptr = facils_lgth;
	}


	if (called_lgth) {
		
		if ((newskb->len + called_lgth + D_OPT) > MAX_CALL_REQ_ASYNC_PKT)
			return -EINVAL;
		
		decode_ptr = decode_buf + TWO_BYTES;
		dest_ptr = skb_put(newskb,(called_lgth+D_OPT));

		/* if applicable, check and format the called address field */
		if(check_insert_address_fields(
			(unsigned char)'d',
			decode_ptr, 
			dest_ptr,
			called_lgth)){
		
			/* if we spot an error, then issue a Clear Request 
			 * with a diagnostic code #67/68 (C3-D1, note 5) */
			return INVALID_CALLED_DTE_ADDRESS;
		}
	}


	if (calling_lgth) {

		if ((newskb->len + calling_lgth + NON_D_OPT) > MAX_CALL_REQ_ASYNC_PKT)
			return -EINVAL;

		dest_ptr = skb_put(newskb,(calling_lgth + NON_D_OPT));
		decode_ptr = decode_buf + TWO_BYTES + called_lgth;

		/* if applicable, check and format the calling address field */
		if (check_insert_address_fields(
			(UCHAR)'s',
			decode_ptr,
			dest_ptr,
			calling_lgth)){

			/* if we spot an error, then issue a Clear Request 
			 * with a diagnostic code #67/68 (C3-D1, note 5) */
			return INVALID_CALLING_DTE_ADDRESS;
		}
	}


	if (facils_lgth) {

		if ((newskb->len + facils_lgth + NON_D_OPT) > MAX_CALL_REQ_ASYNC_PKT)
			return -EINVAL;

		dest_ptr = skb_put(newskb,(facils_lgth + NON_D_OPT));
		decode_ptr = decode_buf + off_facils + TWO_BYTES;
		
		/* if applicable, check and format the facilities field */
		/* insert the '-f' facilities marker in the packet arguments */
		insert_ASCII_prefix(&dest_ptr, (UCHAR)'f');
					
		/* copy the unpacked facilities string to the packet formatting buffer */
		memcpy(dest_ptr,
		       decode_ptr,
		       facils_lgth);
		
		/* increment the pointer to this formatting buffer */
	}

	// ALEX SHORT_ACCEPT_CALL:
	// 		off_facils is currently points to facilities length. Move off_facils
	// 		to real facilities offset.
	off_facils += TWO_BYTES;
	
decode_user_data:
	// ALEX ^^^^
	if ((X25_Data_lgth = X25_Data_lgth - (off_facils + facils_lgth))) {

		*user_data_lgth_ptr = X25_Data_lgth;
		
		if ((newskb->len +  X25_Data_lgth + NON_D_OPT) > MAX_CALL_REQ_ASYNC_PKT)
			return -EINVAL;

		dest_ptr = skb_put(newskb,(X25_Data_lgth+NON_D_OPT));
		
		decode_ptr = decode_buf + off_facils + facils_lgth;
		
		/* insert the '-u' user data marker in the packet arguments */
		insert_ASCII_prefix(&dest_ptr, (UCHAR)'u');
	
		/* copy the actual user data to the packet formatting buffer */
		memcpy(dest_ptr, 
		       decode_ptr, 
			X25_Data_lgth);
	}
#if 0
	if ((X25_Data_lgth = X25_Data_lgth - (off_facils + facils_lgth + TWO_BYTES))) {

		*user_data_lgth_ptr = X25_Data_lgth;
		
		if ((newskb->len +  X25_Data_lgth + NON_D_OPT) > MAX_CALL_REQ_ASYNC_PKT)
			return -EINVAL;

		dest_ptr = skb_put(newskb,(X25_Data_lgth+NON_D_OPT));
		
		decode_ptr = decode_buf + off_facils + facils_lgth + TWO_BYTES;
		
		/* insert the '-u' user data marker in the packet arguments */
		insert_ASCII_prefix(&dest_ptr, (UCHAR)'u');
	
		/* copy the actual user data to the packet formatting buffer */
		memcpy(dest_ptr, 
		       decode_ptr, 
			X25_Data_lgth);
	}
#endif

	dest_ptr = skb_put(newskb,1);
	*dest_ptr = '\0';

	return 0;
}

//============================= PARSE USER CALL STRING ==============================


/* ===================================================================
 * dighex
 * 
 * CONVERT A SINGLE HEX DIGIT TO A BINARY NUMBER 
 *
 */ 

unsigned char dighex(unsigned char n)
{

	if (n >= 'a')	
		/* provide the adjustment if the passed ASCII value is greater */
		return(n + 10 - 'a') ;	/* than 'a' */
	else 
		/* provide the adjustment if the passed ASCII value is greater */
		/* than 'A' or is a numeral */
		return((n >= 'A') ? (n + 10 - 'A') : (n - '0'));
}




unsigned short str_index(unsigned char *s, unsigned char *t, unsigned short max_chars)
{
	int l,i,j;
	/* process all characters in the larger string */
	for(l = 0; l <= max_chars; l ++) {
		for(j = l, i = 0; *(t + i) != '\0' && (*(s + j) == *(t + i)); j++, i++);

		/* if we have matched all the characters in the short string up */
		if(*(t + i) == '\0'){	
			/* to the null byte, then we have a match 
			 * the string was located, so return its offset */
			return(l + ONE_BYTE);		
		}
	}
	
	/* the string was not located */
	return(ZERO);	
}



unsigned short locate_facils_string(unsigned char *usr_data,
				    int off_facils,
		                    unsigned char *facils_string,
				    unsigned short len)
{
	int i;
	/* return the offset of this string if it is 
	 * located, otherwise return zero */
	return((i = str_index((usr_data + off_facils), facils_string, (len-off_facils))));

}


unsigned short locate_pkt_argument(unsigned char *call_string , 
		                   struct sk_buff *skb,  
				   unsigned char *srch_str, 
				   unsigned char shift)
{
	
	/* assume that the located argument length is zero */
	int X25_Data_lgth = ZERO;
	unsigned char util_char;
	unsigned char *dest_ptr,*prev_ptr;
	int i, x, y, data_off=0;
	int data_lgth = 0;
	
	/* search for the prefix within the asynchronous packet argument string */ 
	if ((i = str_index(call_string, srch_str, strlen(call_string)))) {

		/* Data offset withing the user ASCII call data buffer.
		 * This is the start of the address string we are looking
		 * for.*/
		data_off = i+1;
		
		/* find the end of this argument string by skipping 
		 * all characters until we find a whitespace */
		while(*(call_string + data_off + (++ X25_Data_lgth)) > ' ');

		/* read the prefix type in the search string ('d', 's', 'f', 'u') */
		util_char = *(call_string + data_off - ONE_BYTE);

	
		/* multiply the string length by 2, since one ascii
		 * equals two bytes.  
		 *
		 * If the data length is odd, then
		 * make it even by incrementing it, only if
		 * shift (indication that the previous, value
		 * was also odd) is not set.
		 */
		
		data_lgth = X25_Data_lgth >> 1;
		if ((X25_Data_lgth % 2) && shift == 0 && util_char != 'f'){
			data_lgth++;
		}

		/* copy the located argument string into our formatting 
		 * buffer - note that if we are copying a facilities string, then 
		 * we skip two bytes where the facilities length will be inserted */
		if (util_char == 'f'){
			/* Increment the buffer length to account
			 * for the facilites length */
			dest_ptr = skb_put(skb,1);
			*dest_ptr = (data_lgth & 0x3F);
			x=0; 
		}else{
			x=0;
		}

		dest_ptr = skb_put(skb, data_lgth);

		for (y=0; x<X25_Data_lgth ; x++, y++){	
		
			*(dest_ptr+y) = 0;
			*(dest_ptr+y) |= ((dighex(*(call_string+data_off+x))<<4) & 0xF0);
		
			x++;
			
			if (X25_Data_lgth > x){
				*(dest_ptr+y) |= (dighex(*(call_string+data_off+x)) & 0x0F);
			}

			/* If the -d address was odd, the -s address
			 * must be 4 bit shifted to the left. */
			if (shift){
				prev_ptr = (dest_ptr+y)-1;
				*prev_ptr |= (*(dest_ptr+y)>>shift);
				*(dest_ptr+y) = *(dest_ptr+y)<<shift;
			}
		}
		/* return the length of the located argument string */
		return(X25_Data_lgth);	
	}	

	/* the defined prefix string has not been located */
	return(ZERO);	
}


int x25_parse_call_string(x25_lcn_t *x25_lcn, unsigned char type, struct sk_buff *pkt_skb)
{
	unsigned char *dest_ptr;
	unsigned char calling_lgth,called_lgth,user_lgth,facils_lgth;
	unsigned char shift=0;
	
	x25_link_t* x25_link = x25_lcn->x25_link;
	
	if (type == CALL_ACCEPT_PKT &&
		x25_link->cfg.X25_protocol_options & CALLAC_DOES_NOT_INCL_ADDR_LGTH){

		/* SHORT_ACCEPT_CALL
 		 * If bit 6 in X25_protocol_options is set to 1, 
		 * the Accept Call packet will not includes 
		 * Address/Facilities length field.
	 	 */
		
		goto parse_user_data;
	}

	dest_ptr = skb_put(pkt_skb,1);
	*dest_ptr = 0;

	if ((called_lgth = locate_pkt_argument(x25_lcn->call_string,pkt_skb,"-d",0))){
		*dest_ptr |= (called_lgth&0x0F);	
	}else{
		if (type == CALL_REQUEST_PKT){
			printk(KERN_INFO "%s: Error, No destination address in place call string !\n",
					x25_lcn->dev->name);
			return NO_CALL_DESTINATION_ADDR;
		}
	}

	if ((called_lgth)%2){
		shift=4;
	}
	
	if ((calling_lgth = locate_pkt_argument(x25_lcn->call_string, pkt_skb,"-s",shift))){
		*dest_ptr |= (calling_lgth << 4)&0xF0;	
	}	

	/* check for any facility field included by the application */
	if (!(facils_lgth = locate_pkt_argument(x25_lcn->call_string, pkt_skb,"-f",0))) {

		/* if no facilities were included, so we must 
		 * manually insert the facility length field - 
		 * this length will either be '00' or '06' in the case 
		 * where we add flow control facilities 
		 * */
		if (x25_lcn->x25_link->cfg.X25_protocol_options & AUTO_FLOW_CTRL_PARM_FACIL_INS) {
			/* insert the flow control facilities */
			insert_flow_control_facils(x25_lcn,CALL_REQUEST_PKT, pkt_skb);		
		}else{
			if (type == CALL_REQUEST_PKT || type == CALL_ACCEPT_PKT){
				/* otherwise, we must insert the second '0' into the 
				 * facilities length field */
				dest_ptr = skb_put(pkt_skb,1);
				*dest_ptr = 0;	
			}
		}
	}else{
		/* We have facilites */
		


	}

	/* FIXME: Facilities are not supported in the outgoing calls
	 *        This should be fixed/added ASAP :) */
	
parse_user_data:
	if ((user_lgth = locate_pkt_argument(x25_lcn->call_string,pkt_skb,"-u",0))){
		
		if (asy_pkt_Data_lgth_excessive(x25_lcn->config,user_lgth)){
			return CALL_ARGS_LGTH_EXCESSIVE;	
		}
		
		/* if user length is ODD invalid */
		if (user_lgth % 2){
			return USER_DATA_FIELD_LGTH_ODD;
		}
	}
	
	return X25_OK;
}


inline int select_outgoing_call_LCN(x25_lcn_t *x25_lcn)
{
	x25_link_t *x25_link = x25_lcn->x25_link;
	volatile unsigned short LCN_in_use;
	unsigned long flags;
	
	/* LCNs are assigned as follows:
 	 *
	 *	Lowest LCN ----------------------------------------------- Highest LCN
	 *		 PVC      One-way incoming      Two-way      One-way outgoing
         *
	 *	* For a DTE, we place a call on the highest available 
	 *		One-way outgoing or Two-way LCN.
	 *	* For a DCE, we place a call on the lowest available 
	 *	 	Two-way or One-way outgoing LCN.
	 */

	/* for a DTE, we search for an available LCN, starting from the highest configured SVC */
	/* for a DCE, we search for an available LCN, starting from the lowest configured Two-way LCN */
	WP_WRITE_LOCK(&x25_link->map_lock,flags);
	for (LCN_in_use = ((x25_link->cfg.dte) ? 
		 		x25_link->cfg.highest_two_way_channel : 
				x25_link->cfg.lowest_two_way_channel);
			
	     	LCN_in_use != ((x25_link->cfg.dte) ? ZERO : 
			      (x25_link->cfg.highest_two_way_channel + 1));
	
	     	LCN_in_use = ((x25_link->cfg.dte) ? LCN_in_use - 1 : LCN_in_use + 1)) {

		if (x25_link->lcn_to_dev_map[LCN_in_use%MAX_LCN]){ 
			continue;	
		}

		X25_HOLD(x25_lcn->dev);
		x25_link->lcn_to_dev_map[LCN_in_use%MAX_LCN] = x25_lcn->dev;

		atomic_set(&x25_lcn->lcn,LCN_in_use);
		WP_WRITE_UNLOCK(&x25_link->map_lock,flags);
		return 0;
	}

	WP_WRITE_UNLOCK(&x25_link->map_lock,flags);
	return -EINVAL;  /* no available LCN was found */	
}


/*============================================================================
 * Strip leading and trailing spaces off the string str.
 */
char* strstrip (char* str, char* s)
{
	char* eos = str + strlen(str);		/* -> end of string */

	while (*str && strchr(s, *str))
		++str;				/* strip leading spaces */
	while ((eos > str) && strchr(s, *(eos - 1)))
		--eos;				/* strip trailing spaces */
	*eos = '\0';
	return str;
}

/*============================================================================
 * Tokenize string.
 *	Parse a string of the following syntax:
 *		<tag>=<arg1>,<arg2>,...
 *	and fill array of tokens with pointers to string elements.
 *
 *	Return number of tokens.
 */
int tokenize (char* str, char **tokens, char *arg1, char *arg2, char *arg3)
{
	int cnt = 0;
#ifdef LINUX_2_6
	tokens[0] = strsep(&str, arg1);
	while (tokens[cnt] && (cnt < MAX_TOKENS - 1)) {
		tokens[cnt] = strstrip(tokens[cnt], arg2);
		tokens[++cnt] = strsep(&str, arg3);
	}

#else
	tokens[0] = strtok(str, arg1);
	while (tokens[cnt] && (cnt < MAX_TOKENS - 1)) {
		tokens[cnt] = strstrip(tokens[cnt], arg2);
		tokens[++cnt] = strtok(NULL, arg3);
	}
#endif
	return cnt;
}



/*====================================================================
 * 	Pattern matching function in incoming calls.
 *====================================================================*/

static int is_match_found(unsigned char info[], unsigned char chan[])
{
	unsigned char temp[WAN_ADDRESS_SZ+1];

	/* accept all calls */
	if (strcmp(chan, "*") == 0)	return 0;
	
	/* case ?? - must exactly match */
	if (chan[strlen(chan) - 1] != '*' && chan[0] != '*' && strlen(chan)){
		if (strcmp(info, chan) == 0)	return 0;
	}
		
	/* case ??* must match at the begining */
	if ( chan[0] != '*' && chan[strlen(chan) - 1] == '*'){
		if (strncmp(info, &chan[0], strlen(chan) - 1) == 0)	return 0;
	}
		
	/* case *?? must match at the end */
	if ( chan[0] == '*' && chan[strlen(chan) - 1] != '*'){	
		if (strncmp(&info[strlen(info) - strlen(chan)+1], &chan[1], strlen(chan) - 1) == 0)
			return 0;
	}
		
	/* case *??* must match somewhere in the string */
	if ( chan[0] == '*' && chan[strlen(chan) - 1] == '*'){
			
		strncpy(temp, &chan[1], strlen(chan) - 2);
		/* null terminate for strstr() */
		temp[strlen(chan) - 2] = '\0';
			
		if (strstr(info, temp))	return 0;
	}

	/* if got here the string is empty - do not accept call */
	return -EINVAL;
}

int x25_match_accept_criteria(x25_lcn_t *x25_lcn, char **token, int toknum)
{
	int i;
	int match=0;
	int called=0,calling=0,facil=0,data=0;
	
	for (i=0;i<toknum;i++){
#if 0
		printk(KERN_INFO "%s: Type=%c I=%s, L=%s\n",
				x25_lcn->dev->name,token[i][0],
				&token[i][1],
				(token[i][0] == 'd') ? x25_lcn->accept_called:
							(token[i][0] == 's') ? x25_lcn->accept_calling :
										x25_lcn->accept_udata);
#endif
		switch (token[i][0]){

		case 'd':
			match = is_match_found(&token[i][1],x25_lcn->accept_called);
			called=1;
			if (match != X25_OK)
				match=67;
			break;
		case 's':
			calling=1;
			match = is_match_found(&token[i][1],x25_lcn->accept_calling);
			if (match != X25_OK)
				match=68;
			break;	
		case 'f':
			facil=1;
			match = is_match_found(&token[i][1],x25_lcn->accept_facil);
			break;
		case 'u':
			data=1;
			match = is_match_found(&token[i][1],x25_lcn->accept_udata);
			if (match != X25_OK)
				match=64;
			break;			

		default:
			printk(KERN_INFO "%s: Invallid incoming call string!\n",
					x25_lcn->dev->name);
			return 64;
		}

		if (match != X25_OK)
			return match;
	}

	if (!called){
		printk(KERN_INFO "No Called Address in call string\n");
		return 67;
	}
	
	if (!calling && strlen(x25_lcn->accept_calling)){
		printk(KERN_INFO "No Calling Address in call string\n");
		return 71;
	}

	if (!facil && strlen(x25_lcn->accept_facil)){
		printk(KERN_INFO "No Facil Address in call string\n");
		return 69;
	}
	
	if (!data && strlen(x25_lcn->accept_udata)){
		printk(KERN_INFO "No User Data in call string\n");
		return 71;
	}
	
	
	return match;
}

void PTI_err_LCN_0_pkt_on_non_0_LCN(x25_link_t *x25_link, 
				    unsigned char diag_code, 
				    struct x25_frame *frame)
{

	/* This error handler includes the following type of 
	 * received packets:
	 *    - packets having a Packet Type Identifier which is 
	 * 	shorter than 1 octet and having a LCN of 0 (C2-A)
	 *    - packets having a Packet Type Identifier which is 
	 * 	shorter than 1 octet and having an assigned LCN (C2-B,
	 * 	C3-A, C4-A)
	 *    - packets having a Packet Type Identifier which is 
	 *      undefined or not supported and having a LCN of 0 (C2-C)
	 *    - packets having a Packet Type Identifier which is undefined 
	 *      or not supported and having an assigned LCN (C2-D C3-B, C4-B)
	 *    - Restart Indication/Confirmation and Registration 
	 *    - Request/Confirmation packets witha LCN unequal to 0 
	 *      (C2-E, C3-C, C4-C)
	 */

	if(!frame->lcn) {	
		/* handle a packet received on LCN zero */

		switch(x25_link->state) {

		case X25_ST_RESTART_REQ_REC_R3:					
			/* if we have received a Restart Request, 
			 * then we issue a Restart with diagnostic code (C2-A3, C2-C3)*/
			x25_send_control(x25_link,0,0,RESTART_REQUEST_PKT,0,diag_code, NULL,0);		
			break;

		case X25_ST_RESTART_REQ_SNT_R2:		
			/* if we have just issued a Restart Request, 
			 * then we discard the packet (C2-A2, C2-C2) */
			break;

		default:
			/* for state r1, we issue a Diagnostic packet (C2-A1, C2-C1) */
			//FIXME
			//issue_Diagnostic_pkt(diag_code);
			break;
		}
	}
}

int x25_update_stats (x25_link_t *x25_link, unsigned int type, unsigned char direction)
{

	switch (type){

		case DATA_PKT:
			if (!direction)
				x25_link->x25stats.rx_data_pkt++;
			else
				x25_link->x25stats.tx_data_pkt++;
			break;

		case RR_PKT:
			if (!direction)
				x25_link->x25stats.rx_rr++;
			else
				x25_link->x25stats.tx_rr++;
			break;

		case RNR_PKT:
			if (!direction)
				x25_link->x25stats.rx_rnr++;
			else
				x25_link->x25stats.tx_rnr++;
			break;

		case REJ_PKT:
			if (!direction)
				x25_link->x25stats.rx_rej++;
			else
				x25_link->x25stats.tx_rej++;
			break;

		case CALL_REQUEST_PKT:
			if (!direction)
				x25_link->x25stats.rx_call_req++;
			else
				x25_link->x25stats.tx_call_req++;
			break;

		case CALL_ACCEPT_PKT:
			if (!direction)
				x25_link->x25stats.rx_call_acpt++;
			else
				x25_link->x25stats.tx_call_acpt++;
			break;
	
		case CLEAR_REQUEST_PKT:
			if (!direction)
				x25_link->x25stats.rx_clear_req++;
			else
				x25_link->x25stats.tx_clear_req++;
			break;

		case CLEAR_CONF_PKT:
			if (!direction)
				x25_link->x25stats.rx_clear_conf++;
			else
				x25_link->x25stats.tx_clear_conf++;
			break;
	
		case RESET_REQUEST_PKT:
			if (!direction)
				x25_link->x25stats.rx_reset_req++;
			else
				x25_link->x25stats.tx_reset_req++;
			break;
		
		case RESET_CONF_PKT:
			if (!direction)
				x25_link->x25stats.rx_reset_conf++;
			else
				x25_link->x25stats.tx_reset_conf++;
			break;

		case RESTART_REQUEST_PKT:
			
			DEBUG_TEST("%s: %s Restart Request\n",
					x25_link->name,
					direction ? "TX" : "RX");
			
			if (!direction)
				x25_link->x25stats.rx_restrt_req++;
			else
				x25_link->x25stats.tx_restrt_req++;
			break;

		case RESTART_CONF_PKT:

			DEBUG_TEST("%s: %s Restart Conf\n",
					x25_link->name,
					direction ? "TX" : "RX");
			
			if (!direction)
				x25_link->x25stats.rx_restrt_conf++;
			else
				x25_link->x25stats.tx_restrt_conf++;
			break;

		case INTERRUPT_PKT:
			if (!direction)
				x25_link->x25stats.rx_intr_req++;
			else
				x25_link->x25stats.tx_intr_req++;
			break;

		case INTERRUPT_CONF_PKT:
			if (!direction)
				x25_link->x25stats.rx_intr_conf++;
			else
				x25_link->x25stats.tx_intr_conf++;
			break;

		case REGISTRATION_REQUEST_PKT:
			if (!direction)
				x25_link->x25stats.rx_reg_req++;
			else
				x25_link->x25stats.tx_reg_req++;
			break;

		case REGISTRATION_CONF_PKT:
			if (!direction)
				x25_link->x25stats.rx_reg_conf++;
			else
				x25_link->x25stats.tx_reg_conf++;
			break;
			
		case DIAGNOSTIC_PKT:
			if (!direction)
				x25_link->x25stats.rx_diag++;
			else
				x25_link->x25stats.tx_diag++;
			break;
		default:
			if (!direction)
				x25_link->x25stats.rx_ill++;
			else
				x25_link->x25stats.tx_ill++;
			break;
	}

	if (!direction){
		x25_link->x25stats.rx_total++;
	}else{
		x25_link->x25stats.tx_total++;
	}
	return 0;
}

void x25_log_proc_event(x25_link_t *x25_link, unsigned char *lcn_dev_name,
		        unsigned char *event_name, 
		        unsigned short lcn, unsigned char cause, 
			unsigned char diag, unsigned char *call_string)
{
	int cnt=0;
	unsigned char * event_buf;
	unsigned int offset;
	struct timeval tv;

	do_gettimeofday(&tv);
	
	offset=atomic_read(&x25_link->proc_event_offset);
	
	if (offset >= MAX_PROC_EVENTS){
		offset=0;
	}
	event_buf=&x25_link->proc_event_log[offset][0];

	if (call_string && strlen(call_string)){
		cnt=sprintf(event_buf,"%lu:%-7lu : %-10s : %-21s : Lcn=%-3i : Cause=0x%02x : Diag=0x%02x : %s ",
				tv.tv_sec,tv.tv_usec,lcn_dev_name?lcn_dev_name:x25_link->name,
				event_name, lcn, cause, diag, call_string);
	}else{
		cnt=sprintf(event_buf,"%lu:%-7lu : %-10s : %-21s : Lcn=%-3i : Cause=0x%02x : Diag=0x%02x",
				tv.tv_sec,tv.tv_usec,lcn_dev_name?lcn_dev_name:x25_link->name,
				event_name, lcn, cause, diag);
	}

	offset++;
	atomic_set(&x25_link->proc_event_offset,offset);
}

/*============================================================================
 * Convert decimal string to unsigned integer.
 * If len != 0 then only 'len' characters of the string are converted.
 */
unsigned int dec_to_uint (unsigned char* str, int len)
{
	unsigned val;

	if (!len) 
		len = strlen(str);

	for (val = 0; len && is_digit(*str); ++str, --len)
		val = (val * 10) + (*str - (unsigned)'0');

	return val;
}
	
