
#include <linux/version.h>
#include <linux/config.h>	/* OS configuration options */
#include <linux/stddef.h>	/* offsetof(), etc. */
#include <linux/errno.h>	/* return codes */
#include <linux/string.h>	/* inline memset(), etc. */
#include <linux/slab.h>		/* kmalloc(), kfree() */
#include <linux/kernel.h>	/* printk(), and other useful stuff */
#include <linux/module.h>	/* support for loadable modules */
#include <linux/ioport.h>	/* request_region(), release_region() */

#include <linux/in.h>		/* sockaddr_in */
#include <linux/inet.h>	
#include <linux/if.h>

#include <linux/pci.h>
#include <linux/if_arp.h>	/* ARPHRD_* defines */

#include <asm/uaccess.h>
#include <linux/netdevice.h>

#include "sdla_t3_hdlc.h"

MODULE_AUTHOR ("Nenad Corbic <ncorbic@sangoma.com>");
MODULE_DESCRIPTION ("Sangoma WANPIPE: Test Driver");
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,9) 
MODULE_LICENSE("GPL");
#endif

#define MODNAME  "test-driver"

#define MTAG "sdla_t3_hdlc"

static struct net_device *dev;
static struct pci_dev *pci_dev;

static void *txdma_buf;
static unsigned int txdma_buf_len;
static unsigned int txdma_buf_data_sz;

static void *rxdma_buf;
static unsigned int rxdma_buf_len;


#define SDLA_HDLC_T3_PCI_VENDOR_ID 0x10ee
#define SDLA_HDLC_T3_PCI_DEVICE_ID 0x0340

#define	HDLC_DEV_ALLOC 		0x00000001
#define	HDLC_DEV_PRIV_ALLOC 	0x00000002
#define	HDLC_PCI_BAR0_REQ	0x00000004
#define HDLC_PCI_BAR1_REQ	0x00000008
#define HDLC_PCI_BAR0_IOREMAP	0x00000010
#define HDLC_PCI_BAR1_IOREMAP	0x00000020
#define HDLC_PCI_ENABLE		0x00000040
#define HDLC_PCI_IRQ		0x00000080

#define MAX_RX_DMA_BUF_LEN 	5000

static int hdlc_module_status;

void unregister_hw(void)
{
	struct sdla_hdlc_priv *hdlc_priv=NULL;	
	
	if (!dev){
		return;
	}

	if (dev->priv){
		hdlc_priv = dev->priv;
	}

	if (pci_dev){
		
		if (hdlc_module_status & HDLC_PCI_IRQ){
			hdlc_module_status &= ~HDLC_PCI_IRQ;
			free_irq(pci_dev->irq,dev);
		}
		
		if (hdlc_module_status & HDLC_PCI_ENABLE){
			hdlc_module_status &= ~HDLC_PCI_ENABLE;
			pci_disable_device(pci_dev);
		}
		
		if (hdlc_priv && hdlc_module_status & HDLC_PCI_BAR0_IOREMAP){
			hdlc_module_status &= ~HDLC_PCI_BAR0_IOREMAP;
			iounmap(hdlc_priv->bar0);
		}

		if (hdlc_priv && hdlc_module_status & HDLC_PCI_BAR1_IOREMAP){
			hdlc_module_status &= ~HDLC_PCI_BAR1_IOREMAP;
			iounmap(hdlc_priv->bar1);
		}

		if (hdlc_module_status & HDLC_PCI_BAR0_REQ){
			hdlc_module_status &= ~HDLC_PCI_BAR0_REQ;
			release_mem_region(pci_resource_start(pci_dev, 0),
				   pci_resource_len(pci_dev, 0));
		}

		if (hdlc_module_status & HDLC_PCI_BAR1_REQ){
			hdlc_module_status &= ~HDLC_PCI_BAR1_REQ;
			release_mem_region(pci_resource_start(pci_dev, 1),
				   pci_resource_len(pci_dev, 1));
		}

	}
	if (dev->priv){
		kfree(dev->priv);
		dev->priv=NULL;
	}
}

int hdlc_if_open(struct net_device *dev)
{
	if (netif_running(dev))
		return -EBUSY;

	MOD_INC_USE_COUNT;
	return 0;
}

int hdlc_if_close(struct net_device *dev)
{
	MOD_DEC_USE_COUNT;
	return 0;
	
}

int hdlc_if_do_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
	struct sdla_hdlc_priv *chan= (struct sdla_hdlc_priv *)dev->priv;
	int err=-EINVAL;
	struct sdla_hdlc_api api_cmd;
	void *vector;
		
	if (!chan){
		return -ENODEV;
	}

	if (!ifr->ifr_data){
		return -EINVAL;
	}

	if (copy_from_user(&api_cmd,ifr->ifr_data,sizeof(struct sdla_hdlc_api))){
		return -EFAULT;
	}

	switch (cmd){

	case SDLA_HDLC_READ_REG:
		if (api_cmd.bar == 0){
			vector=chan->bar0 + api_cmd.offset;
		}else{
			vector=chan->bar1 + api_cmd.offset;
		}
	
		memcpy_fromio((unsigned char*)&api_cmd.data[0],
				(unsigned char*)vector, api_cmd.len);

		printk(KERN_INFO "%s: Reading Bar%i Offset=%i Len=%i\n",
				MODNAME,api_cmd.bar,api_cmd.offset,api_cmd.len);
		
		err=0;
		break;
	
	case SDLA_HDLC_WRITE_REG:
		if (api_cmd.bar == 0){
			vector=chan->bar0 + api_cmd.offset;
		}else{
			vector=chan->bar1 + api_cmd.offset;
		}

		printk(KERN_INFO "%s: Writting Bar%i Offset=%i Len=%i\n",
				MODNAME,api_cmd.bar,api_cmd.offset,api_cmd.len);

		if (api_cmd.len == 1){
			writeb(*(unsigned char*)&api_cmd.data[0],vector);
		}else if (api_cmd.len == 2){
			writew(*(unsigned short*)&api_cmd.data[0],vector);
		}else if (api_cmd.len == 4){
			printk(KERN_INFO "%s: Write Data 0x%x\n",
					MODNAME,*(unsigned int*)&api_cmd.data[0]);
			writel(*(unsigned int*)&api_cmd.data[0],vector);
		}else{
			memcpy_toio((unsigned char*)vector,(unsigned char*)&api_cmd.data[0],api_cmd.len);
		}
		err=0;
		break;

	case SDLA_HDLC_FREE_RXDMA:

		if (rxdma_buf && rxdma_buf_len){
			free_pages((unsigned long)rxdma_buf, get_order(rxdma_buf_len));	
			rxdma_buf=0;
			rxdma_buf_len=0;
			err=0;
		}else{
			err=-ENOBUFS;
		}
		
		break;

	case SDLA_HDLC_GET_RXDMA:

		if (rxdma_buf && rxdma_buf_len){
			printk(KERN_INFO "%s: Get Rx DMA failed, already allocated\n",
					MODNAME);
			return -EEXIST;
		}
		if (api_cmd.len == 0 || api_cmd.len > MAX_RX_DMA_BUF_LEN){
			printk(KERN_INFO "%s: Get Rx DMA Error, len=%i > Maxlen=%i\n",
					MODNAME,api_cmd.len, MAX_RX_DMA_BUF_LEN);
			return -EINVAL;
		}
		
		rxdma_buf = (void *)__get_free_pages(GFP_ATOMIC, get_order(api_cmd.len));
		if (rxdma_buf){
			rxdma_buf_len=api_cmd.len;
			*(unsigned int*)api_cmd.data = (unsigned int)rxdma_buf;
			api_cmd.len=4;
			err=0;

			memset(rxdma_buf,0,api_cmd.len);

			printk(KERN_INFO "%s: Rx DMA allocate at 0x%x Size=%i!\n",
					MODNAME, (unsigned int)rxdma_buf,rxdma_buf_len);

		}else{
			printk(KERN_INFO "%s: Get Rx DMA failed to allocate memory!\n",
					MODNAME);
			*(unsigned int*)api_cmd.data = 0;
			err=-ENOMEM;
		}
		break;

	case SDLA_HDLC_READ_RXDMA:

		
		if (!rxdma_buf || !rxdma_buf_len){
			printk(KERN_INFO "%s: Read Rx DMA failed, not allocated\n",
					MODNAME);
			return -ENOBUFS;
		}
		
		if (api_cmd.len == 0 || api_cmd.len > rxdma_buf_len){
			printk(KERN_INFO "%s: Read Rx DMA Error, len=%i > Maxlen=%i\n",
					MODNAME,api_cmd.len, rxdma_buf_len);
			return -EINVAL;
		}

		err=0;
		{
			int i;
			printk(KERN_INFO "%s: Read Rx DMA OK: ",MODNAME);
			for (i=0;i<api_cmd.len;i++){
				api_cmd.data[i]=*((unsigned char*)rxdma_buf+i);
				printk("%x ",*((unsigned char*)rxdma_buf+i));
			}
			printk("\n");
		}

		break;

	case SDLA_HDLC_FREE_TXDMA:
		
		if (txdma_buf && txdma_buf_len){
			free_pages((unsigned long)txdma_buf, get_order(txdma_buf_len));	
			txdma_buf=0;
			txdma_buf_len=0;
			err=0;
		}else{
			err=-ENOBUFS;
		}
		break;
			
	case SDLA_HDLC_GET_TXDMA:
		
		if (txdma_buf && txdma_buf_len){
			printk(KERN_INFO "%s: Get Tx DMA failed, already allocated\n",
					MODNAME);
			return -EEXIST;
		}

		
		if (api_cmd.len == 0 || api_cmd.len > MAX_RX_DMA_BUF_LEN){
			printk(KERN_INFO "%s: Get Tx DMA Error, len=%i > Maxlen=%i\n",
					MODNAME,api_cmd.len, MAX_RX_DMA_BUF_LEN);
			return -EINVAL;
		}
		
		txdma_buf = (void *)__get_free_pages(GFP_ATOMIC, get_order(api_cmd.len));
		if (txdma_buf){
			txdma_buf_len=api_cmd.len;
			*(unsigned int*)api_cmd.data = (unsigned int)txdma_buf;
			api_cmd.len=4;

			memset(txdma_buf,0,api_cmd.len);

			printk(KERN_INFO "%s: Tx DMA allocate at 0x%x Size=%i!\n",
					MODNAME,(unsigned int)txdma_buf,txdma_buf_len);
			err=0;
		}else{
			txdma_buf_len=0;
			printk(KERN_INFO "%s: Get Tx DMA failed to allocate memory!\n",
					MODNAME);
			*(unsigned int*)api_cmd.data = 0;
			err=-ENOMEM;
		}
		break;

	case SDLA_HDLC_SET_TXDMA:
	
		if (!txdma_buf || !txdma_buf_len){
			printk(KERN_INFO "%s: Set Tx DMA failed, not allocated\n",MODNAME);
			return -ENOBUFS;
		}
		
		if (api_cmd.len == 0 || api_cmd.len > txdma_buf_len){
			printk(KERN_INFO "%s: Set Tx DMA Error, len=%i > Maxlen=%i\n",
					MODNAME,api_cmd.len, MAX_RX_DMA_BUF_LEN);
			return -EINVAL;
		}

		memcpy(txdma_buf,api_cmd.data,api_cmd.len);
		txdma_buf_data_sz=api_cmd.len;

		{
			int i;
			printk(KERN_INFO "%s: Set Tx DMA OK: ",MODNAME);
			for (i=0;i<txdma_buf_data_sz;i++){
				printk("%x ",*((unsigned char*)txdma_buf+i));
			}
			printk("\n");
		}
		break;
		
	default:
		
		break;
	}

	if (copy_to_user(ifr->ifr_data,&api_cmd,sizeof(struct sdla_hdlc_api))){
		return -EFAULT;
	}

	
	return err;
	
}

void sdla_hdlc_isr (int irq, void* dev_id, struct pt_regs *regs)
{

	struct net_device *dev = (struct net_device*)dev_id;	

	if (!dev->priv){
		if (net_ratelimit()){
			printk(KERN_INFO "%s: ISR: no priv struct\n",MODNAME);
		}
		return;
	}

	if (net_ratelimit()){
		printk(KERN_INFO "%s: ISR Executed for device %s\n",MODNAME,dev->name);
	}
	return;
}


int init_module (void)
{
	int err;
	struct sdla_hdlc_priv *hdlc_priv;	
	dev=NULL;
	pci_dev=NULL;

	hdlc_module_status=0;
	
	printk(KERN_INFO "%s: init_module()\n",
				MODNAME);

	pci_dev = pci_find_device(SDLA_HDLC_T3_PCI_VENDOR_ID, SDLA_HDLC_T3_PCI_DEVICE_ID, pci_dev);
	if (!pci_dev){
		printk(KERN_INFO "%s: Failed to find device\n",
				MODNAME);
		return -ENODEV;
	}
	
	dev = kmalloc(sizeof(struct net_device),GFP_KERNEL);
	if (!dev){
		printk(KERN_INFO "%s: Failed to allocate dev\n",
				MODNAME);
		return -ENOMEM;
	}

	memset(dev,0,sizeof(struct net_device));

	dev->priv=kmalloc(sizeof(struct sdla_hdlc_priv),GFP_KERNEL);
	if (!dev->priv){
		err=-ENOMEM;
		goto sdla_init_exit;
	}

	memset(dev->priv,0,sizeof(struct sdla_hdlc_priv));
	
	hdlc_priv = dev->priv;
		
	sprintf(dev->name,"whdlc0");

	dev->open		=&hdlc_if_open;
	dev->stop		=&hdlc_if_close;
//	dev->init		=&hdlc_if_init;
//	dev->hard_start_xmit	=&hdlc_if_send;
//	dev->get_stats          =&hdlc_if_get_stats;
//	dev->tx_timeout		=&hdlc_if_tx_timeout;
//	dev->watchdog_timeo     = (1*HZ);	
	dev->do_ioctl		=&hdlc_if_do_ioctl;

	/* Initialize media-specific parameters */
	dev->flags		|= IFF_POINTOPOINT;
	dev->flags		|= IFF_NOARP;

	dev->type		= ARPHRD_PPP;
	dev->mtu		= 1500;
	dev->hard_header_len	= 0;
        dev->tx_queue_len 	= 100;

	if (!request_mem_region(pci_resource_start(pci_dev, 0),
			        pci_resource_len(pci_dev, 0), MTAG)){
		printk(KERN_INFO "%s: Error: Unable to reserve Bar0 PCI adapter memory\n",
				MODNAME);
        	err = -EFAULT;
		goto sdla_init_exit;
    	}
	hdlc_module_status|=HDLC_PCI_BAR0_REQ;
	
	if (!request_mem_region(pci_resource_start(pci_dev, 1),
			        pci_resource_len(pci_dev, 1), MTAG)){
		printk(KERN_INFO "%s: Error: Unable to reserve Bar1 PCI adapter memory\n",
				MODNAME);
        	err= -EFAULT;
		goto sdla_init_exit;
    	}
	hdlc_module_status|=HDLC_PCI_BAR1_REQ;

	hdlc_priv->bar0 = ioremap(pci_resource_start(pci_dev, 0),
				   pci_resource_len(pci_dev, 0));
	if (!hdlc_priv->bar0){
		printk(KERN_INFO "%s: Failed to remap Bar0\n",MODNAME);
		err= -EFAULT;
		goto sdla_init_exit;
	}	

	hdlc_module_status|=HDLC_PCI_BAR0_IOREMAP;
	
	hdlc_priv->bar1 = ioremap(pci_resource_start(pci_dev, 1),
				   pci_resource_len(pci_dev, 1));

	if (!hdlc_priv->bar1){
		printk(KERN_INFO "%s: Failed to remap Bar1\n",MODNAME);
		err= -EFAULT;
		goto sdla_init_exit;
	}
	hdlc_module_status|=HDLC_PCI_BAR1_IOREMAP;
		
	if (pci_enable_device(pci_dev)){
        	printk(KERN_INFO "%s: Error: Unable enable PCI device\n",MODNAME);
		err= -EFAULT;
		goto sdla_init_exit;
    	}

	pci_set_master(pci_dev);
	hdlc_module_status|=HDLC_PCI_ENABLE;

	if (request_irq(pci_dev->irq,sdla_hdlc_isr, 
		       SA_SHIRQ, 
		       MODNAME, dev)){
		printk(KERN_INFO "%s: Can't reserve IRQ %d!\n", MODNAME, pci_dev->irq);
		err= -EFAULT;
		goto sdla_init_exit;
	}

	hdlc_module_status|=HDLC_PCI_IRQ;
	
	printk(KERN_INFO "%s: PCI Device loaded OK\n",
			MODNAME);

	printk(KERN_INFO "%s: Reseting Fifo\n",MODNAME);
	writel(0x80000000,hdlc_priv->bar0);
	printk(KERN_INFO "%s: Bar0 Reg 0 = %x\n", MODNAME,readl(hdlc_priv->bar0));
	
	err = register_netdev(dev);
	if (!err){
		return 0;	
	}

sdla_init_exit:

	unregister_hw();
	
	if (dev){
		if (dev->priv){
			kfree(dev->priv);
			dev->priv=NULL;
		}
		kfree(dev);
		dev=NULL;
	}

	return err;
}


void cleanup_module (void)
{
	struct sdla_hdlc_priv *hdlc_priv;

	printk(KERN_INFO "%s: cleanup_module()\n",
				MODNAME);

	if (!dev){
		return; 
	}

	hdlc_priv = dev->priv;
	
	unregister_netdev(dev);

	unregister_hw();
	
	if (dev->priv){
		kfree(dev->priv);
		dev->priv=NULL;
	}

	if (txdma_buf && txdma_buf_len){
		free_pages((unsigned long)txdma_buf, get_order(txdma_buf_len));	
		txdma_buf=0;
		txdma_buf_len=0;
	}

	if (rxdma_buf && rxdma_buf_len){
		free_pages((unsigned long)rxdma_buf, get_order(rxdma_buf_len));	
		rxdma_buf=0;
		rxdma_buf_len=0;
	}
	
	kfree(dev);
	dev=NULL;
}
