/*****************************************************************************
* sdla_xilinx.c	WANPIPE(tm) S51XX Xilinx Hardware Support
* 		
* Authors: 	Nenad Corbic <ncorbic@sangoma.com>
*
* Copyright:	(c) 2003 Sangoma Technologies Inc.
*
*		This program is free software; you can redistribute it and/or
*		modify it under the terms of the GNU General Public License
*		as published by the Free Software Foundation; either version
*		2 of the License, or (at your option) any later version.
* ============================================================================
* Jan 07, 2003	Nenad Corbic	Initial version.
*****************************************************************************/

#include <linux/wanpipe_includes.h>
#include <linux/wanpipe_defines.h>
#include <linux/wanpipe.h>	
#include <linux/wanproc.h>

#include <linux/wanpipe_abstr.h>
#include <linux/sdla_xilinx.h>

#include <linux/if_wanpipe_common.h>    /* Socket Driver common area */
#include <linux/if_wanpipe.h>		


/****** Defines & Macros ****************************************************/

/* Private critical flags */
enum { 
	POLL_CRIT = PRIV_CRIT, 
	TX_BUSY,
	RX_BUSY,
	TASK_POLL
};

#define MAX_IP_ERRORS	10

#define PORT(x)   (x == 0 ? "PRIMARY" : "SECONDARY" )
#define MAX_TX_BUF	10
#define MAX_RX_BUF	10

/******Data Structures*****************************************************/

/* This structure is placed in the private data area of the device structure.
 * The card structure used to occupy the private area but now the following 
 * structure will incorporate the card structure along with Protocol specific data
 */

typedef struct private_area
{
	wanpipe_common_t 	common;
	sdla_t			*card;
	struct net_device	*dev;

	struct sk_buff_head 	wp_tx_free_list;
	struct sk_buff_head 	wp_tx_pending_list;
	struct sk_buff_head 	wp_tx_complete_list;
	struct sk_buff 		*tx_dma_skb;
	void                    *tx_dma_buf;

	struct sk_buff_head 	wp_rx_free_list;
	struct sk_buff_head 	wp_rx_complete_list;
	struct sk_buff 		*rx_dma_skb;
	void			*rx_dma_buf;

	unsigned long 		time_slot_map;
	unsigned char 		num_of_time_slots;
	long          		logic_ch_num;

	wan_tasklet_t 		chan_task;
	unsigned char		hdlc_eng;
	unsigned char		dma_status;

	struct net_device_stats	if_stats;

#if 1
	int 		TracingEnabled;		/* For enabling Tracing */
	unsigned long 	curr_trace_addr;	/* Used for Tracing */
	unsigned long 	start_trace_addr;
	unsigned long 	end_trace_addr;
	unsigned short 	number_trace_elements;
	unsigned  	available_buffer_space;
	unsigned long 	router_start_time;
	unsigned char 	route_status;
	unsigned char 	route_removed;
	unsigned long 	tick_counter;		/* For 5s timeout counter */
	unsigned long 	router_up_time;
        u32             IP_address;		/* IP addressing */
        u32             IP_netmask;
	u32		ip_local;
	u32		ip_remote;
	u32 		ip_local_tmp;
	u32		ip_remote_tmp;
	u8		ip_error;
	u8		config_frmw;
	u8 		config_frmw_timeout;
	unsigned char  	mc;			/* Mulitcast support on/off */
	unsigned char 	udp_pkt_src;		/* udp packet processing */
	unsigned short 	timer_int_enabled;

	unsigned char 	interface_down;

	/* Polling task queue. Each interface
         * has its own task queue, which is used
         * to defer events from the interrupt */
	struct tq_struct 	poll_task;
	struct timer_list 	poll_delay_timer;

	u8 		gateway;
	u8 		true_if_encoding;
	
	//FIXME: add driver stats as per frame relay!
#endif
	
	/* Entry in proc fs per each interface */
	struct proc_dir_entry	*dent;

	unsigned char 	udp_pkt_data[sizeof(wan_udp_pkt_t)+10];
	atomic_t 	udp_pkt_len;

	char 		if_name[WAN_IFNAME_SZ+1];

}private_area_t;

/* Route Status options */
#define NO_ROUTE	0x00
#define ADD_ROUTE	0x01
#define ROUTE_ADDED	0x02
#define REMOVE_ROUTE	0x03


/* variable for keeping track of enabling/disabling FT1 monitor status */
static int rCount;
static int Intr_test_counter;

extern void disable_irq(unsigned int);
extern void enable_irq(unsigned int);

/**SECTOIN**************************************************
 *
 * Function Prototypes 
 *
 ***********************************************************/

/* WAN link driver entry points. These are called by the WAN router module. */
static int 	update (wan_device_t* wandev);
static int 	new_if (wan_device_t* wandev, struct net_device* dev, wanif_conf_t* conf);
static int 	del_if(wan_device_t *wandev, struct net_device *dev);

/* Network device interface */
static int 	if_init   (struct net_device* dev);
static int 	if_open   (struct net_device* dev);
static int 	if_close  (struct net_device* dev);
static int 	if_do_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd);

static struct net_device_stats* if_stats (struct net_device* dev);
  
static int 	if_send (struct sk_buff* skb, struct net_device* dev);

/* Firmware interface functions */
static int 	frmw_error (sdla_t *card, int err, wan_mbox_t *mb);
static void 	handle_front_end_state(sdla_t *card);
static void 	enable_timer(void* card_id);
#ifdef LINUX_2_4
static void 	if_tx_timeout (struct net_device *dev);
#endif

/* Miscellaneous Functions */
static void 	port_set_state (sdla_t *card, int);

static void 	disable_comm (sdla_t *card);

/* Interrupt handlers */
static void 	wp_xilinx_isr (sdla_t* card);

/* Bottom half handlers */
static void 	wp_bh (unsigned long);
static void 	wakeup_sk_bh (struct net_device *dev);

/* Miscellaneous functions */
static int 	chk_bcast_mcast_addr(sdla_t* card, struct net_device* dev,
				struct sk_buff *skb);
#if 0
static int 	process_udp_mgmt_pkt(sdla_t* card, struct net_device* dev,  
				private_area_t*,
				int local_dev);
#endif

static int xilinx_chip_configure(sdla_t *card, wandev_conf_t* conf);
static int xilinx_dev_configure(sdla_t *card, private_area_t *chan, wanif_conf_t* conf);
static void xilinx_dev_unconfigure(sdla_t *card, private_area_t *chan);
static int xilinx_dev_open(sdla_t *card, private_area_t *chan);
static void xilinx_dev_close(sdla_t *card, private_area_t *priv_area);
static int xilinx_dma_tx (sdla_t *card, private_area_t *chan);
static void xilinx_dma_tx_complete (sdla_t *card, private_area_t *chan);
static void xilinx_dma_rx_complete (sdla_t *card, private_area_t *chan);
static void xilinx_dma_max_logic_ch(sdla_t *card);

static char request_xilinx_logical_channel_num (sdla_t *card, private_area_t *chan, long *free_ch);
static void free_xilinx_logical_channel_num (sdla_t *card, int logic_ch);



/* TE1 Control registers  */
static WRITE_FRONT_END_REG_T write_front_end_reg;
static READ_FRONT_END_REG_T  read_front_end_reg;

static void xilinx_delay(int sec)
{
	unsigned long timeout=jiffies;
	while ((jiffies-timeout)<(sec*HZ)){
		schedule();
	}
}

/**SECTION********************************************************* 
 * 
 * Public Functions 
 *
 ******************************************************************/


/*============================================================================
 * wp_xilinx_init - Cisco HDLC protocol initialization routine.
 *
 * @card:	Wanpipe card pointer
 * @conf:	User hardware/firmware/general protocol configuration
 *              pointer. 
 * 
 * This routine is called by the main WANPIPE module 
 * during setup: ROUTER_SETUP ioctl().  
 *
 * At this point adapter is completely initialized 
 * and firmware is running.
 *  o read firmware version (to make sure it's alive)
 *  o configure adapter
 *  o initialize protocol-specific fields of the adapter data space.
 *
 * Return:	0	o.k.
 *		< 0	failure.
 */

int wp_xilinx_init (sdla_t* card, wandev_conf_t* conf)
{
	int err;

	/* Verify configuration ID */

	if (conf->config_id != WANCONFIG_XILINX) {
		DEBUG_EVENT( "%s: invalid configuration ID %u!\n",
				  card->devname, conf->config_id);
		return -EINVAL;
	}

	/* Obtain hardware configuration parameters */
	card->wandev.clocking 			= conf->clocking;
	card->wandev.ignore_front_end_status 	= conf->ignore_front_end_status;
	card->wandev.ttl 			= conf->ttl;
	card->wandev.interface 			= conf->interface; 
	card->wandev.comm_port 			= conf->comm_port;
	card->wandev.udp_port   		= conf->udp_port;
	card->wandev.new_if_cnt 		= 0;
	atomic_set(&card->wandev.if_cnt,0);

	/* TE1 Make special hardware initialization for T1/E1 board */
	if (IS_TE1(conf->te_cfg)){
		
		memcpy(&card->wandev.te_cfg, &conf->te_cfg, sizeof(sdla_te_cfg_t));
		card->wandev.write_front_end_reg = write_front_end_reg;
		card->wandev.read_front_end_reg = read_front_end_reg;
		card->wandev.te_enable_timer = enable_timer;
		conf->interface = 
			(card->wandev.te_cfg.media == WANOPT_MEDIA_T1) ? WANOPT_V35 : WANOPT_RS232;

		if (card->wandev.comm_port == WANOPT_PRI){
			conf->clocking = WANOPT_EXTERNAL;
		}

	}else if (IS_56K(conf->te_cfg)){

		memcpy(&card->wandev.te_cfg, &conf->te_cfg, sizeof(sdla_te_cfg_t));
		card->wandev.write_front_end_reg = write_front_end_reg;
		card->wandev.read_front_end_reg = read_front_end_reg;
		
		if (card->wandev.comm_port == WANOPT_PRI){
			conf->clocking = WANOPT_EXTERNAL;
		}

	}else{
		card->wandev.front_end_status = FE_CONNECTED;
	}

	if (card->wandev.ignore_front_end_status == WANOPT_NO){
		DEBUG_EVENT( 
		  "%s: Enabling front end link monitor\n",
				card->devname);
	}else{
		DEBUG_EVENT( 
		"%s: Disabling front end link monitor\n",
				card->devname);
	}


	card->isr			= &wp_xilinx_isr;
	card->wandev.update		= &update;
 	card->wandev.new_if		= &new_if;
	card->wandev.del_if		= &del_if;
	card->disable_comm		= &disable_comm;

#ifdef WANPIPE_ENABLE_PROC_FILE_HOOKS
	/* Proc fs functions hooks */
	card->wandev.get_config_info 	= &get_config_info;
	card->wandev.get_status_info 	= &get_status_info;
	card->wandev.get_dev_config_info= &get_dev_config_info;
	card->wandev.get_if_info     	= &get_if_info;
	card->wandev.set_dev_config    	= &set_dev_config;
	card->wandev.set_if_info     	= &set_if_info;
#endif

	/* Setup Port Bps */
	if(card->wandev.clocking) {
		card->wandev.bps = conf->bps;
	}else{
        	card->wandev.bps = 0;
  	}

	/* Setup the Port MTU */
	if((card->wandev.comm_port == WANOPT_PRI)) {
		/* For Primary Port 0 */
		card->wandev.mtu =
			(conf->mtu >= MIN_WP_PRI_MTU) ? 
			 wp_min(conf->mtu, MAX_WP_PRI_MTU) : DEFAULT_WP_PRI_MTU;
		
	}

	err=xilinx_chip_configure(card,conf);
	if (err){
		return err;
	}
	
	/* Set protocol link state to disconnected,
	 * After seting the state to DISCONNECTED this
	 * function must return 0 i.e. success */
	card->wandev.state = WAN_CONNECTED;

	DEBUG_EVENT( "\n");
	return 0;
}



/**SECTION************************************************************** 
 *
 * 	WANPIPE Device Driver Entry Points 
 *
 * *********************************************************************/



/*============================================================================
 * update - Update wanpipe device status & statistics
 * 
 * @wandev:	Wanpipe device pointer
 *
 * This procedure is called when updating the PROC file system.
 * It returns various communications statistics.
 * 
 * cat /proc/net/wanrouter/wanpipe#  (where #=1,2,3...)
 * 
 * These statistics are accumulated from 3 
 * different locations:
 * 	1) The 'if_stats' recorded for the device.
 * 	2) Communication error statistics on the adapter.
 *      3) Operational statistics on the adapter.
 * 
 * The board level statistics are read during a timer interrupt. 
 * Note that we read the error and operational statistics 
 * during consecitive timer ticks so as to minimize the time 
 * that we are inside the interrupt handler.
 *
 */
static int update (wan_device_t* wandev)
{
	sdla_t* card = wandev->private;
 	struct net_device* dev;
        volatile private_area_t* priv_area;

	/* sanity checks */
	if((wandev == NULL) || (wandev->private == NULL))
		return -EFAULT;
	
	if(wandev->state == WAN_UNCONFIGURED)
		return -ENODEV;

	if(test_bit(PERI_CRIT, (void*)&card->wandev.critical))
                return -EAGAIN;

	if((dev=card->wandev.dev) == NULL)
		return -ENODEV;

	if((priv_area=dev->priv) == NULL)
		return -ENODEV;

       	if(card->update_comms_stats){
		return -EAGAIN;
	}

/* FIXME: Add update logic */
	
	return 0;
}


/*============================================================================
 * new_if - Create new logical channel.
 * 
 * &wandev: 	Wanpipe device pointer
 * &dev:	Network device pointer
 * &conf:	User configuration options pointer
 *
 * This routine is called by the ROUTER_IFNEW ioctl,
 * in wanmain.c.  The ioctl passes us the user configuration
 * options which we use to configure the driver and
 * firmware.
 *
 * This functions main purpose is to allocate the
 * private structure for protocol and bind it 
 * to dev->priv pointer.  
 *
 * Also the dev->init pointer should also be initialized
 * to the if_init() function.
 *
 * Any allocation necessary for the private strucutre
 * should be done here, as well as proc/ file initializetion
 * for the network interface.
 * 
 * o parse media- and hardware-specific configuration
 * o make sure that a new channel can be created
 * o allocate resources, if necessary
 * o prepare network device structure for registaration.
 * o add network interface to the /proc/net/wanrouter
 * 
 * The opposite of this function is del_if()
 *
 * Return:	0	o.k.
 *		< 0	failure (channel will not be created)
 */
static int new_if (wan_device_t* wandev, struct net_device* dev, wanif_conf_t* conf)
{
	sdla_t* card = wandev->private;
	private_area_t* chan;
	int err = 0;
	int i;
	struct sk_buff *skb;

	DEBUG_EVENT( "%s: Configuring Interface: %s\n",
			card->devname, conf->name);
 
	if ((conf->name[0] == '\0') || (strlen(conf->name) > WAN_IFNAME_SZ)) {
		DEBUG_EVENT( "%s: Invalid interface name!\n",
			card->devname);
		return -EINVAL;
	}

	/* This protocol only supports a single network interface
	 * If developing a multi-interface protocol, one should
	 * alter this check to allow multiple interfaces */
	if (atomic_read(&card->wandev.if_cnt) > 0){
		return -EEXIST;
	}
		
	/* allocate and initialize private data */
	chan = wan_malloc(sizeof(private_area_t));
	if(chan == NULL){ 
		WAN_MEM_ASSERT(card->devname);
		return -ENOMEM;
	}
	memset(chan, 0, sizeof(private_area_t));
	
	strncpy(chan->if_name, conf->name, WAN_IFNAME_SZ);


	chan->rx_dma_buf=wpabs_dma_alloc(card, card->wandev.mtu+100);
	if (!chan->rx_dma_buf){
		wan_free(chan);
		WAN_MEM_ASSERT(card->devname);
                return -ENOMEM;
	}

	memset(wpabs_dma_get_vaddr(card,chan->rx_dma_buf),0,card->wandev.mtu+100);

        chan->tx_dma_buf=wpabs_dma_alloc(card, card->wandev.mtu+100);
        if (!chan->tx_dma_buf){
                wan_free(chan);
		wpabs_dma_free(card,chan->rx_dma_buf);
                WAN_MEM_ASSERT(card->devname);
                return -ENOMEM;
        }

	memset(wpabs_dma_get_vaddr(card,chan->tx_dma_buf),0,card->wandev.mtu+100);
	
	chan->card = card; 
	chan->dev  = dev;

	WAN_TASKLET_INIT(((wan_tasklet_t*)&chan->chan_task), 
			 0, wp_bh, (unsigned long)chan);

	skb_queue_head_init(&chan->wp_tx_free_list);
	skb_queue_head_init(&chan->wp_tx_pending_list);
	skb_queue_head_init(&chan->wp_tx_complete_list);

	skb_queue_head_init(&chan->wp_rx_free_list);
	skb_queue_head_init(&chan->wp_rx_complete_list);

	/* Initialize the socket binding information
	 * These hooks are used by the API sockets to
	 * bind into the network interface */
	chan->common.sk = NULL;
	chan->common.func = NULL;	
	chan->common.state = WAN_CONNECTING;

	chan->TracingEnabled = 0;
	chan->route_status = NO_ROUTE;
	chan->route_removed = 0;

        /* Setup interface as:
	 *    WANPIPE 	  = IP over Protocol (Firmware)
	 *    API     	  = Raw Socket access to Protocol (Firmware)
	 *    BRIDGE  	  = Ethernet over Protocol, no ip info
	 *    BRIDGE_NODE = Ethernet over Protocol, with ip info
	 */
	if(strcmp(conf->usedby, "WANPIPE") == 0) {

		DEBUG_EVENT( "%s: Running in WANPIPE mode!\n",
			wandev->name);
		chan->common.usedby = WANPIPE;

		/* Option to bring down the interface when 
        	 * the link goes down */
		if (conf->if_down){
			set_bit(DYN_OPT_ON,&chan->interface_down);
			DEBUG_EVENT( 
			 "%s:%s: Dynamic interface configuration enabled\n",
			   card->devname,chan->if_name);
		} 

	} else if( strcmp(conf->usedby, "API") == 0) {
		chan->common.usedby = API;
		DEBUG_EVENT( "%s:%s: Running in API mode !\n",
			wandev->name,chan->if_name);

	}else if (strcmp(conf->usedby, "BRIDGE") == 0) {
		chan->common.usedby = BRIDGE;
		DEBUG_EVENT( "%s:%s: Running in WANPIPE (BRIDGE) mode.\n", 
				card->devname,chan->if_name);
	
	}else if (strcmp(conf->usedby, "BRIDGE_N") == 0) {
		chan->common.usedby = BRIDGE_NODE;
		DEBUG_EVENT( "%s:%s: Running in WANPIPE (BRIDGE_NODE) mode.\n", 
				card->devname,chan->if_name);
	
	}else{
		DEBUG_EVENT( "%s:%s: Error: Invalid operation mode [WANPIPE|API|BRIDGE|BRIDGE_NODE]\n",
				card->devname,chan->if_name);
		err=-EINVAL;
		goto new_if_error;
	}

	chan->hdlc_eng=conf->hdlc_streaming;

	err=xilinx_dev_configure(card,chan,conf);
	if (err){
		goto new_if_error;
	}

	for (i=0;i<MAX_RX_BUF;i++){
		skb=wan_skb_alloc(card->wandev.mtu + 100,card);
		if (!skb){
			err=-ENOMEM;
			goto new_if_error;
		}
		wan_skb_queue_tail(&chan->wp_rx_free_list,skb);
	}

#if 0
	for (i=0;i<MAX_TX_BUF;i++){
		skb=wan_skb_alloc(card->wandev.mtu + 100,card);
		if (!skb){
			err=-ENOMEM;
			goto new_if_error;
		}
		wan_skb_queue_tail(&chan->wp_tx_free_list,skb);
	}
#endif

	/* If gateway option is set, then this interface is the
	 * default gateway on this system. We must know that information
	 * in case DYNAMIC interface configuration is enabled. 
	 *
	 * I.E. If the interface is brought down by the driver, the
	 *      default route will also be removed.  Once the interface
	 *      is brought back up, we must know to re-astablish the
	 *      default route.
	 */
	if ((chan->gateway = conf->gateway) == WANOPT_YES){
		DEBUG_EVENT( "%s: Interface %s is set as a gateway.\n",
			card->devname,chan->if_name);
	}

	/* Get Multicast Information from the user
	 * FIXME: This option is not clearly defined
	 */
	chan->mc = conf->mc;

	
	/* The network interface "dev" has been passed as
	 * an argument from the above layer. We must initialize
	 * it so it can be registered into the kernel.
	 *
	 * The "dev" structure is the link between the kernel
	 * stack and the wanpipe driver.  It contains all 
	 * access hooks that kernel uses to communicate to 
	 * the our driver.
	 *
	 * For now, just set the "dev" name to the user
	 * defined name and initialize:
	 * 	dev->if_init : function that will be called
	 * 	               to further initialize
	 * 	               dev structure on "ifconfig up"
	 *
	 * 	dev->priv    : private structure allocated above
	 * 	
	 */
	
#ifdef LINUX_2_4
	strcpy(dev->name,chan->if_name);
#else
	dev->name = (char *)wan_malloc(strlen(chan->if_name) + 2); 
	if (!dev->name){
		err=-ENOMEM;
		goto new_if_error;
	}
	sprintf(dev->name, "%s", chan->if_name);
#endif

#if 0
	/* Create interface file in proc fs.
	 * Once the proc file system is created, the new_if() function
	 * should exit successfuly.  
	 *
	 * DO NOT place code under this function that can return 
	 * anything else but 0.
	 */
	err = wanrouter_proc_add_interface(wandev, 
					   &chan->dent, 
					   chan->if_name, 
					   dev);
	if (err){	
		DEBUG_EVENT(
			"%s: can't create /proc/net/router/frmw/%s entry!\n",
			card->devname, chan->if_name);
		goto new_if_error;
	}
#endif
	/* Only setup the dev pointer once the new_if function has
	 * finished successfully.  DO NOT place any code below that
	 * can return an error */
	dev->init = &if_init;
	dev->priv = chan;

	/* Increment the number of network interfaces 
	 * configured on this card.  
	 */
	atomic_inc(&card->wandev.if_cnt);

	DEBUG_EVENT( "\n");

	return 0;
	
new_if_error:

	while ((skb=wan_skb_dequeue(&chan->wp_tx_free_list)) != NULL){
		wan_skb_free(skb);
	}

	while ((skb=wan_skb_dequeue(&chan->wp_rx_free_list)) != NULL){
		wan_skb_free(skb);
	}

	wan_free(chan);

#ifndef LINUX_2_4
	if (dev->name){
		wan_free(dev->name);
		dev->name=NULL;
	}
#endif
	dev->priv=NULL;

	return err;
}

/*============================================================================
 * del_if - Delete logical channel.
 *
 * @wandev: 	Wanpipe private device pointer
 * @dev:	Netowrk interface pointer
 *
 * This function is called by ROUTER_DELIF ioctl call
 * to deallocate the network interface.
 * 
 * The network interface and the private structure are
 * about to be deallocated by the upper layer. 
 * We have to clean and deallocate any allocated memory.
 *
 * NOTE: DO NOT deallocate dev->priv here! It will be
 *       done by the upper layer.
 * 
 */
static int del_if (wan_device_t* wandev, struct net_device* dev)
{
	private_area_t* 	priv_area = dev->priv;
	sdla_t*			card = priv_area->card;
	unsigned long 		smp_flags;
	struct sk_buff 		*skb;


	lock_adapter_irq(&card->wandev.lock,&smp_flags);
	xilinx_dev_unconfigure(card,priv_area);
	unlock_adapter_irq(&card->wandev.lock,&smp_flags);

	
	WAN_TASKLET_KILL(((wan_tasklet_t*)&priv_area->chan_task));
	
	lock_adapter_irq(&card->wandev.lock,&smp_flags);

	while ((skb=wan_skb_dequeue(&priv_area->wp_rx_free_list)) != NULL){
		wan_skb_free(skb);
	}
	while ((skb=wan_skb_dequeue(&priv_area->wp_rx_complete_list)) != NULL){
		wan_skb_free(skb);
	}

	while ((skb=wan_skb_dequeue(&priv_area->wp_tx_free_list)) != NULL){
		wan_skb_free(skb);
	}
	while ((skb=wan_skb_dequeue(&priv_area->wp_tx_pending_list)) != NULL){
		wan_skb_free(skb);
	}

	if (priv_area->tx_dma_skb){
		DEBUG_EVENT("freeing tx dma skb\n");
		wan_skb_free(priv_area->tx_dma_skb);
		priv_area->tx_dma_skb=NULL;
	}

	if (priv_area->rx_dma_skb){
                DEBUG_EVENT("freeing rx dma skb\n");
                wan_skb_free(priv_area->rx_dma_skb);
                priv_area->rx_dma_skb=NULL;
        }


	wpabs_dma_free(card,priv_area->rx_dma_buf);	
	wpabs_dma_free(card,priv_area->tx_dma_buf);
	
	unlock_adapter_irq(&card->wandev.lock,&smp_flags);

	
	/* Delete interface name from proc fs. */
#if 0
	wanrouter_proc_delete_interface(wandev, priv_area->if_name);
#endif

	/* Decrement the number of network interfaces 
	 * configured on this card.  
	 */
	atomic_dec(&card->wandev.if_cnt);
	return 0;
}


/**SECTION***********************************************************
 *
 * 	KERNEL Device Entry Interfaces 
 *
 ********************************************************************/


 
/*============================================================================
 * if_init - Initialize Linux network interface.
 *
 * @dev:	Network interface pointer
 * 
 * During "ifconfig up" the upper layer calls this function
 * to initialize dev access pointers.  Such as transmit,
 * stats and header.
 * 
 * It is called only once for each interface, 
 * during Linux network interface registration.  
 *
 * Returning anything but zero will fail interface
 * registration.
 */
static int if_init (struct net_device* dev)
{
	private_area_t* priv_area = dev->priv;
	sdla_t* card = priv_area->card;
	wan_device_t* wandev = &card->wandev;

	/* Initialize device driver entry points */
	dev->open		= &if_open;
	dev->stop		= &if_close;
	dev->hard_header	= NULL; 
	dev->rebuild_header	= NULL;
	dev->hard_start_xmit	= &if_send;
	dev->get_stats		= &if_stats;
#ifdef LINUX_2_4
	dev->tx_timeout		= &if_tx_timeout;
	dev->watchdog_timeo	= TX_TIMEOUT;
#endif
	dev->do_ioctl		= if_do_ioctl;

	if (priv_area->common.usedby == BRIDGE || 
            priv_area->common.usedby == BRIDGE_NODE){

		/* Setup the interface for Bridging */
		int hw_addr=0;
		ether_setup(dev);
		
		/* Use a random number to generate the MAC address */
		memcpy(dev->dev_addr, "\xFE\xFC\x00\x00\x00\x00", 6);
		get_random_bytes(&hw_addr, sizeof(hw_addr));
		*(int *)(dev->dev_addr + 2) += hw_addr;

	}else{
		/* Initialize media-specific parameters */
		dev->flags		|= IFF_POINTOPOINT;
		dev->flags		|= IFF_NOARP;

		/* Enable Mulitcasting if user selected */
		if (priv_area->mc == WANOPT_YES){
			dev->flags 	|= IFF_MULTICAST;
		}
		
		if (priv_area->true_if_encoding){
			dev->type	= ARPHRD_HDLC; /* This breaks the tcpdump */
		}else{
			dev->type	= ARPHRD_PPP;
		}
		
		dev->mtu		= card->wandev.mtu;
	 
		dev->hard_header_len	= 0;
	}
	
	/* Initialize hardware parameters */
	dev->irq	= wandev->irq;
	dev->dma	= wandev->dma;
	dev->base_addr	= wandev->ioport;
	dev->mem_start	= wandev->maddr;
	dev->mem_end	= wandev->maddr + wandev->msize - 1;

	/* Set transmit buffer queue length 
	 * If too low packets will not be retransmitted 
         * by stack.
	 */
        dev->tx_queue_len = 100;
   
	return 0;
}

/*============================================================================
 * if_open - Open network interface.
 *
 * @dev: Network device pointer
 *
 * On ifconfig up, this function gets called in order
 * to initialize and configure the private area.  
 * Driver should be configured to send and receive data.
 *
 * This functions starts a timer that will call
 * frmw_config() function. This function must be called
 * because the IP addresses could have been changed
 * for this interface.
 *
 * Return 0 if O.k. or errno.
 */
static int if_open (struct net_device* dev)
{
	private_area_t* priv_area = dev->priv;
	sdla_t* card = priv_area->card;
	struct timeval tv;
	int err = 0;

	/* Only one open per interface is allowed */
	if (open_dev_check(dev))
		return -EBUSY;

	/* Initialize the router start time.
	 * Used by wanpipemon debugger to indicate
	 * how long has the interface been up */
	do_gettimeofday(&tv);
	priv_area->router_start_time = tv.tv_sec;

#ifdef LINUX_2_4
	netif_start_queue(dev);
#else
	dev->interrupt = 0;
	dev->tbusy = 0;
	dev->start = 1;
#endif

	err=xilinx_dev_open(card,priv_area);
	if (err){
		return err;
	}
	
	/* Increment the module usage count */
	wanpipe_open(card);

	priv_area->common.state = WAN_CONNECTED;
	return err;
}

/*============================================================================
 * if_close - Close network interface.
 * 
 * @dev: Network device pointer
 * 
 * On ifconfig down, this function gets called in order
 * to cleanup interace private area.  
 *
 * IMPORTANT: 
 *
 * No deallocation or unconfiguration should ever occur in this 
 * function, because the interface can come back up 
 * (via ifconfig up).  
 * 
 * Furthermore, in dynamic interfacace configuration mode, the 
 * interface will come up and down to reflect the protocol state.
 *
 * Any deallocation and cleanup can occur in del_if() 
 * function.  That function is called before the dev interface
 * itself is deallocated.
 *
 * Thus, we should only stop the net queue and decrement
 * the wanpipe usage counter via wanpipe_close() function.
 */

static int if_close (struct net_device* dev)
{
	private_area_t* priv_area = dev->priv;
	sdla_t* card = priv_area->card;

	stop_net_queue(dev);

#ifndef LINUX_2_4
	dev->start=0;
#endif

	xilinx_dev_close(card,priv_area);

	wanpipe_close(card);
	return 0;
}


/*=============================================================
 * disable_comm - Main shutdown function
 *
 * @card: Wanpipe device pointer
 *
 * The command 'wanrouter stop' has been called
 * and the whole wanpipe device is going down.
 * This is the last function called to disable 
 * all comunications and deallocate any memory
 * that is still allocated.
 *
 * o Disable communications, turn off interrupts
 * o Deallocate memory used, if any
 * o Unconfigure TE1 card
 */ 

static void disable_comm (sdla_t *card)
{
	unsigned long reg;	
	unsigned long flags;

	lock_adapter_irq(&card->wandev.lock,&flags);
	
 	reg=0;
        sdla_pci_writel(&card->hw,XILINX_TIMESLOT_HDLC_CHAN_REG,reg);

	reg=sdla_pci_readl(&card->hw,XILINX_CHIP_CFG_REG);

	/* Enable the chip/hdlc reset condition */
	reg=0;
	set_bit(CHIP_RESET_BIT,&reg);
	set_bit(HDLC_CORE_RESET_BIT,&reg);

	sdla_pci_writel(&card->hw,XILINX_CHIP_CFG_REG,reg);

	unlock_adapter_irq(&card->wandev.lock,&flags);
	
	return;
}



#ifdef LINUX_2_4
/*============================================================================
 * if_tx_timeout
 *
 * Kernel networking stack calls this function in case
 * the interface has been stopped for TX_TIMEOUT seconds.
 *
 * This would occur if we lost TX interrupts or the
 * card has stopped working for some reason.  
 * 
 * Handle transmit timeout event from netif watchdog
 */
static void if_tx_timeout (struct net_device *dev)
{
    	private_area_t* chan = dev->priv;
	sdla_t *card = chan->card;
	
	/* If our device stays busy for at least 5 seconds then we will
	 * kick start the device by making dev->tbusy = 0.  We expect
	 * that our device never stays busy more than 5 seconds. So this                 
	 * is only used as a last resort.
	 */

	++chan->if_stats.collisions;

	DEBUG_EVENT( "%s: Transmit timed out on %s\n", card->devname,dev->name);
	netif_wake_queue (dev);
}
#endif


/*============================================================================
 * if_send - Send a packet on a network interface.
 * 
 * @dev:	Network interface pointer
 * @skb:	Packet obtained from the stack or API
 *              that should be sent out the port.
 *              
 * o Mark interface as stopped
 * 	(marks start of the transmission) to indicate 
 * 	to the stack that the interface is busy. 
 *
 * o Check link state. 
 * 	If link is not up, then drop the packet.
 * 
 * o Copy the tx packet into the protocol tx buffers on
 *   the adapter.
 *   
 * o If tx successful:
 * 	Free the skb buffer and mark interface as running
 * 	and return 0.
 *
 * o If tx failed, busy:
 * 	Keep interface marked as busy 
 * 	Do not free skb buffer
 * 	Enable Tx interrupt (which will tell the stack
 * 	                     that interace is not busy)
 * 	Return a non-zero value to tell the stack
 * 	that the tx should be retried.
 * 	
 * Return:	0	complete (socket buffer must be freed)
 *		non-0	packet may be re-transmitted 
 *
 */
static int if_send (struct sk_buff* skb, struct net_device* dev)
{

	private_area_t *chan = dev->priv;
	sdla_t *card = chan->card;

	/* Mark interface as busy. The kernel will not
	 * attempt to send any more packets until we clear
	 * this condition */
#ifdef LINUX_2_4
	netif_stop_queue(dev);
#endif
	
	if (skb == NULL){
		/* This should never happen. Just a sanity check.
		 */
		DEBUG_EVENT( "%s: interface %s got kicked!\n",
			card->devname, dev->name);

		WAN_NETIF_WAKE_QUEUE(dev);
		return 0;
	}

	/* Non 2.4 kernels used to call if_send() 
	 * after TX_TIMEOUT seconds have passed of interface 
	 * being busy. Same as if_tx_timeout() in 2.4 kernels */
#ifndef LINUX_2_4
	if (dev->tbusy){

		/* If our device stays busy for at least 5 seconds then we will
		 * kick start the device by making dev->tbusy = 0.  We expect 
		 * that our device never stays busy more than 5 seconds. So this
		 * is only used as a last resort. 
		 */
                ++chan->if_stats.collisions;
		if((jiffies - chan->tick_counter) < (5 * HZ)) {
			return 1;
		}

		DEBUG_EVENT( "%s: Transmit timeout !\n",
			card->devname);

		/* unbusy the interface */
		clear_bit(0,&dev->tbusy);
	}
#endif

	/* Sanity check, which blocks the if_send() from being
	 * re-entrant.  The kernel will never do this to us
	 * but its better to be safe */
    	if(test_and_set_bit(SEND_CRIT, (void*)&card->wandev.critical)) {
	
		DEBUG_EVENT( "%s: Critical in if_send: %lx\n",
					card->wandev.name,card->wandev.critical);
                ++chan->if_stats.tx_dropped;
		start_net_queue(dev);
		goto if_send_exit_crit;
	}

	if(!skb->protocol){
		/* Skb buffer without protocol should never
		 * occur in normal operation */
        	++chan->if_stats.tx_errors;
		start_net_queue(dev);
		
	}else {
		if (wan_skb_queue_len(&chan->wp_tx_pending_list) > MAX_TX_BUF){
			stop_net_queue(dev);
			xilinx_dma_tx(card,chan);
		}else{
			skb_unlink(skb);
			wan_skb_queue_tail(&chan->wp_tx_pending_list,skb);
			xilinx_dma_tx(card,chan);
			start_net_queue(dev);
#ifdef LINUX_2_4
		 	dev->trans_start = jiffies;
#endif
		}	
	}

if_send_exit_crit:
	
	/* End of critical area for re-entry and for S508 card */
	clear_bit(SEND_CRIT, (void*)&card->wandev.critical);
	
	return WAN_NETIF_QUEUE_STOPPED(dev);
}


/*============================================================================
 * chk_bcast_mcast_addr - Check for source broadcast addresses
 *
 * Check to see if the packet to be transmitted contains a broadcast or
 * multicast source IP address.
 */

static int chk_bcast_mcast_addr(sdla_t *card, struct net_device* dev,
				struct sk_buff *skb)
{
	u32 src_ip_addr;
        u32 broadcast_ip_addr = 0;
	private_area_t *priv_area=dev->priv;
        struct in_device *in_dev;
        /* read the IP source address from the outgoing packet */
        src_ip_addr = *(u32 *)(skb->data + 12);

	if (priv_area->common.usedby != WANPIPE){
		return 0;
	}
	
	/* read the IP broadcast address for the device */
        in_dev = dev->ip_ptr;
        if(in_dev != NULL) {
                struct in_ifaddr *ifa= in_dev->ifa_list;
                if(ifa != NULL)
                        broadcast_ip_addr = ifa->ifa_broadcast;
                else
                        return 0;
        }
 
        /* check if the IP Source Address is a Broadcast address */
        if((dev->flags & IFF_BROADCAST) && (src_ip_addr == broadcast_ip_addr)) {
                DEBUG_EVENT( "%s: Broadcast Source Address silently discarded\n",
				card->devname);
                return 1;
        } 

        /* check if the IP Source Address is a Multicast address */
        if((ntohl(src_ip_addr) >= 0xE0000001) &&
		(ntohl(src_ip_addr) <= 0xFFFFFFFE)) {
                DEBUG_EVENT( "%s: Multicast Source Address silently discarded\n",
				card->devname);
                return 1;
        }

        return 0;
}

/*============================================================================
 * if_stats
 *
 * Used by /proc/net/dev and ifconfig to obtain interface
 * statistics.
 *
 * Return a pointer to struct net_device_stats.
 */
static struct net_device_stats* if_stats (struct net_device* dev)
{
	private_area_t* priv_area;

	if ((priv_area=dev->priv) == NULL)
		return NULL;

	return &priv_area->if_stats;
}




/*========================================================================
 * 
 * if_do_ioctl - Ioctl handler for fr
 *
 * 	@dev: Device subject to ioctl
 * 	@ifr: Interface request block from the user
 *	@cmd: Command that is being issued
 *	
 *	This function handles the ioctls that may be issued by the user
 *	to control or debug the protocol or hardware . 
 *
 *	It does both busy and security checks. 
 *	This function is intended to be wrapped by callers who wish to 
 *	add additional ioctl calls of their own.
 *
 * Used by:  SNMP Mibs
 * 	     wanpipemon debugger
 *
 */
static int if_do_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
	private_area_t* chan= (private_area_t*)dev->priv;
	sdla_t *card;
#if 0
	wan_udp_pkt_t *wan_udp_pkt;
	unsigned long smp_flags;
#endif
	
	if (!chan){
		return -ENODEV;
	}
	card=chan->card;

	NET_ADMIN_CHECK();

	switch(cmd)
	{

#if 0
		case SIOC_WANPIPE_PIPEMON: 

			
			if (atomic_read(&chan->udp_pkt_len) != 0){
				return -EBUSY;
			}
	
			atomic_set(&chan->udp_pkt_len,MAX_LGTH_UDP_MGNT_PKT);
			
			/* For performance reasons test the critical
			 * here before spin lock */
			if (test_bit(0,&card->in_isr)){
				atomic_set(&chan->udp_pkt_len,0);
				return -EBUSY;
			}
		
			
			wan_udp_pkt=(wan_udp_pkt_t*)chan->udp_pkt_data;
			if (copy_from_user(&wan_udp_pkt->wan_udp_hdr,ifr->ifr_data,sizeof(wan_udp_hdr_t))){
				atomic_set(&chan->udp_pkt_len,0);
				return -EFAULT;
			}

			spin_lock_irqsave(&card->wandev.lock, smp_flags);

			/* We have to check here again because we don't know
			 * what happened during spin_lock */
			if (test_bit(0,&card->in_isr)) {
				DEBUG_EVENT( "%s:%s Pipemon command failed, Driver busy: try again.\n",
						card->devname,dev->name);
				atomic_set(&chan->udp_pkt_len,0);
				spin_unlock_irqrestore(&card->wandev.lock, smp_flags);
				return -EBUSY;
			}
			
			process_udp_mgmt_pkt(card,dev,chan,1);
			
			spin_unlock_irqrestore(&card->wandev.lock, smp_flags);

			/* This area will still be critical to other
			 * PIPEMON commands due to udp_pkt_len
			 * thus we can release the irq */
			
			if (atomic_read(&chan->udp_pkt_len) > sizeof(wan_udp_pkt_t)){
				DEBUG_EVENT( "%s: Error: Pipemon buf too bit on the way up! %i\n",
						card->devname,atomic_read(&chan->udp_pkt_len));
				atomic_set(&chan->udp_pkt_len,0);
				return -EINVAL;
			}

			if (copy_to_user(ifr->ifr_data,&wan_udp_pkt->wan_udp_hdr,sizeof(wan_udp_hdr_t))){
				atomic_set(&chan->udp_pkt_len,0);
				return -EFAULT;
			}
			
			atomic_set(&chan->udp_pkt_len,0);
			return 0;
#endif
			
		default:
			return -EOPNOTSUPP;
	}
	return 0;
}


/**SECTION**********************************************************
 *
 * 	FIRMWARE Specific Interface Functions 
 *
 *******************************************************************/


/*============================================================================
 * xilinx_chip_configure
 *
 * 
 */

static int xilinx_chip_configure(sdla_t *card, wandev_conf_t* conf)
{
	volatile unsigned long reg;
	volatile unsigned char cnt=0;
	int err=0;
	
	DEBUG_EVENT("-- Xilinx_chip_config: --");

/* FIXME: Configure Front End here 
 *
 * Define Clock Source
 * */

	reg=sdla_pci_readl(&card->hw,XILINX_CHIP_CFG_REG);

	/* Enable the chip/hdlc reset condition */
	reg=0;
	set_bit(CHIP_RESET_BIT,&reg);
	set_bit(HDLC_CORE_RESET_BIT,&reg);

	DEBUG_EVENT("RESETING CHIP\n\n");
	xilinx_delay(1);

	sdla_pci_writel(&card->hw,XILINX_CHIP_CFG_REG,reg);

	
	/* Disable the chip/hdlc reset condition */
	clear_bit(CHIP_RESET_BIT,&reg);
	clear_bit(HDLC_CORE_RESET_BIT,&reg);
	
	/* Disable ALL chip interrupts */
	clear_bit(GLOBAL_INTR_ENABLE_BIT,&reg);
	clear_bit(ERROR_INTR_ENABLE_BIT,&reg);
	
	/* Configure for T1 or E1 front end */
	if (IS_T1_CARD(card)){
		card->u.xilinx.num_of_time_slots=NUM_OF_T1_CHANNELS;
		clear_bit(INTERFACE_TYPE_T1_E1_BIT,&reg);	
	}else if (IS_E1_CARD(card)){
		card->u.xilinx.num_of_time_slots=NUM_OF_E1_CHANNELS;
		set_bit(INTERFACE_TYPE_T1_E1_BIT,&reg);
	}else{
		DEBUG_EVENT("%s: Error: Xilinx doesn't support non T1/E1 interface!\n",
				card->devname);
		return -EINVAL;
	}

	DEBUG_EVENT("CONFIGURING/ENABLING CHIP\n\n");
        xilinx_delay(1);

	sdla_pci_writel(&card->hw,XILINX_CHIP_CFG_REG,reg);

	DEBUG_EVENT("WAITING FOR READY \n\n");
        xilinx_delay(1);

	for (;;){
		reg=sdla_pci_readl(&card->hw,XILINX_CHIP_CFG_REG);

		if (!test_bit(HDLC_CORE_READY_FLAG_BIT,&reg)){
			/* The HDLC Core is not ready! we have
			 * an error. */
			if (++cnt > 5){
				err = -EINVAL;
				break;
			}else{
				udelay(500);
				schedule();
			}
		}else{
			DEBUG_EVENT("%s: HDLC Core Ready 0x%08lX\n",
                                        card->devname,reg);
			err=0;
			break;
		}
	}


	if (err != 0){
		DEBUG_EVENT("%s: Error: HDLC Core Not Ready!\n",
					card->devname);
		xilinx_delay(1);

		reg=sdla_pci_readl(&card->hw,XILINX_CHIP_CFG_REG);

		/* Disable the chip/hdlc reset condition */
		set_bit(CHIP_RESET_BIT,&reg);
		set_bit(HDLC_CORE_RESET_BIT,&reg);

		sdla_pci_writel(&card->hw,XILINX_CHIP_CFG_REG,reg);
		return err;
	}

	/* Setup global DMA parameters */
	reg=0;
	reg|=(XILINX_DMA_SIZE    << DMA_SIZE_BIT_SHIFT);
	reg|=(XILINX_DMA_FIFO_UP << DMA_FIFO_HI_MARK_BIT_SHIFT);
	reg|=(XILINX_DMA_FIFO_LO << DMA_FIFO_LO_MARK_BIT_SHIFT);

	/* Enable global DMA engine and set to default
	 * number of active channels. Note: this value will
	 * change in dev configuration */
	reg|=(XILINX_DEFLT_ACTIVE_CH << DMA_ACTIVE_CHANNEL_BIT_SHIFT);
	set_bit(DMA_ENGINE_ENABLE_BIT,&reg);

   	DEBUG_EVENT("WRITING DMA CONTROL \n\n");
	xilinx_delay(1);
	
	sdla_pci_writel(&card->hw,XILINX_DMA_CONTROL_REG,reg);


	DEBUG_EVENT("WRITING TX RX ENABLE \n\n");
        xilinx_delay(1);

	reg=0;
	set_bit(GLOBAL_TX_ENABLE_BIT,&reg);
	set_bit(GLOBAL_RX_ENABLE_BIT,&reg);
	sdla_pci_writel(&card->hw,XILINX_TIMESLOT_HDLC_CHAN_REG,reg);

	reg=sdla_pci_readl(&card->hw,XILINX_CHIP_CFG_REG);
	
	set_bit(GLOBAL_INTR_ENABLE_BIT,&reg);
	/* FIXME: Add this later */
	/* set_bit(ERROR_INTR_ENABLE_BIT,&reg); */

	DEBUG_EVENT("WRITING ENABLE INTR ON CHIP \n\n");
        xilinx_delay(1);

	sdla_pci_writel(&card->hw,XILINX_CHIP_CFG_REG,reg);
	
	return err;
}


/*============================================================================
 * xilinx_dev_configure
 *
 * 
 */

static int xilinx_dev_configure(sdla_t *card, private_area_t *chan, wanif_conf_t* conf)
{
	unsigned long reg;
	long free_logic_ch,i;

	DEBUG_EVENT("-- Xilinx_dev_configure:--");

	chan->logic_ch_num=-1;

	if (!IS_TE1_CARD(card)){
		return -EINVAL;
	}

	/* Channel definition section. If not channels defined 
	 * return error */
	if ((chan->time_slot_map=conf->active_ch) == 0){
		printk(KERN_INFO "%s: Invalid Channel Selection 0x%lX\n",
				card->devname,chan->time_slot_map);
		return -EINVAL;
	}

	DEBUG_EVENT("%s: %s: Interface active channels = 0x%lX\n",
			card->devname,conf->name,chan->time_slot_map);


	/* Check that the time slot is not being used. If it is
	 * stop the interface setup.  Notice, though we proceed
	 * to check for all timeslots before we start binding 
	 * the channels in.  This way, we don't have to go back
	 * and clean the time_slot_map */ 
	for (i=0;i<card->u.xilinx.num_of_time_slots;i++){
		if (test_bit(i,&chan->time_slot_map)){

			printk(KERN_INFO "Confiuring %s for timeslot %li\n",
					conf->name, i+1);

			if (test_bit(i,&card->u.xilinx.time_slot_map)){
				printk(KERN_INFO "%s: Channel/Time Slot resource conflict!\n",
						card->devname); 
				printk(KERN_INFO "%s: %s: Channel/Time Slot %li, aready in use!\n",
						card->devname,chan->if_name,(i+1));
				
				return -EEXIST;
			}
		}
	}

	
	chan->logic_ch_num=request_xilinx_logical_channel_num(card, chan, &free_logic_ch);
	if (chan->logic_ch_num == -1){
		return -EBUSY;
	}

	DEBUG_EVENT("%s:%d: GOT Logic ch %li  Free Logic ch %li\n",
		__FUNCTION__,__LINE__,chan->logic_ch_num,free_logic_ch);

	for (i=0;i<card->u.xilinx.num_of_time_slots;i++){
		if (test_bit(i,&chan->time_slot_map)){

			set_bit(i,&card->u.xilinx.time_slot_map);

			DEBUG_EVENT("Setting Timeslot %li to logic ch %li\n",
				        i, chan->logic_ch_num);

			reg=sdla_pci_readl(&card->hw, XILINX_TIMESLOT_HDLC_CHAN_REG);
			reg&=~TIMESLOT_BIT_MASK;

			/* Select a Timeslot for configuration */
			sdla_pci_writel(&card->hw,
				       XILINX_TIMESLOT_HDLC_CHAN_REG,
				       (reg|(i<<TIMESLOT_BIT_SHIFT)));
			
			/* Bind a logical channel to the above selected
			 * timeslot */
			sdla_pci_writel(&card->hw,
				       XILINX_CONTROL_RAM_ACCESS_BUF,
				       (chan->logic_ch_num&CONTROL_RAM_DATA_MASK));
		}
	}

	if (free_logic_ch != -1){
		for (i=0;i<card->u.xilinx.num_of_time_slots;i++){
			if (!test_bit(i,&card->u.xilinx.time_slot_map)){

				DEBUG_CFG("Setting Timeslot %li to logic ch %li\n",
                                        i, free_logic_ch);

				reg=sdla_pci_readl(&card->hw, XILINX_TIMESLOT_HDLC_CHAN_REG);
	                        reg&=~TIMESLOT_BIT_MASK;

				/* Select a Timeslot for configuration */
				sdla_pci_writel(&card->hw,
					       XILINX_TIMESLOT_HDLC_CHAN_REG,
					       (reg|(i<<TIMESLOT_BIT_SHIFT)));
				
				/* Bind a logical channel to the above selected
				 * timeslot */
				sdla_pci_writel(&card->hw,
					       XILINX_CONTROL_RAM_ACCESS_BUF,
					       (free_logic_ch&CONTROL_RAM_DATA_MASK));
			}
		}
	}

	/* Select an HDLC logic channel for configuration */
	reg=sdla_pci_readl(&card->hw, XILINX_TIMESLOT_HDLC_CHAN_REG);
	reg&=~HDLC_LOGIC_CH_BIT_MASK;
	sdla_pci_writel(&card->hw,
			XILINX_TIMESLOT_HDLC_CHAN_REG,
			(reg|(chan->logic_ch_num&HDLC_LOGIC_CH_BIT_MASK)));

	DEBUG_EVENT("%s: HDLC ENG VALUE %i \n",card->devname,chan->hdlc_eng);

	reg=0;
	if (chan->hdlc_eng){
		/* HDLC engine is enabled on the above logical channels */		
		clear_bit(HDLC_RX_PROT_DISABLE_BIT,&reg);
		clear_bit(HDLC_TX_PROT_DISABLE_BIT,&reg);
	}else{
		/* Transprent Mode
		 * HDLC engine disabled for the above logical channels */	
		set_bit(HDLC_RX_PROT_DISABLE_BIT,&reg);
		set_bit(HDLC_TX_PROT_DISABLE_BIT,&reg);
	}

	set_bit(HDLC_TX_CHAN_ENABLE_BIT,&reg);
	set_bit(HDLC_RX_ADDR_RECOGN_DIS_BIT,&reg);

	sdla_pci_writel(&card->hw,XILINX_HDLC_CONTROL_REG,reg);

	
	return 0;
}


static void xilinx_dev_unconfigure(sdla_t *card, private_area_t *chan)
{
	unsigned long reg;
	int i;

	
	DEBUG_EVENT("-- Xilinx_dev_unconfigure:--");

	/* Select an HDLC logic channel for configuration */
	if (chan->logic_ch_num != -1){

		reg=sdla_pci_readl(&card->hw,XILINX_TIMESLOT_HDLC_CHAN_REG);
		reg&=~HDLC_LOGIC_CH_BIT_MASK;

		sdla_pci_writel(&card->hw,
				XILINX_TIMESLOT_HDLC_CHAN_REG,
				(reg|(chan->logic_ch_num&HDLC_LOGIC_CH_BIT_MASK)));

		reg=0;
		sdla_pci_writel(&card->hw,XILINX_HDLC_CONTROL_REG,reg);

		free_xilinx_logical_channel_num(card,chan->logic_ch_num);
		chan->logic_ch_num=-1;

		for (i=0;i<card->u.xilinx.num_of_time_slots;i++){
			if (test_bit(i,&chan->time_slot_map)){
				clear_bit(i,&card->u.xilinx.time_slot_map);
			}
		}

		udelay(500);
	}
}


#define xilinx_dma_rx xilinx_dev_open
static int xilinx_dev_open(sdla_t *card, private_area_t *chan)
{
	unsigned long reg;
	unsigned long dma_descr;
//	struct sk_buff *skb;
//	void *buf;
	unsigned long bus_addr;

	
	DEBUG_EVENT("-- Xilinx_dma_rx (dev_open):--");

	/* sanity check: make sure that DMA is in ready state */
	dma_descr=(chan->logic_ch_num<<4) + XILINX_RxDMA_DESCRIPTOR_HI;
	reg=sdla_pci_readl(&card->hw,dma_descr);
	
	if (test_bit(RxDMA_HI_DMA_GO_READY_BIT,&reg)){
		DEBUG_EVENT("%s: Error: RxDMA GO Ready bit set on dma Rx\n",
				card->devname);
		return -EFAULT;
	}

#if 0
	skb=wan_skb_dequeue(&chan->wp_rx_free_list);
	if (!skb){
		DEBUG_EVENT("%s: Error: No free rx buffers \n",
				card->devname);
		return -ENOBUFS;
	}
	
	dma_descr=(chan->logic_ch_num<<4) + XILINX_RxDMA_DESCRIPTOR_LO;
	reg=0;

	buf=wan_skb_put(skb,card->wandev.mtu);

	bus_addr = wan_virt2bus((void*)buf);
#endif

	bus_addr=wpabs_dma_get_paddr(card,chan->rx_dma_buf);

	if (bus_addr&0x03){
		DEBUG_EVENT("%s: Error: Rx Ptr not aligned to 32bit boudary!\n",
				card->devname);
#if 0
		wan_skb_init(skb,16);
	        wan_skb_trim(skb,0);
		wan_skb_queue_tail(&chan->wp_rx_free_list,skb);
#endif
		return -EINVAL;
	}


//	chan->rx_dma_skb=skb;
	
	/* Write the pointer of the data packet to the
	 * DMA address register */
	reg=bus_addr;

	/* Set the 32bit alignment of the data length.
	 * Since we are setting up for rx, set this value
	 * to Zero */
	reg&=~(RxDMA_LO_ALIGNMENT_BIT_MASK);
// M.F.
	dma_descr=(chan->logic_ch_num<<4) + XILINX_RxDMA_DESCRIPTOR_LO;
	
	DEBUG_EVENT("!!!! RxDMA LO = 0x%08X\n",reg);		
	sdla_pci_writel(&card->hw,dma_descr,reg);

	dma_descr=(unsigned long)(chan->logic_ch_num<<4) + XILINX_RxDMA_DESCRIPTOR_HI;
	reg=0;

	reg|=(card->wandev.mtu>>2)&RxDMA_HI_DMA_DATA_LENGTH_MASK;

	set_bit(RxDMA_HI_DMA_INTR_ENABLE_BIT,&reg);

	set_bit(RxDMA_HI_DMA_GO_READY_BIT,&reg);

	DEBUG_EVENT("!!!! RxDMA HI = 0x%08X\n",reg);
	sdla_pci_writel(&card->hw,dma_descr,reg);

	return 0;
}

static void xilinx_dev_close(sdla_t *card, private_area_t *chan)
{
	unsigned long reg;
	unsigned long dma_descr;
	unsigned long smp_flags;

	
	DEBUG_EVENT("-- Xilinx_dev_close :--");

	lock_adapter_irq(&card->wandev.lock,&smp_flags);
	
	dma_descr=(chan->logic_ch_num<<4) + XILINX_RxDMA_DESCRIPTOR_HI;
	reg=0;
	sdla_pci_writel(&card->hw,dma_descr,reg);

	dma_descr=(chan->logic_ch_num<<4) + XILINX_TxDMA_DESCRIPTOR_HI;
	reg=0;
	sdla_pci_writel(&card->hw,dma_descr,reg);

	unlock_adapter_irq(&card->wandev.lock,&smp_flags);

	/* FIXME: Cleanp up Tx and Rx buffers */ 
}


static int xilinx_dma_tx (sdla_t *card, private_area_t *chan)
{
	unsigned long reg=0;
	struct sk_buff *skb;
	unsigned long dma_descr;
	unsigned char len_align=0;
	unsigned long bus_addr;
	unsigned char *buf;


	DEBUG_EVENT("-- Xilinx_dma_tx :--");

	if (test_and_set_bit(TX_BUSY,&chan->dma_status)){
		DEBUG_EVENT("%s:%d:  TX_BUSY set!\n",
			__FUNCTION__,__LINE__);
		return -EBUSY;
	}

	/* sanity check: make sure that DMA is in ready state */
	dma_descr=(chan->logic_ch_num<<4) + XILINX_TxDMA_DESCRIPTOR_HI;

	DEBUG_EVENT("%s:%d:  chan logic ch=%li dma_descr=0x%lx set!\n",
                        __FUNCTION__,__LINE__,chan->logic_ch_num,dma_descr);

	reg=sdla_pci_readl(&card->hw,dma_descr);
	if (test_bit(TxDMA_HI_DMA_GO_READY_BIT,&reg)){
		DEBUG_EVENT("%s: Error: TxDMA GO Ready bit set on dma Tx\n",
				card->devname);
		clear_bit(TX_BUSY,&chan->dma_status);
		return -EFAULT;
	}
	
	skb=wan_skb_dequeue(&chan->wp_tx_pending_list);
	if (!skb){
		DEBUG_EVENT("%s:%d: Tx pending list empty \n",
                        __FUNCTION__,__LINE__);
		clear_bit(TX_BUSY,&chan->dma_status);
		return -ENOBUFS;
	}

	if (skb->len > MAX_XILINX_TX_DMA_SIZE){
		/* FIXME: We need to split this frame into
		 *        multiple parts.  For now thought
		 *        just drop it :) */
		DEBUG_EVENT("%s:%d:  Tx pkt len > MAX XILINX TX DMA LEN!\n",
                        __FUNCTION__,__LINE__);
		wan_skb_free(skb);
		clear_bit(TX_BUSY,&chan->dma_status);
		return -EINVAL;
	}

	buf=(unsigned char *)wpabs_dma_get_vaddr(card,chan->tx_dma_buf);
	memcpy(buf,skb->data,skb->len);

	bus_addr = wpabs_dma_get_paddr(card,chan->tx_dma_buf);

	if (bus_addr & 0x03){
		DEBUG_EVENT("%s: Error: Tx Ptr not aligned to 32bit boudary!\n",
				card->devname);
		wan_skb_free(skb);
		clear_bit(TX_BUSY,&chan->dma_status);
		return -EINVAL;
	}

	chan->tx_dma_skb=skb;

	DEBUG_EVENT("TX SKB DATA 0x%08X 0x%08X 0x%08X 0x%08X \n",
			*(unsigned int*)&skb->data[0],
			*(unsigned int*)&skb->data[4],
			*(unsigned int*)&skb->data[8],
			*(unsigned int*)&skb->data[12]);

	DEBUG_EVENT("Tx dma skb bound %p List=%p Data=%p BusPtr=%lx\n",
			skb,skb->list,skb->data,bus_addr);
	
	dma_descr=(chan->logic_ch_num<<4) + XILINX_TxDMA_DESCRIPTOR_LO; 

	/* Write the pointer of the data packet to the
	 * DMA address register */
	reg=bus_addr;

	/* Set the 32bit alignment of the data length.
	 * Used to pad the tx packet to the 32 bit 
	 * boundary */
	reg&=~(TxDMA_LO_ALIGNMENT_BIT_MASK);
	reg|=(skb->len&0x03);
	
	if (skb->len&0x03){
		len_align=1;
	}

	sdla_pci_writel(&card->hw,dma_descr,reg);
	
	
	dma_descr=(chan->logic_ch_num<<4) + XILINX_TxDMA_DESCRIPTOR_HI;
	
	reg=0;
	reg|=(((skb->len>>2)+len_align)&TxDMA_HI_DMA_DATA_LENGTH_MASK);
	
	set_bit(TxDMA_HI_DMA_FRAME_START_BIT,&reg);
	set_bit(TxDMA_HI_DMA_FRAME_END_BIT,&reg);
	set_bit(TxDMA_HI_DMA_INTR_ENABLE_BIT,&reg);
	
	set_bit(TxDMA_HI_DMA_GO_READY_BIT,&reg);

	sdla_pci_writel(&card->hw,dma_descr,reg);

	return 0;
	
}

static void xilinx_dma_tx_complete (sdla_t *card, private_area_t *chan)
{
	unsigned long reg=0;
	unsigned long dma_descr;
	unsigned int len;


	DEBUG_EVENT("-- Xilinx_dma_tx_complite :--");
// M.F.
//dma_descr=(chan->logic_ch_num<<4) + XILINX_RxDMA_DESCRIPTOR_LO;
//reg=sdla_pci_readl(&card->hw,dma_descr);
//DEBUG_EVENT("RXPTR LO=0x%08lX\n",reg);

//dma_descr=(chan->logic_ch_num<<4) + XILINX_RxDMA_DESCRIPTOR_HI;
//reg=sdla_pci_readl(&card->hw,dma_descr);
//DEBUG_EVENT("RXPTR HI=0x%08lX\n",reg);
//


	dma_descr=(chan->logic_ch_num<<4) + XILINX_TxDMA_DESCRIPTOR_LO;
	reg=sdla_pci_readl(&card->hw,dma_descr);
	DEBUG_EVENT("TXPTR LO=0x%08lX\n",reg);
	dma_descr=(chan->logic_ch_num<<4) + XILINX_TxDMA_DESCRIPTOR_HI;
        reg=sdla_pci_readl(&card->hw,dma_descr);
        DEBUG_EVENT("TXPTR HI=0x%08lX\n",reg);
	reg=sdla_pci_readl(&card->hw,0x74);
        DEBUG_EVENT("TX FIFO Status=0x%08lX\n",reg);

	
	dma_descr=(chan->logic_ch_num<<4) + XILINX_TxDMA_DESCRIPTOR_HI;

	reg=sdla_pci_readl(&card->hw,dma_descr);

	if (!chan->tx_dma_skb){
		DEBUG_EVENT("%s: Critical Error: Tx DMA intr: no tx skb !\n",
			card->devname);
		clear_bit(TX_BUSY,&chan->dma_status);
		return;
	}

	len=chan->tx_dma_skb->len;
	/* FIXME: Put in complete list and dealocte in bh,
	 *        maybe :) */
#if 0
	wan_skb_queue_tail(&chan->wp_tx_free_list,skb);
#endif

	wan_skb_free(chan->tx_dma_skb);
	chan->tx_dma_skb=NULL;

	if (test_bit(TxDMA_HI_DMA_GO_READY_BIT,&reg)){
		DEBUG_EVENT("%s: Error: TxDMA Intr: GO Ready bit set on dma Tx intr\n",
				card->devname);
		
		chan->if_stats.tx_dropped++;

		clear_bit(TX_BUSY,&chan->dma_status);
		return;
	}

	if (reg&TxDMA_HI_DMA_PCI_ERROR_MASK){
		DEBUG_EVENT("%s: Error: TxDMA Intr: pci fatal errors on Tx intr 0x%lX\n",
				card->devname,reg&TxDMA_HI_DMA_PCI_ERROR_MASK);

		chan->if_stats.tx_dropped++;

		clear_bit(TX_BUSY,&chan->dma_status);
		return;
	}
	
	chan->if_stats.tx_packets++;	
	chan->if_stats.tx_bytes+=len;
	
	clear_bit(TX_BUSY,&chan->dma_status);

	DEBUG_EVENT("Tx dma done!, try again ?\n");

	xilinx_dma_tx(card,chan);
}

static void xilinx_dma_rx_complete (sdla_t *card, private_area_t *chan)
{
	unsigned long align,reg=0;
	unsigned long dma_descr;
	unsigned char critical=0;
	unsigned int len;


	DEBUG_EVENT("-- Xilinx_dma_rx_complite :--");

	dma_descr=(chan->logic_ch_num<<4) + XILINX_RxDMA_DESCRIPTOR_HI;
	reg=sdla_pci_readl(&card->hw,dma_descr);
	DEBUG_EVENT("RXDMA HI 0x%08lX\n",reg);
	
	dma_descr=(chan->logic_ch_num<<4) + XILINX_RxDMA_DESCRIPTOR_LO;
	align=sdla_pci_readl(&card->hw,dma_descr);

	DEBUG_EVENT("RXDMA LO 0x%08lX\n",align);

	align&=RxDMA_LO_ALIGNMENT_BIT_MASK;

	{
	unsigned long *data=wpabs_dma_get_vaddr(card, chan->rx_dma_buf);
	
        DEBUG_EVENT("RXD- 0x%08X 0x%08X 0x%08X 0x%08X \n",
                        data[0],
                        data[1],
                        data[2],
                        data[3]);
//	}
// M.F.	
        DEBUG_EVENT("RXD- 0x%08X 0x%08X 0x%08X 0x%08X \n",
                        data[4],
                        data[5],
                        data[6],
                        data[7]);

        DEBUG_EVENT("RXD- 0x%08X 0x%08X 0x%08X 0x%08X \n",
                        data[8],
                        data[9],
                        data[10],
                        data[11]);

        DEBUG_EVENT("RXD- 0x%08X 0x%08X 0x%08X 0x%08X \n",
                        data[12],
                        data[13],
                        data[14],
                        data[15]);
	}
//



	if (test_bit(RxDMA_HI_DMA_GO_READY_BIT,&reg)){
		DEBUG_EVENT("%s: Error: RxDMA Intr: GO Ready bit set on Rx intr\n",
				card->devname);

		critical=1;
		goto rx_comp_error;
	}

	if (reg&RxDMA_HI_DMA_PCI_ERROR_MASK){
		DEBUG_EVENT("%s: Error: RxDMA Intr: pci fatal errors on Rx intr 0x%08X\n",
				card->devname, reg&RxDMA_HI_DMA_PCI_ERROR_MASK);

		critical=1;
		goto rx_comp_error;
	}

	if (!test_bit(RxDMA_HI_DMA_FRAME_START_BIT,&reg)){ 

		DEBUG_EVENT("%s: RxDMA Intr: Start flag missing!\n",
				card->devname);
		goto rx_comp_error;
	}


	if (!test_bit(RxDMA_HI_DMA_FRAME_END_BIT,&reg)){

		DEBUG_EVENT("%s: RxDMA Intr: End flag missing!\n",
				card->devname);
		goto rx_comp_error;
        }

	if (test_bit(RxDMA_HI_DMA_CRC_ERROR_BIT,&reg) ||
	    test_bit(RxDMA_HI_DMA_FIFO_OVR_FLW_BIT,&reg)){
		DEBUG_EVENT("%s: RxDMA Intr: Crc/Fifo Overflow!\n",
				card->devname);
		chan->if_stats.rx_errors++;
		goto rx_comp_error;
	}
	

	len=reg&RxDMA_HI_DMA_DATA_LENGTH_MASK;

	DEBUG_EVENT("%s: RX LENGTH = %i\n",chan->if_name,len);
	return;

#if 0
	len=((chan->rx_dma_skb->len-len)<<2) - ~(align);

	DEBUG_EVENT("%s: RX LENGTH = %i\n",chan->if_name,len);

	return;

	wan_skb_trim(chan->rx_dma_skb,len);	
	
	wan_skb_queue_tail(&chan->wp_rx_complete_list,chan->rx_dma_skb);
	chan->rx_dma_skb=NULL;

	WAN_TASKLET_SCHEDULE(((wan_tasklet_t*)&chan->chan_task));
#endif

rx_comp_restart:

	//xilinx_dma_rx(card,chan);
	return;
	
rx_comp_error:

	chan->if_stats.rx_dropped++;

#if 0
	wan_skb_init(chan->rx_dma_skb,16);
	wan_skb_trim(chan->rx_dma_skb,0);
	
	wan_skb_queue_tail(&chan->wp_rx_free_list,chan->rx_dma_skb);
	chan->rx_dma_skb=NULL;
#endif

	if (critical){
		return;
	}
	
	goto rx_comp_restart;
	
}

static char request_xilinx_logical_channel_num (sdla_t *card, private_area_t *chan, long *free_ch)
{
	char logic_ch=-1, free_logic_ch=-1;
	int i;


	DEBUG_EVENT("-- Request_Xilinx_logic_channel_num:--");

	DEBUG_EVENT("%s:%d Global Num Timeslots=%i  Global Logic ch Map 0x%lX \n",
		__FUNCTION__,__LINE__,card->u.xilinx.num_of_time_slots,card->u.xilinx.logic_ch_map);

	for (i=0;i<card->u.xilinx.num_of_time_slots;i++){
		if (!test_and_set_bit(i,&card->u.xilinx.logic_ch_map)){
			logic_ch=i;
			break;
		}
	}

	if (logic_ch == -1){
		return logic_ch;
	}

	for (i=0;i<card->u.xilinx.num_of_time_slots;i++){
		if (!test_bit(i,&card->u.xilinx.logic_ch_map)){
			free_logic_ch=i;
			break;
		}
	}

	if (card->u.xilinx.dev_to_ch_map[(unsigned char)logic_ch]){
		DEBUG_EVENT("%s: Error, request logical ch=%i map busy\n",
				card->devname,logic_ch);
		return -1;
	}
	
	*free_ch=free_logic_ch;

	card->u.xilinx.dev_to_ch_map[(unsigned char)logic_ch]=(void*)chan;

	if (logic_ch > card->u.xilinx.top_logic_ch){
		card->u.xilinx.top_logic_ch=logic_ch;
		xilinx_dma_max_logic_ch(card);
	}

	
	DEBUG_EVENT("Binding logic ch %d  Ptr=%p\n",logic_ch,chan);
	return logic_ch;
}

static void free_xilinx_logical_channel_num (sdla_t *card, int logic_ch)
{
	clear_bit (logic_ch,&card->u.xilinx.logic_ch_map);
	card->u.xilinx.dev_to_ch_map[logic_ch]=NULL;

	if (logic_ch >= card->u.xilinx.top_logic_ch){
		int i;

		card->u.xilinx.top_logic_ch=XILINX_DEFLT_ACTIVE_CH;

		for (i=0;i<card->u.xilinx.num_of_time_slots;i++){
			if (card->u.xilinx.dev_to_ch_map[logic_ch]){
				card->u.xilinx.top_logic_ch=i;
			}
		}

		xilinx_dma_max_logic_ch(card);	
	}
}

static void xilinx_dma_max_logic_ch(sdla_t *card)
{
	unsigned long reg;


	DEBUG_EVENT("-- Xilinx_dma_max_logic_ch :--");

	reg = sdla_pci_readl(&card->hw,XILINX_DMA_CONTROL_REG);

        /* Set up the current highest active logic channel */

	reg&=DMA_ACTIVE_CHANNEL_BIT_MASK;
        reg|=(card->u.xilinx.top_logic_ch << DMA_ACTIVE_CHANNEL_BIT_SHIFT);

        sdla_pci_writel(&card->hw,XILINX_DMA_CONTROL_REG,reg);
}


/*============================================================================
 * Enable timer interrupt  
 */
static void enable_timer (void* card_id)
{
	sdla_t* 		card = (sdla_t*)card_id;

	DEBUG_EVENT( "%s: enabling timer %s\n",card->devname,
			card->wandev.dev?card->wandev.dev->name:"No DEV");
		
	return;
}

/*============================================================================
 * Firmware error handler.
 *	This routine is called whenever firmware command returns non-zero
 *	return code.
 *
 * Return zero if previous command has to be cancelled.
 */
static int frmw_error (sdla_t *card, int err, wan_mbox_t *mb)
{
	unsigned cmd = mb->wan_command;

	switch (err) {

	default:
		DEBUG_EVENT( "%s: command 0x%02X returned 0x%02X!\n",
			card->devname, cmd, err);
	}

	return 0;
}


/**SECTION************************************************** 
 *
 * 	API Bottom Half Handlers 
 *
 **********************************************************/

static void wp_bh (unsigned long data)
{
	private_area_t* chan = (private_area_t *)data;
	struct sk_buff *new_skb,*skb;

	while((skb=wan_skb_dequeue(&chan->wp_rx_complete_list)) != NULL ){
		
		if (chan->common.sk == NULL || chan->common.func == NULL){
			++chan->if_stats.rx_dropped;
			goto wp_bh_skip;
		}

		new_skb=skb_clone(skb,GFP_ATOMIC);
		if (new_skb){

			new_skb->protocol = htons(PVC_PROT);
			new_skb->mac.raw  = new_skb->data;
			new_skb->dev      = chan->dev;
			new_skb->pkt_type = WAN_PACKET_DATA;

			if (chan->common.func(new_skb,chan->dev,chan->common.sk) != 0){
				++chan->if_stats.rx_dropped;
				wan_skb_free(new_skb);
			}

			chan->if_stats.rx_packets++;
			chan->if_stats.rx_bytes+=skb->len;
		}
wp_bh_skip:
		wan_skb_init(skb,16);
		wan_skb_trim(skb,0);
	
		wan_skb_queue_tail(&chan->wp_rx_free_list,skb);
	}

	WAN_TASKLET_END(((wan_tasklet_t*)&chan->chan_task));
	return;
}

static void wakeup_sk_bh (struct net_device *dev)
{

	private_area_t *chan = dev->priv;

	/* If the sock is in the process of unlinking the
	 * driver from the socket, we must get out. 
	 * This never happends but is a sanity check. */
	if (test_bit(0,&chan->common.common_critical)){
		return;
	}

	if (!chan->common.sk || !chan->common.func){
		return;
	}

	chan->common.func(NULL,dev,chan->common.sk);
	return;
}



/**SECTION*************************************************************** 
 * 
 * 	HARDWARE Interrupt Handlers 
 *
 ***********************************************************************/


/*============================================================================
 * wpfw_isr
 *
 * Main interrupt service routine. 
 * Determin the interrupt received and handle it.
 *
 */
static void wp_xilinx_isr (sdla_t* card)
{
	int i;
	unsigned long reg;
	unsigned long dma_tx_reg,dma_rx_reg;
/*	unsigned long err_tx_reg,err_rx_reg;*/
	private_area_t *chan;

        set_bit(0,&card->in_isr);

	DEBUG_EVENT("--ISR: IN ISR --");

// M.F.
reg=sdla_pci_readl(&card->hw,0x50);
DEBUG_EVENT("0x50 =0x%lX",reg);

reg=sdla_pci_readl(&card->hw,0x54);
DEBUG_EVENT("0x54 =0x%lX",reg);


	reg=sdla_pci_readl(&card->hw,XILINX_CHIP_CFG_REG);


DEBUG_EVENT("0x40 -> 0x%lX\n",reg);

	/* Disable ALL chip interrupts */
	clear_bit(GLOBAL_INTR_ENABLE_BIT,&reg);
	clear_bit(ERROR_INTR_ENABLE_BIT,&reg);
	sdla_pci_writel(&card->hw,XILINX_CHIP_CFG_REG,reg);
	
	
//reg=sdla_pci_readl(&card->hw,0x50);
//DEBUG_EVENT("0x50 =0x%lX",reg);

//reg=sdla_pci_readl(&card->hw,0x54);
//DEBUG_EVENT("0x54 =0x%lX",reg);

//reg=sdla_pci_readl(&card->hw,0x58);
//DEBUG_EVENT("0x58 =0x%lX",reg);

//reg=sdla_pci_readl(&card->hw,0x5C);
//DEBUG_EVENT("0x5C =0x%lX\n",reg);

//reg=sdla_pci_readl(&card->hw,XILINX_CHIP_CFG_REG);
//DEBUG_EVENT("End of ISR : 0x40 -> 0x%lX\n",reg);

//card->in_isr = 0;
//return;

	if (test_bit(DMA_INTR_FLAG,&reg)){
		
		dma_rx_reg=sdla_pci_readl(&card->hw,XILINX_DMA_RX_INTR_PENDING_REG);
		dma_tx_reg=sdla_pci_readl(&card->hw,XILINX_DMA_TX_INTR_PENDING_REG);

		DEBUG_EVENT("ISR: DMA RX REG=0x%lX DMA TX REG=0x%lX\n",
			dma_rx_reg,dma_tx_reg);
		
		for (i=0; i<card->u.xilinx.num_of_time_slots ;i++){
			if (test_bit(i,&dma_rx_reg)){
				chan=(private_area_t*)card->u.xilinx.dev_to_ch_map[i];
				if (!chan){
					DEBUG_EVENT("%s: Error: No Dev for Rx logical ch=%i\n",
							card->devname,i);
					continue;
				}

				DEBUG_EVENT("\n\nIN RX INTERRUPT !\n");
				xilinx_dma_rx_complete(card,chan);
			}
		}

		
		for (i=0; i<card->u.xilinx.num_of_time_slots ;i++){

			if (test_bit(i,&dma_tx_reg)){
				chan=(private_area_t*)card->u.xilinx.dev_to_ch_map[i];
				if (!chan){
					DEBUG_EVENT("%s: Error: No Dev for Tx logical ch=%i\n",
							card->devname,i);
					continue;
				}

				DEBUG_EVENT("IN TX INTERRUPT !\n");
				xilinx_dma_tx_complete(card,chan);
			}
		}
	}else{
		DEBUG_EVENT("%s: Suprious Interrupt\n",
				card->devname);
	}

#if 0	
	if (test_bit(ERROR_INTR_FLAG,&reg)){
		
		err_rx_reg=sdla_pci_readl(&card->hw,XILINX_HDLC_RX_INTR_PENDING_REG);
		err_tx_reg=sdla_pci_readl(&card->hw,XILINX_HDLC_TX_INTR_PENDING_REG);

		for (i=0; i<card->u.xilinx.num_of_time_slots ;i++){
			if (test_bit(i,&err_rx_reg)){
				chan=card->u.xilinx.dev_to_ch_map[i];
				if (!chan){
					DEBUG_EVENT("%s: Error: No Dev for Rx logical ch=%i\n",
							card->devname,i);
					continue;
				}

				xilinx_rx_err_intr(card,chan);
			}
		}

		for (i=0; i<card->u.xilinx.num_of_time_slots ;i++){

			if (test_bit(i,&err_tx_reg)){
				chan=card->u.xilinx.dev_to_ch_map[i];
				if (!chan){
					DEBUG_EVENT("%s: Error: No Dev for Tx logical ch=%i\n",
							card->devname,i);
					continue;
				}

				xilinx_tx_err_intr(card,chan);
			}
		}
	}
#endif
	
	clear_bit(0,&card->in_isr);

	reg=sdla_pci_readl(&card->hw,XILINX_CHIP_CFG_REG);
	/* Enable ALL chip interrupts */

	set_bit(GLOBAL_INTR_ENABLE_BIT,&reg);
	/* FIXME: No error intr for now */
	/*set_bit(ERROR_INTR_ENABLE_BIT,&reg);*/
	sdla_pci_writel(&card->hw,XILINX_CHIP_CFG_REG,reg);

	return;
}



/**SECTION***********************************************************
 * 
 * WANPIPE Debugging Interfaces
 *
 ********************************************************************/



/*=============================================================================
 * process_udp_mgmt_pkt
 * 
 * Process all "wanpipemon" debugger commands.  This function
 * performs all debugging tasks: 
 *
 * 	Line Tracing
 * 	Line/Hardware Statistics
 * 	Protocol Statistics
 * 
 * "wanpipemon" utility is a user-space program that 
 * is used to debug the WANPIPE product.
 *
 */

#if 0
static int process_udp_mgmt_pkt(sdla_t* card, struct net_device* dev,
				private_area_t* priv_area, int local_dev ) 
{
	unsigned char *buf;
	unsigned int frames, len;
	struct sk_buff *new_skb;
	unsigned short buffer_length, real_len;
	volatile unsigned long data_ptr;
	unsigned data_length;
	int udp_mgmt_req_valid = 1;
	wan_udp_pkt_t *wan_udp_pkt;
	struct timeval tv;
	int err;
	char ut_char;

	wan_udp_pkt = (wan_udp_pkt_t *) priv_area->udp_pkt_data;

	if (!local_dev){

		if(priv_area->udp_pkt_src == UDP_PKT_FRM_NETWORK){

			/* Only these commands are support for remote debugging.
			 * All others are not */
			switch(wan_udp_pkt->wan_udp_command) {

				case READ_GLOBAL_STATISTICS:
				case READ_MODEM_STATUS:  
				case READ_LINK_STATUS:
				case ROUTER_UP_TIME:
				case READ_COMMS_ERROR_STATS:
				case READ_OPERATIONAL_STATS:
				/* These two commands are executed for
				 * each request */
				case READ_CONFIGURATION:
				case READ_CODE_VERSION:
				case WAN_GET_MEDIA_TYPE:
				case WAN_FE_GET_STAT:
					udp_mgmt_req_valid = 1;
					break;
				default:
					udp_mgmt_req_valid = 0;
					break;
			} 
		}
	}

  	if(!udp_mgmt_req_valid) {

		/* set length to 0 */
		wan_udp_pkt->wan_udp_data_len = 0;

    		/* set return code */
		wan_udp_pkt->wan_udp_return_code = 0xCD;

		if (net_ratelimit()){	
			DEBUG_EVENT( 
			"%s: Warning, Illegal UDP command attempted from network: %x\n",
			card->devname,wan_udp_pkt->wan_udp_command);
		}

   	} else {
	   	unsigned long trace_status_cfg_addr = 0;
		TRACE_STATUS_EL_CFG_STRUCT trace_cfg_struct;
		TRACE_STATUS_ELEMENT_STRUCT trace_element_struct;

		wan_udp_pkt->wan_udp_opp_flag = 0;

		switch(wan_udp_pkt->wan_udp_command) {

		case ENABLE_TRACING:
		     if (!priv_area->TracingEnabled) {

			/* OPERATE_DATALINE_MONITOR */
			mb->wan_data_len = sizeof(LINE_TRACE_CONFIG_STRUCT);
			mb->wan_command = SET_TRACE_CONFIGURATION;

    			((LINE_TRACE_CONFIG_STRUCT *)mb->wan_data)->
				trace_config = TRACE_ACTIVE;
			/* Trace delay mode is not used because it slows
			   down transfer and results in a standoff situation
			   when there is a lot of data */

			/* Configure the Trace based on user inputs */
			((LINE_TRACE_CONFIG_STRUCT *)mb->wan_data)->trace_config |= 
					wan_udp_pkt->wan_udp_data[0];

			((LINE_TRACE_CONFIG_STRUCT *)mb->wan_data)->
			   trace_deactivation_timer = 4000;


			err = sdla_exec(mb) ? mb->wan_return_code : CMD_TIMEOUT;
			if (err != COMMAND_OK) {
				frmw_error(card,err,mb);
				card->TracingEnabled = 0;
				wan_udp_pkt->wan_udp_return_code = err;
				mb->wan_data_len = 0;
				break;
	    		} 

			/* Get the base address of the trace element list */
			mb->wan_data_len = 0;
			mb->wan_command = READ_TRACE_CONFIGURATION;
			err = sdla_exec(mb) ? mb->wan_return_code : CMD_TIMEOUT;

			if (err != COMMAND_OK) {
				frmw_error(card,err,mb);
				priv_area->TracingEnabled = 0;
				wan_udp_pkt->wan_udp_return_code = err;
				mb->wan_data_len = 0;
				break;
	    		} 	

	   		trace_status_cfg_addr =((LINE_TRACE_CONFIG_STRUCT *)
				mb->wan_data) -> ptr_trace_stat_el_cfg_struct;

			sdla_peek(&card->hw, trace_status_cfg_addr,
				 &trace_cfg_struct, sizeof(trace_cfg_struct));
		    
			priv_area->start_trace_addr = trace_cfg_struct.
				base_addr_trace_status_elements;

			priv_area->number_trace_elements = 
					trace_cfg_struct.number_trace_status_elements;

			priv_area->end_trace_addr = (unsigned long)
					((TRACE_STATUS_ELEMENT_STRUCT *)
					 priv_area->start_trace_addr + 
					 (priv_area->number_trace_elements - 1));

			priv_area->base_addr_trace_buffer = 
					trace_cfg_struct.base_addr_trace_buffer;

			priv_area->end_addr_trace_buffer = 
					trace_cfg_struct.end_addr_trace_buffer;

		    	priv_area->curr_trace_addr = 
					trace_cfg_struct.next_trace_element_to_use;

	    		priv_area->available_buffer_space = 2000 - 
								  sizeof(struct iphdr) -
								  sizeof(struct udphdr) -
							      	  sizeof(wan_mgmt_t)-
								  sizeof(wan_cmd_t)-
								  sizeof(wan_trace_info_t);	
	       	     }

		     wan_udp_pkt->wan_udp_return_code = COMMAND_OK;
		     mb->wan_data_len = 0;
	       	     priv_area->TracingEnabled = 1;
	       	     break;
	   

		case DISABLE_TRACING:
		     if (priv_area->TracingEnabled) {

			/* OPERATE_DATALINE_MONITOR */
			mb->wan_data_len = sizeof(LINE_TRACE_CONFIG_STRUCT);
			mb->wan_command = SET_TRACE_CONFIGURATION;
    			((LINE_TRACE_CONFIG_STRUCT *)mb->wan_data)->
				trace_config = TRACE_INACTIVE;
			err = sdla_exec(mb) ? mb->wan_return_code : CMD_TIMEOUT;
		     }		

		     priv_area->TracingEnabled = 0;
		     wan_udp_pkt->wan_udp_return_code = COMMAND_OK;
		     mb->wan_data_len = 0;
		     break;
	   

		case GET_TRACE_INFO:

		     if (!priv_area->TracingEnabled) {
			wan_udp_pkt->wan_udp_return_code = 1;
			mb->wan_data_len = 0;
			break;
		     }

  		     wan_udp_pkt->wan_udp_ismoredata = 0x00;
		     buffer_length = 0;	/* offset of packet already occupied */

		     for (frames=0; frames < priv_area->number_trace_elements; frames++){

			trace_pkt_t *trace_pkt = (trace_pkt_t *)
				&wan_udp_pkt->wan_udp_data[buffer_length];

			sdla_peek(&card->hw, priv_area->curr_trace_addr,
			   	  (unsigned char *)&trace_element_struct,
			   	  sizeof(TRACE_STATUS_ELEMENT_STRUCT));

     			if (trace_element_struct.opp_flag == 0x00) {
			 	break;
			}

			/* get pointer to real data */
			data_ptr = trace_element_struct.ptr_data_bfr;

			/* See if there is actual data on the trace buffer */
			if (data_ptr){
				data_length = trace_element_struct.trace_length;
			}else{
				data_length = 0;
				wan_udp_pkt->wan_udp_ismoredata = 0x01;
			}
	
   			if( (priv_area->available_buffer_space - buffer_length)
				< ( sizeof(trace_pkt_t) + data_length) ) {

                            /* indicate there are more frames on board & exit */
				wan_udp_pkt->wan_udp_ismoredata = 0x01;
                               	break;
                         }

			trace_pkt->status = trace_element_struct.trace_type;

			trace_pkt->time_stamp =
				trace_element_struct.trace_time_stamp;

			trace_pkt->real_length =
				trace_element_struct.trace_length;

			/* see if we can fit the frame into the user buffer */
			real_len = trace_pkt->real_length;

			if (data_ptr == 0) {
			     	trace_pkt->data_avail = 0x00;
			} else {
				unsigned tmp = 0;

				/* get the data from circular buffer
				    must check for end of buffer */
			        trace_pkt->data_avail = 0x01;

				if ((data_ptr + real_len) >
					     priv_area->end_addr_trace_buffer + 1){

				    	tmp = priv_area->end_addr_trace_buffer - data_ptr + 1;
				    	sdla_peek(&card->hw, data_ptr,
					       	  trace_pkt->data,tmp);
				    	data_ptr = priv_area->base_addr_trace_buffer;
				}
	
		        	sdla_peek(&card->hw, data_ptr,
					  &trace_pkt->data[tmp], real_len - tmp);
			}	

			/* zero the opp flag to show we got the frame */
			ut_char = 0x00;
			sdla_poke(&card->hw, priv_area->curr_trace_addr, &ut_char, 1);

       			/* now move onto the next frame */
       			priv_area->curr_trace_addr += sizeof(TRACE_STATUS_ELEMENT_STRUCT);

       			/* check if we went over the last address */
			if ( priv_area->curr_trace_addr > priv_area->end_trace_addr ) {
				priv_area->curr_trace_addr = priv_area->start_trace_addr;
       			}

            		if(trace_pkt->data_avail == 0x01) {
				buffer_length += real_len - 1;
			}
	 
	       	    	/* for the header */
	            	buffer_length += sizeof(trace_pkt_t);

		     }  /* For Loop */

		     if (frames == priv_area->number_trace_elements){
			wan_udp_pkt->wan_udp_ismoredata = 0x01;
	             }
 		     wan_udp_pkt->wan_udp_num_frames = frames;
		 
    		     mb->wan_data_len = buffer_length;
		     wan_udp_pkt->wan_udp_data_len = buffer_length; 
		 
		     wan_udp_pkt->wan_udp_return_code = COMMAND_OK; 
		     
		     break;


		case FT1_READ_STATUS:
			((unsigned char *)wan_udp_pkt->wan_udp_data )[0] =
				flags->FT1_info_struct.parallel_port_A_input;

			((unsigned char *)wan_udp_pkt->wan_udp_data )[1] =
				flags->FT1_info_struct.parallel_port_B_input;
				
			wan_udp_pkt->wan_udp_return_code = COMMAND_OK;
			wan_udp_pkt->wan_udp_data_len = 2;
			mb->wan_data_len = 2;
			break;

		case ROUTER_UP_TIME:
			do_gettimeofday( &tv );
			priv_area->router_up_time = tv.tv_sec - 
					priv_area->router_start_time;
			*(unsigned long *)&wan_udp_pkt->wan_udp_data = 
					priv_area->router_up_time;	
			mb->wan_data_len = sizeof(unsigned long);
			wan_udp_pkt->wan_udp_data_len = sizeof(unsigned long);
			wan_udp_pkt->wan_udp_return_code = COMMAND_OK;
			break;

   		case FT1_MONITOR_STATUS_CTRL:
			/* Enable FT1 MONITOR STATUS */
	        	if ((wan_udp_pkt->wan_udp_data[0] & ENABLE_READ_FT1_STATUS) ||  
				(wan_udp_pkt->wan_udp_data[0] & ENABLE_READ_FT1_OP_STATS)) {
			
			     	if( rCount++ != 0 ) {
					wan_udp_pkt->wan_udp_return_code = COMMAND_OK;
					mb->wan_data_len = 1;
		  			break;
		    	     	}
	      		}

	      		/* Disable FT1 MONITOR STATUS */
	      		if( wan_udp_pkt->wan_udp_data[0] == 0) {

	      	   	     	if( --rCount != 0) {
		  			wan_udp_pkt->wan_udp_return_code = COMMAND_OK;
					mb->wan_data_len = 1;
		  			break;
	   	    	     	} 
	      		} 	
			goto dflt_1;

		case WAN_GET_MEDIA_TYPE:
		case WAN_FE_GET_STAT:
		case WAN_FE_SET_LB_MODE:
 		case WAN_FE_FLUSH_PMON:
		case WAN_FE_GET_CFG:
			if (IS_TE1_CARD(card)){
				sdla_te_udp(card, 
					    &wan_udp_pkt->wan_udp_cmd,
					    &wan_udp_pkt->wan_udp_data[0]);
			}else if (IS_56K_CARD(card)){
				sdla_56k_udp(card, 
					     &wan_udp_pkt->wan_udp_cmd,
					     &wan_udp_pkt->wan_udp_data[0]);
			}else{
				if (wan_udp_pkt->wan_udp_command == WAN_GET_MEDIA_TYPE){
		    			wan_udp_pkt->wan_udp_data_len = sizeof(unsigned char); 
					wan_udp_pkt->wan_udp_return_code = CMD_OK;
				}else{
					wan_udp_pkt->wan_udp_return_code = WAN_UDP_INVALID_CMD;
				}
			}
			mb->wan_data_len = wan_udp_pkt->wan_udp_data_len;
			break;
#if 0
		case WAN_FE_SET_LB_MODE:
		    /* Activate/Deactivate Line Loopback modes */
		    if (IS_TE1_CARD(card)){
			err = sdla_set_te1_lb_modes(card, 
						    wan_udp_pkt->wan_udp_data[0], 
						    wan_udp_pkt->wan_udp_data[1]);
			wan_udp_pkt->wan_udp_return_code = 
					(!err) ? CMD_OK : WAN_UDP_FAILED_CMD;
		    }else{
			wan_udp_pkt->wan_udp_return_code = WAN_UDP_INVALID_CMD;
		    }
		    wan_udp_pkt->wan_udp_data_len = 0x00;
		    break;

		case WAN_GET_MEDIA_TYPE:
	 		wan_udp_pkt->wan_udp_data[0] = 
				(IS_T1_CARD(card) ? WANOPT_MEDIA_T1 :
				 IS_E1_CARD(card) ? WANOPT_MEDIA_E1 :
				 IS_56K_CARD(card) ? WANOPT_MEDIA_56K : 
				 				WANOPT_MEDIA_NONE);
		    	wan_udp_pkt->wan_udp_data_len = sizeof(unsigned char); 
			wan_udp_pkt->wan_udp_return_code = COMMAND_OK;
			mb->wan_data_len = sizeof(unsigned char);
			break;

		case WAN_FE_GET_STAT:
		    	if (IS_TE1_CARD(card)) {	
	 	    		/* TE1_56K Read T1/E1/56K alarms */
			  	*(unsigned long *)&wan_udp_pkt->wan_udp_data = 
		                        sdla_te_alarm(card, 0);
				/* TE1 Update T1/E1 perfomance counters */
    				sdla_te_pmon(card);
	        		memcpy(&wan_udp_pkt->wan_udp_data[sizeof(unsigned long)],
					&card->wandev.te_pmon,
					sizeof(pmc_pmon_t));
		        	wan_udp_pkt->wan_udp_return_code = COMMAND_OK;
		    		wan_udp_pkt->wan_udp_data_len = 
					sizeof(unsigned long) + sizeof(pmc_pmon_t); 
				mb->wan_data_len = wan_udp_pkt->wan_udp_data_len;
			}else if (IS_56K_CARD(card)){
				/* 56K Update CSU/DSU alarms */
				card->wandev.k56_alarm = sdla_56k_alarm(card, 1); 
			 	*(unsigned long *)&wan_udp_pkt->wan_udp_data = 
			                        card->wandev.k56_alarm;
				wan_udp_pkt->wan_udp_return_code = COMMAND_OK;
	    			wan_udp_pkt->wan_udp_data_len = sizeof(unsigned long);
				mb->wan_data_len = wan_udp_pkt->wan_udp_data_len;
			}
		    	break;

 		case WAN_FE_FLUSH_PMON:
 	    		/* TE1 Flush T1/E1 pmon counters */
	    		if (IS_TE1_CARD(card)){	
				sdla_flush_te1_pmon(card);
	        		wan_udp_pkt->wan_udp_return_code = COMMAND_OK;
	    		}
	    		break;

		case WAN_FE_GET_CFG:
			/* Read T1/E1 configuration */
			DEBUG_EVENT("Read T1/E1 configuration!\n");
	    		if (IS_TE1_CARD(card)){	
        			memcpy(&wan_udp_pkt->wan_udp_data[0],
					&card->wandev.te_cfg,
					sizeof(sdla_te_cfg_t));
		        	wan_udp_pkt->wan_udp_return_code = COMMAND_OK;
	    			wan_udp_pkt->wan_udp_data_len = sizeof(sdla_te_cfg_t);
				mb->wan_data_len = wan_udp_pkt->wan_udp_data_len;
			}
			break;
#endif
			
		case WAN_GET_PROTOCOL:
		   	wan_udp_pkt->wan_udp_num_frames = card->wandev.config_id;
		    	wan_udp_pkt->wan_udp_return_code = CMD_OK;
		    	mb->wan_data_len = wan_udp_pkt->wan_udp_data_len = 1;
		    	break;

		case WAN_GET_PLATFORM:
		    	wan_udp_pkt->wan_udp_data[0] = WAN_LINUX_PLATFORM;
		    	wan_udp_pkt->wan_udp_return_code = CMD_OK;
		    	mb->wan_data_len = wan_udp_pkt->wan_udp_data_len = 1;
		    	break;

		default:
dflt_1:
			/* it's a board command */
			mb->wan_command = wan_udp_pkt->wan_udp_command;
			mb->wan_data_len = wan_udp_pkt->wan_udp_data_len;
			if (mb->wan_data_len) {
				memcpy(&mb->wan_data, (unsigned char *) wan_udp_pkt->
							wan_udp_data, mb->wan_data_len);
	      		} 

			/* run the command on the board */
			err = sdla_exec(mb) ? mb->wan_return_code : CMD_TIMEOUT;
			if (err != COMMAND_OK) {
				frmw_error(card,err,mb);
				wan_udp_pkt->wan_udp_return_code = mb->wan_return_code;
				break;
			}
			
			/* copy the result back to our buffer */
	         	memcpy(&wan_udp_pkt->wan_udp_hdr.wan_cmd, &mb->wan_command, sizeof(wan_cmd_t)); 
			
			if (mb->wan_data_len) {
	         		memcpy(&wan_udp_pkt->wan_udp_data, &mb->wan_data, 
								mb->wan_data_len); 
	      		}

		} /* end of switch */
     	} /* end of else */

     	/* Fill UDP TTL */
	wan_udp_pkt->wan_ip_ttl= card->wandev.ttl; 

	if (local_dev){
		wan_udp_pkt->wan_udp_request_reply = UDPMGMT_REPLY;
		return 1;
	}
	
	len = reply_udp(priv_area->udp_pkt_data, mb->wan_data_len);

     	if(priv_area->udp_pkt_src == UDP_PKT_FRM_NETWORK){

		/* Must check if we interrupted if_send() routine. The
		 * tx buffers might be used. If so drop the packet */
	   	if (!test_bit(SEND_CRIT,&card->wandev.critical)) {
		
			if(!frmw_send(card, priv_area->udp_pkt_data, len, 0)) {
				++ chan->if_stats.tx_packets;
				chan->if_stats.tx_bytes += len;
			}
		}
	} else {	
	
		/* Pass it up the stack
    		   Allocate socket buffer */
		if ((new_skb = wan_skb_alloc(len),dev) != NULL) {
			/* copy data into new_skb */

 	    		buf = wan_skb_put(new_skb, len);
  	    		memcpy(buf, priv_area->udp_pkt_data, len);

            		/* Decapsulate pkt and pass it up the protocol stack */
	    		new_skb->protocol = htons(ETH_P_IP);
            		new_skb->dev = dev;
	    		new_skb->mac.raw  = new_skb->data;

			netif_rx(new_skb);
		} else {
	    	
			DEBUG_EVENT( "%s: no socket buffers available!\n",
					card->devname);
  		}
    	}

	atomic_set(&priv_area->udp_pkt_len,0);
 	
	return 0;
}

/*==============================================================================
 * udp_pkt_type
 * 
 * Determine what type of UDP call it is. ?
 *
 */
static int udp_pkt_type(struct sk_buff *skb, sdla_t* card)
{
	 wan_udp_pkt_t *wan_udp_pkt = (wan_udp_pkt_t *)skb->data;

#ifdef _WAN_UDP_DEBUG
		DEBUG_EVENT( "SIG %s = %s\n\
				  UPP %x = %x\n\
				  PRT %x = %x\n\
				  REQ %i = %i\n\
				  36 th = %x 37th = %x\n",
				  wan_udp_pkt->wan_udp_signature,
				  UDPMGMT_SIGNATURE,
				  wan_udp_pkt->wan_udp_dport,
				  ntohs(card->wandev.udp_port),
				  wan_udp_pkt->wan_ip_p,
				  UDPMGMT_UDP_PROTOCOL,
				  wan_udp_pkt->wan_udp_request_reply,
				  UDPMGMT_REQUEST,
				  skb->data[36], skb->data[37]);
#endif	
		
	if ((wan_udp_pkt->wan_udp_dport == ntohs(card->wandev.udp_port)) &&
	   (wan_udp_pkt->wan_ip_p == UDPMGMT_UDP_PROTOCOL) &&
	   (wan_udp_pkt->wan_udp_request_reply == UDPMGMT_REQUEST)) {
		if (!strncmp(wan_udp_pkt->wan_udp_signature,UDPMGMT_SIGNATURE,8)){
			return UDP_TYPE;
		}
		if (!strncmp(wan_udp_pkt->wan_udp_signature,GLOBAL_UDP_SIGNATURE,8)){
			return UDP_TYPE;
		}
	}
	return UDP_INVALID_TYPE;
}
#endif


/**SECTION*************************************************************
 *
 * 	TASK Functions and Triggers
 *
 **********************************************************************/


/*============================================================================
 * port_set_state
 * 
 * Set PORT state.
 *
 */
static void port_set_state (sdla_t *card, int state)
{
        if (card->wandev.state != state)
        {
                switch (state)
                {
                case WAN_CONNECTED:
                        DEBUG_EVENT( "%s: Link connected!\n",
                                card->devname);
                      	break;

                case WAN_CONNECTING:
                        DEBUG_EVENT( "%s: Link connecting...\n",
                                card->devname);
                        break;

                case WAN_DISCONNECTED:
                        DEBUG_EVENT( "%s: Link disconnected!\n",
                                card->devname);
                        break;
                }

                card->wandev.state = state;
		if (card->wandev.dev){
			struct net_device *dev = card->wandev.dev;
			private_area_t *priv_area = dev->priv;
			priv_area->common.state = state;

			if (priv_area->common.usedby == API){
				wakeup_sk_bh(dev);	
			}
		}
        }
}

/*============================================================
 * handle_front_end_state
 *
 * Front end state indicates the physical medium that
 * the Z80 backend connects to.  
 * 
 * S514-1/2/3:		V32/RS232/FT1 Front End
 * 	  		Front end state is determined via 
 * 	 		Modem/Status.
 * S514-4/5/7/8:	56K/T1/E1 Front End
 * 			Front end state is determined via
 * 			link status interrupt received
 * 			from the front end hardware.
 *
 * If the front end state handler is enabed by the
 * user.  The interface state will follow the 
 * front end state. I.E. If the front end goes down
 * the protocol and interface will be declared down.
 *
 * If the front end state is UP, then the interface
 * and protocol will be up ONLY if the protocol is
 * also UP.
 *
 * Therefore, we must have three state variables
 * 1. Front End State (card->wandev.front_end_status)
 * 2. Protocol State  (card->wandev.state)
 * 3. Interface State (dev->flags & IFF_UP)
 *
 */

static void handle_front_end_state(sdla_t *card)
{
	if (card->wandev.ignore_front_end_status == WANOPT_YES){
		return;
	}

	if (card->wandev.front_end_status == FE_CONNECTED){
		if (card->u.c.state == WAN_CONNECTED){
			port_set_state(card,WAN_CONNECTED);
		}
	}else{
		port_set_state(card,WAN_DISCONNECTED);
	}
}

/****** End ****************************************************************/
