#include <linux/version.h>
#include <linux/config.h>	/* OS configuration options */
#include <linux/stddef.h>	/* offsetof(), etc. */
#include <linux/errno.h>	/* return codes */
#include <linux/string.h>	/* inline memset(), etc. */
#include <linux/slab.h>		/* kmalloc(), kfree() */
#include <linux/kernel.h>	/* printk(), and other useful stuff */
#include <linux/module.h>	/* support for loadable modules */
#include <linux/ioport.h>	/* request_region(), release_region() */

#include <linux/in.h>		/* sockaddr_in */
#include <linux/inet.h>	
#include <linux/if.h>

#include <asm/uaccess.h>
#include <linux/netdevice.h>

#include <linux/wanpipe_defines.h>
#include <linux/sdla_sdlc.h>
#include <linux/if_wanpipe.h>


MODULE_AUTHOR ("Nenad Corbic <ncorbic@sangoma.com>");
MODULE_DESCRIPTION ("Sangoma WANPIPE: Test Driver");
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,9) 
MODULE_LICENSE("GPL");
#endif

#define MODNAME  "test-driver"

static int wanpipe_notifier(struct notifier_block *this, unsigned long msg, void *data);


static struct net_device *dev;
static wan_mbox_t mbox;

struct notifier_block wanpipe_netdev_notifier={
	wanpipe_notifier,
	NULL,
	0
};

static int wanpipe_notifier(struct notifier_block *this, unsigned long msg, void *data)
{
	struct net_device *tmpdev = (struct net_device*)data;
	int err;

	if (tmpdev == dev){
		printk(KERN_INFO "%s: Network device %s down!\n",
						MODNAME,dev->name);

		err=wanpipe_sdlc_unregister(dev);
		if (err != 0){
			printk(KERN_INFO "%s: SDLC Un-registration failed rc=0x%0x\n",
					MODNAME,err);
		}else{
			printk(KERN_INFO "%s: SDLC Un-registration OK\n",
					MODNAME);
		}

		dev_put(dev);
		unregister_netdevice_notifier(&wanpipe_netdev_notifier);
		dev=NULL;
	}

	return NOTIFY_DONE;
}


static int do_command(void)
{
	struct ifreq ifr;
	
	if (!dev->do_ioctl){
		return -EFAULT;
	}

	ifr.ifr_data = (char *)&mbox;

	/* Not the do_ioctl uses spin_lock_irqsave() and
	 * spin_lock_irqrestore() */
	return dev->do_ioctl(dev,&ifr,SIOC_WANPIPE_EXEC_CMD);
}

static int config_sdlc(void)
{

	memset(&mbox,0,sizeof(mbox));

	/* The user should add as many station as needed
	 * at this point */
	mbox.wan_command=ADD_STATION;
	mbox.wan_data_len=0;
	mbox.wan_sdlc_address=0xA1;
	mbox.wan_sdlc_poll_interval=30;

	if (do_command() != 0){
		printk(KERN_INFO "Failed to add station!\n");
		return -EINVAL;
	}
	printk(KERN_INFO "Added station %x\n",mbox.wan_sdlc_address);

	mbox.wan_command=ACTIVATE_STATION;
	mbox.wan_data_len=0;
	mbox.wan_sdlc_address=0xA1;
	mbox.wan_sdlc_poll_interval=30;

	if (do_command() != 0){
		printk(KERN_INFO "Failed to activate station!\n");
		return -EINVAL;
	}

	printk(KERN_INFO "Activated station %x\n",mbox.wan_sdlc_address);

	return 0;
}


/* 
 * Return  0:  the lower layer will assume
 *             that the skb buffer will be handled by the
 *             upper layer. 
 * Return <0:  the skb buffer will be deallocated
 *             by the bottom layer.
 */             
static int sdlc_rx (struct sk_buff *skb)
{
	struct net_device *dev;
	wan_api_t *wan_api;
	
	if (!skb || !skb->dev){
		printk(KERN_INFO "%s: No skb dev/ptr\n",
				MODNAME);
		return -ENODEV;
	}

	if (skb->len < sizeof(wan_api_hdr_t)){
		printk(KERN_INFO "%s: Rx skb len too short %i\n",MODNAME,skb->len);
		dev_kfree_skb_any(skb);
		return 0;
	}
	
	wan_api=(wan_api_t*)skb->data;
	dev=skb->dev;
	
	switch (skb->pkt_type){
	
		case WAN_PACKET_DATA:
			printk (KERN_INFO "%s: Got rx packet on dev=%s sdlc_addr=%i len=%i\n",
					MODNAME,dev->name,wan_api->wan_api_sdlc_station,skb->len);
			break;
		case WAN_PACKET_CMD:
			printk (KERN_INFO "%s: Got OOB Message on dev %s\n",
					MODNAME,dev->name);
			printk (KERN_INFO "%s: Command 0x%x  Len=%i\n",
					MODNAME,wan_api->wan_api_pktType,wan_api->wan_api_length);
			break;
	}
	dev_kfree_skb_any(skb);

	return 0;
}

int init_module (void)
{
	int err;
	
	wp_sdlc_reg_t wp_sdlc_reg;
	wp_sdlc_reg.sdlc_stack_rx=sdlc_rx;
	
	dev=NULL;
	
	printk(KERN_INFO "%s: init_module()\n",
				MODNAME);

	dev = dev_get_by_name("sdlc0");
	if (!dev){
		printk(KERN_INFO "%s: Failed to obtain dev sdlc0\n",
				MODNAME);
		return -ENODEV;
	}

	err = wanpipe_sdlc_register(dev,&wp_sdlc_reg);
	if (err != 0){
		dev_put(dev);
		printk(KERN_INFO "%s: SDLC Registration failed rc=%i \n",
				MODNAME,err);
		return -EINVAL;
	}

	printk(KERN_INFO "%s: SDLC Registration successful\n",
				MODNAME);

	err=config_sdlc();
	if (err != 0){
		dev_put(dev);
		printk(KERN_INFO "%s: SDLC configuration failed rc=%i \n",
				MODNAME,err);
		return -EINVAL;
	}

	register_netdevice_notifier(&wanpipe_netdev_notifier);
	
	return 0;
	
}


void cleanup_module (void)
{
	int err;
	printk(KERN_INFO "%s: cleanup_module()\n",
				MODNAME);

	if (!dev){
		return; 
	}
		
	err=wanpipe_sdlc_unregister(dev);
	if (err != 0){
		printk(KERN_INFO "%s: SDLC Un-registration failed rc=0x%0x\n",
				MODNAME,err);
	}else{
		printk(KERN_INFO "%s: SDLC Un-registration OK\n",
				MODNAME);
	}

	dev_put(dev);
	unregister_netdevice_notifier(&wanpipe_netdev_notifier);
}



