/*****************************************************************************
 * sdla_voip.c	WANPIPE(tm) Multiprotocol WAN Link Driver. 
 *				VoIP board configuration.
 *
 * Author: 	Alex Feldman  <al.feldman@sangoma.com>
 *
 * Copyright:	(c) 1995-2001 Sangoma Technologies Inc.
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 * ============================================================================
 * Jul 22, 2001	Nenad Corbic	Initial version.
 * Oct 01, 2001 Gideon Hack	Modifications for interrupt usage.
 ******************************************************************************
 */


/*
 ******************************************************************************
			   INCLUDE FILES
 ******************************************************************************
*/

#if (defined __FreeBSD__) | (defined __OpenBSD__)
# include <net/wanpipe_config.h>
# include <net/wanpipe_includes.h>
# include <net/wanpipe.h>
#elif (defined __WINDOWS__)
# include <wanpipe\csu_dsu.h>
#else
# include "zaptel/zaptel.h"
# include <linux/wanpipe_includes.h>
# include <linux/sdla_voip.h>	/* WANPIPE VoIP definitions */
# include <linux/wanpipe.h>	/* WANPIPE common user API definitions */
#endif

/*
 ******************************************************************************
			  DEFINES AND MACROS
 ******************************************************************************
*/
#define WP_MAX_CARDS	32

#define FIXME_MSG(func)		DEBUG_EVENT("(%s): FIXME: line %d\n", func, __LINE__)
#define DBG_FUNC_START(func)	DEBUG_EVENT("(DEBUG)(VoIP): %s - Start\n", func)
#define DBG_FUNC_END(func)	DEBUG_EVENT("(DEBUG)(VoIP): %s - End\n", func)

#define WP_FLAG_STOP			0x01
#define WP_FLAG_RUNNING			0x02
#define WP_FLAG_DATA_PER_INTERFACE	0x04

/*
 ******************************************************************************
			STRUCTURES AND TYPEDEFS
 ******************************************************************************
*/
typedef struct voip_pvt_area
{
	sdla_t*		card;
	char*		devname;
	// FIXME struct pci_dev *dev;
	spinlock_t	lock;
	spinlock_t	tx_rx_lock;
	int		ise1;
	int		num;
	int		flags;
	int		lbo;
	int 		lcode;
	int 		frame;

	int		usecount;
	int		sync;
	int		dead;
	int		blinktimer;
	int		alarmtimer;
	int		loopupcnt;
	int		loopdowncnt;
#ifdef FANCY_ALARM
	int		alarmpos;
#endif
	/* T1 signalling */
	struct zt_span	span;					/* Span */
	struct zt_chan	chans[31];				/* Channels */
	unsigned long	timeslot_map;
	int		timeslots;
} wp_voip_pvt_t;


/*
*******************************************************************************
**			   GLOBAL VARIABLES
*******************************************************************************
*/
static int 	wp_cards_voip_no = 0;
static unsigned char brt[256];

/*
*******************************************************************************
**			  FUNCTION PROTOTYPES
*******************************************************************************
*/
static int wp_voip_software_init(wp_voip_pvt_t* wp);
static int wp_voip_startup(struct zt_span *span);
static int wp_voip_shutdown(struct zt_span *span);
static int wp_voip_maint(struct zt_span *span, int cmd);
static int wp_voip_chanconfig(struct zt_chan *chan, int sigtype);
static int wp_voip_spanconfig(struct zt_span *span, struct zt_lineconfig *lc);
static int wp_voip_rbsbits(struct zt_chan *chan, int bits);
static int wp_voip_open(struct zt_chan *chan);
static int wp_voip_close(struct zt_chan *chan);
static void wp_voip_release(wp_voip_pvt_t *wp);
static int wp_voip_ioctl(struct zt_chan *chan, unsigned int cmd, unsigned long data);
static int wp_voip_receiveprep(void* voip, unsigned char*, int);
static int  wp_voip_transmitprep(void* voip, unsigned char*, int);

static inline void start_alarm(wp_voip_pvt_t* wp);
static inline void stop_alarm(wp_voip_pvt_t* wp);

/******************************************************************************
** () - 
**
**
*/

/*
**			  FUNCTION DEFINITIONS
**
*/
int wp_voip_check_mtu(void* pcard, unsigned long timeslot_map)
{
	sdla_t	*card = (sdla_t*)pcard;
	int	x, num_of_channels = 0, max_channels;

	if (IS_T1_CARD(card)){
		max_channels = 24;
	}else{
		max_channels = 31;
	}
	for (x = 0; x < max_channels; x++) {
		if (test_bit(x,&timeslot_map)){
			num_of_channels++;
		}
	}
	return (ZT_CHUNKSIZE * num_of_channels);
}

/******************************************************************************
** wp_voip_init() - 
**
**	OK
*/
int wp_voip_init(void* pcard, unsigned long timeslot_map, int data_per_interface)
{
	struct sk_buff*	skb;
	sdla_t*		card = (sdla_t*)pcard;
	wan_voip_t*	wan_voip = &card->wan_voip;
	wp_voip_pvt_t*	wp = NULL;
	int		i = 0, err;
	
	// ALEX-VOIP
	if (!timeslot_map){
		DEBUG_EVENT("%s: No timeslots configured for Voice!\n",
				card->devname);
		err = -EINVAL;
		goto wp_voip_init_error;
	}
	
	wan_voip->max_timeslots = GET_TE_CHANNEL_RANGE(card);
	wan_voip->max_tx_len = card->wandev.mtu;
	skb_queue_head_init(&wan_voip->rx_queue);
	skb_queue_head_init(&wan_voip->rx_free_queue);
	wan_voip->tx_skb = dev_alloc_skb(wan_voip->max_tx_len+1);
	if (wan_voip->tx_skb == NULL){
		err = -ENOMEM;
		goto wp_voip_init_error;
	}
	for (i=0;i<10;i++){
		skb=dev_alloc_skb(wan_voip->max_tx_len+1);
		if (!skb){
			err = -ENOMEM;
			goto wp_voip_init_error;
		}
		skb_queue_tail(&wan_voip->rx_free_queue,skb);
	}
	card->wandev.te_report_rbsbits = wp_voip_report_rbsbits;
	card->wandev.te_report_alarms = wp_voip_report_alarms;	

	if (card->wandev.te_signaling_config && card->wandev.te_read_signaling_config){
		if (card->wandev.te_signaling_config(card, timeslot_map)){
			err = -EINVAL;
			goto wp_voip_init_error;
		}
		if (card->wandev.te_read_signaling_config(card)){
			err = -EINVAL;
			goto wp_voip_init_error;
		}
	}
	wan_voip->timeslot_map = timeslot_map;

	wp = wan_malloc(sizeof(wp_voip_pvt_t));
	if (wp == NULL){
		if (card->wandev.te_disable_signaling){
			card->wandev.te_disable_signaling(card,timeslot_map); 
		}
		err = -ENOMEM;
		goto wp_voip_init_error;
	}
	memset(wp, 0x0, sizeof(wp_voip_pvt_t));

	wp->card		= card;
	wp->devname		= card->devname;
	wp->lcode		= card->wandev.te_cfg.lcode;
	wp->frame		= card->wandev.te_cfg.frame;
	wp->lbo			= card->wandev.te_cfg.lbo;
	wp->timeslot_map	= timeslot_map;
	wp->ise1		= IS_E1_CARD(card) ? 1 : 0;
	spin_lock_init(&wp->lock);
	spin_lock_init(&wp->tx_rx_lock);

	/* Initialize ABCD bits */
//	for(i = 0; i < wan_voip->max_timeslots; i++){
//		sdla_te_set_RBS(wp->card, i+1, 0x00/*BIT_SIGX_SIGDATA_A | BIT_SIGX_SIGDATA_C*/);
//	}
	/* Misc. software stuff */
	if ((err = wp_voip_software_init(wp))){
		if (card->wandev.te_disable_signaling){
			card->wandev.te_disable_signaling(card,timeslot_map); 
		}
		wan_free(wp);
		goto wp_voip_init_error;
		
	}
	wp->flags |= WP_FLAG_RUNNING;
	if (data_per_interface){
		wp->flags |= WP_FLAG_DATA_PER_INTERFACE;
	}

	wan_voip->voip = wp;
	return 0;

wp_voip_init_error:
	wp_voip_free(wan_voip);
	return err;
}


/******************************************************************************
** wp_stop_remove() - 
**
**	OK
*/
void wp_voip_free(wan_voip_t* wan_voip)
{
	struct sk_buff *skb;
	wp_voip_pvt_t*	wp = NULL;
	
	if (wan_voip->voip == NULL){
		return;
	}
	wp = wan_voip->voip;
	if (wan_voip->tx_skb){
		dev_kfree_skb(wan_voip->tx_skb);
		wan_voip->tx_skb = NULL;
	}
	while ((skb=skb_dequeue(&wan_voip->rx_free_queue))!=NULL){
		dev_kfree_skb_any(skb);
	}

	while ((skb=skb_dequeue(&wan_voip->rx_queue))!=NULL){
		dev_kfree_skb_any(skb);
	}
	return;
}

/******************************************************************************
** wp_stop_remove() - 
**
**	OK
*/
int wp_voip_remove(void* pcard)
{
	sdla_t*		card = (sdla_t*)pcard;
	wan_voip_t*	wan_voip = &card->wan_voip;
	wp_voip_pvt_t*	wp = NULL;

	WAN_ASSERT2(wan_voip->voip == NULL, 0);
	wp = wan_voip->voip;
	wp->flags &= ~WP_FLAG_RUNNING;
	
	if (card->wandev.te_disable_signaling){
		card->wandev.te_disable_signaling(card,wan_voip->timeslot_map); 
	}

	card->wandev.te_report_rbsbits = NULL;
	card->wandev.te_report_alarms = NULL;	
	/* Release span, possibly delayed */
	if (!wp->usecount){
		wan_voip->voip = NULL;
		wp_voip_free(wan_voip);
		wp_voip_release(wp);
		wan_voip->voip = NULL;
	}else{
		DEBUG_EVENT("%s: Application still using the driver!\n",
				card->devname);
		wp->dead = 1;
		return -EINVAL;
	}
	return 0;
}

/******************************************************************************
** wp_voip_report_rbsbits() - Report A,B bit status changes to VoIP requests. 
**
**	DONE
*/
void wp_voip_report_rbsbits(void* pcard, int channel, unsigned char status)
{
	sdla_t*	card = (sdla_t*)pcard;
	wan_voip_t*	wan_voip = &card->wan_voip;
	wp_voip_pvt_t*	wp = NULL;
	int 		rxs = 0;
	// int a = 0, b = 0, i = 0, y = 0;

	WAN_ASSERT1(wan_voip->voip == NULL);
	wp = (wp_voip_pvt_t*)wan_voip->voip;
	if (!(wp->flags & WP_FLAG_RUNNING)) return;
	
	if (!test_bit(channel-1, &wan_voip->timeslot_map)){
		return;
	}
	if (wp->ise1){
		FIXME_MSG("wp_voip_report_rbsbits");
#if 0
		/* Read 5 registers at a time, loading 10 channels at a time */
		for (i = (x *5); i < (x * 5) + 5; i++) {
			// FIXME a = __t1_get_reg(wp, 0x31 + i);
			/* Get high channel in low bits */
			rxs = (a & 0xf);
			if (!(wp->chans[i+15].sig & ZT_SIG_CLEAR)) {
				if (wp->chans[i+15].rxsig != rxs)
					zt_rbsbits(&wp->chans[i+15], rxs);
			}
			rxs = (a >> 4) & 0xf;
			if (!(wp->chans[i].sig & ZT_SIG_CLEAR)) {
				if (wp->chans[i].rxsig != rxs)
					zt_rbsbits(&wp->chans[i], rxs);
			}
		}
#endif
	} else {
		int i = 0;
		if (status & BIT_SIGX_SIGDATA_A){
			rxs |= ZT_ABIT;
		}
		if (status & BIT_SIGX_SIGDATA_B){
			rxs |= ZT_BBIT;
		}
		if (status & BIT_SIGX_SIGDATA_C){
			rxs |= ZT_CBIT;
		}
		if (status & BIT_SIGX_SIGDATA_D){
			rxs |= ZT_DBIT;
		}
		for(i=0; i < wp->span.channels;i++){
			if (wp->chans[i].channo == channel)
				break;
		}
		if (i == wp->span.channels){
			return;
		}
		if (test_bit(wp->chans[i].chanpos-1, &wp->timeslot_map)){
			if (wp->chans[i].rxsig != rxs){
				zt_rbsbits(&wp->chans[i], rxs);
				DEBUG_VOIP("%s: T1 Report Channel %d: %s A-bit, %s B-bit!\n",
						wp->devname, 
						wp->chans[i].channo,
						(rxs & ZT_ABIT) ? "Set" : "Clear",
						(rxs & ZT_BBIT) ? "Set" : "Clear");
			}
		}
	}
}

/******************************************************************************
** wp_voip_report_alarms() - 
**
**	DONE
*/
void wp_voip_report_alarms(void* pcard, unsigned long te_alarm)
{
	sdla_t*	card = (sdla_t*)pcard;
	wan_voip_t*	wan_voip = &card->wan_voip;
	wp_voip_pvt_t*	wp = NULL;
	int		alarms;
	int		x,j;

	WAN_ASSERT1(wan_voip->voip == NULL);
	wp = wan_voip->voip;
	if (!(wp->flags & WP_FLAG_RUNNING)) return;
	/* Assume no alarms */
	alarms = 0;

//	DEBUG_EVENT("%s(VoIP): wp_voip_report_alarms!\n", wp->devname);
	/* And consider only carrier alarms */
	wp->span.alarms &= (ZT_ALARM_RED | ZT_ALARM_BLUE | ZT_ALARM_NOTOPEN);

	if (wp->ise1){
		/* XXX Implement me XXX */
		FIXME_MSG("wp_voip_report_alarms");
	}else{
		/* Detect loopup code if we're not sending one */
		if (!wp->span.mainttimer && (te_alarm & BIT_LOOPUP_CODE)){
			/* Loop-up code detected */
			if ((wp->loopupcnt++ > 80)  && (wp->span.maintstat != ZT_MAINT_REMOTELOOP)){
				sdla_set_te1_lb_modes(wp->card, WAN_TE1_DDLB_MODE, WAN_TE1_DEACTIVATE_LB);
				sdla_set_te1_lb_modes(wp->card, WAN_TE1_LINELB_MODE, WAN_TE1_ACTIVATE_LB);
				wp->span.maintstat = ZT_MAINT_REMOTELOOP;
			}
		}else{
			wp->loopupcnt = 0;
		}
		/* Same for loopdown code */
		if (!wp->span.mainttimer && (te_alarm & BIT_LOOPDOWN_CODE)){
			/* Loop-down code detected */
			if ((wp->loopdowncnt++ > 80)  && (wp->span.maintstat == ZT_MAINT_REMOTELOOP)){
				sdla_set_te1_lb_modes(wp->card, WAN_TE1_DDLB_MODE, WAN_TE1_DEACTIVATE_LB);
				sdla_set_te1_lb_modes(wp->card, WAN_TE1_LINELB_MODE, WAN_TE1_DEACTIVATE_LB);
				wp->span.maintstat = ZT_MAINT_NONE;
			}
		}else{
			wp->loopdowncnt = 0;
		}
	}

	if (wp->span.lineconfig & ZT_CONFIG_NOTOPEN) {
		for (x=0,j=0;x < wp->span.channels;x++){
			if ((wp->chans[x].flags & ZT_FLAG_OPEN) ||
			    (wp->chans[x].flags & ZT_FLAG_NETDEV)){
				j++;
			}
		}
		if (!j){
			alarms |= ZT_ALARM_NOTOPEN;
		}
	}

	if (wp->ise1) {
		if (te_alarm & BIT_RED_ALARM) 
			alarms |= ZT_ALARM_RED;
		if (te_alarm & BIT_AIS_ALARM)
			alarms |= ZT_ALARM_BLUE;
	} else {
		/* Check actual alarm status */
		if (te_alarm & BIT_RED_ALARM) 
			alarms |= ZT_ALARM_RED;
		if (te_alarm & BIT_AIS_ALARM)
			alarms |= ZT_ALARM_BLUE;
	}
	/* Keep track of recovering */
	if ((!alarms) && wp->span.alarms)
		wp->alarmtimer = ZT_ALARMSETTLE_TIME;

	/* If receiving alarms, go into Yellow alarm state */
	if (alarms && (!wp->span.alarms)) {
		DEBUG_VOIP("%s: Going into yellow alarm\n",
				wp->devname);
		if (card->fe_iface.set_fe_alarm){
			card->fe_iface.set_fe_alarm(card, BIT_YEL_ALARM);
		}
	}

	if (wp->span.alarms != alarms) {
		// FIXME: Alarm status changed!!!
		// DEBUG_EVENT("%s: Alarm status changed!\n", wp->devname);
	}
	if (wp->alarmtimer)
		alarms |= ZT_ALARM_RECOVER;
	if (te_alarm & BIT_YEL_ALARM)
		alarms |= ZT_ALARM_YELLOW;

	wp->span.alarms = alarms;

	zt_alarm_notify(&wp->span);
}

static int wp_voip_transmitprep(void* voip, unsigned char* txbuf, int max_len)
{
	wp_voip_pvt_t*	wp = (wp_voip_pvt_t*)voip;
	int		x = 0, y = 0, offset = 0;
	unsigned char 	value;

	WAN_ASSERT2(voip == NULL, 0);
	zt_transmit(&wp->span);
	for (y=0;y<ZT_CHUNKSIZE;y++) {
		for (x=0;x<wp->span.channels;x++){
			if (wp->flags & WP_FLAG_DATA_PER_INTERFACE){
				if (!test_bit(x,&wp->timeslot_map)){
					continue;
				}
			}
			if (offset >= max_len) break;
			value = wp->chans[x].writechunk[y];
			txbuf[offset++] = brt[value];
		}
	}
	return (offset+1);
}

/******************************************************************************
** wp_voip_receiveprep() - 
**
**	OK
*/
static int wp_voip_receiveprep(void* voip, unsigned char* rxbuf, int max_len)
{
	wp_voip_pvt_t*	wp = (wp_voip_pvt_t*)voip;
	int		x, y, offset = 0;
	unsigned char	value;
	
	WAN_ASSERT2(voip == NULL, 0);
	for (y=0;y<ZT_CHUNKSIZE;y++){
		for (x=0;x<wp->span.channels;x++){
			if (offset >= max_len) break;
			if (wp->flags & WP_FLAG_DATA_PER_INTERFACE){
				if (test_bit(x,&wp->timeslot_map)){
					value = rxbuf[offset++];
				}else{
					value = WAN_VOIP_IDLE_FLAG;	//0xFE;
				}
			}else{
				value = rxbuf[offset++];
			}
			wp->chans[x].readchunk[y] = brt[value]; 
		}
	}
	zt_receive(&wp->span);
	return (offset);
}


/******************************************************************************
** wp_voip_rx_tx() - 
**
**	OK
*/
int wp_voip_rx_tx(void* pcard)
{
	sdla_t*	card = (sdla_t*)pcard;
	struct sk_buff*	rx_skb = NULL;
	long		smp_flags;
	wan_voip_t*	wan_voip = &card->wan_voip;
	wp_voip_pvt_t*	wp = NULL;
	char*		data = NULL;
	int 		max_len = wan_voip->max_tx_len;
	int rx_ind = 0, rx_len = 0;

	WAN_ASSERT(wan_voip->voip == NULL);
	wp = wan_voip->voip;
	skb_trim(wan_voip->tx_skb, 0);
	if (!(wp->flags & WP_FLAG_RUNNING)) return 0;
	if (!wan_voip->tx_skb){
		DEBUG_EVENT("%s: No TX SKB packet!\n",
				card->devname);
		return 0;
	}

	spin_lock_irqsave(&wp->tx_rx_lock, smp_flags);
	if (skb_queue_len(&wan_voip->rx_queue)){
		rx_skb	= skb_dequeue(&wan_voip->rx_queue);	
		max_len	= rx_skb->len; 
	}

	skb_trim(wan_voip->tx_skb, 0);
	while(rx_ind < max_len){
		if (rx_skb){
			rx_len = wp_voip_receiveprep(wan_voip->voip, 
						  &rx_skb->data[rx_ind],
						  max_len - rx_ind); 
			if (rx_len == 0){
				/* Should never happened! */
				DEBUG_EVENT("%s:%d - Critical error (rx_len=%d,rx_ind=%d,max_len=%d)\n",
						__FILE__,__LINE__, rx_len, rx_ind, max_len);	
				break;
			}
		}else{
			rx_len = wp->timeslots * ZT_CHUNKSIZE; 
		}
		data = skb_put(wan_voip->tx_skb, rx_len);
		wp_voip_transmitprep(wan_voip->voip, data, rx_len);
		rx_ind += rx_len;
	}
	if (rx_ind < max_len){
		/* Asterisk application still not ready */
		skb_trim(wan_voip->tx_skb, 0);
	}
	if (rx_skb){
		skb_trim(rx_skb, 0);
		skb_queue_tail(&wan_voip->rx_free_queue, rx_skb);
	}
	max_len = rx_ind;
#if 0
	if (wp->flags & WP_FLAG_DATA_PER_INTERFACE){
		int rx_ind = 0, rx_len = 0;

		while(rx_ind < max_len){
			if (rx_skb){
				rx_len = wp_voip_receiveprep(wan_voip->voip, 
							  &rx_skb->data[rx_ind],
							  max_len - rx_ind); 
				if (rx_len == 0){
					/* Should never happened! */
					DEBUG_EVENT("%s:%d - Critical error (rx_len=%d,rx_ind=%d,max_len=%d)\n",
							__FILE__,__LINE__, rx_len, rx_ind, max_len);	
					break;
				}
			}else{
				rx_len = wp->timeslots * ZT_CHUNKSIZE; 
			}
			data = skb_put(wan_voip->tx_skb, rx_len);
			wp_voip_transmitprep(wan_voip->voip, data, rx_len);
			rx_ind += rx_len;
		}
		if (rx_ind < max_len){
			/* Asterisk application still not ready */
			skb_trim(wan_voip->tx_skb, 0);
		}
		if (rx_skb){
			skb_trim(rx_skb, 0);
			skb_queue_tail(&wan_voip->rx_free_queue, rx_skb);
		}
		max_len = rx_ind;
	}else{
		if (rx_skb){
			max_len = wp_voip_receiveprep(wan_voip->voip, 
						  rx_skb->data,
						  max_len); 
			skb_trim(rx_skb,0);
			skb_queue_tail(&wan_voip->rx_free_queue, rx_skb);
		}
		skb_trim(wan_voip->tx_skb, 0);
		data = skb_put(wan_voip->tx_skb, max_len);
		wp_voip_transmitprep(wan_voip->voip, data, max_len);
	}
#endif
	spin_unlock_irqrestore(&wp->tx_rx_lock, smp_flags);
	return max_len;
}



/*
**		STATIC FUNCTION DEFINITIONS
**
*/

static void wp_voip_generate_bit_rev_table(void)
{
	unsigned char util_char;
	unsigned char misc_status_byte;
	int i;

	/* generate the bit-reversing table for all unsigned characters */
	for(util_char = 0;; util_char ++) {
		misc_status_byte = 0;			/* zero the character to be 'built' */
		/* process all 8 bits of the source byte and generate the */
		for(i = 0; i <= 7; i ++) {
		      	/* corresponding 'bit-flipped' character */
			if(util_char & (1 << i)) {
				misc_status_byte |= (1 << (7 - i));
			}
		}
		/* insert the 'bit-flipped' character into the table at the appropriate location */
		brt[util_char] = misc_status_byte; 
		/* exit when all unsigned characters have been processed */
		if(util_char == 0xFF) {
			break;
		}
	}
}

/******************************************************************************
** wp_voip_software_init() - 
**
** OK
*/
static int wp_voip_software_init(wp_voip_pvt_t* wp)
{
	int 	x = 0;

	WAN_ASSERT(wp == NULL);
	wp->num = wp_cards_voip_no++;
	sprintf(wp->span.name, "WPT1/%d", wp->num);
	sprintf(wp->span.desc, "%s card %d", wp->card->devname, wp->num);
	wp->span.spanconfig = wp_voip_spanconfig;
	wp->span.chanconfig = wp_voip_chanconfig;
	wp->span.startup = wp_voip_startup;
	wp->span.shutdown = wp_voip_shutdown;
	wp->span.rbsbits = wp_voip_rbsbits;
	wp->span.maint = wp_voip_maint;
	wp->span.open = wp_voip_open;
	wp->span.close = wp_voip_close;
	if (wp->ise1)
		wp->span.channels = 31;
	else
		wp->span.channels = 24;
	wp->span.chans = wp->chans;
	wp->span.flags = ZT_FLAG_RBS;
	wp->span.linecompat = ZT_CONFIG_AMI | ZT_CONFIG_B8ZS | ZT_CONFIG_D4 | ZT_CONFIG_ESF;
	wp->span.ioctl = wp_voip_ioctl;
	wp->span.pvt = wp;
	if (wp->ise1)
		wp->span.deflaw = ZT_LAW_ALAW;
	else
		wp->span.deflaw = ZT_LAW_MULAW;
	init_waitqueue_head(&wp->span.maintq);
	for (x=0;x<wp->span.channels;x++) {
		sprintf(wp->chans[x].name, "WPT1/%d/%d", wp->num, x + 1);
		if (test_bit(x,&wp->timeslot_map)){
			DEBUG_EVENT("%s: Configuring channel %d for voice!\n", 
					wp->card->devname, x + 1);
			wp->chans[x].sigcap = ZT_SIG_EM | ZT_SIG_CLEAR | 
					      ZT_SIG_FXSLS | ZT_SIG_FXSGS | 
					      ZT_SIG_FXSKS | ZT_SIG_FXOLS | 
					      ZT_SIG_FXOGS | ZT_SIG_FXOKS | ZT_SIG_CAS;
			wp->timeslots++;
		}else{
			wp->chans[x].sigcap = ZT_SIG_NONE;
		}
		wp->chans[x].pvt = wp;
		wp->chans[x].chanpos = x + 1;
	}

	DEBUG_EVENT("%s: Registering WANPIPE with Asterisk Software!\n", 
					wp->devname);
	if (zt_register(&wp->span, 0)) {
		DEBUG_EVENT("%s: Unable to register span with zaptel\n",
					wp->devname);
		return -EINVAL;
	}
	wp_voip_generate_bit_rev_table();

	return 0;
}







/******************************************************************************
** wp_voip_startup() - 
**
**	OK
*/
static int wp_voip_startup(struct zt_span *span)
{
	wp_voip_pvt_t*	wp = NULL;

	WAN_ASSERT2(span == NULL, -ENODEV);
	WAN_ASSERT2(span->pvt == NULL, -ENODEV);
	wp		= span->pvt;
	if (!(span->flags & ZT_FLAG_RUNNING)) {
		/* Only if we're not already going */
		span->flags |= ZT_FLAG_RUNNING;
	}

	return 0;
}

/******************************************************************************
** wp_voip_shutdown() - 
**
**	OK
*/
static int wp_voip_shutdown(struct zt_span *span)
{
	wp_voip_pvt_t*	wp = NULL;
	long		flags;

	WAN_ASSERT2(span == NULL, -ENODEV);
	WAN_ASSERT2(span->pvt == NULL, -ENODEV);
	wp		= span->pvt;
	spin_lock_irqsave(&wp->lock, flags);
	span->flags &= ~ZT_FLAG_RUNNING;
	spin_unlock_irqrestore(&wp->lock, flags);
	return 0;
}

/******************************************************************************
** wp_voip_maint() - 
**
**	OK
*/
static int wp_voip_maint(struct zt_span *span, int cmd)
{
	wp_voip_pvt_t*	wp = span->pvt;
	int		res = 0;
	long		flags;
	
	WAN_ASSERT2(span == NULL, -ENODEV);
	WAN_ASSERT2(span->pvt == NULL, -ENODEV);
	wp		= span->pvt;
	spin_lock_irqsave(&wp->lock, flags);
	if (wp->ise1) {
		// FIXME: Support E1
		FIXME_MSG("wp_voip_maint");
#if 0
		switch(cmd) {
		case ZT_MAINT_NONE:
			DEBUG_EVENT("%s: E1: Set to normal mode (no local/remote loops)\n",
				       wp->card->devname);	
			// FIXME __t1_set_reg(wp,0xa8,0); /* no loops */
			break;
		case ZT_MAINT_LOCALLOOP:
			DEBUG_EVENT("%s: E1: Set to local loopback mode\n",
				       wp->card->devname);	
			// FIXME __t1_set_reg(wp,0xa8,0x40); /* local loop */
			break;
		case ZT_MAINT_REMOTELOOP:
			DEBUG_EVENT("%s: E1: Set to remote loopback mode\n",
				       wp->card->devname);	
			// FIXME __t1_set_reg(wp,0xa8,0x80); /* remote loop */
			break;
		case ZT_MAINT_LOOPUP:
		case ZT_MAINT_LOOPDOWN:
		case ZT_MAINT_LOOPSTOP:
			res = -ENOSYS;
			break;
		default:
			DEBUG_EVENT("%s: E1: Unknown maintenance mode (%d)\n", 
					wp->card->devname, cmd);
			res = -EINVAL;
			break;
		}
#endif
	}else{
		switch(cmd) {
		case ZT_MAINT_NONE:
			DEBUG_EVENT("%s: T1: Set to normal mode (no local/remote loop)\n",
				       wp->card->devname);	
			sdla_set_te1_lb_modes(wp->card, WAN_TE1_DDLB_MODE, WAN_TE1_DEACTIVATE_LB);
			sdla_set_te1_lb_modes(wp->card, WAN_TE1_LINELB_MODE, WAN_TE1_DEACTIVATE_LB);
			break;
	    	case ZT_MAINT_LOCALLOOP:
			DEBUG_EVENT("%s: T1: Set to local loopback mode (local/no remote loop)\n",
				       wp->card->devname);	
			sdla_set_te1_lb_modes(wp->card, WAN_TE1_LINELB_MODE, WAN_TE1_DEACTIVATE_LB);
			sdla_set_te1_lb_modes(wp->card, WAN_TE1_DDLB_MODE, WAN_TE1_ACTIVATE_LB);
			break;
	    	case ZT_MAINT_REMOTELOOP:
			DEBUG_EVENT("%s: T1: Set to remote loopback mode (no local/remote loop)\n",
				       wp->card->devname);	
			sdla_set_te1_lb_modes(wp->card, WAN_TE1_LINELB_MODE, WAN_TE1_ACTIVATE_LB);
			sdla_set_te1_lb_modes(wp->card, WAN_TE1_LINELB_MODE, WAN_TE1_DEACTIVATE_LB);
			break;
	    	case ZT_MAINT_LOOPUP:
			DEBUG_EVENT("%s: T1: Send loopup code\n",
				       wp->card->devname);	
			sdla_set_te1_lb_modes(wp->card, WAN_TE1_TX_LB_MODE, WAN_TE1_ACTIVATE_LB);
			break;
	    	case ZT_MAINT_LOOPDOWN:
			DEBUG_EVENT("%s: T1: Send loopdown code\n",
				       wp->card->devname);	
			sdla_set_te1_lb_modes(wp->card, WAN_TE1_TX_LB_MODE, WAN_TE1_DEACTIVATE_LB);
			break;
	    	case ZT_MAINT_LOOPSTOP:
			DEBUG_EVENT("%s: T1: Stop sending loop code\n",
				       wp->card->devname);	
			// FIXME __t1_set_reg(wp,0x30,0);	/* stop sending loopup code */
			break;
	    	default:
			DEBUG_EVENT("%s: T1: Unknown maintenance mode (%d)\n", 
					wp->card->devname, cmd);
			res = -EINVAL;
	   	}
	}
	spin_unlock_irqrestore(&wp->lock, flags);
	return res;
}

static void wp_voip_set_clear(wp_voip_pvt_t* wp)
{
	WAN_ASSERT1(wp == NULL);
	/* No such thing under E1 */
	if (wp->ise1) {
		DEBUG_EVENT("Can't set clear mode on an E1!\n");
		return;
	}

	// FIXME: Add action for this function 
}

/******************************************************************************
** wp_voip_chanconfig() - 
**
**	OK	
*/
static int wp_voip_chanconfig(struct zt_chan *chan, int sigtype)
{
	sdla_t*		card;
	wp_voip_pvt_t*	wp = NULL;

	WAN_ASSERT2(chan == NULL, -ENODEV);
	WAN_ASSERT2(chan->pvt == NULL, -ENODEV);
	wp	= chan->pvt;
	card	= (sdla_t*)wp->card;
	if (wp->ise1){
		// No action for E1
		return 0;
	}
	if (chan->span->flags & ZT_FLAG_RUNNING){
		wp_voip_set_clear(wp);
	}
	if (!(sigtype & ZT_SIG_HDLCFCS)){
		/* Initialize ABCD bits */
		DEBUG_EVENT("%s: Initialize signalling bits for Channel %d!\n",
			       	card->devname, chan->channo);
		sdla_te_set_RBS(wp->card, chan->channo, 0x00/*BIT_SIGX_SIGDATA_A | BIT_SIGX_SIGDATA_C*/);
	}
	return 0;
}

/******************************************************************************
** wp_voip_spanconfig() - 
**
**	OK
*/
static int wp_voip_spanconfig(struct zt_span *span, struct zt_lineconfig *lc)
{
	wp_voip_pvt_t*	wp = NULL;
	sdla_t*		card = NULL;
	int		err = 0;

	WAN_ASSERT2(span == NULL, -ENODEV);
	WAN_ASSERT2(span->pvt == NULL, -ENODEV);
	wp		= span->pvt;
	card = (sdla_t*)wp->card;
	switch(wp->lcode){
	case WANOPT_LC_AMI:
		span->lineconfig |= ZT_CONFIG_AMI;
		break;			
	case WANOPT_LC_B8ZS:
		span->lineconfig |= ZT_CONFIG_B8ZS;
		break;			
	case WANOPT_LC_HDB3:
		span->lineconfig |= ZT_CONFIG_HDB3;
		break;	
	}		
	switch(wp->frame){
	case WANOPT_FR_ESF:
		span->lineconfig |= ZT_CONFIG_ESF;
		break;			
	case WANOPT_FR_D4:
		span->lineconfig |= ZT_CONFIG_D4;
		break;			
	case WANOPT_FR_CRC4:
		span->lineconfig |= ZT_CONFIG_CRC4;
		break;	
	}		
	switch(wp->lbo){
	case WANOPT_T1_LBO_0_DB:  
        	DEBUG_TE1("%s: LBO 0 dB\n", card->devname);
		span->lineconfig = 0;
		break;
	case WANOPT_T1_LBO_75_DB:
        	DEBUG_TE1("%s: LBO 7.5 dB\n", card->devname);
		span->lineconfig = 5;
		break;
	case WANOPT_T1_LBO_15_DB: 
        	DEBUG_TE1("%s: LBO 15 dB\n", card->devname);
		span->lineconfig = 6;
		break;
 	case WANOPT_T1_LBO_225_DB:
        	DEBUG_TE1("%s: LBO 22.5 dB\n", card->devname);
		span->lineconfig = 7;
		break;
	case WANOPT_T1_0_110:    
        	DEBUG_TE1("%s: LBO 0-110 ft.\n", card->devname);
		span->lineconfig = 0;
		break;
 	case WANOPT_T1_110_220:   
        	DEBUG_TE1("%s: LBO 110-220 ft.\n", card->devname);
		span->lineconfig = 1;
		break;
 	case WANOPT_T1_220_330:   
        	DEBUG_TE1("%s: LBO 220-330 ft.\n", card->devname);
		span->lineconfig = 2;
		break;
 	case WANOPT_T1_330_440:   
        	DEBUG_TE1("%s: LBO 330-440 ft.\n", card->devname);
		span->lineconfig = 3;
		break;
 	case WANOPT_T1_440_550:   
        	DEBUG_TE1("%s: LBO 440-550 ft.\n", card->devname);
		span->lineconfig = 4;
		break;
 	case WANOPT_T1_550_660:  
        	DEBUG_TE1("%s: LBO 550-660 ft.\n", card->devname);
		span->lineconfig = 5;
		break;
	}

	span->rxlevel = 0;
	/* Do we want to SYNC on receive or not */
	wp->sync = lc->sync;
	/* If already running, apply changes immediately */
	if (span->flags & ZT_FLAG_RUNNING){
		err = wp_voip_startup(span);
	}

	return err;
}


/******************************************************************************
** wp_voip_rbsbits() - Set A,B bits according VoIP requests. 
**
**	DONE
*/
static int wp_voip_rbsbits(struct zt_chan *chan, int bits)
{
	wp_voip_pvt_t*	wp = NULL;
	sdla_t*		card = NULL;
	
	/* Byte offset */
	WAN_ASSERT2(chan == NULL, 0);
	if ((wp = chan->pvt) == NULL) return 0;
	WAN_ASSERT2(wp->card == NULL, 0);
	card = (sdla_t*)wp->card;
	if (!test_bit(chan->chanpos-1, &wp->timeslot_map)){
		return 0;	
	}
	if (!(wp->span.flags & ZT_FLAG_RUNNING)){
		return -EINVAL;
	}
	if (wp->ise1){
		// FIXME: Support E1
		FIXME_MSG("wp_voip_rbsbits");
#if 0
		if (chan->chanpos > 15){
			mask = (bits | (wp->chans[chan->chanpos - 15].txsig << 4));
			DEBUG_EVENT("(wp_voip_rbsbits): E1 Timeslot %d mask=%02x\n", 
						chan->chanpos, mask);
			// FIXME __t1_set_reg(wp, 0x41 + chan->chanpos - 15, mask);
		}else if (chan->chanpos < 15){
			mask = ((bits << 4) | wp->chans[chan->chanpos + 15].txsig);
			DEBUG_EVENT("(wp_voip_rbsbits): E1 Timeslot %d mask=%02x\n",
				       		chan->chanpos, mask);
			//FIME __t1_set_reg(wp, 0x41 + chan->chanpos, mask);
		}
#endif
	}else{
		unsigned char ABCD_bits = 0x00;

		if (chan->flags & ZT_FLAG_HDLC){
			return 0;
		}
		if (bits & ZT_ABIT){
			/* Set A-bit */
			ABCD_bits |= BIT_SIGX_SIGDATA_A;
		}
		if (bits & ZT_BBIT){
			/* Set B-bit */
			ABCD_bits |= BIT_SIGX_SIGDATA_B;
		}
		if (bits & ZT_CBIT){
			/* Set C-bit */
			ABCD_bits |= BIT_SIGX_SIGDATA_C;
		}
		if (bits & ZT_DBIT){
			/* Set D-bit */
			ABCD_bits |= BIT_SIGX_SIGDATA_D;
		}
		DEBUG_VOIP("%s: T1 Channel %d: %s A-bit, %s B-bit\n", 
				wp->devname, 
				chan->channo,
				(ABCD_bits & BIT_SIGX_SIGDATA_A) ? "Set" : "Clear",
				(ABCD_bits & BIT_SIGX_SIGDATA_B) ? "Set" : "Clear");
		// FIXME: Do we need lock here?
		/* Output new values */
		sdla_te_set_RBS(wp->card, chan->channo, ABCD_bits);
	}

	return 0;
}


/******************************************************************************
** wp_voip_ioctl() - 
**
**	OK
*/
static int wp_voip_ioctl(struct zt_chan *chan, unsigned int cmd, unsigned long data)
{
	switch(cmd) {
	default:
		return -ENOTTY;
	}
}

/******************************************************************************
** wp_voip_open() - 
**
**	OK
*/
static int wp_voip_open(struct zt_chan *chan)
{
	wp_voip_pvt_t*	wp = NULL;
	
	WAN_ASSERT2(chan == NULL, -ENODEV);
	WAN_ASSERT2(chan->pvt == NULL, -ENODEV);
	wp = chan->pvt;
	wp->usecount++;
	DEBUG_VOIP("%s: Open (usecount=%d, channo=%d)...\n", 
				wp->devname, wp->usecount, chan->channo);
	return 0;
}


/******************************************************************************
** wp_voip_close() - 
**
**	OK
*/
static int wp_voip_close(struct zt_chan *chan)
{
	wp_voip_pvt_t*	wp = NULL;
	
	WAN_ASSERT2(chan == NULL, -ENODEV);
	WAN_ASSERT2(chan->pvt == NULL, -ENODEV);
	wp = chan->pvt;
	wp->usecount--;
	DEBUG_VOIP("%s: Close (usecount=%d, channo=%d)...\n", 
				wp->devname, wp->usecount, chan->channo);
	/* If we're dead, release us now */
	if (!wp->usecount && wp->dead){ 
		wp_voip_release(wp);
	}
	return 0;
}

/******************************************************************************
** wp_voip_release() - 
**
**	OK
*/
static void wp_voip_release(wp_voip_pvt_t *wp)
{
	WAN_ASSERT1(wp == NULL);
	DEBUG_VOIP("%s: Unregister from Asterisk software!\n",
			wp->devname);
	zt_unregister(&wp->span);
	kfree(wp);
}

static inline void start_alarm(wp_voip_pvt_t* wp)
{
	WAN_ASSERT1(wp == NULL);
#ifdef FANCY_ALARM
	wp->alarmpos = 0;
#endif
	wp->blinktimer = 0;
}

static inline void stop_alarm(wp_voip_pvt_t* wp)
{
	WAN_ASSERT1(wp == NULL);
#ifdef FANCY_ALARM
	wp->alarmpos = 0;
#endif
	wp->blinktimer = 0;
}

