/*****************************************************************************
* chdlc_api.c	CHDLC API: Receive Module
*
* Author(s):	Gideon Hack & Nenad Corbic <ncorbic@sangoma.com>
*
* Copyright:	(c) 1995-2001 Sangoma Technologies Inc.
*
*		This program is free software; you can redistribute it and/or
*		modify it under the terms of the GNU General Public License
*		as published by the Free Software Foundation; either version
*		2 of the License, or (at your option) any later version.
* ============================================================================
* Description:
*
* 	The chdlc_api.c utility will bind to a socket to a chdlc network
* 	interface, and continously tx and rx packets to an from the sockets.
*
*	This example has been written for a single interface in mind, 
*	where the same process handles tx and rx data.
*
*	A real world example, should use different processes to handle
*	tx and rx spearately.  
*/



#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <linux/if_wanpipe.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <linux/wanpipe.h>
#include <linux/sdla_chdlc.h>

#define FALSE	0
#define TRUE	1

#define LGTH_CRC_BYTES	2
#define MAX_TX_DATA     16000	/* Max Size of tx data */  
#define MAX_RX_DATA     16000	/* Max Size of rx data */  

#define TX_PKT_LEN	1400
#define MAX_FRAMES 	1000000     /* Number of frames to transmit */  

#define TX_START_PATTERN	0x00
#define TX_STOP_PATTERN		0x1F
//#define TX_START_PATTERN	0x20
//#define TX_STOP_PATTERN	0x3F

#define WRITE 1 

/*===================================================
 * Golobal data
 *==================================================*/

unsigned short no_CRC_bytes_Rx;
unsigned char HDLC_streaming = FALSE;
unsigned short Rx_lgth;

unsigned char Rx_data[MAX_RX_DATA+sizeof(api_rx_hdr_t)];
unsigned char Tx_data[MAX_TX_DATA+sizeof(api_tx_hdr_t)];
int 	sock;
char file_name[50];

/*=================================================== 
 * Function Prototypes 
 *==================================================*/
int MakeConnection(char *, char *);
void process_con( void);
void print_packet(char* rx_data, int rx_lgth);


/*===================================================
 * MakeConnection
 *
 *   o Create a Socket
 *   o Bind a socket to a wanpipe network interface
 *       (Interface name is supplied by the user)
 *==================================================*/         

int MakeConnection(char *r_name, char *i_name ) 
{
	struct wan_sockaddr_ll 	sa;

	memset(&sa,0,sizeof(struct wan_sockaddr_ll));
	errno = 0;
   	sock = socket(AF_WANPIPE, SOCK_RAW, 0);
   	if( sock < 0 ) {
      		perror("Socket");
      		return( FALSE );
   	} /* if */
  
	printf("\nConnecting to router %s, interface %s\n", r_name, i_name);
	
	strcpy( sa.sll_device, i_name);
	strcpy( sa.sll_card, r_name);
	sa.sll_protocol = htons(PVC_PROT);
	sa.sll_family=AF_WANPIPE;

        if(bind(sock, (struct sockaddr *)&sa, sizeof(struct wan_sockaddr_ll)) < 0){
                perror("bind");
		printf("Failed to bind a socket to %s interface\n",i_name);
                exit(0);
        }
	printf("Socket bound to %s\n\n",i_name);

	return( TRUE );

}

/*===========================================================
 * process_con 
 *
 *   o Tx/Rx data to and from the socket 
 *   o Cast received data to an api_rx_element_t data type 
 *   o The received packet contains 16 bytes header 
 *
 *	------------------------------------------
 *      |  16 bytes      |        X bytes        ...
 *	------------------------------------------
 * 	   Header              Data
 *
 * RX DATA:
 * --------
 * 	Each rx data packet contains the 16 byte header!
 * 	
 *   	o Rx 16 byte data structure:
 *
 *		typedef struct {
 *       	 	unsigned char   error_flag      PACKED;
 *        		unsigned short  time_stamp      PACKED;
 *        		unsigned char   reserved[13]    PACKED;
 *		} api_rx_hdr_t; 
 *
 *		typedef struct {
 *        		api_rx_hdr_t    api_rx_hdr      PACKED;
 *        		void *          data            PACKED;
 *		} api_rx_element_t;
 *
 * 	error_flag:
 * 		bit 0: 	incoming frame was aborted
 * 		bit 1: 	incoming frame has a CRC error
 * 		bit 2:  incoming frame has an overrun eror 
 *
 * 	time_stamp:
 * 		absolute time value in ms.
 *
 * TX_DATA:
 * --------
 *	Each tx data packet MUST contain a 16 byte header!
 *
 * 	o Tx 16 byte data structure
 * 	
 *		typedef struct {
 *			unsigned char 	attr		PACKED;
 *			unsigned char  	reserved[15]	PACKED;
 *		} api_tx_hdr_t;
 *
 *		typedef struct {
 *			api_tx_hdr_t 	api_tx_hdr	PACKED;
 *			void *		data		PACKED;
 *		} api_tx_element_t;
 *
 *	Currently the chdlc device driver doesn't use any of
 *	the above fields.  Thus, the user can set the 16 bytes
 *	to ZERO.
 * 
 */

void process_con(void) 
{
	unsigned int Rx_count,Tx_count,Tx_length;
	api_rx_element_t* api_rx_el;
	api_tx_element_t * api_tx_el;
	fd_set 	ready,write;
	int err,i;
	unsigned char *data, *rx_data;
	unsigned char	tx_req_pattern = TX_START_PATTERN;
	char	pattern = TX_START_PATTERN;
	struct timeval	tx_req_time, curr_time;
	struct timezone	tx_req_tz;
	int tx_reply = 0, Tx_failed = 0;

        Rx_count = 0;
	Tx_count = 0;
	Tx_length = TX_PKT_LEN;
	//Tx_length = MAX_TX_DATA;
	
	/* If running HDLC_STREAMING then the received CRC bytes
         * will be passed to the application as part of the 
         * received data.  The CRC bytes will be appended as the 
	 * last two bytes in the rx packet.
	 */

	memset(&Tx_data[0],0,MAX_TX_DATA + sizeof(api_tx_hdr_t));


	/* Initialize the tx packet. The 16 byte header must
	 * be inserted before tx data.  In CHDLC protocol,
	 * the tx x25 header is not used, thus it can be
	 * set to zero */
	api_tx_el = (api_tx_element_t*)&Tx_data[0];
	data = (unsigned char *)&api_tx_el->data;

	for (i=0;i<MAX_TX_DATA;i++){
		data[i] = tx_req_pattern;
	}
	tx_reply = 0;
	timerclear(&tx_req_time);

	/* If drivers has been configured in HDLC_STREAMING
	 * mode, the CRC bytes will be included into the 
	 * rx packet.  Thus, the application should remove
	 * the last two bytes of the frame. 
	 * 
	 * The no_CRC_bytes_Rx will indicate to the application
	 * how many bytes to cut from the end of the rx frame.
	 *
	 */
	no_CRC_bytes_Rx = 0; //LGTH_CRC_BYTES;

	for(;;) {	
		FD_ZERO(&ready);
		FD_ZERO(&write);
		if (Tx_failed == 0){
			FD_SET(sock,&ready);
		}
#ifdef WRITE
		FD_SET(sock,&write);
#endif

		/* The select function must be used to implement flow control.
	         * WANPIPE socket will block the user if the socket cannot send
		 * or there is nothing to receive.
		 *
		 * By using the last socket file descriptor +1 select will wait
		 * for all active sockets.
		 */ 

  		if(select(sock + 1,&ready, &write, NULL, NULL)){

			/* Check for rx packets */
                   	if (FD_ISSET(sock,&ready)){

				err = recv(sock, Rx_data, sizeof(Rx_data), 0);

				/* err indicates bytes received */
				if(!(err > 0)) {
					printf("\nError receiving data\n");
					break;
				}

				/* Rx packet recevied OK
				 * Each rx packet will contain 16 
				 * bytes of rx header, that must be 
				 * removed. The first byte of the 16
				 * byte header will indicate an error
				 * condition.
				 */
				api_rx_el = (api_rx_element_t*)Rx_data;

				switch (api_rx_el->api_rx_hdr.error_flag){
			
				case 0:
					/* Rx packet is good */
					break;
					
				case RX_FRM_ABORT:
					/* Frame was aborted */
					break;
				case RX_FRM_CRC_ERROR:
					/* Frame has crc error */
					break;
				case RX_FRM_OVERRUN_ERROR:
					/* Frame has an overrun error */
					break;
				default:
					/* Error with the rx packet 
					 * handle it ... */
					break;
				}
				
					
				/* Check the packet length, after 
				 * we remove the 16 bytes rx header */
				Rx_lgth = err - sizeof(api_rx_hdr_t);
				if(Rx_lgth <= no_CRC_bytes_Rx) {
					printf("\nShort frame received (%d)\n",
						Rx_lgth);
					return;
				}

				/* Remove any CRC bytes from the rx 
				 * packet by adjusting the rx packet
				 * length */
				Rx_lgth -= no_CRC_bytes_Rx;
				rx_data = (unsigned char *)&api_rx_el->data;
				pattern = (char)rx_data[0];
				if (pattern >= TX_START_PATTERN && 
				    pattern <= TX_STOP_PATTERN){
					for (i = 0; i < Rx_lgth; i++){
						if (rx_data[i] != tx_req_pattern){
							print_packet(rx_data,
								     Rx_lgth);
							printf("RX PKT corrupted: offset=%d, pattern=0x%02X (expected=0x%02X)!\n",
								i, rx_data[i],
								tx_req_pattern);
							return;
						}
					}
					printf(".");
					//printf("RX PKT %i, len=%i, pat=%02X\n",
					//	++Tx_count, Rx_lgth, pattern);
					if (tx_req_pattern == TX_STOP_PATTERN){
						tx_req_pattern = TX_START_PATTERN;
					}else{
						tx_req_pattern++;
					}
					timerclear(&tx_req_time);
					for (i=0;i<TX_PKT_LEN;i++){
						data[i] = tx_req_pattern;
					}
					//if (Tx_count < MAX_FRAMES){
						Tx_length = TX_PKT_LEN;
					//}
				}else{
					//printf("\t\tRX REQ PKT: pat=%02X\n",
					//		pattern);
					Tx_length = Rx_lgth+sizeof(api_tx_hdr_t);
					for (i=0;i<Tx_length;i++){
						Tx_data[i] = pattern;
					}
					tx_reply = 1;
				}
			 } 
		  
			/* Check if the socket is ready to tx data */
		   	if (Tx_length && FD_ISSET(sock,&write)){

				err = send(sock,Tx_data, Tx_length, 0);

				if (err > 0){
					/* Packet sent ok */
					if (!tx_reply){
						//printf("TX PKT REQ: %i\n",
						//	err);
						gettimeofday(&tx_req_time, 
								&tx_req_tz);
					}else{
						//printf("\t\tTX PKT REPLY: %i\n",
						//	err);
					}
					Tx_length = 0;
					tx_reply = 0;
					Tx_failed = 0;
				
				}else{
					/* The driver is busy, we must 
					 * requeue this tx packet, and 
					 * try resending it again later. 
					 */
					printf("Failed to send: pat=0x%02X!\n",
							pattern);
					Tx_failed = 1;
				}
		   	}

			//sleep(1);
		   	/* Stop the program afer transmitting MAX_FRAMES */
			//if (Tx_count == MAX_FRAMES)
			//	break;
			//if (Tx_length == MAX_TX_DATA){
			//	Tx_length = 1000;
			//}else{
			//	Tx_length++;
			//}
			gettimeofday(&curr_time, &tx_req_tz);
			if (timerisset(&tx_req_time)){
				if (curr_time.tv_sec > tx_req_time.tv_sec + 5){
					if (Tx_count == 0){
						Tx_length = TX_PKT_LEN;
						for (i=0;i<TX_PKT_LEN;i++){
							data[i] = tx_req_pattern;
						}
						continue;
					}
					printf("Timeout on TX PKT %d\n",
							Tx_count+1);
					timerclear(&tx_req_time);
				}
			}
		}else{
                	printf("\nError selecting socket\n");
			break;
		}
	}
} 

void print_packet(char* rx_data, int rx_lgth)
{
	int i = 0;
	int base = rx_lgth >> 4;
	FILE* fd = NULL;

	fd = fopen(file_name, "w");
	if (fd == NULL){
		printf("Failed to open output file %s\n", file_name);
		return;
	}
	fprintf(fd, "\n");
	for (i=0; i < base;i++){
		fprintf(fd, "%02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X\n",
				rx_data[i*16],    rx_data[i*16+1],
				rx_data[i*16+2],  rx_data[i*16+3],
				rx_data[i*16+4],  rx_data[i*16+5],
				rx_data[i*16+6],  rx_data[i*16+7],
				rx_data[i*16+8],  rx_data[i*16+9],
				rx_data[i*16+10], rx_data[i*16+11],
				rx_data[i*16+12], rx_data[i*16+13],
				rx_data[i*16+14], rx_data[i*16+15]);
	}
	i *= 16;	
	while(i < rx_lgth){
		fprintf(fd, "%02X ", rx_data[i++]);
	}
	fprintf(fd,"\n");
	fclose(fd);
}

/***************************************************************
 * Main:
 *
 *    o Make a socket connection to the driver.
 *    o Call process_con() to read/write the socket 
 *
 **************************************************************/


int main(int argc, char* argv[])
{
	int proceed;

	if (argc != 3){
		printf("Usage: rec_wan_sock <router name> <interface name>\n");
		exit(0);
	}	

	proceed = MakeConnection(argv[argc - 2], argv[argc - 1]);
	if( proceed == TRUE ){
		sprintf(file_name, "%s_api.log", argv[argc-2]);
		process_con( );
		close(sock);
		return 0;
	}

	return 1;
};
