/*================================================================================
 * wpbackupd.cc	: WANPIPE BACKUPD DAEMON
 *
 * Author: Nenad Coribc <ncorbic@sangoma.com>
 * 
 * Copyright:	(c) 2000-2001 Sangoma Technologies Inc.
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 * ----------------------------------------------------------------------------
 *
 * Mar 25, 2001 Nenad Corbic	Re-wrote the backupd program using C++
 *                              Added support for:
 *                              	- user scripts
 *                                      - multiple device monitoring
 *                                      - non WANPIPE device monitoring
 * Jan 11, 2001	Nenad Corbic	Initial version
 *================================================================================
*/ 

#include <string>
#include <vector>
#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cstdio>
#include <pfstream.h>

extern "C" {
	#include <stdlib.h>
	#include <errno.h>
	#include <sys/stat.h>                     //open()
	#include <sys/time.h>
	#include <unistd.h>
	#include <wait.h>
	#include <signal.h>
	#include <sys/socket.h>
	#include <arpa/inet.h>
	#include <net/if.h>
	#include <sys/ioctl.h>
	#include <syslog.h>
	#include <fcntl.h>
}

#define UPDATE_INTERVAL 5
#define MAX_PRIMARY_DEVICES 10
#define IP_ADDR_SIZE 10
#define MAX_TOKENS 32
#define CONFIG_FILE "/etc/wanpipe/wpbackupd.conf"
#define ITOSTATE(x) (x ? "CONNECTED":"DISCONNECTED")

#define MAX_LINK_CHECK_FRQ 60
#define MAX_PING_CHECK_FRQ 60
#define MAX_START_DELAY 100

#define LINK_CHK_DELAY -1
#define START_DELAY -2

#ifdef _DEBUG_
   #define DEBUG _DEBUG_
#else
   #define DEBUG 0
#endif

typedef struct options {
	string s;
	int ret;
}options_t;

class Connection;

//**************************************************
//GOBAL DEFINES
//**************************************************
enum state_t{ DISCONNECTED, CONNECTED, UNCONF };
enum connection_type_t { SINGLE, BALANCED, BOD };
enum {NO, YES};
enum {OK, ERR}; 
enum {ERROR, END, PRIMARY, SECONDARY, WANPIPE, SYNCPPP, DIALUP};

pid_t sec_pid=0;
Connection *pri, *sec;

options_t general_options[] = {
	{"[primary]",PRIMARY},
	{"[secondary]",SECONDARY},
	{"[wanpipe]",WANPIPE},
	{"[syncppp]",SYNCPPP},
	{"[dialup]",DIALUP},
	{"",ERROR}
};

string program_name = "wpbackupd";
string log_file  = "/var/log/wpbackupd";
string prog;
std::ofstream fout;
int session=PRIMARY;
int enable_ping_chk=0;

//**************************************************
//FUNCTION PROTOTYPES
//**************************************************
int get_opt (std::fstream &in_file);
void filter_char(string &s, char ch);
string &prog_name(void);
void usage(void);
void usage_long(void);

class Options {

public:	
	int get_opt(std::fstream &in_file);
	int get_opt(std::fstream &in_file, string *opt, string *val);
	void filter_char(string &s, char ch);
	int verify_exec(string &s);
	int exec_script(string &s);
	int get_ip_addr(string &if_name, string &ip);
	int open_log_file(void);
};



//**************************************************
//OPTIONS CLASS
//
//**************************************************
int Options::open_log_file(void)
{
	fout.open(log_file.c_str(),std::ios::out|std::ios::app);
	if (fout.fail()){
		cout << prog_name() << "Failed to open log file " << log_file.c_str() << "\n";
		return -EINVAL;
	}
	return OK;
}

int Options::get_opt(std::fstream &in_file)
{
	string str;
	int pos;
	int i=0;

	for (;;){
		if (in_file.eof())
			return END;

		if (std::getline(in_file,str)){

			// Erase Comments
			pos = str.find("#");
			if (pos >= 0){
				str.erase(pos,(str.length()-pos));
				if (!str.length())
					continue;
			}

			filter_char(str,' ');
			filter_char(str,'\t');

			if (!str.length())
				continue;

			for (i=0;general_options[i].s.length();i++){ 
				pos = str.find(general_options[i].s);
				if (pos >= 0){
					return general_options[i].ret;
				}
			}
		}
	}
}


int Options::get_opt(std::fstream &in_file, string *opt, string *val)
{
	string str;
	int pos,orig_len;

	for (;;){
		if (in_file.eof())
			return END;

		if (std::getline(in_file,str)){
	
			orig_len = str.length()+1;

			// Erase Comments
			pos = str.find("#");
			if (pos >= 0){
				str.erase(pos,(str.length()-pos));
				if (!str.length())
					continue;
			}

			filter_char(str,' ');
			filter_char(str,'\t');

			if (!str.length())
				continue;

			if ((pos = str.find("=")) >= 0){
				*val = str.substr(pos+1);		
				str.erase(pos,str.length()-pos);
				*opt = str;
				return OK;
				
			}else if ((pos = str.find("[")) >= 0) {
				in_file.seekg(-orig_len, ios::cur);
				return ERR;
			}else{
				*opt = str;
				*val ="";
				return OK;
			}
		}
	}
}

void Options::filter_char (string &str, char ch)
{
	int pos;

	for (;;){
		pos = str.find(ch);
		if (pos >= 0){
			str.erase(pos,1);
		}else{
			break;
		}
	}
}

int Options::verify_exec(string &str)
{
	struct stat sbuf;
	int err=-EINVAL;

	/* Remove any arguments */
	string s = str;
	int pos = s.find(" ");
	if (pos >= 0){
	  	s.erase(pos,(s.length()-pos));
	}

	/* Convert the string into a (char *) */
	char *file = new char[s.length()+1];
	if (!file){
		return -ENOMEM;
	}
	
	s.copy(file,string::npos);
	file[s.length()]=0;


	/* Check that file exists, and that it's execuatble */
	if (stat(file, &sbuf) < 0){
		cout << prog_name() << "Script not found : " << file << "\n";
	}else if (!S_ISREG(sbuf.st_mode)){
		cout << prog_name() << "Script is an invlid file : " << file << "\n";
	}else if ((sbuf.st_mode & (S_IXUSR|S_IXGRP|S_IXOTH)) == 0) {
		cout << prog_name() << "Script is not executable : " << file << "\n";
	}else{
		err=OK;
	}
	
	delete file;
	
	return err;
}


int Options::exec_script(string &s)
{
	int pid;
	int status = -1;

	char *program = new char[s.length()+1];
	if (!program){
		return -ENOMEM;
	}
	
	s.copy(program,string::npos);
	program[s.length()]=0;

	if (verify_exec(s) != OK){
		delete program;
		return -EINVAL;
	}

	pid = fork();

	if (pid < 0) {
		cout << prog_name() << "Failed to create child process: %m \n";
		delete program;
		return -EINVAL;
	}

	if (pid != 0) {
		while (waitpid(pid, &status, 0) < 0) {
			if (errno == EINTR)
				continue;
			cout << prog_name() << "Error waiting for (dis)connection process: %m \n";
		}
		delete program;
		return (status == 0 ? 0 : -1);
	}

	//FIXME Force to root
	uid_t uid=0;	
	setuid(uid);
	if (getuid() != uid) {
		cout << prog_name() << "Setuid failed\n";
		exit(1);
	}

	setgid(getgid());
	execl("/bin/sh", "sh", "-c", program, (char *)0);
	cout << prog_name() << "Error, could not exec /bin/sh: %m \n";
	exit(99);
	/* NOTREACHED */
}

int Options::get_ip_addr(string &if_name,string &ip){

	int skfd;
	struct sockaddr_in *sin;
	struct ifreq ifr;
	int err;

    	if ((skfd = socket(AF_INET, SOCK_STREAM, IPPROTO_IP)) < 0) {
		return -EINVAL;
    	}

	strncpy(ifr.ifr_ifrn.ifrn_name, if_name.c_str(), sizeof(ifr.ifr_ifrn.ifrn_name));

        if ((err=ioctl(skfd, SIOCGIFFLAGS , &ifr)) < 0) {
		return -EINVAL;
    	}

        if (ioctl(skfd, SIOCGIFDSTADDR , &ifr) < 0) {
		return -EINVAL;
	}else{
		sin = (struct sockaddr_in *)&ifr.ifr_ifru.ifru_dstaddr;
		ip = inet_ntoa(sin->sin_addr);
	}

	close (skfd);

	return OK;
}



//**************************************************
//SYNCPPP PROTOCOL CLASS
//**************************************************

class SyncPPP_Prot {

public:
	SyncPPP_Prot () {}
	~SyncPPP_Prot() {}
	
	state_t syncppp_link_chk(string &device);
};

state_t SyncPPP_Prot::syncppp_link_chk(string &device)
{
	int i,pos,found=0;
	
	ipfstream pfd("|cat /proc/net/syncppp", ios::in);
	if (pfd.fail())
		return UNCONF;

	i=0;
	string line;
	while (std::getline(pfd,line)){
		if ((pos = line.find(device)) >= 0){
			found = 1;
			break;
		}
	}

	if (!found){
		return UNCONF;
	}

	if ((pos=line.find("up")) >= 0){
		return CONNECTED;
	}else{
		return DISCONNECTED;
	}
	
	return UNCONF;
}

//**************************************************
//WANPIPE PROTOCOL CLASS
//
//**************************************************

class Wanpipe_Prot {

public:
	Wanpipe_Prot () {}
	~Wanpipe_Prot() {}
	
	state_t wanpipe_link_chk(string &device);
};

state_t Wanpipe_Prot::wanpipe_link_chk(string &device)
{
	char router_dir[] = "/proc/net/wanrouter";	
	struct stat file_stat;
	int pos,i;
	state_t state=UNCONF;

	if (stat(router_dir, &file_stat) < 0){
		cout << prog_name() << "Directory not found : " << router_dir << "\n";
		return state;
	}

	ipfstream pfd("|cat /proc/net/wanrouter/status", ios::in);
	if (pfd.fail()){
		cout << prog_name() << "Directory not found : /proc/net/wanrouter/status \n";
		return state;
	}

	i=0;
	string line;
	while (std::getline(pfd,line)){
		if ((pos=line.find(device)) >= 0){
			if ((pos=line.find("Disconnected")) >= 0){
				state = DISCONNECTED;
			}else if ((pos=line.find("Connected")) >= 0){
				state = CONNECTED;
			}else{
				state = DISCONNECTED;
			}
			break;
		}
	}

	pfd.close();
	
	return state;
}


//**************************************************
//DEVICE CLASS
//
//**************************************************
class Device : public Options, public Wanpipe_Prot, public SyncPPP_Prot{

public:
	//Constructors, Destructors
	Device ();
	~Device();

	//Modifiers
	int up();
	int down();

	int bod_chk(void);
	void print(void);
	
	void set_state(state_t state);
	state_t get_state(void);

	void set_dev_name(string &devname);
	void set_if_name(string &ifname);
	string & get_dev_name(void);
	string & get_if_name(void);

	char* strstrip (char* str, char* s);
	int tokenize (char* str, char **tokens, char *arg1, char *arg2, char *arg3);
	int opt_set(string &opt, string &val);

	int init_link_type(int device_type);
	int link_chk(void);
	int ping_chk(void);

	int link_down(void);
	int link_up(void);

	int start(void);
	int stop(void);

private:
	//Variables
	string dev_name;
	string if_name;
	
	string local_ip;
	string remote_ip;
	string gateway_ip;
	string ping_ip;

	string default_route;
	state_t state;
	state_t prev_state;

	int link_type;
	
	string start_exec;
	string stop_exec;
	string link_down_exec;
	string link_up_exec;
	string link_chk_exec;
	string ping_chk_exec;
};

//==================================================
//Device Class, Constructors, Desctructors and
//              Modifiers
//==================================================

Device::Device()
{
	state = DISCONNECTED;
	prev_state = DISCONNECTED;
	default_route="NO";
	link_type=-1;
}


Device::~Device()
{
}

int Device::init_link_type(int device_type)
{
	switch (device_type){

	case WANPIPE:
	case SYNCPPP:
	case DIALUP:
		link_type = device_type;
		return OK;

	default:
		return ERR;
	}
}

int Device::link_chk(void)
{
	switch (link_type){
		
	case WANPIPE:
		if ((state=wanpipe_link_chk(dev_name)) == UNCONF){
			cout << prog_name() << "Error, Invalid WANPIPE State!\n";
			return -EINVAL;
		}
		break;

	case SYNCPPP:
		if ((state=syncppp_link_chk(if_name)) == UNCONF){
		        cout << prog_name() << "Error, Invalid SYNCPPP State!\n";
			return -EINVAL;
		}
		break;
	
	case DIALUP:
		if (exec_script(link_chk_exec) == OK){
			state=CONNECTED;
		}else{
			state=DISCONNECTED;
		}
		break;
	default:
		cout << prog_name() << "Error, Invalid Link Type!\n";
		return -EINVAL;
	}

	//If the state has changed for this device, run the
	//approprite link state script.
	if (state != prev_state){
		if (state == CONNECTED){
			link_up();
		}else{	
			link_down();
		}	
		prev_state = state;
	}

	if (state == CONNECTED){
		if (!remote_ip.length()){
			if (get_ip_addr(if_name,remote_ip) != OK){
				cout << prog_name() << 
				  "Error failed to obtain remote ip address for : " << if_name << ".\n";
				state=DISCONNECTED;
				return -EINVAL;
			}
		}
	}

	return OK;
}

int Device::ping_chk(void)
{
	if (state != CONNECTED)
		return OK;

	if (!ping_chk_exec.length() ||
	    !ping_ip.length()){
		return OK;
	}

	string ping_cmd;
	if (!enable_ping_chk){

		if (remote_ip.length()){
			ping_cmd = ping_chk_exec+" "+remote_ip;
		}else{
			cout << prog_name() << "Error, Ping check failed, no remote IP\n";
			return -EINVAL;
		}
			
	}else{
		ping_cmd = ping_chk_exec+" "+ping_ip;
	}

	if (exec_script(ping_cmd) == OK){
		if (!enable_ping_chk || DEBUG)
			cout << prog_name() << "Ping Check OK : " << ping_cmd << "\n";
		state=CONNECTED;
	}else{

		if (enable_ping_chk || DEBUG)
			cout << prog_name() << "Ping Check FAILED : " << ping_cmd << "\n";
		state=DISCONNECTED;
	}
	
	return OK; 
}

void Device::print(void)
{
	cout << prog_name() << "Print Device: " << dev_name 	<< " Interface: " << if_name << "\n"; 
}

void Device::set_state(state_t state_tmp)
{
	state = state_tmp;
}

state_t Device::get_state(void)
{
	return state;
}

void Device::set_dev_name(string &device)
{
	dev_name = device;
}

void Device::set_if_name(string &ifname)
{
	if_name = ifname;
}

string &Device::get_dev_name(void)
{
	return dev_name;
}

string &Device::get_if_name(void)
{
	return if_name;
}

//========================================================
// Strip leading and trailing spaces off the string str.
//========================================================
char* Device::strstrip (char* str, char* s)
{
	char* eos = str + strlen(str);		/* -> end of string */

	while (*str && strchr(s, *str))
		++str;				/* strip leading spaces */
	while ((eos > str) && strchr(s, *(eos - 1)))
		--eos;				/* strip trailing spaces */
	*eos = '\0';
	return str;
}


//==================================================================
// Tokenize string.
//	Parse a string of the following syntax:
// 		<tag>=<arg1>,<arg2>,...
// 	and fill array of tokens with pointers to string elements.
// 
// 	Return number of tokens.
//==================================================================

int Device::tokenize (char* str, char **tokens, char *arg1, char *arg2, char *arg3)
{
	int cnt = 0;

	tokens[0] = strtok(str, arg1);
	while (tokens[cnt] && (cnt < MAX_TOKENS - 1)) {
		tokens[cnt] = strstrip(tokens[cnt], arg2);
		tokens[++cnt] = strtok(NULL, arg3);
	}
	return cnt;
}

int Device::opt_set(string &opt, string &val)
{
	int pos;

	if ((pos=opt.find("device")) >=0){
		set_dev_name(val);
		cout << prog_name() << "\tDevice name is " << get_dev_name() << "\n";
		return OK;
	}
	if ((pos=opt.find("interface"))>=0){
		set_if_name(val);
		cout << prog_name() << "\tInterface name is " << get_if_name() << "\n";
		return OK;
	}
	if ((pos=opt.find("default_route"))>=0){
		default_route=val;
		cout << prog_name() << "\tDefault Route =" << default_route << "\n";
		return OK;
	}
	if ((pos=opt.find("ping_ip"))>=0){
		ping_ip=val;
		cout << prog_name() << "\tPing IP =" << ping_ip << "\n";
		return OK;
	}
	if ((pos=opt.find("gateway_ip"))>=0){
		gateway_ip=val;
		cout << prog_name() << "\tGateway IP =" << gateway_ip << "\n";
		return OK;
	}
	if ((pos=opt.find("link_up"))>=0){
		link_up_exec = val;
		cout << prog_name() << "\tLink UP Executable = " << link_up_exec << "\n";
		return verify_exec(link_up_exec);
	}
	if ((pos=opt.find("link_down"))>=0){
		link_down_exec = val;
		cout << prog_name() << "\tLink Down Executable = " << link_down_exec << "\n";
		return verify_exec(link_down_exec);
	}
	if ((pos=opt.find("start_exec"))>=0){
		start_exec = val;
		cout << prog_name() << "\tDevice Startup script = " << start_exec << "\n";
		return verify_exec(start_exec);
	}
	if ((pos=opt.find("stop_exec"))>=0){
		stop_exec = val;
		cout << prog_name() << "\tDevice Stop script = " << stop_exec << "\n";
		return verify_exec(start_exec);
	}
	if ((pos=opt.find("link_chk_exec"))>=0){
		link_chk_exec = val;
		cout << prog_name() << "\tLink check script = " << link_chk_exec << "\n";
		return verify_exec(link_chk_exec);
	}
	if ((pos=opt.find("ping_chk_exec"))>=0){
		ping_chk_exec = val;
		cout << prog_name() << "\tPing check script = " << ping_chk_exec << "\n";
		return verify_exec(ping_chk_exec);
	}

	return ERR;
}

int Device::link_down(void)
{
	if (!link_down_exec.length()){
		return OK;
	}

	cout << prog_name() << "Running Link DOWN Script for " 
	      << dev_name << " : " << if_name << endl;

	return exec_script(link_down_exec);
}

int Device::link_up(void)
{
	if (!link_up_exec.length()){
		return OK;
	}

	cout << prog_name() << "Running Link UP Script for " 
	      << dev_name << " : " << if_name << endl;
	return exec_script(link_up_exec);
}

int Device::start(void)
{
	if (!start_exec.length()){
		return OK;
	}

	cout << prog_name() << "Running Link START Script for "
		<< dev_name << " : " << if_name << endl;
	return exec_script(start_exec);
}
int Device::stop(void)
{
	if (!stop_exec.length()){
		return OK;
	}

	cout << prog_name() << "Running Link STOP Script for"
		<< dev_name << " : " << if_name << endl;
	return exec_script(stop_exec);
}


//**************************************************
//CONNECTION CLASS
//
//**************************************************
class Connection : public Options {

public:
	//Constructor, Destructor
	Connection ();
	~Connection();

	//Modifiers
	int attach_device(Device* dev);
	int update_state(void);
	void print(void);
	state_t get_state(void);
	int link_chk(void);
	int ping_chk(void);
	int opt_set(string &opt, string &val);

	void delay(int val);

	int start(int delay_time);
	int stop(int delay_time);

	int link_down(void);
	int link_up(void);

	//Variables
	vector<Device*>dev_list;
	
	state_t state;
	state_t prev_state;
	connection_type_t type;
	string con_up_exec;
	string con_down_exec;

	unsigned int link_chk_frq;
	unsigned int ping_chk_frq;
	unsigned int ping_cur_delay;

	unsigned int start_delay;
};

Connection::Connection()
{
	link_chk_frq=5;
	ping_chk_frq=5;
	ping_cur_delay=0;
	start_delay=0;
	type=SINGLE;
	state=DISCONNECTED;
	prev_state=DISCONNECTED;
}

Connection::~Connection()
{
	vector<Device*>::iterator p;
	
	for (p = dev_list.begin(); p!=dev_list.end(); ++p){
		delete *p;
	}

	dev_list.clear();
}

int Connection::attach_device(Device* dev_ptr)
{
	dev_list.push_back(dev_ptr);
	return 0;
}

void Connection::print(void)
{
	vector<Device*>::iterator p;
	
	for (p = dev_list.begin(); p!=dev_list.end(); ++p)
		(*p)->print();
}

state_t Connection::get_state(void)
{
	vector<Device*>::iterator p;
	state = DISCONNECTED;

	for (p = dev_list.begin(); p!=dev_list.end(); ++p){
		if ((*p)->get_state() == CONNECTED){
			state=CONNECTED;
		}
	}

	return state;
}

int Connection::link_chk(void)
{
	vector<Device*>::iterator p;

	for (p = dev_list.begin(); p!=dev_list.end(); ++p){

		#if _DEBUG_ > 1
		cout << prog_name() << "Checking Link State for : " << 
					(*p)->get_dev_name() << " : " << 
					(*p)->get_if_name() << endl;
		#endif
		
		if ((*p)->link_chk() < 0){
			return -EINVAL;
		}
		if ((*p)->ping_chk() < 0){
			return -EINVAL;
		}
	}

	return OK;
}

int Connection::opt_set(string &opt, string &val)
{
	int pos;

	if ((pos=opt.find("link_check_frq")) >=0){
		link_chk_frq = atoi(val.c_str());
		if (link_chk_frq < 1 || link_chk_frq > MAX_LINK_CHECK_FRQ)
			return ERR;
			
		cout << prog_name() << "\tLink Check freq " <<  link_chk_frq << "\n";
		return OK;
	}
	if ((pos=opt.find("ping_check_frq")) >=0){
		ping_chk_frq = atoi(val.c_str());
		if (ping_chk_frq < 1 || ping_chk_frq > MAX_PING_CHECK_FRQ)
			return ERR;
			
		cout << prog_name() << "\tPing Check freq " <<  ping_chk_frq << "\n";
		return OK;
	}
	if ((pos=opt.find("delay_after_startup")) >=0){
		start_delay = atoi(val.c_str());
		if (start_delay < 0 || start_delay > MAX_START_DELAY)
			return ERR;
			
		cout << prog_name() << "\tDelay after startup " <<  start_delay << "\n";
		return OK;
	}
	if ((pos=opt.find("bod")) >=0){
		type=BOD;	
		cout << prog_name() << "\tBandwith on Demand" << "\n";
		return OK;
	}
	if ((pos=opt.find("loadbalance")) >=0){
		type=BALANCED;
		cout << prog_name() << "\tLoad balanced" << "\n";
		return OK;
	}
	if ((pos=opt.find("connection_up"))>=0){
		con_up_exec = val;
		cout << prog_name() << "\tConnection Link UP script = " << con_up_exec << "\n";
		return verify_exec(con_up_exec);
	}
	if ((pos=opt.find("connection_down"))>=0){
		con_down_exec = val;
		cout << prog_name() << "\tConnection Link Down script = " << con_down_exec << "\n";
		return verify_exec(con_down_exec);
	}
	
	return ERR;
}

void Connection::delay(int delay)
{
	struct timeval to;

	to.tv_sec=5;
	to.tv_usec = 0;
	
	switch (delay){

	case LINK_CHK_DELAY:
		to.tv_sec = link_chk_frq;
		break;
	case START_DELAY:
		to.tv_sec = start_delay;
		if (start_delay)
			cout << prog_name() << "Post Start Delay " << to.tv_sec << " sec ...\n";
		break;
	default:
		to.tv_sec = delay;
	}

	if (!to.tv_sec)
		return;

	#if _DEBUG_ > 2
	cout << prog_name() << "Delaying for " << to.tv_sec << " sec ...\n";
	#endif

	select(0,NULL,NULL,NULL,&to);
}

int Connection::link_down(void)
{
	if (state == prev_state)
		return OK;

	prev_state = state;
	enable_ping_chk=0;

	if (session == PRIMARY){
		cout << "\n" << prog_name() << "Primary Connection is DOWN !\n";
	}else{
		cout << "\n" << prog_name() << "Secondary Connection is DOWN !\n";
	}

	if (!con_down_exec.length())
		return OK;
		
	if (session == PRIMARY){
		cout << prog_name() << "Running Primary connection Link UP Script\n"; 
	}else{
		cout << prog_name() << "Running Secondary connection Link UP Script\n"; 
	}
	return exec_script(con_down_exec);
}

int Connection::link_up(void)
{

	if (state == prev_state)
		return OK;
		
	prev_state = state;
	enable_ping_chk=1;

	if (session == PRIMARY){
		cout << "\n" << prog_name() << "Primary Connection is UP !\n";
	}else{
		cout << "\n" << prog_name() << "Secondary Connection is UP !\n";
	}
	
	if (!con_up_exec.length())
		return OK;

	if (session == PRIMARY){
		cout << prog_name() << "Running Primary connection Link UP Script\n"; 
	}else{
		cout << prog_name() << "Running Secondary connection Link UP Script\n"; 
	}
	return exec_script(con_up_exec);
}

int Connection::start(int delay_time)
{
	vector<Device*>::iterator p;
	
	for (p = dev_list.begin(); p!=dev_list.end(); ++p)
		(*p)->start();

	delay(delay_time);

	return OK;
}

int Connection::stop(int delay_time)
{
	vector<Device*>::iterator p;
	
	for (p = dev_list.begin(); p!=dev_list.end(); ++p)
		(*p)->stop();

	delay(delay_time);

	return OK;
}


void sig_child(int signum)
{
	pid_t pid;
	while ((pid=waitpid(-1,NULL,WNOHANG)) > 0);
}

void sig_handler(int signum)
{
	cout << prog_name() << "Exit.\n";
	cout.flush();
	if (fout.good()){
		fout.close();
	}
	
	if (pri)
		delete pri;
	if (sec)
		delete sec;
	exit(0);
}

//*****************************************************************
// MAIN
//*****************************************************************

int main (int argc, char **argv)
{
	Connection *cur_conn=NULL;
	Device *dev;
	int dev_type;
	string opt,val;

	pri = new Connection();
	sec = new Connection();
	
	std::fstream in_file(CONFIG_FILE, ios::in);
	if (in_file.fail()){ 
		cout << prog_name() << "Error cannot open configuraiton file : " <<  CONFIG_FILE << endl;
		goto cleanup;
	}

	Options file_options;

	if (argc == 1){
		if (file_options.open_log_file() != OK){
			goto cleanup;
		}
		cout=fout;
	}else if (argc == 2){
		string arg = argv[1];
		if (arg.compare("help")==0){
			usage_long();
			exit(0);
		}
		if (arg.compare("debug")){
			usage();
			exit(0);
		}
	}else{
		usage();
		exit(1);
	}

	signal(SIGTERM,&sig_handler);
	signal(SIGINT,&sig_handler);
	signal(SIGCHLD,&sig_child);

	for (;;){
		cout.flush();
		switch ((dev_type=file_options.get_opt(in_file))){
		
		case PRIMARY:
			cur_conn = pri;
			cout << "\n" << prog_name() << "Parsing PRIMARY Connection ...\n\n";
			goto conn_options;

		case SECONDARY:
			cur_conn = sec;
			cout << "\n" << prog_name() << "Parsing SECONDARY Connection ...\n\n";

		conn_options:
			for (;;){
				if (cur_conn->get_opt(in_file,&opt,&val) == OK){
					if (cur_conn->opt_set(opt,val) != OK){
						cout << prog_name() << 
						"Error in config : " << opt << "=" << val << endl;
						goto cleanup;
					}
					continue;
				}
				break;
			}
			
			continue;
		
		case END:
			break;

		case ERROR:
			cout << "Error in config file : " 
				<<  CONFIG_FILE << endl;
			goto cleanup;

		default:

			if (cur_conn == NULL){
				cout << prog_name() 
				     << "Error no Primary or Seconadry devices found! " << endl
				     << prog_name()
				     << "Error in config file : " 
				     <<  CONFIG_FILE << endl;
				goto cleanup;
			}
		
			dev = new Device();
			
			if (dev->init_link_type(dev_type) != OK){
				cout << prog_name() << "Error Illegal Device " << dev_type << endl;
				goto cleanup;
			}

			for (;;){
				if (dev->get_opt(in_file,&opt,&val) == OK){
					if (dev->opt_set(opt,val) != OK){
						cout << "Error in config : " << opt << "=" << val << endl;
						goto cleanup;
					}
					continue;
				}
				break;
			}

			cur_conn->attach_device(dev);
			continue;

		}
		break;
	}

	//Start the primary connections
	//Implement a dealy to wait for connect.
	cout << endl << prog_name() << "Wpbackupd Monitor started on PRIMARY.\n";
	pri->start(START_DELAY);

	for (;;){
		cout.flush();
		if (pri->link_chk() != OK){
			cout << prog_name() << "Failed to check link state.\n";
			goto cleanup;
		}	
		
		if (pri->get_state() != CONNECTED){
	
			pri->link_down();
	
			if (sec_pid){
				while (waitpid(sec_pid,0,WNOHANG) > 0){
					cout << prog_name() << "Error, Secondary connection terminated!\n";
					sec->stop(0);
					goto cleanup;
				}
			}else{
				if ((sec_pid=fork())==0){
					session=SECONDARY;
					cout << endl << endl 
					<< prog_name() << "Starting Secondary Connection ...\n";
					signal(SIGCHLD,&sig_child);
				sec_conn:
					cout.flush();
					sec->stop(0);
					sec->start(START_DELAY);
					for (;;){
						if (sec->link_chk() != OK){
							cout << prog_name() 
							     << "Secondary Link Check Failed\n";
							break;
						}
						if (sec->get_state() != CONNECTED){
							sec->link_down();
							sec->stop(5);
							goto sec_conn;		
						}else{
							sec->link_up();
						}
						sec->delay(LINK_CHK_DELAY);
					}
					exit(0);
				}
			}
		}else{
			//This means that secondary connection is
			//active.
			if (sec_pid){
				cout << "\n" << prog_name() << "Stopping Secondary Connection ...\n";
				kill(sec_pid,SIGTERM);
				sec_pid=0;
				sec->stop(0);
			}
			
			pri->link_up();
		}
		pri->delay(LINK_CHK_DELAY);
	}
	
cleanup:

	cout << prog_name() << "Exit.\n";
	cout.flush();
	if (fout.good()){
		fout.close();
	}	
	delete pri;
	delete sec;
	exit(0);
}


string &prog_name (void)
{
	char pid_str[10];
	sprintf(pid_str,"%i",getpid());

	time_t t;
	time(&t);
	prog=ctime(&t);
	prog.erase(prog.length()-6,6);
	prog.erase(0,4);

	if (session == PRIMARY){
		prog += " "+program_name+" PRI [";
	}else{
		prog += " "+program_name+" SEC [";
	}
	prog += pid_str;
	prog += "]: ";

	return prog;
}

void usage_long (void)
{
	cout << 

"
Wanpipe Backupd Daemon
----------------------

Author: Nenad Coribc <ncorbic@sangoma.com>

Copyright:	(c) 2000-2001 Sangoma Technologies Inc.

Description: 

	WANPIPE Backup Daemon utility is used to monitor the 
	Primary (HIGH speed) connection.  If the Primary 
	connection goes down for any reason, the wpbackupd 
	will establish a secondary (LOW speed) connection.
	
	At the same time the primary link will continue to 
	be monitored by wpbackupd for the signs of life. 
	Once the priamary link comes up the secondary 
	connection will be terminated and the primary will be 
	re-established. 

Usage: 
	wpbackupd       : log all output in /var/log/wpbackupd
	wpbackupd debug : forces all output into STDOUT 
	wpbackupd help  : displays help screen.	

Configuration Files: 

	/etc/wanpipe/wpbackupd.conf

Configuration File Structure:
	[primary]
		...
	[primary_device1]
		...
	[primary_device2] 
		...

	[secondary]
		...
	[secondary_device1]
		...
	[secondary_device2]
		...

Connection types:
	primary: main high speed connection.
		 combination of any device types.
		 usually wanpipe T1 device.

	secondary: secondary backup link
		   combination of any device types.
		   usually dialup device.
		   
Device types:
	wanpipe: standard wanpipe protocols running
		 over a FT1 line. (PPP, FR, CHDLC...)
		 Refer to WanpipeConfigManual.(pdf/txt)
		 
	syncppp: MultiPort PPP protocol.
		 This is the kernel implementation of PPP
		 running over the WANPIPE card.
		 Refer to WanpipeConfigManual.(pdf/txt)
		 
	dialup:  For now this represents any NON WANPIPE device.
		 User must supply the link_chk() script
		 since the dameon doesn't know anything about
		 this device.

Note: Each connectinon must contain at least a single device.	

Refer to the /etc/wanpipe/wpbackupd.conf.sample.
		
"
<< endl;

}


void usage (void)
{
cout <<
"
Wanpipe Backupd Daemon
----------------------

Usage: 
	wpbackupd       : log all output in /var/log/wpbackupd.
	wpbackupd debug : forces all output into STDOUT. 
	wpbackupd help  : displays help screen.	

Configuration File: 
	/etc/wanpipe/wpbackupd.conf

Refer to the /etc/wanpipe/wpbackupd.conf.sample.
"
<<endl;


}
//End
